#define _CFFI_

/* We try to define Py_LIMITED_API before including Python.h.

   Mess: we can only define it if Py_DEBUG, Py_TRACE_REFS and
   Py_REF_DEBUG are not defined.  This is a best-effort approximation:
   we can learn about Py_DEBUG from pyconfig.h, but it is unclear if
   the same works for the other two macros.  Py_DEBUG implies them,
   but not the other way around.

   The implementation is messy (issue #350): on Windows, with _MSC_VER,
   we have to define Py_LIMITED_API even before including pyconfig.h.
   In that case, we guess what pyconfig.h will do to the macros above,
   and check our guess after the #include.

   Note that on Windows, with CPython 3.x, you need >= 3.5 and virtualenv
   version >= 16.0.0.  With older versions of either, you don't get a
   copy of PYTHON3.DLL in the virtualenv.  We can't check the version of
   CPython *before* we even include pyconfig.h.  ffi.set_source() puts
   a ``#define _CFFI_NO_LIMITED_API'' at the start of this file if it is
   running on Windows < 3.5, as an attempt at fixing it, but that's
   arguably wrong because it may not be the target version of Python.
   Still better than nothing I guess.  As another workaround, you can
   remove the definition of Py_LIMITED_API here.

   See also 'py_limited_api' in cffi/setuptools_ext.py.
*/
#if !defined(_CFFI_USE_EMBEDDING) && !defined(Py_LIMITED_API)
#  ifdef _MSC_VER
#    if !defined(_DEBUG) && !defined(Py_DEBUG) && !defined(Py_TRACE_REFS) && !defined(Py_REF_DEBUG) && !defined(_CFFI_NO_LIMITED_API)
#      define Py_LIMITED_API
#    endif
#    include <pyconfig.h>
     /* sanity-check: Py_LIMITED_API will cause crashes if any of these
        are also defined.  Normally, the Python file PC/pyconfig.h does not
        cause any of these to be defined, with the exception that _DEBUG
        causes Py_DEBUG.  Double-check that. */
#    ifdef Py_LIMITED_API
#      if defined(Py_DEBUG)
#        error "pyconfig.h unexpectedly defines Py_DEBUG, but Py_LIMITED_API is set"
#      endif
#      if defined(Py_TRACE_REFS)
#        error "pyconfig.h unexpectedly defines Py_TRACE_REFS, but Py_LIMITED_API is set"
#      endif
#      if defined(Py_REF_DEBUG)
#        error "pyconfig.h unexpectedly defines Py_REF_DEBUG, but Py_LIMITED_API is set"
#      endif
#    endif
#  else
#    include <pyconfig.h>
#    if !defined(Py_DEBUG) && !defined(Py_TRACE_REFS) && !defined(Py_REF_DEBUG) && !defined(_CFFI_NO_LIMITED_API)
#      define Py_LIMITED_API
#    endif
#  endif
#endif

#include <Python.h>
#ifdef __cplusplus
extern "C" {
#endif
#include <stddef.h>

/* This part is from file 'cffi/parse_c_type.h'.  It is copied at the
   beginning of C sources generated by CFFI's ffi.set_source(). */

typedef void *_cffi_opcode_t;

#define _CFFI_OP(opcode, arg)   (_cffi_opcode_t)(opcode | (((uintptr_t)(arg)) << 8))
#define _CFFI_GETOP(cffi_opcode)    ((unsigned char)(uintptr_t)cffi_opcode)
#define _CFFI_GETARG(cffi_opcode)   (((intptr_t)cffi_opcode) >> 8)

#define _CFFI_OP_PRIMITIVE       1
#define _CFFI_OP_POINTER         3
#define _CFFI_OP_ARRAY           5
#define _CFFI_OP_OPEN_ARRAY      7
#define _CFFI_OP_STRUCT_UNION    9
#define _CFFI_OP_ENUM           11
#define _CFFI_OP_FUNCTION       13
#define _CFFI_OP_FUNCTION_END   15
#define _CFFI_OP_NOOP           17
#define _CFFI_OP_BITFIELD       19
#define _CFFI_OP_TYPENAME       21
#define _CFFI_OP_CPYTHON_BLTN_V 23   // varargs
#define _CFFI_OP_CPYTHON_BLTN_N 25   // noargs
#define _CFFI_OP_CPYTHON_BLTN_O 27   // O  (i.e. a single arg)
#define _CFFI_OP_CONSTANT       29
#define _CFFI_OP_CONSTANT_INT   31
#define _CFFI_OP_GLOBAL_VAR     33
#define _CFFI_OP_DLOPEN_FUNC    35
#define _CFFI_OP_DLOPEN_CONST   37
#define _CFFI_OP_GLOBAL_VAR_F   39
#define _CFFI_OP_EXTERN_PYTHON  41

#define _CFFI_PRIM_VOID          0
#define _CFFI_PRIM_BOOL          1
#define _CFFI_PRIM_CHAR          2
#define _CFFI_PRIM_SCHAR         3
#define _CFFI_PRIM_UCHAR         4
#define _CFFI_PRIM_SHORT         5
#define _CFFI_PRIM_USHORT        6
#define _CFFI_PRIM_INT           7
#define _CFFI_PRIM_UINT          8
#define _CFFI_PRIM_LONG          9
#define _CFFI_PRIM_ULONG        10
#define _CFFI_PRIM_LONGLONG     11
#define _CFFI_PRIM_ULONGLONG    12
#define _CFFI_PRIM_FLOAT        13
#define _CFFI_PRIM_DOUBLE       14
#define _CFFI_PRIM_LONGDOUBLE   15

#define _CFFI_PRIM_WCHAR        16
#define _CFFI_PRIM_INT8         17
#define _CFFI_PRIM_UINT8        18
#define _CFFI_PRIM_INT16        19
#define _CFFI_PRIM_UINT16       20
#define _CFFI_PRIM_INT32        21
#define _CFFI_PRIM_UINT32       22
#define _CFFI_PRIM_INT64        23
#define _CFFI_PRIM_UINT64       24
#define _CFFI_PRIM_INTPTR       25
#define _CFFI_PRIM_UINTPTR      26
#define _CFFI_PRIM_PTRDIFF      27
#define _CFFI_PRIM_SIZE         28
#define _CFFI_PRIM_SSIZE        29
#define _CFFI_PRIM_INT_LEAST8   30
#define _CFFI_PRIM_UINT_LEAST8  31
#define _CFFI_PRIM_INT_LEAST16  32
#define _CFFI_PRIM_UINT_LEAST16 33
#define _CFFI_PRIM_INT_LEAST32  34
#define _CFFI_PRIM_UINT_LEAST32 35
#define _CFFI_PRIM_INT_LEAST64  36
#define _CFFI_PRIM_UINT_LEAST64 37
#define _CFFI_PRIM_INT_FAST8    38
#define _CFFI_PRIM_UINT_FAST8   39
#define _CFFI_PRIM_INT_FAST16   40
#define _CFFI_PRIM_UINT_FAST16  41
#define _CFFI_PRIM_INT_FAST32   42
#define _CFFI_PRIM_UINT_FAST32  43
#define _CFFI_PRIM_INT_FAST64   44
#define _CFFI_PRIM_UINT_FAST64  45
#define _CFFI_PRIM_INTMAX       46
#define _CFFI_PRIM_UINTMAX      47
#define _CFFI_PRIM_FLOATCOMPLEX 48
#define _CFFI_PRIM_DOUBLECOMPLEX 49
#define _CFFI_PRIM_CHAR16       50
#define _CFFI_PRIM_CHAR32       51

#define _CFFI__NUM_PRIM         52
#define _CFFI__UNKNOWN_PRIM           (-1)
#define _CFFI__UNKNOWN_FLOAT_PRIM     (-2)
#define _CFFI__UNKNOWN_LONG_DOUBLE    (-3)

#define _CFFI__IO_FILE_STRUCT         (-1)


struct _cffi_global_s {
    const char *name;
    void *address;
    _cffi_opcode_t type_op;
    void *size_or_direct_fn;  // OP_GLOBAL_VAR: size, or 0 if unknown
                              // OP_CPYTHON_BLTN_*: addr of direct function
};

struct _cffi_getconst_s {
    unsigned long long value;
    const struct _cffi_type_context_s *ctx;
    int gindex;
};

struct _cffi_struct_union_s {
    const char *name;
    int type_index;          // -> _cffi_types, on a OP_STRUCT_UNION
    int flags;               // _CFFI_F_* flags below
    size_t size;
    int alignment;
    int first_field_index;   // -> _cffi_fields array
    int num_fields;
};
#define _CFFI_F_UNION         0x01   // is a union, not a struct
#define _CFFI_F_CHECK_FIELDS  0x02   // complain if fields are not in the
                                     // "standard layout" or if some are missing
#define _CFFI_F_PACKED        0x04   // for CHECK_FIELDS, assume a packed struct
#define _CFFI_F_EXTERNAL      0x08   // in some other ffi.include()
#define _CFFI_F_OPAQUE        0x10   // opaque

struct _cffi_field_s {
    const char *name;
    size_t field_offset;
    size_t field_size;
    _cffi_opcode_t field_type_op;
};

struct _cffi_enum_s {
    const char *name;
    int type_index;          // -> _cffi_types, on a OP_ENUM
    int type_prim;           // _CFFI_PRIM_xxx
    const char *enumerators; // comma-delimited string
};

struct _cffi_typename_s {
    const char *name;
    int type_index;   /* if opaque, points to a possibly artificial
                         OP_STRUCT which is itself opaque */
};

struct _cffi_type_context_s {
    _cffi_opcode_t *types;
    const struct _cffi_global_s *globals;
    const struct _cffi_field_s *fields;
    const struct _cffi_struct_union_s *struct_unions;
    const struct _cffi_enum_s *enums;
    const struct _cffi_typename_s *typenames;
    int num_globals;
    int num_struct_unions;
    int num_enums;
    int num_typenames;
    const char *const *includes;
    int num_types;
    int flags;      /* future extension */
};

struct _cffi_parse_info_s {
    const struct _cffi_type_context_s *ctx;
    _cffi_opcode_t *output;
    unsigned int output_size;
    size_t error_location;
    const char *error_message;
};

struct _cffi_externpy_s {
    const char *name;
    size_t size_of_result;
    void *reserved1, *reserved2;
};

#ifdef _CFFI_INTERNAL
static int parse_c_type(struct _cffi_parse_info_s *info, const char *input);
static int search_in_globals(const struct _cffi_type_context_s *ctx,
                             const char *search, size_t search_len);
static int search_in_struct_unions(const struct _cffi_type_context_s *ctx,
                                   const char *search, size_t search_len);
#endif

/* this block of #ifs should be kept exactly identical between
   c/_cffi_backend.c, cffi/vengine_cpy.py, cffi/vengine_gen.py
   and cffi/_cffi_include.h */
#if defined(_MSC_VER)
# include <malloc.h>   /* for alloca() */
# if _MSC_VER < 1600   /* MSVC < 2010 */
   typedef __int8 int8_t;
   typedef __int16 int16_t;
   typedef __int32 int32_t;
   typedef __int64 int64_t;
   typedef unsigned __int8 uint8_t;
   typedef unsigned __int16 uint16_t;
   typedef unsigned __int32 uint32_t;
   typedef unsigned __int64 uint64_t;
   typedef __int8 int_least8_t;
   typedef __int16 int_least16_t;
   typedef __int32 int_least32_t;
   typedef __int64 int_least64_t;
   typedef unsigned __int8 uint_least8_t;
   typedef unsigned __int16 uint_least16_t;
   typedef unsigned __int32 uint_least32_t;
   typedef unsigned __int64 uint_least64_t;
   typedef __int8 int_fast8_t;
   typedef __int16 int_fast16_t;
   typedef __int32 int_fast32_t;
   typedef __int64 int_fast64_t;
   typedef unsigned __int8 uint_fast8_t;
   typedef unsigned __int16 uint_fast16_t;
   typedef unsigned __int32 uint_fast32_t;
   typedef unsigned __int64 uint_fast64_t;
   typedef __int64 intmax_t;
   typedef unsigned __int64 uintmax_t;
# else
#  include <stdint.h>
# endif
# if _MSC_VER < 1800   /* MSVC < 2013 */
#  ifndef __cplusplus
    typedef unsigned char _Bool;
#  endif
# endif
#else
# include <stdint.h>
# if (defined (__SVR4) && defined (__sun)) || defined(_AIX) || defined(__hpux)
#  include <alloca.h>
# endif
#endif

#ifdef __GNUC__
# define _CFFI_UNUSED_FN  __attribute__((unused))
#else
# define _CFFI_UNUSED_FN  /* nothing */
#endif

#ifdef __cplusplus
# ifndef _Bool
   typedef bool _Bool;   /* semi-hackish: C++ has no _Bool; bool is builtin */
# endif
#endif

/**********  CPython-specific section  **********/
#ifndef PYPY_VERSION


#if PY_MAJOR_VERSION >= 3
# define PyInt_FromLong PyLong_FromLong
#endif

#define _cffi_from_c_double PyFloat_FromDouble
#define _cffi_from_c_float PyFloat_FromDouble
#define _cffi_from_c_long PyInt_FromLong
#define _cffi_from_c_ulong PyLong_FromUnsignedLong
#define _cffi_from_c_longlong PyLong_FromLongLong
#define _cffi_from_c_ulonglong PyLong_FromUnsignedLongLong
#define _cffi_from_c__Bool PyBool_FromLong

#define _cffi_to_c_double PyFloat_AsDouble
#define _cffi_to_c_float PyFloat_AsDouble

#define _cffi_from_c_int(x, type)                                        \
    (((type)-1) > 0 ? /* unsigned */                                     \
        (sizeof(type) < sizeof(long) ?                                   \
            PyInt_FromLong((long)x) :                                    \
         sizeof(type) == sizeof(long) ?                                  \
            PyLong_FromUnsignedLong((unsigned long)x) :                  \
            PyLong_FromUnsignedLongLong((unsigned long long)x)) :        \
        (sizeof(type) <= sizeof(long) ?                                  \
            PyInt_FromLong((long)x) :                                    \
            PyLong_FromLongLong((long long)x)))

#define _cffi_to_c_int(o, type)                                          \
    ((type)(                                                             \
     sizeof(type) == 1 ? (((type)-1) > 0 ? (type)_cffi_to_c_u8(o)        \
                                         : (type)_cffi_to_c_i8(o)) :     \
     sizeof(type) == 2 ? (((type)-1) > 0 ? (type)_cffi_to_c_u16(o)       \
                                         : (type)_cffi_to_c_i16(o)) :    \
     sizeof(type) == 4 ? (((type)-1) > 0 ? (type)_cffi_to_c_u32(o)       \
                                         : (type)_cffi_to_c_i32(o)) :    \
     sizeof(type) == 8 ? (((type)-1) > 0 ? (type)_cffi_to_c_u64(o)       \
                                         : (type)_cffi_to_c_i64(o)) :    \
     (Py_FatalError("unsupported size for type " #type), (type)0)))

#define _cffi_to_c_i8                                                    \
                 ((int(*)(PyObject *))_cffi_exports[1])
#define _cffi_to_c_u8                                                    \
                 ((int(*)(PyObject *))_cffi_exports[2])
#define _cffi_to_c_i16                                                   \
                 ((int(*)(PyObject *))_cffi_exports[3])
#define _cffi_to_c_u16                                                   \
                 ((int(*)(PyObject *))_cffi_exports[4])
#define _cffi_to_c_i32                                                   \
                 ((int(*)(PyObject *))_cffi_exports[5])
#define _cffi_to_c_u32                                                   \
                 ((unsigned int(*)(PyObject *))_cffi_exports[6])
#define _cffi_to_c_i64                                                   \
                 ((long long(*)(PyObject *))_cffi_exports[7])
#define _cffi_to_c_u64                                                   \
                 ((unsigned long long(*)(PyObject *))_cffi_exports[8])
#define _cffi_to_c_char                                                  \
                 ((int(*)(PyObject *))_cffi_exports[9])
#define _cffi_from_c_pointer                                             \
    ((PyObject *(*)(char *, struct _cffi_ctypedescr *))_cffi_exports[10])
#define _cffi_to_c_pointer                                               \
    ((char *(*)(PyObject *, struct _cffi_ctypedescr *))_cffi_exports[11])
#define _cffi_get_struct_layout                                          \
    not used any more
#define _cffi_restore_errno                                              \
    ((void(*)(void))_cffi_exports[13])
#define _cffi_save_errno                                                 \
    ((void(*)(void))_cffi_exports[14])
#define _cffi_from_c_char                                                \
    ((PyObject *(*)(char))_cffi_exports[15])
#define _cffi_from_c_deref                                               \
    ((PyObject *(*)(char *, struct _cffi_ctypedescr *))_cffi_exports[16])
#define _cffi_to_c                                                       \
    ((int(*)(char *, struct _cffi_ctypedescr *, PyObject *))_cffi_exports[17])
#define _cffi_from_c_struct                                              \
    ((PyObject *(*)(char *, struct _cffi_ctypedescr *))_cffi_exports[18])
#define _cffi_to_c_wchar_t                                               \
    ((_cffi_wchar_t(*)(PyObject *))_cffi_exports[19])
#define _cffi_from_c_wchar_t                                             \
    ((PyObject *(*)(_cffi_wchar_t))_cffi_exports[20])
#define _cffi_to_c_long_double                                           \
    ((long double(*)(PyObject *))_cffi_exports[21])
#define _cffi_to_c__Bool                                                 \
    ((_Bool(*)(PyObject *))_cffi_exports[22])
#define _cffi_prepare_pointer_call_argument                              \
    ((Py_ssize_t(*)(struct _cffi_ctypedescr *,                           \
                    PyObject *, char **))_cffi_exports[23])
#define _cffi_convert_array_from_object                                  \
    ((int(*)(char *, struct _cffi_ctypedescr *, PyObject *))_cffi_exports[24])
#define _CFFI_CPIDX  25
#define _cffi_call_python                                                \
    ((void(*)(struct _cffi_externpy_s *, char *))_cffi_exports[_CFFI_CPIDX])
#define _cffi_to_c_wchar3216_t                                           \
    ((int(*)(PyObject *))_cffi_exports[26])
#define _cffi_from_c_wchar3216_t                                         \
    ((PyObject *(*)(int))_cffi_exports[27])
#define _CFFI_NUM_EXPORTS 28

struct _cffi_ctypedescr;

static void *_cffi_exports[_CFFI_NUM_EXPORTS];

#define _cffi_type(index)   (                           \
    assert((((uintptr_t)_cffi_types[index]) & 1) == 0), \
    (struct _cffi_ctypedescr *)_cffi_types[index])

static PyObject *_cffi_init(const char *module_name, Py_ssize_t version,
                            const struct _cffi_type_context_s *ctx)
{
    PyObject *module, *o_arg, *new_module;
    void *raw[] = {
        (void *)module_name,
        (void *)version,
        (void *)_cffi_exports,
        (void *)ctx,
    };

    module = PyImport_ImportModule("_cffi_backend");
    if (module == NULL)
        goto failure;

    o_arg = PyLong_FromVoidPtr((void *)raw);
    if (o_arg == NULL)
        goto failure;

    new_module = PyObject_CallMethod(
        module, (char *)"_init_cffi_1_0_external_module", (char *)"O", o_arg);

    Py_DECREF(o_arg);
    Py_DECREF(module);
    return new_module;

  failure:
    Py_XDECREF(module);
    return NULL;
}


#ifdef HAVE_WCHAR_H
typedef wchar_t _cffi_wchar_t;
#else
typedef uint16_t _cffi_wchar_t;   /* same random pick as _cffi_backend.c */
#endif

_CFFI_UNUSED_FN static uint16_t _cffi_to_c_char16_t(PyObject *o)
{
    if (sizeof(_cffi_wchar_t) == 2)
        return (uint16_t)_cffi_to_c_wchar_t(o);
    else
        return (uint16_t)_cffi_to_c_wchar3216_t(o);
}

_CFFI_UNUSED_FN static PyObject *_cffi_from_c_char16_t(uint16_t x)
{
    if (sizeof(_cffi_wchar_t) == 2)
        return _cffi_from_c_wchar_t((_cffi_wchar_t)x);
    else
        return _cffi_from_c_wchar3216_t((int)x);
}

_CFFI_UNUSED_FN static int _cffi_to_c_char32_t(PyObject *o)
{
    if (sizeof(_cffi_wchar_t) == 4)
        return (int)_cffi_to_c_wchar_t(o);
    else
        return (int)_cffi_to_c_wchar3216_t(o);
}

_CFFI_UNUSED_FN static PyObject *_cffi_from_c_char32_t(unsigned int x)
{
    if (sizeof(_cffi_wchar_t) == 4)
        return _cffi_from_c_wchar_t((_cffi_wchar_t)x);
    else
        return _cffi_from_c_wchar3216_t((int)x);
}

union _cffi_union_alignment_u {
    unsigned char m_char;
    unsigned short m_short;
    unsigned int m_int;
    unsigned long m_long;
    unsigned long long m_longlong;
    float m_float;
    double m_double;
    long double m_longdouble;
};

struct _cffi_freeme_s {
    struct _cffi_freeme_s *next;
    union _cffi_union_alignment_u alignment;
};

_CFFI_UNUSED_FN static int
_cffi_convert_array_argument(struct _cffi_ctypedescr *ctptr, PyObject *arg,
                             char **output_data, Py_ssize_t datasize,
                             struct _cffi_freeme_s **freeme)
{
    char *p;
    if (datasize < 0)
        return -1;

    p = *output_data;
    if (p == NULL) {
        struct _cffi_freeme_s *fp = (struct _cffi_freeme_s *)PyObject_Malloc(
            offsetof(struct _cffi_freeme_s, alignment) + (size_t)datasize);
        if (fp == NULL)
            return -1;
        fp->next = *freeme;
        *freeme = fp;
        p = *output_data = (char *)&fp->alignment;
    }
    memset((void *)p, 0, (size_t)datasize);
    return _cffi_convert_array_from_object(p, ctptr, arg);
}

_CFFI_UNUSED_FN static void
_cffi_free_array_arguments(struct _cffi_freeme_s *freeme)
{
    do {
        void *p = (void *)freeme;
        freeme = freeme->next;
        PyObject_Free(p);
    } while (freeme != NULL);
}

/**********  end CPython-specific section  **********/
#else
_CFFI_UNUSED_FN
static void (*_cffi_call_python_org)(struct _cffi_externpy_s *, char *);
# define _cffi_call_python  _cffi_call_python_org
#endif


#define _cffi_array_len(array)   (sizeof(array) / sizeof((array)[0]))

#define _cffi_prim_int(size, sign)                                      \
    ((size) == 1 ? ((sign) ? _CFFI_PRIM_INT8  : _CFFI_PRIM_UINT8)  :    \
     (size) == 2 ? ((sign) ? _CFFI_PRIM_INT16 : _CFFI_PRIM_UINT16) :    \
     (size) == 4 ? ((sign) ? _CFFI_PRIM_INT32 : _CFFI_PRIM_UINT32) :    \
     (size) == 8 ? ((sign) ? _CFFI_PRIM_INT64 : _CFFI_PRIM_UINT64) :    \
     _CFFI__UNKNOWN_PRIM)

#define _cffi_prim_float(size)                                          \
    ((size) == sizeof(float) ? _CFFI_PRIM_FLOAT :                       \
     (size) == sizeof(double) ? _CFFI_PRIM_DOUBLE :                     \
     (size) == sizeof(long double) ? _CFFI__UNKNOWN_LONG_DOUBLE :       \
     _CFFI__UNKNOWN_FLOAT_PRIM)

#define _cffi_check_int(got, got_nonpos, expected)      \
    ((got_nonpos) == (expected <= 0) &&                 \
     (got) == (unsigned long long)expected)

#ifdef MS_WIN32
# define _cffi_stdcall  __stdcall
#else
# define _cffi_stdcall  /* nothing */
#endif

#ifdef __cplusplus
}
#endif

/************************************************************/

/* This file is automatically generated */
#include "GraphBLAS.h"
char *GxB_IMPLEMENTATION_ABOUT_STR = GxB_IMPLEMENTATION_ABOUT;
char *GxB_IMPLEMENTATION_DATE_STR = GxB_IMPLEMENTATION_DATE;
char *GxB_IMPLEMENTATION_LICENSE_STR = GxB_IMPLEMENTATION_LICENSE;
char *GxB_IMPLEMENTATION_NAME_STR = GxB_IMPLEMENTATION_NAME;
char *GxB_SPEC_ABOUT_STR = GxB_SPEC_ABOUT;
char *GxB_SPEC_DATE_STR = GxB_SPEC_DATE;


/************************************************************/

static void *_cffi_types[] = {
/*  0 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(GrB_Desc_Value *, struct GB_Descriptor_opaque *, GrB_Desc_Field)
/*  1 */ _CFFI_OP(_CFFI_OP_POINTER, 476), // GrB_Desc_Value *
/*  2 */ _CFFI_OP(_CFFI_OP_POINTER, 3986), // struct GB_Descriptor_opaque *
/*  3 */ _CFFI_OP(_CFFI_OP_ENUM, 0), // GrB_Desc_Field
/*  4 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/*  5 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(GrB_Format *, struct GB_Matrix_opaque *)
/*  6 */ _CFFI_OP(_CFFI_OP_POINTER, 702), // GrB_Format *
/*  7 */ _CFFI_OP(_CFFI_OP_POINTER, 3989), // struct GB_Matrix_opaque *
/*  8 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/*  9 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(GrB_Mode)
/* 10 */ _CFFI_OP(_CFFI_OP_ENUM, 4), // GrB_Mode
/* 11 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 12 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(GrB_Mode, void *(*)(size_t), void *(*)(size_t, size_t), void *(*)(void *, size_t), void(*)(void *))
/* 13 */ _CFFI_OP(_CFFI_OP_NOOP, 10),
/* 14 */ _CFFI_OP(_CFFI_OP_POINTER, 3939), // void *(*)(size_t)
/* 15 */ _CFFI_OP(_CFFI_OP_POINTER, 3942), // void *(*)(size_t, size_t)
/* 16 */ _CFFI_OP(_CFFI_OP_POINTER, 3946), // void *(*)(void *, size_t)
/* 17 */ _CFFI_OP(_CFFI_OP_POINTER, 3959), // void(*)(void *)
/* 18 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 19 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(GxB_Option_Field, ...)
/* 20 */ _CFFI_OP(_CFFI_OP_ENUM, 7), // GxB_Option_Field
/* 21 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 1),
/* 22 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(GxB_Option_Field, char * *)
/* 23 */ _CFFI_OP(_CFFI_OP_NOOP, 20),
/* 24 */ _CFFI_OP(_CFFI_OP_POINTER, 97), // char * *
/* 25 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 26 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(GxB_Option_Field, double *)
/* 27 */ _CFFI_OP(_CFFI_OP_NOOP, 20),
/* 28 */ _CFFI_OP(_CFFI_OP_POINTER, 32), // double *
/* 29 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 30 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(GxB_Option_Field, double)
/* 31 */ _CFFI_OP(_CFFI_OP_NOOP, 20),
/* 32 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 14), // double
/* 33 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 34 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(GxB_Option_Field, int32_t *)
/* 35 */ _CFFI_OP(_CFFI_OP_NOOP, 20),
/* 36 */ _CFFI_OP(_CFFI_OP_POINTER, 40), // int32_t *
/* 37 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 38 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(GxB_Option_Field, int32_t)
/* 39 */ _CFFI_OP(_CFFI_OP_NOOP, 20),
/* 40 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 21), // int32_t
/* 41 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 42 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(GxB_Option_Field, int64_t *)
/* 43 */ _CFFI_OP(_CFFI_OP_NOOP, 20),
/* 44 */ _CFFI_OP(_CFFI_OP_POINTER, 899), // int64_t *
/* 45 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 46 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(GxB_Option_Field, void * *)
/* 47 */ _CFFI_OP(_CFFI_OP_NOOP, 20),
/* 48 */ _CFFI_OP(_CFFI_OP_POINTER, 52), // void * *
/* 49 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 50 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(GxB_Option_Field, void *)
/* 51 */ _CFFI_OP(_CFFI_OP_NOOP, 20),
/* 52 */ _CFFI_OP(_CFFI_OP_POINTER, 3998), // void *
/* 53 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 54 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(_Bool *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 55 */ _CFFI_OP(_CFFI_OP_POINTER, 566), // _Bool *
/* 56 */ _CFFI_OP(_CFFI_OP_POINTER, 3985), // struct GB_BinaryOp_opaque *
/* 57 */ _CFFI_OP(_CFFI_OP_POINTER, 3990), // struct GB_Monoid_opaque *
/* 58 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 59 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 60 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 61 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(_Bool *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 62 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 63 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 64 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 65 */ _CFFI_OP(_CFFI_OP_POINTER, 3996), // struct GB_Vector_opaque *
/* 66 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 67 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 68 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(_Bool *, struct GB_Matrix_opaque *)
/* 69 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 70 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 71 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 72 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(_Bool *, struct GB_Matrix_opaque *, uint64_t, uint64_t)
/* 73 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 74 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 75 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24), // uint64_t
/* 76 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 77 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 78 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(_Bool *, struct GB_Scalar_opaque *)
/* 79 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 80 */ _CFFI_OP(_CFFI_OP_POINTER, 3991), // struct GB_Scalar_opaque *
/* 81 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 82 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(_Bool *, struct GB_Vector_opaque *)
/* 83 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 84 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 85 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 86 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(_Bool *, struct GB_Vector_opaque *, uint64_t)
/* 87 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 88 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 89 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 90 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 91 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(_Bool *, void *, struct GB_Monoid_opaque *)
/* 92 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 93 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 94 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 95 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 96 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(char *, struct GB_BinaryOp_opaque *)
/* 97 */ _CFFI_OP(_CFFI_OP_POINTER, 3984), // char *
/* 98 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 99 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 100 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(char *, struct GB_IndexUnaryOp_opaque *)
/* 101 */ _CFFI_OP(_CFFI_OP_NOOP, 97),
/* 102 */ _CFFI_OP(_CFFI_OP_POINTER, 3987), // struct GB_IndexUnaryOp_opaque *
/* 103 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 104 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(char *, struct GB_Matrix_opaque *)
/* 105 */ _CFFI_OP(_CFFI_OP_NOOP, 97),
/* 106 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 107 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 108 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(char *, struct GB_Scalar_opaque *)
/* 109 */ _CFFI_OP(_CFFI_OP_NOOP, 97),
/* 110 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 111 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 112 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(char *, struct GB_Type_opaque *)
/* 113 */ _CFFI_OP(_CFFI_OP_NOOP, 97),
/* 114 */ _CFFI_OP(_CFFI_OP_POINTER, 3994), // struct GB_Type_opaque *
/* 115 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 116 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(char *, struct GB_UnaryOp_opaque *)
/* 117 */ _CFFI_OP(_CFFI_OP_NOOP, 97),
/* 118 */ _CFFI_OP(_CFFI_OP_POINTER, 3995), // struct GB_UnaryOp_opaque *
/* 119 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 120 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(char *, struct GB_Vector_opaque *)
/* 121 */ _CFFI_OP(_CFFI_OP_NOOP, 97),
/* 122 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 123 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 124 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(char *, void const *, uint64_t)
/* 125 */ _CFFI_OP(_CFFI_OP_NOOP, 97),
/* 126 */ _CFFI_OP(_CFFI_OP_POINTER, 3998), // void const *
/* 127 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 128 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 129 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(char const * *, struct GB_BinaryOp_opaque *)
/* 130 */ _CFFI_OP(_CFFI_OP_POINTER, 445), // char const * *
/* 131 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 132 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 133 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(char const * *, struct GB_Descriptor_opaque *)
/* 134 */ _CFFI_OP(_CFFI_OP_NOOP, 130),
/* 135 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 136 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 137 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(char const * *, struct GB_IndexUnaryOp_opaque *)
/* 138 */ _CFFI_OP(_CFFI_OP_NOOP, 130),
/* 139 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 140 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 141 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(char const * *, struct GB_Matrix_opaque *)
/* 142 */ _CFFI_OP(_CFFI_OP_NOOP, 130),
/* 143 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 144 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 145 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(char const * *, struct GB_Monoid_opaque *)
/* 146 */ _CFFI_OP(_CFFI_OP_NOOP, 130),
/* 147 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 148 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 149 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(char const * *, struct GB_Scalar_opaque *)
/* 150 */ _CFFI_OP(_CFFI_OP_NOOP, 130),
/* 151 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 152 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 153 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(char const * *, struct GB_SelectOp_opaque *)
/* 154 */ _CFFI_OP(_CFFI_OP_NOOP, 130),
/* 155 */ _CFFI_OP(_CFFI_OP_POINTER, 3992), // struct GB_SelectOp_opaque *
/* 156 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 157 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(char const * *, struct GB_Semiring_opaque *)
/* 158 */ _CFFI_OP(_CFFI_OP_NOOP, 130),
/* 159 */ _CFFI_OP(_CFFI_OP_POINTER, 3993), // struct GB_Semiring_opaque *
/* 160 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 161 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(char const * *, struct GB_Type_opaque *)
/* 162 */ _CFFI_OP(_CFFI_OP_NOOP, 130),
/* 163 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 164 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 165 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(char const * *, struct GB_UnaryOp_opaque *)
/* 166 */ _CFFI_OP(_CFFI_OP_NOOP, 130),
/* 167 */ _CFFI_OP(_CFFI_OP_NOOP, 118),
/* 168 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 169 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(char const * *, struct GB_Vector_opaque *)
/* 170 */ _CFFI_OP(_CFFI_OP_NOOP, 130),
/* 171 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 172 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 173 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(double *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 174 */ _CFFI_OP(_CFFI_OP_NOOP, 28),
/* 175 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 176 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 177 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 178 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 179 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 180 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(double *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 181 */ _CFFI_OP(_CFFI_OP_NOOP, 28),
/* 182 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 183 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 184 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 185 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 186 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 187 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(double *, struct GB_Matrix_opaque *, uint64_t, uint64_t)
/* 188 */ _CFFI_OP(_CFFI_OP_NOOP, 28),
/* 189 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 190 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 191 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 192 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 193 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(double *, struct GB_Scalar_opaque *)
/* 194 */ _CFFI_OP(_CFFI_OP_NOOP, 28),
/* 195 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 196 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 197 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(double *, struct GB_Vector_opaque *, uint64_t)
/* 198 */ _CFFI_OP(_CFFI_OP_NOOP, 28),
/* 199 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 200 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 201 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 202 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(_Dcomplex *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 203 */ _CFFI_OP(_CFFI_OP_POINTER, 962), // _Dcomplex *
/* 204 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 205 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 206 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 207 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 208 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 209 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(_Dcomplex *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 210 */ _CFFI_OP(_CFFI_OP_NOOP, 203),
/* 211 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 212 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 213 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 214 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 215 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 216 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(_Dcomplex *, struct GB_Matrix_opaque *, uint64_t, uint64_t)
/* 217 */ _CFFI_OP(_CFFI_OP_NOOP, 203),
/* 218 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 219 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 220 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 221 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 222 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(_Dcomplex *, struct GB_Scalar_opaque *)
/* 223 */ _CFFI_OP(_CFFI_OP_NOOP, 203),
/* 224 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 225 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 226 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(_Dcomplex *, struct GB_Vector_opaque *, uint64_t)
/* 227 */ _CFFI_OP(_CFFI_OP_NOOP, 203),
/* 228 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 229 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 230 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 231 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(float *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 232 */ _CFFI_OP(_CFFI_OP_POINTER, 980), // float *
/* 233 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 234 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 235 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 236 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 237 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 238 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(float *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 239 */ _CFFI_OP(_CFFI_OP_NOOP, 232),
/* 240 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 241 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 242 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 243 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 244 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 245 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(float *, struct GB_Matrix_opaque *, uint64_t, uint64_t)
/* 246 */ _CFFI_OP(_CFFI_OP_NOOP, 232),
/* 247 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 248 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 249 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 250 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 251 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(float *, struct GB_Scalar_opaque *)
/* 252 */ _CFFI_OP(_CFFI_OP_NOOP, 232),
/* 253 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 254 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 255 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(float *, struct GB_Vector_opaque *, uint64_t)
/* 256 */ _CFFI_OP(_CFFI_OP_NOOP, 232),
/* 257 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 258 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 259 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 260 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(_Fcomplex *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 261 */ _CFFI_OP(_CFFI_OP_POINTER, 974), // _Fcomplex *
/* 262 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 263 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 264 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 265 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 266 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 267 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(_Fcomplex *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 268 */ _CFFI_OP(_CFFI_OP_NOOP, 261),
/* 269 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 270 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 271 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 272 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 273 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 274 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(_Fcomplex *, struct GB_Matrix_opaque *, uint64_t, uint64_t)
/* 275 */ _CFFI_OP(_CFFI_OP_NOOP, 261),
/* 276 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 277 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 278 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 279 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 280 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(_Fcomplex *, struct GB_Scalar_opaque *)
/* 281 */ _CFFI_OP(_CFFI_OP_NOOP, 261),
/* 282 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 283 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 284 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(_Fcomplex *, struct GB_Vector_opaque *, uint64_t)
/* 285 */ _CFFI_OP(_CFFI_OP_NOOP, 261),
/* 286 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 287 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 288 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 289 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(int16_t *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 290 */ _CFFI_OP(_CFFI_OP_POINTER, 986), // int16_t *
/* 291 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 292 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 293 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 294 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 295 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 296 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(int16_t *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 297 */ _CFFI_OP(_CFFI_OP_NOOP, 290),
/* 298 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 299 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 300 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 301 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 302 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 303 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(int16_t *, struct GB_Matrix_opaque *, uint64_t, uint64_t)
/* 304 */ _CFFI_OP(_CFFI_OP_NOOP, 290),
/* 305 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 306 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 307 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 308 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 309 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(int16_t *, struct GB_Scalar_opaque *)
/* 310 */ _CFFI_OP(_CFFI_OP_NOOP, 290),
/* 311 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 312 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 313 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(int16_t *, struct GB_Vector_opaque *, uint64_t)
/* 314 */ _CFFI_OP(_CFFI_OP_NOOP, 290),
/* 315 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 316 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 317 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 318 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(int32_t *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 319 */ _CFFI_OP(_CFFI_OP_NOOP, 36),
/* 320 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 321 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 322 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 323 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 324 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 325 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(int32_t *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 326 */ _CFFI_OP(_CFFI_OP_NOOP, 36),
/* 327 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 328 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 329 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 330 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 331 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 332 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(int32_t *, struct GB_Matrix_opaque *, uint64_t, uint64_t)
/* 333 */ _CFFI_OP(_CFFI_OP_NOOP, 36),
/* 334 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 335 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 336 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 337 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 338 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(int32_t *, struct GB_Scalar_opaque *)
/* 339 */ _CFFI_OP(_CFFI_OP_NOOP, 36),
/* 340 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 341 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 342 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(int32_t *, struct GB_Vector_opaque *, uint64_t)
/* 343 */ _CFFI_OP(_CFFI_OP_NOOP, 36),
/* 344 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 345 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 346 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 347 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(int64_t *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 348 */ _CFFI_OP(_CFFI_OP_NOOP, 44),
/* 349 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 350 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 351 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 352 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 353 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 354 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(int64_t *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 355 */ _CFFI_OP(_CFFI_OP_NOOP, 44),
/* 356 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 357 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 358 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 359 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 360 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 361 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(int64_t *, struct GB_Matrix_opaque *, uint64_t, uint64_t)
/* 362 */ _CFFI_OP(_CFFI_OP_NOOP, 44),
/* 363 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 364 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 365 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 366 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 367 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(int64_t *, struct GB_Scalar_opaque *)
/* 368 */ _CFFI_OP(_CFFI_OP_NOOP, 44),
/* 369 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 370 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 371 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(int64_t *, struct GB_Vector_opaque *, uint64_t)
/* 372 */ _CFFI_OP(_CFFI_OP_NOOP, 44),
/* 373 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 374 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 375 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 376 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(int8_t *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 377 */ _CFFI_OP(_CFFI_OP_POINTER, 1024), // int8_t *
/* 378 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 379 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 380 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 381 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 382 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 383 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(int8_t *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 384 */ _CFFI_OP(_CFFI_OP_NOOP, 377),
/* 385 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 386 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 387 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 388 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 389 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 390 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(int8_t *, struct GB_Matrix_opaque *, uint64_t, uint64_t)
/* 391 */ _CFFI_OP(_CFFI_OP_NOOP, 377),
/* 392 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 393 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 394 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 395 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 396 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(int8_t *, struct GB_Scalar_opaque *)
/* 397 */ _CFFI_OP(_CFFI_OP_NOOP, 377),
/* 398 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 399 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 400 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(int8_t *, struct GB_Vector_opaque *, uint64_t)
/* 401 */ _CFFI_OP(_CFFI_OP_NOOP, 377),
/* 402 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 403 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 404 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 405 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(size_t *, struct GB_Matrix_opaque *)
/* 406 */ _CFFI_OP(_CFFI_OP_POINTER, 2280), // size_t *
/* 407 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 408 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 409 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(size_t *, struct GB_Scalar_opaque *)
/* 410 */ _CFFI_OP(_CFFI_OP_NOOP, 406),
/* 411 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 412 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 413 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(size_t *, struct GB_Type_opaque *)
/* 414 */ _CFFI_OP(_CFFI_OP_NOOP, 406),
/* 415 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 416 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 417 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(size_t *, struct GB_Vector_opaque *)
/* 418 */ _CFFI_OP(_CFFI_OP_NOOP, 406),
/* 419 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 420 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 421 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_BinaryOp_opaque * *)
/* 422 */ _CFFI_OP(_CFFI_OP_POINTER, 56), // struct GB_BinaryOp_opaque * *
/* 423 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 424 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_BinaryOp_opaque * *, struct GB_Monoid_opaque *)
/* 425 */ _CFFI_OP(_CFFI_OP_NOOP, 422),
/* 426 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 427 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 428 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_BinaryOp_opaque * *, struct GB_Semiring_opaque *)
/* 429 */ _CFFI_OP(_CFFI_OP_NOOP, 422),
/* 430 */ _CFFI_OP(_CFFI_OP_NOOP, 159),
/* 431 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 432 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_BinaryOp_opaque * *, void(*)(void *, void const *, void const *), struct GB_Type_opaque *, struct GB_Type_opaque *, struct GB_Type_opaque *)
/* 433 */ _CFFI_OP(_CFFI_OP_NOOP, 422),
/* 434 */ _CFFI_OP(_CFFI_OP_POINTER, 3973), // void(*)(void *, void const *, void const *)
/* 435 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 436 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 437 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 438 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 439 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_BinaryOp_opaque * *, void(*)(void *, void const *, void const *), struct GB_Type_opaque *, struct GB_Type_opaque *, struct GB_Type_opaque *, char const *)
/* 440 */ _CFFI_OP(_CFFI_OP_NOOP, 422),
/* 441 */ _CFFI_OP(_CFFI_OP_NOOP, 434),
/* 442 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 443 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 444 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 445 */ _CFFI_OP(_CFFI_OP_POINTER, 3984), // char const *
/* 446 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 447 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_BinaryOp_opaque * *, void(*)(void *, void const *, void const *), struct GB_Type_opaque *, struct GB_Type_opaque *, struct GB_Type_opaque *, char const *, char const *)
/* 448 */ _CFFI_OP(_CFFI_OP_NOOP, 422),
/* 449 */ _CFFI_OP(_CFFI_OP_NOOP, 434),
/* 450 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 451 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 452 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 453 */ _CFFI_OP(_CFFI_OP_NOOP, 445),
/* 454 */ _CFFI_OP(_CFFI_OP_NOOP, 445),
/* 455 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 456 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_BinaryOp_opaque *, GrB_WaitMode)
/* 457 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 458 */ _CFFI_OP(_CFFI_OP_ENUM, 5), // GrB_WaitMode
/* 459 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 460 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_BinaryOp_opaque *, char const *, GxB_Print_Level, FILE *)
/* 461 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 462 */ _CFFI_OP(_CFFI_OP_NOOP, 445),
/* 463 */ _CFFI_OP(_CFFI_OP_ENUM, 8), // GxB_Print_Level
/* 464 */ _CFFI_OP(_CFFI_OP_POINTER, 3978), // FILE *
/* 465 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 466 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Descriptor_opaque * *)
/* 467 */ _CFFI_OP(_CFFI_OP_POINTER, 2), // struct GB_Descriptor_opaque * *
/* 468 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 469 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Descriptor_opaque *, GrB_Desc_Field, ...)
/* 470 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 471 */ _CFFI_OP(_CFFI_OP_NOOP, 3),
/* 472 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 1),
/* 473 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Descriptor_opaque *, GrB_Desc_Field, GrB_Desc_Value)
/* 474 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 475 */ _CFFI_OP(_CFFI_OP_NOOP, 3),
/* 476 */ _CFFI_OP(_CFFI_OP_ENUM, 1), // GrB_Desc_Value
/* 477 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 478 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Descriptor_opaque *, GrB_Desc_Field, double *)
/* 479 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 480 */ _CFFI_OP(_CFFI_OP_NOOP, 3),
/* 481 */ _CFFI_OP(_CFFI_OP_NOOP, 28),
/* 482 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 483 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Descriptor_opaque *, GrB_Desc_Field, double)
/* 484 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 485 */ _CFFI_OP(_CFFI_OP_NOOP, 3),
/* 486 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 14),
/* 487 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 488 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Descriptor_opaque *, GrB_Desc_Field, int32_t *)
/* 489 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 490 */ _CFFI_OP(_CFFI_OP_NOOP, 3),
/* 491 */ _CFFI_OP(_CFFI_OP_NOOP, 36),
/* 492 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 493 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Descriptor_opaque *, GrB_Desc_Field, int32_t)
/* 494 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 495 */ _CFFI_OP(_CFFI_OP_NOOP, 3),
/* 496 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 21),
/* 497 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 498 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Descriptor_opaque *, GrB_WaitMode)
/* 499 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 500 */ _CFFI_OP(_CFFI_OP_NOOP, 458),
/* 501 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 502 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Descriptor_opaque *, char const *, GxB_Print_Level, FILE *)
/* 503 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 504 */ _CFFI_OP(_CFFI_OP_NOOP, 445),
/* 505 */ _CFFI_OP(_CFFI_OP_NOOP, 463),
/* 506 */ _CFFI_OP(_CFFI_OP_NOOP, 464),
/* 507 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 508 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_IndexUnaryOp_opaque * *)
/* 509 */ _CFFI_OP(_CFFI_OP_POINTER, 102), // struct GB_IndexUnaryOp_opaque * *
/* 510 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 511 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_IndexUnaryOp_opaque * *, void(*)(void *, void const *, uint64_t, uint64_t, void const *), struct GB_Type_opaque *, struct GB_Type_opaque *, struct GB_Type_opaque *)
/* 512 */ _CFFI_OP(_CFFI_OP_NOOP, 509),
/* 513 */ _CFFI_OP(_CFFI_OP_POINTER, 3966), // void(*)(void *, void const *, uint64_t, uint64_t, void const *)
/* 514 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 515 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 516 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 517 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 518 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_IndexUnaryOp_opaque * *, void(*)(void *, void const *, uint64_t, uint64_t, void const *), struct GB_Type_opaque *, struct GB_Type_opaque *, struct GB_Type_opaque *, char const *, char const *)
/* 519 */ _CFFI_OP(_CFFI_OP_NOOP, 509),
/* 520 */ _CFFI_OP(_CFFI_OP_NOOP, 513),
/* 521 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 522 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 523 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 524 */ _CFFI_OP(_CFFI_OP_NOOP, 445),
/* 525 */ _CFFI_OP(_CFFI_OP_NOOP, 445),
/* 526 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 527 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_IndexUnaryOp_opaque *, GrB_WaitMode)
/* 528 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 529 */ _CFFI_OP(_CFFI_OP_NOOP, 458),
/* 530 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 531 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_IndexUnaryOp_opaque *, char const *, GxB_Print_Level, FILE *)
/* 532 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 533 */ _CFFI_OP(_CFFI_OP_NOOP, 445),
/* 534 */ _CFFI_OP(_CFFI_OP_NOOP, 463),
/* 535 */ _CFFI_OP(_CFFI_OP_NOOP, 464),
/* 536 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 537 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Iterator_opaque * *)
/* 538 */ _CFFI_OP(_CFFI_OP_POINTER, 541), // struct GB_Iterator_opaque * *
/* 539 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 540 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Iterator_opaque *)
/* 541 */ _CFFI_OP(_CFFI_OP_POINTER, 3988), // struct GB_Iterator_opaque *
/* 542 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 543 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Iterator_opaque *, struct GB_Matrix_opaque *, GxB_Format_Value, struct GB_Descriptor_opaque *)
/* 544 */ _CFFI_OP(_CFFI_OP_NOOP, 541),
/* 545 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 546 */ _CFFI_OP(_CFFI_OP_ENUM, 6), // GxB_Format_Value
/* 547 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 548 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 549 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Iterator_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 550 */ _CFFI_OP(_CFFI_OP_NOOP, 541),
/* 551 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 552 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 553 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 554 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Iterator_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 555 */ _CFFI_OP(_CFFI_OP_NOOP, 541),
/* 556 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 557 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 558 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 559 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Iterator_opaque *, uint64_t)
/* 560 */ _CFFI_OP(_CFFI_OP_NOOP, 541),
/* 561 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 562 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 563 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Iterator_opaque *, uint64_t, _Bool)
/* 564 */ _CFFI_OP(_CFFI_OP_NOOP, 541),
/* 565 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 566 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1), // _Bool
/* 567 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 568 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *)
/* 569 */ _CFFI_OP(_CFFI_OP_POINTER, 7), // struct GB_Matrix_opaque * *
/* 570 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 571 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Matrix_opaque *)
/* 572 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 573 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 574 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 575 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Matrix_opaque *, _Bool, uint64_t, uint64_t, struct GB_Descriptor_opaque *)
/* 576 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 577 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 578 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 579 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 580 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 581 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 582 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 583 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque * *, uint64_t *, uint64_t *, int8_t * *, void * *, uint64_t *, uint64_t *, _Bool *, uint64_t *, struct GB_Descriptor_opaque *)
/* 584 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 585 */ _CFFI_OP(_CFFI_OP_POINTER, 114), // struct GB_Type_opaque * *
/* 586 */ _CFFI_OP(_CFFI_OP_POINTER, 75), // uint64_t *
/* 587 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 588 */ _CFFI_OP(_CFFI_OP_POINTER, 377), // int8_t * *
/* 589 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 590 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 591 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 592 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 593 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 594 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 595 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 596 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque * *, uint64_t *, uint64_t *, uint64_t * *, uint64_t * *, uint64_t * *, void * *, uint64_t *, uint64_t *, uint64_t *, uint64_t *, _Bool *, uint64_t *, _Bool *, struct GB_Descriptor_opaque *)
/* 597 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 598 */ _CFFI_OP(_CFFI_OP_NOOP, 585),
/* 599 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 600 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 601 */ _CFFI_OP(_CFFI_OP_POINTER, 586), // uint64_t * *
/* 602 */ _CFFI_OP(_CFFI_OP_NOOP, 601),
/* 603 */ _CFFI_OP(_CFFI_OP_NOOP, 601),
/* 604 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 605 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 606 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 607 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 608 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 609 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 610 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 611 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 612 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 613 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 614 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque * *, uint64_t *, uint64_t *, uint64_t * *, uint64_t * *, void * *, uint64_t *, uint64_t *, uint64_t *, _Bool *, _Bool *, struct GB_Descriptor_opaque *)
/* 615 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 616 */ _CFFI_OP(_CFFI_OP_NOOP, 585),
/* 617 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 618 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 619 */ _CFFI_OP(_CFFI_OP_NOOP, 601),
/* 620 */ _CFFI_OP(_CFFI_OP_NOOP, 601),
/* 621 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 622 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 623 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 624 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 625 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 626 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 627 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 628 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 629 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque * *, uint64_t *, uint64_t *, void * *, uint64_t *, _Bool *, struct GB_Descriptor_opaque *)
/* 630 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 631 */ _CFFI_OP(_CFFI_OP_NOOP, 585),
/* 632 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 633 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 634 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 635 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 636 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 637 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 638 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 639 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque *, uint64_t, uint64_t)
/* 640 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 641 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 642 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 643 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 644 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 645 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque *, uint64_t, uint64_t, int8_t * *, void * *, uint64_t, uint64_t, _Bool, uint64_t, struct GB_Descriptor_opaque *)
/* 646 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 647 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 648 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 649 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 650 */ _CFFI_OP(_CFFI_OP_NOOP, 588),
/* 651 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 652 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 653 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 654 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 655 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 656 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 657 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 658 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque *, uint64_t, uint64_t, uint64_t * *, uint64_t * *, uint64_t * *, void * *, uint64_t, uint64_t, uint64_t, uint64_t, _Bool, uint64_t, _Bool, struct GB_Descriptor_opaque *)
/* 659 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 660 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 661 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 662 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 663 */ _CFFI_OP(_CFFI_OP_NOOP, 601),
/* 664 */ _CFFI_OP(_CFFI_OP_NOOP, 601),
/* 665 */ _CFFI_OP(_CFFI_OP_NOOP, 601),
/* 666 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 667 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 668 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 669 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 670 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 671 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 672 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 673 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 674 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 675 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 676 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque *, uint64_t, uint64_t, uint64_t * *, uint64_t * *, void * *, uint64_t, uint64_t, uint64_t, _Bool, _Bool, struct GB_Descriptor_opaque *)
/* 677 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 678 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 679 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 680 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 681 */ _CFFI_OP(_CFFI_OP_NOOP, 601),
/* 682 */ _CFFI_OP(_CFFI_OP_NOOP, 601),
/* 683 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 684 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 685 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 686 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 687 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 688 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 689 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 690 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 691 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque *, uint64_t, uint64_t, uint64_t const *, uint64_t const *, _Bool const *, uint64_t, uint64_t, uint64_t, GrB_Format)
/* 692 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 693 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 694 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 695 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 696 */ _CFFI_OP(_CFFI_OP_POINTER, 75), // uint64_t const *
/* 697 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 698 */ _CFFI_OP(_CFFI_OP_POINTER, 566), // _Bool const *
/* 699 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 700 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 701 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 702 */ _CFFI_OP(_CFFI_OP_ENUM, 2), // GrB_Format
/* 703 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 704 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque *, uint64_t, uint64_t, uint64_t const *, uint64_t const *, _Dcomplex const *, uint64_t, uint64_t, uint64_t, GrB_Format)
/* 705 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 706 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 707 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 708 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 709 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 710 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 711 */ _CFFI_OP(_CFFI_OP_POINTER, 962), // _Dcomplex const *
/* 712 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 713 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 714 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 715 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 716 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 717 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque *, uint64_t, uint64_t, uint64_t const *, uint64_t const *, double const *, uint64_t, uint64_t, uint64_t, GrB_Format)
/* 718 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 719 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 720 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 721 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 722 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 723 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 724 */ _CFFI_OP(_CFFI_OP_POINTER, 32), // double const *
/* 725 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 726 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 727 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 728 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 729 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 730 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque *, uint64_t, uint64_t, uint64_t const *, uint64_t const *, _Fcomplex const *, uint64_t, uint64_t, uint64_t, GrB_Format)
/* 731 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 732 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 733 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 734 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 735 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 736 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 737 */ _CFFI_OP(_CFFI_OP_POINTER, 974), // _Fcomplex const *
/* 738 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 739 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 740 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 741 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 742 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 743 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque *, uint64_t, uint64_t, uint64_t const *, uint64_t const *, float const *, uint64_t, uint64_t, uint64_t, GrB_Format)
/* 744 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 745 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 746 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 747 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 748 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 749 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 750 */ _CFFI_OP(_CFFI_OP_POINTER, 980), // float const *
/* 751 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 752 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 753 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 754 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 755 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 756 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque *, uint64_t, uint64_t, uint64_t const *, uint64_t const *, int16_t const *, uint64_t, uint64_t, uint64_t, GrB_Format)
/* 757 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 758 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 759 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 760 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 761 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 762 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 763 */ _CFFI_OP(_CFFI_OP_POINTER, 986), // int16_t const *
/* 764 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 765 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 766 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 767 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 768 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 769 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque *, uint64_t, uint64_t, uint64_t const *, uint64_t const *, int32_t const *, uint64_t, uint64_t, uint64_t, GrB_Format)
/* 770 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 771 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 772 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 773 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 774 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 775 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 776 */ _CFFI_OP(_CFFI_OP_POINTER, 40), // int32_t const *
/* 777 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 778 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 779 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 780 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 781 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 782 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque *, uint64_t, uint64_t, uint64_t const *, uint64_t const *, int64_t const *, uint64_t, uint64_t, uint64_t, GrB_Format)
/* 783 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 784 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 785 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 786 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 787 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 788 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 789 */ _CFFI_OP(_CFFI_OP_POINTER, 899), // int64_t const *
/* 790 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 791 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 792 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 793 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 794 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 795 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque *, uint64_t, uint64_t, uint64_t const *, uint64_t const *, int8_t const *, uint64_t, uint64_t, uint64_t, GrB_Format)
/* 796 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 797 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 798 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 799 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 800 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 801 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 802 */ _CFFI_OP(_CFFI_OP_POINTER, 1024), // int8_t const *
/* 803 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 804 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 805 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 806 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 807 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 808 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque *, uint64_t, uint64_t, uint64_t const *, uint64_t const *, uint16_t const *, uint64_t, uint64_t, uint64_t, GrB_Format)
/* 809 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 810 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 811 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 812 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 813 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 814 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 815 */ _CFFI_OP(_CFFI_OP_POINTER, 1336), // uint16_t const *
/* 816 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 817 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 818 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 819 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 820 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 821 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque *, uint64_t, uint64_t, uint64_t const *, uint64_t const *, uint32_t const *, uint64_t, uint64_t, uint64_t, GrB_Format)
/* 822 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 823 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 824 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 825 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 826 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 827 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 828 */ _CFFI_OP(_CFFI_OP_POINTER, 1345), // uint32_t const *
/* 829 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 830 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 831 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 832 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 833 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 834 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque *, uint64_t, uint64_t, uint64_t const *, uint64_t const *, uint64_t const *, uint64_t, uint64_t, uint64_t, GrB_Format)
/* 835 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 836 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 837 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 838 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 839 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 840 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 841 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 842 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 843 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 844 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 845 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 846 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 847 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque *, uint64_t, uint64_t, uint64_t const *, uint64_t const *, uint8_t const *, uint64_t, uint64_t, uint64_t, GrB_Format)
/* 848 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 849 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 850 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 851 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 852 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 853 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 854 */ _CFFI_OP(_CFFI_OP_POINTER, 1363), // uint8_t const *
/* 855 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 856 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 857 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 858 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 859 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 860 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque *, uint64_t, uint64_t, uint64_t const *, uint64_t const *, void const *, uint64_t, uint64_t, uint64_t, GrB_Format)
/* 861 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 862 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 863 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 864 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 865 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 866 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 867 */ _CFFI_OP(_CFFI_OP_NOOP, 126),
/* 868 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 869 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 870 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 871 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 872 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 873 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque *, uint64_t, uint64_t, void * *, uint64_t, _Bool, struct GB_Descriptor_opaque *)
/* 874 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 875 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 876 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 877 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 878 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 879 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 880 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 881 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 882 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 883 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque *, void const *, uint64_t)
/* 884 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 885 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 886 */ _CFFI_OP(_CFFI_OP_NOOP, 126),
/* 887 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 888 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 889 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Type_opaque *, void const *, uint64_t, struct GB_Descriptor_opaque *)
/* 890 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 891 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 892 */ _CFFI_OP(_CFFI_OP_NOOP, 126),
/* 893 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 894 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 895 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 896 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, struct GB_Vector_opaque *, int64_t)
/* 897 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 898 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 899 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 23), // int64_t
/* 900 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 901 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque * *, uint64_t, uint64_t, uint64_t const *, uint64_t const *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 902 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 903 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 904 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 905 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 906 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 907 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 908 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 909 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 910 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *)
/* 911 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 912 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 913 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, GrB_WaitMode)
/* 914 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 915 */ _CFFI_OP(_CFFI_OP_NOOP, 458),
/* 916 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 917 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, GxB_Option_Field, ...)
/* 918 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 919 */ _CFFI_OP(_CFFI_OP_NOOP, 20),
/* 920 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 1),
/* 921 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, GxB_Option_Field, double *)
/* 922 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 923 */ _CFFI_OP(_CFFI_OP_NOOP, 20),
/* 924 */ _CFFI_OP(_CFFI_OP_NOOP, 28),
/* 925 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 926 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, GxB_Option_Field, double)
/* 927 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 928 */ _CFFI_OP(_CFFI_OP_NOOP, 20),
/* 929 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 14),
/* 930 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 931 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, GxB_Option_Field, int32_t *)
/* 932 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 933 */ _CFFI_OP(_CFFI_OP_NOOP, 20),
/* 934 */ _CFFI_OP(_CFFI_OP_NOOP, 36),
/* 935 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 936 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, GxB_Option_Field, int32_t)
/* 937 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 938 */ _CFFI_OP(_CFFI_OP_NOOP, 20),
/* 939 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 21),
/* 940 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 941 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, _Bool, uint64_t, uint64_t)
/* 942 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 943 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 944 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 945 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 946 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 947 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, _Bool, uint64_t, uint64_t, struct GB_Descriptor_opaque *)
/* 948 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 949 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 950 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 951 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 952 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 953 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 954 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, char const *, GxB_Print_Level, FILE *)
/* 955 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 956 */ _CFFI_OP(_CFFI_OP_NOOP, 445),
/* 957 */ _CFFI_OP(_CFFI_OP_NOOP, 463),
/* 958 */ _CFFI_OP(_CFFI_OP_NOOP, 464),
/* 959 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 960 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, _Dcomplex, uint64_t, uint64_t)
/* 961 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 962 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 49), // _Dcomplex
/* 963 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 964 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 965 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 966 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, double, uint64_t, uint64_t)
/* 967 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 968 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 14),
/* 969 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 970 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 971 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 972 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, _Fcomplex, uint64_t, uint64_t)
/* 973 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 974 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 48), // _Fcomplex
/* 975 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 976 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 977 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 978 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, float, uint64_t, uint64_t)
/* 979 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 980 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 13), // float
/* 981 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 982 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 983 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 984 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, int16_t, uint64_t, uint64_t)
/* 985 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 986 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 19), // int16_t
/* 987 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 988 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 989 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 990 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, int32_t, uint64_t, uint64_t)
/* 991 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 992 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 21),
/* 993 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 994 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 995 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 996 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, int64_t, uint64_t, uint64_t)
/* 997 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 998 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 23),
/* 999 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1000 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1001 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1002 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, int8_t * *, void * *, uint64_t *, uint64_t *, _Bool *, uint64_t *, struct GB_Descriptor_opaque *)
/* 1003 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1004 */ _CFFI_OP(_CFFI_OP_NOOP, 588),
/* 1005 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 1006 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 1007 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 1008 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 1009 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 1010 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1011 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1012 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, int8_t * *, void * *, uint64_t, uint64_t, _Bool, uint64_t, struct GB_Descriptor_opaque *)
/* 1013 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1014 */ _CFFI_OP(_CFFI_OP_NOOP, 588),
/* 1015 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 1016 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1017 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1018 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 1019 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1020 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1021 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1022 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, int8_t, uint64_t, uint64_t)
/* 1023 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1024 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 17), // int8_t
/* 1025 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1026 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1027 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1028 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque * *, struct GB_Descriptor_opaque *)
/* 1029 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1030 */ _CFFI_OP(_CFFI_OP_NOOP, 569),
/* 1031 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1032 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1033 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque * const *, uint64_t, uint64_t, struct GB_Descriptor_opaque *)
/* 1034 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1035 */ _CFFI_OP(_CFFI_OP_POINTER, 7), // struct GB_Matrix_opaque * const *
/* 1036 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1037 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1038 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1039 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1040 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, _Bool, uint64_t const *, uint64_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 1041 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1042 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1043 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1044 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 1045 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1046 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1047 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1048 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1049 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1050 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1051 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, _Dcomplex, uint64_t const *, uint64_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 1052 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1053 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1054 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1055 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 49),
/* 1056 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1057 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1058 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1059 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1060 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1061 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1062 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, double, uint64_t const *, uint64_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 1063 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1064 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1065 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1066 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 14),
/* 1067 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1068 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1069 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1070 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1071 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1072 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1073 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, _Fcomplex, uint64_t const *, uint64_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 1074 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1075 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1076 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1077 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 48),
/* 1078 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1079 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1080 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1081 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1082 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1083 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1084 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, float, uint64_t const *, uint64_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 1085 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1086 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1087 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1088 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 13),
/* 1089 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1090 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1091 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1092 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1093 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1094 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1095 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, int16_t, uint64_t const *, uint64_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 1096 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1097 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1098 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1099 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 19),
/* 1100 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1101 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1102 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1103 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1104 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1105 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1106 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, int32_t, uint64_t const *, uint64_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 1107 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1108 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1109 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1110 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 21),
/* 1111 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1112 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1113 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1114 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1115 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1116 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1117 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, int64_t, uint64_t const *, uint64_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 1118 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1119 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1120 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1121 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 23),
/* 1122 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1123 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1124 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1125 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1126 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1127 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1128 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, int8_t, uint64_t const *, uint64_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 1129 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1130 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1131 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1132 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 17),
/* 1133 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1134 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1135 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1136 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1137 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1138 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1139 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, _Bool, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 1140 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1141 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1142 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1143 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1144 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 1145 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1146 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1147 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1148 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, _Dcomplex, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 1149 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1150 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1151 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1152 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1153 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 49),
/* 1154 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1155 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1156 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1157 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, double, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 1158 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1159 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1160 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1161 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1162 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 14),
/* 1163 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1164 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1165 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1166 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, _Fcomplex, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 1167 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1168 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1169 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1170 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1171 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 48),
/* 1172 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1173 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1174 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1175 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, float, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 1176 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1177 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1178 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1179 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1180 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 13),
/* 1181 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1182 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1183 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1184 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, int16_t, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 1185 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1186 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1187 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1188 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1189 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 19),
/* 1190 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1191 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1192 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1193 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, int32_t, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 1194 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1195 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1196 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1197 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1198 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 21),
/* 1199 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1200 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1201 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1202 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, int64_t, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 1203 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1204 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1205 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1206 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1207 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 23),
/* 1208 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1209 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1210 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1211 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, int8_t, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 1212 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1213 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1214 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1215 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1216 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 17),
/* 1217 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1218 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1219 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1220 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Matrix_opaque *, _Bool, struct GB_Descriptor_opaque *)
/* 1221 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1222 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1223 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1224 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1225 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1226 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 1227 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1228 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1229 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Matrix_opaque *, _Dcomplex, struct GB_Descriptor_opaque *)
/* 1230 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1231 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1232 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1233 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1234 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1235 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 49),
/* 1236 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1237 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1238 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Matrix_opaque *, double, struct GB_Descriptor_opaque *)
/* 1239 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1240 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1241 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1242 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1243 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1244 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 14),
/* 1245 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1246 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1247 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Matrix_opaque *, _Fcomplex, struct GB_Descriptor_opaque *)
/* 1248 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1249 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1250 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1251 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1252 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1253 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 48),
/* 1254 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1255 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1256 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Matrix_opaque *, float, struct GB_Descriptor_opaque *)
/* 1257 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1258 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1259 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1260 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1261 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1262 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 13),
/* 1263 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1264 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1265 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Matrix_opaque *, int16_t, struct GB_Descriptor_opaque *)
/* 1266 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1267 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1268 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1269 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1270 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1271 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 19),
/* 1272 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1273 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1274 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Matrix_opaque *, int32_t, struct GB_Descriptor_opaque *)
/* 1275 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1276 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1277 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1278 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1279 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1280 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 21),
/* 1281 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1282 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1283 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Matrix_opaque *, int64_t, struct GB_Descriptor_opaque *)
/* 1284 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1285 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1286 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1287 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1288 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1289 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 23),
/* 1290 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1291 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1292 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Matrix_opaque *, int8_t, struct GB_Descriptor_opaque *)
/* 1293 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1294 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1295 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1296 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1297 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1298 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 17),
/* 1299 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1300 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1301 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 1302 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1303 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1304 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1305 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1306 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1307 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1308 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1309 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1310 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Matrix_opaque *, struct GB_Scalar_opaque *, struct GB_Descriptor_opaque *)
/* 1311 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1312 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1313 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1314 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1315 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1316 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 1317 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1318 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1319 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Matrix_opaque *, struct GB_Scalar_opaque *, struct GB_Matrix_opaque *, struct GB_Scalar_opaque *, struct GB_Descriptor_opaque *)
/* 1320 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1321 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1322 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1323 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1324 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1325 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 1326 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1327 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 1328 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1329 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1330 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Matrix_opaque *, uint16_t, struct GB_Descriptor_opaque *)
/* 1331 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1332 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1333 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1334 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1335 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1336 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 20), // uint16_t
/* 1337 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1338 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1339 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Matrix_opaque *, uint32_t, struct GB_Descriptor_opaque *)
/* 1340 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1341 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1342 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1343 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1344 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1345 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22), // uint32_t
/* 1346 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1347 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1348 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Matrix_opaque *, uint64_t, struct GB_Descriptor_opaque *)
/* 1349 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1350 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1351 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1352 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1353 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1354 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1355 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1356 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1357 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Matrix_opaque *, uint8_t, struct GB_Descriptor_opaque *)
/* 1358 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1359 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1360 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1361 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1362 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1363 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 18), // uint8_t
/* 1364 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1365 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1366 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Matrix_opaque *, void const *, struct GB_Descriptor_opaque *)
/* 1367 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1368 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1369 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1370 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1371 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1372 */ _CFFI_OP(_CFFI_OP_NOOP, 126),
/* 1373 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1374 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1375 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Scalar_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 1376 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1377 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1378 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1379 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1380 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 1381 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1382 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1383 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1384 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, uint16_t, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 1385 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1386 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1387 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1388 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1389 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 20),
/* 1390 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1391 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1392 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1393 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, uint32_t, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 1394 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1395 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1396 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1397 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1398 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 1399 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1400 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1401 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1402 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, uint64_t, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 1403 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1404 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1405 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1406 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1407 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1408 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1409 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1410 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1411 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, uint8_t, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 1412 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1413 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1414 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1415 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1416 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 18),
/* 1417 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1418 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1419 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1420 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, void const *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 1421 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1422 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1423 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1424 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1425 */ _CFFI_OP(_CFFI_OP_NOOP, 126),
/* 1426 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1427 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1428 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1429 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Matrix_opaque *, _Bool, struct GB_Descriptor_opaque *)
/* 1430 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1431 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1432 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1433 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 1434 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1435 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 1436 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1437 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1438 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Matrix_opaque *, _Dcomplex, struct GB_Descriptor_opaque *)
/* 1439 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1440 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1441 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1442 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 1443 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1444 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 49),
/* 1445 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1446 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1447 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Matrix_opaque *, double, struct GB_Descriptor_opaque *)
/* 1448 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1449 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1450 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1451 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 1452 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1453 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 14),
/* 1454 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1455 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1456 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Matrix_opaque *, _Fcomplex, struct GB_Descriptor_opaque *)
/* 1457 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1458 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1459 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1460 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 1461 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1462 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 48),
/* 1463 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1464 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1465 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Matrix_opaque *, float, struct GB_Descriptor_opaque *)
/* 1466 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1467 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1468 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1469 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 1470 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1471 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 13),
/* 1472 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1473 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1474 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Matrix_opaque *, int16_t, struct GB_Descriptor_opaque *)
/* 1475 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1476 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1477 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1478 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 1479 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1480 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 19),
/* 1481 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1482 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1483 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Matrix_opaque *, int32_t, struct GB_Descriptor_opaque *)
/* 1484 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1485 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1486 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1487 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 1488 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1489 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 21),
/* 1490 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1491 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1492 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Matrix_opaque *, int64_t, struct GB_Descriptor_opaque *)
/* 1493 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1494 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1495 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1496 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 1497 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1498 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 23),
/* 1499 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1500 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1501 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Matrix_opaque *, int8_t, struct GB_Descriptor_opaque *)
/* 1502 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1503 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1504 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1505 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 1506 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1507 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 17),
/* 1508 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1509 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1510 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Matrix_opaque *, struct GB_Scalar_opaque *, struct GB_Descriptor_opaque *)
/* 1511 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1512 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1513 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1514 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 1515 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1516 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 1517 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1518 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1519 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Matrix_opaque *, uint16_t, struct GB_Descriptor_opaque *)
/* 1520 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1521 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1522 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1523 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 1524 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1525 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 20),
/* 1526 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1527 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1528 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Matrix_opaque *, uint32_t, struct GB_Descriptor_opaque *)
/* 1529 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1530 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1531 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1532 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 1533 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1534 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 1535 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1536 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1537 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Matrix_opaque *, uint64_t, struct GB_Descriptor_opaque *)
/* 1538 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1539 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1540 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1541 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 1542 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1543 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1544 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1545 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1546 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Matrix_opaque *, uint8_t, struct GB_Descriptor_opaque *)
/* 1547 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1548 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1549 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1550 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 1551 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1552 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 18),
/* 1553 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1554 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1555 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Matrix_opaque *, void const *, struct GB_Descriptor_opaque *)
/* 1556 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1557 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1558 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1559 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 1560 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1561 */ _CFFI_OP(_CFFI_OP_NOOP, 126),
/* 1562 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1563 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1564 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 1565 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1566 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1567 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1568 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1569 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1570 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1571 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_Matrix_opaque *, uint64_t const *, uint64_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 1572 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1573 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1574 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1575 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1576 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1577 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1578 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1579 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1580 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1581 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1582 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 1583 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1584 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1585 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1586 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 1587 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1588 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1589 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1590 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1591 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_Scalar_opaque *, uint64_t const *, uint64_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 1592 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1593 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1594 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1595 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 1596 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1597 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1598 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1599 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1600 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1601 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1602 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_SelectOp_opaque *, struct GB_Matrix_opaque *, struct GB_Scalar_opaque *, struct GB_Descriptor_opaque *)
/* 1603 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1604 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1605 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1606 */ _CFFI_OP(_CFFI_OP_NOOP, 155),
/* 1607 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1608 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 1609 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1610 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1611 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_Semiring_opaque *, struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 1612 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1613 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1614 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1615 */ _CFFI_OP(_CFFI_OP_NOOP, 159),
/* 1616 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1617 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1618 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1619 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1620 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, struct GB_UnaryOp_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 1621 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1622 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1623 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1624 */ _CFFI_OP(_CFFI_OP_NOOP, 118),
/* 1625 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1626 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1627 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1628 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, uint16_t, uint64_t const *, uint64_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 1629 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1630 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1631 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1632 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 20),
/* 1633 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1634 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1635 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1636 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1637 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1638 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1639 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, uint32_t, uint64_t const *, uint64_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 1640 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1641 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1642 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1643 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 1644 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1645 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1646 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1647 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1648 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1649 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1650 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, uint64_t, uint64_t const *, uint64_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 1651 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1652 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1653 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1654 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1655 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1656 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1657 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1658 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1659 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1660 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1661 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, uint8_t, uint64_t const *, uint64_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 1662 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1663 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1664 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1665 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 18),
/* 1666 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1667 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1668 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1669 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1670 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1671 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1672 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Matrix_opaque *, struct GB_BinaryOp_opaque *, void *, uint64_t const *, uint64_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 1673 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1674 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1675 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1676 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 1677 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1678 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1679 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1680 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1681 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1682 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1683 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Scalar_opaque *, uint64_t, uint64_t)
/* 1684 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1685 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 1686 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1687 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1688 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1689 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Vector_opaque *, int64_t, struct GB_Descriptor_opaque *)
/* 1690 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1691 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 1692 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 23),
/* 1693 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1694 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1695 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_Vector_opaque *, uint64_t const *, uint64_t, uint64_t, struct GB_Descriptor_opaque *)
/* 1696 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1697 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 1698 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1699 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 1700 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1701 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1702 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1703 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1704 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1705 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_Vector_opaque *, uint64_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 1706 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1707 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 1708 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1709 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 1710 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1711 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1712 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1713 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1714 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1715 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint16_t, uint64_t, uint64_t)
/* 1716 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1717 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 20),
/* 1718 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1719 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1720 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1721 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint32_t, uint64_t, uint64_t)
/* 1722 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1723 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 1724 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1725 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1726 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1727 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint64_t * *, uint64_t * *, uint64_t * *, void * *, uint64_t *, uint64_t *, uint64_t *, uint64_t *, _Bool *, uint64_t *, _Bool *, struct GB_Descriptor_opaque *)
/* 1728 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1729 */ _CFFI_OP(_CFFI_OP_NOOP, 601),
/* 1730 */ _CFFI_OP(_CFFI_OP_NOOP, 601),
/* 1731 */ _CFFI_OP(_CFFI_OP_NOOP, 601),
/* 1732 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 1733 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 1734 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 1735 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 1736 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 1737 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 1738 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 1739 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 1740 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1741 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1742 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint64_t * *, uint64_t * *, uint64_t * *, void * *, uint64_t, uint64_t, uint64_t, uint64_t, _Bool, uint64_t, _Bool, struct GB_Descriptor_opaque *)
/* 1743 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1744 */ _CFFI_OP(_CFFI_OP_NOOP, 601),
/* 1745 */ _CFFI_OP(_CFFI_OP_NOOP, 601),
/* 1746 */ _CFFI_OP(_CFFI_OP_NOOP, 601),
/* 1747 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 1748 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1749 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1750 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1751 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1752 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 1753 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1754 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 1755 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1756 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1757 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint64_t * *, uint64_t * *, void * *, uint64_t *, uint64_t *, uint64_t *, _Bool *, _Bool *, struct GB_Descriptor_opaque *)
/* 1758 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1759 */ _CFFI_OP(_CFFI_OP_NOOP, 601),
/* 1760 */ _CFFI_OP(_CFFI_OP_NOOP, 601),
/* 1761 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 1762 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 1763 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 1764 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 1765 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 1766 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 1767 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1768 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1769 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint64_t * *, uint64_t * *, void * *, uint64_t, uint64_t, uint64_t, _Bool, _Bool, struct GB_Descriptor_opaque *)
/* 1770 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1771 */ _CFFI_OP(_CFFI_OP_NOOP, 601),
/* 1772 */ _CFFI_OP(_CFFI_OP_NOOP, 601),
/* 1773 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 1774 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1775 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1776 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1777 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 1778 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 1779 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1780 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1781 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint64_t const *, uint64_t const *, _Bool const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 1782 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1783 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1784 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1785 */ _CFFI_OP(_CFFI_OP_NOOP, 698),
/* 1786 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1787 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1788 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1789 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint64_t const *, uint64_t const *, _Dcomplex const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 1790 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1791 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1792 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1793 */ _CFFI_OP(_CFFI_OP_NOOP, 711),
/* 1794 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1795 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1796 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1797 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint64_t const *, uint64_t const *, double const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 1798 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1799 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1800 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1801 */ _CFFI_OP(_CFFI_OP_NOOP, 724),
/* 1802 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1803 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1804 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1805 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint64_t const *, uint64_t const *, _Fcomplex const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 1806 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1807 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1808 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1809 */ _CFFI_OP(_CFFI_OP_NOOP, 737),
/* 1810 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1811 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1812 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1813 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint64_t const *, uint64_t const *, float const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 1814 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1815 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1816 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1817 */ _CFFI_OP(_CFFI_OP_NOOP, 750),
/* 1818 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1819 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1820 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1821 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint64_t const *, uint64_t const *, int16_t const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 1822 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1823 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1824 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1825 */ _CFFI_OP(_CFFI_OP_NOOP, 763),
/* 1826 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1827 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1828 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1829 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint64_t const *, uint64_t const *, int32_t const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 1830 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1831 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1832 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1833 */ _CFFI_OP(_CFFI_OP_NOOP, 776),
/* 1834 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1835 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1836 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1837 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint64_t const *, uint64_t const *, int64_t const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 1838 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1839 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1840 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1841 */ _CFFI_OP(_CFFI_OP_NOOP, 789),
/* 1842 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1843 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1844 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1845 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint64_t const *, uint64_t const *, int8_t const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 1846 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1847 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1848 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1849 */ _CFFI_OP(_CFFI_OP_NOOP, 802),
/* 1850 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1851 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1852 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1853 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint64_t const *, uint64_t const *, struct GB_Scalar_opaque *, uint64_t)
/* 1854 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1855 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1856 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1857 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 1858 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1859 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1860 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint64_t const *, uint64_t const *, uint16_t const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 1861 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1862 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1863 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1864 */ _CFFI_OP(_CFFI_OP_NOOP, 815),
/* 1865 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1866 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1867 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1868 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint64_t const *, uint64_t const *, uint32_t const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 1869 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1870 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1871 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1872 */ _CFFI_OP(_CFFI_OP_NOOP, 828),
/* 1873 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1874 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1875 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1876 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint64_t const *, uint64_t const *, uint64_t const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 1877 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1878 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1879 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1880 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1881 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1882 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1883 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1884 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint64_t const *, uint64_t const *, uint8_t const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 1885 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1886 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1887 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1888 */ _CFFI_OP(_CFFI_OP_NOOP, 854),
/* 1889 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1890 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1891 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1892 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint64_t const *, uint64_t const *, void const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 1893 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1894 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1895 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 1896 */ _CFFI_OP(_CFFI_OP_NOOP, 126),
/* 1897 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1898 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1899 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1900 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint64_t, uint64_t)
/* 1901 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1902 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1903 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1904 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1905 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint64_t, uint64_t, uint64_t)
/* 1906 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1907 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1908 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1909 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1910 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1911 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, uint8_t, uint64_t, uint64_t)
/* 1912 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1913 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 18),
/* 1914 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1915 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1916 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1917 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, void * *, uint64_t *, _Bool *, struct GB_Descriptor_opaque *)
/* 1918 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1919 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 1920 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 1921 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 1922 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1923 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1924 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, void * *, uint64_t, _Bool, struct GB_Descriptor_opaque *)
/* 1925 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1926 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 1927 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1928 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 1929 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 1930 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1931 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Matrix_opaque *, void *, uint64_t, uint64_t)
/* 1932 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 1933 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 1934 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1935 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 1936 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1937 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *)
/* 1938 */ _CFFI_OP(_CFFI_OP_POINTER, 57), // struct GB_Monoid_opaque * *
/* 1939 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1940 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, _Bool)
/* 1941 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 1942 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1943 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 1944 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1945 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, _Bool, _Bool)
/* 1946 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 1947 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1948 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 1949 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 1950 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1951 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, _Dcomplex)
/* 1952 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 1953 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1954 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 49),
/* 1955 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1956 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, _Dcomplex, _Dcomplex)
/* 1957 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 1958 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1959 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 49),
/* 1960 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 49),
/* 1961 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1962 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, double)
/* 1963 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 1964 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1965 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 14),
/* 1966 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1967 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, double, double)
/* 1968 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 1969 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1970 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 14),
/* 1971 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 14),
/* 1972 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1973 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, _Fcomplex)
/* 1974 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 1975 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1976 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 48),
/* 1977 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1978 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, _Fcomplex, _Fcomplex)
/* 1979 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 1980 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1981 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 48),
/* 1982 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 48),
/* 1983 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1984 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, float)
/* 1985 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 1986 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1987 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 13),
/* 1988 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1989 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, float, float)
/* 1990 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 1991 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1992 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 13),
/* 1993 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 13),
/* 1994 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 1995 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, int16_t)
/* 1996 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 1997 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 1998 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 19),
/* 1999 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2000 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, int16_t, int16_t)
/* 2001 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 2002 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2003 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 19),
/* 2004 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 19),
/* 2005 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2006 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, int32_t)
/* 2007 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 2008 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2009 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 21),
/* 2010 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2011 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, int32_t, int32_t)
/* 2012 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 2013 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2014 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 21),
/* 2015 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 21),
/* 2016 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2017 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, int64_t)
/* 2018 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 2019 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2020 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 23),
/* 2021 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2022 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, int64_t, int64_t)
/* 2023 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 2024 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2025 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 23),
/* 2026 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 23),
/* 2027 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2028 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, int8_t)
/* 2029 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 2030 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2031 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 17),
/* 2032 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2033 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, int8_t, int8_t)
/* 2034 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 2035 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2036 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 17),
/* 2037 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 17),
/* 2038 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2039 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, uint16_t)
/* 2040 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 2041 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2042 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 20),
/* 2043 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2044 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, uint16_t, uint16_t)
/* 2045 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 2046 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2047 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 20),
/* 2048 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 20),
/* 2049 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2050 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, uint32_t)
/* 2051 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 2052 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2053 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 2054 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2055 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, uint32_t, uint32_t)
/* 2056 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 2057 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2058 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 2059 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 2060 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2061 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, uint64_t)
/* 2062 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 2063 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2064 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2065 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2066 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, uint64_t, uint64_t)
/* 2067 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 2068 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2069 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2070 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2071 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2072 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, uint8_t)
/* 2073 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 2074 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2075 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 18),
/* 2076 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2077 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, uint8_t, uint8_t)
/* 2078 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 2079 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2080 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 18),
/* 2081 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 18),
/* 2082 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2083 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, void *)
/* 2084 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 2085 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2086 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 2087 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2088 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_BinaryOp_opaque *, void *, void *)
/* 2089 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 2090 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2091 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 2092 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 2093 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2094 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque * *, struct GB_Semiring_opaque *)
/* 2095 */ _CFFI_OP(_CFFI_OP_NOOP, 1938),
/* 2096 */ _CFFI_OP(_CFFI_OP_NOOP, 159),
/* 2097 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2098 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque *, GrB_WaitMode)
/* 2099 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 2100 */ _CFFI_OP(_CFFI_OP_NOOP, 458),
/* 2101 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2102 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Monoid_opaque *, char const *, GxB_Print_Level, FILE *)
/* 2103 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 2104 */ _CFFI_OP(_CFFI_OP_NOOP, 445),
/* 2105 */ _CFFI_OP(_CFFI_OP_NOOP, 463),
/* 2106 */ _CFFI_OP(_CFFI_OP_NOOP, 464),
/* 2107 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2108 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque * *)
/* 2109 */ _CFFI_OP(_CFFI_OP_POINTER, 80), // struct GB_Scalar_opaque * *
/* 2110 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2111 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque * *, struct GB_Scalar_opaque *)
/* 2112 */ _CFFI_OP(_CFFI_OP_NOOP, 2109),
/* 2113 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2114 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2115 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque * *, struct GB_Type_opaque *)
/* 2116 */ _CFFI_OP(_CFFI_OP_NOOP, 2109),
/* 2117 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 2118 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2119 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque *)
/* 2120 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2121 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2122 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque *, GrB_WaitMode)
/* 2123 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2124 */ _CFFI_OP(_CFFI_OP_NOOP, 458),
/* 2125 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2126 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque *, _Bool)
/* 2127 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2128 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 2129 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2130 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque *, char const *, GxB_Print_Level, FILE *)
/* 2131 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2132 */ _CFFI_OP(_CFFI_OP_NOOP, 445),
/* 2133 */ _CFFI_OP(_CFFI_OP_NOOP, 463),
/* 2134 */ _CFFI_OP(_CFFI_OP_NOOP, 464),
/* 2135 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2136 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque *, _Dcomplex)
/* 2137 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2138 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 49),
/* 2139 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2140 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque *, double)
/* 2141 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2142 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 14),
/* 2143 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2144 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque *, _Fcomplex)
/* 2145 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2146 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 48),
/* 2147 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2148 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque *, float)
/* 2149 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2150 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 13),
/* 2151 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2152 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque *, int16_t)
/* 2153 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2154 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 19),
/* 2155 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2156 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque *, int32_t)
/* 2157 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2158 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 21),
/* 2159 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2160 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque *, int64_t)
/* 2161 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2162 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 23),
/* 2163 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2164 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque *, int8_t)
/* 2165 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2166 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 17),
/* 2167 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2168 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 2169 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2170 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2171 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2172 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 2173 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2174 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2175 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 2176 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2177 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2178 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2179 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2180 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2181 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2182 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 2183 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2184 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2185 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 2186 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 2187 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2188 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2189 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 2190 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2191 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2192 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 2193 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2194 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2195 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2196 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque *, struct GB_Matrix_opaque *, uint64_t, uint64_t)
/* 2197 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2198 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 2199 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2200 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2201 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2202 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque *, struct GB_Vector_opaque *, uint64_t)
/* 2203 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2204 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2205 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2206 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2207 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque *, uint16_t)
/* 2208 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2209 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 20),
/* 2210 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2211 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque *, uint32_t)
/* 2212 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2213 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 2214 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2215 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque *, uint64_t)
/* 2216 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2217 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2218 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2219 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque *, uint8_t)
/* 2220 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2221 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 18),
/* 2222 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2223 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Scalar_opaque *, void *)
/* 2224 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2225 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 2226 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2227 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_SelectOp_opaque * *)
/* 2228 */ _CFFI_OP(_CFFI_OP_POINTER, 155), // struct GB_SelectOp_opaque * *
/* 2229 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2230 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_SelectOp_opaque * *, _Bool(*)(uint64_t, uint64_t, void const *, void const *), struct GB_Type_opaque *, struct GB_Type_opaque *)
/* 2231 */ _CFFI_OP(_CFFI_OP_NOOP, 2228),
/* 2232 */ _CFFI_OP(_CFFI_OP_POINTER, 3897), // _Bool(*)(uint64_t, uint64_t, void const *, void const *)
/* 2233 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 2234 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 2235 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2236 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_SelectOp_opaque * *, _Bool(*)(uint64_t, uint64_t, void const *, void const *), struct GB_Type_opaque *, struct GB_Type_opaque *, char const *)
/* 2237 */ _CFFI_OP(_CFFI_OP_NOOP, 2228),
/* 2238 */ _CFFI_OP(_CFFI_OP_NOOP, 2232),
/* 2239 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 2240 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 2241 */ _CFFI_OP(_CFFI_OP_NOOP, 445),
/* 2242 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2243 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_SelectOp_opaque *, GrB_WaitMode)
/* 2244 */ _CFFI_OP(_CFFI_OP_NOOP, 155),
/* 2245 */ _CFFI_OP(_CFFI_OP_NOOP, 458),
/* 2246 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2247 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_SelectOp_opaque *, char const *, GxB_Print_Level, FILE *)
/* 2248 */ _CFFI_OP(_CFFI_OP_NOOP, 155),
/* 2249 */ _CFFI_OP(_CFFI_OP_NOOP, 445),
/* 2250 */ _CFFI_OP(_CFFI_OP_NOOP, 463),
/* 2251 */ _CFFI_OP(_CFFI_OP_NOOP, 464),
/* 2252 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2253 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Semiring_opaque * *)
/* 2254 */ _CFFI_OP(_CFFI_OP_POINTER, 159), // struct GB_Semiring_opaque * *
/* 2255 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2256 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Semiring_opaque * *, struct GB_Monoid_opaque *, struct GB_BinaryOp_opaque *)
/* 2257 */ _CFFI_OP(_CFFI_OP_NOOP, 2254),
/* 2258 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 2259 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2260 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2261 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Semiring_opaque *, GrB_WaitMode)
/* 2262 */ _CFFI_OP(_CFFI_OP_NOOP, 159),
/* 2263 */ _CFFI_OP(_CFFI_OP_NOOP, 458),
/* 2264 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2265 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Semiring_opaque *, char const *, GxB_Print_Level, FILE *)
/* 2266 */ _CFFI_OP(_CFFI_OP_NOOP, 159),
/* 2267 */ _CFFI_OP(_CFFI_OP_NOOP, 445),
/* 2268 */ _CFFI_OP(_CFFI_OP_NOOP, 463),
/* 2269 */ _CFFI_OP(_CFFI_OP_NOOP, 464),
/* 2270 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2271 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Type_opaque * *)
/* 2272 */ _CFFI_OP(_CFFI_OP_NOOP, 585),
/* 2273 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2274 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Type_opaque * *, char const *)
/* 2275 */ _CFFI_OP(_CFFI_OP_NOOP, 585),
/* 2276 */ _CFFI_OP(_CFFI_OP_NOOP, 445),
/* 2277 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2278 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Type_opaque * *, size_t)
/* 2279 */ _CFFI_OP(_CFFI_OP_NOOP, 585),
/* 2280 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 28), // size_t
/* 2281 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2282 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Type_opaque * *, size_t, char const *)
/* 2283 */ _CFFI_OP(_CFFI_OP_NOOP, 585),
/* 2284 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 28),
/* 2285 */ _CFFI_OP(_CFFI_OP_NOOP, 445),
/* 2286 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2287 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Type_opaque * *, size_t, char const *, char const *)
/* 2288 */ _CFFI_OP(_CFFI_OP_NOOP, 585),
/* 2289 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 28),
/* 2290 */ _CFFI_OP(_CFFI_OP_NOOP, 445),
/* 2291 */ _CFFI_OP(_CFFI_OP_NOOP, 445),
/* 2292 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2293 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Type_opaque * *, struct GB_BinaryOp_opaque *)
/* 2294 */ _CFFI_OP(_CFFI_OP_NOOP, 585),
/* 2295 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2296 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2297 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Type_opaque * *, struct GB_Matrix_opaque *)
/* 2298 */ _CFFI_OP(_CFFI_OP_NOOP, 585),
/* 2299 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 2300 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2301 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Type_opaque * *, struct GB_Scalar_opaque *)
/* 2302 */ _CFFI_OP(_CFFI_OP_NOOP, 585),
/* 2303 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2304 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2305 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Type_opaque * *, struct GB_SelectOp_opaque *)
/* 2306 */ _CFFI_OP(_CFFI_OP_NOOP, 585),
/* 2307 */ _CFFI_OP(_CFFI_OP_NOOP, 155),
/* 2308 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2309 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Type_opaque * *, struct GB_UnaryOp_opaque *)
/* 2310 */ _CFFI_OP(_CFFI_OP_NOOP, 585),
/* 2311 */ _CFFI_OP(_CFFI_OP_NOOP, 118),
/* 2312 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2313 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Type_opaque * *, struct GB_Vector_opaque *)
/* 2314 */ _CFFI_OP(_CFFI_OP_NOOP, 585),
/* 2315 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2316 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2317 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Type_opaque *, GrB_WaitMode)
/* 2318 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 2319 */ _CFFI_OP(_CFFI_OP_NOOP, 458),
/* 2320 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2321 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Type_opaque *, char const *, GxB_Print_Level, FILE *)
/* 2322 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 2323 */ _CFFI_OP(_CFFI_OP_NOOP, 445),
/* 2324 */ _CFFI_OP(_CFFI_OP_NOOP, 463),
/* 2325 */ _CFFI_OP(_CFFI_OP_NOOP, 464),
/* 2326 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2327 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_UnaryOp_opaque * *)
/* 2328 */ _CFFI_OP(_CFFI_OP_POINTER, 118), // struct GB_UnaryOp_opaque * *
/* 2329 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2330 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_UnaryOp_opaque * *, void(*)(void *, void const *), struct GB_Type_opaque *, struct GB_Type_opaque *)
/* 2331 */ _CFFI_OP(_CFFI_OP_NOOP, 2328),
/* 2332 */ _CFFI_OP(_CFFI_OP_POINTER, 3962), // void(*)(void *, void const *)
/* 2333 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 2334 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 2335 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2336 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_UnaryOp_opaque * *, void(*)(void *, void const *), struct GB_Type_opaque *, struct GB_Type_opaque *, char const *)
/* 2337 */ _CFFI_OP(_CFFI_OP_NOOP, 2328),
/* 2338 */ _CFFI_OP(_CFFI_OP_NOOP, 2332),
/* 2339 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 2340 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 2341 */ _CFFI_OP(_CFFI_OP_NOOP, 445),
/* 2342 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2343 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_UnaryOp_opaque * *, void(*)(void *, void const *), struct GB_Type_opaque *, struct GB_Type_opaque *, char const *, char const *)
/* 2344 */ _CFFI_OP(_CFFI_OP_NOOP, 2328),
/* 2345 */ _CFFI_OP(_CFFI_OP_NOOP, 2332),
/* 2346 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 2347 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 2348 */ _CFFI_OP(_CFFI_OP_NOOP, 445),
/* 2349 */ _CFFI_OP(_CFFI_OP_NOOP, 445),
/* 2350 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2351 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_UnaryOp_opaque *, GrB_WaitMode)
/* 2352 */ _CFFI_OP(_CFFI_OP_NOOP, 118),
/* 2353 */ _CFFI_OP(_CFFI_OP_NOOP, 458),
/* 2354 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2355 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_UnaryOp_opaque *, char const *, GxB_Print_Level, FILE *)
/* 2356 */ _CFFI_OP(_CFFI_OP_NOOP, 118),
/* 2357 */ _CFFI_OP(_CFFI_OP_NOOP, 445),
/* 2358 */ _CFFI_OP(_CFFI_OP_NOOP, 463),
/* 2359 */ _CFFI_OP(_CFFI_OP_NOOP, 464),
/* 2360 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2361 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque * *)
/* 2362 */ _CFFI_OP(_CFFI_OP_POINTER, 65), // struct GB_Vector_opaque * *
/* 2363 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2364 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque * *, struct GB_Type_opaque * *, uint64_t *, int8_t * *, void * *, uint64_t *, uint64_t *, _Bool *, uint64_t *, struct GB_Descriptor_opaque *)
/* 2365 */ _CFFI_OP(_CFFI_OP_NOOP, 2362),
/* 2366 */ _CFFI_OP(_CFFI_OP_NOOP, 585),
/* 2367 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 2368 */ _CFFI_OP(_CFFI_OP_NOOP, 588),
/* 2369 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 2370 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 2371 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 2372 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 2373 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 2374 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2375 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2376 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque * *, struct GB_Type_opaque * *, uint64_t *, uint64_t * *, void * *, uint64_t *, uint64_t *, _Bool *, uint64_t *, _Bool *, struct GB_Descriptor_opaque *)
/* 2377 */ _CFFI_OP(_CFFI_OP_NOOP, 2362),
/* 2378 */ _CFFI_OP(_CFFI_OP_NOOP, 585),
/* 2379 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 2380 */ _CFFI_OP(_CFFI_OP_NOOP, 601),
/* 2381 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 2382 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 2383 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 2384 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 2385 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 2386 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 2387 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2388 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2389 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque * *, struct GB_Type_opaque * *, uint64_t *, void * *, uint64_t *, _Bool *, struct GB_Descriptor_opaque *)
/* 2390 */ _CFFI_OP(_CFFI_OP_NOOP, 2362),
/* 2391 */ _CFFI_OP(_CFFI_OP_NOOP, 585),
/* 2392 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 2393 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 2394 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 2395 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 2396 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2397 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2398 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque * *, struct GB_Type_opaque *, uint64_t)
/* 2399 */ _CFFI_OP(_CFFI_OP_NOOP, 2362),
/* 2400 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 2401 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2402 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2403 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque * *, struct GB_Type_opaque *, uint64_t, int8_t * *, void * *, uint64_t, uint64_t, _Bool, uint64_t, struct GB_Descriptor_opaque *)
/* 2404 */ _CFFI_OP(_CFFI_OP_NOOP, 2362),
/* 2405 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 2406 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2407 */ _CFFI_OP(_CFFI_OP_NOOP, 588),
/* 2408 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 2409 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2410 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2411 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 2412 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2413 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2414 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2415 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque * *, struct GB_Type_opaque *, uint64_t, uint64_t * *, void * *, uint64_t, uint64_t, _Bool, uint64_t, _Bool, struct GB_Descriptor_opaque *)
/* 2416 */ _CFFI_OP(_CFFI_OP_NOOP, 2362),
/* 2417 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 2418 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2419 */ _CFFI_OP(_CFFI_OP_NOOP, 601),
/* 2420 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 2421 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2422 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2423 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 2424 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2425 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 2426 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2427 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2428 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque * *, struct GB_Type_opaque *, uint64_t, void * *, uint64_t, _Bool, struct GB_Descriptor_opaque *)
/* 2429 */ _CFFI_OP(_CFFI_OP_NOOP, 2362),
/* 2430 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 2431 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2432 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 2433 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2434 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 2435 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2436 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2437 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque * *, struct GB_Type_opaque *, void const *, uint64_t, struct GB_Descriptor_opaque *)
/* 2438 */ _CFFI_OP(_CFFI_OP_NOOP, 2362),
/* 2439 */ _CFFI_OP(_CFFI_OP_NOOP, 114),
/* 2440 */ _CFFI_OP(_CFFI_OP_NOOP, 126),
/* 2441 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2442 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2443 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2444 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque * *, struct GB_Vector_opaque *)
/* 2445 */ _CFFI_OP(_CFFI_OP_NOOP, 2362),
/* 2446 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2447 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2448 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *)
/* 2449 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2450 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2451 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, GrB_WaitMode)
/* 2452 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2453 */ _CFFI_OP(_CFFI_OP_NOOP, 458),
/* 2454 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2455 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, GxB_Option_Field, ...)
/* 2456 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2457 */ _CFFI_OP(_CFFI_OP_NOOP, 20),
/* 2458 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 1),
/* 2459 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, GxB_Option_Field, double *)
/* 2460 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2461 */ _CFFI_OP(_CFFI_OP_NOOP, 20),
/* 2462 */ _CFFI_OP(_CFFI_OP_NOOP, 28),
/* 2463 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2464 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, GxB_Option_Field, double)
/* 2465 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2466 */ _CFFI_OP(_CFFI_OP_NOOP, 20),
/* 2467 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 14),
/* 2468 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2469 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, GxB_Option_Field, int32_t *)
/* 2470 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2471 */ _CFFI_OP(_CFFI_OP_NOOP, 20),
/* 2472 */ _CFFI_OP(_CFFI_OP_NOOP, 36),
/* 2473 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2474 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, GxB_Option_Field, int32_t)
/* 2475 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2476 */ _CFFI_OP(_CFFI_OP_NOOP, 20),
/* 2477 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 21),
/* 2478 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2479 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, _Bool, uint64_t)
/* 2480 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2481 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 2482 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2483 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2484 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, char const *, GxB_Print_Level, FILE *)
/* 2485 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2486 */ _CFFI_OP(_CFFI_OP_NOOP, 445),
/* 2487 */ _CFFI_OP(_CFFI_OP_NOOP, 463),
/* 2488 */ _CFFI_OP(_CFFI_OP_NOOP, 464),
/* 2489 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2490 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, _Dcomplex, uint64_t)
/* 2491 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2492 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 49),
/* 2493 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2494 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2495 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, double, uint64_t)
/* 2496 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2497 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 14),
/* 2498 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2499 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2500 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, _Fcomplex, uint64_t)
/* 2501 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2502 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 48),
/* 2503 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2504 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2505 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, float, uint64_t)
/* 2506 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2507 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 13),
/* 2508 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2509 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2510 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, int16_t, uint64_t)
/* 2511 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2512 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 19),
/* 2513 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2514 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2515 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, int32_t, uint64_t)
/* 2516 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2517 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 21),
/* 2518 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2519 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2520 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, int64_t, uint64_t)
/* 2521 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2522 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 23),
/* 2523 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2524 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2525 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, int8_t * *, void * *, uint64_t *, uint64_t *, _Bool *, uint64_t *, struct GB_Descriptor_opaque *)
/* 2526 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2527 */ _CFFI_OP(_CFFI_OP_NOOP, 588),
/* 2528 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 2529 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 2530 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 2531 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 2532 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 2533 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2534 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2535 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, int8_t * *, void * *, uint64_t, uint64_t, _Bool, uint64_t, struct GB_Descriptor_opaque *)
/* 2536 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2537 */ _CFFI_OP(_CFFI_OP_NOOP, 588),
/* 2538 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 2539 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2540 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2541 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 2542 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2543 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2544 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2545 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, int8_t, uint64_t)
/* 2546 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2547 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 17),
/* 2548 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2549 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2550 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Matrix_opaque *, int64_t, struct GB_Descriptor_opaque *)
/* 2551 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2552 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 2553 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 23),
/* 2554 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2555 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2556 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Scalar_opaque *, uint64_t)
/* 2557 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2558 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2559 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2560 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2561 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, _Bool, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 2562 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2563 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2564 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2565 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 2566 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 2567 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2568 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2569 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2570 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, _Dcomplex, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 2571 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2572 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2573 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2574 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 49),
/* 2575 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 2576 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2577 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2578 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2579 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, double, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 2580 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2581 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2582 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2583 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 14),
/* 2584 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 2585 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2586 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2587 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2588 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, _Fcomplex, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 2589 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2590 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2591 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2592 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 48),
/* 2593 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 2594 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2595 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2596 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2597 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, float, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 2598 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2599 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2600 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2601 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 13),
/* 2602 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 2603 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2604 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2605 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2606 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, int16_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 2607 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2608 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2609 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2610 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 19),
/* 2611 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 2612 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2613 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2614 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2615 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, int32_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 2616 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2617 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2618 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2619 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 21),
/* 2620 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 2621 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2622 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2623 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2624 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, int64_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 2625 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2626 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2627 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2628 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 23),
/* 2629 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 2630 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2631 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2632 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2633 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, int8_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 2634 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2635 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2636 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2637 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 17),
/* 2638 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 2639 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2640 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2641 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2642 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, _Bool, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 2643 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2644 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2645 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2646 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2647 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 2648 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2649 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2650 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2651 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, _Dcomplex, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 2652 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2653 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2654 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2655 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2656 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 49),
/* 2657 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2658 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2659 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2660 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, double, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 2661 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2662 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2663 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2664 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2665 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 14),
/* 2666 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2667 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2668 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2669 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, _Fcomplex, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 2670 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2671 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2672 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2673 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2674 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 48),
/* 2675 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2676 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2677 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2678 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, float, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 2679 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2680 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2681 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2682 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2683 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 13),
/* 2684 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2685 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2686 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2687 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, int16_t, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 2688 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2689 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2690 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2691 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2692 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 19),
/* 2693 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2694 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2695 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2696 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, int32_t, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 2697 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2698 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2699 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2700 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2701 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 21),
/* 2702 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2703 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2704 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2705 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, int64_t, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 2706 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2707 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2708 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2709 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2710 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 23),
/* 2711 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2712 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2713 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2714 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, int8_t, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 2715 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2716 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2717 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2718 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2719 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 17),
/* 2720 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2721 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2722 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2723 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 2724 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2725 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2726 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2727 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2728 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 2729 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2730 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2731 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Scalar_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 2732 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2733 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2734 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2735 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2736 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2737 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2738 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2739 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2740 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Vector_opaque *, _Bool, struct GB_Descriptor_opaque *)
/* 2741 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2742 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2743 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2744 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2745 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2746 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 2747 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2748 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2749 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Vector_opaque *, _Dcomplex, struct GB_Descriptor_opaque *)
/* 2750 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2751 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2752 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2753 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2754 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2755 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 49),
/* 2756 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2757 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2758 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Vector_opaque *, double, struct GB_Descriptor_opaque *)
/* 2759 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2760 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2761 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2762 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2763 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2764 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 14),
/* 2765 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2766 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2767 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Vector_opaque *, _Fcomplex, struct GB_Descriptor_opaque *)
/* 2768 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2769 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2770 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2771 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2772 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2773 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 48),
/* 2774 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2775 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2776 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Vector_opaque *, float, struct GB_Descriptor_opaque *)
/* 2777 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2778 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2779 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2780 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2781 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2782 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 13),
/* 2783 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2784 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2785 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Vector_opaque *, int16_t, struct GB_Descriptor_opaque *)
/* 2786 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2787 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2788 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2789 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2790 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2791 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 19),
/* 2792 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2793 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2794 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Vector_opaque *, int32_t, struct GB_Descriptor_opaque *)
/* 2795 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2796 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2797 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2798 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2799 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2800 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 21),
/* 2801 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2802 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2803 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Vector_opaque *, int64_t, struct GB_Descriptor_opaque *)
/* 2804 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2805 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2806 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2807 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2808 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2809 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 23),
/* 2810 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2811 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2812 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Vector_opaque *, int8_t, struct GB_Descriptor_opaque *)
/* 2813 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2814 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2815 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2816 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2817 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2818 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 17),
/* 2819 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2820 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2821 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Vector_opaque *, struct GB_Scalar_opaque *, struct GB_Descriptor_opaque *)
/* 2822 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2823 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2824 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2825 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2826 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2827 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2828 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2829 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2830 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Vector_opaque *, struct GB_Scalar_opaque *, struct GB_Vector_opaque *, struct GB_Scalar_opaque *, struct GB_Descriptor_opaque *)
/* 2831 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2832 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2833 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2834 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2835 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2836 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2837 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2838 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 2839 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2840 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2841 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 2842 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2843 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2844 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2845 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2846 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2847 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2848 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2849 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2850 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Vector_opaque *, uint16_t, struct GB_Descriptor_opaque *)
/* 2851 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2852 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2853 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2854 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2855 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2856 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 20),
/* 2857 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2858 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2859 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Vector_opaque *, uint32_t, struct GB_Descriptor_opaque *)
/* 2860 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2861 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2862 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2863 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2864 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2865 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 2866 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2867 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2868 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Vector_opaque *, uint64_t, struct GB_Descriptor_opaque *)
/* 2869 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2870 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2871 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2872 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2873 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2874 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2875 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2876 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2877 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Vector_opaque *, uint8_t, struct GB_Descriptor_opaque *)
/* 2878 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2879 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2880 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2881 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2882 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2883 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 18),
/* 2884 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2885 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2886 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, struct GB_Vector_opaque *, void const *, struct GB_Descriptor_opaque *)
/* 2887 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2888 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2889 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2890 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2891 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2892 */ _CFFI_OP(_CFFI_OP_NOOP, 126),
/* 2893 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2894 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2895 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, uint16_t, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 2896 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2897 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2898 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2899 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2900 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 20),
/* 2901 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2902 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2903 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2904 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, uint32_t, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 2905 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2906 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2907 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2908 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2909 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 2910 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2911 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2912 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2913 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, uint64_t, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 2914 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2915 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2916 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2917 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2918 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 2919 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2920 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2921 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2922 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, uint8_t, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 2923 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2924 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2925 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2926 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2927 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 18),
/* 2928 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2929 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2930 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2931 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_BinaryOp_opaque *, void const *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 2932 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2933 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2934 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2935 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2936 */ _CFFI_OP(_CFFI_OP_NOOP, 126),
/* 2937 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2938 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2939 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2940 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Vector_opaque *, _Bool, struct GB_Descriptor_opaque *)
/* 2941 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2942 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2943 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2944 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 2945 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2946 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 2947 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2948 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2949 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Vector_opaque *, _Dcomplex, struct GB_Descriptor_opaque *)
/* 2950 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2951 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2952 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2953 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 2954 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2955 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 49),
/* 2956 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2957 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2958 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Vector_opaque *, double, struct GB_Descriptor_opaque *)
/* 2959 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2960 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2961 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2962 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 2963 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2964 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 14),
/* 2965 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2966 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2967 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Vector_opaque *, _Fcomplex, struct GB_Descriptor_opaque *)
/* 2968 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2969 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2970 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2971 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 2972 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2973 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 48),
/* 2974 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2975 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2976 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Vector_opaque *, float, struct GB_Descriptor_opaque *)
/* 2977 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2978 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2979 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2980 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 2981 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2982 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 13),
/* 2983 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2984 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2985 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Vector_opaque *, int16_t, struct GB_Descriptor_opaque *)
/* 2986 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2987 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2988 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2989 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 2990 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2991 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 19),
/* 2992 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 2993 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 2994 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Vector_opaque *, int32_t, struct GB_Descriptor_opaque *)
/* 2995 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2996 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 2997 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 2998 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 2999 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3000 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 21),
/* 3001 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3002 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3003 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Vector_opaque *, int64_t, struct GB_Descriptor_opaque *)
/* 3004 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3005 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3006 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3007 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 3008 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3009 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 23),
/* 3010 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3011 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3012 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Vector_opaque *, int8_t, struct GB_Descriptor_opaque *)
/* 3013 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3014 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3015 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3016 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 3017 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3018 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 17),
/* 3019 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3020 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3021 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Vector_opaque *, struct GB_Scalar_opaque *, struct GB_Descriptor_opaque *)
/* 3022 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3023 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3024 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3025 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 3026 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3027 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 3028 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3029 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3030 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Vector_opaque *, uint16_t, struct GB_Descriptor_opaque *)
/* 3031 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3032 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3033 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3034 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 3035 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3036 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 20),
/* 3037 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3038 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3039 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Vector_opaque *, uint32_t, struct GB_Descriptor_opaque *)
/* 3040 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3041 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3042 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3043 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 3044 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3045 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 3046 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3047 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3048 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Vector_opaque *, uint64_t, struct GB_Descriptor_opaque *)
/* 3049 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3050 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3051 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3052 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 3053 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3054 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3055 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3056 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3057 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Vector_opaque *, uint8_t, struct GB_Descriptor_opaque *)
/* 3058 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3059 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3060 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3061 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 3062 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3063 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 18),
/* 3064 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3065 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3066 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_IndexUnaryOp_opaque *, struct GB_Vector_opaque *, void const *, struct GB_Descriptor_opaque *)
/* 3067 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3068 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3069 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3070 */ _CFFI_OP(_CFFI_OP_NOOP, 102),
/* 3071 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3072 */ _CFFI_OP(_CFFI_OP_NOOP, 126),
/* 3073 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3074 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3075 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_Matrix_opaque *, uint64_t const *, uint64_t, uint64_t, struct GB_Descriptor_opaque *)
/* 3076 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3077 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3078 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3079 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3080 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3081 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3082 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3083 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3084 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3085 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 3086 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3087 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3088 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3089 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 3090 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3091 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3092 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3093 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 3094 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3095 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3096 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3097 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 3098 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3099 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3100 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3101 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3102 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_Scalar_opaque *, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 3103 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3104 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3105 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3106 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 3107 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3108 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3109 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3110 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3111 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_SelectOp_opaque *, struct GB_Vector_opaque *, struct GB_Scalar_opaque *, struct GB_Descriptor_opaque *)
/* 3112 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3113 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3114 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3115 */ _CFFI_OP(_CFFI_OP_NOOP, 155),
/* 3116 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3117 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 3118 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3119 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3120 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_Semiring_opaque *, struct GB_Matrix_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 3121 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3122 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3123 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3124 */ _CFFI_OP(_CFFI_OP_NOOP, 159),
/* 3125 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3126 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3127 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3128 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3129 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_Semiring_opaque *, struct GB_Vector_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 3130 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3131 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3132 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3133 */ _CFFI_OP(_CFFI_OP_NOOP, 159),
/* 3134 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3135 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3136 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3137 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3138 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_Semiring_opaque *, struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 3139 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3140 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3141 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3142 */ _CFFI_OP(_CFFI_OP_NOOP, 159),
/* 3143 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3144 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3145 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3146 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3147 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_UnaryOp_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 3148 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3149 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3150 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3151 */ _CFFI_OP(_CFFI_OP_NOOP, 118),
/* 3152 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3153 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3154 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3155 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 3156 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3157 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3158 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3159 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3160 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3161 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3162 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, struct GB_Vector_opaque *, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 3163 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3164 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3165 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3166 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3167 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3168 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3169 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3170 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3171 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, uint16_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 3172 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3173 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3174 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3175 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 20),
/* 3176 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3177 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3178 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3179 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3180 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, uint32_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 3181 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3182 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3183 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3184 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 3185 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3186 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3187 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3188 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3189 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, uint64_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 3190 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3191 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3192 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3193 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3194 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3195 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3196 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3197 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3198 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, uint8_t, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 3199 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3200 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3201 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3202 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 18),
/* 3203 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3204 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3205 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3206 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3207 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, struct GB_Vector_opaque *, struct GB_BinaryOp_opaque *, void *, uint64_t const *, uint64_t, struct GB_Descriptor_opaque *)
/* 3208 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3209 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3210 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3211 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 3212 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3213 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3214 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3215 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3216 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, uint16_t, uint64_t)
/* 3217 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3218 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 20),
/* 3219 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3220 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3221 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, uint32_t, uint64_t)
/* 3222 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3223 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 22),
/* 3224 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3225 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3226 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, uint64_t * *, void * *, uint64_t *, uint64_t *, _Bool *, uint64_t *, _Bool *, struct GB_Descriptor_opaque *)
/* 3227 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3228 */ _CFFI_OP(_CFFI_OP_NOOP, 601),
/* 3229 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 3230 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3231 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3232 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 3233 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3234 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 3235 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3236 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3237 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, uint64_t * *, void * *, uint64_t, uint64_t, _Bool, uint64_t, _Bool, struct GB_Descriptor_opaque *)
/* 3238 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3239 */ _CFFI_OP(_CFFI_OP_NOOP, 601),
/* 3240 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 3241 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3242 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3243 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 3244 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3245 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 3246 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3247 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3248 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, uint64_t const *, _Bool const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 3249 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3250 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3251 */ _CFFI_OP(_CFFI_OP_NOOP, 698),
/* 3252 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3253 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3254 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3255 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, uint64_t const *, _Dcomplex const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 3256 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3257 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3258 */ _CFFI_OP(_CFFI_OP_NOOP, 711),
/* 3259 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3260 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3261 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3262 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, uint64_t const *, double const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 3263 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3264 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3265 */ _CFFI_OP(_CFFI_OP_NOOP, 724),
/* 3266 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3267 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3268 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3269 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, uint64_t const *, _Fcomplex const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 3270 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3271 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3272 */ _CFFI_OP(_CFFI_OP_NOOP, 737),
/* 3273 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3274 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3275 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3276 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, uint64_t const *, float const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 3277 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3278 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3279 */ _CFFI_OP(_CFFI_OP_NOOP, 750),
/* 3280 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3281 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3282 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3283 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, uint64_t const *, int16_t const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 3284 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3285 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3286 */ _CFFI_OP(_CFFI_OP_NOOP, 763),
/* 3287 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3288 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3289 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3290 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, uint64_t const *, int32_t const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 3291 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3292 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3293 */ _CFFI_OP(_CFFI_OP_NOOP, 776),
/* 3294 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3295 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3296 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3297 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, uint64_t const *, int64_t const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 3298 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3299 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3300 */ _CFFI_OP(_CFFI_OP_NOOP, 789),
/* 3301 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3302 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3303 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3304 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, uint64_t const *, int8_t const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 3305 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3306 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3307 */ _CFFI_OP(_CFFI_OP_NOOP, 802),
/* 3308 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3309 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3310 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3311 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, uint64_t const *, struct GB_Scalar_opaque *, uint64_t)
/* 3312 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3313 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3314 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 3315 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3316 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3317 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, uint64_t const *, uint16_t const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 3318 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3319 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3320 */ _CFFI_OP(_CFFI_OP_NOOP, 815),
/* 3321 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3322 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3323 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3324 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, uint64_t const *, uint32_t const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 3325 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3326 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3327 */ _CFFI_OP(_CFFI_OP_NOOP, 828),
/* 3328 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3329 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3330 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3331 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, uint64_t const *, uint64_t const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 3332 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3333 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3334 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3335 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3336 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3337 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3338 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, uint64_t const *, uint8_t const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 3339 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3340 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3341 */ _CFFI_OP(_CFFI_OP_NOOP, 854),
/* 3342 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3343 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3344 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3345 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, uint64_t const *, void const *, uint64_t, struct GB_BinaryOp_opaque *)
/* 3346 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3347 */ _CFFI_OP(_CFFI_OP_NOOP, 696),
/* 3348 */ _CFFI_OP(_CFFI_OP_NOOP, 126),
/* 3349 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3350 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3351 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3352 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, uint64_t)
/* 3353 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3354 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3355 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3356 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, uint64_t, uint64_t)
/* 3357 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3358 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3359 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3360 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3361 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, uint8_t, uint64_t)
/* 3362 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3363 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 18),
/* 3364 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3365 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3366 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, void * *, uint64_t *, _Bool *, struct GB_Descriptor_opaque *)
/* 3367 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3368 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 3369 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3370 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 3371 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3372 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3373 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, void * *, uint64_t, _Bool, struct GB_Descriptor_opaque *)
/* 3374 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3375 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 3376 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3377 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 1),
/* 3378 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3379 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3380 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(struct GB_Vector_opaque *, void *, uint64_t)
/* 3381 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3382 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 3383 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3384 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3385 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint16_t *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 3386 */ _CFFI_OP(_CFFI_OP_POINTER, 1336), // uint16_t *
/* 3387 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3388 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 3389 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3390 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3391 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3392 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint16_t *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 3393 */ _CFFI_OP(_CFFI_OP_NOOP, 3386),
/* 3394 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3395 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 3396 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3397 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3398 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3399 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint16_t *, struct GB_Matrix_opaque *, uint64_t, uint64_t)
/* 3400 */ _CFFI_OP(_CFFI_OP_NOOP, 3386),
/* 3401 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3402 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3403 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3404 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3405 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint16_t *, struct GB_Scalar_opaque *)
/* 3406 */ _CFFI_OP(_CFFI_OP_NOOP, 3386),
/* 3407 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 3408 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3409 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint16_t *, struct GB_Vector_opaque *, uint64_t)
/* 3410 */ _CFFI_OP(_CFFI_OP_NOOP, 3386),
/* 3411 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3412 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3413 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3414 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint32_t *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 3415 */ _CFFI_OP(_CFFI_OP_POINTER, 1345), // uint32_t *
/* 3416 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3417 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 3418 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3419 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3420 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3421 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint32_t *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 3422 */ _CFFI_OP(_CFFI_OP_NOOP, 3415),
/* 3423 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3424 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 3425 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3426 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3427 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3428 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint32_t *, struct GB_Matrix_opaque *, uint64_t, uint64_t)
/* 3429 */ _CFFI_OP(_CFFI_OP_NOOP, 3415),
/* 3430 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3431 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3432 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3433 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3434 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint32_t *, struct GB_Scalar_opaque *)
/* 3435 */ _CFFI_OP(_CFFI_OP_NOOP, 3415),
/* 3436 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 3437 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3438 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint32_t *, struct GB_Vector_opaque *, uint64_t)
/* 3439 */ _CFFI_OP(_CFFI_OP_NOOP, 3415),
/* 3440 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3441 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3442 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3443 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, _Bool *, uint64_t *, struct GB_Vector_opaque *)
/* 3444 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3445 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 3446 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3447 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3448 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3449 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, double *, uint64_t *, struct GB_Vector_opaque *)
/* 3450 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3451 */ _CFFI_OP(_CFFI_OP_NOOP, 28),
/* 3452 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3453 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3454 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3455 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, _Dcomplex *, uint64_t *, struct GB_Vector_opaque *)
/* 3456 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3457 */ _CFFI_OP(_CFFI_OP_NOOP, 203),
/* 3458 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3459 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3460 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3461 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, float *, uint64_t *, struct GB_Vector_opaque *)
/* 3462 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3463 */ _CFFI_OP(_CFFI_OP_NOOP, 232),
/* 3464 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3465 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3466 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3467 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, _Fcomplex *, uint64_t *, struct GB_Vector_opaque *)
/* 3468 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3469 */ _CFFI_OP(_CFFI_OP_NOOP, 261),
/* 3470 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3471 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3472 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3473 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, int16_t *, uint64_t *, struct GB_Vector_opaque *)
/* 3474 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3475 */ _CFFI_OP(_CFFI_OP_NOOP, 290),
/* 3476 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3477 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3478 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3479 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, int32_t *, uint64_t *, struct GB_Vector_opaque *)
/* 3480 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3481 */ _CFFI_OP(_CFFI_OP_NOOP, 36),
/* 3482 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3483 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3484 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3485 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, int64_t *, uint64_t *, struct GB_Vector_opaque *)
/* 3486 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3487 */ _CFFI_OP(_CFFI_OP_NOOP, 44),
/* 3488 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3489 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3490 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3491 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, int8_t *, uint64_t *, struct GB_Vector_opaque *)
/* 3492 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3493 */ _CFFI_OP(_CFFI_OP_NOOP, 377),
/* 3494 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3495 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3496 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3497 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 3498 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3499 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3500 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 3501 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3502 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3503 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3504 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 3505 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3506 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3507 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 3508 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3509 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3510 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3511 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, struct GB_Matrix_opaque *)
/* 3512 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3513 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3514 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3515 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, struct GB_Matrix_opaque *, uint64_t, uint64_t)
/* 3516 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3517 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3518 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3519 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3520 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3521 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, struct GB_Scalar_opaque *)
/* 3522 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3523 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 3524 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3525 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, struct GB_Vector_opaque *)
/* 3526 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3527 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3528 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3529 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, struct GB_Vector_opaque *, uint64_t)
/* 3530 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3531 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3532 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3533 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3534 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint16_t *, uint64_t *, struct GB_Vector_opaque *)
/* 3535 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3536 */ _CFFI_OP(_CFFI_OP_NOOP, 3386),
/* 3537 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3538 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3539 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3540 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint32_t *, uint64_t *, struct GB_Vector_opaque *)
/* 3541 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3542 */ _CFFI_OP(_CFFI_OP_NOOP, 3415),
/* 3543 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3544 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3545 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3546 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, _Bool *, uint64_t *, struct GB_Matrix_opaque *)
/* 3547 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3548 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3549 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 3550 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3551 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3552 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3553 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, _Bool *, uint64_t *, uint64_t *, uint64_t *, GrB_Format, struct GB_Matrix_opaque *)
/* 3554 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3555 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3556 */ _CFFI_OP(_CFFI_OP_NOOP, 55),
/* 3557 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3558 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3559 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3560 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 3561 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3562 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3563 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, double *, uint64_t *, struct GB_Matrix_opaque *)
/* 3564 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3565 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3566 */ _CFFI_OP(_CFFI_OP_NOOP, 28),
/* 3567 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3568 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3569 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3570 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, double *, uint64_t *, uint64_t *, uint64_t *, GrB_Format, struct GB_Matrix_opaque *)
/* 3571 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3572 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3573 */ _CFFI_OP(_CFFI_OP_NOOP, 28),
/* 3574 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3575 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3576 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3577 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 3578 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3579 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3580 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, _Dcomplex *, uint64_t *, struct GB_Matrix_opaque *)
/* 3581 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3582 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3583 */ _CFFI_OP(_CFFI_OP_NOOP, 203),
/* 3584 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3585 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3586 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3587 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, _Dcomplex *, uint64_t *, uint64_t *, uint64_t *, GrB_Format, struct GB_Matrix_opaque *)
/* 3588 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3589 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3590 */ _CFFI_OP(_CFFI_OP_NOOP, 203),
/* 3591 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3592 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3593 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3594 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 3595 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3596 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3597 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, float *, uint64_t *, struct GB_Matrix_opaque *)
/* 3598 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3599 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3600 */ _CFFI_OP(_CFFI_OP_NOOP, 232),
/* 3601 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3602 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3603 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3604 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, float *, uint64_t *, uint64_t *, uint64_t *, GrB_Format, struct GB_Matrix_opaque *)
/* 3605 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3606 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3607 */ _CFFI_OP(_CFFI_OP_NOOP, 232),
/* 3608 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3609 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3610 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3611 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 3612 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3613 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3614 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, _Fcomplex *, uint64_t *, struct GB_Matrix_opaque *)
/* 3615 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3616 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3617 */ _CFFI_OP(_CFFI_OP_NOOP, 261),
/* 3618 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3619 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3620 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3621 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, _Fcomplex *, uint64_t *, uint64_t *, uint64_t *, GrB_Format, struct GB_Matrix_opaque *)
/* 3622 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3623 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3624 */ _CFFI_OP(_CFFI_OP_NOOP, 261),
/* 3625 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3626 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3627 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3628 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 3629 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3630 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3631 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, int16_t *, uint64_t *, struct GB_Matrix_opaque *)
/* 3632 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3633 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3634 */ _CFFI_OP(_CFFI_OP_NOOP, 290),
/* 3635 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3636 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3637 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3638 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, int16_t *, uint64_t *, uint64_t *, uint64_t *, GrB_Format, struct GB_Matrix_opaque *)
/* 3639 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3640 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3641 */ _CFFI_OP(_CFFI_OP_NOOP, 290),
/* 3642 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3643 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3644 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3645 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 3646 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3647 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3648 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, int32_t *, uint64_t *, struct GB_Matrix_opaque *)
/* 3649 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3650 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3651 */ _CFFI_OP(_CFFI_OP_NOOP, 36),
/* 3652 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3653 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3654 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3655 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, int32_t *, uint64_t *, uint64_t *, uint64_t *, GrB_Format, struct GB_Matrix_opaque *)
/* 3656 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3657 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3658 */ _CFFI_OP(_CFFI_OP_NOOP, 36),
/* 3659 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3660 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3661 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3662 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 3663 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3664 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3665 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, int64_t *, uint64_t *, struct GB_Matrix_opaque *)
/* 3666 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3667 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3668 */ _CFFI_OP(_CFFI_OP_NOOP, 44),
/* 3669 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3670 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3671 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3672 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, int64_t *, uint64_t *, uint64_t *, uint64_t *, GrB_Format, struct GB_Matrix_opaque *)
/* 3673 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3674 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3675 */ _CFFI_OP(_CFFI_OP_NOOP, 44),
/* 3676 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3677 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3678 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3679 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 3680 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3681 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3682 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, int8_t *, uint64_t *, struct GB_Matrix_opaque *)
/* 3683 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3684 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3685 */ _CFFI_OP(_CFFI_OP_NOOP, 377),
/* 3686 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3687 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3688 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3689 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, int8_t *, uint64_t *, uint64_t *, uint64_t *, GrB_Format, struct GB_Matrix_opaque *)
/* 3690 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3691 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3692 */ _CFFI_OP(_CFFI_OP_NOOP, 377),
/* 3693 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3694 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3695 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3696 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 3697 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3698 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3699 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, uint16_t *, uint64_t *, struct GB_Matrix_opaque *)
/* 3700 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3701 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3702 */ _CFFI_OP(_CFFI_OP_NOOP, 3386),
/* 3703 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3704 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3705 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3706 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, uint16_t *, uint64_t *, uint64_t *, uint64_t *, GrB_Format, struct GB_Matrix_opaque *)
/* 3707 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3708 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3709 */ _CFFI_OP(_CFFI_OP_NOOP, 3386),
/* 3710 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3711 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3712 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3713 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 3714 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3715 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3716 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, uint32_t *, uint64_t *, struct GB_Matrix_opaque *)
/* 3717 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3718 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3719 */ _CFFI_OP(_CFFI_OP_NOOP, 3415),
/* 3720 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3721 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3722 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3723 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, uint32_t *, uint64_t *, uint64_t *, uint64_t *, GrB_Format, struct GB_Matrix_opaque *)
/* 3724 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3725 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3726 */ _CFFI_OP(_CFFI_OP_NOOP, 3415),
/* 3727 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3728 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3729 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3730 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 3731 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3732 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3733 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, uint64_t *, GrB_Format, struct GB_Matrix_opaque *)
/* 3734 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3735 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3736 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3737 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 3738 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3739 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3740 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, uint64_t *, struct GB_Vector_opaque *)
/* 3741 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3742 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3743 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3744 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3745 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3746 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, uint64_t *, uint64_t *, struct GB_Matrix_opaque *)
/* 3747 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3748 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3749 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3750 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3751 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3752 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3753 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, uint64_t *, uint64_t *, uint64_t *, uint64_t *, GrB_Format, struct GB_Matrix_opaque *)
/* 3754 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3755 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3756 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3757 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3758 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3759 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3760 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 3761 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3762 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3763 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, uint8_t *, uint64_t *, struct GB_Matrix_opaque *)
/* 3764 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3765 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3766 */ _CFFI_OP(_CFFI_OP_POINTER, 1363), // uint8_t *
/* 3767 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3768 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3769 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3770 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, uint8_t *, uint64_t *, uint64_t *, uint64_t *, GrB_Format, struct GB_Matrix_opaque *)
/* 3771 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3772 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3773 */ _CFFI_OP(_CFFI_OP_NOOP, 3766),
/* 3774 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3775 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3776 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3777 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 3778 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3779 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3780 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, void *, uint64_t *, struct GB_Matrix_opaque *)
/* 3781 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3782 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3783 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 3784 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3785 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3786 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3787 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint64_t *, void *, uint64_t *, uint64_t *, uint64_t *, GrB_Format, struct GB_Matrix_opaque *)
/* 3788 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3789 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3790 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 3791 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3792 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3793 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3794 */ _CFFI_OP(_CFFI_OP_NOOP, 702),
/* 3795 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3796 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3797 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, uint8_t *, uint64_t *, struct GB_Vector_opaque *)
/* 3798 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3799 */ _CFFI_OP(_CFFI_OP_NOOP, 3766),
/* 3800 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3801 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3802 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3803 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint64_t *, void *, uint64_t *, struct GB_Vector_opaque *)
/* 3804 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3805 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 3806 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3807 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3808 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3809 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint8_t *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 3810 */ _CFFI_OP(_CFFI_OP_NOOP, 3766),
/* 3811 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3812 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 3813 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3814 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3815 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3816 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint8_t *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 3817 */ _CFFI_OP(_CFFI_OP_NOOP, 3766),
/* 3818 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3819 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 3820 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3821 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3822 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3823 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint8_t *, struct GB_Matrix_opaque *, uint64_t, uint64_t)
/* 3824 */ _CFFI_OP(_CFFI_OP_NOOP, 3766),
/* 3825 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3826 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3827 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3828 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3829 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint8_t *, struct GB_Scalar_opaque *)
/* 3830 */ _CFFI_OP(_CFFI_OP_NOOP, 3766),
/* 3831 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 3832 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3833 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(uint8_t *, struct GB_Vector_opaque *, uint64_t)
/* 3834 */ _CFFI_OP(_CFFI_OP_NOOP, 3766),
/* 3835 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3836 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3837 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3838 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(unsigned int *, unsigned int *)
/* 3839 */ _CFFI_OP(_CFFI_OP_POINTER, 3997), // unsigned int *
/* 3840 */ _CFFI_OP(_CFFI_OP_NOOP, 3839),
/* 3841 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3842 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(void * *, uint64_t *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 3843 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 3844 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3845 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3846 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3847 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3848 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(void * *, uint64_t *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 3849 */ _CFFI_OP(_CFFI_OP_NOOP, 48),
/* 3850 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3851 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3852 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3853 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3854 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(void *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Matrix_opaque *, struct GB_Descriptor_opaque *)
/* 3855 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 3856 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3857 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 3858 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3859 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3860 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3861 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(void *, struct GB_BinaryOp_opaque *, struct GB_Monoid_opaque *, struct GB_Vector_opaque *, struct GB_Descriptor_opaque *)
/* 3862 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 3863 */ _CFFI_OP(_CFFI_OP_NOOP, 56),
/* 3864 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 3865 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3866 */ _CFFI_OP(_CFFI_OP_NOOP, 2),
/* 3867 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3868 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(void *, struct GB_Matrix_opaque *, uint64_t, uint64_t)
/* 3869 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 3870 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3871 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3872 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3873 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3874 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(void *, struct GB_Monoid_opaque *)
/* 3875 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 3876 */ _CFFI_OP(_CFFI_OP_NOOP, 57),
/* 3877 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3878 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(void *, struct GB_Scalar_opaque *)
/* 3879 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 3880 */ _CFFI_OP(_CFFI_OP_NOOP, 80),
/* 3881 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3882 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(void *, struct GB_Vector_opaque *, uint64_t)
/* 3883 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 3884 */ _CFFI_OP(_CFFI_OP_NOOP, 65),
/* 3885 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3886 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3887 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(void *, uint64_t *, struct GB_Matrix_opaque *)
/* 3888 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 3889 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3890 */ _CFFI_OP(_CFFI_OP_NOOP, 7),
/* 3891 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3892 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3983), // GrB_Info()(void)
/* 3893 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3894 */ _CFFI_OP(_CFFI_OP_FUNCTION, 566), // _Bool()(struct GB_Iterator_opaque *)
/* 3895 */ _CFFI_OP(_CFFI_OP_NOOP, 541),
/* 3896 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3897 */ _CFFI_OP(_CFFI_OP_FUNCTION, 566), // _Bool()(uint64_t, uint64_t, void const *, void const *)
/* 3898 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3899 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3900 */ _CFFI_OP(_CFFI_OP_NOOP, 126),
/* 3901 */ _CFFI_OP(_CFFI_OP_NOOP, 126),
/* 3902 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3903 */ _CFFI_OP(_CFFI_OP_FUNCTION, 962), // _Dcomplex()(struct GB_Iterator_opaque *)
/* 3904 */ _CFFI_OP(_CFFI_OP_NOOP, 541),
/* 3905 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3906 */ _CFFI_OP(_CFFI_OP_FUNCTION, 32), // double()(struct GB_Iterator_opaque *)
/* 3907 */ _CFFI_OP(_CFFI_OP_NOOP, 541),
/* 3908 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3909 */ _CFFI_OP(_CFFI_OP_FUNCTION, 974), // _Fcomplex()(struct GB_Iterator_opaque *)
/* 3910 */ _CFFI_OP(_CFFI_OP_NOOP, 541),
/* 3911 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3912 */ _CFFI_OP(_CFFI_OP_FUNCTION, 980), // float()(struct GB_Iterator_opaque *)
/* 3913 */ _CFFI_OP(_CFFI_OP_NOOP, 541),
/* 3914 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3915 */ _CFFI_OP(_CFFI_OP_FUNCTION, 986), // int16_t()(struct GB_Iterator_opaque *)
/* 3916 */ _CFFI_OP(_CFFI_OP_NOOP, 541),
/* 3917 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3918 */ _CFFI_OP(_CFFI_OP_FUNCTION, 40), // int32_t()(struct GB_Iterator_opaque *)
/* 3919 */ _CFFI_OP(_CFFI_OP_NOOP, 541),
/* 3920 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3921 */ _CFFI_OP(_CFFI_OP_FUNCTION, 899), // int64_t()(struct GB_Iterator_opaque *)
/* 3922 */ _CFFI_OP(_CFFI_OP_NOOP, 541),
/* 3923 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3924 */ _CFFI_OP(_CFFI_OP_FUNCTION, 1024), // int8_t()(struct GB_Iterator_opaque *)
/* 3925 */ _CFFI_OP(_CFFI_OP_NOOP, 541),
/* 3926 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3927 */ _CFFI_OP(_CFFI_OP_FUNCTION, 1336), // uint16_t()(struct GB_Iterator_opaque *)
/* 3928 */ _CFFI_OP(_CFFI_OP_NOOP, 541),
/* 3929 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3930 */ _CFFI_OP(_CFFI_OP_FUNCTION, 1345), // uint32_t()(struct GB_Iterator_opaque *)
/* 3931 */ _CFFI_OP(_CFFI_OP_NOOP, 541),
/* 3932 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3933 */ _CFFI_OP(_CFFI_OP_FUNCTION, 75), // uint64_t()(struct GB_Iterator_opaque *)
/* 3934 */ _CFFI_OP(_CFFI_OP_NOOP, 541),
/* 3935 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3936 */ _CFFI_OP(_CFFI_OP_FUNCTION, 1363), // uint8_t()(struct GB_Iterator_opaque *)
/* 3937 */ _CFFI_OP(_CFFI_OP_NOOP, 541),
/* 3938 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3939 */ _CFFI_OP(_CFFI_OP_FUNCTION, 52), // void *()(size_t)
/* 3940 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 28),
/* 3941 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3942 */ _CFFI_OP(_CFFI_OP_FUNCTION, 52), // void *()(size_t, size_t)
/* 3943 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 28),
/* 3944 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 28),
/* 3945 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3946 */ _CFFI_OP(_CFFI_OP_FUNCTION, 52), // void *()(void *, size_t)
/* 3947 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 3948 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 28),
/* 3949 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3950 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3998), // void()(struct GB_Iterator_opaque *, uint64_t *, uint64_t *)
/* 3951 */ _CFFI_OP(_CFFI_OP_NOOP, 541),
/* 3952 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3953 */ _CFFI_OP(_CFFI_OP_NOOP, 586),
/* 3954 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3955 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3998), // void()(struct GB_Iterator_opaque *, void *)
/* 3956 */ _CFFI_OP(_CFFI_OP_NOOP, 541),
/* 3957 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 3958 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3959 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3998), // void()(void *)
/* 3960 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 3961 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3962 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3998), // void()(void *, void const *)
/* 3963 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 3964 */ _CFFI_OP(_CFFI_OP_NOOP, 126),
/* 3965 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3966 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3998), // void()(void *, void const *, uint64_t, uint64_t, void const *)
/* 3967 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 3968 */ _CFFI_OP(_CFFI_OP_NOOP, 126),
/* 3969 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3970 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 24),
/* 3971 */ _CFFI_OP(_CFFI_OP_NOOP, 126),
/* 3972 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3973 */ _CFFI_OP(_CFFI_OP_FUNCTION, 3998), // void()(void *, void const *, void const *)
/* 3974 */ _CFFI_OP(_CFFI_OP_NOOP, 52),
/* 3975 */ _CFFI_OP(_CFFI_OP_NOOP, 126),
/* 3976 */ _CFFI_OP(_CFFI_OP_NOOP, 126),
/* 3977 */ _CFFI_OP(_CFFI_OP_FUNCTION_END, 0),
/* 3978 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 12), // FILE
/* 3979 */ _CFFI_OP(_CFFI_OP_POINTER, 19), // GrB_Info(*)(GxB_Option_Field, ...)
/* 3980 */ _CFFI_OP(_CFFI_OP_POINTER, 469), // GrB_Info(*)(struct GB_Descriptor_opaque *, GrB_Desc_Field, ...)
/* 3981 */ _CFFI_OP(_CFFI_OP_POINTER, 917), // GrB_Info(*)(struct GB_Matrix_opaque *, GxB_Option_Field, ...)
/* 3982 */ _CFFI_OP(_CFFI_OP_POINTER, 2455), // GrB_Info(*)(struct GB_Vector_opaque *, GxB_Option_Field, ...)
/* 3983 */ _CFFI_OP(_CFFI_OP_ENUM, 3), // GrB_Info
/* 3984 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 2), // char
/* 3985 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 0), // struct GB_BinaryOp_opaque
/* 3986 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 1), // struct GB_Descriptor_opaque
/* 3987 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 2), // struct GB_IndexUnaryOp_opaque
/* 3988 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 3), // struct GB_Iterator_opaque
/* 3989 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 4), // struct GB_Matrix_opaque
/* 3990 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 5), // struct GB_Monoid_opaque
/* 3991 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 6), // struct GB_Scalar_opaque
/* 3992 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 7), // struct GB_SelectOp_opaque
/* 3993 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 8), // struct GB_Semiring_opaque
/* 3994 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 9), // struct GB_Type_opaque
/* 3995 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 10), // struct GB_UnaryOp_opaque
/* 3996 */ _CFFI_OP(_CFFI_OP_STRUCT_UNION, 11), // struct GB_Vector_opaque
/* 3997 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 8), // unsigned int
/* 3998 */ _CFFI_OP(_CFFI_OP_PRIMITIVE, 0), // void
};

static int _cffi_const_GrB_OUTP(unsigned long long *o)
{
  int n = (GrB_OUTP) <= 0;
  *o = (unsigned long long)((GrB_OUTP) | 0);  /* check that GrB_OUTP is an integer */
  return n;
}

static int _cffi_const_GrB_MASK(unsigned long long *o)
{
  int n = (GrB_MASK) <= 0;
  *o = (unsigned long long)((GrB_MASK) | 0);  /* check that GrB_MASK is an integer */
  return n;
}

static int _cffi_const_GrB_INP0(unsigned long long *o)
{
  int n = (GrB_INP0) <= 0;
  *o = (unsigned long long)((GrB_INP0) | 0);  /* check that GrB_INP0 is an integer */
  return n;
}

static int _cffi_const_GrB_INP1(unsigned long long *o)
{
  int n = (GrB_INP1) <= 0;
  *o = (unsigned long long)((GrB_INP1) | 0);  /* check that GrB_INP1 is an integer */
  return n;
}

static int _cffi_const_GxB_DESCRIPTOR_NTHREADS(unsigned long long *o)
{
  int n = (GxB_DESCRIPTOR_NTHREADS) <= 0;
  *o = (unsigned long long)((GxB_DESCRIPTOR_NTHREADS) | 0);  /* check that GxB_DESCRIPTOR_NTHREADS is an integer */
  return n;
}

static int _cffi_const_GxB_DESCRIPTOR_CHUNK(unsigned long long *o)
{
  int n = (GxB_DESCRIPTOR_CHUNK) <= 0;
  *o = (unsigned long long)((GxB_DESCRIPTOR_CHUNK) | 0);  /* check that GxB_DESCRIPTOR_CHUNK is an integer */
  return n;
}

static int _cffi_const_GxB_DESCRIPTOR_GPU_CONTROL(unsigned long long *o)
{
  int n = (GxB_DESCRIPTOR_GPU_CONTROL) <= 0;
  *o = (unsigned long long)((GxB_DESCRIPTOR_GPU_CONTROL) | 0);  /* check that GxB_DESCRIPTOR_GPU_CONTROL is an integer */
  return n;
}

static int _cffi_const_GxB_DESCRIPTOR_GPU_CHUNK(unsigned long long *o)
{
  int n = (GxB_DESCRIPTOR_GPU_CHUNK) <= 0;
  *o = (unsigned long long)((GxB_DESCRIPTOR_GPU_CHUNK) | 0);  /* check that GxB_DESCRIPTOR_GPU_CHUNK is an integer */
  return n;
}

static int _cffi_const_GxB_AxB_METHOD(unsigned long long *o)
{
  int n = (GxB_AxB_METHOD) <= 0;
  *o = (unsigned long long)((GxB_AxB_METHOD) | 0);  /* check that GxB_AxB_METHOD is an integer */
  return n;
}

static int _cffi_const_GxB_SORT(unsigned long long *o)
{
  int n = (GxB_SORT) <= 0;
  *o = (unsigned long long)((GxB_SORT) | 0);  /* check that GxB_SORT is an integer */
  return n;
}

static int _cffi_const_GxB_COMPRESSION(unsigned long long *o)
{
  int n = (GxB_COMPRESSION) <= 0;
  *o = (unsigned long long)((GxB_COMPRESSION) | 0);  /* check that GxB_COMPRESSION is an integer */
  return n;
}

static int _cffi_const_GxB_IMPORT(unsigned long long *o)
{
  int n = (GxB_IMPORT) <= 0;
  *o = (unsigned long long)((GxB_IMPORT) | 0);  /* check that GxB_IMPORT is an integer */
  return n;
}

static int _cffi_const_GxB_DEFAULT(unsigned long long *o)
{
  int n = (GxB_DEFAULT) <= 0;
  *o = (unsigned long long)((GxB_DEFAULT) | 0);  /* check that GxB_DEFAULT is an integer */
  return n;
}

static int _cffi_const_GrB_REPLACE(unsigned long long *o)
{
  int n = (GrB_REPLACE) <= 0;
  *o = (unsigned long long)((GrB_REPLACE) | 0);  /* check that GrB_REPLACE is an integer */
  return n;
}

static int _cffi_const_GrB_COMP(unsigned long long *o)
{
  int n = (GrB_COMP) <= 0;
  *o = (unsigned long long)((GrB_COMP) | 0);  /* check that GrB_COMP is an integer */
  return n;
}

static int _cffi_const_GrB_STRUCTURE(unsigned long long *o)
{
  int n = (GrB_STRUCTURE) <= 0;
  *o = (unsigned long long)((GrB_STRUCTURE) | 0);  /* check that GrB_STRUCTURE is an integer */
  return n;
}

static int _cffi_const_GrB_TRAN(unsigned long long *o)
{
  int n = (GrB_TRAN) <= 0;
  *o = (unsigned long long)((GrB_TRAN) | 0);  /* check that GrB_TRAN is an integer */
  return n;
}

static int _cffi_const_GxB_GPU_ALWAYS(unsigned long long *o)
{
  int n = (GxB_GPU_ALWAYS) <= 0;
  *o = (unsigned long long)((GxB_GPU_ALWAYS) | 0);  /* check that GxB_GPU_ALWAYS is an integer */
  return n;
}

static int _cffi_const_GxB_GPU_NEVER(unsigned long long *o)
{
  int n = (GxB_GPU_NEVER) <= 0;
  *o = (unsigned long long)((GxB_GPU_NEVER) | 0);  /* check that GxB_GPU_NEVER is an integer */
  return n;
}

static int _cffi_const_GxB_AxB_GUSTAVSON(unsigned long long *o)
{
  int n = (GxB_AxB_GUSTAVSON) <= 0;
  *o = (unsigned long long)((GxB_AxB_GUSTAVSON) | 0);  /* check that GxB_AxB_GUSTAVSON is an integer */
  return n;
}

static int _cffi_const_GxB_AxB_DOT(unsigned long long *o)
{
  int n = (GxB_AxB_DOT) <= 0;
  *o = (unsigned long long)((GxB_AxB_DOT) | 0);  /* check that GxB_AxB_DOT is an integer */
  return n;
}

static int _cffi_const_GxB_AxB_HASH(unsigned long long *o)
{
  int n = (GxB_AxB_HASH) <= 0;
  *o = (unsigned long long)((GxB_AxB_HASH) | 0);  /* check that GxB_AxB_HASH is an integer */
  return n;
}

static int _cffi_const_GxB_AxB_SAXPY(unsigned long long *o)
{
  int n = (GxB_AxB_SAXPY) <= 0;
  *o = (unsigned long long)((GxB_AxB_SAXPY) | 0);  /* check that GxB_AxB_SAXPY is an integer */
  return n;
}

static int _cffi_const_GxB_SECURE_IMPORT(unsigned long long *o)
{
  int n = (GxB_SECURE_IMPORT) <= 0;
  *o = (unsigned long long)((GxB_SECURE_IMPORT) | 0);  /* check that GxB_SECURE_IMPORT is an integer */
  return n;
}

static int _cffi_const_GrB_CSR_FORMAT(unsigned long long *o)
{
  int n = (GrB_CSR_FORMAT) <= 0;
  *o = (unsigned long long)((GrB_CSR_FORMAT) | 0);  /* check that GrB_CSR_FORMAT is an integer */
  return n;
}

static int _cffi_const_GrB_CSC_FORMAT(unsigned long long *o)
{
  int n = (GrB_CSC_FORMAT) <= 0;
  *o = (unsigned long long)((GrB_CSC_FORMAT) | 0);  /* check that GrB_CSC_FORMAT is an integer */
  return n;
}

static int _cffi_const_GrB_COO_FORMAT(unsigned long long *o)
{
  int n = (GrB_COO_FORMAT) <= 0;
  *o = (unsigned long long)((GrB_COO_FORMAT) | 0);  /* check that GrB_COO_FORMAT is an integer */
  return n;
}

static int _cffi_const_GrB_SUCCESS(unsigned long long *o)
{
  int n = (GrB_SUCCESS) <= 0;
  *o = (unsigned long long)((GrB_SUCCESS) | 0);  /* check that GrB_SUCCESS is an integer */
  return n;
}

static int _cffi_const_GrB_NO_VALUE(unsigned long long *o)
{
  int n = (GrB_NO_VALUE) <= 0;
  *o = (unsigned long long)((GrB_NO_VALUE) | 0);  /* check that GrB_NO_VALUE is an integer */
  return n;
}

static int _cffi_const_GxB_EXHAUSTED(unsigned long long *o)
{
  int n = (GxB_EXHAUSTED) <= 0;
  *o = (unsigned long long)((GxB_EXHAUSTED) | 0);  /* check that GxB_EXHAUSTED is an integer */
  return n;
}

static int _cffi_const_GrB_UNINITIALIZED_OBJECT(unsigned long long *o)
{
  int n = (GrB_UNINITIALIZED_OBJECT) <= 0;
  *o = (unsigned long long)((GrB_UNINITIALIZED_OBJECT) | 0);  /* check that GrB_UNINITIALIZED_OBJECT is an integer */
  return n;
}

static int _cffi_const_GrB_NULL_POINTER(unsigned long long *o)
{
  int n = (GrB_NULL_POINTER) <= 0;
  *o = (unsigned long long)((GrB_NULL_POINTER) | 0);  /* check that GrB_NULL_POINTER is an integer */
  return n;
}

static int _cffi_const_GrB_INVALID_VALUE(unsigned long long *o)
{
  int n = (GrB_INVALID_VALUE) <= 0;
  *o = (unsigned long long)((GrB_INVALID_VALUE) | 0);  /* check that GrB_INVALID_VALUE is an integer */
  return n;
}

static int _cffi_const_GrB_INVALID_INDEX(unsigned long long *o)
{
  int n = (GrB_INVALID_INDEX) <= 0;
  *o = (unsigned long long)((GrB_INVALID_INDEX) | 0);  /* check that GrB_INVALID_INDEX is an integer */
  return n;
}

static int _cffi_const_GrB_DOMAIN_MISMATCH(unsigned long long *o)
{
  int n = (GrB_DOMAIN_MISMATCH) <= 0;
  *o = (unsigned long long)((GrB_DOMAIN_MISMATCH) | 0);  /* check that GrB_DOMAIN_MISMATCH is an integer */
  return n;
}

static int _cffi_const_GrB_DIMENSION_MISMATCH(unsigned long long *o)
{
  int n = (GrB_DIMENSION_MISMATCH) <= 0;
  *o = (unsigned long long)((GrB_DIMENSION_MISMATCH) | 0);  /* check that GrB_DIMENSION_MISMATCH is an integer */
  return n;
}

static int _cffi_const_GrB_OUTPUT_NOT_EMPTY(unsigned long long *o)
{
  int n = (GrB_OUTPUT_NOT_EMPTY) <= 0;
  *o = (unsigned long long)((GrB_OUTPUT_NOT_EMPTY) | 0);  /* check that GrB_OUTPUT_NOT_EMPTY is an integer */
  return n;
}

static int _cffi_const_GrB_NOT_IMPLEMENTED(unsigned long long *o)
{
  int n = (GrB_NOT_IMPLEMENTED) <= 0;
  *o = (unsigned long long)((GrB_NOT_IMPLEMENTED) | 0);  /* check that GrB_NOT_IMPLEMENTED is an integer */
  return n;
}

static int _cffi_const_GrB_PANIC(unsigned long long *o)
{
  int n = (GrB_PANIC) <= 0;
  *o = (unsigned long long)((GrB_PANIC) | 0);  /* check that GrB_PANIC is an integer */
  return n;
}

static int _cffi_const_GrB_OUT_OF_MEMORY(unsigned long long *o)
{
  int n = (GrB_OUT_OF_MEMORY) <= 0;
  *o = (unsigned long long)((GrB_OUT_OF_MEMORY) | 0);  /* check that GrB_OUT_OF_MEMORY is an integer */
  return n;
}

static int _cffi_const_GrB_INSUFFICIENT_SPACE(unsigned long long *o)
{
  int n = (GrB_INSUFFICIENT_SPACE) <= 0;
  *o = (unsigned long long)((GrB_INSUFFICIENT_SPACE) | 0);  /* check that GrB_INSUFFICIENT_SPACE is an integer */
  return n;
}

static int _cffi_const_GrB_INVALID_OBJECT(unsigned long long *o)
{
  int n = (GrB_INVALID_OBJECT) <= 0;
  *o = (unsigned long long)((GrB_INVALID_OBJECT) | 0);  /* check that GrB_INVALID_OBJECT is an integer */
  return n;
}

static int _cffi_const_GrB_INDEX_OUT_OF_BOUNDS(unsigned long long *o)
{
  int n = (GrB_INDEX_OUT_OF_BOUNDS) <= 0;
  *o = (unsigned long long)((GrB_INDEX_OUT_OF_BOUNDS) | 0);  /* check that GrB_INDEX_OUT_OF_BOUNDS is an integer */
  return n;
}

static int _cffi_const_GrB_EMPTY_OBJECT(unsigned long long *o)
{
  int n = (GrB_EMPTY_OBJECT) <= 0;
  *o = (unsigned long long)((GrB_EMPTY_OBJECT) | 0);  /* check that GrB_EMPTY_OBJECT is an integer */
  return n;
}

static int _cffi_const_GrB_NONBLOCKING(unsigned long long *o)
{
  int n = (GrB_NONBLOCKING) <= 0;
  *o = (unsigned long long)((GrB_NONBLOCKING) | 0);  /* check that GrB_NONBLOCKING is an integer */
  return n;
}

static int _cffi_const_GrB_BLOCKING(unsigned long long *o)
{
  int n = (GrB_BLOCKING) <= 0;
  *o = (unsigned long long)((GrB_BLOCKING) | 0);  /* check that GrB_BLOCKING is an integer */
  return n;
}

static int _cffi_const_GxB_NONBLOCKING_GPU(unsigned long long *o)
{
  int n = (GxB_NONBLOCKING_GPU) <= 0;
  *o = (unsigned long long)((GxB_NONBLOCKING_GPU) | 0);  /* check that GxB_NONBLOCKING_GPU is an integer */
  return n;
}

static int _cffi_const_GxB_BLOCKING_GPU(unsigned long long *o)
{
  int n = (GxB_BLOCKING_GPU) <= 0;
  *o = (unsigned long long)((GxB_BLOCKING_GPU) | 0);  /* check that GxB_BLOCKING_GPU is an integer */
  return n;
}

static int _cffi_const_GrB_COMPLETE(unsigned long long *o)
{
  int n = (GrB_COMPLETE) <= 0;
  *o = (unsigned long long)((GrB_COMPLETE) | 0);  /* check that GrB_COMPLETE is an integer */
  return n;
}

static int _cffi_const_GrB_MATERIALIZE(unsigned long long *o)
{
  int n = (GrB_MATERIALIZE) <= 0;
  *o = (unsigned long long)((GrB_MATERIALIZE) | 0);  /* check that GrB_MATERIALIZE is an integer */
  return n;
}

static int _cffi_const_GxB_BY_ROW(unsigned long long *o)
{
  int n = (GxB_BY_ROW) <= 0;
  *o = (unsigned long long)((GxB_BY_ROW) | 0);  /* check that GxB_BY_ROW is an integer */
  return n;
}

static int _cffi_const_GxB_BY_COL(unsigned long long *o)
{
  int n = (GxB_BY_COL) <= 0;
  *o = (unsigned long long)((GxB_BY_COL) | 0);  /* check that GxB_BY_COL is an integer */
  return n;
}

static int _cffi_const_GxB_NO_FORMAT(unsigned long long *o)
{
  int n = (GxB_NO_FORMAT) <= 0;
  *o = (unsigned long long)((GxB_NO_FORMAT) | 0);  /* check that GxB_NO_FORMAT is an integer */
  return n;
}

static int _cffi_const_GxB_HYPER_SWITCH(unsigned long long *o)
{
  int n = (GxB_HYPER_SWITCH) <= 0;
  *o = (unsigned long long)((GxB_HYPER_SWITCH) | 0);  /* check that GxB_HYPER_SWITCH is an integer */
  return n;
}

static int _cffi_const_GxB_BITMAP_SWITCH(unsigned long long *o)
{
  int n = (GxB_BITMAP_SWITCH) <= 0;
  *o = (unsigned long long)((GxB_BITMAP_SWITCH) | 0);  /* check that GxB_BITMAP_SWITCH is an integer */
  return n;
}

static int _cffi_const_GxB_FORMAT(unsigned long long *o)
{
  int n = (GxB_FORMAT) <= 0;
  *o = (unsigned long long)((GxB_FORMAT) | 0);  /* check that GxB_FORMAT is an integer */
  return n;
}

static int _cffi_const_GxB_MODE(unsigned long long *o)
{
  int n = (GxB_MODE) <= 0;
  *o = (unsigned long long)((GxB_MODE) | 0);  /* check that GxB_MODE is an integer */
  return n;
}

static int _cffi_const_GxB_LIBRARY_NAME(unsigned long long *o)
{
  int n = (GxB_LIBRARY_NAME) <= 0;
  *o = (unsigned long long)((GxB_LIBRARY_NAME) | 0);  /* check that GxB_LIBRARY_NAME is an integer */
  return n;
}

static int _cffi_const_GxB_LIBRARY_VERSION(unsigned long long *o)
{
  int n = (GxB_LIBRARY_VERSION) <= 0;
  *o = (unsigned long long)((GxB_LIBRARY_VERSION) | 0);  /* check that GxB_LIBRARY_VERSION is an integer */
  return n;
}

static int _cffi_const_GxB_LIBRARY_DATE(unsigned long long *o)
{
  int n = (GxB_LIBRARY_DATE) <= 0;
  *o = (unsigned long long)((GxB_LIBRARY_DATE) | 0);  /* check that GxB_LIBRARY_DATE is an integer */
  return n;
}

static int _cffi_const_GxB_LIBRARY_ABOUT(unsigned long long *o)
{
  int n = (GxB_LIBRARY_ABOUT) <= 0;
  *o = (unsigned long long)((GxB_LIBRARY_ABOUT) | 0);  /* check that GxB_LIBRARY_ABOUT is an integer */
  return n;
}

static int _cffi_const_GxB_LIBRARY_URL(unsigned long long *o)
{
  int n = (GxB_LIBRARY_URL) <= 0;
  *o = (unsigned long long)((GxB_LIBRARY_URL) | 0);  /* check that GxB_LIBRARY_URL is an integer */
  return n;
}

static int _cffi_const_GxB_LIBRARY_LICENSE(unsigned long long *o)
{
  int n = (GxB_LIBRARY_LICENSE) <= 0;
  *o = (unsigned long long)((GxB_LIBRARY_LICENSE) | 0);  /* check that GxB_LIBRARY_LICENSE is an integer */
  return n;
}

static int _cffi_const_GxB_LIBRARY_COMPILE_DATE(unsigned long long *o)
{
  int n = (GxB_LIBRARY_COMPILE_DATE) <= 0;
  *o = (unsigned long long)((GxB_LIBRARY_COMPILE_DATE) | 0);  /* check that GxB_LIBRARY_COMPILE_DATE is an integer */
  return n;
}

static int _cffi_const_GxB_LIBRARY_COMPILE_TIME(unsigned long long *o)
{
  int n = (GxB_LIBRARY_COMPILE_TIME) <= 0;
  *o = (unsigned long long)((GxB_LIBRARY_COMPILE_TIME) | 0);  /* check that GxB_LIBRARY_COMPILE_TIME is an integer */
  return n;
}

static int _cffi_const_GxB_API_VERSION(unsigned long long *o)
{
  int n = (GxB_API_VERSION) <= 0;
  *o = (unsigned long long)((GxB_API_VERSION) | 0);  /* check that GxB_API_VERSION is an integer */
  return n;
}

static int _cffi_const_GxB_API_DATE(unsigned long long *o)
{
  int n = (GxB_API_DATE) <= 0;
  *o = (unsigned long long)((GxB_API_DATE) | 0);  /* check that GxB_API_DATE is an integer */
  return n;
}

static int _cffi_const_GxB_API_ABOUT(unsigned long long *o)
{
  int n = (GxB_API_ABOUT) <= 0;
  *o = (unsigned long long)((GxB_API_ABOUT) | 0);  /* check that GxB_API_ABOUT is an integer */
  return n;
}

static int _cffi_const_GxB_API_URL(unsigned long long *o)
{
  int n = (GxB_API_URL) <= 0;
  *o = (unsigned long long)((GxB_API_URL) | 0);  /* check that GxB_API_URL is an integer */
  return n;
}

static int _cffi_const_GxB_COMPILER_VERSION(unsigned long long *o)
{
  int n = (GxB_COMPILER_VERSION) <= 0;
  *o = (unsigned long long)((GxB_COMPILER_VERSION) | 0);  /* check that GxB_COMPILER_VERSION is an integer */
  return n;
}

static int _cffi_const_GxB_COMPILER_NAME(unsigned long long *o)
{
  int n = (GxB_COMPILER_NAME) <= 0;
  *o = (unsigned long long)((GxB_COMPILER_NAME) | 0);  /* check that GxB_COMPILER_NAME is an integer */
  return n;
}

static int _cffi_const_GxB_LIBRARY_OPENMP(unsigned long long *o)
{
  int n = (GxB_LIBRARY_OPENMP) <= 0;
  *o = (unsigned long long)((GxB_LIBRARY_OPENMP) | 0);  /* check that GxB_LIBRARY_OPENMP is an integer */
  return n;
}

static int _cffi_const_GxB_GLOBAL_NTHREADS(unsigned long long *o)
{
  int n = (GxB_GLOBAL_NTHREADS) <= 0;
  *o = (unsigned long long)((GxB_GLOBAL_NTHREADS) | 0);  /* check that GxB_GLOBAL_NTHREADS is an integer */
  return n;
}

static int _cffi_const_GxB_GLOBAL_CHUNK(unsigned long long *o)
{
  int n = (GxB_GLOBAL_CHUNK) <= 0;
  *o = (unsigned long long)((GxB_GLOBAL_CHUNK) | 0);  /* check that GxB_GLOBAL_CHUNK is an integer */
  return n;
}

static int _cffi_const_GxB_BURBLE(unsigned long long *o)
{
  int n = (GxB_BURBLE) <= 0;
  *o = (unsigned long long)((GxB_BURBLE) | 0);  /* check that GxB_BURBLE is an integer */
  return n;
}

static int _cffi_const_GxB_PRINTF(unsigned long long *o)
{
  int n = (GxB_PRINTF) <= 0;
  *o = (unsigned long long)((GxB_PRINTF) | 0);  /* check that GxB_PRINTF is an integer */
  return n;
}

static int _cffi_const_GxB_FLUSH(unsigned long long *o)
{
  int n = (GxB_FLUSH) <= 0;
  *o = (unsigned long long)((GxB_FLUSH) | 0);  /* check that GxB_FLUSH is an integer */
  return n;
}

static int _cffi_const_GxB_MEMORY_POOL(unsigned long long *o)
{
  int n = (GxB_MEMORY_POOL) <= 0;
  *o = (unsigned long long)((GxB_MEMORY_POOL) | 0);  /* check that GxB_MEMORY_POOL is an integer */
  return n;
}

static int _cffi_const_GxB_PRINT_1BASED(unsigned long long *o)
{
  int n = (GxB_PRINT_1BASED) <= 0;
  *o = (unsigned long long)((GxB_PRINT_1BASED) | 0);  /* check that GxB_PRINT_1BASED is an integer */
  return n;
}

static int _cffi_const_GxB_SPARSITY_STATUS(unsigned long long *o)
{
  int n = (GxB_SPARSITY_STATUS) <= 0;
  *o = (unsigned long long)((GxB_SPARSITY_STATUS) | 0);  /* check that GxB_SPARSITY_STATUS is an integer */
  return n;
}

static int _cffi_const_GxB_SPARSITY_CONTROL(unsigned long long *o)
{
  int n = (GxB_SPARSITY_CONTROL) <= 0;
  *o = (unsigned long long)((GxB_SPARSITY_CONTROL) | 0);  /* check that GxB_SPARSITY_CONTROL is an integer */
  return n;
}

static int _cffi_const_GxB_GLOBAL_GPU_CONTROL(unsigned long long *o)
{
  int n = (GxB_GLOBAL_GPU_CONTROL) <= 0;
  *o = (unsigned long long)((GxB_GLOBAL_GPU_CONTROL) | 0);  /* check that GxB_GLOBAL_GPU_CONTROL is an integer */
  return n;
}

static int _cffi_const_GxB_GLOBAL_GPU_CHUNK(unsigned long long *o)
{
  int n = (GxB_GLOBAL_GPU_CHUNK) <= 0;
  *o = (unsigned long long)((GxB_GLOBAL_GPU_CHUNK) | 0);  /* check that GxB_GLOBAL_GPU_CHUNK is an integer */
  return n;
}

static int _cffi_const_GxB_SILENT(unsigned long long *o)
{
  int n = (GxB_SILENT) <= 0;
  *o = (unsigned long long)((GxB_SILENT) | 0);  /* check that GxB_SILENT is an integer */
  return n;
}

static int _cffi_const_GxB_SUMMARY(unsigned long long *o)
{
  int n = (GxB_SUMMARY) <= 0;
  *o = (unsigned long long)((GxB_SUMMARY) | 0);  /* check that GxB_SUMMARY is an integer */
  return n;
}

static int _cffi_const_GxB_SHORT(unsigned long long *o)
{
  int n = (GxB_SHORT) <= 0;
  *o = (unsigned long long)((GxB_SHORT) | 0);  /* check that GxB_SHORT is an integer */
  return n;
}

static int _cffi_const_GxB_COMPLETE(unsigned long long *o)
{
  int n = (GxB_COMPLETE) <= 0;
  *o = (unsigned long long)((GxB_COMPLETE) | 0);  /* check that GxB_COMPLETE is an integer */
  return n;
}

static int _cffi_const_GxB_SHORT_VERBOSE(unsigned long long *o)
{
  int n = (GxB_SHORT_VERBOSE) <= 0;
  *o = (unsigned long long)((GxB_SHORT_VERBOSE) | 0);  /* check that GxB_SHORT_VERBOSE is an integer */
  return n;
}

static int _cffi_const_GxB_COMPLETE_VERBOSE(unsigned long long *o)
{
  int n = (GxB_COMPLETE_VERBOSE) <= 0;
  *o = (unsigned long long)((GxB_COMPLETE_VERBOSE) | 0);  /* check that GxB_COMPLETE_VERBOSE is an integer */
  return n;
}

static void _cffi_const_GxB_ALWAYS_HYPER(char *o)
{
  *(double *)o = GxB_ALWAYS_HYPER;
}

static void _cffi_const_GxB_FORMAT_DEFAULT(char *o)
{
  *(GxB_Format_Value *)o = GxB_FORMAT_DEFAULT;
}

static void _cffi_const_GxB_HYPER_DEFAULT(char *o)
{
  *(double *)o = GxB_HYPER_DEFAULT;
}

static void _cffi_const_GxB_NEVER_HYPER(char *o)
{
  *(double *)o = GxB_NEVER_HYPER;
}

static GrB_Info _cffi_d_GB_BinaryOp_new(struct GB_BinaryOp_opaque * * x0, void(* x1)(void *, void const *, void const *), struct GB_Type_opaque * x2, struct GB_Type_opaque * x3, struct GB_Type_opaque * x4, char const * x5)
{
  return GB_BinaryOp_new(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GB_BinaryOp_new(PyObject *self, PyObject *args)
{
  struct GB_BinaryOp_opaque * * x0;
  void(* x1)(void *, void const *, void const *);
  struct GB_Type_opaque * x2;
  struct GB_Type_opaque * x3;
  struct GB_Type_opaque * x4;
  char const * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "GB_BinaryOp_new", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(422), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(422), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = (void(*)(void *, void const *, void const *))_cffi_to_c_pointer(arg1, _cffi_type(434));
  if (x1 == (void(*)(void *, void const *, void const *))NULL && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GB_BinaryOp_new(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GB_BinaryOp_new _cffi_d_GB_BinaryOp_new
#endif

static GrB_Info _cffi_d_GB_Iterator_attach(struct GB_Iterator_opaque * x0, struct GB_Matrix_opaque * x1, GxB_Format_Value x2, struct GB_Descriptor_opaque * x3)
{
  return GB_Iterator_attach(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GB_Iterator_attach(PyObject *self, PyObject *args)
{
  struct GB_Iterator_opaque * x0;
  struct GB_Matrix_opaque * x1;
  GxB_Format_Value x2;
  struct GB_Descriptor_opaque * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GB_Iterator_attach", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x2, _cffi_type(546), arg2) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GB_Iterator_attach(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GB_Iterator_attach _cffi_d_GB_Iterator_attach
#endif

static GrB_Info _cffi_d_GB_Iterator_rc_bitmap_next(struct GB_Iterator_opaque * x0)
{
  return GB_Iterator_rc_bitmap_next(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GB_Iterator_rc_bitmap_next(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GB_Iterator_rc_bitmap_next(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GB_Iterator_rc_bitmap_next _cffi_d_GB_Iterator_rc_bitmap_next
#endif

static GrB_Info _cffi_d_GB_Iterator_rc_seek(struct GB_Iterator_opaque * x0, uint64_t x1, _Bool x2)
{
  return GB_Iterator_rc_seek(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GB_Iterator_rc_seek(PyObject *self, PyObject *args)
{
  struct GB_Iterator_opaque * x0;
  uint64_t x1;
  _Bool x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GB_Iterator_rc_seek", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint64_t);
  if (x1 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x2 = (_Bool)_cffi_to_c__Bool(arg2);
  if (x2 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GB_Iterator_rc_seek(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GB_Iterator_rc_seek _cffi_d_GB_Iterator_rc_seek
#endif

static GrB_Info _cffi_d_GB_SelectOp_new(struct GB_SelectOp_opaque * * x0, _Bool(* x1)(uint64_t, uint64_t, void const *, void const *), struct GB_Type_opaque * x2, struct GB_Type_opaque * x3, char const * x4)
{
  return GB_SelectOp_new(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GB_SelectOp_new(PyObject *self, PyObject *args)
{
  struct GB_SelectOp_opaque * * x0;
  _Bool(* x1)(uint64_t, uint64_t, void const *, void const *);
  struct GB_Type_opaque * x2;
  struct GB_Type_opaque * x3;
  char const * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GB_SelectOp_new", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2228), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_SelectOp_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2228), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = (_Bool(*)(uint64_t, uint64_t, void const *, void const *))_cffi_to_c_pointer(arg1, _cffi_type(2232));
  if (x1 == (_Bool(*)(uint64_t, uint64_t, void const *, void const *))NULL && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GB_SelectOp_new(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GB_SelectOp_new _cffi_d_GB_SelectOp_new
#endif

static GrB_Info _cffi_d_GB_Type_new(struct GB_Type_opaque * * x0, size_t x1, char const * x2)
{
  return GB_Type_new(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GB_Type_new(PyObject *self, PyObject *args)
{
  struct GB_Type_opaque * * x0;
  size_t x1;
  char const * x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GB_Type_new", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, size_t);
  if (x1 == (size_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GB_Type_new(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GB_Type_new _cffi_d_GB_Type_new
#endif

static GrB_Info _cffi_d_GB_UnaryOp_new(struct GB_UnaryOp_opaque * * x0, void(* x1)(void *, void const *), struct GB_Type_opaque * x2, struct GB_Type_opaque * x3, char const * x4)
{
  return GB_UnaryOp_new(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GB_UnaryOp_new(PyObject *self, PyObject *args)
{
  struct GB_UnaryOp_opaque * * x0;
  void(* x1)(void *, void const *);
  struct GB_Type_opaque * x2;
  struct GB_Type_opaque * x3;
  char const * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GB_UnaryOp_new", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2328), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_UnaryOp_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2328), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = (void(*)(void *, void const *))_cffi_to_c_pointer(arg1, _cffi_type(2332));
  if (x1 == (void(*)(void *, void const *))NULL && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GB_UnaryOp_new(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GB_UnaryOp_new _cffi_d_GB_UnaryOp_new
#endif

static GrB_Info _cffi_d_GB_Vector_Iterator_bitmap_seek(struct GB_Iterator_opaque * x0, uint64_t x1)
{
  return GB_Vector_Iterator_bitmap_seek(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GB_Vector_Iterator_bitmap_seek(PyObject *self, PyObject *args)
{
  struct GB_Iterator_opaque * x0;
  uint64_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GB_Vector_Iterator_bitmap_seek", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint64_t);
  if (x1 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GB_Vector_Iterator_bitmap_seek(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GB_Vector_Iterator_bitmap_seek _cffi_d_GB_Vector_Iterator_bitmap_seek
#endif

static GrB_Info _cffi_d_GrB_BinaryOp_error(char const * * x0, struct GB_BinaryOp_opaque * x1)
{
  return GrB_BinaryOp_error(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_BinaryOp_error(PyObject *self, PyObject *args)
{
  char const * * x0;
  struct GB_BinaryOp_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_BinaryOp_error", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(130), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char const * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(130), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_BinaryOp_error(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_BinaryOp_error _cffi_d_GrB_BinaryOp_error
#endif

static GrB_Info _cffi_d_GrB_BinaryOp_free(struct GB_BinaryOp_opaque * * x0)
{
  return GrB_BinaryOp_free(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_BinaryOp_free(PyObject *self, PyObject *arg0)
{
  struct GB_BinaryOp_opaque * * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(422), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(422), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_BinaryOp_free(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_BinaryOp_free _cffi_d_GrB_BinaryOp_free
#endif

static GrB_Info _cffi_d_GrB_BinaryOp_new(struct GB_BinaryOp_opaque * * x0, void(* x1)(void *, void const *, void const *), struct GB_Type_opaque * x2, struct GB_Type_opaque * x3, struct GB_Type_opaque * x4)
{
  return GrB_BinaryOp_new(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_BinaryOp_new(PyObject *self, PyObject *args)
{
  struct GB_BinaryOp_opaque * * x0;
  void(* x1)(void *, void const *, void const *);
  struct GB_Type_opaque * x2;
  struct GB_Type_opaque * x3;
  struct GB_Type_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_BinaryOp_new", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(422), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(422), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = (void(*)(void *, void const *, void const *))_cffi_to_c_pointer(arg1, _cffi_type(434));
  if (x1 == (void(*)(void *, void const *, void const *))NULL && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_BinaryOp_new(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_BinaryOp_new _cffi_d_GrB_BinaryOp_new
#endif

static GrB_Info _cffi_d_GrB_BinaryOp_wait(struct GB_BinaryOp_opaque * x0, GrB_WaitMode x1)
{
  return GrB_BinaryOp_wait(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_BinaryOp_wait(PyObject *self, PyObject *args)
{
  struct GB_BinaryOp_opaque * x0;
  GrB_WaitMode x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_BinaryOp_wait", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(458), arg1) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_BinaryOp_wait(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_BinaryOp_wait _cffi_d_GrB_BinaryOp_wait
#endif

static GrB_Info _cffi_d_GrB_Col_assign(struct GB_Matrix_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Vector_opaque * x3, uint64_t const * x4, uint64_t x5, uint64_t x6, struct GB_Descriptor_opaque * x7)
{
  return GrB_Col_assign(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Col_assign(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Vector_opaque * x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t x6;
  struct GB_Descriptor_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GrB_Col_assign", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x6 = _cffi_to_c_int(arg6, uint64_t);
  if (x6 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Col_assign(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Col_assign _cffi_d_GrB_Col_assign
#endif

static GrB_Info _cffi_d_GrB_Col_extract(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Matrix_opaque * x3, uint64_t const * x4, uint64_t x5, uint64_t x6, struct GB_Descriptor_opaque * x7)
{
  return GrB_Col_extract(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Col_extract(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Matrix_opaque * x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t x6;
  struct GB_Descriptor_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GrB_Col_extract", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x6 = _cffi_to_c_int(arg6, uint64_t);
  if (x6 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Col_extract(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Col_extract _cffi_d_GrB_Col_extract
#endif

static GrB_Info _cffi_d_GrB_Descriptor_error(char const * * x0, struct GB_Descriptor_opaque * x1)
{
  return GrB_Descriptor_error(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Descriptor_error(PyObject *self, PyObject *args)
{
  char const * * x0;
  struct GB_Descriptor_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Descriptor_error", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(130), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char const * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(130), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Descriptor_error(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Descriptor_error _cffi_d_GrB_Descriptor_error
#endif

static GrB_Info _cffi_d_GrB_Descriptor_free(struct GB_Descriptor_opaque * * x0)
{
  return GrB_Descriptor_free(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Descriptor_free(PyObject *self, PyObject *arg0)
{
  struct GB_Descriptor_opaque * * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(467), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(467), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Descriptor_free(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Descriptor_free _cffi_d_GrB_Descriptor_free
#endif

static GrB_Info _cffi_d_GrB_Descriptor_new(struct GB_Descriptor_opaque * * x0)
{
  return GrB_Descriptor_new(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Descriptor_new(PyObject *self, PyObject *arg0)
{
  struct GB_Descriptor_opaque * * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(467), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(467), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Descriptor_new(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Descriptor_new _cffi_d_GrB_Descriptor_new
#endif

static GrB_Info _cffi_d_GrB_Descriptor_set(struct GB_Descriptor_opaque * x0, GrB_Desc_Field x1, GrB_Desc_Value x2)
{
  return GrB_Descriptor_set(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Descriptor_set(PyObject *self, PyObject *args)
{
  struct GB_Descriptor_opaque * x0;
  GrB_Desc_Field x1;
  GrB_Desc_Value x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Descriptor_set", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(3), arg1) < 0)
    return NULL;

  if (_cffi_to_c((char *)&x2, _cffi_type(476), arg2) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Descriptor_set(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Descriptor_set _cffi_d_GrB_Descriptor_set
#endif

static GrB_Info _cffi_d_GrB_Descriptor_wait(struct GB_Descriptor_opaque * x0, GrB_WaitMode x1)
{
  return GrB_Descriptor_wait(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Descriptor_wait(PyObject *self, PyObject *args)
{
  struct GB_Descriptor_opaque * x0;
  GrB_WaitMode x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Descriptor_wait", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(458), arg1) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Descriptor_wait(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Descriptor_wait _cffi_d_GrB_Descriptor_wait
#endif

static GrB_Info _cffi_d_GrB_IndexUnaryOp_error(char const * * x0, struct GB_IndexUnaryOp_opaque * x1)
{
  return GrB_IndexUnaryOp_error(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_IndexUnaryOp_error(PyObject *self, PyObject *args)
{
  char const * * x0;
  struct GB_IndexUnaryOp_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_IndexUnaryOp_error", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(130), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char const * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(130), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_IndexUnaryOp_error(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_IndexUnaryOp_error _cffi_d_GrB_IndexUnaryOp_error
#endif

static GrB_Info _cffi_d_GrB_IndexUnaryOp_free(struct GB_IndexUnaryOp_opaque * * x0)
{
  return GrB_IndexUnaryOp_free(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_IndexUnaryOp_free(PyObject *self, PyObject *arg0)
{
  struct GB_IndexUnaryOp_opaque * * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(509), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(509), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_IndexUnaryOp_free(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_IndexUnaryOp_free _cffi_d_GrB_IndexUnaryOp_free
#endif

static GrB_Info _cffi_d_GrB_IndexUnaryOp_new(struct GB_IndexUnaryOp_opaque * * x0, void(* x1)(void *, void const *, uint64_t, uint64_t, void const *), struct GB_Type_opaque * x2, struct GB_Type_opaque * x3, struct GB_Type_opaque * x4)
{
  return GrB_IndexUnaryOp_new(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_IndexUnaryOp_new(PyObject *self, PyObject *args)
{
  struct GB_IndexUnaryOp_opaque * * x0;
  void(* x1)(void *, void const *, uint64_t, uint64_t, void const *);
  struct GB_Type_opaque * x2;
  struct GB_Type_opaque * x3;
  struct GB_Type_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_IndexUnaryOp_new", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(509), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(509), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = (void(*)(void *, void const *, uint64_t, uint64_t, void const *))_cffi_to_c_pointer(arg1, _cffi_type(513));
  if (x1 == (void(*)(void *, void const *, uint64_t, uint64_t, void const *))NULL && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_IndexUnaryOp_new(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_IndexUnaryOp_new _cffi_d_GrB_IndexUnaryOp_new
#endif

static GrB_Info _cffi_d_GrB_IndexUnaryOp_wait(struct GB_IndexUnaryOp_opaque * x0, GrB_WaitMode x1)
{
  return GrB_IndexUnaryOp_wait(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_IndexUnaryOp_wait(PyObject *self, PyObject *args)
{
  struct GB_IndexUnaryOp_opaque * x0;
  GrB_WaitMode x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_IndexUnaryOp_wait", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(458), arg1) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_IndexUnaryOp_wait(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_IndexUnaryOp_wait _cffi_d_GrB_IndexUnaryOp_wait
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_UnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, struct GB_Descriptor_opaque * x5)
{
  return GrB_Matrix_apply(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_UnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  struct GB_Descriptor_opaque * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(118), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_UnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(118), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply _cffi_d_GrB_Matrix_apply
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp1st_BOOL(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, _Bool x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp1st_BOOL(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp1st_BOOL(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  _Bool x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp1st_BOOL", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = (_Bool)_cffi_to_c__Bool(arg4);
  if (x4 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp1st_BOOL(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp1st_BOOL _cffi_d_GrB_Matrix_apply_BinaryOp1st_BOOL
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp1st_FP32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, float x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp1st_FP32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp1st_FP32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  float x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp1st_FP32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = (float)_cffi_to_c_float(arg4);
  if (x4 == (float)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp1st_FP32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp1st_FP32 _cffi_d_GrB_Matrix_apply_BinaryOp1st_FP32
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp1st_FP64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, double x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp1st_FP64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp1st_FP64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  double x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp1st_FP64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = (double)_cffi_to_c_double(arg4);
  if (x4 == (double)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp1st_FP64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp1st_FP64 _cffi_d_GrB_Matrix_apply_BinaryOp1st_FP64
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp1st_INT16(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, int16_t x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp1st_INT16(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp1st_INT16(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  int16_t x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp1st_INT16", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, int16_t);
  if (x4 == (int16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp1st_INT16(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp1st_INT16 _cffi_d_GrB_Matrix_apply_BinaryOp1st_INT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp1st_INT32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, int32_t x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp1st_INT32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp1st_INT32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  int32_t x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp1st_INT32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, int32_t);
  if (x4 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp1st_INT32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp1st_INT32 _cffi_d_GrB_Matrix_apply_BinaryOp1st_INT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp1st_INT64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, int64_t x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp1st_INT64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp1st_INT64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  int64_t x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp1st_INT64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, int64_t);
  if (x4 == (int64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp1st_INT64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp1st_INT64 _cffi_d_GrB_Matrix_apply_BinaryOp1st_INT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp1st_INT8(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, int8_t x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp1st_INT8(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp1st_INT8(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  int8_t x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp1st_INT8", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, int8_t);
  if (x4 == (int8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp1st_INT8(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp1st_INT8 _cffi_d_GrB_Matrix_apply_BinaryOp1st_INT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp1st_Scalar(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Scalar_opaque * x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp1st_Scalar(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp1st_Scalar(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Scalar_opaque * x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp1st_Scalar", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp1st_Scalar(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp1st_Scalar _cffi_d_GrB_Matrix_apply_BinaryOp1st_Scalar
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp1st_UDT(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, void const * x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp1st_UDT(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp1st_UDT(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  void const * x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp1st_UDT", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(126), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (void const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(126), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp1st_UDT(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp1st_UDT _cffi_d_GrB_Matrix_apply_BinaryOp1st_UDT
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp1st_UINT16(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, uint16_t x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp1st_UINT16(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp1st_UINT16(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  uint16_t x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp1st_UINT16", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint16_t);
  if (x4 == (uint16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp1st_UINT16(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp1st_UINT16 _cffi_d_GrB_Matrix_apply_BinaryOp1st_UINT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp1st_UINT32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, uint32_t x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp1st_UINT32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp1st_UINT32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  uint32_t x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp1st_UINT32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint32_t);
  if (x4 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp1st_UINT32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp1st_UINT32 _cffi_d_GrB_Matrix_apply_BinaryOp1st_UINT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp1st_UINT64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, uint64_t x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp1st_UINT64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp1st_UINT64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  uint64_t x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp1st_UINT64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp1st_UINT64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp1st_UINT64 _cffi_d_GrB_Matrix_apply_BinaryOp1st_UINT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp1st_UINT8(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, uint8_t x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp1st_UINT8(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp1st_UINT8(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  uint8_t x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp1st_UINT8", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint8_t);
  if (x4 == (uint8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp1st_UINT8(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp1st_UINT8 _cffi_d_GrB_Matrix_apply_BinaryOp1st_UINT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp2nd_BOOL(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Matrix_opaque * x4, _Bool x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp2nd_BOOL(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp2nd_BOOL(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  _Bool x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp2nd_BOOL", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (_Bool)_cffi_to_c__Bool(arg5);
  if (x5 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp2nd_BOOL(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp2nd_BOOL _cffi_d_GrB_Matrix_apply_BinaryOp2nd_BOOL
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp2nd_FP32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Matrix_opaque * x4, float x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp2nd_FP32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp2nd_FP32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  float x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp2nd_FP32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (float)_cffi_to_c_float(arg5);
  if (x5 == (float)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp2nd_FP32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp2nd_FP32 _cffi_d_GrB_Matrix_apply_BinaryOp2nd_FP32
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp2nd_FP64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Matrix_opaque * x4, double x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp2nd_FP64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp2nd_FP64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  double x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp2nd_FP64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (double)_cffi_to_c_double(arg5);
  if (x5 == (double)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp2nd_FP64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp2nd_FP64 _cffi_d_GrB_Matrix_apply_BinaryOp2nd_FP64
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp2nd_INT16(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Matrix_opaque * x4, int16_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp2nd_INT16(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp2nd_INT16(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  int16_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp2nd_INT16", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int16_t);
  if (x5 == (int16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp2nd_INT16(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp2nd_INT16 _cffi_d_GrB_Matrix_apply_BinaryOp2nd_INT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp2nd_INT32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Matrix_opaque * x4, int32_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp2nd_INT32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp2nd_INT32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  int32_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp2nd_INT32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int32_t);
  if (x5 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp2nd_INT32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp2nd_INT32 _cffi_d_GrB_Matrix_apply_BinaryOp2nd_INT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp2nd_INT64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Matrix_opaque * x4, int64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp2nd_INT64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp2nd_INT64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  int64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp2nd_INT64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int64_t);
  if (x5 == (int64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp2nd_INT64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp2nd_INT64 _cffi_d_GrB_Matrix_apply_BinaryOp2nd_INT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp2nd_INT8(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Matrix_opaque * x4, int8_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp2nd_INT8(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp2nd_INT8(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  int8_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp2nd_INT8", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int8_t);
  if (x5 == (int8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp2nd_INT8(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp2nd_INT8 _cffi_d_GrB_Matrix_apply_BinaryOp2nd_INT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp2nd_Scalar(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Matrix_opaque * x4, struct GB_Scalar_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp2nd_Scalar(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp2nd_Scalar(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  struct GB_Scalar_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp2nd_Scalar", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp2nd_Scalar(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp2nd_Scalar _cffi_d_GrB_Matrix_apply_BinaryOp2nd_Scalar
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp2nd_UDT(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Matrix_opaque * x4, void const * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp2nd_UDT(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp2nd_UDT(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  void const * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp2nd_UDT", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(126), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (void const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(126), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp2nd_UDT(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp2nd_UDT _cffi_d_GrB_Matrix_apply_BinaryOp2nd_UDT
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp2nd_UINT16(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Matrix_opaque * x4, uint16_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp2nd_UINT16(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp2nd_UINT16(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  uint16_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp2nd_UINT16", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint16_t);
  if (x5 == (uint16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp2nd_UINT16(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp2nd_UINT16 _cffi_d_GrB_Matrix_apply_BinaryOp2nd_UINT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp2nd_UINT32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Matrix_opaque * x4, uint32_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp2nd_UINT32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp2nd_UINT32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  uint32_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp2nd_UINT32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint32_t);
  if (x5 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp2nd_UINT32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp2nd_UINT32 _cffi_d_GrB_Matrix_apply_BinaryOp2nd_UINT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp2nd_UINT64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Matrix_opaque * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp2nd_UINT64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp2nd_UINT64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp2nd_UINT64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp2nd_UINT64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp2nd_UINT64 _cffi_d_GrB_Matrix_apply_BinaryOp2nd_UINT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_BinaryOp2nd_UINT8(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Matrix_opaque * x4, uint8_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_BinaryOp2nd_UINT8(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_BinaryOp2nd_UINT8(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  uint8_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_BinaryOp2nd_UINT8", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint8_t);
  if (x5 == (uint8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_BinaryOp2nd_UINT8(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_BinaryOp2nd_UINT8 _cffi_d_GrB_Matrix_apply_BinaryOp2nd_UINT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_IndexOp_BOOL(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, _Bool x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_IndexOp_BOOL(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_IndexOp_BOOL(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  _Bool x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_IndexOp_BOOL", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (_Bool)_cffi_to_c__Bool(arg5);
  if (x5 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_IndexOp_BOOL(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_IndexOp_BOOL _cffi_d_GrB_Matrix_apply_IndexOp_BOOL
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_IndexOp_FP32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, float x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_IndexOp_FP32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_IndexOp_FP32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  float x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_IndexOp_FP32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (float)_cffi_to_c_float(arg5);
  if (x5 == (float)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_IndexOp_FP32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_IndexOp_FP32 _cffi_d_GrB_Matrix_apply_IndexOp_FP32
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_IndexOp_FP64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, double x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_IndexOp_FP64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_IndexOp_FP64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  double x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_IndexOp_FP64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (double)_cffi_to_c_double(arg5);
  if (x5 == (double)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_IndexOp_FP64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_IndexOp_FP64 _cffi_d_GrB_Matrix_apply_IndexOp_FP64
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_IndexOp_INT16(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, int16_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_IndexOp_INT16(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_IndexOp_INT16(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  int16_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_IndexOp_INT16", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int16_t);
  if (x5 == (int16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_IndexOp_INT16(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_IndexOp_INT16 _cffi_d_GrB_Matrix_apply_IndexOp_INT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_IndexOp_INT32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, int32_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_IndexOp_INT32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_IndexOp_INT32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  int32_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_IndexOp_INT32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int32_t);
  if (x5 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_IndexOp_INT32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_IndexOp_INT32 _cffi_d_GrB_Matrix_apply_IndexOp_INT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_IndexOp_INT64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, int64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_IndexOp_INT64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_IndexOp_INT64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  int64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_IndexOp_INT64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int64_t);
  if (x5 == (int64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_IndexOp_INT64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_IndexOp_INT64 _cffi_d_GrB_Matrix_apply_IndexOp_INT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_IndexOp_INT8(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, int8_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_IndexOp_INT8(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_IndexOp_INT8(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  int8_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_IndexOp_INT8", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int8_t);
  if (x5 == (int8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_IndexOp_INT8(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_IndexOp_INT8 _cffi_d_GrB_Matrix_apply_IndexOp_INT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_IndexOp_Scalar(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, struct GB_Scalar_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_IndexOp_Scalar(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_IndexOp_Scalar(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  struct GB_Scalar_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_IndexOp_Scalar", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_IndexOp_Scalar(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_IndexOp_Scalar _cffi_d_GrB_Matrix_apply_IndexOp_Scalar
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_IndexOp_UDT(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, void const * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_IndexOp_UDT(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_IndexOp_UDT(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  void const * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_IndexOp_UDT", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(126), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (void const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(126), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_IndexOp_UDT(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_IndexOp_UDT _cffi_d_GrB_Matrix_apply_IndexOp_UDT
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_IndexOp_UINT16(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, uint16_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_IndexOp_UINT16(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_IndexOp_UINT16(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  uint16_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_IndexOp_UINT16", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint16_t);
  if (x5 == (uint16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_IndexOp_UINT16(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_IndexOp_UINT16 _cffi_d_GrB_Matrix_apply_IndexOp_UINT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_IndexOp_UINT32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, uint32_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_IndexOp_UINT32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_IndexOp_UINT32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  uint32_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_IndexOp_UINT32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint32_t);
  if (x5 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_IndexOp_UINT32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_IndexOp_UINT32 _cffi_d_GrB_Matrix_apply_IndexOp_UINT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_IndexOp_UINT64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_IndexOp_UINT64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_IndexOp_UINT64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_IndexOp_UINT64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_IndexOp_UINT64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_IndexOp_UINT64 _cffi_d_GrB_Matrix_apply_IndexOp_UINT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_apply_IndexOp_UINT8(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, uint8_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_apply_IndexOp_UINT8(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_apply_IndexOp_UINT8(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  uint8_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_apply_IndexOp_UINT8", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint8_t);
  if (x5 == (uint8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_apply_IndexOp_UINT8(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_apply_IndexOp_UINT8 _cffi_d_GrB_Matrix_apply_IndexOp_UINT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_assign(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Matrix_opaque * x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GrB_Matrix_assign(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_assign(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Matrix_opaque * x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_assign", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_assign(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_assign _cffi_d_GrB_Matrix_assign
#endif

static GrB_Info _cffi_d_GrB_Matrix_assign_BOOL(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, _Bool x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GrB_Matrix_assign_BOOL(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_assign_BOOL(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  _Bool x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_assign_BOOL", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = (_Bool)_cffi_to_c__Bool(arg3);
  if (x3 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_assign_BOOL(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_assign_BOOL _cffi_d_GrB_Matrix_assign_BOOL
#endif

static GrB_Info _cffi_d_GrB_Matrix_assign_FP32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, float x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GrB_Matrix_assign_FP32(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_assign_FP32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  float x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_assign_FP32", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = (float)_cffi_to_c_float(arg3);
  if (x3 == (float)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_assign_FP32(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_assign_FP32 _cffi_d_GrB_Matrix_assign_FP32
#endif

static GrB_Info _cffi_d_GrB_Matrix_assign_FP64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, double x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GrB_Matrix_assign_FP64(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_assign_FP64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  double x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_assign_FP64", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = (double)_cffi_to_c_double(arg3);
  if (x3 == (double)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_assign_FP64(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_assign_FP64 _cffi_d_GrB_Matrix_assign_FP64
#endif

static GrB_Info _cffi_d_GrB_Matrix_assign_INT16(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, int16_t x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GrB_Matrix_assign_INT16(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_assign_INT16(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  int16_t x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_assign_INT16", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, int16_t);
  if (x3 == (int16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_assign_INT16(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_assign_INT16 _cffi_d_GrB_Matrix_assign_INT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_assign_INT32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, int32_t x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GrB_Matrix_assign_INT32(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_assign_INT32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  int32_t x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_assign_INT32", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, int32_t);
  if (x3 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_assign_INT32(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_assign_INT32 _cffi_d_GrB_Matrix_assign_INT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_assign_INT64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, int64_t x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GrB_Matrix_assign_INT64(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_assign_INT64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  int64_t x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_assign_INT64", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, int64_t);
  if (x3 == (int64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_assign_INT64(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_assign_INT64 _cffi_d_GrB_Matrix_assign_INT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_assign_INT8(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, int8_t x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GrB_Matrix_assign_INT8(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_assign_INT8(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  int8_t x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_assign_INT8", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, int8_t);
  if (x3 == (int8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_assign_INT8(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_assign_INT8 _cffi_d_GrB_Matrix_assign_INT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_assign_Scalar(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Scalar_opaque * x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GrB_Matrix_assign_Scalar(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_assign_Scalar(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Scalar_opaque * x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_assign_Scalar", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_assign_Scalar(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_assign_Scalar _cffi_d_GrB_Matrix_assign_Scalar
#endif

static GrB_Info _cffi_d_GrB_Matrix_assign_UDT(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, void * x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GrB_Matrix_assign_UDT(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_assign_UDT(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  void * x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_assign_UDT", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_assign_UDT(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_assign_UDT _cffi_d_GrB_Matrix_assign_UDT
#endif

static GrB_Info _cffi_d_GrB_Matrix_assign_UINT16(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, uint16_t x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GrB_Matrix_assign_UINT16(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_assign_UINT16(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  uint16_t x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_assign_UINT16", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint16_t);
  if (x3 == (uint16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_assign_UINT16(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_assign_UINT16 _cffi_d_GrB_Matrix_assign_UINT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_assign_UINT32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, uint32_t x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GrB_Matrix_assign_UINT32(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_assign_UINT32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  uint32_t x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_assign_UINT32", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint32_t);
  if (x3 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_assign_UINT32(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_assign_UINT32 _cffi_d_GrB_Matrix_assign_UINT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_assign_UINT64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, uint64_t x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GrB_Matrix_assign_UINT64(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_assign_UINT64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  uint64_t x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_assign_UINT64", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_assign_UINT64(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_assign_UINT64 _cffi_d_GrB_Matrix_assign_UINT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_assign_UINT8(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, uint8_t x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GrB_Matrix_assign_UINT8(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_assign_UINT8(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  uint8_t x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_assign_UINT8", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint8_t);
  if (x3 == (uint8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_assign_UINT8(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_assign_UINT8 _cffi_d_GrB_Matrix_assign_UINT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_build_BOOL(struct GB_Matrix_opaque * x0, uint64_t const * x1, uint64_t const * x2, _Bool const * x3, uint64_t x4, struct GB_BinaryOp_opaque * x5)
{
  return GrB_Matrix_build_BOOL(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_build_BOOL(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t const * x1;
  uint64_t const * x2;
  _Bool const * x3;
  uint64_t x4;
  struct GB_BinaryOp_opaque * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_build_BOOL", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(698), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (_Bool const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(698), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_build_BOOL(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_build_BOOL _cffi_d_GrB_Matrix_build_BOOL
#endif

static GrB_Info _cffi_d_GrB_Matrix_build_FP32(struct GB_Matrix_opaque * x0, uint64_t const * x1, uint64_t const * x2, float const * x3, uint64_t x4, struct GB_BinaryOp_opaque * x5)
{
  return GrB_Matrix_build_FP32(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_build_FP32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t const * x1;
  uint64_t const * x2;
  float const * x3;
  uint64_t x4;
  struct GB_BinaryOp_opaque * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_build_FP32", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(750), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (float const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(750), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_build_FP32(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_build_FP32 _cffi_d_GrB_Matrix_build_FP32
#endif

static GrB_Info _cffi_d_GrB_Matrix_build_FP64(struct GB_Matrix_opaque * x0, uint64_t const * x1, uint64_t const * x2, double const * x3, uint64_t x4, struct GB_BinaryOp_opaque * x5)
{
  return GrB_Matrix_build_FP64(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_build_FP64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t const * x1;
  uint64_t const * x2;
  double const * x3;
  uint64_t x4;
  struct GB_BinaryOp_opaque * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_build_FP64", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(724), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (double const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(724), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_build_FP64(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_build_FP64 _cffi_d_GrB_Matrix_build_FP64
#endif

static GrB_Info _cffi_d_GrB_Matrix_build_INT16(struct GB_Matrix_opaque * x0, uint64_t const * x1, uint64_t const * x2, int16_t const * x3, uint64_t x4, struct GB_BinaryOp_opaque * x5)
{
  return GrB_Matrix_build_INT16(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_build_INT16(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t const * x1;
  uint64_t const * x2;
  int16_t const * x3;
  uint64_t x4;
  struct GB_BinaryOp_opaque * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_build_INT16", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(763), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (int16_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(763), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_build_INT16(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_build_INT16 _cffi_d_GrB_Matrix_build_INT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_build_INT32(struct GB_Matrix_opaque * x0, uint64_t const * x1, uint64_t const * x2, int32_t const * x3, uint64_t x4, struct GB_BinaryOp_opaque * x5)
{
  return GrB_Matrix_build_INT32(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_build_INT32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t const * x1;
  uint64_t const * x2;
  int32_t const * x3;
  uint64_t x4;
  struct GB_BinaryOp_opaque * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_build_INT32", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(776), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (int32_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(776), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_build_INT32(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_build_INT32 _cffi_d_GrB_Matrix_build_INT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_build_INT64(struct GB_Matrix_opaque * x0, uint64_t const * x1, uint64_t const * x2, int64_t const * x3, uint64_t x4, struct GB_BinaryOp_opaque * x5)
{
  return GrB_Matrix_build_INT64(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_build_INT64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t const * x1;
  uint64_t const * x2;
  int64_t const * x3;
  uint64_t x4;
  struct GB_BinaryOp_opaque * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_build_INT64", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(789), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (int64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(789), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_build_INT64(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_build_INT64 _cffi_d_GrB_Matrix_build_INT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_build_INT8(struct GB_Matrix_opaque * x0, uint64_t const * x1, uint64_t const * x2, int8_t const * x3, uint64_t x4, struct GB_BinaryOp_opaque * x5)
{
  return GrB_Matrix_build_INT8(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_build_INT8(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t const * x1;
  uint64_t const * x2;
  int8_t const * x3;
  uint64_t x4;
  struct GB_BinaryOp_opaque * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_build_INT8", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(802), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (int8_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(802), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_build_INT8(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_build_INT8 _cffi_d_GrB_Matrix_build_INT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_build_UDT(struct GB_Matrix_opaque * x0, uint64_t const * x1, uint64_t const * x2, void const * x3, uint64_t x4, struct GB_BinaryOp_opaque * x5)
{
  return GrB_Matrix_build_UDT(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_build_UDT(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t const * x1;
  uint64_t const * x2;
  void const * x3;
  uint64_t x4;
  struct GB_BinaryOp_opaque * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_build_UDT", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(126), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (void const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(126), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_build_UDT(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_build_UDT _cffi_d_GrB_Matrix_build_UDT
#endif

static GrB_Info _cffi_d_GrB_Matrix_build_UINT16(struct GB_Matrix_opaque * x0, uint64_t const * x1, uint64_t const * x2, uint16_t const * x3, uint64_t x4, struct GB_BinaryOp_opaque * x5)
{
  return GrB_Matrix_build_UINT16(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_build_UINT16(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t const * x1;
  uint64_t const * x2;
  uint16_t const * x3;
  uint64_t x4;
  struct GB_BinaryOp_opaque * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_build_UINT16", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(815), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint16_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(815), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_build_UINT16(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_build_UINT16 _cffi_d_GrB_Matrix_build_UINT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_build_UINT32(struct GB_Matrix_opaque * x0, uint64_t const * x1, uint64_t const * x2, uint32_t const * x3, uint64_t x4, struct GB_BinaryOp_opaque * x5)
{
  return GrB_Matrix_build_UINT32(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_build_UINT32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t const * x1;
  uint64_t const * x2;
  uint32_t const * x3;
  uint64_t x4;
  struct GB_BinaryOp_opaque * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_build_UINT32", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(828), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint32_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(828), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_build_UINT32(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_build_UINT32 _cffi_d_GrB_Matrix_build_UINT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_build_UINT64(struct GB_Matrix_opaque * x0, uint64_t const * x1, uint64_t const * x2, uint64_t const * x3, uint64_t x4, struct GB_BinaryOp_opaque * x5)
{
  return GrB_Matrix_build_UINT64(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_build_UINT64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t const * x1;
  uint64_t const * x2;
  uint64_t const * x3;
  uint64_t x4;
  struct GB_BinaryOp_opaque * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_build_UINT64", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_build_UINT64(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_build_UINT64 _cffi_d_GrB_Matrix_build_UINT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_build_UINT8(struct GB_Matrix_opaque * x0, uint64_t const * x1, uint64_t const * x2, uint8_t const * x3, uint64_t x4, struct GB_BinaryOp_opaque * x5)
{
  return GrB_Matrix_build_UINT8(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_build_UINT8(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t const * x1;
  uint64_t const * x2;
  uint8_t const * x3;
  uint64_t x4;
  struct GB_BinaryOp_opaque * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_build_UINT8", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(854), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint8_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(854), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_build_UINT8(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_build_UINT8 _cffi_d_GrB_Matrix_build_UINT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_clear(struct GB_Matrix_opaque * x0)
{
  return GrB_Matrix_clear(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_clear(PyObject *self, PyObject *arg0)
{
  struct GB_Matrix_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_clear(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_clear _cffi_d_GrB_Matrix_clear
#endif

static GrB_Info _cffi_d_GrB_Matrix_deserialize(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, void const * x2, uint64_t x3)
{
  return GrB_Matrix_deserialize(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_deserialize(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  void const * x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_deserialize", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(126), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (void const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(126), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_deserialize(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_deserialize _cffi_d_GrB_Matrix_deserialize
#endif

static GrB_Info _cffi_d_GrB_Matrix_diag(struct GB_Matrix_opaque * * x0, struct GB_Vector_opaque * x1, int64_t x2)
{
  return GrB_Matrix_diag(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_diag(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Vector_opaque * x1;
  int64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_diag", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, int64_t);
  if (x2 == (int64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_diag(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_diag _cffi_d_GrB_Matrix_diag
#endif

static GrB_Info _cffi_d_GrB_Matrix_dup(struct GB_Matrix_opaque * * x0, struct GB_Matrix_opaque * x1)
{
  return GrB_Matrix_dup(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_dup(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Matrix_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_dup", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_dup(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_dup _cffi_d_GrB_Matrix_dup
#endif

static GrB_Info _cffi_d_GrB_Matrix_eWiseAdd_BinaryOp(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Matrix_opaque * x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_eWiseAdd_BinaryOp(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_eWiseAdd_BinaryOp(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_eWiseAdd_BinaryOp", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_eWiseAdd_BinaryOp(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_eWiseAdd_BinaryOp _cffi_d_GrB_Matrix_eWiseAdd_BinaryOp
#endif

static GrB_Info _cffi_d_GrB_Matrix_eWiseAdd_Monoid(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Monoid_opaque * x3, struct GB_Matrix_opaque * x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_eWiseAdd_Monoid(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_eWiseAdd_Monoid(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Monoid_opaque * x3;
  struct GB_Matrix_opaque * x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_eWiseAdd_Monoid", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_eWiseAdd_Monoid(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_eWiseAdd_Monoid _cffi_d_GrB_Matrix_eWiseAdd_Monoid
#endif

static GrB_Info _cffi_d_GrB_Matrix_eWiseAdd_Semiring(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Semiring_opaque * x3, struct GB_Matrix_opaque * x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_eWiseAdd_Semiring(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_eWiseAdd_Semiring(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Semiring_opaque * x3;
  struct GB_Matrix_opaque * x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_eWiseAdd_Semiring", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(159), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Semiring_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(159), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_eWiseAdd_Semiring(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_eWiseAdd_Semiring _cffi_d_GrB_Matrix_eWiseAdd_Semiring
#endif

static GrB_Info _cffi_d_GrB_Matrix_eWiseMult_BinaryOp(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Matrix_opaque * x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_eWiseMult_BinaryOp(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_eWiseMult_BinaryOp(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_eWiseMult_BinaryOp", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_eWiseMult_BinaryOp(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_eWiseMult_BinaryOp _cffi_d_GrB_Matrix_eWiseMult_BinaryOp
#endif

static GrB_Info _cffi_d_GrB_Matrix_eWiseMult_Monoid(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Monoid_opaque * x3, struct GB_Matrix_opaque * x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_eWiseMult_Monoid(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_eWiseMult_Monoid(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Monoid_opaque * x3;
  struct GB_Matrix_opaque * x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_eWiseMult_Monoid", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_eWiseMult_Monoid(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_eWiseMult_Monoid _cffi_d_GrB_Matrix_eWiseMult_Monoid
#endif

static GrB_Info _cffi_d_GrB_Matrix_eWiseMult_Semiring(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Semiring_opaque * x3, struct GB_Matrix_opaque * x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_eWiseMult_Semiring(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_eWiseMult_Semiring(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Semiring_opaque * x3;
  struct GB_Matrix_opaque * x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_eWiseMult_Semiring", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(159), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Semiring_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(159), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_eWiseMult_Semiring(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_eWiseMult_Semiring _cffi_d_GrB_Matrix_eWiseMult_Semiring
#endif

static GrB_Info _cffi_d_GrB_Matrix_error(char const * * x0, struct GB_Matrix_opaque * x1)
{
  return GrB_Matrix_error(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_error(PyObject *self, PyObject *args)
{
  char const * * x0;
  struct GB_Matrix_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_error", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(130), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char const * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(130), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_error(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_error _cffi_d_GrB_Matrix_error
#endif

static GrB_Info _cffi_d_GrB_Matrix_exportHint(GrB_Format * x0, struct GB_Matrix_opaque * x1)
{
  return GrB_Matrix_exportHint(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_exportHint(PyObject *self, PyObject *args)
{
  GrB_Format * x0;
  struct GB_Matrix_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_exportHint", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(6), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (GrB_Format *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(6), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_exportHint(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_exportHint _cffi_d_GrB_Matrix_exportHint
#endif

static GrB_Info _cffi_d_GrB_Matrix_exportSize(uint64_t * x0, uint64_t * x1, uint64_t * x2, GrB_Format x3, struct GB_Matrix_opaque * x4)
{
  return GrB_Matrix_exportSize(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_exportSize(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  uint64_t * x2;
  GrB_Format x3;
  struct GB_Matrix_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_exportSize", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x3, _cffi_type(702), arg3) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_exportSize(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_exportSize _cffi_d_GrB_Matrix_exportSize
#endif

static GrB_Info _cffi_d_GrB_Matrix_export_BOOL(uint64_t * x0, uint64_t * x1, _Bool * x2, uint64_t * x3, uint64_t * x4, uint64_t * x5, GrB_Format x6, struct GB_Matrix_opaque * x7)
{
  return GrB_Matrix_export_BOOL(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_export_BOOL(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  _Bool * x2;
  uint64_t * x3;
  uint64_t * x4;
  uint64_t * x5;
  GrB_Format x6;
  struct GB_Matrix_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_export_BOOL", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x6, _cffi_type(702), arg6) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_export_BOOL(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_export_BOOL _cffi_d_GrB_Matrix_export_BOOL
#endif

static GrB_Info _cffi_d_GrB_Matrix_export_FP32(uint64_t * x0, uint64_t * x1, float * x2, uint64_t * x3, uint64_t * x4, uint64_t * x5, GrB_Format x6, struct GB_Matrix_opaque * x7)
{
  return GrB_Matrix_export_FP32(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_export_FP32(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  float * x2;
  uint64_t * x3;
  uint64_t * x4;
  uint64_t * x5;
  GrB_Format x6;
  struct GB_Matrix_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_export_FP32", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(232), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (float *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(232), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x6, _cffi_type(702), arg6) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_export_FP32(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_export_FP32 _cffi_d_GrB_Matrix_export_FP32
#endif

static GrB_Info _cffi_d_GrB_Matrix_export_FP64(uint64_t * x0, uint64_t * x1, double * x2, uint64_t * x3, uint64_t * x4, uint64_t * x5, GrB_Format x6, struct GB_Matrix_opaque * x7)
{
  return GrB_Matrix_export_FP64(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_export_FP64(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  double * x2;
  uint64_t * x3;
  uint64_t * x4;
  uint64_t * x5;
  GrB_Format x6;
  struct GB_Matrix_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_export_FP64", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(28), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(28), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x6, _cffi_type(702), arg6) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_export_FP64(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_export_FP64 _cffi_d_GrB_Matrix_export_FP64
#endif

static GrB_Info _cffi_d_GrB_Matrix_export_INT16(uint64_t * x0, uint64_t * x1, int16_t * x2, uint64_t * x3, uint64_t * x4, uint64_t * x5, GrB_Format x6, struct GB_Matrix_opaque * x7)
{
  return GrB_Matrix_export_INT16(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_export_INT16(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  int16_t * x2;
  uint64_t * x3;
  uint64_t * x4;
  uint64_t * x5;
  GrB_Format x6;
  struct GB_Matrix_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_export_INT16", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(290), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (int16_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(290), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x6, _cffi_type(702), arg6) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_export_INT16(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_export_INT16 _cffi_d_GrB_Matrix_export_INT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_export_INT32(uint64_t * x0, uint64_t * x1, int32_t * x2, uint64_t * x3, uint64_t * x4, uint64_t * x5, GrB_Format x6, struct GB_Matrix_opaque * x7)
{
  return GrB_Matrix_export_INT32(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_export_INT32(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  int32_t * x2;
  uint64_t * x3;
  uint64_t * x4;
  uint64_t * x5;
  GrB_Format x6;
  struct GB_Matrix_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_export_INT32", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(36), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (int32_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(36), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x6, _cffi_type(702), arg6) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_export_INT32(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_export_INT32 _cffi_d_GrB_Matrix_export_INT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_export_INT64(uint64_t * x0, uint64_t * x1, int64_t * x2, uint64_t * x3, uint64_t * x4, uint64_t * x5, GrB_Format x6, struct GB_Matrix_opaque * x7)
{
  return GrB_Matrix_export_INT64(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_export_INT64(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  int64_t * x2;
  uint64_t * x3;
  uint64_t * x4;
  uint64_t * x5;
  GrB_Format x6;
  struct GB_Matrix_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_export_INT64", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(44), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (int64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(44), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x6, _cffi_type(702), arg6) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_export_INT64(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_export_INT64 _cffi_d_GrB_Matrix_export_INT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_export_INT8(uint64_t * x0, uint64_t * x1, int8_t * x2, uint64_t * x3, uint64_t * x4, uint64_t * x5, GrB_Format x6, struct GB_Matrix_opaque * x7)
{
  return GrB_Matrix_export_INT8(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_export_INT8(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  int8_t * x2;
  uint64_t * x3;
  uint64_t * x4;
  uint64_t * x5;
  GrB_Format x6;
  struct GB_Matrix_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_export_INT8", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(377), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (int8_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(377), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x6, _cffi_type(702), arg6) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_export_INT8(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_export_INT8 _cffi_d_GrB_Matrix_export_INT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_export_UDT(uint64_t * x0, uint64_t * x1, void * x2, uint64_t * x3, uint64_t * x4, uint64_t * x5, GrB_Format x6, struct GB_Matrix_opaque * x7)
{
  return GrB_Matrix_export_UDT(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_export_UDT(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  void * x2;
  uint64_t * x3;
  uint64_t * x4;
  uint64_t * x5;
  GrB_Format x6;
  struct GB_Matrix_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_export_UDT", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x6, _cffi_type(702), arg6) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_export_UDT(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_export_UDT _cffi_d_GrB_Matrix_export_UDT
#endif

static GrB_Info _cffi_d_GrB_Matrix_export_UINT16(uint64_t * x0, uint64_t * x1, uint16_t * x2, uint64_t * x3, uint64_t * x4, uint64_t * x5, GrB_Format x6, struct GB_Matrix_opaque * x7)
{
  return GrB_Matrix_export_UINT16(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_export_UINT16(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  uint16_t * x2;
  uint64_t * x3;
  uint64_t * x4;
  uint64_t * x5;
  GrB_Format x6;
  struct GB_Matrix_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_export_UINT16", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3386), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint16_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3386), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x6, _cffi_type(702), arg6) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_export_UINT16(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_export_UINT16 _cffi_d_GrB_Matrix_export_UINT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_export_UINT32(uint64_t * x0, uint64_t * x1, uint32_t * x2, uint64_t * x3, uint64_t * x4, uint64_t * x5, GrB_Format x6, struct GB_Matrix_opaque * x7)
{
  return GrB_Matrix_export_UINT32(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_export_UINT32(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  uint32_t * x2;
  uint64_t * x3;
  uint64_t * x4;
  uint64_t * x5;
  GrB_Format x6;
  struct GB_Matrix_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_export_UINT32", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3415), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint32_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3415), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x6, _cffi_type(702), arg6) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_export_UINT32(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_export_UINT32 _cffi_d_GrB_Matrix_export_UINT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_export_UINT64(uint64_t * x0, uint64_t * x1, uint64_t * x2, uint64_t * x3, uint64_t * x4, uint64_t * x5, GrB_Format x6, struct GB_Matrix_opaque * x7)
{
  return GrB_Matrix_export_UINT64(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_export_UINT64(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  uint64_t * x2;
  uint64_t * x3;
  uint64_t * x4;
  uint64_t * x5;
  GrB_Format x6;
  struct GB_Matrix_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_export_UINT64", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x6, _cffi_type(702), arg6) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_export_UINT64(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_export_UINT64 _cffi_d_GrB_Matrix_export_UINT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_export_UINT8(uint64_t * x0, uint64_t * x1, uint8_t * x2, uint64_t * x3, uint64_t * x4, uint64_t * x5, GrB_Format x6, struct GB_Matrix_opaque * x7)
{
  return GrB_Matrix_export_UINT8(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_export_UINT8(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  uint8_t * x2;
  uint64_t * x3;
  uint64_t * x4;
  uint64_t * x5;
  GrB_Format x6;
  struct GB_Matrix_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_export_UINT8", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3766), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint8_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3766), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x6, _cffi_type(702), arg6) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_export_UINT8(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_export_UINT8 _cffi_d_GrB_Matrix_export_UINT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_extract(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Matrix_opaque * x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GrB_Matrix_extract(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extract(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Matrix_opaque * x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extract", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extract(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extract _cffi_d_GrB_Matrix_extract
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractElement_BOOL(_Bool * x0, struct GB_Matrix_opaque * x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_extractElement_BOOL(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractElement_BOOL(PyObject *self, PyObject *args)
{
  _Bool * x0;
  struct GB_Matrix_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractElement_BOOL", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractElement_BOOL(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractElement_BOOL _cffi_d_GrB_Matrix_extractElement_BOOL
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractElement_FP32(float * x0, struct GB_Matrix_opaque * x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_extractElement_FP32(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractElement_FP32(PyObject *self, PyObject *args)
{
  float * x0;
  struct GB_Matrix_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractElement_FP32", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(232), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (float *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(232), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractElement_FP32(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractElement_FP32 _cffi_d_GrB_Matrix_extractElement_FP32
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractElement_FP64(double * x0, struct GB_Matrix_opaque * x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_extractElement_FP64(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractElement_FP64(PyObject *self, PyObject *args)
{
  double * x0;
  struct GB_Matrix_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractElement_FP64", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(28), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(28), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractElement_FP64(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractElement_FP64 _cffi_d_GrB_Matrix_extractElement_FP64
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractElement_INT16(int16_t * x0, struct GB_Matrix_opaque * x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_extractElement_INT16(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractElement_INT16(PyObject *self, PyObject *args)
{
  int16_t * x0;
  struct GB_Matrix_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractElement_INT16", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(290), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int16_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(290), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractElement_INT16(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractElement_INT16 _cffi_d_GrB_Matrix_extractElement_INT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractElement_INT32(int32_t * x0, struct GB_Matrix_opaque * x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_extractElement_INT32(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractElement_INT32(PyObject *self, PyObject *args)
{
  int32_t * x0;
  struct GB_Matrix_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractElement_INT32", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(36), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int32_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(36), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractElement_INT32(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractElement_INT32 _cffi_d_GrB_Matrix_extractElement_INT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractElement_INT64(int64_t * x0, struct GB_Matrix_opaque * x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_extractElement_INT64(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractElement_INT64(PyObject *self, PyObject *args)
{
  int64_t * x0;
  struct GB_Matrix_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractElement_INT64", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(44), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(44), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractElement_INT64(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractElement_INT64 _cffi_d_GrB_Matrix_extractElement_INT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractElement_INT8(int8_t * x0, struct GB_Matrix_opaque * x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_extractElement_INT8(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractElement_INT8(PyObject *self, PyObject *args)
{
  int8_t * x0;
  struct GB_Matrix_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractElement_INT8", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(377), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int8_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(377), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractElement_INT8(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractElement_INT8 _cffi_d_GrB_Matrix_extractElement_INT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractElement_Scalar(struct GB_Scalar_opaque * x0, struct GB_Matrix_opaque * x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_extractElement_Scalar(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractElement_Scalar(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  struct GB_Matrix_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractElement_Scalar", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractElement_Scalar(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractElement_Scalar _cffi_d_GrB_Matrix_extractElement_Scalar
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractElement_UDT(void * x0, struct GB_Matrix_opaque * x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_extractElement_UDT(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractElement_UDT(PyObject *self, PyObject *args)
{
  void * x0;
  struct GB_Matrix_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractElement_UDT", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractElement_UDT(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractElement_UDT _cffi_d_GrB_Matrix_extractElement_UDT
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractElement_UINT16(uint16_t * x0, struct GB_Matrix_opaque * x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_extractElement_UINT16(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractElement_UINT16(PyObject *self, PyObject *args)
{
  uint16_t * x0;
  struct GB_Matrix_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractElement_UINT16", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3386), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint16_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3386), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractElement_UINT16(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractElement_UINT16 _cffi_d_GrB_Matrix_extractElement_UINT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractElement_UINT32(uint32_t * x0, struct GB_Matrix_opaque * x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_extractElement_UINT32(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractElement_UINT32(PyObject *self, PyObject *args)
{
  uint32_t * x0;
  struct GB_Matrix_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractElement_UINT32", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3415), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint32_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3415), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractElement_UINT32(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractElement_UINT32 _cffi_d_GrB_Matrix_extractElement_UINT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractElement_UINT64(uint64_t * x0, struct GB_Matrix_opaque * x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_extractElement_UINT64(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractElement_UINT64(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  struct GB_Matrix_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractElement_UINT64", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractElement_UINT64(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractElement_UINT64 _cffi_d_GrB_Matrix_extractElement_UINT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractElement_UINT8(uint8_t * x0, struct GB_Matrix_opaque * x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_extractElement_UINT8(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractElement_UINT8(PyObject *self, PyObject *args)
{
  uint8_t * x0;
  struct GB_Matrix_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractElement_UINT8", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3766), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint8_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3766), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractElement_UINT8(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractElement_UINT8 _cffi_d_GrB_Matrix_extractElement_UINT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractTuples_BOOL(uint64_t * x0, uint64_t * x1, _Bool * x2, uint64_t * x3, struct GB_Matrix_opaque * x4)
{
  return GrB_Matrix_extractTuples_BOOL(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractTuples_BOOL(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  _Bool * x2;
  uint64_t * x3;
  struct GB_Matrix_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractTuples_BOOL", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractTuples_BOOL(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractTuples_BOOL _cffi_d_GrB_Matrix_extractTuples_BOOL
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractTuples_FP32(uint64_t * x0, uint64_t * x1, float * x2, uint64_t * x3, struct GB_Matrix_opaque * x4)
{
  return GrB_Matrix_extractTuples_FP32(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractTuples_FP32(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  float * x2;
  uint64_t * x3;
  struct GB_Matrix_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractTuples_FP32", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(232), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (float *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(232), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractTuples_FP32(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractTuples_FP32 _cffi_d_GrB_Matrix_extractTuples_FP32
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractTuples_FP64(uint64_t * x0, uint64_t * x1, double * x2, uint64_t * x3, struct GB_Matrix_opaque * x4)
{
  return GrB_Matrix_extractTuples_FP64(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractTuples_FP64(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  double * x2;
  uint64_t * x3;
  struct GB_Matrix_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractTuples_FP64", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(28), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(28), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractTuples_FP64(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractTuples_FP64 _cffi_d_GrB_Matrix_extractTuples_FP64
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractTuples_INT16(uint64_t * x0, uint64_t * x1, int16_t * x2, uint64_t * x3, struct GB_Matrix_opaque * x4)
{
  return GrB_Matrix_extractTuples_INT16(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractTuples_INT16(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  int16_t * x2;
  uint64_t * x3;
  struct GB_Matrix_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractTuples_INT16", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(290), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (int16_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(290), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractTuples_INT16(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractTuples_INT16 _cffi_d_GrB_Matrix_extractTuples_INT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractTuples_INT32(uint64_t * x0, uint64_t * x1, int32_t * x2, uint64_t * x3, struct GB_Matrix_opaque * x4)
{
  return GrB_Matrix_extractTuples_INT32(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractTuples_INT32(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  int32_t * x2;
  uint64_t * x3;
  struct GB_Matrix_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractTuples_INT32", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(36), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (int32_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(36), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractTuples_INT32(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractTuples_INT32 _cffi_d_GrB_Matrix_extractTuples_INT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractTuples_INT64(uint64_t * x0, uint64_t * x1, int64_t * x2, uint64_t * x3, struct GB_Matrix_opaque * x4)
{
  return GrB_Matrix_extractTuples_INT64(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractTuples_INT64(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  int64_t * x2;
  uint64_t * x3;
  struct GB_Matrix_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractTuples_INT64", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(44), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (int64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(44), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractTuples_INT64(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractTuples_INT64 _cffi_d_GrB_Matrix_extractTuples_INT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractTuples_INT8(uint64_t * x0, uint64_t * x1, int8_t * x2, uint64_t * x3, struct GB_Matrix_opaque * x4)
{
  return GrB_Matrix_extractTuples_INT8(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractTuples_INT8(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  int8_t * x2;
  uint64_t * x3;
  struct GB_Matrix_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractTuples_INT8", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(377), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (int8_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(377), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractTuples_INT8(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractTuples_INT8 _cffi_d_GrB_Matrix_extractTuples_INT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractTuples_UDT(uint64_t * x0, uint64_t * x1, void * x2, uint64_t * x3, struct GB_Matrix_opaque * x4)
{
  return GrB_Matrix_extractTuples_UDT(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractTuples_UDT(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  void * x2;
  uint64_t * x3;
  struct GB_Matrix_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractTuples_UDT", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractTuples_UDT(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractTuples_UDT _cffi_d_GrB_Matrix_extractTuples_UDT
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractTuples_UINT16(uint64_t * x0, uint64_t * x1, uint16_t * x2, uint64_t * x3, struct GB_Matrix_opaque * x4)
{
  return GrB_Matrix_extractTuples_UINT16(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractTuples_UINT16(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  uint16_t * x2;
  uint64_t * x3;
  struct GB_Matrix_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractTuples_UINT16", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3386), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint16_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3386), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractTuples_UINT16(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractTuples_UINT16 _cffi_d_GrB_Matrix_extractTuples_UINT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractTuples_UINT32(uint64_t * x0, uint64_t * x1, uint32_t * x2, uint64_t * x3, struct GB_Matrix_opaque * x4)
{
  return GrB_Matrix_extractTuples_UINT32(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractTuples_UINT32(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  uint32_t * x2;
  uint64_t * x3;
  struct GB_Matrix_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractTuples_UINT32", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3415), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint32_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3415), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractTuples_UINT32(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractTuples_UINT32 _cffi_d_GrB_Matrix_extractTuples_UINT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractTuples_UINT64(uint64_t * x0, uint64_t * x1, uint64_t * x2, uint64_t * x3, struct GB_Matrix_opaque * x4)
{
  return GrB_Matrix_extractTuples_UINT64(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractTuples_UINT64(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  uint64_t * x2;
  uint64_t * x3;
  struct GB_Matrix_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractTuples_UINT64", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractTuples_UINT64(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractTuples_UINT64 _cffi_d_GrB_Matrix_extractTuples_UINT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_extractTuples_UINT8(uint64_t * x0, uint64_t * x1, uint8_t * x2, uint64_t * x3, struct GB_Matrix_opaque * x4)
{
  return GrB_Matrix_extractTuples_UINT8(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_extractTuples_UINT8(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  uint8_t * x2;
  uint64_t * x3;
  struct GB_Matrix_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_extractTuples_UINT8", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3766), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint8_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3766), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_extractTuples_UINT8(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_extractTuples_UINT8 _cffi_d_GrB_Matrix_extractTuples_UINT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_free(struct GB_Matrix_opaque * * x0)
{
  return GrB_Matrix_free(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_free(PyObject *self, PyObject *arg0)
{
  struct GB_Matrix_opaque * * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_free(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_free _cffi_d_GrB_Matrix_free
#endif

static GrB_Info _cffi_d_GrB_Matrix_import_BOOL(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t x3, uint64_t const * x4, uint64_t const * x5, _Bool const * x6, uint64_t x7, uint64_t x8, uint64_t x9, GrB_Format x10)
{
  return GrB_Matrix_import_BOOL(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_import_BOOL(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  uint64_t const * x4;
  uint64_t const * x5;
  _Bool const * x6;
  uint64_t x7;
  uint64_t x8;
  uint64_t x9;
  GrB_Format x10;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_import_BOOL", 11, 11, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(698), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (_Bool const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(698), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x8 = _cffi_to_c_int(arg8, uint64_t);
  if (x8 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x9 = _cffi_to_c_int(arg9, uint64_t);
  if (x9 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  if (_cffi_to_c((char *)&x10, _cffi_type(702), arg10) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_import_BOOL(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_import_BOOL _cffi_d_GrB_Matrix_import_BOOL
#endif

static GrB_Info _cffi_d_GrB_Matrix_import_FP32(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t x3, uint64_t const * x4, uint64_t const * x5, float const * x6, uint64_t x7, uint64_t x8, uint64_t x9, GrB_Format x10)
{
  return GrB_Matrix_import_FP32(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_import_FP32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  uint64_t const * x4;
  uint64_t const * x5;
  float const * x6;
  uint64_t x7;
  uint64_t x8;
  uint64_t x9;
  GrB_Format x10;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_import_FP32", 11, 11, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(750), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (float const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(750), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x8 = _cffi_to_c_int(arg8, uint64_t);
  if (x8 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x9 = _cffi_to_c_int(arg9, uint64_t);
  if (x9 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  if (_cffi_to_c((char *)&x10, _cffi_type(702), arg10) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_import_FP32(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_import_FP32 _cffi_d_GrB_Matrix_import_FP32
#endif

static GrB_Info _cffi_d_GrB_Matrix_import_FP64(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t x3, uint64_t const * x4, uint64_t const * x5, double const * x6, uint64_t x7, uint64_t x8, uint64_t x9, GrB_Format x10)
{
  return GrB_Matrix_import_FP64(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_import_FP64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  uint64_t const * x4;
  uint64_t const * x5;
  double const * x6;
  uint64_t x7;
  uint64_t x8;
  uint64_t x9;
  GrB_Format x10;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_import_FP64", 11, 11, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(724), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (double const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(724), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x8 = _cffi_to_c_int(arg8, uint64_t);
  if (x8 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x9 = _cffi_to_c_int(arg9, uint64_t);
  if (x9 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  if (_cffi_to_c((char *)&x10, _cffi_type(702), arg10) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_import_FP64(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_import_FP64 _cffi_d_GrB_Matrix_import_FP64
#endif

static GrB_Info _cffi_d_GrB_Matrix_import_INT16(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t x3, uint64_t const * x4, uint64_t const * x5, int16_t const * x6, uint64_t x7, uint64_t x8, uint64_t x9, GrB_Format x10)
{
  return GrB_Matrix_import_INT16(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_import_INT16(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  uint64_t const * x4;
  uint64_t const * x5;
  int16_t const * x6;
  uint64_t x7;
  uint64_t x8;
  uint64_t x9;
  GrB_Format x10;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_import_INT16", 11, 11, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(763), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (int16_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(763), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x8 = _cffi_to_c_int(arg8, uint64_t);
  if (x8 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x9 = _cffi_to_c_int(arg9, uint64_t);
  if (x9 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  if (_cffi_to_c((char *)&x10, _cffi_type(702), arg10) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_import_INT16(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_import_INT16 _cffi_d_GrB_Matrix_import_INT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_import_INT32(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t x3, uint64_t const * x4, uint64_t const * x5, int32_t const * x6, uint64_t x7, uint64_t x8, uint64_t x9, GrB_Format x10)
{
  return GrB_Matrix_import_INT32(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_import_INT32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  uint64_t const * x4;
  uint64_t const * x5;
  int32_t const * x6;
  uint64_t x7;
  uint64_t x8;
  uint64_t x9;
  GrB_Format x10;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_import_INT32", 11, 11, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(776), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (int32_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(776), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x8 = _cffi_to_c_int(arg8, uint64_t);
  if (x8 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x9 = _cffi_to_c_int(arg9, uint64_t);
  if (x9 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  if (_cffi_to_c((char *)&x10, _cffi_type(702), arg10) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_import_INT32(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_import_INT32 _cffi_d_GrB_Matrix_import_INT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_import_INT64(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t x3, uint64_t const * x4, uint64_t const * x5, int64_t const * x6, uint64_t x7, uint64_t x8, uint64_t x9, GrB_Format x10)
{
  return GrB_Matrix_import_INT64(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_import_INT64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  uint64_t const * x4;
  uint64_t const * x5;
  int64_t const * x6;
  uint64_t x7;
  uint64_t x8;
  uint64_t x9;
  GrB_Format x10;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_import_INT64", 11, 11, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(789), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (int64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(789), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x8 = _cffi_to_c_int(arg8, uint64_t);
  if (x8 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x9 = _cffi_to_c_int(arg9, uint64_t);
  if (x9 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  if (_cffi_to_c((char *)&x10, _cffi_type(702), arg10) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_import_INT64(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_import_INT64 _cffi_d_GrB_Matrix_import_INT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_import_INT8(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t x3, uint64_t const * x4, uint64_t const * x5, int8_t const * x6, uint64_t x7, uint64_t x8, uint64_t x9, GrB_Format x10)
{
  return GrB_Matrix_import_INT8(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_import_INT8(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  uint64_t const * x4;
  uint64_t const * x5;
  int8_t const * x6;
  uint64_t x7;
  uint64_t x8;
  uint64_t x9;
  GrB_Format x10;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_import_INT8", 11, 11, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(802), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (int8_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(802), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x8 = _cffi_to_c_int(arg8, uint64_t);
  if (x8 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x9 = _cffi_to_c_int(arg9, uint64_t);
  if (x9 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  if (_cffi_to_c((char *)&x10, _cffi_type(702), arg10) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_import_INT8(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_import_INT8 _cffi_d_GrB_Matrix_import_INT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_import_UDT(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t x3, uint64_t const * x4, uint64_t const * x5, void const * x6, uint64_t x7, uint64_t x8, uint64_t x9, GrB_Format x10)
{
  return GrB_Matrix_import_UDT(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_import_UDT(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  uint64_t const * x4;
  uint64_t const * x5;
  void const * x6;
  uint64_t x7;
  uint64_t x8;
  uint64_t x9;
  GrB_Format x10;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_import_UDT", 11, 11, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(126), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (void const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(126), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x8 = _cffi_to_c_int(arg8, uint64_t);
  if (x8 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x9 = _cffi_to_c_int(arg9, uint64_t);
  if (x9 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  if (_cffi_to_c((char *)&x10, _cffi_type(702), arg10) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_import_UDT(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_import_UDT _cffi_d_GrB_Matrix_import_UDT
#endif

static GrB_Info _cffi_d_GrB_Matrix_import_UINT16(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t x3, uint64_t const * x4, uint64_t const * x5, uint16_t const * x6, uint64_t x7, uint64_t x8, uint64_t x9, GrB_Format x10)
{
  return GrB_Matrix_import_UINT16(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_import_UINT16(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  uint64_t const * x4;
  uint64_t const * x5;
  uint16_t const * x6;
  uint64_t x7;
  uint64_t x8;
  uint64_t x9;
  GrB_Format x10;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_import_UINT16", 11, 11, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(815), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint16_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(815), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x8 = _cffi_to_c_int(arg8, uint64_t);
  if (x8 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x9 = _cffi_to_c_int(arg9, uint64_t);
  if (x9 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  if (_cffi_to_c((char *)&x10, _cffi_type(702), arg10) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_import_UINT16(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_import_UINT16 _cffi_d_GrB_Matrix_import_UINT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_import_UINT32(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t x3, uint64_t const * x4, uint64_t const * x5, uint32_t const * x6, uint64_t x7, uint64_t x8, uint64_t x9, GrB_Format x10)
{
  return GrB_Matrix_import_UINT32(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_import_UINT32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  uint64_t const * x4;
  uint64_t const * x5;
  uint32_t const * x6;
  uint64_t x7;
  uint64_t x8;
  uint64_t x9;
  GrB_Format x10;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_import_UINT32", 11, 11, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(828), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint32_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(828), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x8 = _cffi_to_c_int(arg8, uint64_t);
  if (x8 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x9 = _cffi_to_c_int(arg9, uint64_t);
  if (x9 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  if (_cffi_to_c((char *)&x10, _cffi_type(702), arg10) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_import_UINT32(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_import_UINT32 _cffi_d_GrB_Matrix_import_UINT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_import_UINT64(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t x3, uint64_t const * x4, uint64_t const * x5, uint64_t const * x6, uint64_t x7, uint64_t x8, uint64_t x9, GrB_Format x10)
{
  return GrB_Matrix_import_UINT64(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_import_UINT64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  uint64_t const * x4;
  uint64_t const * x5;
  uint64_t const * x6;
  uint64_t x7;
  uint64_t x8;
  uint64_t x9;
  GrB_Format x10;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_import_UINT64", 11, 11, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x8 = _cffi_to_c_int(arg8, uint64_t);
  if (x8 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x9 = _cffi_to_c_int(arg9, uint64_t);
  if (x9 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  if (_cffi_to_c((char *)&x10, _cffi_type(702), arg10) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_import_UINT64(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_import_UINT64 _cffi_d_GrB_Matrix_import_UINT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_import_UINT8(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t x3, uint64_t const * x4, uint64_t const * x5, uint8_t const * x6, uint64_t x7, uint64_t x8, uint64_t x9, GrB_Format x10)
{
  return GrB_Matrix_import_UINT8(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_import_UINT8(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  uint64_t const * x4;
  uint64_t const * x5;
  uint8_t const * x6;
  uint64_t x7;
  uint64_t x8;
  uint64_t x9;
  GrB_Format x10;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_import_UINT8", 11, 11, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(854), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint8_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(854), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x8 = _cffi_to_c_int(arg8, uint64_t);
  if (x8 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x9 = _cffi_to_c_int(arg9, uint64_t);
  if (x9 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  if (_cffi_to_c((char *)&x10, _cffi_type(702), arg10) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_import_UINT8(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_import_UINT8 _cffi_d_GrB_Matrix_import_UINT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_kronecker_BinaryOp(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Matrix_opaque * x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_kronecker_BinaryOp(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_kronecker_BinaryOp(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_kronecker_BinaryOp", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_kronecker_BinaryOp(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_kronecker_BinaryOp _cffi_d_GrB_Matrix_kronecker_BinaryOp
#endif

static GrB_Info _cffi_d_GrB_Matrix_kronecker_Monoid(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Monoid_opaque * x3, struct GB_Matrix_opaque * x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_kronecker_Monoid(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_kronecker_Monoid(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Monoid_opaque * x3;
  struct GB_Matrix_opaque * x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_kronecker_Monoid", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_kronecker_Monoid(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_kronecker_Monoid _cffi_d_GrB_Matrix_kronecker_Monoid
#endif

static GrB_Info _cffi_d_GrB_Matrix_kronecker_Semiring(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Semiring_opaque * x3, struct GB_Matrix_opaque * x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_kronecker_Semiring(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_kronecker_Semiring(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Semiring_opaque * x3;
  struct GB_Matrix_opaque * x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_kronecker_Semiring", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(159), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Semiring_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(159), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_kronecker_Semiring(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_kronecker_Semiring _cffi_d_GrB_Matrix_kronecker_Semiring
#endif

static GrB_Info _cffi_d_GrB_Matrix_ncols(uint64_t * x0, struct GB_Matrix_opaque * x1)
{
  return GrB_Matrix_ncols(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_ncols(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  struct GB_Matrix_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_ncols", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_ncols(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_ncols _cffi_d_GrB_Matrix_ncols
#endif

static GrB_Info _cffi_d_GrB_Matrix_new(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_new(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_new(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_new", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_new(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_new _cffi_d_GrB_Matrix_new
#endif

static GrB_Info _cffi_d_GrB_Matrix_nrows(uint64_t * x0, struct GB_Matrix_opaque * x1)
{
  return GrB_Matrix_nrows(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_nrows(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  struct GB_Matrix_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_nrows", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_nrows(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_nrows _cffi_d_GrB_Matrix_nrows
#endif

static GrB_Info _cffi_d_GrB_Matrix_nvals(uint64_t * x0, struct GB_Matrix_opaque * x1)
{
  return GrB_Matrix_nvals(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_nvals(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  struct GB_Matrix_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_nvals", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_nvals(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_nvals _cffi_d_GrB_Matrix_nvals
#endif

static GrB_Info _cffi_d_GrB_Matrix_reduce_BOOL(_Bool * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Matrix_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Matrix_reduce_BOOL(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_reduce_BOOL(PyObject *self, PyObject *args)
{
  _Bool * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Matrix_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_reduce_BOOL", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_reduce_BOOL(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_reduce_BOOL _cffi_d_GrB_Matrix_reduce_BOOL
#endif

static GrB_Info _cffi_d_GrB_Matrix_reduce_BinaryOp(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Matrix_opaque * x4, struct GB_Descriptor_opaque * x5)
{
  return GrB_Matrix_reduce_BinaryOp(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_reduce_BinaryOp(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  struct GB_Descriptor_opaque * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_reduce_BinaryOp", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_reduce_BinaryOp(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_reduce_BinaryOp _cffi_d_GrB_Matrix_reduce_BinaryOp
#endif

static GrB_Info _cffi_d_GrB_Matrix_reduce_BinaryOp_Scalar(struct GB_Scalar_opaque * x0, struct GB_BinaryOp_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Matrix_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Matrix_reduce_BinaryOp_Scalar(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_reduce_BinaryOp_Scalar(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Matrix_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_reduce_BinaryOp_Scalar", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_reduce_BinaryOp_Scalar(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_reduce_BinaryOp_Scalar _cffi_d_GrB_Matrix_reduce_BinaryOp_Scalar
#endif

static GrB_Info _cffi_d_GrB_Matrix_reduce_FP32(float * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Matrix_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Matrix_reduce_FP32(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_reduce_FP32(PyObject *self, PyObject *args)
{
  float * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Matrix_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_reduce_FP32", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(232), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (float *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(232), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_reduce_FP32(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_reduce_FP32 _cffi_d_GrB_Matrix_reduce_FP32
#endif

static GrB_Info _cffi_d_GrB_Matrix_reduce_FP64(double * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Matrix_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Matrix_reduce_FP64(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_reduce_FP64(PyObject *self, PyObject *args)
{
  double * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Matrix_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_reduce_FP64", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(28), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(28), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_reduce_FP64(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_reduce_FP64 _cffi_d_GrB_Matrix_reduce_FP64
#endif

static GrB_Info _cffi_d_GrB_Matrix_reduce_INT16(int16_t * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Matrix_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Matrix_reduce_INT16(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_reduce_INT16(PyObject *self, PyObject *args)
{
  int16_t * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Matrix_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_reduce_INT16", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(290), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int16_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(290), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_reduce_INT16(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_reduce_INT16 _cffi_d_GrB_Matrix_reduce_INT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_reduce_INT32(int32_t * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Matrix_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Matrix_reduce_INT32(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_reduce_INT32(PyObject *self, PyObject *args)
{
  int32_t * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Matrix_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_reduce_INT32", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(36), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int32_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(36), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_reduce_INT32(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_reduce_INT32 _cffi_d_GrB_Matrix_reduce_INT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_reduce_INT64(int64_t * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Matrix_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Matrix_reduce_INT64(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_reduce_INT64(PyObject *self, PyObject *args)
{
  int64_t * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Matrix_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_reduce_INT64", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(44), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(44), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_reduce_INT64(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_reduce_INT64 _cffi_d_GrB_Matrix_reduce_INT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_reduce_INT8(int8_t * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Matrix_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Matrix_reduce_INT8(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_reduce_INT8(PyObject *self, PyObject *args)
{
  int8_t * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Matrix_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_reduce_INT8", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(377), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int8_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(377), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_reduce_INT8(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_reduce_INT8 _cffi_d_GrB_Matrix_reduce_INT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_reduce_Monoid(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Monoid_opaque * x3, struct GB_Matrix_opaque * x4, struct GB_Descriptor_opaque * x5)
{
  return GrB_Matrix_reduce_Monoid(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_reduce_Monoid(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Monoid_opaque * x3;
  struct GB_Matrix_opaque * x4;
  struct GB_Descriptor_opaque * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_reduce_Monoid", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_reduce_Monoid(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_reduce_Monoid _cffi_d_GrB_Matrix_reduce_Monoid
#endif

static GrB_Info _cffi_d_GrB_Matrix_reduce_Monoid_Scalar(struct GB_Scalar_opaque * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Matrix_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Matrix_reduce_Monoid_Scalar(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_reduce_Monoid_Scalar(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Matrix_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_reduce_Monoid_Scalar", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_reduce_Monoid_Scalar(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_reduce_Monoid_Scalar _cffi_d_GrB_Matrix_reduce_Monoid_Scalar
#endif

static GrB_Info _cffi_d_GrB_Matrix_reduce_UDT(void * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Matrix_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Matrix_reduce_UDT(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_reduce_UDT(PyObject *self, PyObject *args)
{
  void * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Matrix_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_reduce_UDT", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_reduce_UDT(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_reduce_UDT _cffi_d_GrB_Matrix_reduce_UDT
#endif

static GrB_Info _cffi_d_GrB_Matrix_reduce_UINT16(uint16_t * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Matrix_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Matrix_reduce_UINT16(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_reduce_UINT16(PyObject *self, PyObject *args)
{
  uint16_t * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Matrix_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_reduce_UINT16", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3386), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint16_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3386), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_reduce_UINT16(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_reduce_UINT16 _cffi_d_GrB_Matrix_reduce_UINT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_reduce_UINT32(uint32_t * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Matrix_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Matrix_reduce_UINT32(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_reduce_UINT32(PyObject *self, PyObject *args)
{
  uint32_t * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Matrix_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_reduce_UINT32", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3415), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint32_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3415), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_reduce_UINT32(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_reduce_UINT32 _cffi_d_GrB_Matrix_reduce_UINT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_reduce_UINT64(uint64_t * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Matrix_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Matrix_reduce_UINT64(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_reduce_UINT64(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Matrix_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_reduce_UINT64", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_reduce_UINT64(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_reduce_UINT64 _cffi_d_GrB_Matrix_reduce_UINT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_reduce_UINT8(uint8_t * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Matrix_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Matrix_reduce_UINT8(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_reduce_UINT8(PyObject *self, PyObject *args)
{
  uint8_t * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Matrix_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_reduce_UINT8", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3766), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint8_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3766), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_reduce_UINT8(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_reduce_UINT8 _cffi_d_GrB_Matrix_reduce_UINT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_removeElement(struct GB_Matrix_opaque * x0, uint64_t x1, uint64_t x2)
{
  return GrB_Matrix_removeElement(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_removeElement(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_removeElement", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint64_t);
  if (x1 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_removeElement(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_removeElement _cffi_d_GrB_Matrix_removeElement
#endif

static GrB_Info _cffi_d_GrB_Matrix_resize(struct GB_Matrix_opaque * x0, uint64_t x1, uint64_t x2)
{
  return GrB_Matrix_resize(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_resize(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_resize", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint64_t);
  if (x1 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_resize(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_resize _cffi_d_GrB_Matrix_resize
#endif

static GrB_Info _cffi_d_GrB_Matrix_select_BOOL(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, _Bool x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_select_BOOL(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_select_BOOL(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  _Bool x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_select_BOOL", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (_Bool)_cffi_to_c__Bool(arg5);
  if (x5 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_select_BOOL(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_select_BOOL _cffi_d_GrB_Matrix_select_BOOL
#endif

static GrB_Info _cffi_d_GrB_Matrix_select_FP32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, float x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_select_FP32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_select_FP32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  float x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_select_FP32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (float)_cffi_to_c_float(arg5);
  if (x5 == (float)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_select_FP32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_select_FP32 _cffi_d_GrB_Matrix_select_FP32
#endif

static GrB_Info _cffi_d_GrB_Matrix_select_FP64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, double x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_select_FP64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_select_FP64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  double x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_select_FP64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (double)_cffi_to_c_double(arg5);
  if (x5 == (double)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_select_FP64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_select_FP64 _cffi_d_GrB_Matrix_select_FP64
#endif

static GrB_Info _cffi_d_GrB_Matrix_select_INT16(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, int16_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_select_INT16(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_select_INT16(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  int16_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_select_INT16", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int16_t);
  if (x5 == (int16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_select_INT16(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_select_INT16 _cffi_d_GrB_Matrix_select_INT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_select_INT32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, int32_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_select_INT32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_select_INT32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  int32_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_select_INT32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int32_t);
  if (x5 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_select_INT32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_select_INT32 _cffi_d_GrB_Matrix_select_INT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_select_INT64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, int64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_select_INT64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_select_INT64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  int64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_select_INT64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int64_t);
  if (x5 == (int64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_select_INT64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_select_INT64 _cffi_d_GrB_Matrix_select_INT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_select_INT8(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, int8_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_select_INT8(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_select_INT8(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  int8_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_select_INT8", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int8_t);
  if (x5 == (int8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_select_INT8(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_select_INT8 _cffi_d_GrB_Matrix_select_INT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_select_Scalar(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, struct GB_Scalar_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_select_Scalar(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_select_Scalar(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  struct GB_Scalar_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_select_Scalar", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_select_Scalar(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_select_Scalar _cffi_d_GrB_Matrix_select_Scalar
#endif

static GrB_Info _cffi_d_GrB_Matrix_select_UDT(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, void const * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_select_UDT(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_select_UDT(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  void const * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_select_UDT", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(126), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (void const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(126), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_select_UDT(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_select_UDT _cffi_d_GrB_Matrix_select_UDT
#endif

static GrB_Info _cffi_d_GrB_Matrix_select_UINT16(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, uint16_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_select_UINT16(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_select_UINT16(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  uint16_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_select_UINT16", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint16_t);
  if (x5 == (uint16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_select_UINT16(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_select_UINT16 _cffi_d_GrB_Matrix_select_UINT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_select_UINT32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, uint32_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_select_UINT32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_select_UINT32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  uint32_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_select_UINT32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint32_t);
  if (x5 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_select_UINT32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_select_UINT32 _cffi_d_GrB_Matrix_select_UINT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_select_UINT64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_select_UINT64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_select_UINT64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_select_UINT64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_select_UINT64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_select_UINT64 _cffi_d_GrB_Matrix_select_UINT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_select_UINT8(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, uint8_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Matrix_select_UINT8(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_select_UINT8(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  uint8_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_select_UINT8", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint8_t);
  if (x5 == (uint8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_select_UINT8(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_select_UINT8 _cffi_d_GrB_Matrix_select_UINT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_serialize(void * x0, uint64_t * x1, struct GB_Matrix_opaque * x2)
{
  return GrB_Matrix_serialize(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_serialize(PyObject *self, PyObject *args)
{
  void * x0;
  uint64_t * x1;
  struct GB_Matrix_opaque * x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_serialize", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_serialize(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_serialize _cffi_d_GrB_Matrix_serialize
#endif

static GrB_Info _cffi_d_GrB_Matrix_serializeSize(uint64_t * x0, struct GB_Matrix_opaque * x1)
{
  return GrB_Matrix_serializeSize(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_serializeSize(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  struct GB_Matrix_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_serializeSize", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_serializeSize(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_serializeSize _cffi_d_GrB_Matrix_serializeSize
#endif

static GrB_Info _cffi_d_GrB_Matrix_setElement_BOOL(struct GB_Matrix_opaque * x0, _Bool x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_setElement_BOOL(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_setElement_BOOL(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  _Bool x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_setElement_BOOL", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = (_Bool)_cffi_to_c__Bool(arg1);
  if (x1 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_setElement_BOOL(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_setElement_BOOL _cffi_d_GrB_Matrix_setElement_BOOL
#endif

static GrB_Info _cffi_d_GrB_Matrix_setElement_FP32(struct GB_Matrix_opaque * x0, float x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_setElement_FP32(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_setElement_FP32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  float x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_setElement_FP32", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = (float)_cffi_to_c_float(arg1);
  if (x1 == (float)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_setElement_FP32(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_setElement_FP32 _cffi_d_GrB_Matrix_setElement_FP32
#endif

static GrB_Info _cffi_d_GrB_Matrix_setElement_FP64(struct GB_Matrix_opaque * x0, double x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_setElement_FP64(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_setElement_FP64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  double x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_setElement_FP64", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = (double)_cffi_to_c_double(arg1);
  if (x1 == (double)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_setElement_FP64(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_setElement_FP64 _cffi_d_GrB_Matrix_setElement_FP64
#endif

static GrB_Info _cffi_d_GrB_Matrix_setElement_INT16(struct GB_Matrix_opaque * x0, int16_t x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_setElement_INT16(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_setElement_INT16(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  int16_t x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_setElement_INT16", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, int16_t);
  if (x1 == (int16_t)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_setElement_INT16(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_setElement_INT16 _cffi_d_GrB_Matrix_setElement_INT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_setElement_INT32(struct GB_Matrix_opaque * x0, int32_t x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_setElement_INT32(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_setElement_INT32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  int32_t x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_setElement_INT32", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, int32_t);
  if (x1 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_setElement_INT32(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_setElement_INT32 _cffi_d_GrB_Matrix_setElement_INT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_setElement_INT64(struct GB_Matrix_opaque * x0, int64_t x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_setElement_INT64(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_setElement_INT64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  int64_t x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_setElement_INT64", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, int64_t);
  if (x1 == (int64_t)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_setElement_INT64(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_setElement_INT64 _cffi_d_GrB_Matrix_setElement_INT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_setElement_INT8(struct GB_Matrix_opaque * x0, int8_t x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_setElement_INT8(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_setElement_INT8(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  int8_t x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_setElement_INT8", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, int8_t);
  if (x1 == (int8_t)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_setElement_INT8(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_setElement_INT8 _cffi_d_GrB_Matrix_setElement_INT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_setElement_Scalar(struct GB_Matrix_opaque * x0, struct GB_Scalar_opaque * x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_setElement_Scalar(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_setElement_Scalar(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Scalar_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_setElement_Scalar", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_setElement_Scalar(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_setElement_Scalar _cffi_d_GrB_Matrix_setElement_Scalar
#endif

static GrB_Info _cffi_d_GrB_Matrix_setElement_UDT(struct GB_Matrix_opaque * x0, void * x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_setElement_UDT(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_setElement_UDT(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  void * x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_setElement_UDT", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_setElement_UDT(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_setElement_UDT _cffi_d_GrB_Matrix_setElement_UDT
#endif

static GrB_Info _cffi_d_GrB_Matrix_setElement_UINT16(struct GB_Matrix_opaque * x0, uint16_t x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_setElement_UINT16(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_setElement_UINT16(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint16_t x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_setElement_UINT16", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint16_t);
  if (x1 == (uint16_t)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_setElement_UINT16(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_setElement_UINT16 _cffi_d_GrB_Matrix_setElement_UINT16
#endif

static GrB_Info _cffi_d_GrB_Matrix_setElement_UINT32(struct GB_Matrix_opaque * x0, uint32_t x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_setElement_UINT32(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_setElement_UINT32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint32_t x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_setElement_UINT32", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint32_t);
  if (x1 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_setElement_UINT32(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_setElement_UINT32 _cffi_d_GrB_Matrix_setElement_UINT32
#endif

static GrB_Info _cffi_d_GrB_Matrix_setElement_UINT64(struct GB_Matrix_opaque * x0, uint64_t x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_setElement_UINT64(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_setElement_UINT64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_setElement_UINT64", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint64_t);
  if (x1 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_setElement_UINT64(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_setElement_UINT64 _cffi_d_GrB_Matrix_setElement_UINT64
#endif

static GrB_Info _cffi_d_GrB_Matrix_setElement_UINT8(struct GB_Matrix_opaque * x0, uint8_t x1, uint64_t x2, uint64_t x3)
{
  return GrB_Matrix_setElement_UINT8(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_setElement_UINT8(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint8_t x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_setElement_UINT8", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint8_t);
  if (x1 == (uint8_t)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_setElement_UINT8(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_setElement_UINT8 _cffi_d_GrB_Matrix_setElement_UINT8
#endif

static GrB_Info _cffi_d_GrB_Matrix_wait(struct GB_Matrix_opaque * x0, GrB_WaitMode x1)
{
  return GrB_Matrix_wait(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Matrix_wait(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  GrB_WaitMode x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Matrix_wait", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(458), arg1) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Matrix_wait(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Matrix_wait _cffi_d_GrB_Matrix_wait
#endif

static GrB_Info _cffi_d_GrB_Monoid_error(char const * * x0, struct GB_Monoid_opaque * x1)
{
  return GrB_Monoid_error(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Monoid_error(PyObject *self, PyObject *args)
{
  char const * * x0;
  struct GB_Monoid_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Monoid_error", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(130), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char const * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(130), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Monoid_error(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Monoid_error _cffi_d_GrB_Monoid_error
#endif

static GrB_Info _cffi_d_GrB_Monoid_free(struct GB_Monoid_opaque * * x0)
{
  return GrB_Monoid_free(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Monoid_free(PyObject *self, PyObject *arg0)
{
  struct GB_Monoid_opaque * * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Monoid_free(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Monoid_free _cffi_d_GrB_Monoid_free
#endif

static GrB_Info _cffi_d_GrB_Monoid_new_BOOL(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, _Bool x2)
{
  return GrB_Monoid_new_BOOL(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Monoid_new_BOOL(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  _Bool x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Monoid_new_BOOL", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = (_Bool)_cffi_to_c__Bool(arg2);
  if (x2 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Monoid_new_BOOL(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Monoid_new_BOOL _cffi_d_GrB_Monoid_new_BOOL
#endif

static GrB_Info _cffi_d_GrB_Monoid_new_FP32(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, float x2)
{
  return GrB_Monoid_new_FP32(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Monoid_new_FP32(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  float x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Monoid_new_FP32", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = (float)_cffi_to_c_float(arg2);
  if (x2 == (float)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Monoid_new_FP32(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Monoid_new_FP32 _cffi_d_GrB_Monoid_new_FP32
#endif

static GrB_Info _cffi_d_GrB_Monoid_new_FP64(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, double x2)
{
  return GrB_Monoid_new_FP64(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Monoid_new_FP64(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  double x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Monoid_new_FP64", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = (double)_cffi_to_c_double(arg2);
  if (x2 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Monoid_new_FP64(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Monoid_new_FP64 _cffi_d_GrB_Monoid_new_FP64
#endif

static GrB_Info _cffi_d_GrB_Monoid_new_INT16(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, int16_t x2)
{
  return GrB_Monoid_new_INT16(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Monoid_new_INT16(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  int16_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Monoid_new_INT16", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, int16_t);
  if (x2 == (int16_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Monoid_new_INT16(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Monoid_new_INT16 _cffi_d_GrB_Monoid_new_INT16
#endif

static GrB_Info _cffi_d_GrB_Monoid_new_INT32(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, int32_t x2)
{
  return GrB_Monoid_new_INT32(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Monoid_new_INT32(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  int32_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Monoid_new_INT32", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, int32_t);
  if (x2 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Monoid_new_INT32(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Monoid_new_INT32 _cffi_d_GrB_Monoid_new_INT32
#endif

static GrB_Info _cffi_d_GrB_Monoid_new_INT64(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, int64_t x2)
{
  return GrB_Monoid_new_INT64(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Monoid_new_INT64(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  int64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Monoid_new_INT64", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, int64_t);
  if (x2 == (int64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Monoid_new_INT64(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Monoid_new_INT64 _cffi_d_GrB_Monoid_new_INT64
#endif

static GrB_Info _cffi_d_GrB_Monoid_new_INT8(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, int8_t x2)
{
  return GrB_Monoid_new_INT8(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Monoid_new_INT8(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  int8_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Monoid_new_INT8", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, int8_t);
  if (x2 == (int8_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Monoid_new_INT8(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Monoid_new_INT8 _cffi_d_GrB_Monoid_new_INT8
#endif

static GrB_Info _cffi_d_GrB_Monoid_new_UDT(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, void * x2)
{
  return GrB_Monoid_new_UDT(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Monoid_new_UDT(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  void * x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Monoid_new_UDT", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Monoid_new_UDT(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Monoid_new_UDT _cffi_d_GrB_Monoid_new_UDT
#endif

static GrB_Info _cffi_d_GrB_Monoid_new_UINT16(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, uint16_t x2)
{
  return GrB_Monoid_new_UINT16(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Monoid_new_UINT16(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  uint16_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Monoid_new_UINT16", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint16_t);
  if (x2 == (uint16_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Monoid_new_UINT16(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Monoid_new_UINT16 _cffi_d_GrB_Monoid_new_UINT16
#endif

static GrB_Info _cffi_d_GrB_Monoid_new_UINT32(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, uint32_t x2)
{
  return GrB_Monoid_new_UINT32(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Monoid_new_UINT32(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  uint32_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Monoid_new_UINT32", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint32_t);
  if (x2 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Monoid_new_UINT32(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Monoid_new_UINT32 _cffi_d_GrB_Monoid_new_UINT32
#endif

static GrB_Info _cffi_d_GrB_Monoid_new_UINT64(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, uint64_t x2)
{
  return GrB_Monoid_new_UINT64(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Monoid_new_UINT64(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Monoid_new_UINT64", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Monoid_new_UINT64(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Monoid_new_UINT64 _cffi_d_GrB_Monoid_new_UINT64
#endif

static GrB_Info _cffi_d_GrB_Monoid_new_UINT8(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, uint8_t x2)
{
  return GrB_Monoid_new_UINT8(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Monoid_new_UINT8(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  uint8_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Monoid_new_UINT8", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint8_t);
  if (x2 == (uint8_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Monoid_new_UINT8(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Monoid_new_UINT8 _cffi_d_GrB_Monoid_new_UINT8
#endif

static GrB_Info _cffi_d_GrB_Monoid_wait(struct GB_Monoid_opaque * x0, GrB_WaitMode x1)
{
  return GrB_Monoid_wait(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Monoid_wait(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * x0;
  GrB_WaitMode x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Monoid_wait", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(458), arg1) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Monoid_wait(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Monoid_wait _cffi_d_GrB_Monoid_wait
#endif

static GrB_Info _cffi_d_GrB_Row_assign(struct GB_Matrix_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Vector_opaque * x3, uint64_t x4, uint64_t const * x5, uint64_t x6, struct GB_Descriptor_opaque * x7)
{
  return GrB_Row_assign(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Row_assign(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Vector_opaque * x3;
  uint64_t x4;
  uint64_t const * x5;
  uint64_t x6;
  struct GB_Descriptor_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GrB_Row_assign", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x6 = _cffi_to_c_int(arg6, uint64_t);
  if (x6 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Row_assign(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Row_assign _cffi_d_GrB_Row_assign
#endif

static GrB_Info _cffi_d_GrB_Scalar_clear(struct GB_Scalar_opaque * x0)
{
  return GrB_Scalar_clear(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_clear(PyObject *self, PyObject *arg0)
{
  struct GB_Scalar_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_clear(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_clear _cffi_d_GrB_Scalar_clear
#endif

static GrB_Info _cffi_d_GrB_Scalar_dup(struct GB_Scalar_opaque * * x0, struct GB_Scalar_opaque * x1)
{
  return GrB_Scalar_dup(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_dup(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_dup", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2109), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2109), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_dup(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_dup _cffi_d_GrB_Scalar_dup
#endif

static GrB_Info _cffi_d_GrB_Scalar_error(char const * * x0, struct GB_Scalar_opaque * x1)
{
  return GrB_Scalar_error(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_error(PyObject *self, PyObject *args)
{
  char const * * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_error", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(130), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char const * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(130), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_error(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_error _cffi_d_GrB_Scalar_error
#endif

static GrB_Info _cffi_d_GrB_Scalar_extractElement_BOOL(_Bool * x0, struct GB_Scalar_opaque * x1)
{
  return GrB_Scalar_extractElement_BOOL(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_extractElement_BOOL(PyObject *self, PyObject *args)
{
  _Bool * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_extractElement_BOOL", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_extractElement_BOOL(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_extractElement_BOOL _cffi_d_GrB_Scalar_extractElement_BOOL
#endif

static GrB_Info _cffi_d_GrB_Scalar_extractElement_FP32(float * x0, struct GB_Scalar_opaque * x1)
{
  return GrB_Scalar_extractElement_FP32(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_extractElement_FP32(PyObject *self, PyObject *args)
{
  float * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_extractElement_FP32", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(232), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (float *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(232), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_extractElement_FP32(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_extractElement_FP32 _cffi_d_GrB_Scalar_extractElement_FP32
#endif

static GrB_Info _cffi_d_GrB_Scalar_extractElement_FP64(double * x0, struct GB_Scalar_opaque * x1)
{
  return GrB_Scalar_extractElement_FP64(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_extractElement_FP64(PyObject *self, PyObject *args)
{
  double * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_extractElement_FP64", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(28), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(28), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_extractElement_FP64(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_extractElement_FP64 _cffi_d_GrB_Scalar_extractElement_FP64
#endif

static GrB_Info _cffi_d_GrB_Scalar_extractElement_INT16(int16_t * x0, struct GB_Scalar_opaque * x1)
{
  return GrB_Scalar_extractElement_INT16(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_extractElement_INT16(PyObject *self, PyObject *args)
{
  int16_t * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_extractElement_INT16", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(290), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int16_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(290), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_extractElement_INT16(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_extractElement_INT16 _cffi_d_GrB_Scalar_extractElement_INT16
#endif

static GrB_Info _cffi_d_GrB_Scalar_extractElement_INT32(int32_t * x0, struct GB_Scalar_opaque * x1)
{
  return GrB_Scalar_extractElement_INT32(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_extractElement_INT32(PyObject *self, PyObject *args)
{
  int32_t * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_extractElement_INT32", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(36), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int32_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(36), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_extractElement_INT32(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_extractElement_INT32 _cffi_d_GrB_Scalar_extractElement_INT32
#endif

static GrB_Info _cffi_d_GrB_Scalar_extractElement_INT64(int64_t * x0, struct GB_Scalar_opaque * x1)
{
  return GrB_Scalar_extractElement_INT64(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_extractElement_INT64(PyObject *self, PyObject *args)
{
  int64_t * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_extractElement_INT64", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(44), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(44), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_extractElement_INT64(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_extractElement_INT64 _cffi_d_GrB_Scalar_extractElement_INT64
#endif

static GrB_Info _cffi_d_GrB_Scalar_extractElement_INT8(int8_t * x0, struct GB_Scalar_opaque * x1)
{
  return GrB_Scalar_extractElement_INT8(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_extractElement_INT8(PyObject *self, PyObject *args)
{
  int8_t * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_extractElement_INT8", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(377), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int8_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(377), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_extractElement_INT8(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_extractElement_INT8 _cffi_d_GrB_Scalar_extractElement_INT8
#endif

static GrB_Info _cffi_d_GrB_Scalar_extractElement_UDT(void * x0, struct GB_Scalar_opaque * x1)
{
  return GrB_Scalar_extractElement_UDT(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_extractElement_UDT(PyObject *self, PyObject *args)
{
  void * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_extractElement_UDT", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_extractElement_UDT(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_extractElement_UDT _cffi_d_GrB_Scalar_extractElement_UDT
#endif

static GrB_Info _cffi_d_GrB_Scalar_extractElement_UINT16(uint16_t * x0, struct GB_Scalar_opaque * x1)
{
  return GrB_Scalar_extractElement_UINT16(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_extractElement_UINT16(PyObject *self, PyObject *args)
{
  uint16_t * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_extractElement_UINT16", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3386), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint16_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3386), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_extractElement_UINT16(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_extractElement_UINT16 _cffi_d_GrB_Scalar_extractElement_UINT16
#endif

static GrB_Info _cffi_d_GrB_Scalar_extractElement_UINT32(uint32_t * x0, struct GB_Scalar_opaque * x1)
{
  return GrB_Scalar_extractElement_UINT32(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_extractElement_UINT32(PyObject *self, PyObject *args)
{
  uint32_t * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_extractElement_UINT32", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3415), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint32_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3415), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_extractElement_UINT32(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_extractElement_UINT32 _cffi_d_GrB_Scalar_extractElement_UINT32
#endif

static GrB_Info _cffi_d_GrB_Scalar_extractElement_UINT64(uint64_t * x0, struct GB_Scalar_opaque * x1)
{
  return GrB_Scalar_extractElement_UINT64(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_extractElement_UINT64(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_extractElement_UINT64", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_extractElement_UINT64(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_extractElement_UINT64 _cffi_d_GrB_Scalar_extractElement_UINT64
#endif

static GrB_Info _cffi_d_GrB_Scalar_extractElement_UINT8(uint8_t * x0, struct GB_Scalar_opaque * x1)
{
  return GrB_Scalar_extractElement_UINT8(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_extractElement_UINT8(PyObject *self, PyObject *args)
{
  uint8_t * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_extractElement_UINT8", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3766), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint8_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3766), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_extractElement_UINT8(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_extractElement_UINT8 _cffi_d_GrB_Scalar_extractElement_UINT8
#endif

static GrB_Info _cffi_d_GrB_Scalar_free(struct GB_Scalar_opaque * * x0)
{
  return GrB_Scalar_free(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_free(PyObject *self, PyObject *arg0)
{
  struct GB_Scalar_opaque * * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2109), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2109), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_free(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_free _cffi_d_GrB_Scalar_free
#endif

static GrB_Info _cffi_d_GrB_Scalar_new(struct GB_Scalar_opaque * * x0, struct GB_Type_opaque * x1)
{
  return GrB_Scalar_new(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_new(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * * x0;
  struct GB_Type_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_new", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2109), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2109), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_new(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_new _cffi_d_GrB_Scalar_new
#endif

static GrB_Info _cffi_d_GrB_Scalar_nvals(uint64_t * x0, struct GB_Scalar_opaque * x1)
{
  return GrB_Scalar_nvals(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_nvals(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_nvals", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_nvals(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_nvals _cffi_d_GrB_Scalar_nvals
#endif

static GrB_Info _cffi_d_GrB_Scalar_setElement_BOOL(struct GB_Scalar_opaque * x0, _Bool x1)
{
  return GrB_Scalar_setElement_BOOL(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_setElement_BOOL(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  _Bool x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_setElement_BOOL", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = (_Bool)_cffi_to_c__Bool(arg1);
  if (x1 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_setElement_BOOL(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_setElement_BOOL _cffi_d_GrB_Scalar_setElement_BOOL
#endif

static GrB_Info _cffi_d_GrB_Scalar_setElement_FP32(struct GB_Scalar_opaque * x0, float x1)
{
  return GrB_Scalar_setElement_FP32(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_setElement_FP32(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  float x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_setElement_FP32", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = (float)_cffi_to_c_float(arg1);
  if (x1 == (float)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_setElement_FP32(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_setElement_FP32 _cffi_d_GrB_Scalar_setElement_FP32
#endif

static GrB_Info _cffi_d_GrB_Scalar_setElement_FP64(struct GB_Scalar_opaque * x0, double x1)
{
  return GrB_Scalar_setElement_FP64(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_setElement_FP64(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  double x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_setElement_FP64", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = (double)_cffi_to_c_double(arg1);
  if (x1 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_setElement_FP64(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_setElement_FP64 _cffi_d_GrB_Scalar_setElement_FP64
#endif

static GrB_Info _cffi_d_GrB_Scalar_setElement_INT16(struct GB_Scalar_opaque * x0, int16_t x1)
{
  return GrB_Scalar_setElement_INT16(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_setElement_INT16(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  int16_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_setElement_INT16", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, int16_t);
  if (x1 == (int16_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_setElement_INT16(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_setElement_INT16 _cffi_d_GrB_Scalar_setElement_INT16
#endif

static GrB_Info _cffi_d_GrB_Scalar_setElement_INT32(struct GB_Scalar_opaque * x0, int32_t x1)
{
  return GrB_Scalar_setElement_INT32(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_setElement_INT32(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  int32_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_setElement_INT32", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, int32_t);
  if (x1 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_setElement_INT32(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_setElement_INT32 _cffi_d_GrB_Scalar_setElement_INT32
#endif

static GrB_Info _cffi_d_GrB_Scalar_setElement_INT64(struct GB_Scalar_opaque * x0, int64_t x1)
{
  return GrB_Scalar_setElement_INT64(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_setElement_INT64(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  int64_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_setElement_INT64", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, int64_t);
  if (x1 == (int64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_setElement_INT64(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_setElement_INT64 _cffi_d_GrB_Scalar_setElement_INT64
#endif

static GrB_Info _cffi_d_GrB_Scalar_setElement_INT8(struct GB_Scalar_opaque * x0, int8_t x1)
{
  return GrB_Scalar_setElement_INT8(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_setElement_INT8(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  int8_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_setElement_INT8", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, int8_t);
  if (x1 == (int8_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_setElement_INT8(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_setElement_INT8 _cffi_d_GrB_Scalar_setElement_INT8
#endif

static GrB_Info _cffi_d_GrB_Scalar_setElement_UDT(struct GB_Scalar_opaque * x0, void * x1)
{
  return GrB_Scalar_setElement_UDT(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_setElement_UDT(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  void * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_setElement_UDT", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_setElement_UDT(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_setElement_UDT _cffi_d_GrB_Scalar_setElement_UDT
#endif

static GrB_Info _cffi_d_GrB_Scalar_setElement_UINT16(struct GB_Scalar_opaque * x0, uint16_t x1)
{
  return GrB_Scalar_setElement_UINT16(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_setElement_UINT16(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  uint16_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_setElement_UINT16", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint16_t);
  if (x1 == (uint16_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_setElement_UINT16(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_setElement_UINT16 _cffi_d_GrB_Scalar_setElement_UINT16
#endif

static GrB_Info _cffi_d_GrB_Scalar_setElement_UINT32(struct GB_Scalar_opaque * x0, uint32_t x1)
{
  return GrB_Scalar_setElement_UINT32(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_setElement_UINT32(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  uint32_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_setElement_UINT32", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint32_t);
  if (x1 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_setElement_UINT32(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_setElement_UINT32 _cffi_d_GrB_Scalar_setElement_UINT32
#endif

static GrB_Info _cffi_d_GrB_Scalar_setElement_UINT64(struct GB_Scalar_opaque * x0, uint64_t x1)
{
  return GrB_Scalar_setElement_UINT64(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_setElement_UINT64(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  uint64_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_setElement_UINT64", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint64_t);
  if (x1 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_setElement_UINT64(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_setElement_UINT64 _cffi_d_GrB_Scalar_setElement_UINT64
#endif

static GrB_Info _cffi_d_GrB_Scalar_setElement_UINT8(struct GB_Scalar_opaque * x0, uint8_t x1)
{
  return GrB_Scalar_setElement_UINT8(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_setElement_UINT8(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  uint8_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_setElement_UINT8", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint8_t);
  if (x1 == (uint8_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_setElement_UINT8(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_setElement_UINT8 _cffi_d_GrB_Scalar_setElement_UINT8
#endif

static GrB_Info _cffi_d_GrB_Scalar_wait(struct GB_Scalar_opaque * x0, GrB_WaitMode x1)
{
  return GrB_Scalar_wait(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Scalar_wait(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  GrB_WaitMode x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Scalar_wait", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(458), arg1) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Scalar_wait(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Scalar_wait _cffi_d_GrB_Scalar_wait
#endif

static GrB_Info _cffi_d_GrB_Semiring_error(char const * * x0, struct GB_Semiring_opaque * x1)
{
  return GrB_Semiring_error(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Semiring_error(PyObject *self, PyObject *args)
{
  char const * * x0;
  struct GB_Semiring_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Semiring_error", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(130), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char const * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(130), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(159), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Semiring_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(159), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Semiring_error(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Semiring_error _cffi_d_GrB_Semiring_error
#endif

static GrB_Info _cffi_d_GrB_Semiring_free(struct GB_Semiring_opaque * * x0)
{
  return GrB_Semiring_free(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Semiring_free(PyObject *self, PyObject *arg0)
{
  struct GB_Semiring_opaque * * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2254), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Semiring_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2254), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Semiring_free(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Semiring_free _cffi_d_GrB_Semiring_free
#endif

static GrB_Info _cffi_d_GrB_Semiring_new(struct GB_Semiring_opaque * * x0, struct GB_Monoid_opaque * x1, struct GB_BinaryOp_opaque * x2)
{
  return GrB_Semiring_new(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Semiring_new(PyObject *self, PyObject *args)
{
  struct GB_Semiring_opaque * * x0;
  struct GB_Monoid_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Semiring_new", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2254), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Semiring_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2254), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Semiring_new(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Semiring_new _cffi_d_GrB_Semiring_new
#endif

static GrB_Info _cffi_d_GrB_Semiring_wait(struct GB_Semiring_opaque * x0, GrB_WaitMode x1)
{
  return GrB_Semiring_wait(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Semiring_wait(PyObject *self, PyObject *args)
{
  struct GB_Semiring_opaque * x0;
  GrB_WaitMode x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Semiring_wait", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(159), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Semiring_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(159), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(458), arg1) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Semiring_wait(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Semiring_wait _cffi_d_GrB_Semiring_wait
#endif

static GrB_Info _cffi_d_GrB_Type_error(char const * * x0, struct GB_Type_opaque * x1)
{
  return GrB_Type_error(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Type_error(PyObject *self, PyObject *args)
{
  char const * * x0;
  struct GB_Type_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Type_error", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(130), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char const * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(130), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Type_error(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Type_error _cffi_d_GrB_Type_error
#endif

static GrB_Info _cffi_d_GrB_Type_free(struct GB_Type_opaque * * x0)
{
  return GrB_Type_free(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Type_free(PyObject *self, PyObject *arg0)
{
  struct GB_Type_opaque * * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Type_free(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Type_free _cffi_d_GrB_Type_free
#endif

static GrB_Info _cffi_d_GrB_Type_new(struct GB_Type_opaque * * x0, size_t x1)
{
  return GrB_Type_new(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Type_new(PyObject *self, PyObject *args)
{
  struct GB_Type_opaque * * x0;
  size_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Type_new", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, size_t);
  if (x1 == (size_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Type_new(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Type_new _cffi_d_GrB_Type_new
#endif

static GrB_Info _cffi_d_GrB_Type_wait(struct GB_Type_opaque * x0, GrB_WaitMode x1)
{
  return GrB_Type_wait(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Type_wait(PyObject *self, PyObject *args)
{
  struct GB_Type_opaque * x0;
  GrB_WaitMode x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Type_wait", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(458), arg1) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Type_wait(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Type_wait _cffi_d_GrB_Type_wait
#endif

static GrB_Info _cffi_d_GrB_UnaryOp_error(char const * * x0, struct GB_UnaryOp_opaque * x1)
{
  return GrB_UnaryOp_error(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_UnaryOp_error(PyObject *self, PyObject *args)
{
  char const * * x0;
  struct GB_UnaryOp_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_UnaryOp_error", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(130), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char const * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(130), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(118), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_UnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(118), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_UnaryOp_error(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_UnaryOp_error _cffi_d_GrB_UnaryOp_error
#endif

static GrB_Info _cffi_d_GrB_UnaryOp_free(struct GB_UnaryOp_opaque * * x0)
{
  return GrB_UnaryOp_free(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_UnaryOp_free(PyObject *self, PyObject *arg0)
{
  struct GB_UnaryOp_opaque * * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2328), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_UnaryOp_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2328), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_UnaryOp_free(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_UnaryOp_free _cffi_d_GrB_UnaryOp_free
#endif

static GrB_Info _cffi_d_GrB_UnaryOp_new(struct GB_UnaryOp_opaque * * x0, void(* x1)(void *, void const *), struct GB_Type_opaque * x2, struct GB_Type_opaque * x3)
{
  return GrB_UnaryOp_new(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_UnaryOp_new(PyObject *self, PyObject *args)
{
  struct GB_UnaryOp_opaque * * x0;
  void(* x1)(void *, void const *);
  struct GB_Type_opaque * x2;
  struct GB_Type_opaque * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_UnaryOp_new", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2328), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_UnaryOp_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2328), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = (void(*)(void *, void const *))_cffi_to_c_pointer(arg1, _cffi_type(2332));
  if (x1 == (void(*)(void *, void const *))NULL && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_UnaryOp_new(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_UnaryOp_new _cffi_d_GrB_UnaryOp_new
#endif

static GrB_Info _cffi_d_GrB_UnaryOp_wait(struct GB_UnaryOp_opaque * x0, GrB_WaitMode x1)
{
  return GrB_UnaryOp_wait(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_UnaryOp_wait(PyObject *self, PyObject *args)
{
  struct GB_UnaryOp_opaque * x0;
  GrB_WaitMode x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_UnaryOp_wait", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(118), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_UnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(118), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(458), arg1) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_UnaryOp_wait(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_UnaryOp_wait _cffi_d_GrB_UnaryOp_wait
#endif

static GrB_Info _cffi_d_GrB_Vector_apply(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_UnaryOp_opaque * x3, struct GB_Vector_opaque * x4, struct GB_Descriptor_opaque * x5)
{
  return GrB_Vector_apply(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_UnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  struct GB_Descriptor_opaque * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(118), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_UnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(118), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply _cffi_d_GrB_Vector_apply
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp1st_BOOL(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, _Bool x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp1st_BOOL(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp1st_BOOL(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  _Bool x4;
  struct GB_Vector_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp1st_BOOL", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = (_Bool)_cffi_to_c__Bool(arg4);
  if (x4 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp1st_BOOL(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp1st_BOOL _cffi_d_GrB_Vector_apply_BinaryOp1st_BOOL
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp1st_FP32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, float x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp1st_FP32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp1st_FP32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  float x4;
  struct GB_Vector_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp1st_FP32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = (float)_cffi_to_c_float(arg4);
  if (x4 == (float)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp1st_FP32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp1st_FP32 _cffi_d_GrB_Vector_apply_BinaryOp1st_FP32
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp1st_FP64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, double x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp1st_FP64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp1st_FP64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  double x4;
  struct GB_Vector_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp1st_FP64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = (double)_cffi_to_c_double(arg4);
  if (x4 == (double)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp1st_FP64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp1st_FP64 _cffi_d_GrB_Vector_apply_BinaryOp1st_FP64
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp1st_INT16(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, int16_t x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp1st_INT16(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp1st_INT16(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  int16_t x4;
  struct GB_Vector_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp1st_INT16", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, int16_t);
  if (x4 == (int16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp1st_INT16(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp1st_INT16 _cffi_d_GrB_Vector_apply_BinaryOp1st_INT16
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp1st_INT32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, int32_t x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp1st_INT32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp1st_INT32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  int32_t x4;
  struct GB_Vector_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp1st_INT32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, int32_t);
  if (x4 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp1st_INT32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp1st_INT32 _cffi_d_GrB_Vector_apply_BinaryOp1st_INT32
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp1st_INT64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, int64_t x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp1st_INT64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp1st_INT64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  int64_t x4;
  struct GB_Vector_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp1st_INT64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, int64_t);
  if (x4 == (int64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp1st_INT64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp1st_INT64 _cffi_d_GrB_Vector_apply_BinaryOp1st_INT64
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp1st_INT8(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, int8_t x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp1st_INT8(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp1st_INT8(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  int8_t x4;
  struct GB_Vector_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp1st_INT8", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, int8_t);
  if (x4 == (int8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp1st_INT8(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp1st_INT8 _cffi_d_GrB_Vector_apply_BinaryOp1st_INT8
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp1st_Scalar(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Scalar_opaque * x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp1st_Scalar(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp1st_Scalar(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Scalar_opaque * x4;
  struct GB_Vector_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp1st_Scalar", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp1st_Scalar(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp1st_Scalar _cffi_d_GrB_Vector_apply_BinaryOp1st_Scalar
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp1st_UDT(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, void const * x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp1st_UDT(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp1st_UDT(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  void const * x4;
  struct GB_Vector_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp1st_UDT", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(126), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (void const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(126), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp1st_UDT(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp1st_UDT _cffi_d_GrB_Vector_apply_BinaryOp1st_UDT
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp1st_UINT16(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, uint16_t x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp1st_UINT16(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp1st_UINT16(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  uint16_t x4;
  struct GB_Vector_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp1st_UINT16", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint16_t);
  if (x4 == (uint16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp1st_UINT16(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp1st_UINT16 _cffi_d_GrB_Vector_apply_BinaryOp1st_UINT16
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp1st_UINT32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, uint32_t x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp1st_UINT32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp1st_UINT32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  uint32_t x4;
  struct GB_Vector_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp1st_UINT32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint32_t);
  if (x4 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp1st_UINT32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp1st_UINT32 _cffi_d_GrB_Vector_apply_BinaryOp1st_UINT32
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp1st_UINT64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, uint64_t x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp1st_UINT64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp1st_UINT64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  uint64_t x4;
  struct GB_Vector_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp1st_UINT64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp1st_UINT64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp1st_UINT64 _cffi_d_GrB_Vector_apply_BinaryOp1st_UINT64
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp1st_UINT8(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, uint8_t x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp1st_UINT8(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp1st_UINT8(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  uint8_t x4;
  struct GB_Vector_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp1st_UINT8", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint8_t);
  if (x4 == (uint8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp1st_UINT8(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp1st_UINT8 _cffi_d_GrB_Vector_apply_BinaryOp1st_UINT8
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp2nd_BOOL(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Vector_opaque * x4, _Bool x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp2nd_BOOL(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp2nd_BOOL(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  _Bool x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp2nd_BOOL", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (_Bool)_cffi_to_c__Bool(arg5);
  if (x5 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp2nd_BOOL(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp2nd_BOOL _cffi_d_GrB_Vector_apply_BinaryOp2nd_BOOL
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp2nd_FP32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Vector_opaque * x4, float x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp2nd_FP32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp2nd_FP32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  float x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp2nd_FP32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (float)_cffi_to_c_float(arg5);
  if (x5 == (float)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp2nd_FP32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp2nd_FP32 _cffi_d_GrB_Vector_apply_BinaryOp2nd_FP32
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp2nd_FP64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Vector_opaque * x4, double x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp2nd_FP64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp2nd_FP64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  double x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp2nd_FP64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (double)_cffi_to_c_double(arg5);
  if (x5 == (double)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp2nd_FP64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp2nd_FP64 _cffi_d_GrB_Vector_apply_BinaryOp2nd_FP64
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp2nd_INT16(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Vector_opaque * x4, int16_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp2nd_INT16(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp2nd_INT16(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  int16_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp2nd_INT16", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int16_t);
  if (x5 == (int16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp2nd_INT16(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp2nd_INT16 _cffi_d_GrB_Vector_apply_BinaryOp2nd_INT16
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp2nd_INT32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Vector_opaque * x4, int32_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp2nd_INT32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp2nd_INT32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  int32_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp2nd_INT32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int32_t);
  if (x5 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp2nd_INT32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp2nd_INT32 _cffi_d_GrB_Vector_apply_BinaryOp2nd_INT32
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp2nd_INT64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Vector_opaque * x4, int64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp2nd_INT64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp2nd_INT64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  int64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp2nd_INT64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int64_t);
  if (x5 == (int64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp2nd_INT64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp2nd_INT64 _cffi_d_GrB_Vector_apply_BinaryOp2nd_INT64
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp2nd_INT8(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Vector_opaque * x4, int8_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp2nd_INT8(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp2nd_INT8(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  int8_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp2nd_INT8", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int8_t);
  if (x5 == (int8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp2nd_INT8(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp2nd_INT8 _cffi_d_GrB_Vector_apply_BinaryOp2nd_INT8
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp2nd_Scalar(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Vector_opaque * x4, struct GB_Scalar_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp2nd_Scalar(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp2nd_Scalar(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  struct GB_Scalar_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp2nd_Scalar", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp2nd_Scalar(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp2nd_Scalar _cffi_d_GrB_Vector_apply_BinaryOp2nd_Scalar
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp2nd_UDT(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Vector_opaque * x4, void const * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp2nd_UDT(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp2nd_UDT(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  void const * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp2nd_UDT", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(126), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (void const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(126), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp2nd_UDT(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp2nd_UDT _cffi_d_GrB_Vector_apply_BinaryOp2nd_UDT
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp2nd_UINT16(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Vector_opaque * x4, uint16_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp2nd_UINT16(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp2nd_UINT16(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  uint16_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp2nd_UINT16", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint16_t);
  if (x5 == (uint16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp2nd_UINT16(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp2nd_UINT16 _cffi_d_GrB_Vector_apply_BinaryOp2nd_UINT16
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp2nd_UINT32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Vector_opaque * x4, uint32_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp2nd_UINT32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp2nd_UINT32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  uint32_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp2nd_UINT32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint32_t);
  if (x5 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp2nd_UINT32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp2nd_UINT32 _cffi_d_GrB_Vector_apply_BinaryOp2nd_UINT32
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp2nd_UINT64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Vector_opaque * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp2nd_UINT64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp2nd_UINT64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp2nd_UINT64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp2nd_UINT64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp2nd_UINT64 _cffi_d_GrB_Vector_apply_BinaryOp2nd_UINT64
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_BinaryOp2nd_UINT8(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Vector_opaque * x4, uint8_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_BinaryOp2nd_UINT8(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_BinaryOp2nd_UINT8(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  uint8_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_BinaryOp2nd_UINT8", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint8_t);
  if (x5 == (uint8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_BinaryOp2nd_UINT8(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_BinaryOp2nd_UINT8 _cffi_d_GrB_Vector_apply_BinaryOp2nd_UINT8
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_IndexOp_BOOL(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, _Bool x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_IndexOp_BOOL(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_IndexOp_BOOL(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  _Bool x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_IndexOp_BOOL", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (_Bool)_cffi_to_c__Bool(arg5);
  if (x5 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_IndexOp_BOOL(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_IndexOp_BOOL _cffi_d_GrB_Vector_apply_IndexOp_BOOL
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_IndexOp_FP32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, float x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_IndexOp_FP32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_IndexOp_FP32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  float x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_IndexOp_FP32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (float)_cffi_to_c_float(arg5);
  if (x5 == (float)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_IndexOp_FP32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_IndexOp_FP32 _cffi_d_GrB_Vector_apply_IndexOp_FP32
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_IndexOp_FP64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, double x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_IndexOp_FP64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_IndexOp_FP64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  double x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_IndexOp_FP64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (double)_cffi_to_c_double(arg5);
  if (x5 == (double)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_IndexOp_FP64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_IndexOp_FP64 _cffi_d_GrB_Vector_apply_IndexOp_FP64
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_IndexOp_INT16(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, int16_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_IndexOp_INT16(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_IndexOp_INT16(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  int16_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_IndexOp_INT16", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int16_t);
  if (x5 == (int16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_IndexOp_INT16(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_IndexOp_INT16 _cffi_d_GrB_Vector_apply_IndexOp_INT16
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_IndexOp_INT32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, int32_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_IndexOp_INT32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_IndexOp_INT32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  int32_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_IndexOp_INT32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int32_t);
  if (x5 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_IndexOp_INT32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_IndexOp_INT32 _cffi_d_GrB_Vector_apply_IndexOp_INT32
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_IndexOp_INT64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, int64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_IndexOp_INT64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_IndexOp_INT64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  int64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_IndexOp_INT64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int64_t);
  if (x5 == (int64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_IndexOp_INT64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_IndexOp_INT64 _cffi_d_GrB_Vector_apply_IndexOp_INT64
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_IndexOp_INT8(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, int8_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_IndexOp_INT8(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_IndexOp_INT8(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  int8_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_IndexOp_INT8", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int8_t);
  if (x5 == (int8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_IndexOp_INT8(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_IndexOp_INT8 _cffi_d_GrB_Vector_apply_IndexOp_INT8
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_IndexOp_Scalar(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, struct GB_Scalar_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_IndexOp_Scalar(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_IndexOp_Scalar(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  struct GB_Scalar_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_IndexOp_Scalar", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_IndexOp_Scalar(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_IndexOp_Scalar _cffi_d_GrB_Vector_apply_IndexOp_Scalar
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_IndexOp_UDT(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, void const * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_IndexOp_UDT(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_IndexOp_UDT(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  void const * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_IndexOp_UDT", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(126), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (void const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(126), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_IndexOp_UDT(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_IndexOp_UDT _cffi_d_GrB_Vector_apply_IndexOp_UDT
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_IndexOp_UINT16(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, uint16_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_IndexOp_UINT16(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_IndexOp_UINT16(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  uint16_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_IndexOp_UINT16", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint16_t);
  if (x5 == (uint16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_IndexOp_UINT16(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_IndexOp_UINT16 _cffi_d_GrB_Vector_apply_IndexOp_UINT16
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_IndexOp_UINT32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, uint32_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_IndexOp_UINT32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_IndexOp_UINT32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  uint32_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_IndexOp_UINT32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint32_t);
  if (x5 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_IndexOp_UINT32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_IndexOp_UINT32 _cffi_d_GrB_Vector_apply_IndexOp_UINT32
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_IndexOp_UINT64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_IndexOp_UINT64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_IndexOp_UINT64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_IndexOp_UINT64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_IndexOp_UINT64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_IndexOp_UINT64 _cffi_d_GrB_Vector_apply_IndexOp_UINT64
#endif

static GrB_Info _cffi_d_GrB_Vector_apply_IndexOp_UINT8(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, uint8_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_apply_IndexOp_UINT8(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_apply_IndexOp_UINT8(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  uint8_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_apply_IndexOp_UINT8", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint8_t);
  if (x5 == (uint8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_apply_IndexOp_UINT8(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_apply_IndexOp_UINT8 _cffi_d_GrB_Vector_apply_IndexOp_UINT8
#endif

static GrB_Info _cffi_d_GrB_Vector_assign(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Vector_opaque * x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_assign(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_assign(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Vector_opaque * x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_assign", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_assign(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_assign _cffi_d_GrB_Vector_assign
#endif

static GrB_Info _cffi_d_GrB_Vector_assign_BOOL(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, _Bool x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_assign_BOOL(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_assign_BOOL(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  _Bool x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_assign_BOOL", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = (_Bool)_cffi_to_c__Bool(arg3);
  if (x3 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_assign_BOOL(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_assign_BOOL _cffi_d_GrB_Vector_assign_BOOL
#endif

static GrB_Info _cffi_d_GrB_Vector_assign_FP32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, float x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_assign_FP32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_assign_FP32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  float x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_assign_FP32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = (float)_cffi_to_c_float(arg3);
  if (x3 == (float)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_assign_FP32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_assign_FP32 _cffi_d_GrB_Vector_assign_FP32
#endif

static GrB_Info _cffi_d_GrB_Vector_assign_FP64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, double x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_assign_FP64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_assign_FP64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  double x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_assign_FP64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = (double)_cffi_to_c_double(arg3);
  if (x3 == (double)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_assign_FP64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_assign_FP64 _cffi_d_GrB_Vector_assign_FP64
#endif

static GrB_Info _cffi_d_GrB_Vector_assign_INT16(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, int16_t x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_assign_INT16(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_assign_INT16(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  int16_t x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_assign_INT16", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, int16_t);
  if (x3 == (int16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_assign_INT16(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_assign_INT16 _cffi_d_GrB_Vector_assign_INT16
#endif

static GrB_Info _cffi_d_GrB_Vector_assign_INT32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, int32_t x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_assign_INT32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_assign_INT32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  int32_t x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_assign_INT32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, int32_t);
  if (x3 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_assign_INT32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_assign_INT32 _cffi_d_GrB_Vector_assign_INT32
#endif

static GrB_Info _cffi_d_GrB_Vector_assign_INT64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, int64_t x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_assign_INT64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_assign_INT64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  int64_t x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_assign_INT64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, int64_t);
  if (x3 == (int64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_assign_INT64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_assign_INT64 _cffi_d_GrB_Vector_assign_INT64
#endif

static GrB_Info _cffi_d_GrB_Vector_assign_INT8(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, int8_t x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_assign_INT8(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_assign_INT8(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  int8_t x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_assign_INT8", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, int8_t);
  if (x3 == (int8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_assign_INT8(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_assign_INT8 _cffi_d_GrB_Vector_assign_INT8
#endif

static GrB_Info _cffi_d_GrB_Vector_assign_Scalar(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Scalar_opaque * x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_assign_Scalar(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_assign_Scalar(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Scalar_opaque * x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_assign_Scalar", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_assign_Scalar(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_assign_Scalar _cffi_d_GrB_Vector_assign_Scalar
#endif

static GrB_Info _cffi_d_GrB_Vector_assign_UDT(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, void * x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_assign_UDT(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_assign_UDT(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  void * x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_assign_UDT", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_assign_UDT(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_assign_UDT _cffi_d_GrB_Vector_assign_UDT
#endif

static GrB_Info _cffi_d_GrB_Vector_assign_UINT16(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, uint16_t x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_assign_UINT16(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_assign_UINT16(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  uint16_t x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_assign_UINT16", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint16_t);
  if (x3 == (uint16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_assign_UINT16(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_assign_UINT16 _cffi_d_GrB_Vector_assign_UINT16
#endif

static GrB_Info _cffi_d_GrB_Vector_assign_UINT32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, uint32_t x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_assign_UINT32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_assign_UINT32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  uint32_t x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_assign_UINT32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint32_t);
  if (x3 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_assign_UINT32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_assign_UINT32 _cffi_d_GrB_Vector_assign_UINT32
#endif

static GrB_Info _cffi_d_GrB_Vector_assign_UINT64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, uint64_t x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_assign_UINT64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_assign_UINT64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  uint64_t x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_assign_UINT64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_assign_UINT64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_assign_UINT64 _cffi_d_GrB_Vector_assign_UINT64
#endif

static GrB_Info _cffi_d_GrB_Vector_assign_UINT8(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, uint8_t x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_assign_UINT8(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_assign_UINT8(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  uint8_t x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_assign_UINT8", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint8_t);
  if (x3 == (uint8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_assign_UINT8(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_assign_UINT8 _cffi_d_GrB_Vector_assign_UINT8
#endif

static GrB_Info _cffi_d_GrB_Vector_build_BOOL(struct GB_Vector_opaque * x0, uint64_t const * x1, _Bool const * x2, uint64_t x3, struct GB_BinaryOp_opaque * x4)
{
  return GrB_Vector_build_BOOL(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_build_BOOL(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint64_t const * x1;
  _Bool const * x2;
  uint64_t x3;
  struct GB_BinaryOp_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_build_BOOL", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(698), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (_Bool const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(698), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_build_BOOL(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_build_BOOL _cffi_d_GrB_Vector_build_BOOL
#endif

static GrB_Info _cffi_d_GrB_Vector_build_FP32(struct GB_Vector_opaque * x0, uint64_t const * x1, float const * x2, uint64_t x3, struct GB_BinaryOp_opaque * x4)
{
  return GrB_Vector_build_FP32(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_build_FP32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint64_t const * x1;
  float const * x2;
  uint64_t x3;
  struct GB_BinaryOp_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_build_FP32", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(750), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (float const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(750), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_build_FP32(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_build_FP32 _cffi_d_GrB_Vector_build_FP32
#endif

static GrB_Info _cffi_d_GrB_Vector_build_FP64(struct GB_Vector_opaque * x0, uint64_t const * x1, double const * x2, uint64_t x3, struct GB_BinaryOp_opaque * x4)
{
  return GrB_Vector_build_FP64(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_build_FP64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint64_t const * x1;
  double const * x2;
  uint64_t x3;
  struct GB_BinaryOp_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_build_FP64", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(724), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (double const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(724), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_build_FP64(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_build_FP64 _cffi_d_GrB_Vector_build_FP64
#endif

static GrB_Info _cffi_d_GrB_Vector_build_INT16(struct GB_Vector_opaque * x0, uint64_t const * x1, int16_t const * x2, uint64_t x3, struct GB_BinaryOp_opaque * x4)
{
  return GrB_Vector_build_INT16(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_build_INT16(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint64_t const * x1;
  int16_t const * x2;
  uint64_t x3;
  struct GB_BinaryOp_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_build_INT16", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(763), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (int16_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(763), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_build_INT16(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_build_INT16 _cffi_d_GrB_Vector_build_INT16
#endif

static GrB_Info _cffi_d_GrB_Vector_build_INT32(struct GB_Vector_opaque * x0, uint64_t const * x1, int32_t const * x2, uint64_t x3, struct GB_BinaryOp_opaque * x4)
{
  return GrB_Vector_build_INT32(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_build_INT32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint64_t const * x1;
  int32_t const * x2;
  uint64_t x3;
  struct GB_BinaryOp_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_build_INT32", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(776), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (int32_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(776), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_build_INT32(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_build_INT32 _cffi_d_GrB_Vector_build_INT32
#endif

static GrB_Info _cffi_d_GrB_Vector_build_INT64(struct GB_Vector_opaque * x0, uint64_t const * x1, int64_t const * x2, uint64_t x3, struct GB_BinaryOp_opaque * x4)
{
  return GrB_Vector_build_INT64(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_build_INT64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint64_t const * x1;
  int64_t const * x2;
  uint64_t x3;
  struct GB_BinaryOp_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_build_INT64", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(789), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (int64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(789), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_build_INT64(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_build_INT64 _cffi_d_GrB_Vector_build_INT64
#endif

static GrB_Info _cffi_d_GrB_Vector_build_INT8(struct GB_Vector_opaque * x0, uint64_t const * x1, int8_t const * x2, uint64_t x3, struct GB_BinaryOp_opaque * x4)
{
  return GrB_Vector_build_INT8(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_build_INT8(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint64_t const * x1;
  int8_t const * x2;
  uint64_t x3;
  struct GB_BinaryOp_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_build_INT8", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(802), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (int8_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(802), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_build_INT8(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_build_INT8 _cffi_d_GrB_Vector_build_INT8
#endif

static GrB_Info _cffi_d_GrB_Vector_build_UDT(struct GB_Vector_opaque * x0, uint64_t const * x1, void const * x2, uint64_t x3, struct GB_BinaryOp_opaque * x4)
{
  return GrB_Vector_build_UDT(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_build_UDT(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint64_t const * x1;
  void const * x2;
  uint64_t x3;
  struct GB_BinaryOp_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_build_UDT", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(126), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (void const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(126), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_build_UDT(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_build_UDT _cffi_d_GrB_Vector_build_UDT
#endif

static GrB_Info _cffi_d_GrB_Vector_build_UINT16(struct GB_Vector_opaque * x0, uint64_t const * x1, uint16_t const * x2, uint64_t x3, struct GB_BinaryOp_opaque * x4)
{
  return GrB_Vector_build_UINT16(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_build_UINT16(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint64_t const * x1;
  uint16_t const * x2;
  uint64_t x3;
  struct GB_BinaryOp_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_build_UINT16", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(815), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint16_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(815), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_build_UINT16(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_build_UINT16 _cffi_d_GrB_Vector_build_UINT16
#endif

static GrB_Info _cffi_d_GrB_Vector_build_UINT32(struct GB_Vector_opaque * x0, uint64_t const * x1, uint32_t const * x2, uint64_t x3, struct GB_BinaryOp_opaque * x4)
{
  return GrB_Vector_build_UINT32(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_build_UINT32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint64_t const * x1;
  uint32_t const * x2;
  uint64_t x3;
  struct GB_BinaryOp_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_build_UINT32", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(828), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint32_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(828), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_build_UINT32(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_build_UINT32 _cffi_d_GrB_Vector_build_UINT32
#endif

static GrB_Info _cffi_d_GrB_Vector_build_UINT64(struct GB_Vector_opaque * x0, uint64_t const * x1, uint64_t const * x2, uint64_t x3, struct GB_BinaryOp_opaque * x4)
{
  return GrB_Vector_build_UINT64(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_build_UINT64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint64_t const * x1;
  uint64_t const * x2;
  uint64_t x3;
  struct GB_BinaryOp_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_build_UINT64", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_build_UINT64(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_build_UINT64 _cffi_d_GrB_Vector_build_UINT64
#endif

static GrB_Info _cffi_d_GrB_Vector_build_UINT8(struct GB_Vector_opaque * x0, uint64_t const * x1, uint8_t const * x2, uint64_t x3, struct GB_BinaryOp_opaque * x4)
{
  return GrB_Vector_build_UINT8(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_build_UINT8(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint64_t const * x1;
  uint8_t const * x2;
  uint64_t x3;
  struct GB_BinaryOp_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_build_UINT8", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(854), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint8_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(854), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_build_UINT8(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_build_UINT8 _cffi_d_GrB_Vector_build_UINT8
#endif

static GrB_Info _cffi_d_GrB_Vector_clear(struct GB_Vector_opaque * x0)
{
  return GrB_Vector_clear(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_clear(PyObject *self, PyObject *arg0)
{
  struct GB_Vector_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_clear(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_clear _cffi_d_GrB_Vector_clear
#endif

static GrB_Info _cffi_d_GrB_Vector_dup(struct GB_Vector_opaque * * x0, struct GB_Vector_opaque * x1)
{
  return GrB_Vector_dup(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_dup(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * * x0;
  struct GB_Vector_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_dup", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2362), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2362), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_dup(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_dup _cffi_d_GrB_Vector_dup
#endif

static GrB_Info _cffi_d_GrB_Vector_eWiseAdd_BinaryOp(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Vector_opaque * x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_eWiseAdd_BinaryOp(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_eWiseAdd_BinaryOp(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  struct GB_Vector_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_eWiseAdd_BinaryOp", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_eWiseAdd_BinaryOp(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_eWiseAdd_BinaryOp _cffi_d_GrB_Vector_eWiseAdd_BinaryOp
#endif

static GrB_Info _cffi_d_GrB_Vector_eWiseAdd_Monoid(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Monoid_opaque * x3, struct GB_Vector_opaque * x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_eWiseAdd_Monoid(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_eWiseAdd_Monoid(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Monoid_opaque * x3;
  struct GB_Vector_opaque * x4;
  struct GB_Vector_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_eWiseAdd_Monoid", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_eWiseAdd_Monoid(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_eWiseAdd_Monoid _cffi_d_GrB_Vector_eWiseAdd_Monoid
#endif

static GrB_Info _cffi_d_GrB_Vector_eWiseAdd_Semiring(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Semiring_opaque * x3, struct GB_Vector_opaque * x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_eWiseAdd_Semiring(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_eWiseAdd_Semiring(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Semiring_opaque * x3;
  struct GB_Vector_opaque * x4;
  struct GB_Vector_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_eWiseAdd_Semiring", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(159), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Semiring_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(159), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_eWiseAdd_Semiring(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_eWiseAdd_Semiring _cffi_d_GrB_Vector_eWiseAdd_Semiring
#endif

static GrB_Info _cffi_d_GrB_Vector_eWiseMult_BinaryOp(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Vector_opaque * x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_eWiseMult_BinaryOp(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_eWiseMult_BinaryOp(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  struct GB_Vector_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_eWiseMult_BinaryOp", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_eWiseMult_BinaryOp(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_eWiseMult_BinaryOp _cffi_d_GrB_Vector_eWiseMult_BinaryOp
#endif

static GrB_Info _cffi_d_GrB_Vector_eWiseMult_Monoid(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Monoid_opaque * x3, struct GB_Vector_opaque * x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_eWiseMult_Monoid(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_eWiseMult_Monoid(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Monoid_opaque * x3;
  struct GB_Vector_opaque * x4;
  struct GB_Vector_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_eWiseMult_Monoid", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_eWiseMult_Monoid(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_eWiseMult_Monoid _cffi_d_GrB_Vector_eWiseMult_Monoid
#endif

static GrB_Info _cffi_d_GrB_Vector_eWiseMult_Semiring(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Semiring_opaque * x3, struct GB_Vector_opaque * x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_eWiseMult_Semiring(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_eWiseMult_Semiring(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Semiring_opaque * x3;
  struct GB_Vector_opaque * x4;
  struct GB_Vector_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_eWiseMult_Semiring", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(159), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Semiring_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(159), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_eWiseMult_Semiring(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_eWiseMult_Semiring _cffi_d_GrB_Vector_eWiseMult_Semiring
#endif

static GrB_Info _cffi_d_GrB_Vector_error(char const * * x0, struct GB_Vector_opaque * x1)
{
  return GrB_Vector_error(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_error(PyObject *self, PyObject *args)
{
  char const * * x0;
  struct GB_Vector_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_error", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(130), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char const * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(130), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_error(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_error _cffi_d_GrB_Vector_error
#endif

static GrB_Info _cffi_d_GrB_Vector_extract(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Vector_opaque * x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_extract(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extract(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Vector_opaque * x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extract", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extract(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extract _cffi_d_GrB_Vector_extract
#endif

static GrB_Info _cffi_d_GrB_Vector_extractElement_BOOL(_Bool * x0, struct GB_Vector_opaque * x1, uint64_t x2)
{
  return GrB_Vector_extractElement_BOOL(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractElement_BOOL(PyObject *self, PyObject *args)
{
  _Bool * x0;
  struct GB_Vector_opaque * x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractElement_BOOL", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractElement_BOOL(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractElement_BOOL _cffi_d_GrB_Vector_extractElement_BOOL
#endif

static GrB_Info _cffi_d_GrB_Vector_extractElement_FP32(float * x0, struct GB_Vector_opaque * x1, uint64_t x2)
{
  return GrB_Vector_extractElement_FP32(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractElement_FP32(PyObject *self, PyObject *args)
{
  float * x0;
  struct GB_Vector_opaque * x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractElement_FP32", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(232), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (float *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(232), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractElement_FP32(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractElement_FP32 _cffi_d_GrB_Vector_extractElement_FP32
#endif

static GrB_Info _cffi_d_GrB_Vector_extractElement_FP64(double * x0, struct GB_Vector_opaque * x1, uint64_t x2)
{
  return GrB_Vector_extractElement_FP64(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractElement_FP64(PyObject *self, PyObject *args)
{
  double * x0;
  struct GB_Vector_opaque * x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractElement_FP64", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(28), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(28), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractElement_FP64(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractElement_FP64 _cffi_d_GrB_Vector_extractElement_FP64
#endif

static GrB_Info _cffi_d_GrB_Vector_extractElement_INT16(int16_t * x0, struct GB_Vector_opaque * x1, uint64_t x2)
{
  return GrB_Vector_extractElement_INT16(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractElement_INT16(PyObject *self, PyObject *args)
{
  int16_t * x0;
  struct GB_Vector_opaque * x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractElement_INT16", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(290), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int16_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(290), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractElement_INT16(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractElement_INT16 _cffi_d_GrB_Vector_extractElement_INT16
#endif

static GrB_Info _cffi_d_GrB_Vector_extractElement_INT32(int32_t * x0, struct GB_Vector_opaque * x1, uint64_t x2)
{
  return GrB_Vector_extractElement_INT32(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractElement_INT32(PyObject *self, PyObject *args)
{
  int32_t * x0;
  struct GB_Vector_opaque * x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractElement_INT32", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(36), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int32_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(36), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractElement_INT32(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractElement_INT32 _cffi_d_GrB_Vector_extractElement_INT32
#endif

static GrB_Info _cffi_d_GrB_Vector_extractElement_INT64(int64_t * x0, struct GB_Vector_opaque * x1, uint64_t x2)
{
  return GrB_Vector_extractElement_INT64(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractElement_INT64(PyObject *self, PyObject *args)
{
  int64_t * x0;
  struct GB_Vector_opaque * x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractElement_INT64", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(44), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(44), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractElement_INT64(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractElement_INT64 _cffi_d_GrB_Vector_extractElement_INT64
#endif

static GrB_Info _cffi_d_GrB_Vector_extractElement_INT8(int8_t * x0, struct GB_Vector_opaque * x1, uint64_t x2)
{
  return GrB_Vector_extractElement_INT8(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractElement_INT8(PyObject *self, PyObject *args)
{
  int8_t * x0;
  struct GB_Vector_opaque * x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractElement_INT8", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(377), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int8_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(377), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractElement_INT8(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractElement_INT8 _cffi_d_GrB_Vector_extractElement_INT8
#endif

static GrB_Info _cffi_d_GrB_Vector_extractElement_Scalar(struct GB_Scalar_opaque * x0, struct GB_Vector_opaque * x1, uint64_t x2)
{
  return GrB_Vector_extractElement_Scalar(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractElement_Scalar(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  struct GB_Vector_opaque * x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractElement_Scalar", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractElement_Scalar(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractElement_Scalar _cffi_d_GrB_Vector_extractElement_Scalar
#endif

static GrB_Info _cffi_d_GrB_Vector_extractElement_UDT(void * x0, struct GB_Vector_opaque * x1, uint64_t x2)
{
  return GrB_Vector_extractElement_UDT(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractElement_UDT(PyObject *self, PyObject *args)
{
  void * x0;
  struct GB_Vector_opaque * x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractElement_UDT", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractElement_UDT(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractElement_UDT _cffi_d_GrB_Vector_extractElement_UDT
#endif

static GrB_Info _cffi_d_GrB_Vector_extractElement_UINT16(uint16_t * x0, struct GB_Vector_opaque * x1, uint64_t x2)
{
  return GrB_Vector_extractElement_UINT16(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractElement_UINT16(PyObject *self, PyObject *args)
{
  uint16_t * x0;
  struct GB_Vector_opaque * x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractElement_UINT16", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3386), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint16_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3386), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractElement_UINT16(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractElement_UINT16 _cffi_d_GrB_Vector_extractElement_UINT16
#endif

static GrB_Info _cffi_d_GrB_Vector_extractElement_UINT32(uint32_t * x0, struct GB_Vector_opaque * x1, uint64_t x2)
{
  return GrB_Vector_extractElement_UINT32(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractElement_UINT32(PyObject *self, PyObject *args)
{
  uint32_t * x0;
  struct GB_Vector_opaque * x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractElement_UINT32", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3415), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint32_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3415), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractElement_UINT32(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractElement_UINT32 _cffi_d_GrB_Vector_extractElement_UINT32
#endif

static GrB_Info _cffi_d_GrB_Vector_extractElement_UINT64(uint64_t * x0, struct GB_Vector_opaque * x1, uint64_t x2)
{
  return GrB_Vector_extractElement_UINT64(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractElement_UINT64(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  struct GB_Vector_opaque * x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractElement_UINT64", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractElement_UINT64(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractElement_UINT64 _cffi_d_GrB_Vector_extractElement_UINT64
#endif

static GrB_Info _cffi_d_GrB_Vector_extractElement_UINT8(uint8_t * x0, struct GB_Vector_opaque * x1, uint64_t x2)
{
  return GrB_Vector_extractElement_UINT8(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractElement_UINT8(PyObject *self, PyObject *args)
{
  uint8_t * x0;
  struct GB_Vector_opaque * x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractElement_UINT8", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3766), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint8_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3766), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractElement_UINT8(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractElement_UINT8 _cffi_d_GrB_Vector_extractElement_UINT8
#endif

static GrB_Info _cffi_d_GrB_Vector_extractTuples_BOOL(uint64_t * x0, _Bool * x1, uint64_t * x2, struct GB_Vector_opaque * x3)
{
  return GrB_Vector_extractTuples_BOOL(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractTuples_BOOL(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  _Bool * x1;
  uint64_t * x2;
  struct GB_Vector_opaque * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractTuples_BOOL", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractTuples_BOOL(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractTuples_BOOL _cffi_d_GrB_Vector_extractTuples_BOOL
#endif

static GrB_Info _cffi_d_GrB_Vector_extractTuples_FP32(uint64_t * x0, float * x1, uint64_t * x2, struct GB_Vector_opaque * x3)
{
  return GrB_Vector_extractTuples_FP32(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractTuples_FP32(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  float * x1;
  uint64_t * x2;
  struct GB_Vector_opaque * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractTuples_FP32", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(232), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (float *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(232), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractTuples_FP32(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractTuples_FP32 _cffi_d_GrB_Vector_extractTuples_FP32
#endif

static GrB_Info _cffi_d_GrB_Vector_extractTuples_FP64(uint64_t * x0, double * x1, uint64_t * x2, struct GB_Vector_opaque * x3)
{
  return GrB_Vector_extractTuples_FP64(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractTuples_FP64(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  double * x1;
  uint64_t * x2;
  struct GB_Vector_opaque * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractTuples_FP64", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(28), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(28), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractTuples_FP64(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractTuples_FP64 _cffi_d_GrB_Vector_extractTuples_FP64
#endif

static GrB_Info _cffi_d_GrB_Vector_extractTuples_INT16(uint64_t * x0, int16_t * x1, uint64_t * x2, struct GB_Vector_opaque * x3)
{
  return GrB_Vector_extractTuples_INT16(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractTuples_INT16(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  int16_t * x1;
  uint64_t * x2;
  struct GB_Vector_opaque * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractTuples_INT16", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(290), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (int16_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(290), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractTuples_INT16(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractTuples_INT16 _cffi_d_GrB_Vector_extractTuples_INT16
#endif

static GrB_Info _cffi_d_GrB_Vector_extractTuples_INT32(uint64_t * x0, int32_t * x1, uint64_t * x2, struct GB_Vector_opaque * x3)
{
  return GrB_Vector_extractTuples_INT32(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractTuples_INT32(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  int32_t * x1;
  uint64_t * x2;
  struct GB_Vector_opaque * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractTuples_INT32", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(36), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (int32_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(36), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractTuples_INT32(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractTuples_INT32 _cffi_d_GrB_Vector_extractTuples_INT32
#endif

static GrB_Info _cffi_d_GrB_Vector_extractTuples_INT64(uint64_t * x0, int64_t * x1, uint64_t * x2, struct GB_Vector_opaque * x3)
{
  return GrB_Vector_extractTuples_INT64(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractTuples_INT64(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  int64_t * x1;
  uint64_t * x2;
  struct GB_Vector_opaque * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractTuples_INT64", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(44), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (int64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(44), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractTuples_INT64(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractTuples_INT64 _cffi_d_GrB_Vector_extractTuples_INT64
#endif

static GrB_Info _cffi_d_GrB_Vector_extractTuples_INT8(uint64_t * x0, int8_t * x1, uint64_t * x2, struct GB_Vector_opaque * x3)
{
  return GrB_Vector_extractTuples_INT8(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractTuples_INT8(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  int8_t * x1;
  uint64_t * x2;
  struct GB_Vector_opaque * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractTuples_INT8", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(377), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (int8_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(377), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractTuples_INT8(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractTuples_INT8 _cffi_d_GrB_Vector_extractTuples_INT8
#endif

static GrB_Info _cffi_d_GrB_Vector_extractTuples_UDT(uint64_t * x0, void * x1, uint64_t * x2, struct GB_Vector_opaque * x3)
{
  return GrB_Vector_extractTuples_UDT(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractTuples_UDT(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  void * x1;
  uint64_t * x2;
  struct GB_Vector_opaque * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractTuples_UDT", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractTuples_UDT(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractTuples_UDT _cffi_d_GrB_Vector_extractTuples_UDT
#endif

static GrB_Info _cffi_d_GrB_Vector_extractTuples_UINT16(uint64_t * x0, uint16_t * x1, uint64_t * x2, struct GB_Vector_opaque * x3)
{
  return GrB_Vector_extractTuples_UINT16(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractTuples_UINT16(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint16_t * x1;
  uint64_t * x2;
  struct GB_Vector_opaque * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractTuples_UINT16", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3386), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint16_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3386), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractTuples_UINT16(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractTuples_UINT16 _cffi_d_GrB_Vector_extractTuples_UINT16
#endif

static GrB_Info _cffi_d_GrB_Vector_extractTuples_UINT32(uint64_t * x0, uint32_t * x1, uint64_t * x2, struct GB_Vector_opaque * x3)
{
  return GrB_Vector_extractTuples_UINT32(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractTuples_UINT32(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint32_t * x1;
  uint64_t * x2;
  struct GB_Vector_opaque * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractTuples_UINT32", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3415), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint32_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3415), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractTuples_UINT32(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractTuples_UINT32 _cffi_d_GrB_Vector_extractTuples_UINT32
#endif

static GrB_Info _cffi_d_GrB_Vector_extractTuples_UINT64(uint64_t * x0, uint64_t * x1, uint64_t * x2, struct GB_Vector_opaque * x3)
{
  return GrB_Vector_extractTuples_UINT64(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractTuples_UINT64(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  uint64_t * x2;
  struct GB_Vector_opaque * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractTuples_UINT64", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractTuples_UINT64(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractTuples_UINT64 _cffi_d_GrB_Vector_extractTuples_UINT64
#endif

static GrB_Info _cffi_d_GrB_Vector_extractTuples_UINT8(uint64_t * x0, uint8_t * x1, uint64_t * x2, struct GB_Vector_opaque * x3)
{
  return GrB_Vector_extractTuples_UINT8(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_extractTuples_UINT8(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint8_t * x1;
  uint64_t * x2;
  struct GB_Vector_opaque * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_extractTuples_UINT8", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3766), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint8_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3766), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_extractTuples_UINT8(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_extractTuples_UINT8 _cffi_d_GrB_Vector_extractTuples_UINT8
#endif

static GrB_Info _cffi_d_GrB_Vector_free(struct GB_Vector_opaque * * x0)
{
  return GrB_Vector_free(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_free(PyObject *self, PyObject *arg0)
{
  struct GB_Vector_opaque * * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2362), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2362), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_free(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_free _cffi_d_GrB_Vector_free
#endif

static GrB_Info _cffi_d_GrB_Vector_new(struct GB_Vector_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2)
{
  return GrB_Vector_new(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_new(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_new", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2362), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2362), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_new(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_new _cffi_d_GrB_Vector_new
#endif

static GrB_Info _cffi_d_GrB_Vector_nvals(uint64_t * x0, struct GB_Vector_opaque * x1)
{
  return GrB_Vector_nvals(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_nvals(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  struct GB_Vector_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_nvals", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_nvals(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_nvals _cffi_d_GrB_Vector_nvals
#endif

static GrB_Info _cffi_d_GrB_Vector_reduce_BOOL(_Bool * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Vector_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Vector_reduce_BOOL(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_reduce_BOOL(PyObject *self, PyObject *args)
{
  _Bool * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Vector_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_reduce_BOOL", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_reduce_BOOL(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_reduce_BOOL _cffi_d_GrB_Vector_reduce_BOOL
#endif

static GrB_Info _cffi_d_GrB_Vector_reduce_BinaryOp_Scalar(struct GB_Scalar_opaque * x0, struct GB_BinaryOp_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Vector_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Vector_reduce_BinaryOp_Scalar(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_reduce_BinaryOp_Scalar(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Vector_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_reduce_BinaryOp_Scalar", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_reduce_BinaryOp_Scalar(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_reduce_BinaryOp_Scalar _cffi_d_GrB_Vector_reduce_BinaryOp_Scalar
#endif

static GrB_Info _cffi_d_GrB_Vector_reduce_FP32(float * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Vector_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Vector_reduce_FP32(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_reduce_FP32(PyObject *self, PyObject *args)
{
  float * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Vector_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_reduce_FP32", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(232), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (float *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(232), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_reduce_FP32(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_reduce_FP32 _cffi_d_GrB_Vector_reduce_FP32
#endif

static GrB_Info _cffi_d_GrB_Vector_reduce_FP64(double * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Vector_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Vector_reduce_FP64(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_reduce_FP64(PyObject *self, PyObject *args)
{
  double * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Vector_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_reduce_FP64", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(28), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(28), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_reduce_FP64(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_reduce_FP64 _cffi_d_GrB_Vector_reduce_FP64
#endif

static GrB_Info _cffi_d_GrB_Vector_reduce_INT16(int16_t * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Vector_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Vector_reduce_INT16(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_reduce_INT16(PyObject *self, PyObject *args)
{
  int16_t * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Vector_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_reduce_INT16", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(290), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int16_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(290), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_reduce_INT16(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_reduce_INT16 _cffi_d_GrB_Vector_reduce_INT16
#endif

static GrB_Info _cffi_d_GrB_Vector_reduce_INT32(int32_t * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Vector_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Vector_reduce_INT32(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_reduce_INT32(PyObject *self, PyObject *args)
{
  int32_t * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Vector_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_reduce_INT32", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(36), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int32_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(36), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_reduce_INT32(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_reduce_INT32 _cffi_d_GrB_Vector_reduce_INT32
#endif

static GrB_Info _cffi_d_GrB_Vector_reduce_INT64(int64_t * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Vector_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Vector_reduce_INT64(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_reduce_INT64(PyObject *self, PyObject *args)
{
  int64_t * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Vector_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_reduce_INT64", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(44), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(44), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_reduce_INT64(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_reduce_INT64 _cffi_d_GrB_Vector_reduce_INT64
#endif

static GrB_Info _cffi_d_GrB_Vector_reduce_INT8(int8_t * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Vector_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Vector_reduce_INT8(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_reduce_INT8(PyObject *self, PyObject *args)
{
  int8_t * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Vector_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_reduce_INT8", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(377), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int8_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(377), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_reduce_INT8(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_reduce_INT8 _cffi_d_GrB_Vector_reduce_INT8
#endif

static GrB_Info _cffi_d_GrB_Vector_reduce_Monoid_Scalar(struct GB_Scalar_opaque * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Vector_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Vector_reduce_Monoid_Scalar(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_reduce_Monoid_Scalar(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Vector_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_reduce_Monoid_Scalar", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_reduce_Monoid_Scalar(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_reduce_Monoid_Scalar _cffi_d_GrB_Vector_reduce_Monoid_Scalar
#endif

static GrB_Info _cffi_d_GrB_Vector_reduce_UDT(void * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Vector_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Vector_reduce_UDT(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_reduce_UDT(PyObject *self, PyObject *args)
{
  void * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Vector_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_reduce_UDT", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_reduce_UDT(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_reduce_UDT _cffi_d_GrB_Vector_reduce_UDT
#endif

static GrB_Info _cffi_d_GrB_Vector_reduce_UINT16(uint16_t * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Vector_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Vector_reduce_UINT16(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_reduce_UINT16(PyObject *self, PyObject *args)
{
  uint16_t * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Vector_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_reduce_UINT16", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3386), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint16_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3386), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_reduce_UINT16(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_reduce_UINT16 _cffi_d_GrB_Vector_reduce_UINT16
#endif

static GrB_Info _cffi_d_GrB_Vector_reduce_UINT32(uint32_t * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Vector_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Vector_reduce_UINT32(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_reduce_UINT32(PyObject *self, PyObject *args)
{
  uint32_t * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Vector_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_reduce_UINT32", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3415), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint32_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3415), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_reduce_UINT32(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_reduce_UINT32 _cffi_d_GrB_Vector_reduce_UINT32
#endif

static GrB_Info _cffi_d_GrB_Vector_reduce_UINT64(uint64_t * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Vector_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Vector_reduce_UINT64(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_reduce_UINT64(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Vector_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_reduce_UINT64", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_reduce_UINT64(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_reduce_UINT64 _cffi_d_GrB_Vector_reduce_UINT64
#endif

static GrB_Info _cffi_d_GrB_Vector_reduce_UINT8(uint8_t * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Vector_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_Vector_reduce_UINT8(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_reduce_UINT8(PyObject *self, PyObject *args)
{
  uint8_t * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Vector_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_reduce_UINT8", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3766), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint8_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3766), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_reduce_UINT8(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_reduce_UINT8 _cffi_d_GrB_Vector_reduce_UINT8
#endif

static GrB_Info _cffi_d_GrB_Vector_removeElement(struct GB_Vector_opaque * x0, uint64_t x1)
{
  return GrB_Vector_removeElement(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_removeElement(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint64_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_removeElement", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint64_t);
  if (x1 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_removeElement(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_removeElement _cffi_d_GrB_Vector_removeElement
#endif

static GrB_Info _cffi_d_GrB_Vector_resize(struct GB_Vector_opaque * x0, uint64_t x1)
{
  return GrB_Vector_resize(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_resize(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint64_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_resize", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint64_t);
  if (x1 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_resize(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_resize _cffi_d_GrB_Vector_resize
#endif

static GrB_Info _cffi_d_GrB_Vector_select_BOOL(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, _Bool x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_select_BOOL(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_select_BOOL(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  _Bool x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_select_BOOL", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (_Bool)_cffi_to_c__Bool(arg5);
  if (x5 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_select_BOOL(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_select_BOOL _cffi_d_GrB_Vector_select_BOOL
#endif

static GrB_Info _cffi_d_GrB_Vector_select_FP32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, float x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_select_FP32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_select_FP32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  float x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_select_FP32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (float)_cffi_to_c_float(arg5);
  if (x5 == (float)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_select_FP32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_select_FP32 _cffi_d_GrB_Vector_select_FP32
#endif

static GrB_Info _cffi_d_GrB_Vector_select_FP64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, double x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_select_FP64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_select_FP64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  double x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_select_FP64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = (double)_cffi_to_c_double(arg5);
  if (x5 == (double)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_select_FP64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_select_FP64 _cffi_d_GrB_Vector_select_FP64
#endif

static GrB_Info _cffi_d_GrB_Vector_select_INT16(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, int16_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_select_INT16(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_select_INT16(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  int16_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_select_INT16", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int16_t);
  if (x5 == (int16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_select_INT16(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_select_INT16 _cffi_d_GrB_Vector_select_INT16
#endif

static GrB_Info _cffi_d_GrB_Vector_select_INT32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, int32_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_select_INT32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_select_INT32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  int32_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_select_INT32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int32_t);
  if (x5 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_select_INT32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_select_INT32 _cffi_d_GrB_Vector_select_INT32
#endif

static GrB_Info _cffi_d_GrB_Vector_select_INT64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, int64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_select_INT64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_select_INT64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  int64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_select_INT64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int64_t);
  if (x5 == (int64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_select_INT64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_select_INT64 _cffi_d_GrB_Vector_select_INT64
#endif

static GrB_Info _cffi_d_GrB_Vector_select_INT8(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, int8_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_select_INT8(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_select_INT8(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  int8_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_select_INT8", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, int8_t);
  if (x5 == (int8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_select_INT8(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_select_INT8 _cffi_d_GrB_Vector_select_INT8
#endif

static GrB_Info _cffi_d_GrB_Vector_select_Scalar(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, struct GB_Scalar_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_select_Scalar(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_select_Scalar(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  struct GB_Scalar_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_select_Scalar", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_select_Scalar(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_select_Scalar _cffi_d_GrB_Vector_select_Scalar
#endif

static GrB_Info _cffi_d_GrB_Vector_select_UDT(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, void const * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_select_UDT(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_select_UDT(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  void const * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_select_UDT", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(126), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (void const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(126), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_select_UDT(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_select_UDT _cffi_d_GrB_Vector_select_UDT
#endif

static GrB_Info _cffi_d_GrB_Vector_select_UINT16(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, uint16_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_select_UINT16(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_select_UINT16(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  uint16_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_select_UINT16", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint16_t);
  if (x5 == (uint16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_select_UINT16(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_select_UINT16 _cffi_d_GrB_Vector_select_UINT16
#endif

static GrB_Info _cffi_d_GrB_Vector_select_UINT32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, uint32_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_select_UINT32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_select_UINT32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  uint32_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_select_UINT32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint32_t);
  if (x5 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_select_UINT32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_select_UINT32 _cffi_d_GrB_Vector_select_UINT32
#endif

static GrB_Info _cffi_d_GrB_Vector_select_UINT64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_select_UINT64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_select_UINT64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_select_UINT64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_select_UINT64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_select_UINT64 _cffi_d_GrB_Vector_select_UINT64
#endif

static GrB_Info _cffi_d_GrB_Vector_select_UINT8(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, uint8_t x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_Vector_select_UINT8(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_select_UINT8(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  uint8_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_select_UINT8", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint8_t);
  if (x5 == (uint8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_select_UINT8(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_select_UINT8 _cffi_d_GrB_Vector_select_UINT8
#endif

static GrB_Info _cffi_d_GrB_Vector_setElement_BOOL(struct GB_Vector_opaque * x0, _Bool x1, uint64_t x2)
{
  return GrB_Vector_setElement_BOOL(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_setElement_BOOL(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  _Bool x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_setElement_BOOL", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = (_Bool)_cffi_to_c__Bool(arg1);
  if (x1 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_setElement_BOOL(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_setElement_BOOL _cffi_d_GrB_Vector_setElement_BOOL
#endif

static GrB_Info _cffi_d_GrB_Vector_setElement_FP32(struct GB_Vector_opaque * x0, float x1, uint64_t x2)
{
  return GrB_Vector_setElement_FP32(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_setElement_FP32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  float x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_setElement_FP32", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = (float)_cffi_to_c_float(arg1);
  if (x1 == (float)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_setElement_FP32(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_setElement_FP32 _cffi_d_GrB_Vector_setElement_FP32
#endif

static GrB_Info _cffi_d_GrB_Vector_setElement_FP64(struct GB_Vector_opaque * x0, double x1, uint64_t x2)
{
  return GrB_Vector_setElement_FP64(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_setElement_FP64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  double x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_setElement_FP64", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = (double)_cffi_to_c_double(arg1);
  if (x1 == (double)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_setElement_FP64(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_setElement_FP64 _cffi_d_GrB_Vector_setElement_FP64
#endif

static GrB_Info _cffi_d_GrB_Vector_setElement_INT16(struct GB_Vector_opaque * x0, int16_t x1, uint64_t x2)
{
  return GrB_Vector_setElement_INT16(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_setElement_INT16(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  int16_t x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_setElement_INT16", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, int16_t);
  if (x1 == (int16_t)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_setElement_INT16(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_setElement_INT16 _cffi_d_GrB_Vector_setElement_INT16
#endif

static GrB_Info _cffi_d_GrB_Vector_setElement_INT32(struct GB_Vector_opaque * x0, int32_t x1, uint64_t x2)
{
  return GrB_Vector_setElement_INT32(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_setElement_INT32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  int32_t x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_setElement_INT32", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, int32_t);
  if (x1 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_setElement_INT32(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_setElement_INT32 _cffi_d_GrB_Vector_setElement_INT32
#endif

static GrB_Info _cffi_d_GrB_Vector_setElement_INT64(struct GB_Vector_opaque * x0, int64_t x1, uint64_t x2)
{
  return GrB_Vector_setElement_INT64(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_setElement_INT64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  int64_t x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_setElement_INT64", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, int64_t);
  if (x1 == (int64_t)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_setElement_INT64(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_setElement_INT64 _cffi_d_GrB_Vector_setElement_INT64
#endif

static GrB_Info _cffi_d_GrB_Vector_setElement_INT8(struct GB_Vector_opaque * x0, int8_t x1, uint64_t x2)
{
  return GrB_Vector_setElement_INT8(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_setElement_INT8(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  int8_t x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_setElement_INT8", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, int8_t);
  if (x1 == (int8_t)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_setElement_INT8(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_setElement_INT8 _cffi_d_GrB_Vector_setElement_INT8
#endif

static GrB_Info _cffi_d_GrB_Vector_setElement_Scalar(struct GB_Vector_opaque * x0, struct GB_Scalar_opaque * x1, uint64_t x2)
{
  return GrB_Vector_setElement_Scalar(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_setElement_Scalar(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Scalar_opaque * x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_setElement_Scalar", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_setElement_Scalar(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_setElement_Scalar _cffi_d_GrB_Vector_setElement_Scalar
#endif

static GrB_Info _cffi_d_GrB_Vector_setElement_UDT(struct GB_Vector_opaque * x0, void * x1, uint64_t x2)
{
  return GrB_Vector_setElement_UDT(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_setElement_UDT(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  void * x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_setElement_UDT", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_setElement_UDT(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_setElement_UDT _cffi_d_GrB_Vector_setElement_UDT
#endif

static GrB_Info _cffi_d_GrB_Vector_setElement_UINT16(struct GB_Vector_opaque * x0, uint16_t x1, uint64_t x2)
{
  return GrB_Vector_setElement_UINT16(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_setElement_UINT16(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint16_t x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_setElement_UINT16", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint16_t);
  if (x1 == (uint16_t)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_setElement_UINT16(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_setElement_UINT16 _cffi_d_GrB_Vector_setElement_UINT16
#endif

static GrB_Info _cffi_d_GrB_Vector_setElement_UINT32(struct GB_Vector_opaque * x0, uint32_t x1, uint64_t x2)
{
  return GrB_Vector_setElement_UINT32(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_setElement_UINT32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint32_t x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_setElement_UINT32", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint32_t);
  if (x1 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_setElement_UINT32(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_setElement_UINT32 _cffi_d_GrB_Vector_setElement_UINT32
#endif

static GrB_Info _cffi_d_GrB_Vector_setElement_UINT64(struct GB_Vector_opaque * x0, uint64_t x1, uint64_t x2)
{
  return GrB_Vector_setElement_UINT64(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_setElement_UINT64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint64_t x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_setElement_UINT64", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint64_t);
  if (x1 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_setElement_UINT64(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_setElement_UINT64 _cffi_d_GrB_Vector_setElement_UINT64
#endif

static GrB_Info _cffi_d_GrB_Vector_setElement_UINT8(struct GB_Vector_opaque * x0, uint8_t x1, uint64_t x2)
{
  return GrB_Vector_setElement_UINT8(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_setElement_UINT8(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint8_t x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_setElement_UINT8", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint8_t);
  if (x1 == (uint8_t)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_setElement_UINT8(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_setElement_UINT8 _cffi_d_GrB_Vector_setElement_UINT8
#endif

static GrB_Info _cffi_d_GrB_Vector_size(uint64_t * x0, struct GB_Vector_opaque * x1)
{
  return GrB_Vector_size(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_size(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  struct GB_Vector_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_size", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_size(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_size _cffi_d_GrB_Vector_size
#endif

static GrB_Info _cffi_d_GrB_Vector_wait(struct GB_Vector_opaque * x0, GrB_WaitMode x1)
{
  return GrB_Vector_wait(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_Vector_wait(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  GrB_WaitMode x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_Vector_wait", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(458), arg1) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_Vector_wait(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_Vector_wait _cffi_d_GrB_Vector_wait
#endif

static GrB_Info _cffi_d_GrB_finalize(void)
{
  return GrB_finalize();
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_finalize(PyObject *self, PyObject *noarg)
{
  GrB_Info result;
  PyObject *pyresult;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_finalize(); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  (void)noarg; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  return pyresult;
}
#else
#  define _cffi_f_GrB_finalize _cffi_d_GrB_finalize
#endif

static GrB_Info _cffi_d_GrB_getVersion(unsigned int * x0, unsigned int * x1)
{
  return GrB_getVersion(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_getVersion(PyObject *self, PyObject *args)
{
  unsigned int * x0;
  unsigned int * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GrB_getVersion", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3839), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (unsigned int *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3839), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3839), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (unsigned int *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3839), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_getVersion(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_getVersion _cffi_d_GrB_getVersion
#endif

static GrB_Info _cffi_d_GrB_init(GrB_Mode x0)
{
  return GrB_init(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_init(PyObject *self, PyObject *arg0)
{
  GrB_Mode x0;
  GrB_Info result;
  PyObject *pyresult;

  if (_cffi_to_c((char *)&x0, _cffi_type(10), arg0) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_init(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  return pyresult;
}
#else
#  define _cffi_f_GrB_init _cffi_d_GrB_init
#endif

static GrB_Info _cffi_d_GrB_mxm(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Semiring_opaque * x3, struct GB_Matrix_opaque * x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_mxm(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_mxm(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Semiring_opaque * x3;
  struct GB_Matrix_opaque * x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_mxm", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(159), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Semiring_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(159), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_mxm(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_mxm _cffi_d_GrB_mxm
#endif

static GrB_Info _cffi_d_GrB_mxv(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Semiring_opaque * x3, struct GB_Matrix_opaque * x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_mxv(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_mxv(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Semiring_opaque * x3;
  struct GB_Matrix_opaque * x4;
  struct GB_Vector_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_mxv", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(159), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Semiring_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(159), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_mxv(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_mxv _cffi_d_GrB_mxv
#endif

static GrB_Info _cffi_d_GrB_transpose(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Matrix_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GrB_transpose(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_transpose(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Matrix_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GrB_transpose", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_transpose(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_transpose _cffi_d_GrB_transpose
#endif

static GrB_Info _cffi_d_GrB_vxm(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Semiring_opaque * x3, struct GB_Vector_opaque * x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GrB_vxm(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GrB_vxm(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Semiring_opaque * x3;
  struct GB_Vector_opaque * x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GrB_vxm", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(159), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Semiring_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(159), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GrB_vxm(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GrB_vxm _cffi_d_GrB_vxm
#endif

static GrB_Info _cffi_d_GxB_BinaryOp_fprint(struct GB_BinaryOp_opaque * x0, char const * x1, GxB_Print_Level x2, FILE * x3)
{
  return GxB_BinaryOp_fprint(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_BinaryOp_fprint(PyObject *self, PyObject *args)
{
  struct GB_BinaryOp_opaque * x0;
  char const * x1;
  GxB_Print_Level x2;
  FILE * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_BinaryOp_fprint", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x2, _cffi_type(463), arg2) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(464), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (FILE *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(464), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_BinaryOp_fprint(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_BinaryOp_fprint _cffi_d_GxB_BinaryOp_fprint
#endif

static GrB_Info _cffi_d_GxB_BinaryOp_new(struct GB_BinaryOp_opaque * * x0, void(* x1)(void *, void const *, void const *), struct GB_Type_opaque * x2, struct GB_Type_opaque * x3, struct GB_Type_opaque * x4, char const * x5, char const * x6)
{
  return GxB_BinaryOp_new(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_BinaryOp_new(PyObject *self, PyObject *args)
{
  struct GB_BinaryOp_opaque * * x0;
  void(* x1)(void *, void const *, void const *);
  struct GB_Type_opaque * x2;
  struct GB_Type_opaque * x3;
  struct GB_Type_opaque * x4;
  char const * x5;
  char const * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_BinaryOp_new", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(422), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(422), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = (void(*)(void *, void const *, void const *))_cffi_to_c_pointer(arg1, _cffi_type(434));
  if (x1 == (void(*)(void *, void const *, void const *))NULL && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_BinaryOp_new(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_BinaryOp_new _cffi_d_GxB_BinaryOp_new
#endif

static GrB_Info _cffi_d_GxB_BinaryOp_xtype(struct GB_Type_opaque * * x0, struct GB_BinaryOp_opaque * x1)
{
  return GxB_BinaryOp_xtype(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_BinaryOp_xtype(PyObject *self, PyObject *args)
{
  struct GB_Type_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_BinaryOp_xtype", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_BinaryOp_xtype(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_BinaryOp_xtype _cffi_d_GxB_BinaryOp_xtype
#endif

static GrB_Info _cffi_d_GxB_BinaryOp_xtype_name(char * x0, struct GB_BinaryOp_opaque * x1)
{
  return GxB_BinaryOp_xtype_name(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_BinaryOp_xtype_name(PyObject *self, PyObject *args)
{
  char * x0;
  struct GB_BinaryOp_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_BinaryOp_xtype_name", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(97), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(97), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_BinaryOp_xtype_name(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_BinaryOp_xtype_name _cffi_d_GxB_BinaryOp_xtype_name
#endif

static GrB_Info _cffi_d_GxB_BinaryOp_ytype(struct GB_Type_opaque * * x0, struct GB_BinaryOp_opaque * x1)
{
  return GxB_BinaryOp_ytype(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_BinaryOp_ytype(PyObject *self, PyObject *args)
{
  struct GB_Type_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_BinaryOp_ytype", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_BinaryOp_ytype(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_BinaryOp_ytype _cffi_d_GxB_BinaryOp_ytype
#endif

static GrB_Info _cffi_d_GxB_BinaryOp_ytype_name(char * x0, struct GB_BinaryOp_opaque * x1)
{
  return GxB_BinaryOp_ytype_name(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_BinaryOp_ytype_name(PyObject *self, PyObject *args)
{
  char * x0;
  struct GB_BinaryOp_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_BinaryOp_ytype_name", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(97), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(97), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_BinaryOp_ytype_name(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_BinaryOp_ytype_name _cffi_d_GxB_BinaryOp_ytype_name
#endif

static GrB_Info _cffi_d_GxB_BinaryOp_ztype(struct GB_Type_opaque * * x0, struct GB_BinaryOp_opaque * x1)
{
  return GxB_BinaryOp_ztype(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_BinaryOp_ztype(PyObject *self, PyObject *args)
{
  struct GB_Type_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_BinaryOp_ztype", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_BinaryOp_ztype(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_BinaryOp_ztype _cffi_d_GxB_BinaryOp_ztype
#endif

static GrB_Info _cffi_d_GxB_BinaryOp_ztype_name(char * x0, struct GB_BinaryOp_opaque * x1)
{
  return GxB_BinaryOp_ztype_name(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_BinaryOp_ztype_name(PyObject *self, PyObject *args)
{
  char * x0;
  struct GB_BinaryOp_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_BinaryOp_ztype_name", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(97), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(97), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_BinaryOp_ztype_name(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_BinaryOp_ztype_name _cffi_d_GxB_BinaryOp_ztype_name
#endif

static GrB_Info _cffi_d_GxB_Col_subassign(struct GB_Matrix_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Vector_opaque * x3, uint64_t const * x4, uint64_t x5, uint64_t x6, struct GB_Descriptor_opaque * x7)
{
  return GxB_Col_subassign(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Col_subassign(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Vector_opaque * x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t x6;
  struct GB_Descriptor_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GxB_Col_subassign", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x6 = _cffi_to_c_int(arg6, uint64_t);
  if (x6 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Col_subassign(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Col_subassign _cffi_d_GxB_Col_subassign
#endif

static void _cffi_const_GxB_Desc_get(char *o)
{
  *(GrB_Info(* *)(struct GB_Descriptor_opaque *, GrB_Desc_Field, ...))o = GxB_Desc_get;
}

static GrB_Info _cffi_d_GxB_Desc_get_FP64(struct GB_Descriptor_opaque * x0, GrB_Desc_Field x1, double * x2)
{
  return GxB_Desc_get_FP64(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Desc_get_FP64(PyObject *self, PyObject *args)
{
  struct GB_Descriptor_opaque * x0;
  GrB_Desc_Field x1;
  double * x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Desc_get_FP64", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(3), arg1) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(28), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(28), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Desc_get_FP64(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Desc_get_FP64 _cffi_d_GxB_Desc_get_FP64
#endif

static GrB_Info _cffi_d_GxB_Desc_get_INT32(struct GB_Descriptor_opaque * x0, GrB_Desc_Field x1, int32_t * x2)
{
  return GxB_Desc_get_INT32(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Desc_get_INT32(PyObject *self, PyObject *args)
{
  struct GB_Descriptor_opaque * x0;
  GrB_Desc_Field x1;
  int32_t * x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Desc_get_INT32", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(3), arg1) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(36), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (int32_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(36), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Desc_get_INT32(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Desc_get_INT32 _cffi_d_GxB_Desc_get_INT32
#endif

static void _cffi_const_GxB_Desc_set(char *o)
{
  *(GrB_Info(* *)(struct GB_Descriptor_opaque *, GrB_Desc_Field, ...))o = GxB_Desc_set;
}

static GrB_Info _cffi_d_GxB_Desc_set_FP64(struct GB_Descriptor_opaque * x0, GrB_Desc_Field x1, double x2)
{
  return GxB_Desc_set_FP64(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Desc_set_FP64(PyObject *self, PyObject *args)
{
  struct GB_Descriptor_opaque * x0;
  GrB_Desc_Field x1;
  double x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Desc_set_FP64", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(3), arg1) < 0)
    return NULL;

  x2 = (double)_cffi_to_c_double(arg2);
  if (x2 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Desc_set_FP64(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Desc_set_FP64 _cffi_d_GxB_Desc_set_FP64
#endif

static GrB_Info _cffi_d_GxB_Desc_set_INT32(struct GB_Descriptor_opaque * x0, GrB_Desc_Field x1, int32_t x2)
{
  return GxB_Desc_set_INT32(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Desc_set_INT32(PyObject *self, PyObject *args)
{
  struct GB_Descriptor_opaque * x0;
  GrB_Desc_Field x1;
  int32_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Desc_set_INT32", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(3), arg1) < 0)
    return NULL;

  x2 = _cffi_to_c_int(arg2, int32_t);
  if (x2 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Desc_set_INT32(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Desc_set_INT32 _cffi_d_GxB_Desc_set_INT32
#endif

static GrB_Info _cffi_d_GxB_Descriptor_fprint(struct GB_Descriptor_opaque * x0, char const * x1, GxB_Print_Level x2, FILE * x3)
{
  return GxB_Descriptor_fprint(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Descriptor_fprint(PyObject *self, PyObject *args)
{
  struct GB_Descriptor_opaque * x0;
  char const * x1;
  GxB_Print_Level x2;
  FILE * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Descriptor_fprint", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x2, _cffi_type(463), arg2) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(464), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (FILE *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(464), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Descriptor_fprint(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Descriptor_fprint _cffi_d_GxB_Descriptor_fprint
#endif

static GrB_Info _cffi_d_GxB_Descriptor_get(GrB_Desc_Value * x0, struct GB_Descriptor_opaque * x1, GrB_Desc_Field x2)
{
  return GxB_Descriptor_get(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Descriptor_get(PyObject *self, PyObject *args)
{
  GrB_Desc_Value * x0;
  struct GB_Descriptor_opaque * x1;
  GrB_Desc_Field x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Descriptor_get", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (GrB_Desc_Value *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x2, _cffi_type(3), arg2) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Descriptor_get(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Descriptor_get _cffi_d_GxB_Descriptor_get
#endif

static void _cffi_const_GxB_Global_Option_get(char *o)
{
  *(GrB_Info(* *)(GxB_Option_Field, ...))o = GxB_Global_Option_get;
}

static GrB_Info _cffi_d_GxB_Global_Option_get_CHAR(GxB_Option_Field x0, char * * x1)
{
  return GxB_Global_Option_get_CHAR(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Global_Option_get_CHAR(PyObject *self, PyObject *args)
{
  GxB_Option_Field x0;
  char * * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Global_Option_get_CHAR", 2, 2, &arg0, &arg1))
    return NULL;

  if (_cffi_to_c((char *)&x0, _cffi_type(20), arg0) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(24), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (char * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(24), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Global_Option_get_CHAR(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Global_Option_get_CHAR _cffi_d_GxB_Global_Option_get_CHAR
#endif

static GrB_Info _cffi_d_GxB_Global_Option_get_FP64(GxB_Option_Field x0, double * x1)
{
  return GxB_Global_Option_get_FP64(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Global_Option_get_FP64(PyObject *self, PyObject *args)
{
  GxB_Option_Field x0;
  double * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Global_Option_get_FP64", 2, 2, &arg0, &arg1))
    return NULL;

  if (_cffi_to_c((char *)&x0, _cffi_type(20), arg0) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(28), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(28), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Global_Option_get_FP64(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Global_Option_get_FP64 _cffi_d_GxB_Global_Option_get_FP64
#endif

static GrB_Info _cffi_d_GxB_Global_Option_get_FUNCTION(GxB_Option_Field x0, void * * x1)
{
  return GxB_Global_Option_get_FUNCTION(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Global_Option_get_FUNCTION(PyObject *self, PyObject *args)
{
  GxB_Option_Field x0;
  void * * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Global_Option_get_FUNCTION", 2, 2, &arg0, &arg1))
    return NULL;

  if (_cffi_to_c((char *)&x0, _cffi_type(20), arg0) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Global_Option_get_FUNCTION(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Global_Option_get_FUNCTION _cffi_d_GxB_Global_Option_get_FUNCTION
#endif

static GrB_Info _cffi_d_GxB_Global_Option_get_INT32(GxB_Option_Field x0, int32_t * x1)
{
  return GxB_Global_Option_get_INT32(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Global_Option_get_INT32(PyObject *self, PyObject *args)
{
  GxB_Option_Field x0;
  int32_t * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Global_Option_get_INT32", 2, 2, &arg0, &arg1))
    return NULL;

  if (_cffi_to_c((char *)&x0, _cffi_type(20), arg0) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(36), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (int32_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(36), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Global_Option_get_INT32(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Global_Option_get_INT32 _cffi_d_GxB_Global_Option_get_INT32
#endif

static GrB_Info _cffi_d_GxB_Global_Option_get_INT64(GxB_Option_Field x0, int64_t * x1)
{
  return GxB_Global_Option_get_INT64(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Global_Option_get_INT64(PyObject *self, PyObject *args)
{
  GxB_Option_Field x0;
  int64_t * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Global_Option_get_INT64", 2, 2, &arg0, &arg1))
    return NULL;

  if (_cffi_to_c((char *)&x0, _cffi_type(20), arg0) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(44), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (int64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(44), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Global_Option_get_INT64(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Global_Option_get_INT64 _cffi_d_GxB_Global_Option_get_INT64
#endif

static void _cffi_const_GxB_Global_Option_set(char *o)
{
  *(GrB_Info(* *)(GxB_Option_Field, ...))o = GxB_Global_Option_set;
}

static GrB_Info _cffi_d_GxB_Global_Option_set_FP64(GxB_Option_Field x0, double x1)
{
  return GxB_Global_Option_set_FP64(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Global_Option_set_FP64(PyObject *self, PyObject *args)
{
  GxB_Option_Field x0;
  double x1;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Global_Option_set_FP64", 2, 2, &arg0, &arg1))
    return NULL;

  if (_cffi_to_c((char *)&x0, _cffi_type(20), arg0) < 0)
    return NULL;

  x1 = (double)_cffi_to_c_double(arg1);
  if (x1 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Global_Option_set_FP64(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  return pyresult;
}
#else
#  define _cffi_f_GxB_Global_Option_set_FP64 _cffi_d_GxB_Global_Option_set_FP64
#endif

static GrB_Info _cffi_d_GxB_Global_Option_set_FP64_ARRAY(GxB_Option_Field x0, double * x1)
{
  return GxB_Global_Option_set_FP64_ARRAY(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Global_Option_set_FP64_ARRAY(PyObject *self, PyObject *args)
{
  GxB_Option_Field x0;
  double * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Global_Option_set_FP64_ARRAY", 2, 2, &arg0, &arg1))
    return NULL;

  if (_cffi_to_c((char *)&x0, _cffi_type(20), arg0) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(28), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(28), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Global_Option_set_FP64_ARRAY(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Global_Option_set_FP64_ARRAY _cffi_d_GxB_Global_Option_set_FP64_ARRAY
#endif

static GrB_Info _cffi_d_GxB_Global_Option_set_FUNCTION(GxB_Option_Field x0, void * x1)
{
  return GxB_Global_Option_set_FUNCTION(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Global_Option_set_FUNCTION(PyObject *self, PyObject *args)
{
  GxB_Option_Field x0;
  void * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Global_Option_set_FUNCTION", 2, 2, &arg0, &arg1))
    return NULL;

  if (_cffi_to_c((char *)&x0, _cffi_type(20), arg0) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Global_Option_set_FUNCTION(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Global_Option_set_FUNCTION _cffi_d_GxB_Global_Option_set_FUNCTION
#endif

static GrB_Info _cffi_d_GxB_Global_Option_set_INT32(GxB_Option_Field x0, int32_t x1)
{
  return GxB_Global_Option_set_INT32(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Global_Option_set_INT32(PyObject *self, PyObject *args)
{
  GxB_Option_Field x0;
  int32_t x1;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Global_Option_set_INT32", 2, 2, &arg0, &arg1))
    return NULL;

  if (_cffi_to_c((char *)&x0, _cffi_type(20), arg0) < 0)
    return NULL;

  x1 = _cffi_to_c_int(arg1, int32_t);
  if (x1 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Global_Option_set_INT32(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  return pyresult;
}
#else
#  define _cffi_f_GxB_Global_Option_set_INT32 _cffi_d_GxB_Global_Option_set_INT32
#endif

static GrB_Info _cffi_d_GxB_Global_Option_set_INT64_ARRAY(GxB_Option_Field x0, int64_t * x1)
{
  return GxB_Global_Option_set_INT64_ARRAY(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Global_Option_set_INT64_ARRAY(PyObject *self, PyObject *args)
{
  GxB_Option_Field x0;
  int64_t * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Global_Option_set_INT64_ARRAY", 2, 2, &arg0, &arg1))
    return NULL;

  if (_cffi_to_c((char *)&x0, _cffi_type(20), arg0) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(44), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (int64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(44), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Global_Option_set_INT64_ARRAY(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Global_Option_set_INT64_ARRAY _cffi_d_GxB_Global_Option_set_INT64_ARRAY
#endif

static GrB_Info _cffi_d_GxB_IndexUnaryOp_fprint(struct GB_IndexUnaryOp_opaque * x0, char const * x1, GxB_Print_Level x2, FILE * x3)
{
  return GxB_IndexUnaryOp_fprint(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_IndexUnaryOp_fprint(PyObject *self, PyObject *args)
{
  struct GB_IndexUnaryOp_opaque * x0;
  char const * x1;
  GxB_Print_Level x2;
  FILE * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_IndexUnaryOp_fprint", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x2, _cffi_type(463), arg2) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(464), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (FILE *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(464), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_IndexUnaryOp_fprint(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_IndexUnaryOp_fprint _cffi_d_GxB_IndexUnaryOp_fprint
#endif

static GrB_Info _cffi_d_GxB_IndexUnaryOp_new(struct GB_IndexUnaryOp_opaque * * x0, void(* x1)(void *, void const *, uint64_t, uint64_t, void const *), struct GB_Type_opaque * x2, struct GB_Type_opaque * x3, struct GB_Type_opaque * x4, char const * x5, char const * x6)
{
  return GxB_IndexUnaryOp_new(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_IndexUnaryOp_new(PyObject *self, PyObject *args)
{
  struct GB_IndexUnaryOp_opaque * * x0;
  void(* x1)(void *, void const *, uint64_t, uint64_t, void const *);
  struct GB_Type_opaque * x2;
  struct GB_Type_opaque * x3;
  struct GB_Type_opaque * x4;
  char const * x5;
  char const * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_IndexUnaryOp_new", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(509), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(509), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = (void(*)(void *, void const *, uint64_t, uint64_t, void const *))_cffi_to_c_pointer(arg1, _cffi_type(513));
  if (x1 == (void(*)(void *, void const *, uint64_t, uint64_t, void const *))NULL && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_IndexUnaryOp_new(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_IndexUnaryOp_new _cffi_d_GxB_IndexUnaryOp_new
#endif

static GrB_Info _cffi_d_GxB_IndexUnaryOp_xtype_name(char * x0, struct GB_IndexUnaryOp_opaque * x1)
{
  return GxB_IndexUnaryOp_xtype_name(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_IndexUnaryOp_xtype_name(PyObject *self, PyObject *args)
{
  char * x0;
  struct GB_IndexUnaryOp_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_IndexUnaryOp_xtype_name", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(97), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(97), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_IndexUnaryOp_xtype_name(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_IndexUnaryOp_xtype_name _cffi_d_GxB_IndexUnaryOp_xtype_name
#endif

static GrB_Info _cffi_d_GxB_IndexUnaryOp_ytype_name(char * x0, struct GB_IndexUnaryOp_opaque * x1)
{
  return GxB_IndexUnaryOp_ytype_name(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_IndexUnaryOp_ytype_name(PyObject *self, PyObject *args)
{
  char * x0;
  struct GB_IndexUnaryOp_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_IndexUnaryOp_ytype_name", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(97), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(97), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_IndexUnaryOp_ytype_name(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_IndexUnaryOp_ytype_name _cffi_d_GxB_IndexUnaryOp_ytype_name
#endif

static GrB_Info _cffi_d_GxB_IndexUnaryOp_ztype_name(char * x0, struct GB_IndexUnaryOp_opaque * x1)
{
  return GxB_IndexUnaryOp_ztype_name(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_IndexUnaryOp_ztype_name(PyObject *self, PyObject *args)
{
  char * x0;
  struct GB_IndexUnaryOp_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_IndexUnaryOp_ztype_name", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(97), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(97), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_IndexUnaryOp_ztype_name(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_IndexUnaryOp_ztype_name _cffi_d_GxB_IndexUnaryOp_ztype_name
#endif

static GrB_Info _cffi_d_GxB_Iterator_free(struct GB_Iterator_opaque * * x0)
{
  return GxB_Iterator_free(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Iterator_free(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(538), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(538), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Iterator_free(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Iterator_free _cffi_d_GxB_Iterator_free
#endif

static _Bool _cffi_d_GxB_Iterator_get_BOOL(struct GB_Iterator_opaque * x0)
{
  return GxB_Iterator_get_BOOL(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Iterator_get_BOOL(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  _Bool result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Iterator_get_BOOL(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c__Bool(result);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Iterator_get_BOOL _cffi_d_GxB_Iterator_get_BOOL
#endif

static _Fcomplex _cffi_d_GxB_Iterator_get_FC32(struct GB_Iterator_opaque * x0)
{
  return GxB_Iterator_get_FC32(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Iterator_get_FC32(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  _Fcomplex result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Iterator_get_FC32(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(974));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static void _cffi_f_GxB_Iterator_get_FC32(_Fcomplex *result, struct GB_Iterator_opaque * x0)
{
  { *result = GxB_Iterator_get_FC32(x0); }
}
#endif

static _Dcomplex _cffi_d_GxB_Iterator_get_FC64(struct GB_Iterator_opaque * x0)
{
  return GxB_Iterator_get_FC64(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Iterator_get_FC64(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  _Dcomplex result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Iterator_get_FC64(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(962));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static void _cffi_f_GxB_Iterator_get_FC64(_Dcomplex *result, struct GB_Iterator_opaque * x0)
{
  { *result = GxB_Iterator_get_FC64(x0); }
}
#endif

static float _cffi_d_GxB_Iterator_get_FP32(struct GB_Iterator_opaque * x0)
{
  return GxB_Iterator_get_FP32(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Iterator_get_FP32(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  float result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Iterator_get_FP32(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_float(result);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Iterator_get_FP32 _cffi_d_GxB_Iterator_get_FP32
#endif

static double _cffi_d_GxB_Iterator_get_FP64(struct GB_Iterator_opaque * x0)
{
  return GxB_Iterator_get_FP64(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Iterator_get_FP64(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  double result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Iterator_get_FP64(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_double(result);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Iterator_get_FP64 _cffi_d_GxB_Iterator_get_FP64
#endif

static int16_t _cffi_d_GxB_Iterator_get_INT16(struct GB_Iterator_opaque * x0)
{
  return GxB_Iterator_get_INT16(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Iterator_get_INT16(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  int16_t result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Iterator_get_INT16(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, int16_t);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Iterator_get_INT16 _cffi_d_GxB_Iterator_get_INT16
#endif

static int32_t _cffi_d_GxB_Iterator_get_INT32(struct GB_Iterator_opaque * x0)
{
  return GxB_Iterator_get_INT32(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Iterator_get_INT32(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  int32_t result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Iterator_get_INT32(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, int32_t);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Iterator_get_INT32 _cffi_d_GxB_Iterator_get_INT32
#endif

static int64_t _cffi_d_GxB_Iterator_get_INT64(struct GB_Iterator_opaque * x0)
{
  return GxB_Iterator_get_INT64(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Iterator_get_INT64(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  int64_t result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Iterator_get_INT64(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, int64_t);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Iterator_get_INT64 _cffi_d_GxB_Iterator_get_INT64
#endif

static int8_t _cffi_d_GxB_Iterator_get_INT8(struct GB_Iterator_opaque * x0)
{
  return GxB_Iterator_get_INT8(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Iterator_get_INT8(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  int8_t result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Iterator_get_INT8(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, int8_t);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Iterator_get_INT8 _cffi_d_GxB_Iterator_get_INT8
#endif

static void _cffi_d_GxB_Iterator_get_UDT(struct GB_Iterator_opaque * x0, void * x1)
{
  GxB_Iterator_get_UDT(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Iterator_get_UDT(PyObject *self, PyObject *args)
{
  struct GB_Iterator_opaque * x0;
  void * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Iterator_get_UDT", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { GxB_Iterator_get_UDT(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_GxB_Iterator_get_UDT _cffi_d_GxB_Iterator_get_UDT
#endif

static uint16_t _cffi_d_GxB_Iterator_get_UINT16(struct GB_Iterator_opaque * x0)
{
  return GxB_Iterator_get_UINT16(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Iterator_get_UINT16(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  uint16_t result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Iterator_get_UINT16(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, uint16_t);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Iterator_get_UINT16 _cffi_d_GxB_Iterator_get_UINT16
#endif

static uint32_t _cffi_d_GxB_Iterator_get_UINT32(struct GB_Iterator_opaque * x0)
{
  return GxB_Iterator_get_UINT32(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Iterator_get_UINT32(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  uint32_t result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Iterator_get_UINT32(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, uint32_t);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Iterator_get_UINT32 _cffi_d_GxB_Iterator_get_UINT32
#endif

static uint64_t _cffi_d_GxB_Iterator_get_UINT64(struct GB_Iterator_opaque * x0)
{
  return GxB_Iterator_get_UINT64(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Iterator_get_UINT64(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  uint64_t result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Iterator_get_UINT64(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, uint64_t);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Iterator_get_UINT64 _cffi_d_GxB_Iterator_get_UINT64
#endif

static uint8_t _cffi_d_GxB_Iterator_get_UINT8(struct GB_Iterator_opaque * x0)
{
  return GxB_Iterator_get_UINT8(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Iterator_get_UINT8(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  uint8_t result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Iterator_get_UINT8(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, uint8_t);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Iterator_get_UINT8 _cffi_d_GxB_Iterator_get_UINT8
#endif

static GrB_Info _cffi_d_GxB_Iterator_new(struct GB_Iterator_opaque * * x0)
{
  return GxB_Iterator_new(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Iterator_new(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(538), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(538), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Iterator_new(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Iterator_new _cffi_d_GxB_Iterator_new
#endif

static GrB_Info _cffi_d_GxB_Matrix_Iterator_attach(struct GB_Iterator_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_Descriptor_opaque * x2)
{
  return GxB_Matrix_Iterator_attach(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_Iterator_attach(PyObject *self, PyObject *args)
{
  struct GB_Iterator_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_Descriptor_opaque * x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_Iterator_attach", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_Iterator_attach(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_Iterator_attach _cffi_d_GxB_Matrix_Iterator_attach
#endif

static void _cffi_d_GxB_Matrix_Iterator_getIndex(struct GB_Iterator_opaque * x0, uint64_t * x1, uint64_t * x2)
{
  GxB_Matrix_Iterator_getIndex(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_Iterator_getIndex(PyObject *self, PyObject *args)
{
  struct GB_Iterator_opaque * x0;
  uint64_t * x1;
  uint64_t * x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_Iterator_getIndex", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { GxB_Matrix_Iterator_getIndex(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  Py_INCREF(Py_None);
  return Py_None;
}
#else
#  define _cffi_f_GxB_Matrix_Iterator_getIndex _cffi_d_GxB_Matrix_Iterator_getIndex
#endif

static uint64_t _cffi_d_GxB_Matrix_Iterator_getp(struct GB_Iterator_opaque * x0)
{
  return GxB_Matrix_Iterator_getp(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_Iterator_getp(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  uint64_t result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_Iterator_getp(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, uint64_t);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_Iterator_getp _cffi_d_GxB_Matrix_Iterator_getp
#endif

static uint64_t _cffi_d_GxB_Matrix_Iterator_getpmax(struct GB_Iterator_opaque * x0)
{
  return GxB_Matrix_Iterator_getpmax(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_Iterator_getpmax(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  uint64_t result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_Iterator_getpmax(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, uint64_t);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_Iterator_getpmax _cffi_d_GxB_Matrix_Iterator_getpmax
#endif

static GrB_Info _cffi_d_GxB_Matrix_Iterator_next(struct GB_Iterator_opaque * x0)
{
  return GxB_Matrix_Iterator_next(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_Iterator_next(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_Iterator_next(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_Iterator_next _cffi_d_GxB_Matrix_Iterator_next
#endif

static GrB_Info _cffi_d_GxB_Matrix_Iterator_seek(struct GB_Iterator_opaque * x0, uint64_t x1)
{
  return GxB_Matrix_Iterator_seek(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_Iterator_seek(PyObject *self, PyObject *args)
{
  struct GB_Iterator_opaque * x0;
  uint64_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_Iterator_seek", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint64_t);
  if (x1 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_Iterator_seek(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_Iterator_seek _cffi_d_GxB_Matrix_Iterator_seek
#endif

static void _cffi_const_GxB_Matrix_Option_get(char *o)
{
  *(GrB_Info(* *)(struct GB_Matrix_opaque *, GxB_Option_Field, ...))o = GxB_Matrix_Option_get;
}

static GrB_Info _cffi_d_GxB_Matrix_Option_get_FP64(struct GB_Matrix_opaque * x0, GxB_Option_Field x1, double * x2)
{
  return GxB_Matrix_Option_get_FP64(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_Option_get_FP64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  GxB_Option_Field x1;
  double * x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_Option_get_FP64", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(20), arg1) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(28), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(28), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_Option_get_FP64(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_Option_get_FP64 _cffi_d_GxB_Matrix_Option_get_FP64
#endif

static GrB_Info _cffi_d_GxB_Matrix_Option_get_INT32(struct GB_Matrix_opaque * x0, GxB_Option_Field x1, int32_t * x2)
{
  return GxB_Matrix_Option_get_INT32(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_Option_get_INT32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  GxB_Option_Field x1;
  int32_t * x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_Option_get_INT32", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(20), arg1) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(36), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (int32_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(36), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_Option_get_INT32(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_Option_get_INT32 _cffi_d_GxB_Matrix_Option_get_INT32
#endif

static void _cffi_const_GxB_Matrix_Option_set(char *o)
{
  *(GrB_Info(* *)(struct GB_Matrix_opaque *, GxB_Option_Field, ...))o = GxB_Matrix_Option_set;
}

static GrB_Info _cffi_d_GxB_Matrix_Option_set_FP64(struct GB_Matrix_opaque * x0, GxB_Option_Field x1, double x2)
{
  return GxB_Matrix_Option_set_FP64(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_Option_set_FP64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  GxB_Option_Field x1;
  double x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_Option_set_FP64", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(20), arg1) < 0)
    return NULL;

  x2 = (double)_cffi_to_c_double(arg2);
  if (x2 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_Option_set_FP64(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_Option_set_FP64 _cffi_d_GxB_Matrix_Option_set_FP64
#endif

static GrB_Info _cffi_d_GxB_Matrix_Option_set_INT32(struct GB_Matrix_opaque * x0, GxB_Option_Field x1, int32_t x2)
{
  return GxB_Matrix_Option_set_INT32(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_Option_set_INT32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  GxB_Option_Field x1;
  int32_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_Option_set_INT32", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(20), arg1) < 0)
    return NULL;

  x2 = _cffi_to_c_int(arg2, int32_t);
  if (x2 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_Option_set_INT32(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_Option_set_INT32 _cffi_d_GxB_Matrix_Option_set_INT32
#endif

static GrB_Info _cffi_d_GxB_Matrix_apply_BinaryOp1st(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Scalar_opaque * x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Matrix_apply_BinaryOp1st(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_apply_BinaryOp1st(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Scalar_opaque * x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_apply_BinaryOp1st", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_apply_BinaryOp1st(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_apply_BinaryOp1st _cffi_d_GxB_Matrix_apply_BinaryOp1st
#endif

static GrB_Info _cffi_d_GxB_Matrix_apply_BinaryOp1st_FC32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, _Fcomplex x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Matrix_apply_BinaryOp1st_FC32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_apply_BinaryOp1st_FC32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  _Fcomplex x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_apply_BinaryOp1st_FC32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x4, _cffi_type(974), arg4) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_apply_BinaryOp1st_FC32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Matrix_apply_BinaryOp1st_FC32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, _Fcomplex *x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  GrB_Info result;
  { result = GxB_Matrix_apply_BinaryOp1st_FC32(x0, x1, x2, x3, *x4, x5, x6); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Matrix_apply_BinaryOp1st_FC64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, _Dcomplex x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Matrix_apply_BinaryOp1st_FC64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_apply_BinaryOp1st_FC64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  _Dcomplex x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_apply_BinaryOp1st_FC64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x4, _cffi_type(962), arg4) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_apply_BinaryOp1st_FC64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Matrix_apply_BinaryOp1st_FC64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, _Dcomplex *x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  GrB_Info result;
  { result = GxB_Matrix_apply_BinaryOp1st_FC64(x0, x1, x2, x3, *x4, x5, x6); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Matrix_apply_BinaryOp2nd(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Matrix_opaque * x4, struct GB_Scalar_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Matrix_apply_BinaryOp2nd(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_apply_BinaryOp2nd(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  struct GB_Scalar_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_apply_BinaryOp2nd", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_apply_BinaryOp2nd(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_apply_BinaryOp2nd _cffi_d_GxB_Matrix_apply_BinaryOp2nd
#endif

static GrB_Info _cffi_d_GxB_Matrix_apply_BinaryOp2nd_FC32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Matrix_opaque * x4, _Fcomplex x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Matrix_apply_BinaryOp2nd_FC32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_apply_BinaryOp2nd_FC32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  _Fcomplex x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_apply_BinaryOp2nd_FC32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x5, _cffi_type(974), arg5) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_apply_BinaryOp2nd_FC32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Matrix_apply_BinaryOp2nd_FC32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Matrix_opaque * x4, _Fcomplex *x5, struct GB_Descriptor_opaque * x6)
{
  GrB_Info result;
  { result = GxB_Matrix_apply_BinaryOp2nd_FC32(x0, x1, x2, x3, x4, *x5, x6); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Matrix_apply_BinaryOp2nd_FC64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Matrix_opaque * x4, _Dcomplex x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Matrix_apply_BinaryOp2nd_FC64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_apply_BinaryOp2nd_FC64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  _Dcomplex x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_apply_BinaryOp2nd_FC64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x5, _cffi_type(962), arg5) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_apply_BinaryOp2nd_FC64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Matrix_apply_BinaryOp2nd_FC64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Matrix_opaque * x4, _Dcomplex *x5, struct GB_Descriptor_opaque * x6)
{
  GrB_Info result;
  { result = GxB_Matrix_apply_BinaryOp2nd_FC64(x0, x1, x2, x3, x4, *x5, x6); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Matrix_apply_IndexOp_FC32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, _Fcomplex x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Matrix_apply_IndexOp_FC32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_apply_IndexOp_FC32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  _Fcomplex x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_apply_IndexOp_FC32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x5, _cffi_type(974), arg5) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_apply_IndexOp_FC32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Matrix_apply_IndexOp_FC32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, _Fcomplex *x5, struct GB_Descriptor_opaque * x6)
{
  GrB_Info result;
  { result = GxB_Matrix_apply_IndexOp_FC32(x0, x1, x2, x3, x4, *x5, x6); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Matrix_apply_IndexOp_FC64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, _Dcomplex x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Matrix_apply_IndexOp_FC64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_apply_IndexOp_FC64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  _Dcomplex x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_apply_IndexOp_FC64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x5, _cffi_type(962), arg5) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_apply_IndexOp_FC64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Matrix_apply_IndexOp_FC64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, _Dcomplex *x5, struct GB_Descriptor_opaque * x6)
{
  GrB_Info result;
  { result = GxB_Matrix_apply_IndexOp_FC64(x0, x1, x2, x3, x4, *x5, x6); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Matrix_assign_FC32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, _Fcomplex x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GxB_Matrix_assign_FC32(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_assign_FC32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  _Fcomplex x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_assign_FC32", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x3, _cffi_type(974), arg3) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_assign_FC32(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Matrix_assign_FC32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, _Fcomplex *x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  GrB_Info result;
  { result = GxB_Matrix_assign_FC32(x0, x1, x2, *x3, x4, x5, x6, x7, x8); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Matrix_assign_FC64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, _Dcomplex x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GxB_Matrix_assign_FC64(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_assign_FC64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  _Dcomplex x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_assign_FC64", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x3, _cffi_type(962), arg3) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_assign_FC64(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Matrix_assign_FC64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, _Dcomplex *x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  GrB_Info result;
  { result = GxB_Matrix_assign_FC64(x0, x1, x2, *x3, x4, x5, x6, x7, x8); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Matrix_build_FC32(struct GB_Matrix_opaque * x0, uint64_t const * x1, uint64_t const * x2, _Fcomplex const * x3, uint64_t x4, struct GB_BinaryOp_opaque * x5)
{
  return GxB_Matrix_build_FC32(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_build_FC32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t const * x1;
  uint64_t const * x2;
  _Fcomplex const * x3;
  uint64_t x4;
  struct GB_BinaryOp_opaque * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_build_FC32", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(737), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (_Fcomplex const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(737), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_build_FC32(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_build_FC32 _cffi_d_GxB_Matrix_build_FC32
#endif

static GrB_Info _cffi_d_GxB_Matrix_build_FC64(struct GB_Matrix_opaque * x0, uint64_t const * x1, uint64_t const * x2, _Dcomplex const * x3, uint64_t x4, struct GB_BinaryOp_opaque * x5)
{
  return GxB_Matrix_build_FC64(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_build_FC64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t const * x1;
  uint64_t const * x2;
  _Dcomplex const * x3;
  uint64_t x4;
  struct GB_BinaryOp_opaque * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_build_FC64", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(711), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (_Dcomplex const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(711), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_build_FC64(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_build_FC64 _cffi_d_GxB_Matrix_build_FC64
#endif

static GrB_Info _cffi_d_GxB_Matrix_build_Scalar(struct GB_Matrix_opaque * x0, uint64_t const * x1, uint64_t const * x2, struct GB_Scalar_opaque * x3, uint64_t x4)
{
  return GxB_Matrix_build_Scalar(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_build_Scalar(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t const * x1;
  uint64_t const * x2;
  struct GB_Scalar_opaque * x3;
  uint64_t x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_build_Scalar", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_build_Scalar(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_build_Scalar _cffi_d_GxB_Matrix_build_Scalar
#endif

static GrB_Info _cffi_d_GxB_Matrix_concat(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * const * x1, uint64_t x2, uint64_t x3, struct GB_Descriptor_opaque * x4)
{
  return GxB_Matrix_concat(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_concat(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * const * x1;
  uint64_t x2;
  uint64_t x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_concat", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1035), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1035), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_concat(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_concat _cffi_d_GxB_Matrix_concat
#endif

static GrB_Info _cffi_d_GxB_Matrix_deserialize(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, void const * x2, uint64_t x3, struct GB_Descriptor_opaque * x4)
{
  return GxB_Matrix_deserialize(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_deserialize(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  void const * x2;
  uint64_t x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_deserialize", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(126), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (void const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(126), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_deserialize(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_deserialize _cffi_d_GxB_Matrix_deserialize
#endif

static GrB_Info _cffi_d_GxB_Matrix_diag(struct GB_Matrix_opaque * x0, struct GB_Vector_opaque * x1, int64_t x2, struct GB_Descriptor_opaque * x3)
{
  return GxB_Matrix_diag(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_diag(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Vector_opaque * x1;
  int64_t x2;
  struct GB_Descriptor_opaque * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_diag", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, int64_t);
  if (x2 == (int64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_diag(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_diag _cffi_d_GxB_Matrix_diag
#endif

static GrB_Info _cffi_d_GxB_Matrix_eWiseUnion(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Matrix_opaque * x4, struct GB_Scalar_opaque * x5, struct GB_Matrix_opaque * x6, struct GB_Scalar_opaque * x7, struct GB_Descriptor_opaque * x8)
{
  return GxB_Matrix_eWiseUnion(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_eWiseUnion(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  struct GB_Scalar_opaque * x5;
  struct GB_Matrix_opaque * x6;
  struct GB_Scalar_opaque * x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_eWiseUnion", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_eWiseUnion(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_eWiseUnion _cffi_d_GxB_Matrix_eWiseUnion
#endif

static GrB_Info _cffi_d_GxB_Matrix_export_BitmapC(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * * x1, uint64_t * x2, uint64_t * x3, int8_t * * x4, void * * x5, uint64_t * x6, uint64_t * x7, _Bool * x8, uint64_t * x9, struct GB_Descriptor_opaque * x10)
{
  return GxB_Matrix_export_BitmapC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_export_BitmapC(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * * x1;
  uint64_t * x2;
  uint64_t * x3;
  int8_t * * x4;
  void * * x5;
  uint64_t * x6;
  uint64_t * x7;
  _Bool * x8;
  uint64_t * x9;
  struct GB_Descriptor_opaque * x10;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_export_BitmapC", 11, 11, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(588), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (int8_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(588), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg9, (char **)&x9);
  if (datasize != 0) {
    x9 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg9, (char **)&x9,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg10, (char **)&x10);
  if (datasize != 0) {
    x10 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg10, (char **)&x10,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_export_BitmapC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_export_BitmapC _cffi_d_GxB_Matrix_export_BitmapC
#endif

static GrB_Info _cffi_d_GxB_Matrix_export_BitmapR(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * * x1, uint64_t * x2, uint64_t * x3, int8_t * * x4, void * * x5, uint64_t * x6, uint64_t * x7, _Bool * x8, uint64_t * x9, struct GB_Descriptor_opaque * x10)
{
  return GxB_Matrix_export_BitmapR(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_export_BitmapR(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * * x1;
  uint64_t * x2;
  uint64_t * x3;
  int8_t * * x4;
  void * * x5;
  uint64_t * x6;
  uint64_t * x7;
  _Bool * x8;
  uint64_t * x9;
  struct GB_Descriptor_opaque * x10;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_export_BitmapR", 11, 11, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(588), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (int8_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(588), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg9, (char **)&x9);
  if (datasize != 0) {
    x9 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg9, (char **)&x9,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg10, (char **)&x10);
  if (datasize != 0) {
    x10 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg10, (char **)&x10,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_export_BitmapR(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_export_BitmapR _cffi_d_GxB_Matrix_export_BitmapR
#endif

static GrB_Info _cffi_d_GxB_Matrix_export_CSC(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * * x1, uint64_t * x2, uint64_t * x3, uint64_t * * x4, uint64_t * * x5, void * * x6, uint64_t * x7, uint64_t * x8, uint64_t * x9, _Bool * x10, _Bool * x11, struct GB_Descriptor_opaque * x12)
{
  return GxB_Matrix_export_CSC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_export_CSC(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * * x1;
  uint64_t * x2;
  uint64_t * x3;
  uint64_t * * x4;
  uint64_t * * x5;
  void * * x6;
  uint64_t * x7;
  uint64_t * x8;
  uint64_t * x9;
  _Bool * x10;
  _Bool * x11;
  struct GB_Descriptor_opaque * x12;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;
  PyObject *arg11;
  PyObject *arg12;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_export_CSC", 13, 13, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10, &arg11, &arg12))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg9, (char **)&x9);
  if (datasize != 0) {
    x9 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg9, (char **)&x9,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg10, (char **)&x10);
  if (datasize != 0) {
    x10 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg10, (char **)&x10,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg11, (char **)&x11);
  if (datasize != 0) {
    x11 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg11, (char **)&x11,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg12, (char **)&x12);
  if (datasize != 0) {
    x12 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg12, (char **)&x12,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_export_CSC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_export_CSC _cffi_d_GxB_Matrix_export_CSC
#endif

static GrB_Info _cffi_d_GxB_Matrix_export_CSR(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * * x1, uint64_t * x2, uint64_t * x3, uint64_t * * x4, uint64_t * * x5, void * * x6, uint64_t * x7, uint64_t * x8, uint64_t * x9, _Bool * x10, _Bool * x11, struct GB_Descriptor_opaque * x12)
{
  return GxB_Matrix_export_CSR(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_export_CSR(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * * x1;
  uint64_t * x2;
  uint64_t * x3;
  uint64_t * * x4;
  uint64_t * * x5;
  void * * x6;
  uint64_t * x7;
  uint64_t * x8;
  uint64_t * x9;
  _Bool * x10;
  _Bool * x11;
  struct GB_Descriptor_opaque * x12;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;
  PyObject *arg11;
  PyObject *arg12;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_export_CSR", 13, 13, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10, &arg11, &arg12))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg9, (char **)&x9);
  if (datasize != 0) {
    x9 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg9, (char **)&x9,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg10, (char **)&x10);
  if (datasize != 0) {
    x10 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg10, (char **)&x10,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg11, (char **)&x11);
  if (datasize != 0) {
    x11 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg11, (char **)&x11,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg12, (char **)&x12);
  if (datasize != 0) {
    x12 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg12, (char **)&x12,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_export_CSR(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_export_CSR _cffi_d_GxB_Matrix_export_CSR
#endif

static GrB_Info _cffi_d_GxB_Matrix_export_FC32(uint64_t * x0, uint64_t * x1, _Fcomplex * x2, uint64_t * x3, uint64_t * x4, uint64_t * x5, GrB_Format x6, struct GB_Matrix_opaque * x7)
{
  return GxB_Matrix_export_FC32(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_export_FC32(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  _Fcomplex * x2;
  uint64_t * x3;
  uint64_t * x4;
  uint64_t * x5;
  GrB_Format x6;
  struct GB_Matrix_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_export_FC32", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(261), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (_Fcomplex *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(261), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x6, _cffi_type(702), arg6) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_export_FC32(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_export_FC32 _cffi_d_GxB_Matrix_export_FC32
#endif

static GrB_Info _cffi_d_GxB_Matrix_export_FC64(uint64_t * x0, uint64_t * x1, _Dcomplex * x2, uint64_t * x3, uint64_t * x4, uint64_t * x5, GrB_Format x6, struct GB_Matrix_opaque * x7)
{
  return GxB_Matrix_export_FC64(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_export_FC64(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  _Dcomplex * x2;
  uint64_t * x3;
  uint64_t * x4;
  uint64_t * x5;
  GrB_Format x6;
  struct GB_Matrix_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_export_FC64", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(203), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (_Dcomplex *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(203), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x6, _cffi_type(702), arg6) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_export_FC64(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_export_FC64 _cffi_d_GxB_Matrix_export_FC64
#endif

static GrB_Info _cffi_d_GxB_Matrix_export_FullC(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * * x1, uint64_t * x2, uint64_t * x3, void * * x4, uint64_t * x5, _Bool * x6, struct GB_Descriptor_opaque * x7)
{
  return GxB_Matrix_export_FullC(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_export_FullC(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * * x1;
  uint64_t * x2;
  uint64_t * x3;
  void * * x4;
  uint64_t * x5;
  _Bool * x6;
  struct GB_Descriptor_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_export_FullC", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_export_FullC(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_export_FullC _cffi_d_GxB_Matrix_export_FullC
#endif

static GrB_Info _cffi_d_GxB_Matrix_export_FullR(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * * x1, uint64_t * x2, uint64_t * x3, void * * x4, uint64_t * x5, _Bool * x6, struct GB_Descriptor_opaque * x7)
{
  return GxB_Matrix_export_FullR(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_export_FullR(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * * x1;
  uint64_t * x2;
  uint64_t * x3;
  void * * x4;
  uint64_t * x5;
  _Bool * x6;
  struct GB_Descriptor_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_export_FullR", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_export_FullR(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_export_FullR _cffi_d_GxB_Matrix_export_FullR
#endif

static GrB_Info _cffi_d_GxB_Matrix_export_HyperCSC(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * * x1, uint64_t * x2, uint64_t * x3, uint64_t * * x4, uint64_t * * x5, uint64_t * * x6, void * * x7, uint64_t * x8, uint64_t * x9, uint64_t * x10, uint64_t * x11, _Bool * x12, uint64_t * x13, _Bool * x14, struct GB_Descriptor_opaque * x15)
{
  return GxB_Matrix_export_HyperCSC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12, x13, x14, x15);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_export_HyperCSC(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * * x1;
  uint64_t * x2;
  uint64_t * x3;
  uint64_t * * x4;
  uint64_t * * x5;
  uint64_t * * x6;
  void * * x7;
  uint64_t * x8;
  uint64_t * x9;
  uint64_t * x10;
  uint64_t * x11;
  _Bool * x12;
  uint64_t * x13;
  _Bool * x14;
  struct GB_Descriptor_opaque * x15;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;
  PyObject *arg11;
  PyObject *arg12;
  PyObject *arg13;
  PyObject *arg14;
  PyObject *arg15;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_export_HyperCSC", 16, 16, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10, &arg11, &arg12, &arg13, &arg14, &arg15))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg9, (char **)&x9);
  if (datasize != 0) {
    x9 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg9, (char **)&x9,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg10, (char **)&x10);
  if (datasize != 0) {
    x10 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg10, (char **)&x10,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg11, (char **)&x11);
  if (datasize != 0) {
    x11 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg11, (char **)&x11,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg12, (char **)&x12);
  if (datasize != 0) {
    x12 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg12, (char **)&x12,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg13, (char **)&x13);
  if (datasize != 0) {
    x13 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg13, (char **)&x13,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg14, (char **)&x14);
  if (datasize != 0) {
    x14 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg14, (char **)&x14,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg15, (char **)&x15);
  if (datasize != 0) {
    x15 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg15, (char **)&x15,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_export_HyperCSC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12, x13, x14, x15); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_export_HyperCSC _cffi_d_GxB_Matrix_export_HyperCSC
#endif

static GrB_Info _cffi_d_GxB_Matrix_export_HyperCSR(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * * x1, uint64_t * x2, uint64_t * x3, uint64_t * * x4, uint64_t * * x5, uint64_t * * x6, void * * x7, uint64_t * x8, uint64_t * x9, uint64_t * x10, uint64_t * x11, _Bool * x12, uint64_t * x13, _Bool * x14, struct GB_Descriptor_opaque * x15)
{
  return GxB_Matrix_export_HyperCSR(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12, x13, x14, x15);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_export_HyperCSR(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * * x1;
  uint64_t * x2;
  uint64_t * x3;
  uint64_t * * x4;
  uint64_t * * x5;
  uint64_t * * x6;
  void * * x7;
  uint64_t * x8;
  uint64_t * x9;
  uint64_t * x10;
  uint64_t * x11;
  _Bool * x12;
  uint64_t * x13;
  _Bool * x14;
  struct GB_Descriptor_opaque * x15;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;
  PyObject *arg11;
  PyObject *arg12;
  PyObject *arg13;
  PyObject *arg14;
  PyObject *arg15;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_export_HyperCSR", 16, 16, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10, &arg11, &arg12, &arg13, &arg14, &arg15))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg9, (char **)&x9);
  if (datasize != 0) {
    x9 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg9, (char **)&x9,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg10, (char **)&x10);
  if (datasize != 0) {
    x10 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg10, (char **)&x10,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg11, (char **)&x11);
  if (datasize != 0) {
    x11 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg11, (char **)&x11,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg12, (char **)&x12);
  if (datasize != 0) {
    x12 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg12, (char **)&x12,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg13, (char **)&x13);
  if (datasize != 0) {
    x13 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg13, (char **)&x13,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg14, (char **)&x14);
  if (datasize != 0) {
    x14 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg14, (char **)&x14,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg15, (char **)&x15);
  if (datasize != 0) {
    x15 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg15, (char **)&x15,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_export_HyperCSR(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12, x13, x14, x15); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_export_HyperCSR _cffi_d_GxB_Matrix_export_HyperCSR
#endif

static GrB_Info _cffi_d_GxB_Matrix_extractElement_FC32(_Fcomplex * x0, struct GB_Matrix_opaque * x1, uint64_t x2, uint64_t x3)
{
  return GxB_Matrix_extractElement_FC32(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_extractElement_FC32(PyObject *self, PyObject *args)
{
  _Fcomplex * x0;
  struct GB_Matrix_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_extractElement_FC32", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(261), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (_Fcomplex *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(261), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_extractElement_FC32(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_extractElement_FC32 _cffi_d_GxB_Matrix_extractElement_FC32
#endif

static GrB_Info _cffi_d_GxB_Matrix_extractElement_FC64(_Dcomplex * x0, struct GB_Matrix_opaque * x1, uint64_t x2, uint64_t x3)
{
  return GxB_Matrix_extractElement_FC64(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_extractElement_FC64(PyObject *self, PyObject *args)
{
  _Dcomplex * x0;
  struct GB_Matrix_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_extractElement_FC64", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(203), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (_Dcomplex *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(203), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_extractElement_FC64(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_extractElement_FC64 _cffi_d_GxB_Matrix_extractElement_FC64
#endif

static GrB_Info _cffi_d_GxB_Matrix_extractTuples_FC32(uint64_t * x0, uint64_t * x1, _Fcomplex * x2, uint64_t * x3, struct GB_Matrix_opaque * x4)
{
  return GxB_Matrix_extractTuples_FC32(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_extractTuples_FC32(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  _Fcomplex * x2;
  uint64_t * x3;
  struct GB_Matrix_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_extractTuples_FC32", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(261), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (_Fcomplex *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(261), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_extractTuples_FC32(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_extractTuples_FC32 _cffi_d_GxB_Matrix_extractTuples_FC32
#endif

static GrB_Info _cffi_d_GxB_Matrix_extractTuples_FC64(uint64_t * x0, uint64_t * x1, _Dcomplex * x2, uint64_t * x3, struct GB_Matrix_opaque * x4)
{
  return GxB_Matrix_extractTuples_FC64(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_extractTuples_FC64(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  uint64_t * x1;
  _Dcomplex * x2;
  uint64_t * x3;
  struct GB_Matrix_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_extractTuples_FC64", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(203), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (_Dcomplex *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(203), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_extractTuples_FC64(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_extractTuples_FC64 _cffi_d_GxB_Matrix_extractTuples_FC64
#endif

static GrB_Info _cffi_d_GxB_Matrix_fprint(struct GB_Matrix_opaque * x0, char const * x1, GxB_Print_Level x2, FILE * x3)
{
  return GxB_Matrix_fprint(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_fprint(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  char const * x1;
  GxB_Print_Level x2;
  FILE * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_fprint", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x2, _cffi_type(463), arg2) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(464), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (FILE *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(464), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_fprint(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_fprint _cffi_d_GxB_Matrix_fprint
#endif

static GrB_Info _cffi_d_GxB_Matrix_import_BitmapC(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t x3, int8_t * * x4, void * * x5, uint64_t x6, uint64_t x7, _Bool x8, uint64_t x9, struct GB_Descriptor_opaque * x10)
{
  return GxB_Matrix_import_BitmapC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_import_BitmapC(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  int8_t * * x4;
  void * * x5;
  uint64_t x6;
  uint64_t x7;
  _Bool x8;
  uint64_t x9;
  struct GB_Descriptor_opaque * x10;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_import_BitmapC", 11, 11, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(588), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (int8_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(588), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x6 = _cffi_to_c_int(arg6, uint64_t);
  if (x6 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x8 = (_Bool)_cffi_to_c__Bool(arg8);
  if (x8 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  x9 = _cffi_to_c_int(arg9, uint64_t);
  if (x9 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg10, (char **)&x10);
  if (datasize != 0) {
    x10 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg10, (char **)&x10,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_import_BitmapC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_import_BitmapC _cffi_d_GxB_Matrix_import_BitmapC
#endif

static GrB_Info _cffi_d_GxB_Matrix_import_BitmapR(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t x3, int8_t * * x4, void * * x5, uint64_t x6, uint64_t x7, _Bool x8, uint64_t x9, struct GB_Descriptor_opaque * x10)
{
  return GxB_Matrix_import_BitmapR(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_import_BitmapR(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  int8_t * * x4;
  void * * x5;
  uint64_t x6;
  uint64_t x7;
  _Bool x8;
  uint64_t x9;
  struct GB_Descriptor_opaque * x10;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_import_BitmapR", 11, 11, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(588), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (int8_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(588), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x6 = _cffi_to_c_int(arg6, uint64_t);
  if (x6 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x8 = (_Bool)_cffi_to_c__Bool(arg8);
  if (x8 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  x9 = _cffi_to_c_int(arg9, uint64_t);
  if (x9 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg10, (char **)&x10);
  if (datasize != 0) {
    x10 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg10, (char **)&x10,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_import_BitmapR(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_import_BitmapR _cffi_d_GxB_Matrix_import_BitmapR
#endif

static GrB_Info _cffi_d_GxB_Matrix_import_CSC(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t x3, uint64_t * * x4, uint64_t * * x5, void * * x6, uint64_t x7, uint64_t x8, uint64_t x9, _Bool x10, _Bool x11, struct GB_Descriptor_opaque * x12)
{
  return GxB_Matrix_import_CSC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_import_CSC(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  uint64_t * * x4;
  uint64_t * * x5;
  void * * x6;
  uint64_t x7;
  uint64_t x8;
  uint64_t x9;
  _Bool x10;
  _Bool x11;
  struct GB_Descriptor_opaque * x12;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;
  PyObject *arg11;
  PyObject *arg12;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_import_CSC", 13, 13, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10, &arg11, &arg12))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x8 = _cffi_to_c_int(arg8, uint64_t);
  if (x8 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x9 = _cffi_to_c_int(arg9, uint64_t);
  if (x9 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x10 = (_Bool)_cffi_to_c__Bool(arg10);
  if (x10 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  x11 = (_Bool)_cffi_to_c__Bool(arg11);
  if (x11 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg12, (char **)&x12);
  if (datasize != 0) {
    x12 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg12, (char **)&x12,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_import_CSC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_import_CSC _cffi_d_GxB_Matrix_import_CSC
#endif

static GrB_Info _cffi_d_GxB_Matrix_import_CSR(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t x3, uint64_t * * x4, uint64_t * * x5, void * * x6, uint64_t x7, uint64_t x8, uint64_t x9, _Bool x10, _Bool x11, struct GB_Descriptor_opaque * x12)
{
  return GxB_Matrix_import_CSR(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_import_CSR(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  uint64_t * * x4;
  uint64_t * * x5;
  void * * x6;
  uint64_t x7;
  uint64_t x8;
  uint64_t x9;
  _Bool x10;
  _Bool x11;
  struct GB_Descriptor_opaque * x12;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;
  PyObject *arg11;
  PyObject *arg12;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_import_CSR", 13, 13, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10, &arg11, &arg12))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x8 = _cffi_to_c_int(arg8, uint64_t);
  if (x8 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x9 = _cffi_to_c_int(arg9, uint64_t);
  if (x9 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x10 = (_Bool)_cffi_to_c__Bool(arg10);
  if (x10 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  x11 = (_Bool)_cffi_to_c__Bool(arg11);
  if (x11 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg12, (char **)&x12);
  if (datasize != 0) {
    x12 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg12, (char **)&x12,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_import_CSR(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_import_CSR _cffi_d_GxB_Matrix_import_CSR
#endif

static GrB_Info _cffi_d_GxB_Matrix_import_FC32(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t x3, uint64_t const * x4, uint64_t const * x5, _Fcomplex const * x6, uint64_t x7, uint64_t x8, uint64_t x9, GrB_Format x10)
{
  return GxB_Matrix_import_FC32(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_import_FC32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  uint64_t const * x4;
  uint64_t const * x5;
  _Fcomplex const * x6;
  uint64_t x7;
  uint64_t x8;
  uint64_t x9;
  GrB_Format x10;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_import_FC32", 11, 11, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(737), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (_Fcomplex const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(737), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x8 = _cffi_to_c_int(arg8, uint64_t);
  if (x8 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x9 = _cffi_to_c_int(arg9, uint64_t);
  if (x9 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  if (_cffi_to_c((char *)&x10, _cffi_type(702), arg10) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_import_FC32(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_import_FC32 _cffi_d_GxB_Matrix_import_FC32
#endif

static GrB_Info _cffi_d_GxB_Matrix_import_FC64(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t x3, uint64_t const * x4, uint64_t const * x5, _Dcomplex const * x6, uint64_t x7, uint64_t x8, uint64_t x9, GrB_Format x10)
{
  return GxB_Matrix_import_FC64(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_import_FC64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  uint64_t const * x4;
  uint64_t const * x5;
  _Dcomplex const * x6;
  uint64_t x7;
  uint64_t x8;
  uint64_t x9;
  GrB_Format x10;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_import_FC64", 11, 11, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(711), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (_Dcomplex const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(711), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x8 = _cffi_to_c_int(arg8, uint64_t);
  if (x8 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x9 = _cffi_to_c_int(arg9, uint64_t);
  if (x9 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  if (_cffi_to_c((char *)&x10, _cffi_type(702), arg10) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_import_FC64(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_import_FC64 _cffi_d_GxB_Matrix_import_FC64
#endif

static GrB_Info _cffi_d_GxB_Matrix_import_FullC(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t x3, void * * x4, uint64_t x5, _Bool x6, struct GB_Descriptor_opaque * x7)
{
  return GxB_Matrix_import_FullC(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_import_FullC(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  void * * x4;
  uint64_t x5;
  _Bool x6;
  struct GB_Descriptor_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_import_FullC", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x6 = (_Bool)_cffi_to_c__Bool(arg6);
  if (x6 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_import_FullC(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_import_FullC _cffi_d_GxB_Matrix_import_FullC
#endif

static GrB_Info _cffi_d_GxB_Matrix_import_FullR(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t x3, void * * x4, uint64_t x5, _Bool x6, struct GB_Descriptor_opaque * x7)
{
  return GxB_Matrix_import_FullR(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_import_FullR(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  void * * x4;
  uint64_t x5;
  _Bool x6;
  struct GB_Descriptor_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_import_FullR", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x6 = (_Bool)_cffi_to_c__Bool(arg6);
  if (x6 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_import_FullR(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_import_FullR _cffi_d_GxB_Matrix_import_FullR
#endif

static GrB_Info _cffi_d_GxB_Matrix_import_HyperCSC(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t x3, uint64_t * * x4, uint64_t * * x5, uint64_t * * x6, void * * x7, uint64_t x8, uint64_t x9, uint64_t x10, uint64_t x11, _Bool x12, uint64_t x13, _Bool x14, struct GB_Descriptor_opaque * x15)
{
  return GxB_Matrix_import_HyperCSC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12, x13, x14, x15);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_import_HyperCSC(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  uint64_t * * x4;
  uint64_t * * x5;
  uint64_t * * x6;
  void * * x7;
  uint64_t x8;
  uint64_t x9;
  uint64_t x10;
  uint64_t x11;
  _Bool x12;
  uint64_t x13;
  _Bool x14;
  struct GB_Descriptor_opaque * x15;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;
  PyObject *arg11;
  PyObject *arg12;
  PyObject *arg13;
  PyObject *arg14;
  PyObject *arg15;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_import_HyperCSC", 16, 16, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10, &arg11, &arg12, &arg13, &arg14, &arg15))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x8 = _cffi_to_c_int(arg8, uint64_t);
  if (x8 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x9 = _cffi_to_c_int(arg9, uint64_t);
  if (x9 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x10 = _cffi_to_c_int(arg10, uint64_t);
  if (x10 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x11 = _cffi_to_c_int(arg11, uint64_t);
  if (x11 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x12 = (_Bool)_cffi_to_c__Bool(arg12);
  if (x12 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  x13 = _cffi_to_c_int(arg13, uint64_t);
  if (x13 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x14 = (_Bool)_cffi_to_c__Bool(arg14);
  if (x14 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg15, (char **)&x15);
  if (datasize != 0) {
    x15 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg15, (char **)&x15,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_import_HyperCSC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12, x13, x14, x15); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_import_HyperCSC _cffi_d_GxB_Matrix_import_HyperCSC
#endif

static GrB_Info _cffi_d_GxB_Matrix_import_HyperCSR(struct GB_Matrix_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t x3, uint64_t * * x4, uint64_t * * x5, uint64_t * * x6, void * * x7, uint64_t x8, uint64_t x9, uint64_t x10, uint64_t x11, _Bool x12, uint64_t x13, _Bool x14, struct GB_Descriptor_opaque * x15)
{
  return GxB_Matrix_import_HyperCSR(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12, x13, x14, x15);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_import_HyperCSR(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  uint64_t * * x4;
  uint64_t * * x5;
  uint64_t * * x6;
  void * * x7;
  uint64_t x8;
  uint64_t x9;
  uint64_t x10;
  uint64_t x11;
  _Bool x12;
  uint64_t x13;
  _Bool x14;
  struct GB_Descriptor_opaque * x15;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;
  PyObject *arg11;
  PyObject *arg12;
  PyObject *arg13;
  PyObject *arg14;
  PyObject *arg15;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_import_HyperCSR", 16, 16, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10, &arg11, &arg12, &arg13, &arg14, &arg15))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x8 = _cffi_to_c_int(arg8, uint64_t);
  if (x8 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x9 = _cffi_to_c_int(arg9, uint64_t);
  if (x9 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x10 = _cffi_to_c_int(arg10, uint64_t);
  if (x10 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x11 = _cffi_to_c_int(arg11, uint64_t);
  if (x11 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x12 = (_Bool)_cffi_to_c__Bool(arg12);
  if (x12 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  x13 = _cffi_to_c_int(arg13, uint64_t);
  if (x13 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x14 = (_Bool)_cffi_to_c__Bool(arg14);
  if (x14 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg15, (char **)&x15);
  if (datasize != 0) {
    x15 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg15, (char **)&x15,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_import_HyperCSR(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12, x13, x14, x15); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_import_HyperCSR _cffi_d_GxB_Matrix_import_HyperCSR
#endif

static GrB_Info _cffi_d_GxB_Matrix_isStoredElement(struct GB_Matrix_opaque * x0, uint64_t x1, uint64_t x2)
{
  return GxB_Matrix_isStoredElement(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_isStoredElement(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_isStoredElement", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint64_t);
  if (x1 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_isStoredElement(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_isStoredElement _cffi_d_GxB_Matrix_isStoredElement
#endif

static GrB_Info _cffi_d_GxB_Matrix_iso(_Bool * x0, struct GB_Matrix_opaque * x1)
{
  return GxB_Matrix_iso(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_iso(PyObject *self, PyObject *args)
{
  _Bool * x0;
  struct GB_Matrix_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_iso", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_iso(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_iso _cffi_d_GxB_Matrix_iso
#endif

static GrB_Info _cffi_d_GxB_Matrix_memoryUsage(size_t * x0, struct GB_Matrix_opaque * x1)
{
  return GxB_Matrix_memoryUsage(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_memoryUsage(PyObject *self, PyObject *args)
{
  size_t * x0;
  struct GB_Matrix_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_memoryUsage", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(406), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (size_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(406), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_memoryUsage(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_memoryUsage _cffi_d_GxB_Matrix_memoryUsage
#endif

static GrB_Info _cffi_d_GxB_Matrix_pack_BitmapC(struct GB_Matrix_opaque * x0, int8_t * * x1, void * * x2, uint64_t x3, uint64_t x4, _Bool x5, uint64_t x6, struct GB_Descriptor_opaque * x7)
{
  return GxB_Matrix_pack_BitmapC(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_pack_BitmapC(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  int8_t * * x1;
  void * * x2;
  uint64_t x3;
  uint64_t x4;
  _Bool x5;
  uint64_t x6;
  struct GB_Descriptor_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_pack_BitmapC", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(588), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (int8_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(588), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x5 = (_Bool)_cffi_to_c__Bool(arg5);
  if (x5 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  x6 = _cffi_to_c_int(arg6, uint64_t);
  if (x6 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_pack_BitmapC(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_pack_BitmapC _cffi_d_GxB_Matrix_pack_BitmapC
#endif

static GrB_Info _cffi_d_GxB_Matrix_pack_BitmapR(struct GB_Matrix_opaque * x0, int8_t * * x1, void * * x2, uint64_t x3, uint64_t x4, _Bool x5, uint64_t x6, struct GB_Descriptor_opaque * x7)
{
  return GxB_Matrix_pack_BitmapR(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_pack_BitmapR(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  int8_t * * x1;
  void * * x2;
  uint64_t x3;
  uint64_t x4;
  _Bool x5;
  uint64_t x6;
  struct GB_Descriptor_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_pack_BitmapR", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(588), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (int8_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(588), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x5 = (_Bool)_cffi_to_c__Bool(arg5);
  if (x5 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  x6 = _cffi_to_c_int(arg6, uint64_t);
  if (x6 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_pack_BitmapR(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_pack_BitmapR _cffi_d_GxB_Matrix_pack_BitmapR
#endif

static GrB_Info _cffi_d_GxB_Matrix_pack_CSC(struct GB_Matrix_opaque * x0, uint64_t * * x1, uint64_t * * x2, void * * x3, uint64_t x4, uint64_t x5, uint64_t x6, _Bool x7, _Bool x8, struct GB_Descriptor_opaque * x9)
{
  return GxB_Matrix_pack_CSC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_pack_CSC(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t * * x1;
  uint64_t * * x2;
  void * * x3;
  uint64_t x4;
  uint64_t x5;
  uint64_t x6;
  _Bool x7;
  _Bool x8;
  struct GB_Descriptor_opaque * x9;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_pack_CSC", 10, 10, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x6 = _cffi_to_c_int(arg6, uint64_t);
  if (x6 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x7 = (_Bool)_cffi_to_c__Bool(arg7);
  if (x7 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  x8 = (_Bool)_cffi_to_c__Bool(arg8);
  if (x8 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg9, (char **)&x9);
  if (datasize != 0) {
    x9 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg9, (char **)&x9,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_pack_CSC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_pack_CSC _cffi_d_GxB_Matrix_pack_CSC
#endif

static GrB_Info _cffi_d_GxB_Matrix_pack_CSR(struct GB_Matrix_opaque * x0, uint64_t * * x1, uint64_t * * x2, void * * x3, uint64_t x4, uint64_t x5, uint64_t x6, _Bool x7, _Bool x8, struct GB_Descriptor_opaque * x9)
{
  return GxB_Matrix_pack_CSR(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_pack_CSR(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t * * x1;
  uint64_t * * x2;
  void * * x3;
  uint64_t x4;
  uint64_t x5;
  uint64_t x6;
  _Bool x7;
  _Bool x8;
  struct GB_Descriptor_opaque * x9;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_pack_CSR", 10, 10, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x6 = _cffi_to_c_int(arg6, uint64_t);
  if (x6 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x7 = (_Bool)_cffi_to_c__Bool(arg7);
  if (x7 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  x8 = (_Bool)_cffi_to_c__Bool(arg8);
  if (x8 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg9, (char **)&x9);
  if (datasize != 0) {
    x9 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg9, (char **)&x9,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_pack_CSR(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_pack_CSR _cffi_d_GxB_Matrix_pack_CSR
#endif

static GrB_Info _cffi_d_GxB_Matrix_pack_FullC(struct GB_Matrix_opaque * x0, void * * x1, uint64_t x2, _Bool x3, struct GB_Descriptor_opaque * x4)
{
  return GxB_Matrix_pack_FullC(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_pack_FullC(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  void * * x1;
  uint64_t x2;
  _Bool x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_pack_FullC", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = (_Bool)_cffi_to_c__Bool(arg3);
  if (x3 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_pack_FullC(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_pack_FullC _cffi_d_GxB_Matrix_pack_FullC
#endif

static GrB_Info _cffi_d_GxB_Matrix_pack_FullR(struct GB_Matrix_opaque * x0, void * * x1, uint64_t x2, _Bool x3, struct GB_Descriptor_opaque * x4)
{
  return GxB_Matrix_pack_FullR(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_pack_FullR(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  void * * x1;
  uint64_t x2;
  _Bool x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_pack_FullR", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = (_Bool)_cffi_to_c__Bool(arg3);
  if (x3 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_pack_FullR(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_pack_FullR _cffi_d_GxB_Matrix_pack_FullR
#endif

static GrB_Info _cffi_d_GxB_Matrix_pack_HyperCSC(struct GB_Matrix_opaque * x0, uint64_t * * x1, uint64_t * * x2, uint64_t * * x3, void * * x4, uint64_t x5, uint64_t x6, uint64_t x7, uint64_t x8, _Bool x9, uint64_t x10, _Bool x11, struct GB_Descriptor_opaque * x12)
{
  return GxB_Matrix_pack_HyperCSC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_pack_HyperCSC(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t * * x1;
  uint64_t * * x2;
  uint64_t * * x3;
  void * * x4;
  uint64_t x5;
  uint64_t x6;
  uint64_t x7;
  uint64_t x8;
  _Bool x9;
  uint64_t x10;
  _Bool x11;
  struct GB_Descriptor_opaque * x12;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;
  PyObject *arg11;
  PyObject *arg12;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_pack_HyperCSC", 13, 13, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10, &arg11, &arg12))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x6 = _cffi_to_c_int(arg6, uint64_t);
  if (x6 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x8 = _cffi_to_c_int(arg8, uint64_t);
  if (x8 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x9 = (_Bool)_cffi_to_c__Bool(arg9);
  if (x9 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  x10 = _cffi_to_c_int(arg10, uint64_t);
  if (x10 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x11 = (_Bool)_cffi_to_c__Bool(arg11);
  if (x11 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg12, (char **)&x12);
  if (datasize != 0) {
    x12 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg12, (char **)&x12,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_pack_HyperCSC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_pack_HyperCSC _cffi_d_GxB_Matrix_pack_HyperCSC
#endif

static GrB_Info _cffi_d_GxB_Matrix_pack_HyperCSR(struct GB_Matrix_opaque * x0, uint64_t * * x1, uint64_t * * x2, uint64_t * * x3, void * * x4, uint64_t x5, uint64_t x6, uint64_t x7, uint64_t x8, _Bool x9, uint64_t x10, _Bool x11, struct GB_Descriptor_opaque * x12)
{
  return GxB_Matrix_pack_HyperCSR(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_pack_HyperCSR(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t * * x1;
  uint64_t * * x2;
  uint64_t * * x3;
  void * * x4;
  uint64_t x5;
  uint64_t x6;
  uint64_t x7;
  uint64_t x8;
  _Bool x9;
  uint64_t x10;
  _Bool x11;
  struct GB_Descriptor_opaque * x12;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;
  PyObject *arg11;
  PyObject *arg12;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_pack_HyperCSR", 13, 13, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10, &arg11, &arg12))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x6 = _cffi_to_c_int(arg6, uint64_t);
  if (x6 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x8 = _cffi_to_c_int(arg8, uint64_t);
  if (x8 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x9 = (_Bool)_cffi_to_c__Bool(arg9);
  if (x9 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  x10 = _cffi_to_c_int(arg10, uint64_t);
  if (x10 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x11 = (_Bool)_cffi_to_c__Bool(arg11);
  if (x11 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg12, (char **)&x12);
  if (datasize != 0) {
    x12 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg12, (char **)&x12,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_pack_HyperCSR(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_pack_HyperCSR _cffi_d_GxB_Matrix_pack_HyperCSR
#endif

static GrB_Info _cffi_d_GxB_Matrix_reduce_FC32(_Fcomplex * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Matrix_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GxB_Matrix_reduce_FC32(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_reduce_FC32(PyObject *self, PyObject *args)
{
  _Fcomplex * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Matrix_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_reduce_FC32", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(261), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (_Fcomplex *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(261), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_reduce_FC32(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_reduce_FC32 _cffi_d_GxB_Matrix_reduce_FC32
#endif

static GrB_Info _cffi_d_GxB_Matrix_reduce_FC64(_Dcomplex * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Matrix_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GxB_Matrix_reduce_FC64(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_reduce_FC64(PyObject *self, PyObject *args)
{
  _Dcomplex * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Matrix_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_reduce_FC64", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(203), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (_Dcomplex *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(203), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_reduce_FC64(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_reduce_FC64 _cffi_d_GxB_Matrix_reduce_FC64
#endif

static GrB_Info _cffi_d_GxB_Matrix_reshape(struct GB_Matrix_opaque * x0, _Bool x1, uint64_t x2, uint64_t x3, struct GB_Descriptor_opaque * x4)
{
  return GxB_Matrix_reshape(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_reshape(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  _Bool x1;
  uint64_t x2;
  uint64_t x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_reshape", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = (_Bool)_cffi_to_c__Bool(arg1);
  if (x1 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_reshape(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_reshape _cffi_d_GxB_Matrix_reshape
#endif

static GrB_Info _cffi_d_GxB_Matrix_reshapeDup(struct GB_Matrix_opaque * * x0, struct GB_Matrix_opaque * x1, _Bool x2, uint64_t x3, uint64_t x4, struct GB_Descriptor_opaque * x5)
{
  return GxB_Matrix_reshapeDup(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_reshapeDup(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  struct GB_Matrix_opaque * x1;
  _Bool x2;
  uint64_t x3;
  uint64_t x4;
  struct GB_Descriptor_opaque * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_reshapeDup", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = (_Bool)_cffi_to_c__Bool(arg2);
  if (x2 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_reshapeDup(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_reshapeDup _cffi_d_GxB_Matrix_reshapeDup
#endif

static GrB_Info _cffi_d_GxB_Matrix_select(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_SelectOp_opaque * x3, struct GB_Matrix_opaque * x4, struct GB_Scalar_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Matrix_select(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_select(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_SelectOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  struct GB_Scalar_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_select", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(155), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_SelectOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(155), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_select(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_select _cffi_d_GxB_Matrix_select
#endif

static GrB_Info _cffi_d_GxB_Matrix_select_FC32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, _Fcomplex x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Matrix_select_FC32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_select_FC32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  _Fcomplex x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_select_FC32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x5, _cffi_type(974), arg5) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_select_FC32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Matrix_select_FC32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, _Fcomplex *x5, struct GB_Descriptor_opaque * x6)
{
  GrB_Info result;
  { result = GxB_Matrix_select_FC32(x0, x1, x2, x3, x4, *x5, x6); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Matrix_select_FC64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, _Dcomplex x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Matrix_select_FC64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_select_FC64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Matrix_opaque * x4;
  _Dcomplex x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_select_FC64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x5, _cffi_type(962), arg5) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_select_FC64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Matrix_select_FC64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Matrix_opaque * x4, _Dcomplex *x5, struct GB_Descriptor_opaque * x6)
{
  GrB_Info result;
  { result = GxB_Matrix_select_FC64(x0, x1, x2, x3, x4, *x5, x6); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Matrix_serialize(void * * x0, uint64_t * x1, struct GB_Matrix_opaque * x2, struct GB_Descriptor_opaque * x3)
{
  return GxB_Matrix_serialize(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_serialize(PyObject *self, PyObject *args)
{
  void * * x0;
  uint64_t * x1;
  struct GB_Matrix_opaque * x2;
  struct GB_Descriptor_opaque * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_serialize", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_serialize(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_serialize _cffi_d_GxB_Matrix_serialize
#endif

static GrB_Info _cffi_d_GxB_Matrix_setElement_FC32(struct GB_Matrix_opaque * x0, _Fcomplex x1, uint64_t x2, uint64_t x3)
{
  return GxB_Matrix_setElement_FC32(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_setElement_FC32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  _Fcomplex x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_setElement_FC32", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(974), arg1) < 0)
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_setElement_FC32(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Matrix_setElement_FC32(struct GB_Matrix_opaque * x0, _Fcomplex *x1, uint64_t x2, uint64_t x3)
{
  GrB_Info result;
  { result = GxB_Matrix_setElement_FC32(x0, *x1, x2, x3); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Matrix_setElement_FC64(struct GB_Matrix_opaque * x0, _Dcomplex x1, uint64_t x2, uint64_t x3)
{
  return GxB_Matrix_setElement_FC64(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_setElement_FC64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  _Dcomplex x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_setElement_FC64", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(962), arg1) < 0)
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_setElement_FC64(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Matrix_setElement_FC64(struct GB_Matrix_opaque * x0, _Dcomplex *x1, uint64_t x2, uint64_t x3)
{
  GrB_Info result;
  { result = GxB_Matrix_setElement_FC64(x0, *x1, x2, x3); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Matrix_sort(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Matrix_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GxB_Matrix_sort(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_sort(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Matrix_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_sort", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_sort(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_sort _cffi_d_GxB_Matrix_sort
#endif

static GrB_Info _cffi_d_GxB_Matrix_split(struct GB_Matrix_opaque * * x0, uint64_t x1, uint64_t x2, uint64_t const * x3, uint64_t const * x4, struct GB_Matrix_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Matrix_split(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_split(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * * x0;
  uint64_t x1;
  uint64_t x2;
  uint64_t const * x3;
  uint64_t const * x4;
  struct GB_Matrix_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_split", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint64_t);
  if (x1 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_split(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_split _cffi_d_GxB_Matrix_split
#endif

static GrB_Info _cffi_d_GxB_Matrix_subassign(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Matrix_opaque * x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GxB_Matrix_subassign(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_subassign(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Matrix_opaque * x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_subassign", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_subassign(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_subassign _cffi_d_GxB_Matrix_subassign
#endif

static GrB_Info _cffi_d_GxB_Matrix_subassign_BOOL(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, _Bool x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GxB_Matrix_subassign_BOOL(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_subassign_BOOL(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  _Bool x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_subassign_BOOL", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = (_Bool)_cffi_to_c__Bool(arg3);
  if (x3 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_subassign_BOOL(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_subassign_BOOL _cffi_d_GxB_Matrix_subassign_BOOL
#endif

static GrB_Info _cffi_d_GxB_Matrix_subassign_FC32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, _Fcomplex x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GxB_Matrix_subassign_FC32(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_subassign_FC32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  _Fcomplex x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_subassign_FC32", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x3, _cffi_type(974), arg3) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_subassign_FC32(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Matrix_subassign_FC32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, _Fcomplex *x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  GrB_Info result;
  { result = GxB_Matrix_subassign_FC32(x0, x1, x2, *x3, x4, x5, x6, x7, x8); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Matrix_subassign_FC64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, _Dcomplex x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GxB_Matrix_subassign_FC64(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_subassign_FC64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  _Dcomplex x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_subassign_FC64", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x3, _cffi_type(962), arg3) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_subassign_FC64(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Matrix_subassign_FC64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, _Dcomplex *x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  GrB_Info result;
  { result = GxB_Matrix_subassign_FC64(x0, x1, x2, *x3, x4, x5, x6, x7, x8); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Matrix_subassign_FP32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, float x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GxB_Matrix_subassign_FP32(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_subassign_FP32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  float x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_subassign_FP32", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = (float)_cffi_to_c_float(arg3);
  if (x3 == (float)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_subassign_FP32(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_subassign_FP32 _cffi_d_GxB_Matrix_subassign_FP32
#endif

static GrB_Info _cffi_d_GxB_Matrix_subassign_FP64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, double x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GxB_Matrix_subassign_FP64(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_subassign_FP64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  double x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_subassign_FP64", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = (double)_cffi_to_c_double(arg3);
  if (x3 == (double)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_subassign_FP64(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_subassign_FP64 _cffi_d_GxB_Matrix_subassign_FP64
#endif

static GrB_Info _cffi_d_GxB_Matrix_subassign_INT16(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, int16_t x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GxB_Matrix_subassign_INT16(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_subassign_INT16(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  int16_t x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_subassign_INT16", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, int16_t);
  if (x3 == (int16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_subassign_INT16(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_subassign_INT16 _cffi_d_GxB_Matrix_subassign_INT16
#endif

static GrB_Info _cffi_d_GxB_Matrix_subassign_INT32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, int32_t x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GxB_Matrix_subassign_INT32(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_subassign_INT32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  int32_t x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_subassign_INT32", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, int32_t);
  if (x3 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_subassign_INT32(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_subassign_INT32 _cffi_d_GxB_Matrix_subassign_INT32
#endif

static GrB_Info _cffi_d_GxB_Matrix_subassign_INT64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, int64_t x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GxB_Matrix_subassign_INT64(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_subassign_INT64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  int64_t x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_subassign_INT64", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, int64_t);
  if (x3 == (int64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_subassign_INT64(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_subassign_INT64 _cffi_d_GxB_Matrix_subassign_INT64
#endif

static GrB_Info _cffi_d_GxB_Matrix_subassign_INT8(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, int8_t x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GxB_Matrix_subassign_INT8(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_subassign_INT8(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  int8_t x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_subassign_INT8", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, int8_t);
  if (x3 == (int8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_subassign_INT8(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_subassign_INT8 _cffi_d_GxB_Matrix_subassign_INT8
#endif

static GrB_Info _cffi_d_GxB_Matrix_subassign_Scalar(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Scalar_opaque * x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GxB_Matrix_subassign_Scalar(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_subassign_Scalar(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Scalar_opaque * x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_subassign_Scalar", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_subassign_Scalar(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_subassign_Scalar _cffi_d_GxB_Matrix_subassign_Scalar
#endif

static GrB_Info _cffi_d_GxB_Matrix_subassign_UDT(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, void * x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GxB_Matrix_subassign_UDT(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_subassign_UDT(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  void * x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_subassign_UDT", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_subassign_UDT(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_subassign_UDT _cffi_d_GxB_Matrix_subassign_UDT
#endif

static GrB_Info _cffi_d_GxB_Matrix_subassign_UINT16(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, uint16_t x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GxB_Matrix_subassign_UINT16(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_subassign_UINT16(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  uint16_t x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_subassign_UINT16", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint16_t);
  if (x3 == (uint16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_subassign_UINT16(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_subassign_UINT16 _cffi_d_GxB_Matrix_subassign_UINT16
#endif

static GrB_Info _cffi_d_GxB_Matrix_subassign_UINT32(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, uint32_t x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GxB_Matrix_subassign_UINT32(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_subassign_UINT32(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  uint32_t x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_subassign_UINT32", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint32_t);
  if (x3 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_subassign_UINT32(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_subassign_UINT32 _cffi_d_GxB_Matrix_subassign_UINT32
#endif

static GrB_Info _cffi_d_GxB_Matrix_subassign_UINT64(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, uint64_t x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GxB_Matrix_subassign_UINT64(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_subassign_UINT64(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  uint64_t x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_subassign_UINT64", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_subassign_UINT64(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_subassign_UINT64 _cffi_d_GxB_Matrix_subassign_UINT64
#endif

static GrB_Info _cffi_d_GxB_Matrix_subassign_UINT8(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_BinaryOp_opaque * x2, uint8_t x3, uint64_t const * x4, uint64_t x5, uint64_t const * x6, uint64_t x7, struct GB_Descriptor_opaque * x8)
{
  return GxB_Matrix_subassign_UINT8(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_subassign_UINT8(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  uint8_t x3;
  uint64_t const * x4;
  uint64_t x5;
  uint64_t const * x6;
  uint64_t x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_subassign_UINT8", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint8_t);
  if (x3 == (uint8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x7 = _cffi_to_c_int(arg7, uint64_t);
  if (x7 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_subassign_UINT8(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_subassign_UINT8 _cffi_d_GxB_Matrix_subassign_UINT8
#endif

static GrB_Info _cffi_d_GxB_Matrix_type(struct GB_Type_opaque * * x0, struct GB_Matrix_opaque * x1)
{
  return GxB_Matrix_type(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_type(PyObject *self, PyObject *args)
{
  struct GB_Type_opaque * * x0;
  struct GB_Matrix_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_type", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_type(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_type _cffi_d_GxB_Matrix_type
#endif

static GrB_Info _cffi_d_GxB_Matrix_type_name(char * x0, struct GB_Matrix_opaque * x1)
{
  return GxB_Matrix_type_name(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_type_name(PyObject *self, PyObject *args)
{
  char * x0;
  struct GB_Matrix_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_type_name", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(97), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(97), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_type_name(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_type_name _cffi_d_GxB_Matrix_type_name
#endif

static GrB_Info _cffi_d_GxB_Matrix_unpack_BitmapC(struct GB_Matrix_opaque * x0, int8_t * * x1, void * * x2, uint64_t * x3, uint64_t * x4, _Bool * x5, uint64_t * x6, struct GB_Descriptor_opaque * x7)
{
  return GxB_Matrix_unpack_BitmapC(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_unpack_BitmapC(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  int8_t * * x1;
  void * * x2;
  uint64_t * x3;
  uint64_t * x4;
  _Bool * x5;
  uint64_t * x6;
  struct GB_Descriptor_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_unpack_BitmapC", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(588), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (int8_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(588), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_unpack_BitmapC(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_unpack_BitmapC _cffi_d_GxB_Matrix_unpack_BitmapC
#endif

static GrB_Info _cffi_d_GxB_Matrix_unpack_BitmapR(struct GB_Matrix_opaque * x0, int8_t * * x1, void * * x2, uint64_t * x3, uint64_t * x4, _Bool * x5, uint64_t * x6, struct GB_Descriptor_opaque * x7)
{
  return GxB_Matrix_unpack_BitmapR(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_unpack_BitmapR(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  int8_t * * x1;
  void * * x2;
  uint64_t * x3;
  uint64_t * x4;
  _Bool * x5;
  uint64_t * x6;
  struct GB_Descriptor_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_unpack_BitmapR", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(588), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (int8_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(588), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_unpack_BitmapR(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_unpack_BitmapR _cffi_d_GxB_Matrix_unpack_BitmapR
#endif

static GrB_Info _cffi_d_GxB_Matrix_unpack_CSC(struct GB_Matrix_opaque * x0, uint64_t * * x1, uint64_t * * x2, void * * x3, uint64_t * x4, uint64_t * x5, uint64_t * x6, _Bool * x7, _Bool * x8, struct GB_Descriptor_opaque * x9)
{
  return GxB_Matrix_unpack_CSC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_unpack_CSC(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t * * x1;
  uint64_t * * x2;
  void * * x3;
  uint64_t * x4;
  uint64_t * x5;
  uint64_t * x6;
  _Bool * x7;
  _Bool * x8;
  struct GB_Descriptor_opaque * x9;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_unpack_CSC", 10, 10, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg9, (char **)&x9);
  if (datasize != 0) {
    x9 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg9, (char **)&x9,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_unpack_CSC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_unpack_CSC _cffi_d_GxB_Matrix_unpack_CSC
#endif

static GrB_Info _cffi_d_GxB_Matrix_unpack_CSR(struct GB_Matrix_opaque * x0, uint64_t * * x1, uint64_t * * x2, void * * x3, uint64_t * x4, uint64_t * x5, uint64_t * x6, _Bool * x7, _Bool * x8, struct GB_Descriptor_opaque * x9)
{
  return GxB_Matrix_unpack_CSR(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_unpack_CSR(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t * * x1;
  uint64_t * * x2;
  void * * x3;
  uint64_t * x4;
  uint64_t * x5;
  uint64_t * x6;
  _Bool * x7;
  _Bool * x8;
  struct GB_Descriptor_opaque * x9;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_unpack_CSR", 10, 10, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg9, (char **)&x9);
  if (datasize != 0) {
    x9 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg9, (char **)&x9,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_unpack_CSR(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_unpack_CSR _cffi_d_GxB_Matrix_unpack_CSR
#endif

static GrB_Info _cffi_d_GxB_Matrix_unpack_FullC(struct GB_Matrix_opaque * x0, void * * x1, uint64_t * x2, _Bool * x3, struct GB_Descriptor_opaque * x4)
{
  return GxB_Matrix_unpack_FullC(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_unpack_FullC(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  void * * x1;
  uint64_t * x2;
  _Bool * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_unpack_FullC", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_unpack_FullC(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_unpack_FullC _cffi_d_GxB_Matrix_unpack_FullC
#endif

static GrB_Info _cffi_d_GxB_Matrix_unpack_FullR(struct GB_Matrix_opaque * x0, void * * x1, uint64_t * x2, _Bool * x3, struct GB_Descriptor_opaque * x4)
{
  return GxB_Matrix_unpack_FullR(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_unpack_FullR(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  void * * x1;
  uint64_t * x2;
  _Bool * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_unpack_FullR", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_unpack_FullR(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_unpack_FullR _cffi_d_GxB_Matrix_unpack_FullR
#endif

static GrB_Info _cffi_d_GxB_Matrix_unpack_HyperCSC(struct GB_Matrix_opaque * x0, uint64_t * * x1, uint64_t * * x2, uint64_t * * x3, void * * x4, uint64_t * x5, uint64_t * x6, uint64_t * x7, uint64_t * x8, _Bool * x9, uint64_t * x10, _Bool * x11, struct GB_Descriptor_opaque * x12)
{
  return GxB_Matrix_unpack_HyperCSC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_unpack_HyperCSC(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t * * x1;
  uint64_t * * x2;
  uint64_t * * x3;
  void * * x4;
  uint64_t * x5;
  uint64_t * x6;
  uint64_t * x7;
  uint64_t * x8;
  _Bool * x9;
  uint64_t * x10;
  _Bool * x11;
  struct GB_Descriptor_opaque * x12;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;
  PyObject *arg11;
  PyObject *arg12;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_unpack_HyperCSC", 13, 13, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10, &arg11, &arg12))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg9, (char **)&x9);
  if (datasize != 0) {
    x9 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg9, (char **)&x9,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg10, (char **)&x10);
  if (datasize != 0) {
    x10 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg10, (char **)&x10,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg11, (char **)&x11);
  if (datasize != 0) {
    x11 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg11, (char **)&x11,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg12, (char **)&x12);
  if (datasize != 0) {
    x12 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg12, (char **)&x12,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_unpack_HyperCSC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_unpack_HyperCSC _cffi_d_GxB_Matrix_unpack_HyperCSC
#endif

static GrB_Info _cffi_d_GxB_Matrix_unpack_HyperCSR(struct GB_Matrix_opaque * x0, uint64_t * * x1, uint64_t * * x2, uint64_t * * x3, void * * x4, uint64_t * x5, uint64_t * x6, uint64_t * x7, uint64_t * x8, _Bool * x9, uint64_t * x10, _Bool * x11, struct GB_Descriptor_opaque * x12)
{
  return GxB_Matrix_unpack_HyperCSR(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Matrix_unpack_HyperCSR(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  uint64_t * * x1;
  uint64_t * * x2;
  uint64_t * * x3;
  void * * x4;
  uint64_t * x5;
  uint64_t * x6;
  uint64_t * x7;
  uint64_t * x8;
  _Bool * x9;
  uint64_t * x10;
  _Bool * x11;
  struct GB_Descriptor_opaque * x12;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;
  PyObject *arg11;
  PyObject *arg12;

  if (!PyArg_UnpackTuple(args, "GxB_Matrix_unpack_HyperCSR", 13, 13, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10, &arg11, &arg12))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg9, (char **)&x9);
  if (datasize != 0) {
    x9 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg9, (char **)&x9,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg10, (char **)&x10);
  if (datasize != 0) {
    x10 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg10, (char **)&x10,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg11, (char **)&x11);
  if (datasize != 0) {
    x11 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg11, (char **)&x11,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg12, (char **)&x12);
  if (datasize != 0) {
    x12 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg12, (char **)&x12,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Matrix_unpack_HyperCSR(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Matrix_unpack_HyperCSR _cffi_d_GxB_Matrix_unpack_HyperCSR
#endif

static GrB_Info _cffi_d_GxB_Monoid_fprint(struct GB_Monoid_opaque * x0, char const * x1, GxB_Print_Level x2, FILE * x3)
{
  return GxB_Monoid_fprint(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Monoid_fprint(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * x0;
  char const * x1;
  GxB_Print_Level x2;
  FILE * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Monoid_fprint", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x2, _cffi_type(463), arg2) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(464), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (FILE *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(464), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Monoid_fprint(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Monoid_fprint _cffi_d_GxB_Monoid_fprint
#endif

static GrB_Info _cffi_d_GxB_Monoid_identity(void * x0, struct GB_Monoid_opaque * x1)
{
  return GxB_Monoid_identity(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Monoid_identity(PyObject *self, PyObject *args)
{
  void * x0;
  struct GB_Monoid_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Monoid_identity", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Monoid_identity(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Monoid_identity _cffi_d_GxB_Monoid_identity
#endif

static GrB_Info _cffi_d_GxB_Monoid_new_FC32(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, _Fcomplex x2)
{
  return GxB_Monoid_new_FC32(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Monoid_new_FC32(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  _Fcomplex x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Monoid_new_FC32", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x2, _cffi_type(974), arg2) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Monoid_new_FC32(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Monoid_new_FC32(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, _Fcomplex *x2)
{
  GrB_Info result;
  { result = GxB_Monoid_new_FC32(x0, x1, *x2); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Monoid_new_FC64(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, _Dcomplex x2)
{
  return GxB_Monoid_new_FC64(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Monoid_new_FC64(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  _Dcomplex x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Monoid_new_FC64", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x2, _cffi_type(962), arg2) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Monoid_new_FC64(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Monoid_new_FC64(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, _Dcomplex *x2)
{
  GrB_Info result;
  { result = GxB_Monoid_new_FC64(x0, x1, *x2); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Monoid_operator(struct GB_BinaryOp_opaque * * x0, struct GB_Monoid_opaque * x1)
{
  return GxB_Monoid_operator(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Monoid_operator(PyObject *self, PyObject *args)
{
  struct GB_BinaryOp_opaque * * x0;
  struct GB_Monoid_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Monoid_operator", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(422), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(422), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Monoid_operator(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Monoid_operator _cffi_d_GxB_Monoid_operator
#endif

static GrB_Info _cffi_d_GxB_Monoid_terminal(_Bool * x0, void * x1, struct GB_Monoid_opaque * x2)
{
  return GxB_Monoid_terminal(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Monoid_terminal(PyObject *self, PyObject *args)
{
  _Bool * x0;
  void * x1;
  struct GB_Monoid_opaque * x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Monoid_terminal", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Monoid_terminal(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Monoid_terminal _cffi_d_GxB_Monoid_terminal
#endif

static GrB_Info _cffi_d_GxB_Monoid_terminal_new_BOOL(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, _Bool x2, _Bool x3)
{
  return GxB_Monoid_terminal_new_BOOL(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Monoid_terminal_new_BOOL(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  _Bool x2;
  _Bool x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Monoid_terminal_new_BOOL", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = (_Bool)_cffi_to_c__Bool(arg2);
  if (x2 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  x3 = (_Bool)_cffi_to_c__Bool(arg3);
  if (x3 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Monoid_terminal_new_BOOL(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Monoid_terminal_new_BOOL _cffi_d_GxB_Monoid_terminal_new_BOOL
#endif

static GrB_Info _cffi_d_GxB_Monoid_terminal_new_FC32(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, _Fcomplex x2, _Fcomplex x3)
{
  return GxB_Monoid_terminal_new_FC32(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Monoid_terminal_new_FC32(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  _Fcomplex x2;
  _Fcomplex x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Monoid_terminal_new_FC32", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x2, _cffi_type(974), arg2) < 0)
    return NULL;

  if (_cffi_to_c((char *)&x3, _cffi_type(974), arg3) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Monoid_terminal_new_FC32(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Monoid_terminal_new_FC32(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, _Fcomplex *x2, _Fcomplex *x3)
{
  GrB_Info result;
  { result = GxB_Monoid_terminal_new_FC32(x0, x1, *x2, *x3); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Monoid_terminal_new_FC64(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, _Dcomplex x2, _Dcomplex x3)
{
  return GxB_Monoid_terminal_new_FC64(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Monoid_terminal_new_FC64(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  _Dcomplex x2;
  _Dcomplex x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Monoid_terminal_new_FC64", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x2, _cffi_type(962), arg2) < 0)
    return NULL;

  if (_cffi_to_c((char *)&x3, _cffi_type(962), arg3) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Monoid_terminal_new_FC64(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Monoid_terminal_new_FC64(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, _Dcomplex *x2, _Dcomplex *x3)
{
  GrB_Info result;
  { result = GxB_Monoid_terminal_new_FC64(x0, x1, *x2, *x3); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Monoid_terminal_new_FP32(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, float x2, float x3)
{
  return GxB_Monoid_terminal_new_FP32(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Monoid_terminal_new_FP32(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  float x2;
  float x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Monoid_terminal_new_FP32", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = (float)_cffi_to_c_float(arg2);
  if (x2 == (float)-1 && PyErr_Occurred())
    return NULL;

  x3 = (float)_cffi_to_c_float(arg3);
  if (x3 == (float)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Monoid_terminal_new_FP32(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Monoid_terminal_new_FP32 _cffi_d_GxB_Monoid_terminal_new_FP32
#endif

static GrB_Info _cffi_d_GxB_Monoid_terminal_new_FP64(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, double x2, double x3)
{
  return GxB_Monoid_terminal_new_FP64(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Monoid_terminal_new_FP64(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  double x2;
  double x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Monoid_terminal_new_FP64", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = (double)_cffi_to_c_double(arg2);
  if (x2 == (double)-1 && PyErr_Occurred())
    return NULL;

  x3 = (double)_cffi_to_c_double(arg3);
  if (x3 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Monoid_terminal_new_FP64(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Monoid_terminal_new_FP64 _cffi_d_GxB_Monoid_terminal_new_FP64
#endif

static GrB_Info _cffi_d_GxB_Monoid_terminal_new_INT16(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, int16_t x2, int16_t x3)
{
  return GxB_Monoid_terminal_new_INT16(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Monoid_terminal_new_INT16(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  int16_t x2;
  int16_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Monoid_terminal_new_INT16", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, int16_t);
  if (x2 == (int16_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, int16_t);
  if (x3 == (int16_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Monoid_terminal_new_INT16(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Monoid_terminal_new_INT16 _cffi_d_GxB_Monoid_terminal_new_INT16
#endif

static GrB_Info _cffi_d_GxB_Monoid_terminal_new_INT32(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, int32_t x2, int32_t x3)
{
  return GxB_Monoid_terminal_new_INT32(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Monoid_terminal_new_INT32(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  int32_t x2;
  int32_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Monoid_terminal_new_INT32", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, int32_t);
  if (x2 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, int32_t);
  if (x3 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Monoid_terminal_new_INT32(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Monoid_terminal_new_INT32 _cffi_d_GxB_Monoid_terminal_new_INT32
#endif

static GrB_Info _cffi_d_GxB_Monoid_terminal_new_INT64(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, int64_t x2, int64_t x3)
{
  return GxB_Monoid_terminal_new_INT64(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Monoid_terminal_new_INT64(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  int64_t x2;
  int64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Monoid_terminal_new_INT64", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, int64_t);
  if (x2 == (int64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, int64_t);
  if (x3 == (int64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Monoid_terminal_new_INT64(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Monoid_terminal_new_INT64 _cffi_d_GxB_Monoid_terminal_new_INT64
#endif

static GrB_Info _cffi_d_GxB_Monoid_terminal_new_INT8(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, int8_t x2, int8_t x3)
{
  return GxB_Monoid_terminal_new_INT8(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Monoid_terminal_new_INT8(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  int8_t x2;
  int8_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Monoid_terminal_new_INT8", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, int8_t);
  if (x2 == (int8_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, int8_t);
  if (x3 == (int8_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Monoid_terminal_new_INT8(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Monoid_terminal_new_INT8 _cffi_d_GxB_Monoid_terminal_new_INT8
#endif

static GrB_Info _cffi_d_GxB_Monoid_terminal_new_UDT(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, void * x2, void * x3)
{
  return GxB_Monoid_terminal_new_UDT(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Monoid_terminal_new_UDT(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  void * x2;
  void * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Monoid_terminal_new_UDT", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Monoid_terminal_new_UDT(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Monoid_terminal_new_UDT _cffi_d_GxB_Monoid_terminal_new_UDT
#endif

static GrB_Info _cffi_d_GxB_Monoid_terminal_new_UINT16(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, uint16_t x2, uint16_t x3)
{
  return GxB_Monoid_terminal_new_UINT16(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Monoid_terminal_new_UINT16(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  uint16_t x2;
  uint16_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Monoid_terminal_new_UINT16", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint16_t);
  if (x2 == (uint16_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint16_t);
  if (x3 == (uint16_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Monoid_terminal_new_UINT16(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Monoid_terminal_new_UINT16 _cffi_d_GxB_Monoid_terminal_new_UINT16
#endif

static GrB_Info _cffi_d_GxB_Monoid_terminal_new_UINT32(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, uint32_t x2, uint32_t x3)
{
  return GxB_Monoid_terminal_new_UINT32(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Monoid_terminal_new_UINT32(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  uint32_t x2;
  uint32_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Monoid_terminal_new_UINT32", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint32_t);
  if (x2 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint32_t);
  if (x3 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Monoid_terminal_new_UINT32(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Monoid_terminal_new_UINT32 _cffi_d_GxB_Monoid_terminal_new_UINT32
#endif

static GrB_Info _cffi_d_GxB_Monoid_terminal_new_UINT64(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, uint64_t x2, uint64_t x3)
{
  return GxB_Monoid_terminal_new_UINT64(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Monoid_terminal_new_UINT64(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  uint64_t x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Monoid_terminal_new_UINT64", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Monoid_terminal_new_UINT64(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Monoid_terminal_new_UINT64 _cffi_d_GxB_Monoid_terminal_new_UINT64
#endif

static GrB_Info _cffi_d_GxB_Monoid_terminal_new_UINT8(struct GB_Monoid_opaque * * x0, struct GB_BinaryOp_opaque * x1, uint8_t x2, uint8_t x3)
{
  return GxB_Monoid_terminal_new_UINT8(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Monoid_terminal_new_UINT8(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_BinaryOp_opaque * x1;
  uint8_t x2;
  uint8_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Monoid_terminal_new_UINT8", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint8_t);
  if (x2 == (uint8_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = _cffi_to_c_int(arg3, uint8_t);
  if (x3 == (uint8_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Monoid_terminal_new_UINT8(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Monoid_terminal_new_UINT8 _cffi_d_GxB_Monoid_terminal_new_UINT8
#endif

static GrB_Info _cffi_d_GxB_Row_subassign(struct GB_Matrix_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Vector_opaque * x3, uint64_t x4, uint64_t const * x5, uint64_t x6, struct GB_Descriptor_opaque * x7)
{
  return GxB_Row_subassign(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Row_subassign(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Vector_opaque * x3;
  uint64_t x4;
  uint64_t const * x5;
  uint64_t x6;
  struct GB_Descriptor_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GxB_Row_subassign", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x6 = _cffi_to_c_int(arg6, uint64_t);
  if (x6 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Row_subassign(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Row_subassign _cffi_d_GxB_Row_subassign
#endif

static GrB_Info _cffi_d_GxB_Scalar_clear(struct GB_Scalar_opaque * x0)
{
  return GxB_Scalar_clear(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_clear(PyObject *self, PyObject *arg0)
{
  struct GB_Scalar_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_clear(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_clear _cffi_d_GxB_Scalar_clear
#endif

static GrB_Info _cffi_d_GxB_Scalar_dup(struct GB_Scalar_opaque * * x0, struct GB_Scalar_opaque * x1)
{
  return GxB_Scalar_dup(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_dup(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_dup", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2109), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2109), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_dup(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_dup _cffi_d_GxB_Scalar_dup
#endif

static GrB_Info _cffi_d_GxB_Scalar_error(char const * * x0, struct GB_Scalar_opaque * x1)
{
  return GxB_Scalar_error(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_error(PyObject *self, PyObject *args)
{
  char const * * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_error", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(130), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char const * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(130), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_error(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_error _cffi_d_GxB_Scalar_error
#endif

static GrB_Info _cffi_d_GxB_Scalar_extractElement_BOOL(_Bool * x0, struct GB_Scalar_opaque * x1)
{
  return GxB_Scalar_extractElement_BOOL(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_extractElement_BOOL(PyObject *self, PyObject *args)
{
  _Bool * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_extractElement_BOOL", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_extractElement_BOOL(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_extractElement_BOOL _cffi_d_GxB_Scalar_extractElement_BOOL
#endif

static GrB_Info _cffi_d_GxB_Scalar_extractElement_FC32(_Fcomplex * x0, struct GB_Scalar_opaque * x1)
{
  return GxB_Scalar_extractElement_FC32(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_extractElement_FC32(PyObject *self, PyObject *args)
{
  _Fcomplex * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_extractElement_FC32", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(261), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (_Fcomplex *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(261), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_extractElement_FC32(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_extractElement_FC32 _cffi_d_GxB_Scalar_extractElement_FC32
#endif

static GrB_Info _cffi_d_GxB_Scalar_extractElement_FC64(_Dcomplex * x0, struct GB_Scalar_opaque * x1)
{
  return GxB_Scalar_extractElement_FC64(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_extractElement_FC64(PyObject *self, PyObject *args)
{
  _Dcomplex * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_extractElement_FC64", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(203), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (_Dcomplex *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(203), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_extractElement_FC64(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_extractElement_FC64 _cffi_d_GxB_Scalar_extractElement_FC64
#endif

static GrB_Info _cffi_d_GxB_Scalar_extractElement_FP32(float * x0, struct GB_Scalar_opaque * x1)
{
  return GxB_Scalar_extractElement_FP32(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_extractElement_FP32(PyObject *self, PyObject *args)
{
  float * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_extractElement_FP32", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(232), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (float *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(232), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_extractElement_FP32(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_extractElement_FP32 _cffi_d_GxB_Scalar_extractElement_FP32
#endif

static GrB_Info _cffi_d_GxB_Scalar_extractElement_FP64(double * x0, struct GB_Scalar_opaque * x1)
{
  return GxB_Scalar_extractElement_FP64(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_extractElement_FP64(PyObject *self, PyObject *args)
{
  double * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_extractElement_FP64", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(28), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(28), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_extractElement_FP64(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_extractElement_FP64 _cffi_d_GxB_Scalar_extractElement_FP64
#endif

static GrB_Info _cffi_d_GxB_Scalar_extractElement_INT16(int16_t * x0, struct GB_Scalar_opaque * x1)
{
  return GxB_Scalar_extractElement_INT16(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_extractElement_INT16(PyObject *self, PyObject *args)
{
  int16_t * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_extractElement_INT16", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(290), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int16_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(290), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_extractElement_INT16(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_extractElement_INT16 _cffi_d_GxB_Scalar_extractElement_INT16
#endif

static GrB_Info _cffi_d_GxB_Scalar_extractElement_INT32(int32_t * x0, struct GB_Scalar_opaque * x1)
{
  return GxB_Scalar_extractElement_INT32(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_extractElement_INT32(PyObject *self, PyObject *args)
{
  int32_t * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_extractElement_INT32", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(36), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int32_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(36), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_extractElement_INT32(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_extractElement_INT32 _cffi_d_GxB_Scalar_extractElement_INT32
#endif

static GrB_Info _cffi_d_GxB_Scalar_extractElement_INT64(int64_t * x0, struct GB_Scalar_opaque * x1)
{
  return GxB_Scalar_extractElement_INT64(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_extractElement_INT64(PyObject *self, PyObject *args)
{
  int64_t * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_extractElement_INT64", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(44), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(44), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_extractElement_INT64(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_extractElement_INT64 _cffi_d_GxB_Scalar_extractElement_INT64
#endif

static GrB_Info _cffi_d_GxB_Scalar_extractElement_INT8(int8_t * x0, struct GB_Scalar_opaque * x1)
{
  return GxB_Scalar_extractElement_INT8(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_extractElement_INT8(PyObject *self, PyObject *args)
{
  int8_t * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_extractElement_INT8", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(377), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (int8_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(377), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_extractElement_INT8(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_extractElement_INT8 _cffi_d_GxB_Scalar_extractElement_INT8
#endif

static GrB_Info _cffi_d_GxB_Scalar_extractElement_UDT(void * x0, struct GB_Scalar_opaque * x1)
{
  return GxB_Scalar_extractElement_UDT(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_extractElement_UDT(PyObject *self, PyObject *args)
{
  void * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_extractElement_UDT", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_extractElement_UDT(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_extractElement_UDT _cffi_d_GxB_Scalar_extractElement_UDT
#endif

static GrB_Info _cffi_d_GxB_Scalar_extractElement_UINT16(uint16_t * x0, struct GB_Scalar_opaque * x1)
{
  return GxB_Scalar_extractElement_UINT16(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_extractElement_UINT16(PyObject *self, PyObject *args)
{
  uint16_t * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_extractElement_UINT16", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3386), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint16_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3386), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_extractElement_UINT16(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_extractElement_UINT16 _cffi_d_GxB_Scalar_extractElement_UINT16
#endif

static GrB_Info _cffi_d_GxB_Scalar_extractElement_UINT32(uint32_t * x0, struct GB_Scalar_opaque * x1)
{
  return GxB_Scalar_extractElement_UINT32(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_extractElement_UINT32(PyObject *self, PyObject *args)
{
  uint32_t * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_extractElement_UINT32", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3415), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint32_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3415), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_extractElement_UINT32(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_extractElement_UINT32 _cffi_d_GxB_Scalar_extractElement_UINT32
#endif

static GrB_Info _cffi_d_GxB_Scalar_extractElement_UINT64(uint64_t * x0, struct GB_Scalar_opaque * x1)
{
  return GxB_Scalar_extractElement_UINT64(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_extractElement_UINT64(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_extractElement_UINT64", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_extractElement_UINT64(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_extractElement_UINT64 _cffi_d_GxB_Scalar_extractElement_UINT64
#endif

static GrB_Info _cffi_d_GxB_Scalar_extractElement_UINT8(uint8_t * x0, struct GB_Scalar_opaque * x1)
{
  return GxB_Scalar_extractElement_UINT8(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_extractElement_UINT8(PyObject *self, PyObject *args)
{
  uint8_t * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_extractElement_UINT8", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(3766), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint8_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(3766), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_extractElement_UINT8(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_extractElement_UINT8 _cffi_d_GxB_Scalar_extractElement_UINT8
#endif

static GrB_Info _cffi_d_GxB_Scalar_fprint(struct GB_Scalar_opaque * x0, char const * x1, GxB_Print_Level x2, FILE * x3)
{
  return GxB_Scalar_fprint(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_fprint(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  char const * x1;
  GxB_Print_Level x2;
  FILE * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_fprint", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x2, _cffi_type(463), arg2) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(464), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (FILE *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(464), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_fprint(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_fprint _cffi_d_GxB_Scalar_fprint
#endif

static GrB_Info _cffi_d_GxB_Scalar_free(struct GB_Scalar_opaque * * x0)
{
  return GxB_Scalar_free(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_free(PyObject *self, PyObject *arg0)
{
  struct GB_Scalar_opaque * * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2109), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2109), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_free(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_free _cffi_d_GxB_Scalar_free
#endif

static GrB_Info _cffi_d_GxB_Scalar_memoryUsage(size_t * x0, struct GB_Scalar_opaque * x1)
{
  return GxB_Scalar_memoryUsage(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_memoryUsage(PyObject *self, PyObject *args)
{
  size_t * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_memoryUsage", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(406), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (size_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(406), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_memoryUsage(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_memoryUsage _cffi_d_GxB_Scalar_memoryUsage
#endif

static GrB_Info _cffi_d_GxB_Scalar_new(struct GB_Scalar_opaque * * x0, struct GB_Type_opaque * x1)
{
  return GxB_Scalar_new(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_new(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * * x0;
  struct GB_Type_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_new", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2109), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2109), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_new(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_new _cffi_d_GxB_Scalar_new
#endif

static GrB_Info _cffi_d_GxB_Scalar_nvals(uint64_t * x0, struct GB_Scalar_opaque * x1)
{
  return GxB_Scalar_nvals(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_nvals(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_nvals", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_nvals(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_nvals _cffi_d_GxB_Scalar_nvals
#endif

static GrB_Info _cffi_d_GxB_Scalar_setElement_BOOL(struct GB_Scalar_opaque * x0, _Bool x1)
{
  return GxB_Scalar_setElement_BOOL(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_setElement_BOOL(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  _Bool x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_setElement_BOOL", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = (_Bool)_cffi_to_c__Bool(arg1);
  if (x1 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_setElement_BOOL(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_setElement_BOOL _cffi_d_GxB_Scalar_setElement_BOOL
#endif

static GrB_Info _cffi_d_GxB_Scalar_setElement_FC32(struct GB_Scalar_opaque * x0, _Fcomplex x1)
{
  return GxB_Scalar_setElement_FC32(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_setElement_FC32(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  _Fcomplex x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_setElement_FC32", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(974), arg1) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_setElement_FC32(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Scalar_setElement_FC32(struct GB_Scalar_opaque * x0, _Fcomplex *x1)
{
  GrB_Info result;
  { result = GxB_Scalar_setElement_FC32(x0, *x1); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Scalar_setElement_FC64(struct GB_Scalar_opaque * x0, _Dcomplex x1)
{
  return GxB_Scalar_setElement_FC64(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_setElement_FC64(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  _Dcomplex x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_setElement_FC64", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(962), arg1) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_setElement_FC64(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Scalar_setElement_FC64(struct GB_Scalar_opaque * x0, _Dcomplex *x1)
{
  GrB_Info result;
  { result = GxB_Scalar_setElement_FC64(x0, *x1); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Scalar_setElement_FP32(struct GB_Scalar_opaque * x0, float x1)
{
  return GxB_Scalar_setElement_FP32(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_setElement_FP32(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  float x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_setElement_FP32", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = (float)_cffi_to_c_float(arg1);
  if (x1 == (float)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_setElement_FP32(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_setElement_FP32 _cffi_d_GxB_Scalar_setElement_FP32
#endif

static GrB_Info _cffi_d_GxB_Scalar_setElement_FP64(struct GB_Scalar_opaque * x0, double x1)
{
  return GxB_Scalar_setElement_FP64(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_setElement_FP64(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  double x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_setElement_FP64", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = (double)_cffi_to_c_double(arg1);
  if (x1 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_setElement_FP64(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_setElement_FP64 _cffi_d_GxB_Scalar_setElement_FP64
#endif

static GrB_Info _cffi_d_GxB_Scalar_setElement_INT16(struct GB_Scalar_opaque * x0, int16_t x1)
{
  return GxB_Scalar_setElement_INT16(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_setElement_INT16(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  int16_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_setElement_INT16", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, int16_t);
  if (x1 == (int16_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_setElement_INT16(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_setElement_INT16 _cffi_d_GxB_Scalar_setElement_INT16
#endif

static GrB_Info _cffi_d_GxB_Scalar_setElement_INT32(struct GB_Scalar_opaque * x0, int32_t x1)
{
  return GxB_Scalar_setElement_INT32(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_setElement_INT32(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  int32_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_setElement_INT32", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, int32_t);
  if (x1 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_setElement_INT32(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_setElement_INT32 _cffi_d_GxB_Scalar_setElement_INT32
#endif

static GrB_Info _cffi_d_GxB_Scalar_setElement_INT64(struct GB_Scalar_opaque * x0, int64_t x1)
{
  return GxB_Scalar_setElement_INT64(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_setElement_INT64(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  int64_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_setElement_INT64", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, int64_t);
  if (x1 == (int64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_setElement_INT64(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_setElement_INT64 _cffi_d_GxB_Scalar_setElement_INT64
#endif

static GrB_Info _cffi_d_GxB_Scalar_setElement_INT8(struct GB_Scalar_opaque * x0, int8_t x1)
{
  return GxB_Scalar_setElement_INT8(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_setElement_INT8(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  int8_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_setElement_INT8", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, int8_t);
  if (x1 == (int8_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_setElement_INT8(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_setElement_INT8 _cffi_d_GxB_Scalar_setElement_INT8
#endif

static GrB_Info _cffi_d_GxB_Scalar_setElement_UDT(struct GB_Scalar_opaque * x0, void * x1)
{
  return GxB_Scalar_setElement_UDT(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_setElement_UDT(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  void * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_setElement_UDT", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_setElement_UDT(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_setElement_UDT _cffi_d_GxB_Scalar_setElement_UDT
#endif

static GrB_Info _cffi_d_GxB_Scalar_setElement_UINT16(struct GB_Scalar_opaque * x0, uint16_t x1)
{
  return GxB_Scalar_setElement_UINT16(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_setElement_UINT16(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  uint16_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_setElement_UINT16", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint16_t);
  if (x1 == (uint16_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_setElement_UINT16(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_setElement_UINT16 _cffi_d_GxB_Scalar_setElement_UINT16
#endif

static GrB_Info _cffi_d_GxB_Scalar_setElement_UINT32(struct GB_Scalar_opaque * x0, uint32_t x1)
{
  return GxB_Scalar_setElement_UINT32(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_setElement_UINT32(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  uint32_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_setElement_UINT32", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint32_t);
  if (x1 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_setElement_UINT32(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_setElement_UINT32 _cffi_d_GxB_Scalar_setElement_UINT32
#endif

static GrB_Info _cffi_d_GxB_Scalar_setElement_UINT64(struct GB_Scalar_opaque * x0, uint64_t x1)
{
  return GxB_Scalar_setElement_UINT64(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_setElement_UINT64(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  uint64_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_setElement_UINT64", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint64_t);
  if (x1 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_setElement_UINT64(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_setElement_UINT64 _cffi_d_GxB_Scalar_setElement_UINT64
#endif

static GrB_Info _cffi_d_GxB_Scalar_setElement_UINT8(struct GB_Scalar_opaque * x0, uint8_t x1)
{
  return GxB_Scalar_setElement_UINT8(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_setElement_UINT8(PyObject *self, PyObject *args)
{
  struct GB_Scalar_opaque * x0;
  uint8_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_setElement_UINT8", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint8_t);
  if (x1 == (uint8_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_setElement_UINT8(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_setElement_UINT8 _cffi_d_GxB_Scalar_setElement_UINT8
#endif

static GrB_Info _cffi_d_GxB_Scalar_type(struct GB_Type_opaque * * x0, struct GB_Scalar_opaque * x1)
{
  return GxB_Scalar_type(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_type(PyObject *self, PyObject *args)
{
  struct GB_Type_opaque * * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_type", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_type(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_type _cffi_d_GxB_Scalar_type
#endif

static GrB_Info _cffi_d_GxB_Scalar_type_name(char * x0, struct GB_Scalar_opaque * x1)
{
  return GxB_Scalar_type_name(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_type_name(PyObject *self, PyObject *args)
{
  char * x0;
  struct GB_Scalar_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Scalar_type_name", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(97), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(97), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_type_name(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_type_name _cffi_d_GxB_Scalar_type_name
#endif

static GrB_Info _cffi_d_GxB_Scalar_wait(struct GB_Scalar_opaque * * x0)
{
  return GxB_Scalar_wait(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Scalar_wait(PyObject *self, PyObject *arg0)
{
  struct GB_Scalar_opaque * * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2109), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2109), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Scalar_wait(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Scalar_wait _cffi_d_GxB_Scalar_wait
#endif

static GrB_Info _cffi_d_GxB_SelectOp_error(char const * * x0, struct GB_SelectOp_opaque * x1)
{
  return GxB_SelectOp_error(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_SelectOp_error(PyObject *self, PyObject *args)
{
  char const * * x0;
  struct GB_SelectOp_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_SelectOp_error", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(130), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char const * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(130), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(155), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_SelectOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(155), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_SelectOp_error(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_SelectOp_error _cffi_d_GxB_SelectOp_error
#endif

static GrB_Info _cffi_d_GxB_SelectOp_fprint(struct GB_SelectOp_opaque * x0, char const * x1, GxB_Print_Level x2, FILE * x3)
{
  return GxB_SelectOp_fprint(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_SelectOp_fprint(PyObject *self, PyObject *args)
{
  struct GB_SelectOp_opaque * x0;
  char const * x1;
  GxB_Print_Level x2;
  FILE * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_SelectOp_fprint", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(155), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_SelectOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(155), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x2, _cffi_type(463), arg2) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(464), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (FILE *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(464), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_SelectOp_fprint(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_SelectOp_fprint _cffi_d_GxB_SelectOp_fprint
#endif

static GrB_Info _cffi_d_GxB_SelectOp_free(struct GB_SelectOp_opaque * * x0)
{
  return GxB_SelectOp_free(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_SelectOp_free(PyObject *self, PyObject *arg0)
{
  struct GB_SelectOp_opaque * * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2228), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_SelectOp_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2228), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_SelectOp_free(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_SelectOp_free _cffi_d_GxB_SelectOp_free
#endif

static GrB_Info _cffi_d_GxB_SelectOp_new(struct GB_SelectOp_opaque * * x0, _Bool(* x1)(uint64_t, uint64_t, void const *, void const *), struct GB_Type_opaque * x2, struct GB_Type_opaque * x3)
{
  return GxB_SelectOp_new(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_SelectOp_new(PyObject *self, PyObject *args)
{
  struct GB_SelectOp_opaque * * x0;
  _Bool(* x1)(uint64_t, uint64_t, void const *, void const *);
  struct GB_Type_opaque * x2;
  struct GB_Type_opaque * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_SelectOp_new", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2228), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_SelectOp_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2228), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = (_Bool(*)(uint64_t, uint64_t, void const *, void const *))_cffi_to_c_pointer(arg1, _cffi_type(2232));
  if (x1 == (_Bool(*)(uint64_t, uint64_t, void const *, void const *))NULL && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_SelectOp_new(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_SelectOp_new _cffi_d_GxB_SelectOp_new
#endif

static GrB_Info _cffi_d_GxB_SelectOp_ttype(struct GB_Type_opaque * * x0, struct GB_SelectOp_opaque * x1)
{
  return GxB_SelectOp_ttype(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_SelectOp_ttype(PyObject *self, PyObject *args)
{
  struct GB_Type_opaque * * x0;
  struct GB_SelectOp_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_SelectOp_ttype", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(155), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_SelectOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(155), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_SelectOp_ttype(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_SelectOp_ttype _cffi_d_GxB_SelectOp_ttype
#endif

static GrB_Info _cffi_d_GxB_SelectOp_wait(struct GB_SelectOp_opaque * x0, GrB_WaitMode x1)
{
  return GxB_SelectOp_wait(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_SelectOp_wait(PyObject *self, PyObject *args)
{
  struct GB_SelectOp_opaque * x0;
  GrB_WaitMode x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_SelectOp_wait", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(155), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_SelectOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(155), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(458), arg1) < 0)
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_SelectOp_wait(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_SelectOp_wait _cffi_d_GxB_SelectOp_wait
#endif

static GrB_Info _cffi_d_GxB_SelectOp_xtype(struct GB_Type_opaque * * x0, struct GB_SelectOp_opaque * x1)
{
  return GxB_SelectOp_xtype(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_SelectOp_xtype(PyObject *self, PyObject *args)
{
  struct GB_Type_opaque * * x0;
  struct GB_SelectOp_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_SelectOp_xtype", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(155), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_SelectOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(155), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_SelectOp_xtype(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_SelectOp_xtype _cffi_d_GxB_SelectOp_xtype
#endif

static GrB_Info _cffi_d_GxB_Semiring_add(struct GB_Monoid_opaque * * x0, struct GB_Semiring_opaque * x1)
{
  return GxB_Semiring_add(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Semiring_add(PyObject *self, PyObject *args)
{
  struct GB_Monoid_opaque * * x0;
  struct GB_Semiring_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Semiring_add", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(1938), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(1938), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(159), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Semiring_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(159), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Semiring_add(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Semiring_add _cffi_d_GxB_Semiring_add
#endif

static GrB_Info _cffi_d_GxB_Semiring_fprint(struct GB_Semiring_opaque * x0, char const * x1, GxB_Print_Level x2, FILE * x3)
{
  return GxB_Semiring_fprint(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Semiring_fprint(PyObject *self, PyObject *args)
{
  struct GB_Semiring_opaque * x0;
  char const * x1;
  GxB_Print_Level x2;
  FILE * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Semiring_fprint", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(159), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Semiring_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(159), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x2, _cffi_type(463), arg2) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(464), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (FILE *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(464), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Semiring_fprint(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Semiring_fprint _cffi_d_GxB_Semiring_fprint
#endif

static GrB_Info _cffi_d_GxB_Semiring_multiply(struct GB_BinaryOp_opaque * * x0, struct GB_Semiring_opaque * x1)
{
  return GxB_Semiring_multiply(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Semiring_multiply(PyObject *self, PyObject *args)
{
  struct GB_BinaryOp_opaque * * x0;
  struct GB_Semiring_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Semiring_multiply", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(422), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(422), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(159), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Semiring_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(159), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Semiring_multiply(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Semiring_multiply _cffi_d_GxB_Semiring_multiply
#endif

static GrB_Info _cffi_d_GxB_Type_fprint(struct GB_Type_opaque * x0, char const * x1, GxB_Print_Level x2, FILE * x3)
{
  return GxB_Type_fprint(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Type_fprint(PyObject *self, PyObject *args)
{
  struct GB_Type_opaque * x0;
  char const * x1;
  GxB_Print_Level x2;
  FILE * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Type_fprint", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x2, _cffi_type(463), arg2) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(464), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (FILE *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(464), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Type_fprint(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Type_fprint _cffi_d_GxB_Type_fprint
#endif

static GrB_Info _cffi_d_GxB_Type_from_name(struct GB_Type_opaque * * x0, char const * x1)
{
  return GxB_Type_from_name(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Type_from_name(PyObject *self, PyObject *args)
{
  struct GB_Type_opaque * * x0;
  char const * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Type_from_name", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Type_from_name(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Type_from_name _cffi_d_GxB_Type_from_name
#endif

static GrB_Info _cffi_d_GxB_Type_name(char * x0, struct GB_Type_opaque * x1)
{
  return GxB_Type_name(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Type_name(PyObject *self, PyObject *args)
{
  char * x0;
  struct GB_Type_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Type_name", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(97), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(97), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Type_name(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Type_name _cffi_d_GxB_Type_name
#endif

static GrB_Info _cffi_d_GxB_Type_new(struct GB_Type_opaque * * x0, size_t x1, char const * x2, char const * x3)
{
  return GxB_Type_new(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Type_new(PyObject *self, PyObject *args)
{
  struct GB_Type_opaque * * x0;
  size_t x1;
  char const * x2;
  char const * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Type_new", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, size_t);
  if (x1 == (size_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Type_new(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Type_new _cffi_d_GxB_Type_new
#endif

static GrB_Info _cffi_d_GxB_Type_size(size_t * x0, struct GB_Type_opaque * x1)
{
  return GxB_Type_size(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Type_size(PyObject *self, PyObject *args)
{
  size_t * x0;
  struct GB_Type_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Type_size", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(406), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (size_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(406), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Type_size(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Type_size _cffi_d_GxB_Type_size
#endif

static GrB_Info _cffi_d_GxB_UnaryOp_fprint(struct GB_UnaryOp_opaque * x0, char const * x1, GxB_Print_Level x2, FILE * x3)
{
  return GxB_UnaryOp_fprint(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_UnaryOp_fprint(PyObject *self, PyObject *args)
{
  struct GB_UnaryOp_opaque * x0;
  char const * x1;
  GxB_Print_Level x2;
  FILE * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_UnaryOp_fprint", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(118), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_UnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(118), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x2, _cffi_type(463), arg2) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(464), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (FILE *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(464), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_UnaryOp_fprint(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_UnaryOp_fprint _cffi_d_GxB_UnaryOp_fprint
#endif

static GrB_Info _cffi_d_GxB_UnaryOp_new(struct GB_UnaryOp_opaque * * x0, void(* x1)(void *, void const *), struct GB_Type_opaque * x2, struct GB_Type_opaque * x3, char const * x4, char const * x5)
{
  return GxB_UnaryOp_new(x0, x1, x2, x3, x4, x5);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_UnaryOp_new(PyObject *self, PyObject *args)
{
  struct GB_UnaryOp_opaque * * x0;
  void(* x1)(void *, void const *);
  struct GB_Type_opaque * x2;
  struct GB_Type_opaque * x3;
  char const * x4;
  char const * x5;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;

  if (!PyArg_UnpackTuple(args, "GxB_UnaryOp_new", 6, 6, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2328), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_UnaryOp_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2328), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = (void(*)(void *, void const *))_cffi_to_c_pointer(arg1, _cffi_type(2332));
  if (x1 == (void(*)(void *, void const *))NULL && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_UnaryOp_new(x0, x1, x2, x3, x4, x5); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_UnaryOp_new _cffi_d_GxB_UnaryOp_new
#endif

static GrB_Info _cffi_d_GxB_UnaryOp_xtype(struct GB_Type_opaque * * x0, struct GB_UnaryOp_opaque * x1)
{
  return GxB_UnaryOp_xtype(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_UnaryOp_xtype(PyObject *self, PyObject *args)
{
  struct GB_Type_opaque * * x0;
  struct GB_UnaryOp_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_UnaryOp_xtype", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(118), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_UnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(118), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_UnaryOp_xtype(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_UnaryOp_xtype _cffi_d_GxB_UnaryOp_xtype
#endif

static GrB_Info _cffi_d_GxB_UnaryOp_xtype_name(char * x0, struct GB_UnaryOp_opaque * x1)
{
  return GxB_UnaryOp_xtype_name(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_UnaryOp_xtype_name(PyObject *self, PyObject *args)
{
  char * x0;
  struct GB_UnaryOp_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_UnaryOp_xtype_name", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(97), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(97), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(118), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_UnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(118), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_UnaryOp_xtype_name(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_UnaryOp_xtype_name _cffi_d_GxB_UnaryOp_xtype_name
#endif

static GrB_Info _cffi_d_GxB_UnaryOp_ztype(struct GB_Type_opaque * * x0, struct GB_UnaryOp_opaque * x1)
{
  return GxB_UnaryOp_ztype(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_UnaryOp_ztype(PyObject *self, PyObject *args)
{
  struct GB_Type_opaque * * x0;
  struct GB_UnaryOp_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_UnaryOp_ztype", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(118), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_UnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(118), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_UnaryOp_ztype(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_UnaryOp_ztype _cffi_d_GxB_UnaryOp_ztype
#endif

static GrB_Info _cffi_d_GxB_UnaryOp_ztype_name(char * x0, struct GB_UnaryOp_opaque * x1)
{
  return GxB_UnaryOp_ztype_name(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_UnaryOp_ztype_name(PyObject *self, PyObject *args)
{
  char * x0;
  struct GB_UnaryOp_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_UnaryOp_ztype_name", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(97), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(97), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(118), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_UnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(118), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_UnaryOp_ztype_name(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_UnaryOp_ztype_name _cffi_d_GxB_UnaryOp_ztype_name
#endif

static GrB_Info _cffi_d_GxB_Vector_Iterator_attach(struct GB_Iterator_opaque * x0, struct GB_Vector_opaque * x1, struct GB_Descriptor_opaque * x2)
{
  return GxB_Vector_Iterator_attach(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_Iterator_attach(PyObject *self, PyObject *args)
{
  struct GB_Iterator_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_Descriptor_opaque * x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_Iterator_attach", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_Iterator_attach(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_Iterator_attach _cffi_d_GxB_Vector_Iterator_attach
#endif

static uint64_t _cffi_d_GxB_Vector_Iterator_getIndex(struct GB_Iterator_opaque * x0)
{
  return GxB_Vector_Iterator_getIndex(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_Iterator_getIndex(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  uint64_t result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_Iterator_getIndex(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, uint64_t);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_Iterator_getIndex _cffi_d_GxB_Vector_Iterator_getIndex
#endif

static uint64_t _cffi_d_GxB_Vector_Iterator_getp(struct GB_Iterator_opaque * x0)
{
  return GxB_Vector_Iterator_getp(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_Iterator_getp(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  uint64_t result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_Iterator_getp(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, uint64_t);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_Iterator_getp _cffi_d_GxB_Vector_Iterator_getp
#endif

static uint64_t _cffi_d_GxB_Vector_Iterator_getpmax(struct GB_Iterator_opaque * x0)
{
  return GxB_Vector_Iterator_getpmax(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_Iterator_getpmax(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  uint64_t result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_Iterator_getpmax(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, uint64_t);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_Iterator_getpmax _cffi_d_GxB_Vector_Iterator_getpmax
#endif

static GrB_Info _cffi_d_GxB_Vector_Iterator_next(struct GB_Iterator_opaque * x0)
{
  return GxB_Vector_Iterator_next(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_Iterator_next(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_Iterator_next(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_Iterator_next _cffi_d_GxB_Vector_Iterator_next
#endif

static GrB_Info _cffi_d_GxB_Vector_Iterator_seek(struct GB_Iterator_opaque * x0, uint64_t x1)
{
  return GxB_Vector_Iterator_seek(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_Iterator_seek(PyObject *self, PyObject *args)
{
  struct GB_Iterator_opaque * x0;
  uint64_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_Iterator_seek", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint64_t);
  if (x1 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_Iterator_seek(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_Iterator_seek _cffi_d_GxB_Vector_Iterator_seek
#endif

static void _cffi_const_GxB_Vector_Option_get(char *o)
{
  *(GrB_Info(* *)(struct GB_Vector_opaque *, GxB_Option_Field, ...))o = GxB_Vector_Option_get;
}

static GrB_Info _cffi_d_GxB_Vector_Option_get_FP64(struct GB_Vector_opaque * x0, GxB_Option_Field x1, double * x2)
{
  return GxB_Vector_Option_get_FP64(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_Option_get_FP64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  GxB_Option_Field x1;
  double * x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_Option_get_FP64", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(20), arg1) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(28), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (double *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(28), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_Option_get_FP64(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_Option_get_FP64 _cffi_d_GxB_Vector_Option_get_FP64
#endif

static GrB_Info _cffi_d_GxB_Vector_Option_get_INT32(struct GB_Vector_opaque * x0, GxB_Option_Field x1, int32_t * x2)
{
  return GxB_Vector_Option_get_INT32(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_Option_get_INT32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  GxB_Option_Field x1;
  int32_t * x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_Option_get_INT32", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(20), arg1) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(36), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (int32_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(36), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_Option_get_INT32(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_Option_get_INT32 _cffi_d_GxB_Vector_Option_get_INT32
#endif

static void _cffi_const_GxB_Vector_Option_set(char *o)
{
  *(GrB_Info(* *)(struct GB_Vector_opaque *, GxB_Option_Field, ...))o = GxB_Vector_Option_set;
}

static GrB_Info _cffi_d_GxB_Vector_Option_set_FP64(struct GB_Vector_opaque * x0, GxB_Option_Field x1, double x2)
{
  return GxB_Vector_Option_set_FP64(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_Option_set_FP64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  GxB_Option_Field x1;
  double x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_Option_set_FP64", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(20), arg1) < 0)
    return NULL;

  x2 = (double)_cffi_to_c_double(arg2);
  if (x2 == (double)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_Option_set_FP64(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_Option_set_FP64 _cffi_d_GxB_Vector_Option_set_FP64
#endif

static GrB_Info _cffi_d_GxB_Vector_Option_set_INT32(struct GB_Vector_opaque * x0, GxB_Option_Field x1, int32_t x2)
{
  return GxB_Vector_Option_set_INT32(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_Option_set_INT32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  GxB_Option_Field x1;
  int32_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_Option_set_INT32", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(20), arg1) < 0)
    return NULL;

  x2 = _cffi_to_c_int(arg2, int32_t);
  if (x2 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_Option_set_INT32(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_Option_set_INT32 _cffi_d_GxB_Vector_Option_set_INT32
#endif

static GrB_Info _cffi_d_GxB_Vector_apply_BinaryOp1st(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Scalar_opaque * x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_apply_BinaryOp1st(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_apply_BinaryOp1st(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Scalar_opaque * x4;
  struct GB_Vector_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_apply_BinaryOp1st", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_apply_BinaryOp1st(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_apply_BinaryOp1st _cffi_d_GxB_Vector_apply_BinaryOp1st
#endif

static GrB_Info _cffi_d_GxB_Vector_apply_BinaryOp1st_FC32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, _Fcomplex x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_apply_BinaryOp1st_FC32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_apply_BinaryOp1st_FC32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  _Fcomplex x4;
  struct GB_Vector_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_apply_BinaryOp1st_FC32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x4, _cffi_type(974), arg4) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_apply_BinaryOp1st_FC32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Vector_apply_BinaryOp1st_FC32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, _Fcomplex *x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  GrB_Info result;
  { result = GxB_Vector_apply_BinaryOp1st_FC32(x0, x1, x2, x3, *x4, x5, x6); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Vector_apply_BinaryOp1st_FC64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, _Dcomplex x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_apply_BinaryOp1st_FC64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_apply_BinaryOp1st_FC64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  _Dcomplex x4;
  struct GB_Vector_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_apply_BinaryOp1st_FC64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x4, _cffi_type(962), arg4) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_apply_BinaryOp1st_FC64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Vector_apply_BinaryOp1st_FC64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, _Dcomplex *x4, struct GB_Vector_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  GrB_Info result;
  { result = GxB_Vector_apply_BinaryOp1st_FC64(x0, x1, x2, x3, *x4, x5, x6); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Vector_apply_BinaryOp2nd(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Vector_opaque * x4, struct GB_Scalar_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_apply_BinaryOp2nd(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_apply_BinaryOp2nd(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  struct GB_Scalar_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_apply_BinaryOp2nd", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_apply_BinaryOp2nd(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_apply_BinaryOp2nd _cffi_d_GxB_Vector_apply_BinaryOp2nd
#endif

static GrB_Info _cffi_d_GxB_Vector_apply_BinaryOp2nd_FC32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Vector_opaque * x4, _Fcomplex x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_apply_BinaryOp2nd_FC32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_apply_BinaryOp2nd_FC32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  _Fcomplex x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_apply_BinaryOp2nd_FC32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x5, _cffi_type(974), arg5) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_apply_BinaryOp2nd_FC32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Vector_apply_BinaryOp2nd_FC32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Vector_opaque * x4, _Fcomplex *x5, struct GB_Descriptor_opaque * x6)
{
  GrB_Info result;
  { result = GxB_Vector_apply_BinaryOp2nd_FC32(x0, x1, x2, x3, x4, *x5, x6); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Vector_apply_BinaryOp2nd_FC64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Vector_opaque * x4, _Dcomplex x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_apply_BinaryOp2nd_FC64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_apply_BinaryOp2nd_FC64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  _Dcomplex x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_apply_BinaryOp2nd_FC64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x5, _cffi_type(962), arg5) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_apply_BinaryOp2nd_FC64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Vector_apply_BinaryOp2nd_FC64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Vector_opaque * x4, _Dcomplex *x5, struct GB_Descriptor_opaque * x6)
{
  GrB_Info result;
  { result = GxB_Vector_apply_BinaryOp2nd_FC64(x0, x1, x2, x3, x4, *x5, x6); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Vector_apply_IndexOp_FC32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, _Fcomplex x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_apply_IndexOp_FC32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_apply_IndexOp_FC32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  _Fcomplex x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_apply_IndexOp_FC32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x5, _cffi_type(974), arg5) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_apply_IndexOp_FC32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Vector_apply_IndexOp_FC32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, _Fcomplex *x5, struct GB_Descriptor_opaque * x6)
{
  GrB_Info result;
  { result = GxB_Vector_apply_IndexOp_FC32(x0, x1, x2, x3, x4, *x5, x6); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Vector_apply_IndexOp_FC64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, _Dcomplex x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_apply_IndexOp_FC64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_apply_IndexOp_FC64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  _Dcomplex x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_apply_IndexOp_FC64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x5, _cffi_type(962), arg5) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_apply_IndexOp_FC64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Vector_apply_IndexOp_FC64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, _Dcomplex *x5, struct GB_Descriptor_opaque * x6)
{
  GrB_Info result;
  { result = GxB_Vector_apply_IndexOp_FC64(x0, x1, x2, x3, x4, *x5, x6); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Vector_assign_FC32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, _Fcomplex x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_assign_FC32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_assign_FC32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  _Fcomplex x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_assign_FC32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x3, _cffi_type(974), arg3) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_assign_FC32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Vector_assign_FC32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, _Fcomplex *x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  GrB_Info result;
  { result = GxB_Vector_assign_FC32(x0, x1, x2, *x3, x4, x5, x6); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Vector_assign_FC64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, _Dcomplex x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_assign_FC64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_assign_FC64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  _Dcomplex x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_assign_FC64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x3, _cffi_type(962), arg3) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_assign_FC64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Vector_assign_FC64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, _Dcomplex *x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  GrB_Info result;
  { result = GxB_Vector_assign_FC64(x0, x1, x2, *x3, x4, x5, x6); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Vector_build_FC32(struct GB_Vector_opaque * x0, uint64_t const * x1, _Fcomplex const * x2, uint64_t x3, struct GB_BinaryOp_opaque * x4)
{
  return GxB_Vector_build_FC32(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_build_FC32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint64_t const * x1;
  _Fcomplex const * x2;
  uint64_t x3;
  struct GB_BinaryOp_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_build_FC32", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(737), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (_Fcomplex const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(737), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_build_FC32(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_build_FC32 _cffi_d_GxB_Vector_build_FC32
#endif

static GrB_Info _cffi_d_GxB_Vector_build_FC64(struct GB_Vector_opaque * x0, uint64_t const * x1, _Dcomplex const * x2, uint64_t x3, struct GB_BinaryOp_opaque * x4)
{
  return GxB_Vector_build_FC64(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_build_FC64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint64_t const * x1;
  _Dcomplex const * x2;
  uint64_t x3;
  struct GB_BinaryOp_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_build_FC64", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(711), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (_Dcomplex const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(711), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_build_FC64(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_build_FC64 _cffi_d_GxB_Vector_build_FC64
#endif

static GrB_Info _cffi_d_GxB_Vector_build_Scalar(struct GB_Vector_opaque * x0, uint64_t const * x1, struct GB_Scalar_opaque * x2, uint64_t x3)
{
  return GxB_Vector_build_Scalar(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_build_Scalar(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint64_t const * x1;
  struct GB_Scalar_opaque * x2;
  uint64_t x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_build_Scalar", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_build_Scalar(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_build_Scalar _cffi_d_GxB_Vector_build_Scalar
#endif

static GrB_Info _cffi_d_GxB_Vector_deserialize(struct GB_Vector_opaque * * x0, struct GB_Type_opaque * x1, void const * x2, uint64_t x3, struct GB_Descriptor_opaque * x4)
{
  return GxB_Vector_deserialize(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_deserialize(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * * x0;
  struct GB_Type_opaque * x1;
  void const * x2;
  uint64_t x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_deserialize", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2362), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2362), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(126), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (void const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(126), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_deserialize(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_deserialize _cffi_d_GxB_Vector_deserialize
#endif

static GrB_Info _cffi_d_GxB_Vector_diag(struct GB_Vector_opaque * x0, struct GB_Matrix_opaque * x1, int64_t x2, struct GB_Descriptor_opaque * x3)
{
  return GxB_Vector_diag(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_diag(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Matrix_opaque * x1;
  int64_t x2;
  struct GB_Descriptor_opaque * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_diag", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, int64_t);
  if (x2 == (int64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_diag(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_diag _cffi_d_GxB_Vector_diag
#endif

static GrB_Info _cffi_d_GxB_Vector_eWiseUnion(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_BinaryOp_opaque * x3, struct GB_Vector_opaque * x4, struct GB_Scalar_opaque * x5, struct GB_Vector_opaque * x6, struct GB_Scalar_opaque * x7, struct GB_Descriptor_opaque * x8)
{
  return GxB_Vector_eWiseUnion(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_eWiseUnion(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_BinaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  struct GB_Scalar_opaque * x5;
  struct GB_Vector_opaque * x6;
  struct GB_Scalar_opaque * x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_eWiseUnion", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_eWiseUnion(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_eWiseUnion _cffi_d_GxB_Vector_eWiseUnion
#endif

static GrB_Info _cffi_d_GxB_Vector_export_Bitmap(struct GB_Vector_opaque * * x0, struct GB_Type_opaque * * x1, uint64_t * x2, int8_t * * x3, void * * x4, uint64_t * x5, uint64_t * x6, _Bool * x7, uint64_t * x8, struct GB_Descriptor_opaque * x9)
{
  return GxB_Vector_export_Bitmap(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_export_Bitmap(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * * x0;
  struct GB_Type_opaque * * x1;
  uint64_t * x2;
  int8_t * * x3;
  void * * x4;
  uint64_t * x5;
  uint64_t * x6;
  _Bool * x7;
  uint64_t * x8;
  struct GB_Descriptor_opaque * x9;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_export_Bitmap", 10, 10, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2362), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2362), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(588), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (int8_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(588), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg9, (char **)&x9);
  if (datasize != 0) {
    x9 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg9, (char **)&x9,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_export_Bitmap(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_export_Bitmap _cffi_d_GxB_Vector_export_Bitmap
#endif

static GrB_Info _cffi_d_GxB_Vector_export_CSC(struct GB_Vector_opaque * * x0, struct GB_Type_opaque * * x1, uint64_t * x2, uint64_t * * x3, void * * x4, uint64_t * x5, uint64_t * x6, _Bool * x7, uint64_t * x8, _Bool * x9, struct GB_Descriptor_opaque * x10)
{
  return GxB_Vector_export_CSC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_export_CSC(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * * x0;
  struct GB_Type_opaque * * x1;
  uint64_t * x2;
  uint64_t * * x3;
  void * * x4;
  uint64_t * x5;
  uint64_t * x6;
  _Bool * x7;
  uint64_t * x8;
  _Bool * x9;
  struct GB_Descriptor_opaque * x10;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_export_CSC", 11, 11, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2362), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2362), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg9, (char **)&x9);
  if (datasize != 0) {
    x9 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg9, (char **)&x9,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg10, (char **)&x10);
  if (datasize != 0) {
    x10 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg10, (char **)&x10,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_export_CSC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_export_CSC _cffi_d_GxB_Vector_export_CSC
#endif

static GrB_Info _cffi_d_GxB_Vector_export_Full(struct GB_Vector_opaque * * x0, struct GB_Type_opaque * * x1, uint64_t * x2, void * * x3, uint64_t * x4, _Bool * x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_export_Full(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_export_Full(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * * x0;
  struct GB_Type_opaque * * x1;
  uint64_t * x2;
  void * * x3;
  uint64_t * x4;
  _Bool * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_export_Full", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2362), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2362), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_export_Full(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_export_Full _cffi_d_GxB_Vector_export_Full
#endif

static GrB_Info _cffi_d_GxB_Vector_extractElement_FC32(_Fcomplex * x0, struct GB_Vector_opaque * x1, uint64_t x2)
{
  return GxB_Vector_extractElement_FC32(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_extractElement_FC32(PyObject *self, PyObject *args)
{
  _Fcomplex * x0;
  struct GB_Vector_opaque * x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_extractElement_FC32", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(261), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (_Fcomplex *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(261), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_extractElement_FC32(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_extractElement_FC32 _cffi_d_GxB_Vector_extractElement_FC32
#endif

static GrB_Info _cffi_d_GxB_Vector_extractElement_FC64(_Dcomplex * x0, struct GB_Vector_opaque * x1, uint64_t x2)
{
  return GxB_Vector_extractElement_FC64(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_extractElement_FC64(PyObject *self, PyObject *args)
{
  _Dcomplex * x0;
  struct GB_Vector_opaque * x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_extractElement_FC64", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(203), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (_Dcomplex *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(203), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_extractElement_FC64(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_extractElement_FC64 _cffi_d_GxB_Vector_extractElement_FC64
#endif

static GrB_Info _cffi_d_GxB_Vector_extractTuples_FC32(uint64_t * x0, _Fcomplex * x1, uint64_t * x2, struct GB_Vector_opaque * x3)
{
  return GxB_Vector_extractTuples_FC32(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_extractTuples_FC32(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  _Fcomplex * x1;
  uint64_t * x2;
  struct GB_Vector_opaque * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_extractTuples_FC32", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(261), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (_Fcomplex *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(261), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_extractTuples_FC32(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_extractTuples_FC32 _cffi_d_GxB_Vector_extractTuples_FC32
#endif

static GrB_Info _cffi_d_GxB_Vector_extractTuples_FC64(uint64_t * x0, _Dcomplex * x1, uint64_t * x2, struct GB_Vector_opaque * x3)
{
  return GxB_Vector_extractTuples_FC64(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_extractTuples_FC64(PyObject *self, PyObject *args)
{
  uint64_t * x0;
  _Dcomplex * x1;
  uint64_t * x2;
  struct GB_Vector_opaque * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_extractTuples_FC64", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(203), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (_Dcomplex *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(203), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_extractTuples_FC64(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_extractTuples_FC64 _cffi_d_GxB_Vector_extractTuples_FC64
#endif

static GrB_Info _cffi_d_GxB_Vector_fprint(struct GB_Vector_opaque * x0, char const * x1, GxB_Print_Level x2, FILE * x3)
{
  return GxB_Vector_fprint(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_fprint(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  char const * x1;
  GxB_Print_Level x2;
  FILE * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_fprint", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(445), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (char const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(445), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x2, _cffi_type(463), arg2) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(464), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (FILE *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(464), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_fprint(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_fprint _cffi_d_GxB_Vector_fprint
#endif

static GrB_Info _cffi_d_GxB_Vector_import_Bitmap(struct GB_Vector_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, int8_t * * x3, void * * x4, uint64_t x5, uint64_t x6, _Bool x7, uint64_t x8, struct GB_Descriptor_opaque * x9)
{
  return GxB_Vector_import_Bitmap(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_import_Bitmap(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  int8_t * * x3;
  void * * x4;
  uint64_t x5;
  uint64_t x6;
  _Bool x7;
  uint64_t x8;
  struct GB_Descriptor_opaque * x9;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_import_Bitmap", 10, 10, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2362), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2362), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(588), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (int8_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(588), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x6 = _cffi_to_c_int(arg6, uint64_t);
  if (x6 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x7 = (_Bool)_cffi_to_c__Bool(arg7);
  if (x7 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  x8 = _cffi_to_c_int(arg8, uint64_t);
  if (x8 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg9, (char **)&x9);
  if (datasize != 0) {
    x9 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg9, (char **)&x9,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_import_Bitmap(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_import_Bitmap _cffi_d_GxB_Vector_import_Bitmap
#endif

static GrB_Info _cffi_d_GxB_Vector_import_CSC(struct GB_Vector_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, uint64_t * * x3, void * * x4, uint64_t x5, uint64_t x6, _Bool x7, uint64_t x8, _Bool x9, struct GB_Descriptor_opaque * x10)
{
  return GxB_Vector_import_CSC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_import_CSC(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  uint64_t * * x3;
  void * * x4;
  uint64_t x5;
  uint64_t x6;
  _Bool x7;
  uint64_t x8;
  _Bool x9;
  struct GB_Descriptor_opaque * x10;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;
  PyObject *arg9;
  PyObject *arg10;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_import_CSC", 11, 11, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8, &arg9, &arg10))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2362), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2362), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x6 = _cffi_to_c_int(arg6, uint64_t);
  if (x6 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x7 = (_Bool)_cffi_to_c__Bool(arg7);
  if (x7 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  x8 = _cffi_to_c_int(arg8, uint64_t);
  if (x8 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x9 = (_Bool)_cffi_to_c__Bool(arg9);
  if (x9 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg10, (char **)&x10);
  if (datasize != 0) {
    x10 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg10, (char **)&x10,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_import_CSC(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_import_CSC _cffi_d_GxB_Vector_import_CSC
#endif

static GrB_Info _cffi_d_GxB_Vector_import_Full(struct GB_Vector_opaque * * x0, struct GB_Type_opaque * x1, uint64_t x2, void * * x3, uint64_t x4, _Bool x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_import_Full(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_import_Full(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * * x0;
  struct GB_Type_opaque * x1;
  uint64_t x2;
  void * * x3;
  uint64_t x4;
  _Bool x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_import_Full", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2362), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2362), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(114), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(114), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x5 = (_Bool)_cffi_to_c__Bool(arg5);
  if (x5 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_import_Full(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_import_Full _cffi_d_GxB_Vector_import_Full
#endif

static GrB_Info _cffi_d_GxB_Vector_isStoredElement(struct GB_Vector_opaque * x0, uint64_t x1)
{
  return GxB_Vector_isStoredElement(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_isStoredElement(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint64_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_isStoredElement", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint64_t);
  if (x1 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_isStoredElement(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_isStoredElement _cffi_d_GxB_Vector_isStoredElement
#endif

static GrB_Info _cffi_d_GxB_Vector_iso(_Bool * x0, struct GB_Vector_opaque * x1)
{
  return GxB_Vector_iso(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_iso(PyObject *self, PyObject *args)
{
  _Bool * x0;
  struct GB_Vector_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_iso", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_iso(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_iso _cffi_d_GxB_Vector_iso
#endif

static GrB_Info _cffi_d_GxB_Vector_memoryUsage(size_t * x0, struct GB_Vector_opaque * x1)
{
  return GxB_Vector_memoryUsage(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_memoryUsage(PyObject *self, PyObject *args)
{
  size_t * x0;
  struct GB_Vector_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_memoryUsage", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(406), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (size_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(406), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_memoryUsage(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_memoryUsage _cffi_d_GxB_Vector_memoryUsage
#endif

static GrB_Info _cffi_d_GxB_Vector_pack_Bitmap(struct GB_Vector_opaque * x0, int8_t * * x1, void * * x2, uint64_t x3, uint64_t x4, _Bool x5, uint64_t x6, struct GB_Descriptor_opaque * x7)
{
  return GxB_Vector_pack_Bitmap(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_pack_Bitmap(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  int8_t * * x1;
  void * * x2;
  uint64_t x3;
  uint64_t x4;
  _Bool x5;
  uint64_t x6;
  struct GB_Descriptor_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_pack_Bitmap", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(588), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (int8_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(588), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x5 = (_Bool)_cffi_to_c__Bool(arg5);
  if (x5 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  x6 = _cffi_to_c_int(arg6, uint64_t);
  if (x6 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_pack_Bitmap(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_pack_Bitmap _cffi_d_GxB_Vector_pack_Bitmap
#endif

static GrB_Info _cffi_d_GxB_Vector_pack_CSC(struct GB_Vector_opaque * x0, uint64_t * * x1, void * * x2, uint64_t x3, uint64_t x4, _Bool x5, uint64_t x6, _Bool x7, struct GB_Descriptor_opaque * x8)
{
  return GxB_Vector_pack_CSC(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_pack_CSC(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint64_t * * x1;
  void * * x2;
  uint64_t x3;
  uint64_t x4;
  _Bool x5;
  uint64_t x6;
  _Bool x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_pack_CSC", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x4 = _cffi_to_c_int(arg4, uint64_t);
  if (x4 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x5 = (_Bool)_cffi_to_c__Bool(arg5);
  if (x5 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  x6 = _cffi_to_c_int(arg6, uint64_t);
  if (x6 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x7 = (_Bool)_cffi_to_c__Bool(arg7);
  if (x7 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_pack_CSC(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_pack_CSC _cffi_d_GxB_Vector_pack_CSC
#endif

static GrB_Info _cffi_d_GxB_Vector_pack_Full(struct GB_Vector_opaque * x0, void * * x1, uint64_t x2, _Bool x3, struct GB_Descriptor_opaque * x4)
{
  return GxB_Vector_pack_Full(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_pack_Full(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  void * * x1;
  uint64_t x2;
  _Bool x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_pack_Full", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  x3 = (_Bool)_cffi_to_c__Bool(arg3);
  if (x3 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_pack_Full(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_pack_Full _cffi_d_GxB_Vector_pack_Full
#endif

static GrB_Info _cffi_d_GxB_Vector_reduce_FC32(_Fcomplex * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Vector_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GxB_Vector_reduce_FC32(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_reduce_FC32(PyObject *self, PyObject *args)
{
  _Fcomplex * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Vector_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_reduce_FC32", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(261), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (_Fcomplex *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(261), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_reduce_FC32(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_reduce_FC32 _cffi_d_GxB_Vector_reduce_FC32
#endif

static GrB_Info _cffi_d_GxB_Vector_reduce_FC64(_Dcomplex * x0, struct GB_BinaryOp_opaque * x1, struct GB_Monoid_opaque * x2, struct GB_Vector_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GxB_Vector_reduce_FC64(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_reduce_FC64(PyObject *self, PyObject *args)
{
  _Dcomplex * x0;
  struct GB_BinaryOp_opaque * x1;
  struct GB_Monoid_opaque * x2;
  struct GB_Vector_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_reduce_FC64", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(203), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (_Dcomplex *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(203), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(57), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Monoid_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(57), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_reduce_FC64(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_reduce_FC64 _cffi_d_GxB_Vector_reduce_FC64
#endif

static GrB_Info _cffi_d_GxB_Vector_select(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_SelectOp_opaque * x3, struct GB_Vector_opaque * x4, struct GB_Scalar_opaque * x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_select(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_select(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_SelectOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  struct GB_Scalar_opaque * x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_select", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(155), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_SelectOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(155), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_select(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_select _cffi_d_GxB_Vector_select
#endif

static GrB_Info _cffi_d_GxB_Vector_select_FC32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, _Fcomplex x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_select_FC32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_select_FC32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  _Fcomplex x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_select_FC32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x5, _cffi_type(974), arg5) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_select_FC32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Vector_select_FC32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, _Fcomplex *x5, struct GB_Descriptor_opaque * x6)
{
  GrB_Info result;
  { result = GxB_Vector_select_FC32(x0, x1, x2, x3, x4, *x5, x6); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Vector_select_FC64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, _Dcomplex x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_select_FC64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_select_FC64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_IndexUnaryOp_opaque * x3;
  struct GB_Vector_opaque * x4;
  _Dcomplex x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_select_FC64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(102), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_IndexUnaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(102), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x5, _cffi_type(962), arg5) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_select_FC64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Vector_select_FC64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_IndexUnaryOp_opaque * x3, struct GB_Vector_opaque * x4, _Dcomplex *x5, struct GB_Descriptor_opaque * x6)
{
  GrB_Info result;
  { result = GxB_Vector_select_FC64(x0, x1, x2, x3, x4, *x5, x6); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Vector_serialize(void * * x0, uint64_t * x1, struct GB_Vector_opaque * x2, struct GB_Descriptor_opaque * x3)
{
  return GxB_Vector_serialize(x0, x1, x2, x3);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_serialize(PyObject *self, PyObject *args)
{
  void * * x0;
  uint64_t * x1;
  struct GB_Vector_opaque * x2;
  struct GB_Descriptor_opaque * x3;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_serialize", 4, 4, &arg0, &arg1, &arg2, &arg3))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_serialize(x0, x1, x2, x3); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_serialize _cffi_d_GxB_Vector_serialize
#endif

static GrB_Info _cffi_d_GxB_Vector_setElement_FC32(struct GB_Vector_opaque * x0, _Fcomplex x1, uint64_t x2)
{
  return GxB_Vector_setElement_FC32(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_setElement_FC32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  _Fcomplex x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_setElement_FC32", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(974), arg1) < 0)
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_setElement_FC32(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Vector_setElement_FC32(struct GB_Vector_opaque * x0, _Fcomplex *x1, uint64_t x2)
{
  GrB_Info result;
  { result = GxB_Vector_setElement_FC32(x0, *x1, x2); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Vector_setElement_FC64(struct GB_Vector_opaque * x0, _Dcomplex x1, uint64_t x2)
{
  return GxB_Vector_setElement_FC64(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_setElement_FC64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  _Dcomplex x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_setElement_FC64", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x1, _cffi_type(962), arg1) < 0)
    return NULL;

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_setElement_FC64(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Vector_setElement_FC64(struct GB_Vector_opaque * x0, _Dcomplex *x1, uint64_t x2)
{
  GrB_Info result;
  { result = GxB_Vector_setElement_FC64(x0, *x1, x2); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Vector_sort(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Vector_opaque * x3, struct GB_Descriptor_opaque * x4)
{
  return GxB_Vector_sort(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_sort(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Vector_opaque * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_sort", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_sort(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_sort _cffi_d_GxB_Vector_sort
#endif

static GrB_Info _cffi_d_GxB_Vector_subassign(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Vector_opaque * x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_subassign(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_subassign(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Vector_opaque * x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_subassign", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_subassign(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_subassign _cffi_d_GxB_Vector_subassign
#endif

static GrB_Info _cffi_d_GxB_Vector_subassign_BOOL(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, _Bool x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_subassign_BOOL(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_subassign_BOOL(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  _Bool x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_subassign_BOOL", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = (_Bool)_cffi_to_c__Bool(arg3);
  if (x3 == (_Bool)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_subassign_BOOL(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_subassign_BOOL _cffi_d_GxB_Vector_subassign_BOOL
#endif

static GrB_Info _cffi_d_GxB_Vector_subassign_FC32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, _Fcomplex x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_subassign_FC32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_subassign_FC32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  _Fcomplex x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_subassign_FC32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x3, _cffi_type(974), arg3) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_subassign_FC32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Vector_subassign_FC32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, _Fcomplex *x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  GrB_Info result;
  { result = GxB_Vector_subassign_FC32(x0, x1, x2, *x3, x4, x5, x6); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Vector_subassign_FC64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, _Dcomplex x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_subassign_FC64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_subassign_FC64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  _Dcomplex x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_subassign_FC64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  if (_cffi_to_c((char *)&x3, _cffi_type(962), arg3) < 0)
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_subassign_FC64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
static GrB_Info _cffi_f_GxB_Vector_subassign_FC64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, _Dcomplex *x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  GrB_Info result;
  { result = GxB_Vector_subassign_FC64(x0, x1, x2, *x3, x4, x5, x6); }
  return result;
}
#endif

static GrB_Info _cffi_d_GxB_Vector_subassign_FP32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, float x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_subassign_FP32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_subassign_FP32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  float x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_subassign_FP32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = (float)_cffi_to_c_float(arg3);
  if (x3 == (float)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_subassign_FP32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_subassign_FP32 _cffi_d_GxB_Vector_subassign_FP32
#endif

static GrB_Info _cffi_d_GxB_Vector_subassign_FP64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, double x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_subassign_FP64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_subassign_FP64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  double x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_subassign_FP64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = (double)_cffi_to_c_double(arg3);
  if (x3 == (double)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_subassign_FP64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_subassign_FP64 _cffi_d_GxB_Vector_subassign_FP64
#endif

static GrB_Info _cffi_d_GxB_Vector_subassign_INT16(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, int16_t x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_subassign_INT16(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_subassign_INT16(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  int16_t x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_subassign_INT16", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, int16_t);
  if (x3 == (int16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_subassign_INT16(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_subassign_INT16 _cffi_d_GxB_Vector_subassign_INT16
#endif

static GrB_Info _cffi_d_GxB_Vector_subassign_INT32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, int32_t x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_subassign_INT32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_subassign_INT32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  int32_t x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_subassign_INT32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, int32_t);
  if (x3 == (int32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_subassign_INT32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_subassign_INT32 _cffi_d_GxB_Vector_subassign_INT32
#endif

static GrB_Info _cffi_d_GxB_Vector_subassign_INT64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, int64_t x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_subassign_INT64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_subassign_INT64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  int64_t x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_subassign_INT64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, int64_t);
  if (x3 == (int64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_subassign_INT64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_subassign_INT64 _cffi_d_GxB_Vector_subassign_INT64
#endif

static GrB_Info _cffi_d_GxB_Vector_subassign_INT8(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, int8_t x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_subassign_INT8(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_subassign_INT8(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  int8_t x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_subassign_INT8", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, int8_t);
  if (x3 == (int8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_subassign_INT8(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_subassign_INT8 _cffi_d_GxB_Vector_subassign_INT8
#endif

static GrB_Info _cffi_d_GxB_Vector_subassign_Scalar(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, struct GB_Scalar_opaque * x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_subassign_Scalar(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_subassign_Scalar(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  struct GB_Scalar_opaque * x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_subassign_Scalar", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(80), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (struct GB_Scalar_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(80), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_subassign_Scalar(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_subassign_Scalar _cffi_d_GxB_Vector_subassign_Scalar
#endif

static GrB_Info _cffi_d_GxB_Vector_subassign_UDT(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, void * x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_subassign_UDT(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_subassign_UDT(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  void * x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_subassign_UDT", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(52), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (void *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(52), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_subassign_UDT(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_subassign_UDT _cffi_d_GxB_Vector_subassign_UDT
#endif

static GrB_Info _cffi_d_GxB_Vector_subassign_UINT16(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, uint16_t x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_subassign_UINT16(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_subassign_UINT16(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  uint16_t x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_subassign_UINT16", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint16_t);
  if (x3 == (uint16_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_subassign_UINT16(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_subassign_UINT16 _cffi_d_GxB_Vector_subassign_UINT16
#endif

static GrB_Info _cffi_d_GxB_Vector_subassign_UINT32(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, uint32_t x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_subassign_UINT32(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_subassign_UINT32(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  uint32_t x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_subassign_UINT32", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint32_t);
  if (x3 == (uint32_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_subassign_UINT32(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_subassign_UINT32 _cffi_d_GxB_Vector_subassign_UINT32
#endif

static GrB_Info _cffi_d_GxB_Vector_subassign_UINT64(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, uint64_t x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_subassign_UINT64(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_subassign_UINT64(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  uint64_t x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_subassign_UINT64", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint64_t);
  if (x3 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_subassign_UINT64(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_subassign_UINT64 _cffi_d_GxB_Vector_subassign_UINT64
#endif

static GrB_Info _cffi_d_GxB_Vector_subassign_UINT8(struct GB_Vector_opaque * x0, struct GB_Vector_opaque * x1, struct GB_BinaryOp_opaque * x2, uint8_t x3, uint64_t const * x4, uint64_t x5, struct GB_Descriptor_opaque * x6)
{
  return GxB_Vector_subassign_UINT8(x0, x1, x2, x3, x4, x5, x6);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_subassign_UINT8(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  struct GB_Vector_opaque * x1;
  struct GB_BinaryOp_opaque * x2;
  uint8_t x3;
  uint64_t const * x4;
  uint64_t x5;
  struct GB_Descriptor_opaque * x6;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_subassign_UINT8", 7, 7, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(56), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_BinaryOp_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(56), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x3 = _cffi_to_c_int(arg3, uint8_t);
  if (x3 == (uint8_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(696), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(696), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x5 = _cffi_to_c_int(arg5, uint64_t);
  if (x5 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_subassign_UINT8(x0, x1, x2, x3, x4, x5, x6); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_subassign_UINT8 _cffi_d_GxB_Vector_subassign_UINT8
#endif

static GrB_Info _cffi_d_GxB_Vector_type(struct GB_Type_opaque * * x0, struct GB_Vector_opaque * x1)
{
  return GxB_Vector_type(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_type(PyObject *self, PyObject *args)
{
  struct GB_Type_opaque * * x0;
  struct GB_Vector_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_type", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(585), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Type_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(585), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_type(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_type _cffi_d_GxB_Vector_type
#endif

static GrB_Info _cffi_d_GxB_Vector_type_name(char * x0, struct GB_Vector_opaque * x1)
{
  return GxB_Vector_type_name(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_type_name(PyObject *self, PyObject *args)
{
  char * x0;
  struct GB_Vector_opaque * x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_type_name", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(97), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(97), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_type_name(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_type_name _cffi_d_GxB_Vector_type_name
#endif

static GrB_Info _cffi_d_GxB_Vector_unpack_Bitmap(struct GB_Vector_opaque * x0, int8_t * * x1, void * * x2, uint64_t * x3, uint64_t * x4, _Bool * x5, uint64_t * x6, struct GB_Descriptor_opaque * x7)
{
  return GxB_Vector_unpack_Bitmap(x0, x1, x2, x3, x4, x5, x6, x7);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_unpack_Bitmap(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  int8_t * * x1;
  void * * x2;
  uint64_t * x3;
  uint64_t * x4;
  _Bool * x5;
  uint64_t * x6;
  struct GB_Descriptor_opaque * x7;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_unpack_Bitmap", 8, 8, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(588), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (int8_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(588), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_unpack_Bitmap(x0, x1, x2, x3, x4, x5, x6, x7); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_unpack_Bitmap _cffi_d_GxB_Vector_unpack_Bitmap
#endif

static GrB_Info _cffi_d_GxB_Vector_unpack_CSC(struct GB_Vector_opaque * x0, uint64_t * * x1, void * * x2, uint64_t * x3, uint64_t * x4, _Bool * x5, uint64_t * x6, _Bool * x7, struct GB_Descriptor_opaque * x8)
{
  return GxB_Vector_unpack_CSC(x0, x1, x2, x3, x4, x5, x6, x7, x8);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_unpack_CSC(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  uint64_t * * x1;
  void * * x2;
  uint64_t * x3;
  uint64_t * x4;
  _Bool * x5;
  uint64_t * x6;
  _Bool * x7;
  struct GB_Descriptor_opaque * x8;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;
  PyObject *arg5;
  PyObject *arg6;
  PyObject *arg7;
  PyObject *arg8;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_unpack_CSC", 9, 9, &arg0, &arg1, &arg2, &arg3, &arg4, &arg5, &arg6, &arg7, &arg8))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(601), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (uint64_t * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(601), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg5, (char **)&x5);
  if (datasize != 0) {
    x5 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg5, (char **)&x5,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg6, (char **)&x6);
  if (datasize != 0) {
    x6 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg6, (char **)&x6,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg7, (char **)&x7);
  if (datasize != 0) {
    x7 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg7, (char **)&x7,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg8, (char **)&x8);
  if (datasize != 0) {
    x8 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg8, (char **)&x8,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_unpack_CSC(x0, x1, x2, x3, x4, x5, x6, x7, x8); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_unpack_CSC _cffi_d_GxB_Vector_unpack_CSC
#endif

static GrB_Info _cffi_d_GxB_Vector_unpack_Full(struct GB_Vector_opaque * x0, void * * x1, uint64_t * x2, _Bool * x3, struct GB_Descriptor_opaque * x4)
{
  return GxB_Vector_unpack_Full(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_Vector_unpack_Full(PyObject *self, PyObject *args)
{
  struct GB_Vector_opaque * x0;
  void * * x1;
  uint64_t * x2;
  _Bool * x3;
  struct GB_Descriptor_opaque * x4;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GxB_Vector_unpack_Full", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(65), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Vector_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(65), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(48), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (void * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(48), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(586), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (uint64_t *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(586), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(55), arg3, (char **)&x3);
  if (datasize != 0) {
    x3 = ((size_t)datasize) <= 640 ? (_Bool *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(55), arg3, (char **)&x3,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg4, (char **)&x4);
  if (datasize != 0) {
    x4 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg4, (char **)&x4,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_Vector_unpack_Full(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_Vector_unpack_Full _cffi_d_GxB_Vector_unpack_Full
#endif

static GrB_Info _cffi_d_GxB_colIterator_attach(struct GB_Iterator_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_Descriptor_opaque * x2)
{
  return GxB_colIterator_attach(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_colIterator_attach(PyObject *self, PyObject *args)
{
  struct GB_Iterator_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_Descriptor_opaque * x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_colIterator_attach", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_colIterator_attach(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_colIterator_attach _cffi_d_GxB_colIterator_attach
#endif

static uint64_t _cffi_d_GxB_colIterator_getColIndex(struct GB_Iterator_opaque * x0)
{
  return GxB_colIterator_getColIndex(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_colIterator_getColIndex(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  uint64_t result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_colIterator_getColIndex(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, uint64_t);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_colIterator_getColIndex _cffi_d_GxB_colIterator_getColIndex
#endif

static uint64_t _cffi_d_GxB_colIterator_getRowIndex(struct GB_Iterator_opaque * x0)
{
  return GxB_colIterator_getRowIndex(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_colIterator_getRowIndex(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  uint64_t result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_colIterator_getRowIndex(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, uint64_t);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_colIterator_getRowIndex _cffi_d_GxB_colIterator_getRowIndex
#endif

static uint64_t _cffi_d_GxB_colIterator_kount(struct GB_Iterator_opaque * x0)
{
  return GxB_colIterator_kount(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_colIterator_kount(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  uint64_t result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_colIterator_kount(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, uint64_t);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_colIterator_kount _cffi_d_GxB_colIterator_kount
#endif

static GrB_Info _cffi_d_GxB_colIterator_kseek(struct GB_Iterator_opaque * x0, uint64_t x1)
{
  return GxB_colIterator_kseek(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_colIterator_kseek(PyObject *self, PyObject *args)
{
  struct GB_Iterator_opaque * x0;
  uint64_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_colIterator_kseek", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint64_t);
  if (x1 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_colIterator_kseek(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_colIterator_kseek _cffi_d_GxB_colIterator_kseek
#endif

static GrB_Info _cffi_d_GxB_colIterator_nextCol(struct GB_Iterator_opaque * x0)
{
  return GxB_colIterator_nextCol(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_colIterator_nextCol(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_colIterator_nextCol(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_colIterator_nextCol _cffi_d_GxB_colIterator_nextCol
#endif

static GrB_Info _cffi_d_GxB_colIterator_nextRow(struct GB_Iterator_opaque * x0)
{
  return GxB_colIterator_nextRow(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_colIterator_nextRow(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_colIterator_nextRow(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_colIterator_nextRow _cffi_d_GxB_colIterator_nextRow
#endif

static GrB_Info _cffi_d_GxB_colIterator_seekCol(struct GB_Iterator_opaque * x0, uint64_t x1)
{
  return GxB_colIterator_seekCol(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_colIterator_seekCol(PyObject *self, PyObject *args)
{
  struct GB_Iterator_opaque * x0;
  uint64_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_colIterator_seekCol", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint64_t);
  if (x1 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_colIterator_seekCol(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_colIterator_seekCol _cffi_d_GxB_colIterator_seekCol
#endif

static GrB_Info _cffi_d_GxB_deserialize_type_name(char * x0, void const * x1, uint64_t x2)
{
  return GxB_deserialize_type_name(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_deserialize_type_name(PyObject *self, PyObject *args)
{
  char * x0;
  void const * x1;
  uint64_t x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_deserialize_type_name", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(97), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (char *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(97), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(126), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (void const *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(126), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x2 = _cffi_to_c_int(arg2, uint64_t);
  if (x2 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_deserialize_type_name(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_deserialize_type_name _cffi_d_GxB_deserialize_type_name
#endif

static GrB_Info _cffi_d_GxB_init(GrB_Mode x0, void *(* x1)(size_t), void *(* x2)(size_t, size_t), void *(* x3)(void *, size_t), void(* x4)(void *))
{
  return GxB_init(x0, x1, x2, x3, x4);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_init(PyObject *self, PyObject *args)
{
  GrB_Mode x0;
  void *(* x1)(size_t);
  void *(* x2)(size_t, size_t);
  void *(* x3)(void *, size_t);
  void(* x4)(void *);
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;
  PyObject *arg3;
  PyObject *arg4;

  if (!PyArg_UnpackTuple(args, "GxB_init", 5, 5, &arg0, &arg1, &arg2, &arg3, &arg4))
    return NULL;

  if (_cffi_to_c((char *)&x0, _cffi_type(10), arg0) < 0)
    return NULL;

  x1 = (void *(*)(size_t))_cffi_to_c_pointer(arg1, _cffi_type(14));
  if (x1 == (void *(*)(size_t))NULL && PyErr_Occurred())
    return NULL;

  x2 = (void *(*)(size_t, size_t))_cffi_to_c_pointer(arg2, _cffi_type(15));
  if (x2 == (void *(*)(size_t, size_t))NULL && PyErr_Occurred())
    return NULL;

  x3 = (void *(*)(void *, size_t))_cffi_to_c_pointer(arg3, _cffi_type(16));
  if (x3 == (void *(*)(void *, size_t))NULL && PyErr_Occurred())
    return NULL;

  x4 = (void(*)(void *))_cffi_to_c_pointer(arg4, _cffi_type(17));
  if (x4 == (void(*)(void *))NULL && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_init(x0, x1, x2, x3, x4); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  return pyresult;
}
#else
#  define _cffi_f_GxB_init _cffi_d_GxB_init
#endif

static GrB_Info _cffi_d_GxB_pack_HyperHash(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * * x1, struct GB_Descriptor_opaque * x2)
{
  return GxB_pack_HyperHash(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_pack_HyperHash(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * * x1;
  struct GB_Descriptor_opaque * x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_pack_HyperHash", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_pack_HyperHash(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_pack_HyperHash _cffi_d_GxB_pack_HyperHash
#endif

static GrB_Info _cffi_d_GxB_rowIterator_attach(struct GB_Iterator_opaque * x0, struct GB_Matrix_opaque * x1, struct GB_Descriptor_opaque * x2)
{
  return GxB_rowIterator_attach(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_rowIterator_attach(PyObject *self, PyObject *args)
{
  struct GB_Iterator_opaque * x0;
  struct GB_Matrix_opaque * x1;
  struct GB_Descriptor_opaque * x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_rowIterator_attach", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_rowIterator_attach(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_rowIterator_attach _cffi_d_GxB_rowIterator_attach
#endif

static uint64_t _cffi_d_GxB_rowIterator_getColIndex(struct GB_Iterator_opaque * x0)
{
  return GxB_rowIterator_getColIndex(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_rowIterator_getColIndex(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  uint64_t result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_rowIterator_getColIndex(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, uint64_t);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_rowIterator_getColIndex _cffi_d_GxB_rowIterator_getColIndex
#endif

static uint64_t _cffi_d_GxB_rowIterator_getRowIndex(struct GB_Iterator_opaque * x0)
{
  return GxB_rowIterator_getRowIndex(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_rowIterator_getRowIndex(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  uint64_t result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_rowIterator_getRowIndex(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, uint64_t);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_rowIterator_getRowIndex _cffi_d_GxB_rowIterator_getRowIndex
#endif

static uint64_t _cffi_d_GxB_rowIterator_kount(struct GB_Iterator_opaque * x0)
{
  return GxB_rowIterator_kount(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_rowIterator_kount(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  uint64_t result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_rowIterator_kount(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_int(result, uint64_t);
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_rowIterator_kount _cffi_d_GxB_rowIterator_kount
#endif

static GrB_Info _cffi_d_GxB_rowIterator_kseek(struct GB_Iterator_opaque * x0, uint64_t x1)
{
  return GxB_rowIterator_kseek(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_rowIterator_kseek(PyObject *self, PyObject *args)
{
  struct GB_Iterator_opaque * x0;
  uint64_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_rowIterator_kseek", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint64_t);
  if (x1 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_rowIterator_kseek(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_rowIterator_kseek _cffi_d_GxB_rowIterator_kseek
#endif

static GrB_Info _cffi_d_GxB_rowIterator_nextCol(struct GB_Iterator_opaque * x0)
{
  return GxB_rowIterator_nextCol(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_rowIterator_nextCol(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_rowIterator_nextCol(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_rowIterator_nextCol _cffi_d_GxB_rowIterator_nextCol
#endif

static GrB_Info _cffi_d_GxB_rowIterator_nextRow(struct GB_Iterator_opaque * x0)
{
  return GxB_rowIterator_nextRow(x0);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_rowIterator_nextRow(PyObject *self, PyObject *arg0)
{
  struct GB_Iterator_opaque * x0;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_rowIterator_nextRow(x0); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_rowIterator_nextRow _cffi_d_GxB_rowIterator_nextRow
#endif

static GrB_Info _cffi_d_GxB_rowIterator_seekRow(struct GB_Iterator_opaque * x0, uint64_t x1)
{
  return GxB_rowIterator_seekRow(x0, x1);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_rowIterator_seekRow(PyObject *self, PyObject *args)
{
  struct GB_Iterator_opaque * x0;
  uint64_t x1;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;

  if (!PyArg_UnpackTuple(args, "GxB_rowIterator_seekRow", 2, 2, &arg0, &arg1))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(541), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Iterator_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(541), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  x1 = _cffi_to_c_int(arg1, uint64_t);
  if (x1 == (uint64_t)-1 && PyErr_Occurred())
    return NULL;

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_rowIterator_seekRow(x0, x1); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_rowIterator_seekRow _cffi_d_GxB_rowIterator_seekRow
#endif

static GrB_Info _cffi_d_GxB_unpack_HyperHash(struct GB_Matrix_opaque * x0, struct GB_Matrix_opaque * * x1, struct GB_Descriptor_opaque * x2)
{
  return GxB_unpack_HyperHash(x0, x1, x2);
}
#ifndef PYPY_VERSION
static PyObject *
_cffi_f_GxB_unpack_HyperHash(PyObject *self, PyObject *args)
{
  struct GB_Matrix_opaque * x0;
  struct GB_Matrix_opaque * * x1;
  struct GB_Descriptor_opaque * x2;
  Py_ssize_t datasize;
  struct _cffi_freeme_s *large_args_free = NULL;
  GrB_Info result;
  PyObject *pyresult;
  PyObject *arg0;
  PyObject *arg1;
  PyObject *arg2;

  if (!PyArg_UnpackTuple(args, "GxB_unpack_HyperHash", 3, 3, &arg0, &arg1, &arg2))
    return NULL;

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(7), arg0, (char **)&x0);
  if (datasize != 0) {
    x0 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(7), arg0, (char **)&x0,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(569), arg1, (char **)&x1);
  if (datasize != 0) {
    x1 = ((size_t)datasize) <= 640 ? (struct GB_Matrix_opaque * *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(569), arg1, (char **)&x1,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  datasize = _cffi_prepare_pointer_call_argument(
      _cffi_type(2), arg2, (char **)&x2);
  if (datasize != 0) {
    x2 = ((size_t)datasize) <= 640 ? (struct GB_Descriptor_opaque *)alloca((size_t)datasize) : NULL;
    if (_cffi_convert_array_argument(_cffi_type(2), arg2, (char **)&x2,
            datasize, &large_args_free) < 0)
      return NULL;
  }

  Py_BEGIN_ALLOW_THREADS
  _cffi_restore_errno();
  { result = GxB_unpack_HyperHash(x0, x1, x2); }
  _cffi_save_errno();
  Py_END_ALLOW_THREADS

  (void)self; /* unused */
  pyresult = _cffi_from_c_deref((char *)&result, _cffi_type(3983));
  if (large_args_free != NULL) _cffi_free_array_arguments(large_args_free);
  return pyresult;
}
#else
#  define _cffi_f_GxB_unpack_HyperHash _cffi_d_GxB_unpack_HyperHash
#endif

static int _cffi_const_GRB_SUBVERSION(unsigned long long *o)
{
  int n = (GRB_SUBVERSION) <= 0;
  *o = (unsigned long long)((GRB_SUBVERSION) | 0);  /* check that GRB_SUBVERSION is an integer */
  return n;
}

static int _cffi_const_GRB_VERSION(unsigned long long *o)
{
  int n = (GRB_VERSION) <= 0;
  *o = (unsigned long long)((GRB_VERSION) | 0);  /* check that GRB_VERSION is an integer */
  return n;
}

static int _cffi_const_GrB_INDEX_MAX(unsigned long long *o)
{
  int n = (GrB_INDEX_MAX) <= 0;
  *o = (unsigned long long)((GrB_INDEX_MAX) | 0);  /* check that GrB_INDEX_MAX is an integer */
  return n;
}

static int _cffi_const_GxB_ANY_SPARSITY(unsigned long long *o)
{
  int n = (GxB_ANY_SPARSITY) <= 0;
  *o = (unsigned long long)((GxB_ANY_SPARSITY) | 0);  /* check that GxB_ANY_SPARSITY is an integer */
  return n;
}

static int _cffi_const_GxB_AUTO_SPARSITY(unsigned long long *o)
{
  int n = (GxB_AUTO_SPARSITY) <= 0;
  *o = (unsigned long long)((GxB_AUTO_SPARSITY) | 0);  /* check that GxB_AUTO_SPARSITY is an integer */
  return n;
}

static int _cffi_const_GxB_BACKWARDS(unsigned long long *o)
{
  int n = (GxB_BACKWARDS) <= 0;
  *o = (unsigned long long)((GxB_BACKWARDS) | 0);  /* check that GxB_BACKWARDS is an integer */
  return n;
}

static int _cffi_const_GxB_BEGIN(unsigned long long *o)
{
  int n = (GxB_BEGIN) <= 0;
  *o = (unsigned long long)((GxB_BEGIN) | 0);  /* check that GxB_BEGIN is an integer */
  return n;
}

static int _cffi_const_GxB_BITMAP(unsigned long long *o)
{
  int n = (GxB_BITMAP) <= 0;
  *o = (unsigned long long)((GxB_BITMAP) | 0);  /* check that GxB_BITMAP is an integer */
  return n;
}

static int _cffi_const_GxB_CHUNK(unsigned long long *o)
{
  int n = (GxB_CHUNK) <= 0;
  *o = (unsigned long long)((GxB_CHUNK) | 0);  /* check that GxB_CHUNK is an integer */
  return n;
}

static int _cffi_const_GxB_COMPRESSION_DEFAULT(unsigned long long *o)
{
  int n = (GxB_COMPRESSION_DEFAULT) <= 0;
  *o = (unsigned long long)((GxB_COMPRESSION_DEFAULT) | 0);  /* check that GxB_COMPRESSION_DEFAULT is an integer */
  return n;
}

static int _cffi_const_GxB_COMPRESSION_INTEL(unsigned long long *o)
{
  int n = (GxB_COMPRESSION_INTEL) <= 0;
  *o = (unsigned long long)((GxB_COMPRESSION_INTEL) | 0);  /* check that GxB_COMPRESSION_INTEL is an integer */
  return n;
}

static int _cffi_const_GxB_COMPRESSION_LZ4(unsigned long long *o)
{
  int n = (GxB_COMPRESSION_LZ4) <= 0;
  *o = (unsigned long long)((GxB_COMPRESSION_LZ4) | 0);  /* check that GxB_COMPRESSION_LZ4 is an integer */
  return n;
}

static int _cffi_const_GxB_COMPRESSION_LZ4HC(unsigned long long *o)
{
  int n = (GxB_COMPRESSION_LZ4HC) <= 0;
  *o = (unsigned long long)((GxB_COMPRESSION_LZ4HC) | 0);  /* check that GxB_COMPRESSION_LZ4HC is an integer */
  return n;
}

static int _cffi_const_GxB_COMPRESSION_NONE(unsigned long long *o)
{
  int n = (GxB_COMPRESSION_NONE) <= 0;
  *o = (unsigned long long)((GxB_COMPRESSION_NONE) | 0);  /* check that GxB_COMPRESSION_NONE is an integer */
  return n;
}

static int _cffi_const_GxB_COMPRESSION_ZSTD(unsigned long long *o)
{
  int n = (GxB_COMPRESSION_ZSTD) <= 0;
  *o = (unsigned long long)((GxB_COMPRESSION_ZSTD) | 0);  /* check that GxB_COMPRESSION_ZSTD is an integer */
  return n;
}

static int _cffi_const_GxB_END(unsigned long long *o)
{
  int n = (GxB_END) <= 0;
  *o = (unsigned long long)((GxB_END) | 0);  /* check that GxB_END is an integer */
  return n;
}

static int _cffi_const_GxB_FAST_IMPORT(unsigned long long *o)
{
  int n = (GxB_FAST_IMPORT) <= 0;
  *o = (unsigned long long)((GxB_FAST_IMPORT) | 0);  /* check that GxB_FAST_IMPORT is an integer */
  return n;
}

static int _cffi_const_GxB_FULL(unsigned long long *o)
{
  int n = (GxB_FULL) <= 0;
  *o = (unsigned long long)((GxB_FULL) | 0);  /* check that GxB_FULL is an integer */
  return n;
}

static int _cffi_const_GxB_GPU_CHUNK(unsigned long long *o)
{
  int n = (GxB_GPU_CHUNK) <= 0;
  *o = (unsigned long long)((GxB_GPU_CHUNK) | 0);  /* check that GxB_GPU_CHUNK is an integer */
  return n;
}

static int _cffi_const_GxB_GPU_CONTROL(unsigned long long *o)
{
  int n = (GxB_GPU_CONTROL) <= 0;
  *o = (unsigned long long)((GxB_GPU_CONTROL) | 0);  /* check that GxB_GPU_CONTROL is an integer */
  return n;
}

static int _cffi_const_GxB_HYPERSPARSE(unsigned long long *o)
{
  int n = (GxB_HYPERSPARSE) <= 0;
  *o = (unsigned long long)((GxB_HYPERSPARSE) | 0);  /* check that GxB_HYPERSPARSE is an integer */
  return n;
}

static int _cffi_const_GxB_IMPLEMENTATION(unsigned long long *o)
{
  int n = (GxB_IMPLEMENTATION) <= 0;
  *o = (unsigned long long)((GxB_IMPLEMENTATION) | 0);  /* check that GxB_IMPLEMENTATION is an integer */
  return n;
}

static int _cffi_const_GxB_IMPLEMENTATION_MAJOR(unsigned long long *o)
{
  int n = (GxB_IMPLEMENTATION_MAJOR) <= 0;
  *o = (unsigned long long)((GxB_IMPLEMENTATION_MAJOR) | 0);  /* check that GxB_IMPLEMENTATION_MAJOR is an integer */
  return n;
}

static int _cffi_const_GxB_IMPLEMENTATION_MINOR(unsigned long long *o)
{
  int n = (GxB_IMPLEMENTATION_MINOR) <= 0;
  *o = (unsigned long long)((GxB_IMPLEMENTATION_MINOR) | 0);  /* check that GxB_IMPLEMENTATION_MINOR is an integer */
  return n;
}

static int _cffi_const_GxB_IMPLEMENTATION_SUB(unsigned long long *o)
{
  int n = (GxB_IMPLEMENTATION_SUB) <= 0;
  *o = (unsigned long long)((GxB_IMPLEMENTATION_SUB) | 0);  /* check that GxB_IMPLEMENTATION_SUB is an integer */
  return n;
}

static int _cffi_const_GxB_INC(unsigned long long *o)
{
  int n = (GxB_INC) <= 0;
  *o = (unsigned long long)((GxB_INC) | 0);  /* check that GxB_INC is an integer */
  return n;
}

static int _cffi_const_GxB_INDEX_MAX(unsigned long long *o)
{
  int n = (GxB_INDEX_MAX) <= 0;
  *o = (unsigned long long)((GxB_INDEX_MAX) | 0);  /* check that GxB_INDEX_MAX is an integer */
  return n;
}

static int _cffi_const_GxB_MAX_NAME_LEN(unsigned long long *o)
{
  int n = (GxB_MAX_NAME_LEN) <= 0;
  *o = (unsigned long long)((GxB_MAX_NAME_LEN) | 0);  /* check that GxB_MAX_NAME_LEN is an integer */
  return n;
}

static int _cffi_const_GxB_NBITMAP_SWITCH(unsigned long long *o)
{
  int n = (GxB_NBITMAP_SWITCH) <= 0;
  *o = (unsigned long long)((GxB_NBITMAP_SWITCH) | 0);  /* check that GxB_NBITMAP_SWITCH is an integer */
  return n;
}

static int _cffi_const_GxB_NTHREADS(unsigned long long *o)
{
  int n = (GxB_NTHREADS) <= 0;
  *o = (unsigned long long)((GxB_NTHREADS) | 0);  /* check that GxB_NTHREADS is an integer */
  return n;
}

static int _cffi_const_GxB_RANGE(unsigned long long *o)
{
  int n = (GxB_RANGE) <= 0;
  *o = (unsigned long long)((GxB_RANGE) | 0);  /* check that GxB_RANGE is an integer */
  return n;
}

static int _cffi_const_GxB_SPARSE(unsigned long long *o)
{
  int n = (GxB_SPARSE) <= 0;
  *o = (unsigned long long)((GxB_SPARSE) | 0);  /* check that GxB_SPARSE is an integer */
  return n;
}

static int _cffi_const_GxB_SPEC_MAJOR(unsigned long long *o)
{
  int n = (GxB_SPEC_MAJOR) <= 0;
  *o = (unsigned long long)((GxB_SPEC_MAJOR) | 0);  /* check that GxB_SPEC_MAJOR is an integer */
  return n;
}

static int _cffi_const_GxB_SPEC_MINOR(unsigned long long *o)
{
  int n = (GxB_SPEC_MINOR) <= 0;
  *o = (unsigned long long)((GxB_SPEC_MINOR) | 0);  /* check that GxB_SPEC_MINOR is an integer */
  return n;
}

static int _cffi_const_GxB_SPEC_SUB(unsigned long long *o)
{
  int n = (GxB_SPEC_SUB) <= 0;
  *o = (unsigned long long)((GxB_SPEC_SUB) | 0);  /* check that GxB_SPEC_SUB is an integer */
  return n;
}

static int _cffi_const_GxB_SPEC_VERSION(unsigned long long *o)
{
  int n = (GxB_SPEC_VERSION) <= 0;
  *o = (unsigned long long)((GxB_SPEC_VERSION) | 0);  /* check that GxB_SPEC_VERSION is an integer */
  return n;
}

static int _cffi_const_GxB_STDC_VERSION(unsigned long long *o)
{
  int n = (GxB_STDC_VERSION) <= 0;
  *o = (unsigned long long)((GxB_STDC_VERSION) | 0);  /* check that GxB_STDC_VERSION is an integer */
  return n;
}

static int _cffi_const_GxB_STRIDE(unsigned long long *o)
{
  int n = (GxB_STRIDE) <= 0;
  *o = (unsigned long long)((GxB_STRIDE) | 0);  /* check that GxB_STRIDE is an integer */
  return n;
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_ABS_BOOL(void)
{
  return &(GrB_ABS_BOOL);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_ABS_FP32(void)
{
  return &(GrB_ABS_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_ABS_FP64(void)
{
  return &(GrB_ABS_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_ABS_INT16(void)
{
  return &(GrB_ABS_INT16);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_ABS_INT32(void)
{
  return &(GrB_ABS_INT32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_ABS_INT64(void)
{
  return &(GrB_ABS_INT64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_ABS_INT8(void)
{
  return &(GrB_ABS_INT8);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_ABS_UINT16(void)
{
  return &(GrB_ABS_UINT16);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_ABS_UINT32(void)
{
  return &(GrB_ABS_UINT32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_ABS_UINT64(void)
{
  return &(GrB_ABS_UINT64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_ABS_UINT8(void)
{
  return &(GrB_ABS_UINT8);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_AINV_BOOL(void)
{
  return &(GrB_AINV_BOOL);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_AINV_FP32(void)
{
  return &(GrB_AINV_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_AINV_FP64(void)
{
  return &(GrB_AINV_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_AINV_INT16(void)
{
  return &(GrB_AINV_INT16);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_AINV_INT32(void)
{
  return &(GrB_AINV_INT32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_AINV_INT64(void)
{
  return &(GrB_AINV_INT64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_AINV_INT8(void)
{
  return &(GrB_AINV_INT8);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_AINV_UINT16(void)
{
  return &(GrB_AINV_UINT16);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_AINV_UINT32(void)
{
  return &(GrB_AINV_UINT32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_AINV_UINT64(void)
{
  return &(GrB_AINV_UINT64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_AINV_UINT8(void)
{
  return &(GrB_AINV_UINT8);
}

static uint64_t const * *_cffi_var_GrB_ALL(void)
{
  return &(GrB_ALL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BAND_INT16(void)
{
  return &(GrB_BAND_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BAND_INT32(void)
{
  return &(GrB_BAND_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BAND_INT64(void)
{
  return &(GrB_BAND_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BAND_INT8(void)
{
  return &(GrB_BAND_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BAND_UINT16(void)
{
  return &(GrB_BAND_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BAND_UINT32(void)
{
  return &(GrB_BAND_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BAND_UINT64(void)
{
  return &(GrB_BAND_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BAND_UINT8(void)
{
  return &(GrB_BAND_UINT8);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_BNOT_INT16(void)
{
  return &(GrB_BNOT_INT16);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_BNOT_INT32(void)
{
  return &(GrB_BNOT_INT32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_BNOT_INT64(void)
{
  return &(GrB_BNOT_INT64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_BNOT_INT8(void)
{
  return &(GrB_BNOT_INT8);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_BNOT_UINT16(void)
{
  return &(GrB_BNOT_UINT16);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_BNOT_UINT32(void)
{
  return &(GrB_BNOT_UINT32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_BNOT_UINT64(void)
{
  return &(GrB_BNOT_UINT64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_BNOT_UINT8(void)
{
  return &(GrB_BNOT_UINT8);
}

static struct GB_Type_opaque * *_cffi_var_GrB_BOOL(void)
{
  return &(GrB_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BOR_INT16(void)
{
  return &(GrB_BOR_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BOR_INT32(void)
{
  return &(GrB_BOR_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BOR_INT64(void)
{
  return &(GrB_BOR_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BOR_INT8(void)
{
  return &(GrB_BOR_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BOR_UINT16(void)
{
  return &(GrB_BOR_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BOR_UINT32(void)
{
  return &(GrB_BOR_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BOR_UINT64(void)
{
  return &(GrB_BOR_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BOR_UINT8(void)
{
  return &(GrB_BOR_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BXNOR_INT16(void)
{
  return &(GrB_BXNOR_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BXNOR_INT32(void)
{
  return &(GrB_BXNOR_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BXNOR_INT64(void)
{
  return &(GrB_BXNOR_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BXNOR_INT8(void)
{
  return &(GrB_BXNOR_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BXNOR_UINT16(void)
{
  return &(GrB_BXNOR_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BXNOR_UINT32(void)
{
  return &(GrB_BXNOR_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BXNOR_UINT64(void)
{
  return &(GrB_BXNOR_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BXNOR_UINT8(void)
{
  return &(GrB_BXNOR_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BXOR_INT16(void)
{
  return &(GrB_BXOR_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BXOR_INT32(void)
{
  return &(GrB_BXOR_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BXOR_INT64(void)
{
  return &(GrB_BXOR_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BXOR_INT8(void)
{
  return &(GrB_BXOR_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BXOR_UINT16(void)
{
  return &(GrB_BXOR_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BXOR_UINT32(void)
{
  return &(GrB_BXOR_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BXOR_UINT64(void)
{
  return &(GrB_BXOR_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_BXOR_UINT8(void)
{
  return &(GrB_BXOR_UINT8);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_COLGT(void)
{
  return &(GrB_COLGT);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_COLINDEX_INT32(void)
{
  return &(GrB_COLINDEX_INT32);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_COLINDEX_INT64(void)
{
  return &(GrB_COLINDEX_INT64);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_COLLE(void)
{
  return &(GrB_COLLE);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_C(void)
{
  return &(GrB_DESC_C);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_CT0(void)
{
  return &(GrB_DESC_CT0);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_CT0T1(void)
{
  return &(GrB_DESC_CT0T1);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_CT1(void)
{
  return &(GrB_DESC_CT1);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_R(void)
{
  return &(GrB_DESC_R);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_RC(void)
{
  return &(GrB_DESC_RC);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_RCT0(void)
{
  return &(GrB_DESC_RCT0);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_RCT0T1(void)
{
  return &(GrB_DESC_RCT0T1);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_RCT1(void)
{
  return &(GrB_DESC_RCT1);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_RS(void)
{
  return &(GrB_DESC_RS);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_RSC(void)
{
  return &(GrB_DESC_RSC);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_RSCT0(void)
{
  return &(GrB_DESC_RSCT0);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_RSCT0T1(void)
{
  return &(GrB_DESC_RSCT0T1);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_RSCT1(void)
{
  return &(GrB_DESC_RSCT1);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_RST0(void)
{
  return &(GrB_DESC_RST0);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_RST0T1(void)
{
  return &(GrB_DESC_RST0T1);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_RST1(void)
{
  return &(GrB_DESC_RST1);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_RT0(void)
{
  return &(GrB_DESC_RT0);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_RT0T1(void)
{
  return &(GrB_DESC_RT0T1);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_RT1(void)
{
  return &(GrB_DESC_RT1);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_S(void)
{
  return &(GrB_DESC_S);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_SC(void)
{
  return &(GrB_DESC_SC);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_SCT0(void)
{
  return &(GrB_DESC_SCT0);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_SCT0T1(void)
{
  return &(GrB_DESC_SCT0T1);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_SCT1(void)
{
  return &(GrB_DESC_SCT1);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_ST0(void)
{
  return &(GrB_DESC_ST0);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_ST0T1(void)
{
  return &(GrB_DESC_ST0T1);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_ST1(void)
{
  return &(GrB_DESC_ST1);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_T0(void)
{
  return &(GrB_DESC_T0);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_T0T1(void)
{
  return &(GrB_DESC_T0T1);
}

static struct GB_Descriptor_opaque * *_cffi_var_GrB_DESC_T1(void)
{
  return &(GrB_DESC_T1);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_DIAG(void)
{
  return &(GrB_DIAG);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_DIAGINDEX_INT32(void)
{
  return &(GrB_DIAGINDEX_INT32);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_DIAGINDEX_INT64(void)
{
  return &(GrB_DIAGINDEX_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_DIV_BOOL(void)
{
  return &(GrB_DIV_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_DIV_FP32(void)
{
  return &(GrB_DIV_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_DIV_FP64(void)
{
  return &(GrB_DIV_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_DIV_INT16(void)
{
  return &(GrB_DIV_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_DIV_INT32(void)
{
  return &(GrB_DIV_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_DIV_INT64(void)
{
  return &(GrB_DIV_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_DIV_INT8(void)
{
  return &(GrB_DIV_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_DIV_UINT16(void)
{
  return &(GrB_DIV_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_DIV_UINT32(void)
{
  return &(GrB_DIV_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_DIV_UINT64(void)
{
  return &(GrB_DIV_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_DIV_UINT8(void)
{
  return &(GrB_DIV_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_EQ_BOOL(void)
{
  return &(GrB_EQ_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_EQ_FP32(void)
{
  return &(GrB_EQ_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_EQ_FP64(void)
{
  return &(GrB_EQ_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_EQ_INT16(void)
{
  return &(GrB_EQ_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_EQ_INT32(void)
{
  return &(GrB_EQ_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_EQ_INT64(void)
{
  return &(GrB_EQ_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_EQ_INT8(void)
{
  return &(GrB_EQ_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_EQ_UINT16(void)
{
  return &(GrB_EQ_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_EQ_UINT32(void)
{
  return &(GrB_EQ_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_EQ_UINT64(void)
{
  return &(GrB_EQ_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_EQ_UINT8(void)
{
  return &(GrB_EQ_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_FIRST_BOOL(void)
{
  return &(GrB_FIRST_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_FIRST_FP32(void)
{
  return &(GrB_FIRST_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_FIRST_FP64(void)
{
  return &(GrB_FIRST_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_FIRST_INT16(void)
{
  return &(GrB_FIRST_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_FIRST_INT32(void)
{
  return &(GrB_FIRST_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_FIRST_INT64(void)
{
  return &(GrB_FIRST_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_FIRST_INT8(void)
{
  return &(GrB_FIRST_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_FIRST_UINT16(void)
{
  return &(GrB_FIRST_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_FIRST_UINT32(void)
{
  return &(GrB_FIRST_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_FIRST_UINT64(void)
{
  return &(GrB_FIRST_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_FIRST_UINT8(void)
{
  return &(GrB_FIRST_UINT8);
}

static struct GB_Type_opaque * *_cffi_var_GrB_FP32(void)
{
  return &(GrB_FP32);
}

static struct GB_Type_opaque * *_cffi_var_GrB_FP64(void)
{
  return &(GrB_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_GE_BOOL(void)
{
  return &(GrB_GE_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_GE_FP32(void)
{
  return &(GrB_GE_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_GE_FP64(void)
{
  return &(GrB_GE_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_GE_INT16(void)
{
  return &(GrB_GE_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_GE_INT32(void)
{
  return &(GrB_GE_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_GE_INT64(void)
{
  return &(GrB_GE_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_GE_INT8(void)
{
  return &(GrB_GE_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_GE_UINT16(void)
{
  return &(GrB_GE_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_GE_UINT32(void)
{
  return &(GrB_GE_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_GE_UINT64(void)
{
  return &(GrB_GE_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_GE_UINT8(void)
{
  return &(GrB_GE_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_GT_BOOL(void)
{
  return &(GrB_GT_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_GT_FP32(void)
{
  return &(GrB_GT_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_GT_FP64(void)
{
  return &(GrB_GT_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_GT_INT16(void)
{
  return &(GrB_GT_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_GT_INT32(void)
{
  return &(GrB_GT_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_GT_INT64(void)
{
  return &(GrB_GT_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_GT_INT8(void)
{
  return &(GrB_GT_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_GT_UINT16(void)
{
  return &(GrB_GT_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_GT_UINT32(void)
{
  return &(GrB_GT_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_GT_UINT64(void)
{
  return &(GrB_GT_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_GT_UINT8(void)
{
  return &(GrB_GT_UINT8);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_IDENTITY_BOOL(void)
{
  return &(GrB_IDENTITY_BOOL);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_IDENTITY_FP32(void)
{
  return &(GrB_IDENTITY_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_IDENTITY_FP64(void)
{
  return &(GrB_IDENTITY_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_IDENTITY_INT16(void)
{
  return &(GrB_IDENTITY_INT16);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_IDENTITY_INT32(void)
{
  return &(GrB_IDENTITY_INT32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_IDENTITY_INT64(void)
{
  return &(GrB_IDENTITY_INT64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_IDENTITY_INT8(void)
{
  return &(GrB_IDENTITY_INT8);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_IDENTITY_UINT16(void)
{
  return &(GrB_IDENTITY_UINT16);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_IDENTITY_UINT32(void)
{
  return &(GrB_IDENTITY_UINT32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_IDENTITY_UINT64(void)
{
  return &(GrB_IDENTITY_UINT64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_IDENTITY_UINT8(void)
{
  return &(GrB_IDENTITY_UINT8);
}

static struct GB_Type_opaque * *_cffi_var_GrB_INT16(void)
{
  return &(GrB_INT16);
}

static struct GB_Type_opaque * *_cffi_var_GrB_INT32(void)
{
  return &(GrB_INT32);
}

static struct GB_Type_opaque * *_cffi_var_GrB_INT64(void)
{
  return &(GrB_INT64);
}

static struct GB_Type_opaque * *_cffi_var_GrB_INT8(void)
{
  return &(GrB_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LAND(void)
{
  return &(GrB_LAND);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_LAND_LOR_SEMIRING_BOOL(void)
{
  return &(GrB_LAND_LOR_SEMIRING_BOOL);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_LAND_MONOID_BOOL(void)
{
  return &(GrB_LAND_MONOID_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LE_BOOL(void)
{
  return &(GrB_LE_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LE_FP32(void)
{
  return &(GrB_LE_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LE_FP64(void)
{
  return &(GrB_LE_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LE_INT16(void)
{
  return &(GrB_LE_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LE_INT32(void)
{
  return &(GrB_LE_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LE_INT64(void)
{
  return &(GrB_LE_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LE_INT8(void)
{
  return &(GrB_LE_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LE_UINT16(void)
{
  return &(GrB_LE_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LE_UINT32(void)
{
  return &(GrB_LE_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LE_UINT64(void)
{
  return &(GrB_LE_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LE_UINT8(void)
{
  return &(GrB_LE_UINT8);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_LNOT(void)
{
  return &(GrB_LNOT);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LOR(void)
{
  return &(GrB_LOR);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_LOR_LAND_SEMIRING_BOOL(void)
{
  return &(GrB_LOR_LAND_SEMIRING_BOOL);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_LOR_MONOID_BOOL(void)
{
  return &(GrB_LOR_MONOID_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LT_BOOL(void)
{
  return &(GrB_LT_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LT_FP32(void)
{
  return &(GrB_LT_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LT_FP64(void)
{
  return &(GrB_LT_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LT_INT16(void)
{
  return &(GrB_LT_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LT_INT32(void)
{
  return &(GrB_LT_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LT_INT64(void)
{
  return &(GrB_LT_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LT_INT8(void)
{
  return &(GrB_LT_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LT_UINT16(void)
{
  return &(GrB_LT_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LT_UINT32(void)
{
  return &(GrB_LT_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LT_UINT64(void)
{
  return &(GrB_LT_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LT_UINT8(void)
{
  return &(GrB_LT_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LXNOR(void)
{
  return &(GrB_LXNOR);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_LXNOR_LOR_SEMIRING_BOOL(void)
{
  return &(GrB_LXNOR_LOR_SEMIRING_BOOL);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_LXNOR_MONOID_BOOL(void)
{
  return &(GrB_LXNOR_MONOID_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_LXOR(void)
{
  return &(GrB_LXOR);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_LXOR_LAND_SEMIRING_BOOL(void)
{
  return &(GrB_LXOR_LAND_SEMIRING_BOOL);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_LXOR_MONOID_BOOL(void)
{
  return &(GrB_LXOR_MONOID_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MAX_BOOL(void)
{
  return &(GrB_MAX_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_FIRST_SEMIRING_FP32(void)
{
  return &(GrB_MAX_FIRST_SEMIRING_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_FIRST_SEMIRING_FP64(void)
{
  return &(GrB_MAX_FIRST_SEMIRING_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_FIRST_SEMIRING_INT16(void)
{
  return &(GrB_MAX_FIRST_SEMIRING_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_FIRST_SEMIRING_INT32(void)
{
  return &(GrB_MAX_FIRST_SEMIRING_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_FIRST_SEMIRING_INT64(void)
{
  return &(GrB_MAX_FIRST_SEMIRING_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_FIRST_SEMIRING_INT8(void)
{
  return &(GrB_MAX_FIRST_SEMIRING_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_FIRST_SEMIRING_UINT16(void)
{
  return &(GrB_MAX_FIRST_SEMIRING_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_FIRST_SEMIRING_UINT32(void)
{
  return &(GrB_MAX_FIRST_SEMIRING_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_FIRST_SEMIRING_UINT64(void)
{
  return &(GrB_MAX_FIRST_SEMIRING_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_FIRST_SEMIRING_UINT8(void)
{
  return &(GrB_MAX_FIRST_SEMIRING_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MAX_FP32(void)
{
  return &(GrB_MAX_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MAX_FP64(void)
{
  return &(GrB_MAX_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MAX_INT16(void)
{
  return &(GrB_MAX_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MAX_INT32(void)
{
  return &(GrB_MAX_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MAX_INT64(void)
{
  return &(GrB_MAX_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MAX_INT8(void)
{
  return &(GrB_MAX_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_MIN_SEMIRING_FP32(void)
{
  return &(GrB_MAX_MIN_SEMIRING_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_MIN_SEMIRING_FP64(void)
{
  return &(GrB_MAX_MIN_SEMIRING_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_MIN_SEMIRING_INT16(void)
{
  return &(GrB_MAX_MIN_SEMIRING_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_MIN_SEMIRING_INT32(void)
{
  return &(GrB_MAX_MIN_SEMIRING_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_MIN_SEMIRING_INT64(void)
{
  return &(GrB_MAX_MIN_SEMIRING_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_MIN_SEMIRING_INT8(void)
{
  return &(GrB_MAX_MIN_SEMIRING_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_MIN_SEMIRING_UINT16(void)
{
  return &(GrB_MAX_MIN_SEMIRING_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_MIN_SEMIRING_UINT32(void)
{
  return &(GrB_MAX_MIN_SEMIRING_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_MIN_SEMIRING_UINT64(void)
{
  return &(GrB_MAX_MIN_SEMIRING_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_MIN_SEMIRING_UINT8(void)
{
  return &(GrB_MAX_MIN_SEMIRING_UINT8);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_MAX_MONOID_FP32(void)
{
  return &(GrB_MAX_MONOID_FP32);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_MAX_MONOID_FP64(void)
{
  return &(GrB_MAX_MONOID_FP64);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_MAX_MONOID_INT16(void)
{
  return &(GrB_MAX_MONOID_INT16);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_MAX_MONOID_INT32(void)
{
  return &(GrB_MAX_MONOID_INT32);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_MAX_MONOID_INT64(void)
{
  return &(GrB_MAX_MONOID_INT64);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_MAX_MONOID_INT8(void)
{
  return &(GrB_MAX_MONOID_INT8);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_MAX_MONOID_UINT16(void)
{
  return &(GrB_MAX_MONOID_UINT16);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_MAX_MONOID_UINT32(void)
{
  return &(GrB_MAX_MONOID_UINT32);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_MAX_MONOID_UINT64(void)
{
  return &(GrB_MAX_MONOID_UINT64);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_MAX_MONOID_UINT8(void)
{
  return &(GrB_MAX_MONOID_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_PLUS_SEMIRING_FP32(void)
{
  return &(GrB_MAX_PLUS_SEMIRING_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_PLUS_SEMIRING_FP64(void)
{
  return &(GrB_MAX_PLUS_SEMIRING_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_PLUS_SEMIRING_INT16(void)
{
  return &(GrB_MAX_PLUS_SEMIRING_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_PLUS_SEMIRING_INT32(void)
{
  return &(GrB_MAX_PLUS_SEMIRING_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_PLUS_SEMIRING_INT64(void)
{
  return &(GrB_MAX_PLUS_SEMIRING_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_PLUS_SEMIRING_INT8(void)
{
  return &(GrB_MAX_PLUS_SEMIRING_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_PLUS_SEMIRING_UINT16(void)
{
  return &(GrB_MAX_PLUS_SEMIRING_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_PLUS_SEMIRING_UINT32(void)
{
  return &(GrB_MAX_PLUS_SEMIRING_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_PLUS_SEMIRING_UINT64(void)
{
  return &(GrB_MAX_PLUS_SEMIRING_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_PLUS_SEMIRING_UINT8(void)
{
  return &(GrB_MAX_PLUS_SEMIRING_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_SECOND_SEMIRING_FP32(void)
{
  return &(GrB_MAX_SECOND_SEMIRING_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_SECOND_SEMIRING_FP64(void)
{
  return &(GrB_MAX_SECOND_SEMIRING_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_SECOND_SEMIRING_INT16(void)
{
  return &(GrB_MAX_SECOND_SEMIRING_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_SECOND_SEMIRING_INT32(void)
{
  return &(GrB_MAX_SECOND_SEMIRING_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_SECOND_SEMIRING_INT64(void)
{
  return &(GrB_MAX_SECOND_SEMIRING_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_SECOND_SEMIRING_INT8(void)
{
  return &(GrB_MAX_SECOND_SEMIRING_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_SECOND_SEMIRING_UINT16(void)
{
  return &(GrB_MAX_SECOND_SEMIRING_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_SECOND_SEMIRING_UINT32(void)
{
  return &(GrB_MAX_SECOND_SEMIRING_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_SECOND_SEMIRING_UINT64(void)
{
  return &(GrB_MAX_SECOND_SEMIRING_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_SECOND_SEMIRING_UINT8(void)
{
  return &(GrB_MAX_SECOND_SEMIRING_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_TIMES_SEMIRING_FP32(void)
{
  return &(GrB_MAX_TIMES_SEMIRING_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_TIMES_SEMIRING_FP64(void)
{
  return &(GrB_MAX_TIMES_SEMIRING_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_TIMES_SEMIRING_INT16(void)
{
  return &(GrB_MAX_TIMES_SEMIRING_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_TIMES_SEMIRING_INT32(void)
{
  return &(GrB_MAX_TIMES_SEMIRING_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_TIMES_SEMIRING_INT64(void)
{
  return &(GrB_MAX_TIMES_SEMIRING_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_TIMES_SEMIRING_INT8(void)
{
  return &(GrB_MAX_TIMES_SEMIRING_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_TIMES_SEMIRING_UINT16(void)
{
  return &(GrB_MAX_TIMES_SEMIRING_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_TIMES_SEMIRING_UINT32(void)
{
  return &(GrB_MAX_TIMES_SEMIRING_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_TIMES_SEMIRING_UINT64(void)
{
  return &(GrB_MAX_TIMES_SEMIRING_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MAX_TIMES_SEMIRING_UINT8(void)
{
  return &(GrB_MAX_TIMES_SEMIRING_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MAX_UINT16(void)
{
  return &(GrB_MAX_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MAX_UINT32(void)
{
  return &(GrB_MAX_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MAX_UINT64(void)
{
  return &(GrB_MAX_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MAX_UINT8(void)
{
  return &(GrB_MAX_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MINUS_BOOL(void)
{
  return &(GrB_MINUS_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MINUS_FP32(void)
{
  return &(GrB_MINUS_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MINUS_FP64(void)
{
  return &(GrB_MINUS_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MINUS_INT16(void)
{
  return &(GrB_MINUS_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MINUS_INT32(void)
{
  return &(GrB_MINUS_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MINUS_INT64(void)
{
  return &(GrB_MINUS_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MINUS_INT8(void)
{
  return &(GrB_MINUS_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MINUS_UINT16(void)
{
  return &(GrB_MINUS_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MINUS_UINT32(void)
{
  return &(GrB_MINUS_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MINUS_UINT64(void)
{
  return &(GrB_MINUS_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MINUS_UINT8(void)
{
  return &(GrB_MINUS_UINT8);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_MINV_BOOL(void)
{
  return &(GrB_MINV_BOOL);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_MINV_FP32(void)
{
  return &(GrB_MINV_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_MINV_FP64(void)
{
  return &(GrB_MINV_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_MINV_INT16(void)
{
  return &(GrB_MINV_INT16);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_MINV_INT32(void)
{
  return &(GrB_MINV_INT32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_MINV_INT64(void)
{
  return &(GrB_MINV_INT64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_MINV_INT8(void)
{
  return &(GrB_MINV_INT8);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_MINV_UINT16(void)
{
  return &(GrB_MINV_UINT16);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_MINV_UINT32(void)
{
  return &(GrB_MINV_UINT32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_MINV_UINT64(void)
{
  return &(GrB_MINV_UINT64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GrB_MINV_UINT8(void)
{
  return &(GrB_MINV_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MIN_BOOL(void)
{
  return &(GrB_MIN_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_FIRST_SEMIRING_FP32(void)
{
  return &(GrB_MIN_FIRST_SEMIRING_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_FIRST_SEMIRING_FP64(void)
{
  return &(GrB_MIN_FIRST_SEMIRING_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_FIRST_SEMIRING_INT16(void)
{
  return &(GrB_MIN_FIRST_SEMIRING_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_FIRST_SEMIRING_INT32(void)
{
  return &(GrB_MIN_FIRST_SEMIRING_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_FIRST_SEMIRING_INT64(void)
{
  return &(GrB_MIN_FIRST_SEMIRING_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_FIRST_SEMIRING_INT8(void)
{
  return &(GrB_MIN_FIRST_SEMIRING_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_FIRST_SEMIRING_UINT16(void)
{
  return &(GrB_MIN_FIRST_SEMIRING_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_FIRST_SEMIRING_UINT32(void)
{
  return &(GrB_MIN_FIRST_SEMIRING_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_FIRST_SEMIRING_UINT64(void)
{
  return &(GrB_MIN_FIRST_SEMIRING_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_FIRST_SEMIRING_UINT8(void)
{
  return &(GrB_MIN_FIRST_SEMIRING_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MIN_FP32(void)
{
  return &(GrB_MIN_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MIN_FP64(void)
{
  return &(GrB_MIN_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MIN_INT16(void)
{
  return &(GrB_MIN_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MIN_INT32(void)
{
  return &(GrB_MIN_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MIN_INT64(void)
{
  return &(GrB_MIN_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MIN_INT8(void)
{
  return &(GrB_MIN_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_MAX_SEMIRING_FP32(void)
{
  return &(GrB_MIN_MAX_SEMIRING_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_MAX_SEMIRING_FP64(void)
{
  return &(GrB_MIN_MAX_SEMIRING_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_MAX_SEMIRING_INT16(void)
{
  return &(GrB_MIN_MAX_SEMIRING_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_MAX_SEMIRING_INT32(void)
{
  return &(GrB_MIN_MAX_SEMIRING_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_MAX_SEMIRING_INT64(void)
{
  return &(GrB_MIN_MAX_SEMIRING_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_MAX_SEMIRING_INT8(void)
{
  return &(GrB_MIN_MAX_SEMIRING_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_MAX_SEMIRING_UINT16(void)
{
  return &(GrB_MIN_MAX_SEMIRING_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_MAX_SEMIRING_UINT32(void)
{
  return &(GrB_MIN_MAX_SEMIRING_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_MAX_SEMIRING_UINT64(void)
{
  return &(GrB_MIN_MAX_SEMIRING_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_MAX_SEMIRING_UINT8(void)
{
  return &(GrB_MIN_MAX_SEMIRING_UINT8);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_MIN_MONOID_FP32(void)
{
  return &(GrB_MIN_MONOID_FP32);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_MIN_MONOID_FP64(void)
{
  return &(GrB_MIN_MONOID_FP64);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_MIN_MONOID_INT16(void)
{
  return &(GrB_MIN_MONOID_INT16);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_MIN_MONOID_INT32(void)
{
  return &(GrB_MIN_MONOID_INT32);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_MIN_MONOID_INT64(void)
{
  return &(GrB_MIN_MONOID_INT64);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_MIN_MONOID_INT8(void)
{
  return &(GrB_MIN_MONOID_INT8);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_MIN_MONOID_UINT16(void)
{
  return &(GrB_MIN_MONOID_UINT16);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_MIN_MONOID_UINT32(void)
{
  return &(GrB_MIN_MONOID_UINT32);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_MIN_MONOID_UINT64(void)
{
  return &(GrB_MIN_MONOID_UINT64);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_MIN_MONOID_UINT8(void)
{
  return &(GrB_MIN_MONOID_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_PLUS_SEMIRING_FP32(void)
{
  return &(GrB_MIN_PLUS_SEMIRING_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_PLUS_SEMIRING_FP64(void)
{
  return &(GrB_MIN_PLUS_SEMIRING_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_PLUS_SEMIRING_INT16(void)
{
  return &(GrB_MIN_PLUS_SEMIRING_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_PLUS_SEMIRING_INT32(void)
{
  return &(GrB_MIN_PLUS_SEMIRING_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_PLUS_SEMIRING_INT64(void)
{
  return &(GrB_MIN_PLUS_SEMIRING_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_PLUS_SEMIRING_INT8(void)
{
  return &(GrB_MIN_PLUS_SEMIRING_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_PLUS_SEMIRING_UINT16(void)
{
  return &(GrB_MIN_PLUS_SEMIRING_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_PLUS_SEMIRING_UINT32(void)
{
  return &(GrB_MIN_PLUS_SEMIRING_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_PLUS_SEMIRING_UINT64(void)
{
  return &(GrB_MIN_PLUS_SEMIRING_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_PLUS_SEMIRING_UINT8(void)
{
  return &(GrB_MIN_PLUS_SEMIRING_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_SECOND_SEMIRING_FP32(void)
{
  return &(GrB_MIN_SECOND_SEMIRING_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_SECOND_SEMIRING_FP64(void)
{
  return &(GrB_MIN_SECOND_SEMIRING_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_SECOND_SEMIRING_INT16(void)
{
  return &(GrB_MIN_SECOND_SEMIRING_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_SECOND_SEMIRING_INT32(void)
{
  return &(GrB_MIN_SECOND_SEMIRING_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_SECOND_SEMIRING_INT64(void)
{
  return &(GrB_MIN_SECOND_SEMIRING_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_SECOND_SEMIRING_INT8(void)
{
  return &(GrB_MIN_SECOND_SEMIRING_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_SECOND_SEMIRING_UINT16(void)
{
  return &(GrB_MIN_SECOND_SEMIRING_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_SECOND_SEMIRING_UINT32(void)
{
  return &(GrB_MIN_SECOND_SEMIRING_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_SECOND_SEMIRING_UINT64(void)
{
  return &(GrB_MIN_SECOND_SEMIRING_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_SECOND_SEMIRING_UINT8(void)
{
  return &(GrB_MIN_SECOND_SEMIRING_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_TIMES_SEMIRING_FP32(void)
{
  return &(GrB_MIN_TIMES_SEMIRING_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_TIMES_SEMIRING_FP64(void)
{
  return &(GrB_MIN_TIMES_SEMIRING_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_TIMES_SEMIRING_INT16(void)
{
  return &(GrB_MIN_TIMES_SEMIRING_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_TIMES_SEMIRING_INT32(void)
{
  return &(GrB_MIN_TIMES_SEMIRING_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_TIMES_SEMIRING_INT64(void)
{
  return &(GrB_MIN_TIMES_SEMIRING_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_TIMES_SEMIRING_INT8(void)
{
  return &(GrB_MIN_TIMES_SEMIRING_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_TIMES_SEMIRING_UINT16(void)
{
  return &(GrB_MIN_TIMES_SEMIRING_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_TIMES_SEMIRING_UINT32(void)
{
  return &(GrB_MIN_TIMES_SEMIRING_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_TIMES_SEMIRING_UINT64(void)
{
  return &(GrB_MIN_TIMES_SEMIRING_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_MIN_TIMES_SEMIRING_UINT8(void)
{
  return &(GrB_MIN_TIMES_SEMIRING_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MIN_UINT16(void)
{
  return &(GrB_MIN_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MIN_UINT32(void)
{
  return &(GrB_MIN_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MIN_UINT64(void)
{
  return &(GrB_MIN_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_MIN_UINT8(void)
{
  return &(GrB_MIN_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_NE_BOOL(void)
{
  return &(GrB_NE_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_NE_FP32(void)
{
  return &(GrB_NE_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_NE_FP64(void)
{
  return &(GrB_NE_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_NE_INT16(void)
{
  return &(GrB_NE_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_NE_INT32(void)
{
  return &(GrB_NE_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_NE_INT64(void)
{
  return &(GrB_NE_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_NE_INT8(void)
{
  return &(GrB_NE_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_NE_UINT16(void)
{
  return &(GrB_NE_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_NE_UINT32(void)
{
  return &(GrB_NE_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_NE_UINT64(void)
{
  return &(GrB_NE_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_NE_UINT8(void)
{
  return &(GrB_NE_UINT8);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_OFFDIAG(void)
{
  return &(GrB_OFFDIAG);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_ONEB_BOOL(void)
{
  return &(GrB_ONEB_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_ONEB_FP32(void)
{
  return &(GrB_ONEB_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_ONEB_FP64(void)
{
  return &(GrB_ONEB_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_ONEB_INT16(void)
{
  return &(GrB_ONEB_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_ONEB_INT32(void)
{
  return &(GrB_ONEB_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_ONEB_INT64(void)
{
  return &(GrB_ONEB_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_ONEB_INT8(void)
{
  return &(GrB_ONEB_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_ONEB_UINT16(void)
{
  return &(GrB_ONEB_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_ONEB_UINT32(void)
{
  return &(GrB_ONEB_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_ONEB_UINT64(void)
{
  return &(GrB_ONEB_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_ONEB_UINT8(void)
{
  return &(GrB_ONEB_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_PLUS_BOOL(void)
{
  return &(GrB_PLUS_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_PLUS_FP32(void)
{
  return &(GrB_PLUS_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_PLUS_FP64(void)
{
  return &(GrB_PLUS_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_PLUS_INT16(void)
{
  return &(GrB_PLUS_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_PLUS_INT32(void)
{
  return &(GrB_PLUS_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_PLUS_INT64(void)
{
  return &(GrB_PLUS_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_PLUS_INT8(void)
{
  return &(GrB_PLUS_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_PLUS_MIN_SEMIRING_FP32(void)
{
  return &(GrB_PLUS_MIN_SEMIRING_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_PLUS_MIN_SEMIRING_FP64(void)
{
  return &(GrB_PLUS_MIN_SEMIRING_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_PLUS_MIN_SEMIRING_INT16(void)
{
  return &(GrB_PLUS_MIN_SEMIRING_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_PLUS_MIN_SEMIRING_INT32(void)
{
  return &(GrB_PLUS_MIN_SEMIRING_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_PLUS_MIN_SEMIRING_INT64(void)
{
  return &(GrB_PLUS_MIN_SEMIRING_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_PLUS_MIN_SEMIRING_INT8(void)
{
  return &(GrB_PLUS_MIN_SEMIRING_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_PLUS_MIN_SEMIRING_UINT16(void)
{
  return &(GrB_PLUS_MIN_SEMIRING_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_PLUS_MIN_SEMIRING_UINT32(void)
{
  return &(GrB_PLUS_MIN_SEMIRING_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_PLUS_MIN_SEMIRING_UINT64(void)
{
  return &(GrB_PLUS_MIN_SEMIRING_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_PLUS_MIN_SEMIRING_UINT8(void)
{
  return &(GrB_PLUS_MIN_SEMIRING_UINT8);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_PLUS_MONOID_FP32(void)
{
  return &(GrB_PLUS_MONOID_FP32);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_PLUS_MONOID_FP64(void)
{
  return &(GrB_PLUS_MONOID_FP64);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_PLUS_MONOID_INT16(void)
{
  return &(GrB_PLUS_MONOID_INT16);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_PLUS_MONOID_INT32(void)
{
  return &(GrB_PLUS_MONOID_INT32);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_PLUS_MONOID_INT64(void)
{
  return &(GrB_PLUS_MONOID_INT64);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_PLUS_MONOID_INT8(void)
{
  return &(GrB_PLUS_MONOID_INT8);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_PLUS_MONOID_UINT16(void)
{
  return &(GrB_PLUS_MONOID_UINT16);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_PLUS_MONOID_UINT32(void)
{
  return &(GrB_PLUS_MONOID_UINT32);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_PLUS_MONOID_UINT64(void)
{
  return &(GrB_PLUS_MONOID_UINT64);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_PLUS_MONOID_UINT8(void)
{
  return &(GrB_PLUS_MONOID_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_PLUS_TIMES_SEMIRING_FP32(void)
{
  return &(GrB_PLUS_TIMES_SEMIRING_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_PLUS_TIMES_SEMIRING_FP64(void)
{
  return &(GrB_PLUS_TIMES_SEMIRING_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_PLUS_TIMES_SEMIRING_INT16(void)
{
  return &(GrB_PLUS_TIMES_SEMIRING_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_PLUS_TIMES_SEMIRING_INT32(void)
{
  return &(GrB_PLUS_TIMES_SEMIRING_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_PLUS_TIMES_SEMIRING_INT64(void)
{
  return &(GrB_PLUS_TIMES_SEMIRING_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_PLUS_TIMES_SEMIRING_INT8(void)
{
  return &(GrB_PLUS_TIMES_SEMIRING_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_PLUS_TIMES_SEMIRING_UINT16(void)
{
  return &(GrB_PLUS_TIMES_SEMIRING_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_PLUS_TIMES_SEMIRING_UINT32(void)
{
  return &(GrB_PLUS_TIMES_SEMIRING_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_PLUS_TIMES_SEMIRING_UINT64(void)
{
  return &(GrB_PLUS_TIMES_SEMIRING_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GrB_PLUS_TIMES_SEMIRING_UINT8(void)
{
  return &(GrB_PLUS_TIMES_SEMIRING_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_PLUS_UINT16(void)
{
  return &(GrB_PLUS_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_PLUS_UINT32(void)
{
  return &(GrB_PLUS_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_PLUS_UINT64(void)
{
  return &(GrB_PLUS_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_PLUS_UINT8(void)
{
  return &(GrB_PLUS_UINT8);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_ROWGT(void)
{
  return &(GrB_ROWGT);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_ROWINDEX_INT32(void)
{
  return &(GrB_ROWINDEX_INT32);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_ROWINDEX_INT64(void)
{
  return &(GrB_ROWINDEX_INT64);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_ROWLE(void)
{
  return &(GrB_ROWLE);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_SECOND_BOOL(void)
{
  return &(GrB_SECOND_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_SECOND_FP32(void)
{
  return &(GrB_SECOND_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_SECOND_FP64(void)
{
  return &(GrB_SECOND_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_SECOND_INT16(void)
{
  return &(GrB_SECOND_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_SECOND_INT32(void)
{
  return &(GrB_SECOND_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_SECOND_INT64(void)
{
  return &(GrB_SECOND_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_SECOND_INT8(void)
{
  return &(GrB_SECOND_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_SECOND_UINT16(void)
{
  return &(GrB_SECOND_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_SECOND_UINT32(void)
{
  return &(GrB_SECOND_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_SECOND_UINT64(void)
{
  return &(GrB_SECOND_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_SECOND_UINT8(void)
{
  return &(GrB_SECOND_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_TIMES_BOOL(void)
{
  return &(GrB_TIMES_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_TIMES_FP32(void)
{
  return &(GrB_TIMES_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_TIMES_FP64(void)
{
  return &(GrB_TIMES_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_TIMES_INT16(void)
{
  return &(GrB_TIMES_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_TIMES_INT32(void)
{
  return &(GrB_TIMES_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_TIMES_INT64(void)
{
  return &(GrB_TIMES_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_TIMES_INT8(void)
{
  return &(GrB_TIMES_INT8);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_TIMES_MONOID_FP32(void)
{
  return &(GrB_TIMES_MONOID_FP32);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_TIMES_MONOID_FP64(void)
{
  return &(GrB_TIMES_MONOID_FP64);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_TIMES_MONOID_INT16(void)
{
  return &(GrB_TIMES_MONOID_INT16);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_TIMES_MONOID_INT32(void)
{
  return &(GrB_TIMES_MONOID_INT32);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_TIMES_MONOID_INT64(void)
{
  return &(GrB_TIMES_MONOID_INT64);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_TIMES_MONOID_INT8(void)
{
  return &(GrB_TIMES_MONOID_INT8);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_TIMES_MONOID_UINT16(void)
{
  return &(GrB_TIMES_MONOID_UINT16);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_TIMES_MONOID_UINT32(void)
{
  return &(GrB_TIMES_MONOID_UINT32);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_TIMES_MONOID_UINT64(void)
{
  return &(GrB_TIMES_MONOID_UINT64);
}

static struct GB_Monoid_opaque * *_cffi_var_GrB_TIMES_MONOID_UINT8(void)
{
  return &(GrB_TIMES_MONOID_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_TIMES_UINT16(void)
{
  return &(GrB_TIMES_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_TIMES_UINT32(void)
{
  return &(GrB_TIMES_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_TIMES_UINT64(void)
{
  return &(GrB_TIMES_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GrB_TIMES_UINT8(void)
{
  return &(GrB_TIMES_UINT8);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_TRIL(void)
{
  return &(GrB_TRIL);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_TRIU(void)
{
  return &(GrB_TRIU);
}

static struct GB_Type_opaque * *_cffi_var_GrB_UINT16(void)
{
  return &(GrB_UINT16);
}

static struct GB_Type_opaque * *_cffi_var_GrB_UINT32(void)
{
  return &(GrB_UINT32);
}

static struct GB_Type_opaque * *_cffi_var_GrB_UINT64(void)
{
  return &(GrB_UINT64);
}

static struct GB_Type_opaque * *_cffi_var_GrB_UINT8(void)
{
  return &(GrB_UINT8);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEEQ_BOOL(void)
{
  return &(GrB_VALUEEQ_BOOL);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEEQ_FP32(void)
{
  return &(GrB_VALUEEQ_FP32);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEEQ_FP64(void)
{
  return &(GrB_VALUEEQ_FP64);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEEQ_INT16(void)
{
  return &(GrB_VALUEEQ_INT16);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEEQ_INT32(void)
{
  return &(GrB_VALUEEQ_INT32);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEEQ_INT64(void)
{
  return &(GrB_VALUEEQ_INT64);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEEQ_INT8(void)
{
  return &(GrB_VALUEEQ_INT8);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEEQ_UINT16(void)
{
  return &(GrB_VALUEEQ_UINT16);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEEQ_UINT32(void)
{
  return &(GrB_VALUEEQ_UINT32);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEEQ_UINT64(void)
{
  return &(GrB_VALUEEQ_UINT64);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEEQ_UINT8(void)
{
  return &(GrB_VALUEEQ_UINT8);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEGE_BOOL(void)
{
  return &(GrB_VALUEGE_BOOL);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEGE_FP32(void)
{
  return &(GrB_VALUEGE_FP32);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEGE_FP64(void)
{
  return &(GrB_VALUEGE_FP64);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEGE_INT16(void)
{
  return &(GrB_VALUEGE_INT16);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEGE_INT32(void)
{
  return &(GrB_VALUEGE_INT32);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEGE_INT64(void)
{
  return &(GrB_VALUEGE_INT64);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEGE_INT8(void)
{
  return &(GrB_VALUEGE_INT8);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEGE_UINT16(void)
{
  return &(GrB_VALUEGE_UINT16);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEGE_UINT32(void)
{
  return &(GrB_VALUEGE_UINT32);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEGE_UINT64(void)
{
  return &(GrB_VALUEGE_UINT64);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEGE_UINT8(void)
{
  return &(GrB_VALUEGE_UINT8);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEGT_BOOL(void)
{
  return &(GrB_VALUEGT_BOOL);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEGT_FP32(void)
{
  return &(GrB_VALUEGT_FP32);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEGT_FP64(void)
{
  return &(GrB_VALUEGT_FP64);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEGT_INT16(void)
{
  return &(GrB_VALUEGT_INT16);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEGT_INT32(void)
{
  return &(GrB_VALUEGT_INT32);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEGT_INT64(void)
{
  return &(GrB_VALUEGT_INT64);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEGT_INT8(void)
{
  return &(GrB_VALUEGT_INT8);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEGT_UINT16(void)
{
  return &(GrB_VALUEGT_UINT16);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEGT_UINT32(void)
{
  return &(GrB_VALUEGT_UINT32);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEGT_UINT64(void)
{
  return &(GrB_VALUEGT_UINT64);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUEGT_UINT8(void)
{
  return &(GrB_VALUEGT_UINT8);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUELE_BOOL(void)
{
  return &(GrB_VALUELE_BOOL);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUELE_FP32(void)
{
  return &(GrB_VALUELE_FP32);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUELE_FP64(void)
{
  return &(GrB_VALUELE_FP64);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUELE_INT16(void)
{
  return &(GrB_VALUELE_INT16);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUELE_INT32(void)
{
  return &(GrB_VALUELE_INT32);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUELE_INT64(void)
{
  return &(GrB_VALUELE_INT64);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUELE_INT8(void)
{
  return &(GrB_VALUELE_INT8);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUELE_UINT16(void)
{
  return &(GrB_VALUELE_UINT16);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUELE_UINT32(void)
{
  return &(GrB_VALUELE_UINT32);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUELE_UINT64(void)
{
  return &(GrB_VALUELE_UINT64);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUELE_UINT8(void)
{
  return &(GrB_VALUELE_UINT8);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUELT_BOOL(void)
{
  return &(GrB_VALUELT_BOOL);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUELT_FP32(void)
{
  return &(GrB_VALUELT_FP32);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUELT_FP64(void)
{
  return &(GrB_VALUELT_FP64);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUELT_INT16(void)
{
  return &(GrB_VALUELT_INT16);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUELT_INT32(void)
{
  return &(GrB_VALUELT_INT32);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUELT_INT64(void)
{
  return &(GrB_VALUELT_INT64);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUELT_INT8(void)
{
  return &(GrB_VALUELT_INT8);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUELT_UINT16(void)
{
  return &(GrB_VALUELT_UINT16);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUELT_UINT32(void)
{
  return &(GrB_VALUELT_UINT32);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUELT_UINT64(void)
{
  return &(GrB_VALUELT_UINT64);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUELT_UINT8(void)
{
  return &(GrB_VALUELT_UINT8);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUENE_BOOL(void)
{
  return &(GrB_VALUENE_BOOL);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUENE_FP32(void)
{
  return &(GrB_VALUENE_FP32);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUENE_FP64(void)
{
  return &(GrB_VALUENE_FP64);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUENE_INT16(void)
{
  return &(GrB_VALUENE_INT16);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUENE_INT32(void)
{
  return &(GrB_VALUENE_INT32);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUENE_INT64(void)
{
  return &(GrB_VALUENE_INT64);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUENE_INT8(void)
{
  return &(GrB_VALUENE_INT8);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUENE_UINT16(void)
{
  return &(GrB_VALUENE_UINT16);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUENE_UINT32(void)
{
  return &(GrB_VALUENE_UINT32);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUENE_UINT64(void)
{
  return &(GrB_VALUENE_UINT64);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GrB_VALUENE_UINT8(void)
{
  return &(GrB_VALUENE_UINT8);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ABS_FC32(void)
{
  return &(GxB_ABS_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ABS_FC64(void)
{
  return &(GxB_ABS_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ACOSH_FC32(void)
{
  return &(GxB_ACOSH_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ACOSH_FC64(void)
{
  return &(GxB_ACOSH_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ACOSH_FP32(void)
{
  return &(GxB_ACOSH_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ACOSH_FP64(void)
{
  return &(GxB_ACOSH_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ACOS_FC32(void)
{
  return &(GxB_ACOS_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ACOS_FC64(void)
{
  return &(GxB_ACOS_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ACOS_FP32(void)
{
  return &(GxB_ACOS_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ACOS_FP64(void)
{
  return &(GxB_ACOS_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_AINV_FC32(void)
{
  return &(GxB_AINV_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_AINV_FC64(void)
{
  return &(GxB_AINV_FC64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ANY_BOOL(void)
{
  return &(GxB_ANY_BOOL);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_ANY_BOOL_MONOID(void)
{
  return &(GxB_ANY_BOOL_MONOID);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_DIV_FC32(void)
{
  return &(GxB_ANY_DIV_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_DIV_FC64(void)
{
  return &(GxB_ANY_DIV_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_DIV_FP32(void)
{
  return &(GxB_ANY_DIV_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_DIV_FP64(void)
{
  return &(GxB_ANY_DIV_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_DIV_INT16(void)
{
  return &(GxB_ANY_DIV_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_DIV_INT32(void)
{
  return &(GxB_ANY_DIV_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_DIV_INT64(void)
{
  return &(GxB_ANY_DIV_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_DIV_INT8(void)
{
  return &(GxB_ANY_DIV_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_DIV_UINT16(void)
{
  return &(GxB_ANY_DIV_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_DIV_UINT32(void)
{
  return &(GxB_ANY_DIV_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_DIV_UINT64(void)
{
  return &(GxB_ANY_DIV_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_DIV_UINT8(void)
{
  return &(GxB_ANY_DIV_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_EQ_BOOL(void)
{
  return &(GxB_ANY_EQ_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_EQ_FP32(void)
{
  return &(GxB_ANY_EQ_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_EQ_FP64(void)
{
  return &(GxB_ANY_EQ_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_EQ_INT16(void)
{
  return &(GxB_ANY_EQ_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_EQ_INT32(void)
{
  return &(GxB_ANY_EQ_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_EQ_INT64(void)
{
  return &(GxB_ANY_EQ_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_EQ_INT8(void)
{
  return &(GxB_ANY_EQ_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_EQ_UINT16(void)
{
  return &(GxB_ANY_EQ_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_EQ_UINT32(void)
{
  return &(GxB_ANY_EQ_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_EQ_UINT64(void)
{
  return &(GxB_ANY_EQ_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_EQ_UINT8(void)
{
  return &(GxB_ANY_EQ_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ANY_FC32(void)
{
  return &(GxB_ANY_FC32);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_ANY_FC32_MONOID(void)
{
  return &(GxB_ANY_FC32_MONOID);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ANY_FC64(void)
{
  return &(GxB_ANY_FC64);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_ANY_FC64_MONOID(void)
{
  return &(GxB_ANY_FC64_MONOID);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_FIRSTI1_INT32(void)
{
  return &(GxB_ANY_FIRSTI1_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_FIRSTI1_INT64(void)
{
  return &(GxB_ANY_FIRSTI1_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_FIRSTI_INT32(void)
{
  return &(GxB_ANY_FIRSTI_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_FIRSTI_INT64(void)
{
  return &(GxB_ANY_FIRSTI_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_FIRSTJ1_INT32(void)
{
  return &(GxB_ANY_FIRSTJ1_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_FIRSTJ1_INT64(void)
{
  return &(GxB_ANY_FIRSTJ1_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_FIRSTJ_INT32(void)
{
  return &(GxB_ANY_FIRSTJ_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_FIRSTJ_INT64(void)
{
  return &(GxB_ANY_FIRSTJ_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_FIRST_BOOL(void)
{
  return &(GxB_ANY_FIRST_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_FIRST_FC32(void)
{
  return &(GxB_ANY_FIRST_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_FIRST_FC64(void)
{
  return &(GxB_ANY_FIRST_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_FIRST_FP32(void)
{
  return &(GxB_ANY_FIRST_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_FIRST_FP64(void)
{
  return &(GxB_ANY_FIRST_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_FIRST_INT16(void)
{
  return &(GxB_ANY_FIRST_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_FIRST_INT32(void)
{
  return &(GxB_ANY_FIRST_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_FIRST_INT64(void)
{
  return &(GxB_ANY_FIRST_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_FIRST_INT8(void)
{
  return &(GxB_ANY_FIRST_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_FIRST_UINT16(void)
{
  return &(GxB_ANY_FIRST_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_FIRST_UINT32(void)
{
  return &(GxB_ANY_FIRST_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_FIRST_UINT64(void)
{
  return &(GxB_ANY_FIRST_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_FIRST_UINT8(void)
{
  return &(GxB_ANY_FIRST_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ANY_FP32(void)
{
  return &(GxB_ANY_FP32);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_ANY_FP32_MONOID(void)
{
  return &(GxB_ANY_FP32_MONOID);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ANY_FP64(void)
{
  return &(GxB_ANY_FP64);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_ANY_FP64_MONOID(void)
{
  return &(GxB_ANY_FP64_MONOID);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_GE_BOOL(void)
{
  return &(GxB_ANY_GE_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_GE_FP32(void)
{
  return &(GxB_ANY_GE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_GE_FP64(void)
{
  return &(GxB_ANY_GE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_GE_INT16(void)
{
  return &(GxB_ANY_GE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_GE_INT32(void)
{
  return &(GxB_ANY_GE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_GE_INT64(void)
{
  return &(GxB_ANY_GE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_GE_INT8(void)
{
  return &(GxB_ANY_GE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_GE_UINT16(void)
{
  return &(GxB_ANY_GE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_GE_UINT32(void)
{
  return &(GxB_ANY_GE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_GE_UINT64(void)
{
  return &(GxB_ANY_GE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_GE_UINT8(void)
{
  return &(GxB_ANY_GE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_GT_BOOL(void)
{
  return &(GxB_ANY_GT_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_GT_FP32(void)
{
  return &(GxB_ANY_GT_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_GT_FP64(void)
{
  return &(GxB_ANY_GT_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_GT_INT16(void)
{
  return &(GxB_ANY_GT_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_GT_INT32(void)
{
  return &(GxB_ANY_GT_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_GT_INT64(void)
{
  return &(GxB_ANY_GT_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_GT_INT8(void)
{
  return &(GxB_ANY_GT_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_GT_UINT16(void)
{
  return &(GxB_ANY_GT_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_GT_UINT32(void)
{
  return &(GxB_ANY_GT_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_GT_UINT64(void)
{
  return &(GxB_ANY_GT_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_GT_UINT8(void)
{
  return &(GxB_ANY_GT_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ANY_INT16(void)
{
  return &(GxB_ANY_INT16);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_ANY_INT16_MONOID(void)
{
  return &(GxB_ANY_INT16_MONOID);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ANY_INT32(void)
{
  return &(GxB_ANY_INT32);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_ANY_INT32_MONOID(void)
{
  return &(GxB_ANY_INT32_MONOID);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ANY_INT64(void)
{
  return &(GxB_ANY_INT64);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_ANY_INT64_MONOID(void)
{
  return &(GxB_ANY_INT64_MONOID);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ANY_INT8(void)
{
  return &(GxB_ANY_INT8);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_ANY_INT8_MONOID(void)
{
  return &(GxB_ANY_INT8_MONOID);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISEQ_FP32(void)
{
  return &(GxB_ANY_ISEQ_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISEQ_FP64(void)
{
  return &(GxB_ANY_ISEQ_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISEQ_INT16(void)
{
  return &(GxB_ANY_ISEQ_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISEQ_INT32(void)
{
  return &(GxB_ANY_ISEQ_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISEQ_INT64(void)
{
  return &(GxB_ANY_ISEQ_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISEQ_INT8(void)
{
  return &(GxB_ANY_ISEQ_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISEQ_UINT16(void)
{
  return &(GxB_ANY_ISEQ_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISEQ_UINT32(void)
{
  return &(GxB_ANY_ISEQ_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISEQ_UINT64(void)
{
  return &(GxB_ANY_ISEQ_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISEQ_UINT8(void)
{
  return &(GxB_ANY_ISEQ_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISGE_FP32(void)
{
  return &(GxB_ANY_ISGE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISGE_FP64(void)
{
  return &(GxB_ANY_ISGE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISGE_INT16(void)
{
  return &(GxB_ANY_ISGE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISGE_INT32(void)
{
  return &(GxB_ANY_ISGE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISGE_INT64(void)
{
  return &(GxB_ANY_ISGE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISGE_INT8(void)
{
  return &(GxB_ANY_ISGE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISGE_UINT16(void)
{
  return &(GxB_ANY_ISGE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISGE_UINT32(void)
{
  return &(GxB_ANY_ISGE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISGE_UINT64(void)
{
  return &(GxB_ANY_ISGE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISGE_UINT8(void)
{
  return &(GxB_ANY_ISGE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISGT_FP32(void)
{
  return &(GxB_ANY_ISGT_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISGT_FP64(void)
{
  return &(GxB_ANY_ISGT_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISGT_INT16(void)
{
  return &(GxB_ANY_ISGT_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISGT_INT32(void)
{
  return &(GxB_ANY_ISGT_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISGT_INT64(void)
{
  return &(GxB_ANY_ISGT_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISGT_INT8(void)
{
  return &(GxB_ANY_ISGT_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISGT_UINT16(void)
{
  return &(GxB_ANY_ISGT_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISGT_UINT32(void)
{
  return &(GxB_ANY_ISGT_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISGT_UINT64(void)
{
  return &(GxB_ANY_ISGT_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISGT_UINT8(void)
{
  return &(GxB_ANY_ISGT_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISLE_FP32(void)
{
  return &(GxB_ANY_ISLE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISLE_FP64(void)
{
  return &(GxB_ANY_ISLE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISLE_INT16(void)
{
  return &(GxB_ANY_ISLE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISLE_INT32(void)
{
  return &(GxB_ANY_ISLE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISLE_INT64(void)
{
  return &(GxB_ANY_ISLE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISLE_INT8(void)
{
  return &(GxB_ANY_ISLE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISLE_UINT16(void)
{
  return &(GxB_ANY_ISLE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISLE_UINT32(void)
{
  return &(GxB_ANY_ISLE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISLE_UINT64(void)
{
  return &(GxB_ANY_ISLE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISLE_UINT8(void)
{
  return &(GxB_ANY_ISLE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISLT_FP32(void)
{
  return &(GxB_ANY_ISLT_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISLT_FP64(void)
{
  return &(GxB_ANY_ISLT_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISLT_INT16(void)
{
  return &(GxB_ANY_ISLT_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISLT_INT32(void)
{
  return &(GxB_ANY_ISLT_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISLT_INT64(void)
{
  return &(GxB_ANY_ISLT_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISLT_INT8(void)
{
  return &(GxB_ANY_ISLT_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISLT_UINT16(void)
{
  return &(GxB_ANY_ISLT_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISLT_UINT32(void)
{
  return &(GxB_ANY_ISLT_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISLT_UINT64(void)
{
  return &(GxB_ANY_ISLT_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISLT_UINT8(void)
{
  return &(GxB_ANY_ISLT_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISNE_FP32(void)
{
  return &(GxB_ANY_ISNE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISNE_FP64(void)
{
  return &(GxB_ANY_ISNE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISNE_INT16(void)
{
  return &(GxB_ANY_ISNE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISNE_INT32(void)
{
  return &(GxB_ANY_ISNE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISNE_INT64(void)
{
  return &(GxB_ANY_ISNE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISNE_INT8(void)
{
  return &(GxB_ANY_ISNE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISNE_UINT16(void)
{
  return &(GxB_ANY_ISNE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISNE_UINT32(void)
{
  return &(GxB_ANY_ISNE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISNE_UINT64(void)
{
  return &(GxB_ANY_ISNE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_ISNE_UINT8(void)
{
  return &(GxB_ANY_ISNE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LAND_BOOL(void)
{
  return &(GxB_ANY_LAND_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LAND_FP32(void)
{
  return &(GxB_ANY_LAND_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LAND_FP64(void)
{
  return &(GxB_ANY_LAND_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LAND_INT16(void)
{
  return &(GxB_ANY_LAND_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LAND_INT32(void)
{
  return &(GxB_ANY_LAND_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LAND_INT64(void)
{
  return &(GxB_ANY_LAND_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LAND_INT8(void)
{
  return &(GxB_ANY_LAND_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LAND_UINT16(void)
{
  return &(GxB_ANY_LAND_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LAND_UINT32(void)
{
  return &(GxB_ANY_LAND_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LAND_UINT64(void)
{
  return &(GxB_ANY_LAND_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LAND_UINT8(void)
{
  return &(GxB_ANY_LAND_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LE_BOOL(void)
{
  return &(GxB_ANY_LE_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LE_FP32(void)
{
  return &(GxB_ANY_LE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LE_FP64(void)
{
  return &(GxB_ANY_LE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LE_INT16(void)
{
  return &(GxB_ANY_LE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LE_INT32(void)
{
  return &(GxB_ANY_LE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LE_INT64(void)
{
  return &(GxB_ANY_LE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LE_INT8(void)
{
  return &(GxB_ANY_LE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LE_UINT16(void)
{
  return &(GxB_ANY_LE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LE_UINT32(void)
{
  return &(GxB_ANY_LE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LE_UINT64(void)
{
  return &(GxB_ANY_LE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LE_UINT8(void)
{
  return &(GxB_ANY_LE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LOR_BOOL(void)
{
  return &(GxB_ANY_LOR_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LOR_FP32(void)
{
  return &(GxB_ANY_LOR_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LOR_FP64(void)
{
  return &(GxB_ANY_LOR_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LOR_INT16(void)
{
  return &(GxB_ANY_LOR_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LOR_INT32(void)
{
  return &(GxB_ANY_LOR_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LOR_INT64(void)
{
  return &(GxB_ANY_LOR_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LOR_INT8(void)
{
  return &(GxB_ANY_LOR_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LOR_UINT16(void)
{
  return &(GxB_ANY_LOR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LOR_UINT32(void)
{
  return &(GxB_ANY_LOR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LOR_UINT64(void)
{
  return &(GxB_ANY_LOR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LOR_UINT8(void)
{
  return &(GxB_ANY_LOR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LT_BOOL(void)
{
  return &(GxB_ANY_LT_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LT_FP32(void)
{
  return &(GxB_ANY_LT_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LT_FP64(void)
{
  return &(GxB_ANY_LT_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LT_INT16(void)
{
  return &(GxB_ANY_LT_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LT_INT32(void)
{
  return &(GxB_ANY_LT_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LT_INT64(void)
{
  return &(GxB_ANY_LT_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LT_INT8(void)
{
  return &(GxB_ANY_LT_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LT_UINT16(void)
{
  return &(GxB_ANY_LT_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LT_UINT32(void)
{
  return &(GxB_ANY_LT_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LT_UINT64(void)
{
  return &(GxB_ANY_LT_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LT_UINT8(void)
{
  return &(GxB_ANY_LT_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LXOR_BOOL(void)
{
  return &(GxB_ANY_LXOR_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LXOR_FP32(void)
{
  return &(GxB_ANY_LXOR_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LXOR_FP64(void)
{
  return &(GxB_ANY_LXOR_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LXOR_INT16(void)
{
  return &(GxB_ANY_LXOR_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LXOR_INT32(void)
{
  return &(GxB_ANY_LXOR_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LXOR_INT64(void)
{
  return &(GxB_ANY_LXOR_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LXOR_INT8(void)
{
  return &(GxB_ANY_LXOR_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LXOR_UINT16(void)
{
  return &(GxB_ANY_LXOR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LXOR_UINT32(void)
{
  return &(GxB_ANY_LXOR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LXOR_UINT64(void)
{
  return &(GxB_ANY_LXOR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_LXOR_UINT8(void)
{
  return &(GxB_ANY_LXOR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MAX_FP32(void)
{
  return &(GxB_ANY_MAX_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MAX_FP64(void)
{
  return &(GxB_ANY_MAX_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MAX_INT16(void)
{
  return &(GxB_ANY_MAX_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MAX_INT32(void)
{
  return &(GxB_ANY_MAX_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MAX_INT64(void)
{
  return &(GxB_ANY_MAX_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MAX_INT8(void)
{
  return &(GxB_ANY_MAX_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MAX_UINT16(void)
{
  return &(GxB_ANY_MAX_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MAX_UINT32(void)
{
  return &(GxB_ANY_MAX_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MAX_UINT64(void)
{
  return &(GxB_ANY_MAX_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MAX_UINT8(void)
{
  return &(GxB_ANY_MAX_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MINUS_FC32(void)
{
  return &(GxB_ANY_MINUS_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MINUS_FC64(void)
{
  return &(GxB_ANY_MINUS_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MINUS_FP32(void)
{
  return &(GxB_ANY_MINUS_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MINUS_FP64(void)
{
  return &(GxB_ANY_MINUS_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MINUS_INT16(void)
{
  return &(GxB_ANY_MINUS_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MINUS_INT32(void)
{
  return &(GxB_ANY_MINUS_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MINUS_INT64(void)
{
  return &(GxB_ANY_MINUS_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MINUS_INT8(void)
{
  return &(GxB_ANY_MINUS_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MINUS_UINT16(void)
{
  return &(GxB_ANY_MINUS_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MINUS_UINT32(void)
{
  return &(GxB_ANY_MINUS_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MINUS_UINT64(void)
{
  return &(GxB_ANY_MINUS_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MINUS_UINT8(void)
{
  return &(GxB_ANY_MINUS_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MIN_FP32(void)
{
  return &(GxB_ANY_MIN_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MIN_FP64(void)
{
  return &(GxB_ANY_MIN_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MIN_INT16(void)
{
  return &(GxB_ANY_MIN_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MIN_INT32(void)
{
  return &(GxB_ANY_MIN_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MIN_INT64(void)
{
  return &(GxB_ANY_MIN_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MIN_INT8(void)
{
  return &(GxB_ANY_MIN_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MIN_UINT16(void)
{
  return &(GxB_ANY_MIN_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MIN_UINT32(void)
{
  return &(GxB_ANY_MIN_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MIN_UINT64(void)
{
  return &(GxB_ANY_MIN_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_MIN_UINT8(void)
{
  return &(GxB_ANY_MIN_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_NE_FP32(void)
{
  return &(GxB_ANY_NE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_NE_FP64(void)
{
  return &(GxB_ANY_NE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_NE_INT16(void)
{
  return &(GxB_ANY_NE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_NE_INT32(void)
{
  return &(GxB_ANY_NE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_NE_INT64(void)
{
  return &(GxB_ANY_NE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_NE_INT8(void)
{
  return &(GxB_ANY_NE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_NE_UINT16(void)
{
  return &(GxB_ANY_NE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_NE_UINT32(void)
{
  return &(GxB_ANY_NE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_NE_UINT64(void)
{
  return &(GxB_ANY_NE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_NE_UINT8(void)
{
  return &(GxB_ANY_NE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PAIR_BOOL(void)
{
  return &(GxB_ANY_PAIR_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PAIR_FC32(void)
{
  return &(GxB_ANY_PAIR_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PAIR_FC64(void)
{
  return &(GxB_ANY_PAIR_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PAIR_FP32(void)
{
  return &(GxB_ANY_PAIR_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PAIR_FP64(void)
{
  return &(GxB_ANY_PAIR_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PAIR_INT16(void)
{
  return &(GxB_ANY_PAIR_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PAIR_INT32(void)
{
  return &(GxB_ANY_PAIR_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PAIR_INT64(void)
{
  return &(GxB_ANY_PAIR_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PAIR_INT8(void)
{
  return &(GxB_ANY_PAIR_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PAIR_UINT16(void)
{
  return &(GxB_ANY_PAIR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PAIR_UINT32(void)
{
  return &(GxB_ANY_PAIR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PAIR_UINT64(void)
{
  return &(GxB_ANY_PAIR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PAIR_UINT8(void)
{
  return &(GxB_ANY_PAIR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PLUS_FC32(void)
{
  return &(GxB_ANY_PLUS_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PLUS_FC64(void)
{
  return &(GxB_ANY_PLUS_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PLUS_FP32(void)
{
  return &(GxB_ANY_PLUS_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PLUS_FP64(void)
{
  return &(GxB_ANY_PLUS_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PLUS_INT16(void)
{
  return &(GxB_ANY_PLUS_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PLUS_INT32(void)
{
  return &(GxB_ANY_PLUS_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PLUS_INT64(void)
{
  return &(GxB_ANY_PLUS_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PLUS_INT8(void)
{
  return &(GxB_ANY_PLUS_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PLUS_UINT16(void)
{
  return &(GxB_ANY_PLUS_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PLUS_UINT32(void)
{
  return &(GxB_ANY_PLUS_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PLUS_UINT64(void)
{
  return &(GxB_ANY_PLUS_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_PLUS_UINT8(void)
{
  return &(GxB_ANY_PLUS_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RDIV_FC32(void)
{
  return &(GxB_ANY_RDIV_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RDIV_FC64(void)
{
  return &(GxB_ANY_RDIV_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RDIV_FP32(void)
{
  return &(GxB_ANY_RDIV_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RDIV_FP64(void)
{
  return &(GxB_ANY_RDIV_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RDIV_INT16(void)
{
  return &(GxB_ANY_RDIV_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RDIV_INT32(void)
{
  return &(GxB_ANY_RDIV_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RDIV_INT64(void)
{
  return &(GxB_ANY_RDIV_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RDIV_INT8(void)
{
  return &(GxB_ANY_RDIV_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RDIV_UINT16(void)
{
  return &(GxB_ANY_RDIV_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RDIV_UINT32(void)
{
  return &(GxB_ANY_RDIV_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RDIV_UINT64(void)
{
  return &(GxB_ANY_RDIV_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RDIV_UINT8(void)
{
  return &(GxB_ANY_RDIV_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RMINUS_FC32(void)
{
  return &(GxB_ANY_RMINUS_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RMINUS_FC64(void)
{
  return &(GxB_ANY_RMINUS_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RMINUS_FP32(void)
{
  return &(GxB_ANY_RMINUS_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RMINUS_FP64(void)
{
  return &(GxB_ANY_RMINUS_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RMINUS_INT16(void)
{
  return &(GxB_ANY_RMINUS_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RMINUS_INT32(void)
{
  return &(GxB_ANY_RMINUS_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RMINUS_INT64(void)
{
  return &(GxB_ANY_RMINUS_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RMINUS_INT8(void)
{
  return &(GxB_ANY_RMINUS_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RMINUS_UINT16(void)
{
  return &(GxB_ANY_RMINUS_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RMINUS_UINT32(void)
{
  return &(GxB_ANY_RMINUS_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RMINUS_UINT64(void)
{
  return &(GxB_ANY_RMINUS_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_RMINUS_UINT8(void)
{
  return &(GxB_ANY_RMINUS_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_SECONDI1_INT32(void)
{
  return &(GxB_ANY_SECONDI1_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_SECONDI1_INT64(void)
{
  return &(GxB_ANY_SECONDI1_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_SECONDI_INT32(void)
{
  return &(GxB_ANY_SECONDI_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_SECONDI_INT64(void)
{
  return &(GxB_ANY_SECONDI_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_SECONDJ1_INT32(void)
{
  return &(GxB_ANY_SECONDJ1_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_SECONDJ1_INT64(void)
{
  return &(GxB_ANY_SECONDJ1_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_SECONDJ_INT32(void)
{
  return &(GxB_ANY_SECONDJ_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_SECONDJ_INT64(void)
{
  return &(GxB_ANY_SECONDJ_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_SECOND_BOOL(void)
{
  return &(GxB_ANY_SECOND_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_SECOND_FC32(void)
{
  return &(GxB_ANY_SECOND_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_SECOND_FC64(void)
{
  return &(GxB_ANY_SECOND_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_SECOND_FP32(void)
{
  return &(GxB_ANY_SECOND_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_SECOND_FP64(void)
{
  return &(GxB_ANY_SECOND_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_SECOND_INT16(void)
{
  return &(GxB_ANY_SECOND_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_SECOND_INT32(void)
{
  return &(GxB_ANY_SECOND_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_SECOND_INT64(void)
{
  return &(GxB_ANY_SECOND_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_SECOND_INT8(void)
{
  return &(GxB_ANY_SECOND_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_SECOND_UINT16(void)
{
  return &(GxB_ANY_SECOND_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_SECOND_UINT32(void)
{
  return &(GxB_ANY_SECOND_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_SECOND_UINT64(void)
{
  return &(GxB_ANY_SECOND_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_SECOND_UINT8(void)
{
  return &(GxB_ANY_SECOND_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_TIMES_FC32(void)
{
  return &(GxB_ANY_TIMES_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_TIMES_FC64(void)
{
  return &(GxB_ANY_TIMES_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_TIMES_FP32(void)
{
  return &(GxB_ANY_TIMES_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_TIMES_FP64(void)
{
  return &(GxB_ANY_TIMES_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_TIMES_INT16(void)
{
  return &(GxB_ANY_TIMES_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_TIMES_INT32(void)
{
  return &(GxB_ANY_TIMES_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_TIMES_INT64(void)
{
  return &(GxB_ANY_TIMES_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_TIMES_INT8(void)
{
  return &(GxB_ANY_TIMES_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_TIMES_UINT16(void)
{
  return &(GxB_ANY_TIMES_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_TIMES_UINT32(void)
{
  return &(GxB_ANY_TIMES_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_TIMES_UINT64(void)
{
  return &(GxB_ANY_TIMES_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_ANY_TIMES_UINT8(void)
{
  return &(GxB_ANY_TIMES_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ANY_UINT16(void)
{
  return &(GxB_ANY_UINT16);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_ANY_UINT16_MONOID(void)
{
  return &(GxB_ANY_UINT16_MONOID);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ANY_UINT32(void)
{
  return &(GxB_ANY_UINT32);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_ANY_UINT32_MONOID(void)
{
  return &(GxB_ANY_UINT32_MONOID);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ANY_UINT64(void)
{
  return &(GxB_ANY_UINT64);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_ANY_UINT64_MONOID(void)
{
  return &(GxB_ANY_UINT64_MONOID);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ANY_UINT8(void)
{
  return &(GxB_ANY_UINT8);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_ANY_UINT8_MONOID(void)
{
  return &(GxB_ANY_UINT8_MONOID);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ASINH_FC32(void)
{
  return &(GxB_ASINH_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ASINH_FC64(void)
{
  return &(GxB_ASINH_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ASINH_FP32(void)
{
  return &(GxB_ASINH_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ASINH_FP64(void)
{
  return &(GxB_ASINH_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ASIN_FC32(void)
{
  return &(GxB_ASIN_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ASIN_FC64(void)
{
  return &(GxB_ASIN_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ASIN_FP32(void)
{
  return &(GxB_ASIN_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ASIN_FP64(void)
{
  return &(GxB_ASIN_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ATAN2_FP32(void)
{
  return &(GxB_ATAN2_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ATAN2_FP64(void)
{
  return &(GxB_ATAN2_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ATANH_FC32(void)
{
  return &(GxB_ATANH_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ATANH_FC64(void)
{
  return &(GxB_ATANH_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ATANH_FP32(void)
{
  return &(GxB_ATANH_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ATANH_FP64(void)
{
  return &(GxB_ATANH_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ATAN_FC32(void)
{
  return &(GxB_ATAN_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ATAN_FC64(void)
{
  return &(GxB_ATAN_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ATAN_FP32(void)
{
  return &(GxB_ATAN_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ATAN_FP64(void)
{
  return &(GxB_ATAN_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BAND_BAND_UINT16(void)
{
  return &(GxB_BAND_BAND_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BAND_BAND_UINT32(void)
{
  return &(GxB_BAND_BAND_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BAND_BAND_UINT64(void)
{
  return &(GxB_BAND_BAND_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BAND_BAND_UINT8(void)
{
  return &(GxB_BAND_BAND_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BAND_BOR_UINT16(void)
{
  return &(GxB_BAND_BOR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BAND_BOR_UINT32(void)
{
  return &(GxB_BAND_BOR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BAND_BOR_UINT64(void)
{
  return &(GxB_BAND_BOR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BAND_BOR_UINT8(void)
{
  return &(GxB_BAND_BOR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BAND_BXNOR_UINT16(void)
{
  return &(GxB_BAND_BXNOR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BAND_BXNOR_UINT32(void)
{
  return &(GxB_BAND_BXNOR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BAND_BXNOR_UINT64(void)
{
  return &(GxB_BAND_BXNOR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BAND_BXNOR_UINT8(void)
{
  return &(GxB_BAND_BXNOR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BAND_BXOR_UINT16(void)
{
  return &(GxB_BAND_BXOR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BAND_BXOR_UINT32(void)
{
  return &(GxB_BAND_BXOR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BAND_BXOR_UINT64(void)
{
  return &(GxB_BAND_BXOR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BAND_BXOR_UINT8(void)
{
  return &(GxB_BAND_BXOR_UINT8);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_BAND_UINT16_MONOID(void)
{
  return &(GxB_BAND_UINT16_MONOID);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_BAND_UINT32_MONOID(void)
{
  return &(GxB_BAND_UINT32_MONOID);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_BAND_UINT64_MONOID(void)
{
  return &(GxB_BAND_UINT64_MONOID);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_BAND_UINT8_MONOID(void)
{
  return &(GxB_BAND_UINT8_MONOID);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BCLR_INT16(void)
{
  return &(GxB_BCLR_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BCLR_INT32(void)
{
  return &(GxB_BCLR_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BCLR_INT64(void)
{
  return &(GxB_BCLR_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BCLR_INT8(void)
{
  return &(GxB_BCLR_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BCLR_UINT16(void)
{
  return &(GxB_BCLR_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BCLR_UINT32(void)
{
  return &(GxB_BCLR_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BCLR_UINT64(void)
{
  return &(GxB_BCLR_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BCLR_UINT8(void)
{
  return &(GxB_BCLR_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BGET_INT16(void)
{
  return &(GxB_BGET_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BGET_INT32(void)
{
  return &(GxB_BGET_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BGET_INT64(void)
{
  return &(GxB_BGET_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BGET_INT8(void)
{
  return &(GxB_BGET_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BGET_UINT16(void)
{
  return &(GxB_BGET_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BGET_UINT32(void)
{
  return &(GxB_BGET_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BGET_UINT64(void)
{
  return &(GxB_BGET_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BGET_UINT8(void)
{
  return &(GxB_BGET_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BOR_BAND_UINT16(void)
{
  return &(GxB_BOR_BAND_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BOR_BAND_UINT32(void)
{
  return &(GxB_BOR_BAND_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BOR_BAND_UINT64(void)
{
  return &(GxB_BOR_BAND_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BOR_BAND_UINT8(void)
{
  return &(GxB_BOR_BAND_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BOR_BOR_UINT16(void)
{
  return &(GxB_BOR_BOR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BOR_BOR_UINT32(void)
{
  return &(GxB_BOR_BOR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BOR_BOR_UINT64(void)
{
  return &(GxB_BOR_BOR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BOR_BOR_UINT8(void)
{
  return &(GxB_BOR_BOR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BOR_BXNOR_UINT16(void)
{
  return &(GxB_BOR_BXNOR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BOR_BXNOR_UINT32(void)
{
  return &(GxB_BOR_BXNOR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BOR_BXNOR_UINT64(void)
{
  return &(GxB_BOR_BXNOR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BOR_BXNOR_UINT8(void)
{
  return &(GxB_BOR_BXNOR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BOR_BXOR_UINT16(void)
{
  return &(GxB_BOR_BXOR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BOR_BXOR_UINT32(void)
{
  return &(GxB_BOR_BXOR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BOR_BXOR_UINT64(void)
{
  return &(GxB_BOR_BXOR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BOR_BXOR_UINT8(void)
{
  return &(GxB_BOR_BXOR_UINT8);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_BOR_UINT16_MONOID(void)
{
  return &(GxB_BOR_UINT16_MONOID);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_BOR_UINT32_MONOID(void)
{
  return &(GxB_BOR_UINT32_MONOID);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_BOR_UINT64_MONOID(void)
{
  return &(GxB_BOR_UINT64_MONOID);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_BOR_UINT8_MONOID(void)
{
  return &(GxB_BOR_UINT8_MONOID);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BSET_INT16(void)
{
  return &(GxB_BSET_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BSET_INT32(void)
{
  return &(GxB_BSET_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BSET_INT64(void)
{
  return &(GxB_BSET_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BSET_INT8(void)
{
  return &(GxB_BSET_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BSET_UINT16(void)
{
  return &(GxB_BSET_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BSET_UINT32(void)
{
  return &(GxB_BSET_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BSET_UINT64(void)
{
  return &(GxB_BSET_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BSET_UINT8(void)
{
  return &(GxB_BSET_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BSHIFT_INT16(void)
{
  return &(GxB_BSHIFT_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BSHIFT_INT32(void)
{
  return &(GxB_BSHIFT_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BSHIFT_INT64(void)
{
  return &(GxB_BSHIFT_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BSHIFT_INT8(void)
{
  return &(GxB_BSHIFT_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BSHIFT_UINT16(void)
{
  return &(GxB_BSHIFT_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BSHIFT_UINT32(void)
{
  return &(GxB_BSHIFT_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BSHIFT_UINT64(void)
{
  return &(GxB_BSHIFT_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_BSHIFT_UINT8(void)
{
  return &(GxB_BSHIFT_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXNOR_BAND_UINT16(void)
{
  return &(GxB_BXNOR_BAND_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXNOR_BAND_UINT32(void)
{
  return &(GxB_BXNOR_BAND_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXNOR_BAND_UINT64(void)
{
  return &(GxB_BXNOR_BAND_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXNOR_BAND_UINT8(void)
{
  return &(GxB_BXNOR_BAND_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXNOR_BOR_UINT16(void)
{
  return &(GxB_BXNOR_BOR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXNOR_BOR_UINT32(void)
{
  return &(GxB_BXNOR_BOR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXNOR_BOR_UINT64(void)
{
  return &(GxB_BXNOR_BOR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXNOR_BOR_UINT8(void)
{
  return &(GxB_BXNOR_BOR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXNOR_BXNOR_UINT16(void)
{
  return &(GxB_BXNOR_BXNOR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXNOR_BXNOR_UINT32(void)
{
  return &(GxB_BXNOR_BXNOR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXNOR_BXNOR_UINT64(void)
{
  return &(GxB_BXNOR_BXNOR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXNOR_BXNOR_UINT8(void)
{
  return &(GxB_BXNOR_BXNOR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXNOR_BXOR_UINT16(void)
{
  return &(GxB_BXNOR_BXOR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXNOR_BXOR_UINT32(void)
{
  return &(GxB_BXNOR_BXOR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXNOR_BXOR_UINT64(void)
{
  return &(GxB_BXNOR_BXOR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXNOR_BXOR_UINT8(void)
{
  return &(GxB_BXNOR_BXOR_UINT8);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_BXNOR_UINT16_MONOID(void)
{
  return &(GxB_BXNOR_UINT16_MONOID);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_BXNOR_UINT32_MONOID(void)
{
  return &(GxB_BXNOR_UINT32_MONOID);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_BXNOR_UINT64_MONOID(void)
{
  return &(GxB_BXNOR_UINT64_MONOID);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_BXNOR_UINT8_MONOID(void)
{
  return &(GxB_BXNOR_UINT8_MONOID);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXOR_BAND_UINT16(void)
{
  return &(GxB_BXOR_BAND_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXOR_BAND_UINT32(void)
{
  return &(GxB_BXOR_BAND_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXOR_BAND_UINT64(void)
{
  return &(GxB_BXOR_BAND_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXOR_BAND_UINT8(void)
{
  return &(GxB_BXOR_BAND_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXOR_BOR_UINT16(void)
{
  return &(GxB_BXOR_BOR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXOR_BOR_UINT32(void)
{
  return &(GxB_BXOR_BOR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXOR_BOR_UINT64(void)
{
  return &(GxB_BXOR_BOR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXOR_BOR_UINT8(void)
{
  return &(GxB_BXOR_BOR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXOR_BXNOR_UINT16(void)
{
  return &(GxB_BXOR_BXNOR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXOR_BXNOR_UINT32(void)
{
  return &(GxB_BXOR_BXNOR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXOR_BXNOR_UINT64(void)
{
  return &(GxB_BXOR_BXNOR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXOR_BXNOR_UINT8(void)
{
  return &(GxB_BXOR_BXNOR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXOR_BXOR_UINT16(void)
{
  return &(GxB_BXOR_BXOR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXOR_BXOR_UINT32(void)
{
  return &(GxB_BXOR_BXOR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXOR_BXOR_UINT64(void)
{
  return &(GxB_BXOR_BXOR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_BXOR_BXOR_UINT8(void)
{
  return &(GxB_BXOR_BXOR_UINT8);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_BXOR_UINT16_MONOID(void)
{
  return &(GxB_BXOR_UINT16_MONOID);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_BXOR_UINT32_MONOID(void)
{
  return &(GxB_BXOR_UINT32_MONOID);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_BXOR_UINT64_MONOID(void)
{
  return &(GxB_BXOR_UINT64_MONOID);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_BXOR_UINT8_MONOID(void)
{
  return &(GxB_BXOR_UINT8_MONOID);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_CARG_FC32(void)
{
  return &(GxB_CARG_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_CARG_FC64(void)
{
  return &(GxB_CARG_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_CBRT_FP32(void)
{
  return &(GxB_CBRT_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_CBRT_FP64(void)
{
  return &(GxB_CBRT_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_CEIL_FC32(void)
{
  return &(GxB_CEIL_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_CEIL_FC64(void)
{
  return &(GxB_CEIL_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_CEIL_FP32(void)
{
  return &(GxB_CEIL_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_CEIL_FP64(void)
{
  return &(GxB_CEIL_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_CIMAG_FC32(void)
{
  return &(GxB_CIMAG_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_CIMAG_FC64(void)
{
  return &(GxB_CIMAG_FC64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_CMPLX_FP32(void)
{
  return &(GxB_CMPLX_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_CMPLX_FP64(void)
{
  return &(GxB_CMPLX_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_CONJ_FC32(void)
{
  return &(GxB_CONJ_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_CONJ_FC64(void)
{
  return &(GxB_CONJ_FC64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_COPYSIGN_FP32(void)
{
  return &(GxB_COPYSIGN_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_COPYSIGN_FP64(void)
{
  return &(GxB_COPYSIGN_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_COSH_FC32(void)
{
  return &(GxB_COSH_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_COSH_FC64(void)
{
  return &(GxB_COSH_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_COSH_FP32(void)
{
  return &(GxB_COSH_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_COSH_FP64(void)
{
  return &(GxB_COSH_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_COS_FC32(void)
{
  return &(GxB_COS_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_COS_FC64(void)
{
  return &(GxB_COS_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_COS_FP32(void)
{
  return &(GxB_COS_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_COS_FP64(void)
{
  return &(GxB_COS_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_CREAL_FC32(void)
{
  return &(GxB_CREAL_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_CREAL_FC64(void)
{
  return &(GxB_CREAL_FC64);
}

static struct GB_SelectOp_opaque * *_cffi_var_GxB_DIAG(void)
{
  return &(GxB_DIAG);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_DIV_FC32(void)
{
  return &(GxB_DIV_FC32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_DIV_FC64(void)
{
  return &(GxB_DIV_FC64);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_EQ_BOOL_MONOID(void)
{
  return &(GxB_EQ_BOOL_MONOID);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_EQ_BOOL(void)
{
  return &(GxB_EQ_EQ_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_EQ_FP32(void)
{
  return &(GxB_EQ_EQ_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_EQ_FP64(void)
{
  return &(GxB_EQ_EQ_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_EQ_INT16(void)
{
  return &(GxB_EQ_EQ_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_EQ_INT32(void)
{
  return &(GxB_EQ_EQ_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_EQ_INT64(void)
{
  return &(GxB_EQ_EQ_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_EQ_INT8(void)
{
  return &(GxB_EQ_EQ_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_EQ_UINT16(void)
{
  return &(GxB_EQ_EQ_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_EQ_UINT32(void)
{
  return &(GxB_EQ_EQ_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_EQ_UINT64(void)
{
  return &(GxB_EQ_EQ_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_EQ_UINT8(void)
{
  return &(GxB_EQ_EQ_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_EQ_FC32(void)
{
  return &(GxB_EQ_FC32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_EQ_FC64(void)
{
  return &(GxB_EQ_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_FIRST_BOOL(void)
{
  return &(GxB_EQ_FIRST_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_GE_BOOL(void)
{
  return &(GxB_EQ_GE_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_GE_FP32(void)
{
  return &(GxB_EQ_GE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_GE_FP64(void)
{
  return &(GxB_EQ_GE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_GE_INT16(void)
{
  return &(GxB_EQ_GE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_GE_INT32(void)
{
  return &(GxB_EQ_GE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_GE_INT64(void)
{
  return &(GxB_EQ_GE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_GE_INT8(void)
{
  return &(GxB_EQ_GE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_GE_UINT16(void)
{
  return &(GxB_EQ_GE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_GE_UINT32(void)
{
  return &(GxB_EQ_GE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_GE_UINT64(void)
{
  return &(GxB_EQ_GE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_GE_UINT8(void)
{
  return &(GxB_EQ_GE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_GT_BOOL(void)
{
  return &(GxB_EQ_GT_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_GT_FP32(void)
{
  return &(GxB_EQ_GT_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_GT_FP64(void)
{
  return &(GxB_EQ_GT_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_GT_INT16(void)
{
  return &(GxB_EQ_GT_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_GT_INT32(void)
{
  return &(GxB_EQ_GT_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_GT_INT64(void)
{
  return &(GxB_EQ_GT_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_GT_INT8(void)
{
  return &(GxB_EQ_GT_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_GT_UINT16(void)
{
  return &(GxB_EQ_GT_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_GT_UINT32(void)
{
  return &(GxB_EQ_GT_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_GT_UINT64(void)
{
  return &(GxB_EQ_GT_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_GT_UINT8(void)
{
  return &(GxB_EQ_GT_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LAND_BOOL(void)
{
  return &(GxB_EQ_LAND_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LE_BOOL(void)
{
  return &(GxB_EQ_LE_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LE_FP32(void)
{
  return &(GxB_EQ_LE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LE_FP64(void)
{
  return &(GxB_EQ_LE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LE_INT16(void)
{
  return &(GxB_EQ_LE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LE_INT32(void)
{
  return &(GxB_EQ_LE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LE_INT64(void)
{
  return &(GxB_EQ_LE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LE_INT8(void)
{
  return &(GxB_EQ_LE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LE_UINT16(void)
{
  return &(GxB_EQ_LE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LE_UINT32(void)
{
  return &(GxB_EQ_LE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LE_UINT64(void)
{
  return &(GxB_EQ_LE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LE_UINT8(void)
{
  return &(GxB_EQ_LE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LOR_BOOL(void)
{
  return &(GxB_EQ_LOR_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LT_BOOL(void)
{
  return &(GxB_EQ_LT_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LT_FP32(void)
{
  return &(GxB_EQ_LT_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LT_FP64(void)
{
  return &(GxB_EQ_LT_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LT_INT16(void)
{
  return &(GxB_EQ_LT_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LT_INT32(void)
{
  return &(GxB_EQ_LT_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LT_INT64(void)
{
  return &(GxB_EQ_LT_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LT_INT8(void)
{
  return &(GxB_EQ_LT_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LT_UINT16(void)
{
  return &(GxB_EQ_LT_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LT_UINT32(void)
{
  return &(GxB_EQ_LT_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LT_UINT64(void)
{
  return &(GxB_EQ_LT_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LT_UINT8(void)
{
  return &(GxB_EQ_LT_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_LXOR_BOOL(void)
{
  return &(GxB_EQ_LXOR_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_NE_FP32(void)
{
  return &(GxB_EQ_NE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_NE_FP64(void)
{
  return &(GxB_EQ_NE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_NE_INT16(void)
{
  return &(GxB_EQ_NE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_NE_INT32(void)
{
  return &(GxB_EQ_NE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_NE_INT64(void)
{
  return &(GxB_EQ_NE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_NE_INT8(void)
{
  return &(GxB_EQ_NE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_NE_UINT16(void)
{
  return &(GxB_EQ_NE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_NE_UINT32(void)
{
  return &(GxB_EQ_NE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_NE_UINT64(void)
{
  return &(GxB_EQ_NE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_NE_UINT8(void)
{
  return &(GxB_EQ_NE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_PAIR_BOOL(void)
{
  return &(GxB_EQ_PAIR_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_EQ_SECOND_BOOL(void)
{
  return &(GxB_EQ_SECOND_BOOL);
}

static struct GB_SelectOp_opaque * *_cffi_var_GxB_EQ_THUNK(void)
{
  return &(GxB_EQ_THUNK);
}

static struct GB_SelectOp_opaque * *_cffi_var_GxB_EQ_ZERO(void)
{
  return &(GxB_EQ_ZERO);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ERFC_FP32(void)
{
  return &(GxB_ERFC_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ERFC_FP64(void)
{
  return &(GxB_ERFC_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ERF_FP32(void)
{
  return &(GxB_ERF_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ERF_FP64(void)
{
  return &(GxB_ERF_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_EXP2_FC32(void)
{
  return &(GxB_EXP2_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_EXP2_FC64(void)
{
  return &(GxB_EXP2_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_EXP2_FP32(void)
{
  return &(GxB_EXP2_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_EXP2_FP64(void)
{
  return &(GxB_EXP2_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_EXPM1_FC32(void)
{
  return &(GxB_EXPM1_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_EXPM1_FC64(void)
{
  return &(GxB_EXPM1_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_EXPM1_FP32(void)
{
  return &(GxB_EXPM1_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_EXPM1_FP64(void)
{
  return &(GxB_EXPM1_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_EXP_FC32(void)
{
  return &(GxB_EXP_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_EXP_FC64(void)
{
  return &(GxB_EXP_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_EXP_FP32(void)
{
  return &(GxB_EXP_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_EXP_FP64(void)
{
  return &(GxB_EXP_FP64);
}

static struct GB_Type_opaque * *_cffi_var_GxB_FC32(void)
{
  return &(GxB_FC32);
}

static struct GB_Type_opaque * *_cffi_var_GxB_FC64(void)
{
  return &(GxB_FC64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_FIRSTI1_INT32(void)
{
  return &(GxB_FIRSTI1_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_FIRSTI1_INT64(void)
{
  return &(GxB_FIRSTI1_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_FIRSTI_INT32(void)
{
  return &(GxB_FIRSTI_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_FIRSTI_INT64(void)
{
  return &(GxB_FIRSTI_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_FIRSTJ1_INT32(void)
{
  return &(GxB_FIRSTJ1_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_FIRSTJ1_INT64(void)
{
  return &(GxB_FIRSTJ1_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_FIRSTJ_INT32(void)
{
  return &(GxB_FIRSTJ_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_FIRSTJ_INT64(void)
{
  return &(GxB_FIRSTJ_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_FIRST_FC32(void)
{
  return &(GxB_FIRST_FC32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_FIRST_FC64(void)
{
  return &(GxB_FIRST_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_FLOOR_FC32(void)
{
  return &(GxB_FLOOR_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_FLOOR_FC64(void)
{
  return &(GxB_FLOOR_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_FLOOR_FP32(void)
{
  return &(GxB_FLOOR_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_FLOOR_FP64(void)
{
  return &(GxB_FLOOR_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_FMOD_FP32(void)
{
  return &(GxB_FMOD_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_FMOD_FP64(void)
{
  return &(GxB_FMOD_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_FREXPE_FP32(void)
{
  return &(GxB_FREXPE_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_FREXPE_FP64(void)
{
  return &(GxB_FREXPE_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_FREXPX_FP32(void)
{
  return &(GxB_FREXPX_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_FREXPX_FP64(void)
{
  return &(GxB_FREXPX_FP64);
}

static struct GB_SelectOp_opaque * *_cffi_var_GxB_GE_THUNK(void)
{
  return &(GxB_GE_THUNK);
}

static struct GB_SelectOp_opaque * *_cffi_var_GxB_GE_ZERO(void)
{
  return &(GxB_GE_ZERO);
}

static struct GB_SelectOp_opaque * *_cffi_var_GxB_GT_THUNK(void)
{
  return &(GxB_GT_THUNK);
}

static struct GB_SelectOp_opaque * *_cffi_var_GxB_GT_ZERO(void)
{
  return &(GxB_GT_ZERO);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_HYPOT_FP32(void)
{
  return &(GxB_HYPOT_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_HYPOT_FP64(void)
{
  return &(GxB_HYPOT_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_IDENTITY_FC32(void)
{
  return &(GxB_IDENTITY_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_IDENTITY_FC64(void)
{
  return &(GxB_IDENTITY_FC64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_IGNORE_DUP(void)
{
  return &(GxB_IGNORE_DUP);
}

static char * *_cffi_var_GxB_IMPLEMENTATION_ABOUT_STR(void)
{
  return &(GxB_IMPLEMENTATION_ABOUT_STR);
}

static char * *_cffi_var_GxB_IMPLEMENTATION_DATE_STR(void)
{
  return &(GxB_IMPLEMENTATION_DATE_STR);
}

static char * *_cffi_var_GxB_IMPLEMENTATION_LICENSE_STR(void)
{
  return &(GxB_IMPLEMENTATION_LICENSE_STR);
}

static char * *_cffi_var_GxB_IMPLEMENTATION_NAME_STR(void)
{
  return &(GxB_IMPLEMENTATION_NAME_STR);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISEQ_BOOL(void)
{
  return &(GxB_ISEQ_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISEQ_FC32(void)
{
  return &(GxB_ISEQ_FC32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISEQ_FC64(void)
{
  return &(GxB_ISEQ_FC64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISEQ_FP32(void)
{
  return &(GxB_ISEQ_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISEQ_FP64(void)
{
  return &(GxB_ISEQ_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISEQ_INT16(void)
{
  return &(GxB_ISEQ_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISEQ_INT32(void)
{
  return &(GxB_ISEQ_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISEQ_INT64(void)
{
  return &(GxB_ISEQ_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISEQ_INT8(void)
{
  return &(GxB_ISEQ_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISEQ_UINT16(void)
{
  return &(GxB_ISEQ_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISEQ_UINT32(void)
{
  return &(GxB_ISEQ_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISEQ_UINT64(void)
{
  return &(GxB_ISEQ_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISEQ_UINT8(void)
{
  return &(GxB_ISEQ_UINT8);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ISFINITE_FC32(void)
{
  return &(GxB_ISFINITE_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ISFINITE_FC64(void)
{
  return &(GxB_ISFINITE_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ISFINITE_FP32(void)
{
  return &(GxB_ISFINITE_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ISFINITE_FP64(void)
{
  return &(GxB_ISFINITE_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISGE_BOOL(void)
{
  return &(GxB_ISGE_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISGE_FP32(void)
{
  return &(GxB_ISGE_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISGE_FP64(void)
{
  return &(GxB_ISGE_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISGE_INT16(void)
{
  return &(GxB_ISGE_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISGE_INT32(void)
{
  return &(GxB_ISGE_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISGE_INT64(void)
{
  return &(GxB_ISGE_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISGE_INT8(void)
{
  return &(GxB_ISGE_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISGE_UINT16(void)
{
  return &(GxB_ISGE_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISGE_UINT32(void)
{
  return &(GxB_ISGE_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISGE_UINT64(void)
{
  return &(GxB_ISGE_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISGE_UINT8(void)
{
  return &(GxB_ISGE_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISGT_BOOL(void)
{
  return &(GxB_ISGT_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISGT_FP32(void)
{
  return &(GxB_ISGT_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISGT_FP64(void)
{
  return &(GxB_ISGT_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISGT_INT16(void)
{
  return &(GxB_ISGT_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISGT_INT32(void)
{
  return &(GxB_ISGT_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISGT_INT64(void)
{
  return &(GxB_ISGT_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISGT_INT8(void)
{
  return &(GxB_ISGT_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISGT_UINT16(void)
{
  return &(GxB_ISGT_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISGT_UINT32(void)
{
  return &(GxB_ISGT_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISGT_UINT64(void)
{
  return &(GxB_ISGT_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISGT_UINT8(void)
{
  return &(GxB_ISGT_UINT8);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ISINF_FC32(void)
{
  return &(GxB_ISINF_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ISINF_FC64(void)
{
  return &(GxB_ISINF_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ISINF_FP32(void)
{
  return &(GxB_ISINF_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ISINF_FP64(void)
{
  return &(GxB_ISINF_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISLE_BOOL(void)
{
  return &(GxB_ISLE_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISLE_FP32(void)
{
  return &(GxB_ISLE_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISLE_FP64(void)
{
  return &(GxB_ISLE_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISLE_INT16(void)
{
  return &(GxB_ISLE_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISLE_INT32(void)
{
  return &(GxB_ISLE_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISLE_INT64(void)
{
  return &(GxB_ISLE_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISLE_INT8(void)
{
  return &(GxB_ISLE_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISLE_UINT16(void)
{
  return &(GxB_ISLE_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISLE_UINT32(void)
{
  return &(GxB_ISLE_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISLE_UINT64(void)
{
  return &(GxB_ISLE_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISLE_UINT8(void)
{
  return &(GxB_ISLE_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISLT_BOOL(void)
{
  return &(GxB_ISLT_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISLT_FP32(void)
{
  return &(GxB_ISLT_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISLT_FP64(void)
{
  return &(GxB_ISLT_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISLT_INT16(void)
{
  return &(GxB_ISLT_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISLT_INT32(void)
{
  return &(GxB_ISLT_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISLT_INT64(void)
{
  return &(GxB_ISLT_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISLT_INT8(void)
{
  return &(GxB_ISLT_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISLT_UINT16(void)
{
  return &(GxB_ISLT_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISLT_UINT32(void)
{
  return &(GxB_ISLT_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISLT_UINT64(void)
{
  return &(GxB_ISLT_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISLT_UINT8(void)
{
  return &(GxB_ISLT_UINT8);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ISNAN_FC32(void)
{
  return &(GxB_ISNAN_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ISNAN_FC64(void)
{
  return &(GxB_ISNAN_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ISNAN_FP32(void)
{
  return &(GxB_ISNAN_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ISNAN_FP64(void)
{
  return &(GxB_ISNAN_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISNE_BOOL(void)
{
  return &(GxB_ISNE_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISNE_FC32(void)
{
  return &(GxB_ISNE_FC32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISNE_FC64(void)
{
  return &(GxB_ISNE_FC64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISNE_FP32(void)
{
  return &(GxB_ISNE_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISNE_FP64(void)
{
  return &(GxB_ISNE_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISNE_INT16(void)
{
  return &(GxB_ISNE_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISNE_INT32(void)
{
  return &(GxB_ISNE_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISNE_INT64(void)
{
  return &(GxB_ISNE_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISNE_INT8(void)
{
  return &(GxB_ISNE_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISNE_UINT16(void)
{
  return &(GxB_ISNE_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISNE_UINT32(void)
{
  return &(GxB_ISNE_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISNE_UINT64(void)
{
  return &(GxB_ISNE_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ISNE_UINT8(void)
{
  return &(GxB_ISNE_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LAND_BOOL(void)
{
  return &(GxB_LAND_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_EQ_BOOL(void)
{
  return &(GxB_LAND_EQ_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_EQ_FP32(void)
{
  return &(GxB_LAND_EQ_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_EQ_FP64(void)
{
  return &(GxB_LAND_EQ_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_EQ_INT16(void)
{
  return &(GxB_LAND_EQ_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_EQ_INT32(void)
{
  return &(GxB_LAND_EQ_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_EQ_INT64(void)
{
  return &(GxB_LAND_EQ_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_EQ_INT8(void)
{
  return &(GxB_LAND_EQ_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_EQ_UINT16(void)
{
  return &(GxB_LAND_EQ_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_EQ_UINT32(void)
{
  return &(GxB_LAND_EQ_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_EQ_UINT64(void)
{
  return &(GxB_LAND_EQ_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_EQ_UINT8(void)
{
  return &(GxB_LAND_EQ_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_FIRST_BOOL(void)
{
  return &(GxB_LAND_FIRST_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LAND_FP32(void)
{
  return &(GxB_LAND_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LAND_FP64(void)
{
  return &(GxB_LAND_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_GE_BOOL(void)
{
  return &(GxB_LAND_GE_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_GE_FP32(void)
{
  return &(GxB_LAND_GE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_GE_FP64(void)
{
  return &(GxB_LAND_GE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_GE_INT16(void)
{
  return &(GxB_LAND_GE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_GE_INT32(void)
{
  return &(GxB_LAND_GE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_GE_INT64(void)
{
  return &(GxB_LAND_GE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_GE_INT8(void)
{
  return &(GxB_LAND_GE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_GE_UINT16(void)
{
  return &(GxB_LAND_GE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_GE_UINT32(void)
{
  return &(GxB_LAND_GE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_GE_UINT64(void)
{
  return &(GxB_LAND_GE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_GE_UINT8(void)
{
  return &(GxB_LAND_GE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_GT_BOOL(void)
{
  return &(GxB_LAND_GT_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_GT_FP32(void)
{
  return &(GxB_LAND_GT_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_GT_FP64(void)
{
  return &(GxB_LAND_GT_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_GT_INT16(void)
{
  return &(GxB_LAND_GT_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_GT_INT32(void)
{
  return &(GxB_LAND_GT_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_GT_INT64(void)
{
  return &(GxB_LAND_GT_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_GT_INT8(void)
{
  return &(GxB_LAND_GT_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_GT_UINT16(void)
{
  return &(GxB_LAND_GT_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_GT_UINT32(void)
{
  return &(GxB_LAND_GT_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_GT_UINT64(void)
{
  return &(GxB_LAND_GT_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_GT_UINT8(void)
{
  return &(GxB_LAND_GT_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LAND_INT16(void)
{
  return &(GxB_LAND_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LAND_INT32(void)
{
  return &(GxB_LAND_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LAND_INT64(void)
{
  return &(GxB_LAND_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LAND_INT8(void)
{
  return &(GxB_LAND_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LAND_BOOL(void)
{
  return &(GxB_LAND_LAND_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LE_BOOL(void)
{
  return &(GxB_LAND_LE_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LE_FP32(void)
{
  return &(GxB_LAND_LE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LE_FP64(void)
{
  return &(GxB_LAND_LE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LE_INT16(void)
{
  return &(GxB_LAND_LE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LE_INT32(void)
{
  return &(GxB_LAND_LE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LE_INT64(void)
{
  return &(GxB_LAND_LE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LE_INT8(void)
{
  return &(GxB_LAND_LE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LE_UINT16(void)
{
  return &(GxB_LAND_LE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LE_UINT32(void)
{
  return &(GxB_LAND_LE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LE_UINT64(void)
{
  return &(GxB_LAND_LE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LE_UINT8(void)
{
  return &(GxB_LAND_LE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LT_BOOL(void)
{
  return &(GxB_LAND_LT_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LT_FP32(void)
{
  return &(GxB_LAND_LT_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LT_FP64(void)
{
  return &(GxB_LAND_LT_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LT_INT16(void)
{
  return &(GxB_LAND_LT_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LT_INT32(void)
{
  return &(GxB_LAND_LT_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LT_INT64(void)
{
  return &(GxB_LAND_LT_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LT_INT8(void)
{
  return &(GxB_LAND_LT_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LT_UINT16(void)
{
  return &(GxB_LAND_LT_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LT_UINT32(void)
{
  return &(GxB_LAND_LT_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LT_UINT64(void)
{
  return &(GxB_LAND_LT_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LT_UINT8(void)
{
  return &(GxB_LAND_LT_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_LXOR_BOOL(void)
{
  return &(GxB_LAND_LXOR_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_NE_FP32(void)
{
  return &(GxB_LAND_NE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_NE_FP64(void)
{
  return &(GxB_LAND_NE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_NE_INT16(void)
{
  return &(GxB_LAND_NE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_NE_INT32(void)
{
  return &(GxB_LAND_NE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_NE_INT64(void)
{
  return &(GxB_LAND_NE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_NE_INT8(void)
{
  return &(GxB_LAND_NE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_NE_UINT16(void)
{
  return &(GxB_LAND_NE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_NE_UINT32(void)
{
  return &(GxB_LAND_NE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_NE_UINT64(void)
{
  return &(GxB_LAND_NE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_NE_UINT8(void)
{
  return &(GxB_LAND_NE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_PAIR_BOOL(void)
{
  return &(GxB_LAND_PAIR_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LAND_SECOND_BOOL(void)
{
  return &(GxB_LAND_SECOND_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LAND_UINT16(void)
{
  return &(GxB_LAND_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LAND_UINT32(void)
{
  return &(GxB_LAND_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LAND_UINT64(void)
{
  return &(GxB_LAND_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LAND_UINT8(void)
{
  return &(GxB_LAND_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LDEXP_FP32(void)
{
  return &(GxB_LDEXP_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LDEXP_FP64(void)
{
  return &(GxB_LDEXP_FP64);
}

static struct GB_SelectOp_opaque * *_cffi_var_GxB_LE_THUNK(void)
{
  return &(GxB_LE_THUNK);
}

static struct GB_SelectOp_opaque * *_cffi_var_GxB_LE_ZERO(void)
{
  return &(GxB_LE_ZERO);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LGAMMA_FP32(void)
{
  return &(GxB_LGAMMA_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LGAMMA_FP64(void)
{
  return &(GxB_LGAMMA_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LNOT_BOOL(void)
{
  return &(GxB_LNOT_BOOL);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LNOT_FP32(void)
{
  return &(GxB_LNOT_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LNOT_FP64(void)
{
  return &(GxB_LNOT_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LNOT_INT16(void)
{
  return &(GxB_LNOT_INT16);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LNOT_INT32(void)
{
  return &(GxB_LNOT_INT32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LNOT_INT64(void)
{
  return &(GxB_LNOT_INT64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LNOT_INT8(void)
{
  return &(GxB_LNOT_INT8);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LNOT_UINT16(void)
{
  return &(GxB_LNOT_UINT16);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LNOT_UINT32(void)
{
  return &(GxB_LNOT_UINT32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LNOT_UINT64(void)
{
  return &(GxB_LNOT_UINT64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LNOT_UINT8(void)
{
  return &(GxB_LNOT_UINT8);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LOG10_FC32(void)
{
  return &(GxB_LOG10_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LOG10_FC64(void)
{
  return &(GxB_LOG10_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LOG10_FP32(void)
{
  return &(GxB_LOG10_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LOG10_FP64(void)
{
  return &(GxB_LOG10_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LOG1P_FC32(void)
{
  return &(GxB_LOG1P_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LOG1P_FC64(void)
{
  return &(GxB_LOG1P_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LOG1P_FP32(void)
{
  return &(GxB_LOG1P_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LOG1P_FP64(void)
{
  return &(GxB_LOG1P_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LOG2_FC32(void)
{
  return &(GxB_LOG2_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LOG2_FC64(void)
{
  return &(GxB_LOG2_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LOG2_FP32(void)
{
  return &(GxB_LOG2_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LOG2_FP64(void)
{
  return &(GxB_LOG2_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LOG_FC32(void)
{
  return &(GxB_LOG_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LOG_FC64(void)
{
  return &(GxB_LOG_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LOG_FP32(void)
{
  return &(GxB_LOG_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_LOG_FP64(void)
{
  return &(GxB_LOG_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LOR_BOOL(void)
{
  return &(GxB_LOR_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_EQ_BOOL(void)
{
  return &(GxB_LOR_EQ_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_EQ_FP32(void)
{
  return &(GxB_LOR_EQ_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_EQ_FP64(void)
{
  return &(GxB_LOR_EQ_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_EQ_INT16(void)
{
  return &(GxB_LOR_EQ_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_EQ_INT32(void)
{
  return &(GxB_LOR_EQ_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_EQ_INT64(void)
{
  return &(GxB_LOR_EQ_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_EQ_INT8(void)
{
  return &(GxB_LOR_EQ_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_EQ_UINT16(void)
{
  return &(GxB_LOR_EQ_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_EQ_UINT32(void)
{
  return &(GxB_LOR_EQ_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_EQ_UINT64(void)
{
  return &(GxB_LOR_EQ_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_EQ_UINT8(void)
{
  return &(GxB_LOR_EQ_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_FIRST_BOOL(void)
{
  return &(GxB_LOR_FIRST_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LOR_FP32(void)
{
  return &(GxB_LOR_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LOR_FP64(void)
{
  return &(GxB_LOR_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_GE_BOOL(void)
{
  return &(GxB_LOR_GE_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_GE_FP32(void)
{
  return &(GxB_LOR_GE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_GE_FP64(void)
{
  return &(GxB_LOR_GE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_GE_INT16(void)
{
  return &(GxB_LOR_GE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_GE_INT32(void)
{
  return &(GxB_LOR_GE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_GE_INT64(void)
{
  return &(GxB_LOR_GE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_GE_INT8(void)
{
  return &(GxB_LOR_GE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_GE_UINT16(void)
{
  return &(GxB_LOR_GE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_GE_UINT32(void)
{
  return &(GxB_LOR_GE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_GE_UINT64(void)
{
  return &(GxB_LOR_GE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_GE_UINT8(void)
{
  return &(GxB_LOR_GE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_GT_BOOL(void)
{
  return &(GxB_LOR_GT_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_GT_FP32(void)
{
  return &(GxB_LOR_GT_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_GT_FP64(void)
{
  return &(GxB_LOR_GT_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_GT_INT16(void)
{
  return &(GxB_LOR_GT_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_GT_INT32(void)
{
  return &(GxB_LOR_GT_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_GT_INT64(void)
{
  return &(GxB_LOR_GT_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_GT_INT8(void)
{
  return &(GxB_LOR_GT_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_GT_UINT16(void)
{
  return &(GxB_LOR_GT_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_GT_UINT32(void)
{
  return &(GxB_LOR_GT_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_GT_UINT64(void)
{
  return &(GxB_LOR_GT_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_GT_UINT8(void)
{
  return &(GxB_LOR_GT_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LOR_INT16(void)
{
  return &(GxB_LOR_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LOR_INT32(void)
{
  return &(GxB_LOR_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LOR_INT64(void)
{
  return &(GxB_LOR_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LOR_INT8(void)
{
  return &(GxB_LOR_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LE_BOOL(void)
{
  return &(GxB_LOR_LE_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LE_FP32(void)
{
  return &(GxB_LOR_LE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LE_FP64(void)
{
  return &(GxB_LOR_LE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LE_INT16(void)
{
  return &(GxB_LOR_LE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LE_INT32(void)
{
  return &(GxB_LOR_LE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LE_INT64(void)
{
  return &(GxB_LOR_LE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LE_INT8(void)
{
  return &(GxB_LOR_LE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LE_UINT16(void)
{
  return &(GxB_LOR_LE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LE_UINT32(void)
{
  return &(GxB_LOR_LE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LE_UINT64(void)
{
  return &(GxB_LOR_LE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LE_UINT8(void)
{
  return &(GxB_LOR_LE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LOR_BOOL(void)
{
  return &(GxB_LOR_LOR_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LT_BOOL(void)
{
  return &(GxB_LOR_LT_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LT_FP32(void)
{
  return &(GxB_LOR_LT_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LT_FP64(void)
{
  return &(GxB_LOR_LT_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LT_INT16(void)
{
  return &(GxB_LOR_LT_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LT_INT32(void)
{
  return &(GxB_LOR_LT_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LT_INT64(void)
{
  return &(GxB_LOR_LT_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LT_INT8(void)
{
  return &(GxB_LOR_LT_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LT_UINT16(void)
{
  return &(GxB_LOR_LT_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LT_UINT32(void)
{
  return &(GxB_LOR_LT_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LT_UINT64(void)
{
  return &(GxB_LOR_LT_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LT_UINT8(void)
{
  return &(GxB_LOR_LT_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_LXOR_BOOL(void)
{
  return &(GxB_LOR_LXOR_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_NE_FP32(void)
{
  return &(GxB_LOR_NE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_NE_FP64(void)
{
  return &(GxB_LOR_NE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_NE_INT16(void)
{
  return &(GxB_LOR_NE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_NE_INT32(void)
{
  return &(GxB_LOR_NE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_NE_INT64(void)
{
  return &(GxB_LOR_NE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_NE_INT8(void)
{
  return &(GxB_LOR_NE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_NE_UINT16(void)
{
  return &(GxB_LOR_NE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_NE_UINT32(void)
{
  return &(GxB_LOR_NE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_NE_UINT64(void)
{
  return &(GxB_LOR_NE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_NE_UINT8(void)
{
  return &(GxB_LOR_NE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_PAIR_BOOL(void)
{
  return &(GxB_LOR_PAIR_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LOR_SECOND_BOOL(void)
{
  return &(GxB_LOR_SECOND_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LOR_UINT16(void)
{
  return &(GxB_LOR_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LOR_UINT32(void)
{
  return &(GxB_LOR_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LOR_UINT64(void)
{
  return &(GxB_LOR_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LOR_UINT8(void)
{
  return &(GxB_LOR_UINT8);
}

static struct GB_SelectOp_opaque * *_cffi_var_GxB_LT_THUNK(void)
{
  return &(GxB_LT_THUNK);
}

static struct GB_SelectOp_opaque * *_cffi_var_GxB_LT_ZERO(void)
{
  return &(GxB_LT_ZERO);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LXOR_BOOL(void)
{
  return &(GxB_LXOR_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_EQ_BOOL(void)
{
  return &(GxB_LXOR_EQ_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_EQ_FP32(void)
{
  return &(GxB_LXOR_EQ_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_EQ_FP64(void)
{
  return &(GxB_LXOR_EQ_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_EQ_INT16(void)
{
  return &(GxB_LXOR_EQ_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_EQ_INT32(void)
{
  return &(GxB_LXOR_EQ_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_EQ_INT64(void)
{
  return &(GxB_LXOR_EQ_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_EQ_INT8(void)
{
  return &(GxB_LXOR_EQ_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_EQ_UINT16(void)
{
  return &(GxB_LXOR_EQ_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_EQ_UINT32(void)
{
  return &(GxB_LXOR_EQ_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_EQ_UINT64(void)
{
  return &(GxB_LXOR_EQ_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_EQ_UINT8(void)
{
  return &(GxB_LXOR_EQ_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_FIRST_BOOL(void)
{
  return &(GxB_LXOR_FIRST_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LXOR_FP32(void)
{
  return &(GxB_LXOR_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LXOR_FP64(void)
{
  return &(GxB_LXOR_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_GE_BOOL(void)
{
  return &(GxB_LXOR_GE_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_GE_FP32(void)
{
  return &(GxB_LXOR_GE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_GE_FP64(void)
{
  return &(GxB_LXOR_GE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_GE_INT16(void)
{
  return &(GxB_LXOR_GE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_GE_INT32(void)
{
  return &(GxB_LXOR_GE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_GE_INT64(void)
{
  return &(GxB_LXOR_GE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_GE_INT8(void)
{
  return &(GxB_LXOR_GE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_GE_UINT16(void)
{
  return &(GxB_LXOR_GE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_GE_UINT32(void)
{
  return &(GxB_LXOR_GE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_GE_UINT64(void)
{
  return &(GxB_LXOR_GE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_GE_UINT8(void)
{
  return &(GxB_LXOR_GE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_GT_BOOL(void)
{
  return &(GxB_LXOR_GT_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_GT_FP32(void)
{
  return &(GxB_LXOR_GT_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_GT_FP64(void)
{
  return &(GxB_LXOR_GT_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_GT_INT16(void)
{
  return &(GxB_LXOR_GT_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_GT_INT32(void)
{
  return &(GxB_LXOR_GT_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_GT_INT64(void)
{
  return &(GxB_LXOR_GT_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_GT_INT8(void)
{
  return &(GxB_LXOR_GT_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_GT_UINT16(void)
{
  return &(GxB_LXOR_GT_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_GT_UINT32(void)
{
  return &(GxB_LXOR_GT_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_GT_UINT64(void)
{
  return &(GxB_LXOR_GT_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_GT_UINT8(void)
{
  return &(GxB_LXOR_GT_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LXOR_INT16(void)
{
  return &(GxB_LXOR_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LXOR_INT32(void)
{
  return &(GxB_LXOR_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LXOR_INT64(void)
{
  return &(GxB_LXOR_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LXOR_INT8(void)
{
  return &(GxB_LXOR_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LE_BOOL(void)
{
  return &(GxB_LXOR_LE_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LE_FP32(void)
{
  return &(GxB_LXOR_LE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LE_FP64(void)
{
  return &(GxB_LXOR_LE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LE_INT16(void)
{
  return &(GxB_LXOR_LE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LE_INT32(void)
{
  return &(GxB_LXOR_LE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LE_INT64(void)
{
  return &(GxB_LXOR_LE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LE_INT8(void)
{
  return &(GxB_LXOR_LE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LE_UINT16(void)
{
  return &(GxB_LXOR_LE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LE_UINT32(void)
{
  return &(GxB_LXOR_LE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LE_UINT64(void)
{
  return &(GxB_LXOR_LE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LE_UINT8(void)
{
  return &(GxB_LXOR_LE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LOR_BOOL(void)
{
  return &(GxB_LXOR_LOR_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LT_BOOL(void)
{
  return &(GxB_LXOR_LT_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LT_FP32(void)
{
  return &(GxB_LXOR_LT_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LT_FP64(void)
{
  return &(GxB_LXOR_LT_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LT_INT16(void)
{
  return &(GxB_LXOR_LT_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LT_INT32(void)
{
  return &(GxB_LXOR_LT_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LT_INT64(void)
{
  return &(GxB_LXOR_LT_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LT_INT8(void)
{
  return &(GxB_LXOR_LT_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LT_UINT16(void)
{
  return &(GxB_LXOR_LT_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LT_UINT32(void)
{
  return &(GxB_LXOR_LT_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LT_UINT64(void)
{
  return &(GxB_LXOR_LT_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LT_UINT8(void)
{
  return &(GxB_LXOR_LT_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_LXOR_BOOL(void)
{
  return &(GxB_LXOR_LXOR_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_NE_FP32(void)
{
  return &(GxB_LXOR_NE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_NE_FP64(void)
{
  return &(GxB_LXOR_NE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_NE_INT16(void)
{
  return &(GxB_LXOR_NE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_NE_INT32(void)
{
  return &(GxB_LXOR_NE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_NE_INT64(void)
{
  return &(GxB_LXOR_NE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_NE_INT8(void)
{
  return &(GxB_LXOR_NE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_NE_UINT16(void)
{
  return &(GxB_LXOR_NE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_NE_UINT32(void)
{
  return &(GxB_LXOR_NE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_NE_UINT64(void)
{
  return &(GxB_LXOR_NE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_NE_UINT8(void)
{
  return &(GxB_LXOR_NE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_PAIR_BOOL(void)
{
  return &(GxB_LXOR_PAIR_BOOL);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_LXOR_SECOND_BOOL(void)
{
  return &(GxB_LXOR_SECOND_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LXOR_UINT16(void)
{
  return &(GxB_LXOR_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LXOR_UINT32(void)
{
  return &(GxB_LXOR_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LXOR_UINT64(void)
{
  return &(GxB_LXOR_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_LXOR_UINT8(void)
{
  return &(GxB_LXOR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_DIV_FP32(void)
{
  return &(GxB_MAX_DIV_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_DIV_FP64(void)
{
  return &(GxB_MAX_DIV_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_DIV_INT16(void)
{
  return &(GxB_MAX_DIV_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_DIV_INT32(void)
{
  return &(GxB_MAX_DIV_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_DIV_INT64(void)
{
  return &(GxB_MAX_DIV_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_DIV_INT8(void)
{
  return &(GxB_MAX_DIV_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_DIV_UINT16(void)
{
  return &(GxB_MAX_DIV_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_DIV_UINT32(void)
{
  return &(GxB_MAX_DIV_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_DIV_UINT64(void)
{
  return &(GxB_MAX_DIV_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_DIV_UINT8(void)
{
  return &(GxB_MAX_DIV_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_FIRSTI1_INT32(void)
{
  return &(GxB_MAX_FIRSTI1_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_FIRSTI1_INT64(void)
{
  return &(GxB_MAX_FIRSTI1_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_FIRSTI_INT32(void)
{
  return &(GxB_MAX_FIRSTI_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_FIRSTI_INT64(void)
{
  return &(GxB_MAX_FIRSTI_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_FIRSTJ1_INT32(void)
{
  return &(GxB_MAX_FIRSTJ1_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_FIRSTJ1_INT64(void)
{
  return &(GxB_MAX_FIRSTJ1_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_FIRSTJ_INT32(void)
{
  return &(GxB_MAX_FIRSTJ_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_FIRSTJ_INT64(void)
{
  return &(GxB_MAX_FIRSTJ_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISEQ_FP32(void)
{
  return &(GxB_MAX_ISEQ_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISEQ_FP64(void)
{
  return &(GxB_MAX_ISEQ_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISEQ_INT16(void)
{
  return &(GxB_MAX_ISEQ_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISEQ_INT32(void)
{
  return &(GxB_MAX_ISEQ_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISEQ_INT64(void)
{
  return &(GxB_MAX_ISEQ_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISEQ_INT8(void)
{
  return &(GxB_MAX_ISEQ_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISEQ_UINT16(void)
{
  return &(GxB_MAX_ISEQ_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISEQ_UINT32(void)
{
  return &(GxB_MAX_ISEQ_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISEQ_UINT64(void)
{
  return &(GxB_MAX_ISEQ_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISEQ_UINT8(void)
{
  return &(GxB_MAX_ISEQ_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISGE_FP32(void)
{
  return &(GxB_MAX_ISGE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISGE_FP64(void)
{
  return &(GxB_MAX_ISGE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISGE_INT16(void)
{
  return &(GxB_MAX_ISGE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISGE_INT32(void)
{
  return &(GxB_MAX_ISGE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISGE_INT64(void)
{
  return &(GxB_MAX_ISGE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISGE_INT8(void)
{
  return &(GxB_MAX_ISGE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISGE_UINT16(void)
{
  return &(GxB_MAX_ISGE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISGE_UINT32(void)
{
  return &(GxB_MAX_ISGE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISGE_UINT64(void)
{
  return &(GxB_MAX_ISGE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISGE_UINT8(void)
{
  return &(GxB_MAX_ISGE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISGT_FP32(void)
{
  return &(GxB_MAX_ISGT_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISGT_FP64(void)
{
  return &(GxB_MAX_ISGT_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISGT_INT16(void)
{
  return &(GxB_MAX_ISGT_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISGT_INT32(void)
{
  return &(GxB_MAX_ISGT_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISGT_INT64(void)
{
  return &(GxB_MAX_ISGT_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISGT_INT8(void)
{
  return &(GxB_MAX_ISGT_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISGT_UINT16(void)
{
  return &(GxB_MAX_ISGT_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISGT_UINT32(void)
{
  return &(GxB_MAX_ISGT_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISGT_UINT64(void)
{
  return &(GxB_MAX_ISGT_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISGT_UINT8(void)
{
  return &(GxB_MAX_ISGT_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISLE_FP32(void)
{
  return &(GxB_MAX_ISLE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISLE_FP64(void)
{
  return &(GxB_MAX_ISLE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISLE_INT16(void)
{
  return &(GxB_MAX_ISLE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISLE_INT32(void)
{
  return &(GxB_MAX_ISLE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISLE_INT64(void)
{
  return &(GxB_MAX_ISLE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISLE_INT8(void)
{
  return &(GxB_MAX_ISLE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISLE_UINT16(void)
{
  return &(GxB_MAX_ISLE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISLE_UINT32(void)
{
  return &(GxB_MAX_ISLE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISLE_UINT64(void)
{
  return &(GxB_MAX_ISLE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISLE_UINT8(void)
{
  return &(GxB_MAX_ISLE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISLT_FP32(void)
{
  return &(GxB_MAX_ISLT_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISLT_FP64(void)
{
  return &(GxB_MAX_ISLT_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISLT_INT16(void)
{
  return &(GxB_MAX_ISLT_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISLT_INT32(void)
{
  return &(GxB_MAX_ISLT_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISLT_INT64(void)
{
  return &(GxB_MAX_ISLT_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISLT_INT8(void)
{
  return &(GxB_MAX_ISLT_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISLT_UINT16(void)
{
  return &(GxB_MAX_ISLT_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISLT_UINT32(void)
{
  return &(GxB_MAX_ISLT_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISLT_UINT64(void)
{
  return &(GxB_MAX_ISLT_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISLT_UINT8(void)
{
  return &(GxB_MAX_ISLT_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISNE_FP32(void)
{
  return &(GxB_MAX_ISNE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISNE_FP64(void)
{
  return &(GxB_MAX_ISNE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISNE_INT16(void)
{
  return &(GxB_MAX_ISNE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISNE_INT32(void)
{
  return &(GxB_MAX_ISNE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISNE_INT64(void)
{
  return &(GxB_MAX_ISNE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISNE_INT8(void)
{
  return &(GxB_MAX_ISNE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISNE_UINT16(void)
{
  return &(GxB_MAX_ISNE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISNE_UINT32(void)
{
  return &(GxB_MAX_ISNE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISNE_UINT64(void)
{
  return &(GxB_MAX_ISNE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_ISNE_UINT8(void)
{
  return &(GxB_MAX_ISNE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LAND_FP32(void)
{
  return &(GxB_MAX_LAND_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LAND_FP64(void)
{
  return &(GxB_MAX_LAND_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LAND_INT16(void)
{
  return &(GxB_MAX_LAND_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LAND_INT32(void)
{
  return &(GxB_MAX_LAND_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LAND_INT64(void)
{
  return &(GxB_MAX_LAND_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LAND_INT8(void)
{
  return &(GxB_MAX_LAND_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LAND_UINT16(void)
{
  return &(GxB_MAX_LAND_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LAND_UINT32(void)
{
  return &(GxB_MAX_LAND_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LAND_UINT64(void)
{
  return &(GxB_MAX_LAND_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LAND_UINT8(void)
{
  return &(GxB_MAX_LAND_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LOR_FP32(void)
{
  return &(GxB_MAX_LOR_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LOR_FP64(void)
{
  return &(GxB_MAX_LOR_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LOR_INT16(void)
{
  return &(GxB_MAX_LOR_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LOR_INT32(void)
{
  return &(GxB_MAX_LOR_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LOR_INT64(void)
{
  return &(GxB_MAX_LOR_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LOR_INT8(void)
{
  return &(GxB_MAX_LOR_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LOR_UINT16(void)
{
  return &(GxB_MAX_LOR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LOR_UINT32(void)
{
  return &(GxB_MAX_LOR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LOR_UINT64(void)
{
  return &(GxB_MAX_LOR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LOR_UINT8(void)
{
  return &(GxB_MAX_LOR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LXOR_FP32(void)
{
  return &(GxB_MAX_LXOR_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LXOR_FP64(void)
{
  return &(GxB_MAX_LXOR_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LXOR_INT16(void)
{
  return &(GxB_MAX_LXOR_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LXOR_INT32(void)
{
  return &(GxB_MAX_LXOR_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LXOR_INT64(void)
{
  return &(GxB_MAX_LXOR_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LXOR_INT8(void)
{
  return &(GxB_MAX_LXOR_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LXOR_UINT16(void)
{
  return &(GxB_MAX_LXOR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LXOR_UINT32(void)
{
  return &(GxB_MAX_LXOR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LXOR_UINT64(void)
{
  return &(GxB_MAX_LXOR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_LXOR_UINT8(void)
{
  return &(GxB_MAX_LXOR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_MAX_FP32(void)
{
  return &(GxB_MAX_MAX_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_MAX_FP64(void)
{
  return &(GxB_MAX_MAX_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_MAX_INT16(void)
{
  return &(GxB_MAX_MAX_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_MAX_INT32(void)
{
  return &(GxB_MAX_MAX_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_MAX_INT64(void)
{
  return &(GxB_MAX_MAX_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_MAX_INT8(void)
{
  return &(GxB_MAX_MAX_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_MAX_UINT16(void)
{
  return &(GxB_MAX_MAX_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_MAX_UINT32(void)
{
  return &(GxB_MAX_MAX_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_MAX_UINT64(void)
{
  return &(GxB_MAX_MAX_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_MAX_UINT8(void)
{
  return &(GxB_MAX_MAX_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_MINUS_FP32(void)
{
  return &(GxB_MAX_MINUS_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_MINUS_FP64(void)
{
  return &(GxB_MAX_MINUS_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_MINUS_INT16(void)
{
  return &(GxB_MAX_MINUS_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_MINUS_INT32(void)
{
  return &(GxB_MAX_MINUS_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_MINUS_INT64(void)
{
  return &(GxB_MAX_MINUS_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_MINUS_INT8(void)
{
  return &(GxB_MAX_MINUS_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_MINUS_UINT16(void)
{
  return &(GxB_MAX_MINUS_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_MINUS_UINT32(void)
{
  return &(GxB_MAX_MINUS_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_MINUS_UINT64(void)
{
  return &(GxB_MAX_MINUS_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_MINUS_UINT8(void)
{
  return &(GxB_MAX_MINUS_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_PAIR_FP32(void)
{
  return &(GxB_MAX_PAIR_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_PAIR_FP64(void)
{
  return &(GxB_MAX_PAIR_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_PAIR_INT16(void)
{
  return &(GxB_MAX_PAIR_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_PAIR_INT32(void)
{
  return &(GxB_MAX_PAIR_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_PAIR_INT64(void)
{
  return &(GxB_MAX_PAIR_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_PAIR_INT8(void)
{
  return &(GxB_MAX_PAIR_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_PAIR_UINT16(void)
{
  return &(GxB_MAX_PAIR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_PAIR_UINT32(void)
{
  return &(GxB_MAX_PAIR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_PAIR_UINT64(void)
{
  return &(GxB_MAX_PAIR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_PAIR_UINT8(void)
{
  return &(GxB_MAX_PAIR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_RDIV_FP32(void)
{
  return &(GxB_MAX_RDIV_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_RDIV_FP64(void)
{
  return &(GxB_MAX_RDIV_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_RDIV_INT16(void)
{
  return &(GxB_MAX_RDIV_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_RDIV_INT32(void)
{
  return &(GxB_MAX_RDIV_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_RDIV_INT64(void)
{
  return &(GxB_MAX_RDIV_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_RDIV_INT8(void)
{
  return &(GxB_MAX_RDIV_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_RDIV_UINT16(void)
{
  return &(GxB_MAX_RDIV_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_RDIV_UINT32(void)
{
  return &(GxB_MAX_RDIV_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_RDIV_UINT64(void)
{
  return &(GxB_MAX_RDIV_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_RDIV_UINT8(void)
{
  return &(GxB_MAX_RDIV_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_RMINUS_FP32(void)
{
  return &(GxB_MAX_RMINUS_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_RMINUS_FP64(void)
{
  return &(GxB_MAX_RMINUS_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_RMINUS_INT16(void)
{
  return &(GxB_MAX_RMINUS_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_RMINUS_INT32(void)
{
  return &(GxB_MAX_RMINUS_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_RMINUS_INT64(void)
{
  return &(GxB_MAX_RMINUS_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_RMINUS_INT8(void)
{
  return &(GxB_MAX_RMINUS_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_RMINUS_UINT16(void)
{
  return &(GxB_MAX_RMINUS_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_RMINUS_UINT32(void)
{
  return &(GxB_MAX_RMINUS_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_RMINUS_UINT64(void)
{
  return &(GxB_MAX_RMINUS_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_RMINUS_UINT8(void)
{
  return &(GxB_MAX_RMINUS_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_SECONDI1_INT32(void)
{
  return &(GxB_MAX_SECONDI1_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_SECONDI1_INT64(void)
{
  return &(GxB_MAX_SECONDI1_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_SECONDI_INT32(void)
{
  return &(GxB_MAX_SECONDI_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_SECONDI_INT64(void)
{
  return &(GxB_MAX_SECONDI_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_SECONDJ1_INT32(void)
{
  return &(GxB_MAX_SECONDJ1_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_SECONDJ1_INT64(void)
{
  return &(GxB_MAX_SECONDJ1_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_SECONDJ_INT32(void)
{
  return &(GxB_MAX_SECONDJ_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MAX_SECONDJ_INT64(void)
{
  return &(GxB_MAX_SECONDJ_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_MINUS_FC32(void)
{
  return &(GxB_MINUS_FC32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_MINUS_FC64(void)
{
  return &(GxB_MINUS_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_MINV_FC32(void)
{
  return &(GxB_MINV_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_MINV_FC64(void)
{
  return &(GxB_MINV_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_DIV_FP32(void)
{
  return &(GxB_MIN_DIV_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_DIV_FP64(void)
{
  return &(GxB_MIN_DIV_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_DIV_INT16(void)
{
  return &(GxB_MIN_DIV_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_DIV_INT32(void)
{
  return &(GxB_MIN_DIV_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_DIV_INT64(void)
{
  return &(GxB_MIN_DIV_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_DIV_INT8(void)
{
  return &(GxB_MIN_DIV_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_DIV_UINT16(void)
{
  return &(GxB_MIN_DIV_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_DIV_UINT32(void)
{
  return &(GxB_MIN_DIV_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_DIV_UINT64(void)
{
  return &(GxB_MIN_DIV_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_DIV_UINT8(void)
{
  return &(GxB_MIN_DIV_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_FIRSTI1_INT32(void)
{
  return &(GxB_MIN_FIRSTI1_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_FIRSTI1_INT64(void)
{
  return &(GxB_MIN_FIRSTI1_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_FIRSTI_INT32(void)
{
  return &(GxB_MIN_FIRSTI_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_FIRSTI_INT64(void)
{
  return &(GxB_MIN_FIRSTI_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_FIRSTJ1_INT32(void)
{
  return &(GxB_MIN_FIRSTJ1_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_FIRSTJ1_INT64(void)
{
  return &(GxB_MIN_FIRSTJ1_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_FIRSTJ_INT32(void)
{
  return &(GxB_MIN_FIRSTJ_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_FIRSTJ_INT64(void)
{
  return &(GxB_MIN_FIRSTJ_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISEQ_FP32(void)
{
  return &(GxB_MIN_ISEQ_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISEQ_FP64(void)
{
  return &(GxB_MIN_ISEQ_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISEQ_INT16(void)
{
  return &(GxB_MIN_ISEQ_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISEQ_INT32(void)
{
  return &(GxB_MIN_ISEQ_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISEQ_INT64(void)
{
  return &(GxB_MIN_ISEQ_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISEQ_INT8(void)
{
  return &(GxB_MIN_ISEQ_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISEQ_UINT16(void)
{
  return &(GxB_MIN_ISEQ_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISEQ_UINT32(void)
{
  return &(GxB_MIN_ISEQ_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISEQ_UINT64(void)
{
  return &(GxB_MIN_ISEQ_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISEQ_UINT8(void)
{
  return &(GxB_MIN_ISEQ_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISGE_FP32(void)
{
  return &(GxB_MIN_ISGE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISGE_FP64(void)
{
  return &(GxB_MIN_ISGE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISGE_INT16(void)
{
  return &(GxB_MIN_ISGE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISGE_INT32(void)
{
  return &(GxB_MIN_ISGE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISGE_INT64(void)
{
  return &(GxB_MIN_ISGE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISGE_INT8(void)
{
  return &(GxB_MIN_ISGE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISGE_UINT16(void)
{
  return &(GxB_MIN_ISGE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISGE_UINT32(void)
{
  return &(GxB_MIN_ISGE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISGE_UINT64(void)
{
  return &(GxB_MIN_ISGE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISGE_UINT8(void)
{
  return &(GxB_MIN_ISGE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISGT_FP32(void)
{
  return &(GxB_MIN_ISGT_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISGT_FP64(void)
{
  return &(GxB_MIN_ISGT_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISGT_INT16(void)
{
  return &(GxB_MIN_ISGT_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISGT_INT32(void)
{
  return &(GxB_MIN_ISGT_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISGT_INT64(void)
{
  return &(GxB_MIN_ISGT_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISGT_INT8(void)
{
  return &(GxB_MIN_ISGT_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISGT_UINT16(void)
{
  return &(GxB_MIN_ISGT_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISGT_UINT32(void)
{
  return &(GxB_MIN_ISGT_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISGT_UINT64(void)
{
  return &(GxB_MIN_ISGT_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISGT_UINT8(void)
{
  return &(GxB_MIN_ISGT_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISLE_FP32(void)
{
  return &(GxB_MIN_ISLE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISLE_FP64(void)
{
  return &(GxB_MIN_ISLE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISLE_INT16(void)
{
  return &(GxB_MIN_ISLE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISLE_INT32(void)
{
  return &(GxB_MIN_ISLE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISLE_INT64(void)
{
  return &(GxB_MIN_ISLE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISLE_INT8(void)
{
  return &(GxB_MIN_ISLE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISLE_UINT16(void)
{
  return &(GxB_MIN_ISLE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISLE_UINT32(void)
{
  return &(GxB_MIN_ISLE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISLE_UINT64(void)
{
  return &(GxB_MIN_ISLE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISLE_UINT8(void)
{
  return &(GxB_MIN_ISLE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISLT_FP32(void)
{
  return &(GxB_MIN_ISLT_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISLT_FP64(void)
{
  return &(GxB_MIN_ISLT_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISLT_INT16(void)
{
  return &(GxB_MIN_ISLT_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISLT_INT32(void)
{
  return &(GxB_MIN_ISLT_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISLT_INT64(void)
{
  return &(GxB_MIN_ISLT_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISLT_INT8(void)
{
  return &(GxB_MIN_ISLT_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISLT_UINT16(void)
{
  return &(GxB_MIN_ISLT_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISLT_UINT32(void)
{
  return &(GxB_MIN_ISLT_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISLT_UINT64(void)
{
  return &(GxB_MIN_ISLT_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISLT_UINT8(void)
{
  return &(GxB_MIN_ISLT_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISNE_FP32(void)
{
  return &(GxB_MIN_ISNE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISNE_FP64(void)
{
  return &(GxB_MIN_ISNE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISNE_INT16(void)
{
  return &(GxB_MIN_ISNE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISNE_INT32(void)
{
  return &(GxB_MIN_ISNE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISNE_INT64(void)
{
  return &(GxB_MIN_ISNE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISNE_INT8(void)
{
  return &(GxB_MIN_ISNE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISNE_UINT16(void)
{
  return &(GxB_MIN_ISNE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISNE_UINT32(void)
{
  return &(GxB_MIN_ISNE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISNE_UINT64(void)
{
  return &(GxB_MIN_ISNE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_ISNE_UINT8(void)
{
  return &(GxB_MIN_ISNE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LAND_FP32(void)
{
  return &(GxB_MIN_LAND_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LAND_FP64(void)
{
  return &(GxB_MIN_LAND_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LAND_INT16(void)
{
  return &(GxB_MIN_LAND_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LAND_INT32(void)
{
  return &(GxB_MIN_LAND_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LAND_INT64(void)
{
  return &(GxB_MIN_LAND_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LAND_INT8(void)
{
  return &(GxB_MIN_LAND_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LAND_UINT16(void)
{
  return &(GxB_MIN_LAND_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LAND_UINT32(void)
{
  return &(GxB_MIN_LAND_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LAND_UINT64(void)
{
  return &(GxB_MIN_LAND_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LAND_UINT8(void)
{
  return &(GxB_MIN_LAND_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LOR_FP32(void)
{
  return &(GxB_MIN_LOR_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LOR_FP64(void)
{
  return &(GxB_MIN_LOR_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LOR_INT16(void)
{
  return &(GxB_MIN_LOR_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LOR_INT32(void)
{
  return &(GxB_MIN_LOR_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LOR_INT64(void)
{
  return &(GxB_MIN_LOR_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LOR_INT8(void)
{
  return &(GxB_MIN_LOR_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LOR_UINT16(void)
{
  return &(GxB_MIN_LOR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LOR_UINT32(void)
{
  return &(GxB_MIN_LOR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LOR_UINT64(void)
{
  return &(GxB_MIN_LOR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LOR_UINT8(void)
{
  return &(GxB_MIN_LOR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LXOR_FP32(void)
{
  return &(GxB_MIN_LXOR_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LXOR_FP64(void)
{
  return &(GxB_MIN_LXOR_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LXOR_INT16(void)
{
  return &(GxB_MIN_LXOR_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LXOR_INT32(void)
{
  return &(GxB_MIN_LXOR_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LXOR_INT64(void)
{
  return &(GxB_MIN_LXOR_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LXOR_INT8(void)
{
  return &(GxB_MIN_LXOR_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LXOR_UINT16(void)
{
  return &(GxB_MIN_LXOR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LXOR_UINT32(void)
{
  return &(GxB_MIN_LXOR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LXOR_UINT64(void)
{
  return &(GxB_MIN_LXOR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_LXOR_UINT8(void)
{
  return &(GxB_MIN_LXOR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_MINUS_FP32(void)
{
  return &(GxB_MIN_MINUS_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_MINUS_FP64(void)
{
  return &(GxB_MIN_MINUS_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_MINUS_INT16(void)
{
  return &(GxB_MIN_MINUS_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_MINUS_INT32(void)
{
  return &(GxB_MIN_MINUS_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_MINUS_INT64(void)
{
  return &(GxB_MIN_MINUS_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_MINUS_INT8(void)
{
  return &(GxB_MIN_MINUS_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_MINUS_UINT16(void)
{
  return &(GxB_MIN_MINUS_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_MINUS_UINT32(void)
{
  return &(GxB_MIN_MINUS_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_MINUS_UINT64(void)
{
  return &(GxB_MIN_MINUS_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_MINUS_UINT8(void)
{
  return &(GxB_MIN_MINUS_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_MIN_FP32(void)
{
  return &(GxB_MIN_MIN_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_MIN_FP64(void)
{
  return &(GxB_MIN_MIN_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_MIN_INT16(void)
{
  return &(GxB_MIN_MIN_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_MIN_INT32(void)
{
  return &(GxB_MIN_MIN_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_MIN_INT64(void)
{
  return &(GxB_MIN_MIN_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_MIN_INT8(void)
{
  return &(GxB_MIN_MIN_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_MIN_UINT16(void)
{
  return &(GxB_MIN_MIN_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_MIN_UINT32(void)
{
  return &(GxB_MIN_MIN_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_MIN_UINT64(void)
{
  return &(GxB_MIN_MIN_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_MIN_UINT8(void)
{
  return &(GxB_MIN_MIN_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_PAIR_FP32(void)
{
  return &(GxB_MIN_PAIR_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_PAIR_FP64(void)
{
  return &(GxB_MIN_PAIR_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_PAIR_INT16(void)
{
  return &(GxB_MIN_PAIR_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_PAIR_INT32(void)
{
  return &(GxB_MIN_PAIR_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_PAIR_INT64(void)
{
  return &(GxB_MIN_PAIR_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_PAIR_INT8(void)
{
  return &(GxB_MIN_PAIR_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_PAIR_UINT16(void)
{
  return &(GxB_MIN_PAIR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_PAIR_UINT32(void)
{
  return &(GxB_MIN_PAIR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_PAIR_UINT64(void)
{
  return &(GxB_MIN_PAIR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_PAIR_UINT8(void)
{
  return &(GxB_MIN_PAIR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_RDIV_FP32(void)
{
  return &(GxB_MIN_RDIV_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_RDIV_FP64(void)
{
  return &(GxB_MIN_RDIV_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_RDIV_INT16(void)
{
  return &(GxB_MIN_RDIV_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_RDIV_INT32(void)
{
  return &(GxB_MIN_RDIV_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_RDIV_INT64(void)
{
  return &(GxB_MIN_RDIV_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_RDIV_INT8(void)
{
  return &(GxB_MIN_RDIV_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_RDIV_UINT16(void)
{
  return &(GxB_MIN_RDIV_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_RDIV_UINT32(void)
{
  return &(GxB_MIN_RDIV_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_RDIV_UINT64(void)
{
  return &(GxB_MIN_RDIV_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_RDIV_UINT8(void)
{
  return &(GxB_MIN_RDIV_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_RMINUS_FP32(void)
{
  return &(GxB_MIN_RMINUS_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_RMINUS_FP64(void)
{
  return &(GxB_MIN_RMINUS_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_RMINUS_INT16(void)
{
  return &(GxB_MIN_RMINUS_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_RMINUS_INT32(void)
{
  return &(GxB_MIN_RMINUS_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_RMINUS_INT64(void)
{
  return &(GxB_MIN_RMINUS_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_RMINUS_INT8(void)
{
  return &(GxB_MIN_RMINUS_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_RMINUS_UINT16(void)
{
  return &(GxB_MIN_RMINUS_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_RMINUS_UINT32(void)
{
  return &(GxB_MIN_RMINUS_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_RMINUS_UINT64(void)
{
  return &(GxB_MIN_RMINUS_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_RMINUS_UINT8(void)
{
  return &(GxB_MIN_RMINUS_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_SECONDI1_INT32(void)
{
  return &(GxB_MIN_SECONDI1_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_SECONDI1_INT64(void)
{
  return &(GxB_MIN_SECONDI1_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_SECONDI_INT32(void)
{
  return &(GxB_MIN_SECONDI_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_SECONDI_INT64(void)
{
  return &(GxB_MIN_SECONDI_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_SECONDJ1_INT32(void)
{
  return &(GxB_MIN_SECONDJ1_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_SECONDJ1_INT64(void)
{
  return &(GxB_MIN_SECONDJ1_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_SECONDJ_INT32(void)
{
  return &(GxB_MIN_SECONDJ_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_MIN_SECONDJ_INT64(void)
{
  return &(GxB_MIN_SECONDJ_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_NE_FC32(void)
{
  return &(GxB_NE_FC32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_NE_FC64(void)
{
  return &(GxB_NE_FC64);
}

static struct GB_SelectOp_opaque * *_cffi_var_GxB_NE_THUNK(void)
{
  return &(GxB_NE_THUNK);
}

static struct GB_SelectOp_opaque * *_cffi_var_GxB_NONZERO(void)
{
  return &(GxB_NONZERO);
}

static struct GB_SelectOp_opaque * *_cffi_var_GxB_OFFDIAG(void)
{
  return &(GxB_OFFDIAG);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ONEB_FC32(void)
{
  return &(GxB_ONEB_FC32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_ONEB_FC64(void)
{
  return &(GxB_ONEB_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ONE_BOOL(void)
{
  return &(GxB_ONE_BOOL);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ONE_FC32(void)
{
  return &(GxB_ONE_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ONE_FC64(void)
{
  return &(GxB_ONE_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ONE_FP32(void)
{
  return &(GxB_ONE_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ONE_FP64(void)
{
  return &(GxB_ONE_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ONE_INT16(void)
{
  return &(GxB_ONE_INT16);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ONE_INT32(void)
{
  return &(GxB_ONE_INT32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ONE_INT64(void)
{
  return &(GxB_ONE_INT64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ONE_INT8(void)
{
  return &(GxB_ONE_INT8);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ONE_UINT16(void)
{
  return &(GxB_ONE_UINT16);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ONE_UINT32(void)
{
  return &(GxB_ONE_UINT32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ONE_UINT64(void)
{
  return &(GxB_ONE_UINT64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ONE_UINT8(void)
{
  return &(GxB_ONE_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_PAIR_BOOL(void)
{
  return &(GxB_PAIR_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_PAIR_FC32(void)
{
  return &(GxB_PAIR_FC32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_PAIR_FC64(void)
{
  return &(GxB_PAIR_FC64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_PAIR_FP32(void)
{
  return &(GxB_PAIR_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_PAIR_FP64(void)
{
  return &(GxB_PAIR_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_PAIR_INT16(void)
{
  return &(GxB_PAIR_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_PAIR_INT32(void)
{
  return &(GxB_PAIR_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_PAIR_INT64(void)
{
  return &(GxB_PAIR_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_PAIR_INT8(void)
{
  return &(GxB_PAIR_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_PAIR_UINT16(void)
{
  return &(GxB_PAIR_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_PAIR_UINT32(void)
{
  return &(GxB_PAIR_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_PAIR_UINT64(void)
{
  return &(GxB_PAIR_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_PAIR_UINT8(void)
{
  return &(GxB_PAIR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_DIV_FC32(void)
{
  return &(GxB_PLUS_DIV_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_DIV_FC64(void)
{
  return &(GxB_PLUS_DIV_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_DIV_FP32(void)
{
  return &(GxB_PLUS_DIV_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_DIV_FP64(void)
{
  return &(GxB_PLUS_DIV_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_DIV_INT16(void)
{
  return &(GxB_PLUS_DIV_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_DIV_INT32(void)
{
  return &(GxB_PLUS_DIV_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_DIV_INT64(void)
{
  return &(GxB_PLUS_DIV_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_DIV_INT8(void)
{
  return &(GxB_PLUS_DIV_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_DIV_UINT16(void)
{
  return &(GxB_PLUS_DIV_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_DIV_UINT32(void)
{
  return &(GxB_PLUS_DIV_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_DIV_UINT64(void)
{
  return &(GxB_PLUS_DIV_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_DIV_UINT8(void)
{
  return &(GxB_PLUS_DIV_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_PLUS_FC32(void)
{
  return &(GxB_PLUS_FC32);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_PLUS_FC32_MONOID(void)
{
  return &(GxB_PLUS_FC32_MONOID);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_PLUS_FC64(void)
{
  return &(GxB_PLUS_FC64);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_PLUS_FC64_MONOID(void)
{
  return &(GxB_PLUS_FC64_MONOID);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_FIRSTI1_INT32(void)
{
  return &(GxB_PLUS_FIRSTI1_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_FIRSTI1_INT64(void)
{
  return &(GxB_PLUS_FIRSTI1_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_FIRSTI_INT32(void)
{
  return &(GxB_PLUS_FIRSTI_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_FIRSTI_INT64(void)
{
  return &(GxB_PLUS_FIRSTI_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_FIRSTJ1_INT32(void)
{
  return &(GxB_PLUS_FIRSTJ1_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_FIRSTJ1_INT64(void)
{
  return &(GxB_PLUS_FIRSTJ1_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_FIRSTJ_INT32(void)
{
  return &(GxB_PLUS_FIRSTJ_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_FIRSTJ_INT64(void)
{
  return &(GxB_PLUS_FIRSTJ_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_FIRST_FC32(void)
{
  return &(GxB_PLUS_FIRST_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_FIRST_FC64(void)
{
  return &(GxB_PLUS_FIRST_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_FIRST_FP32(void)
{
  return &(GxB_PLUS_FIRST_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_FIRST_FP64(void)
{
  return &(GxB_PLUS_FIRST_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_FIRST_INT16(void)
{
  return &(GxB_PLUS_FIRST_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_FIRST_INT32(void)
{
  return &(GxB_PLUS_FIRST_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_FIRST_INT64(void)
{
  return &(GxB_PLUS_FIRST_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_FIRST_INT8(void)
{
  return &(GxB_PLUS_FIRST_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_FIRST_UINT16(void)
{
  return &(GxB_PLUS_FIRST_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_FIRST_UINT32(void)
{
  return &(GxB_PLUS_FIRST_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_FIRST_UINT64(void)
{
  return &(GxB_PLUS_FIRST_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_FIRST_UINT8(void)
{
  return &(GxB_PLUS_FIRST_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISEQ_FP32(void)
{
  return &(GxB_PLUS_ISEQ_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISEQ_FP64(void)
{
  return &(GxB_PLUS_ISEQ_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISEQ_INT16(void)
{
  return &(GxB_PLUS_ISEQ_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISEQ_INT32(void)
{
  return &(GxB_PLUS_ISEQ_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISEQ_INT64(void)
{
  return &(GxB_PLUS_ISEQ_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISEQ_INT8(void)
{
  return &(GxB_PLUS_ISEQ_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISEQ_UINT16(void)
{
  return &(GxB_PLUS_ISEQ_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISEQ_UINT32(void)
{
  return &(GxB_PLUS_ISEQ_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISEQ_UINT64(void)
{
  return &(GxB_PLUS_ISEQ_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISEQ_UINT8(void)
{
  return &(GxB_PLUS_ISEQ_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISGE_FP32(void)
{
  return &(GxB_PLUS_ISGE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISGE_FP64(void)
{
  return &(GxB_PLUS_ISGE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISGE_INT16(void)
{
  return &(GxB_PLUS_ISGE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISGE_INT32(void)
{
  return &(GxB_PLUS_ISGE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISGE_INT64(void)
{
  return &(GxB_PLUS_ISGE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISGE_INT8(void)
{
  return &(GxB_PLUS_ISGE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISGE_UINT16(void)
{
  return &(GxB_PLUS_ISGE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISGE_UINT32(void)
{
  return &(GxB_PLUS_ISGE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISGE_UINT64(void)
{
  return &(GxB_PLUS_ISGE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISGE_UINT8(void)
{
  return &(GxB_PLUS_ISGE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISGT_FP32(void)
{
  return &(GxB_PLUS_ISGT_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISGT_FP64(void)
{
  return &(GxB_PLUS_ISGT_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISGT_INT16(void)
{
  return &(GxB_PLUS_ISGT_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISGT_INT32(void)
{
  return &(GxB_PLUS_ISGT_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISGT_INT64(void)
{
  return &(GxB_PLUS_ISGT_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISGT_INT8(void)
{
  return &(GxB_PLUS_ISGT_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISGT_UINT16(void)
{
  return &(GxB_PLUS_ISGT_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISGT_UINT32(void)
{
  return &(GxB_PLUS_ISGT_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISGT_UINT64(void)
{
  return &(GxB_PLUS_ISGT_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISGT_UINT8(void)
{
  return &(GxB_PLUS_ISGT_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISLE_FP32(void)
{
  return &(GxB_PLUS_ISLE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISLE_FP64(void)
{
  return &(GxB_PLUS_ISLE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISLE_INT16(void)
{
  return &(GxB_PLUS_ISLE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISLE_INT32(void)
{
  return &(GxB_PLUS_ISLE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISLE_INT64(void)
{
  return &(GxB_PLUS_ISLE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISLE_INT8(void)
{
  return &(GxB_PLUS_ISLE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISLE_UINT16(void)
{
  return &(GxB_PLUS_ISLE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISLE_UINT32(void)
{
  return &(GxB_PLUS_ISLE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISLE_UINT64(void)
{
  return &(GxB_PLUS_ISLE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISLE_UINT8(void)
{
  return &(GxB_PLUS_ISLE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISLT_FP32(void)
{
  return &(GxB_PLUS_ISLT_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISLT_FP64(void)
{
  return &(GxB_PLUS_ISLT_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISLT_INT16(void)
{
  return &(GxB_PLUS_ISLT_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISLT_INT32(void)
{
  return &(GxB_PLUS_ISLT_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISLT_INT64(void)
{
  return &(GxB_PLUS_ISLT_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISLT_INT8(void)
{
  return &(GxB_PLUS_ISLT_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISLT_UINT16(void)
{
  return &(GxB_PLUS_ISLT_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISLT_UINT32(void)
{
  return &(GxB_PLUS_ISLT_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISLT_UINT64(void)
{
  return &(GxB_PLUS_ISLT_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISLT_UINT8(void)
{
  return &(GxB_PLUS_ISLT_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISNE_FP32(void)
{
  return &(GxB_PLUS_ISNE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISNE_FP64(void)
{
  return &(GxB_PLUS_ISNE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISNE_INT16(void)
{
  return &(GxB_PLUS_ISNE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISNE_INT32(void)
{
  return &(GxB_PLUS_ISNE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISNE_INT64(void)
{
  return &(GxB_PLUS_ISNE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISNE_INT8(void)
{
  return &(GxB_PLUS_ISNE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISNE_UINT16(void)
{
  return &(GxB_PLUS_ISNE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISNE_UINT32(void)
{
  return &(GxB_PLUS_ISNE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISNE_UINT64(void)
{
  return &(GxB_PLUS_ISNE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_ISNE_UINT8(void)
{
  return &(GxB_PLUS_ISNE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LAND_FP32(void)
{
  return &(GxB_PLUS_LAND_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LAND_FP64(void)
{
  return &(GxB_PLUS_LAND_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LAND_INT16(void)
{
  return &(GxB_PLUS_LAND_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LAND_INT32(void)
{
  return &(GxB_PLUS_LAND_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LAND_INT64(void)
{
  return &(GxB_PLUS_LAND_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LAND_INT8(void)
{
  return &(GxB_PLUS_LAND_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LAND_UINT16(void)
{
  return &(GxB_PLUS_LAND_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LAND_UINT32(void)
{
  return &(GxB_PLUS_LAND_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LAND_UINT64(void)
{
  return &(GxB_PLUS_LAND_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LAND_UINT8(void)
{
  return &(GxB_PLUS_LAND_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LOR_FP32(void)
{
  return &(GxB_PLUS_LOR_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LOR_FP64(void)
{
  return &(GxB_PLUS_LOR_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LOR_INT16(void)
{
  return &(GxB_PLUS_LOR_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LOR_INT32(void)
{
  return &(GxB_PLUS_LOR_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LOR_INT64(void)
{
  return &(GxB_PLUS_LOR_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LOR_INT8(void)
{
  return &(GxB_PLUS_LOR_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LOR_UINT16(void)
{
  return &(GxB_PLUS_LOR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LOR_UINT32(void)
{
  return &(GxB_PLUS_LOR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LOR_UINT64(void)
{
  return &(GxB_PLUS_LOR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LOR_UINT8(void)
{
  return &(GxB_PLUS_LOR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LXOR_FP32(void)
{
  return &(GxB_PLUS_LXOR_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LXOR_FP64(void)
{
  return &(GxB_PLUS_LXOR_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LXOR_INT16(void)
{
  return &(GxB_PLUS_LXOR_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LXOR_INT32(void)
{
  return &(GxB_PLUS_LXOR_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LXOR_INT64(void)
{
  return &(GxB_PLUS_LXOR_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LXOR_INT8(void)
{
  return &(GxB_PLUS_LXOR_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LXOR_UINT16(void)
{
  return &(GxB_PLUS_LXOR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LXOR_UINT32(void)
{
  return &(GxB_PLUS_LXOR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LXOR_UINT64(void)
{
  return &(GxB_PLUS_LXOR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_LXOR_UINT8(void)
{
  return &(GxB_PLUS_LXOR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_MAX_FP32(void)
{
  return &(GxB_PLUS_MAX_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_MAX_FP64(void)
{
  return &(GxB_PLUS_MAX_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_MAX_INT16(void)
{
  return &(GxB_PLUS_MAX_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_MAX_INT32(void)
{
  return &(GxB_PLUS_MAX_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_MAX_INT64(void)
{
  return &(GxB_PLUS_MAX_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_MAX_INT8(void)
{
  return &(GxB_PLUS_MAX_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_MAX_UINT16(void)
{
  return &(GxB_PLUS_MAX_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_MAX_UINT32(void)
{
  return &(GxB_PLUS_MAX_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_MAX_UINT64(void)
{
  return &(GxB_PLUS_MAX_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_MAX_UINT8(void)
{
  return &(GxB_PLUS_MAX_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_MINUS_FC32(void)
{
  return &(GxB_PLUS_MINUS_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_MINUS_FC64(void)
{
  return &(GxB_PLUS_MINUS_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_MINUS_FP32(void)
{
  return &(GxB_PLUS_MINUS_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_MINUS_FP64(void)
{
  return &(GxB_PLUS_MINUS_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_MINUS_INT16(void)
{
  return &(GxB_PLUS_MINUS_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_MINUS_INT32(void)
{
  return &(GxB_PLUS_MINUS_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_MINUS_INT64(void)
{
  return &(GxB_PLUS_MINUS_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_MINUS_INT8(void)
{
  return &(GxB_PLUS_MINUS_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_MINUS_UINT16(void)
{
  return &(GxB_PLUS_MINUS_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_MINUS_UINT32(void)
{
  return &(GxB_PLUS_MINUS_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_MINUS_UINT64(void)
{
  return &(GxB_PLUS_MINUS_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_MINUS_UINT8(void)
{
  return &(GxB_PLUS_MINUS_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PAIR_FC32(void)
{
  return &(GxB_PLUS_PAIR_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PAIR_FC64(void)
{
  return &(GxB_PLUS_PAIR_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PAIR_FP32(void)
{
  return &(GxB_PLUS_PAIR_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PAIR_FP64(void)
{
  return &(GxB_PLUS_PAIR_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PAIR_INT16(void)
{
  return &(GxB_PLUS_PAIR_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PAIR_INT32(void)
{
  return &(GxB_PLUS_PAIR_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PAIR_INT64(void)
{
  return &(GxB_PLUS_PAIR_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PAIR_INT8(void)
{
  return &(GxB_PLUS_PAIR_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PAIR_UINT16(void)
{
  return &(GxB_PLUS_PAIR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PAIR_UINT32(void)
{
  return &(GxB_PLUS_PAIR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PAIR_UINT64(void)
{
  return &(GxB_PLUS_PAIR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PAIR_UINT8(void)
{
  return &(GxB_PLUS_PAIR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PLUS_FC32(void)
{
  return &(GxB_PLUS_PLUS_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PLUS_FC64(void)
{
  return &(GxB_PLUS_PLUS_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PLUS_FP32(void)
{
  return &(GxB_PLUS_PLUS_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PLUS_FP64(void)
{
  return &(GxB_PLUS_PLUS_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PLUS_INT16(void)
{
  return &(GxB_PLUS_PLUS_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PLUS_INT32(void)
{
  return &(GxB_PLUS_PLUS_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PLUS_INT64(void)
{
  return &(GxB_PLUS_PLUS_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PLUS_INT8(void)
{
  return &(GxB_PLUS_PLUS_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PLUS_UINT16(void)
{
  return &(GxB_PLUS_PLUS_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PLUS_UINT32(void)
{
  return &(GxB_PLUS_PLUS_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PLUS_UINT64(void)
{
  return &(GxB_PLUS_PLUS_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_PLUS_UINT8(void)
{
  return &(GxB_PLUS_PLUS_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RDIV_FC32(void)
{
  return &(GxB_PLUS_RDIV_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RDIV_FC64(void)
{
  return &(GxB_PLUS_RDIV_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RDIV_FP32(void)
{
  return &(GxB_PLUS_RDIV_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RDIV_FP64(void)
{
  return &(GxB_PLUS_RDIV_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RDIV_INT16(void)
{
  return &(GxB_PLUS_RDIV_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RDIV_INT32(void)
{
  return &(GxB_PLUS_RDIV_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RDIV_INT64(void)
{
  return &(GxB_PLUS_RDIV_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RDIV_INT8(void)
{
  return &(GxB_PLUS_RDIV_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RDIV_UINT16(void)
{
  return &(GxB_PLUS_RDIV_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RDIV_UINT32(void)
{
  return &(GxB_PLUS_RDIV_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RDIV_UINT64(void)
{
  return &(GxB_PLUS_RDIV_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RDIV_UINT8(void)
{
  return &(GxB_PLUS_RDIV_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RMINUS_FC32(void)
{
  return &(GxB_PLUS_RMINUS_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RMINUS_FC64(void)
{
  return &(GxB_PLUS_RMINUS_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RMINUS_FP32(void)
{
  return &(GxB_PLUS_RMINUS_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RMINUS_FP64(void)
{
  return &(GxB_PLUS_RMINUS_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RMINUS_INT16(void)
{
  return &(GxB_PLUS_RMINUS_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RMINUS_INT32(void)
{
  return &(GxB_PLUS_RMINUS_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RMINUS_INT64(void)
{
  return &(GxB_PLUS_RMINUS_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RMINUS_INT8(void)
{
  return &(GxB_PLUS_RMINUS_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RMINUS_UINT16(void)
{
  return &(GxB_PLUS_RMINUS_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RMINUS_UINT32(void)
{
  return &(GxB_PLUS_RMINUS_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RMINUS_UINT64(void)
{
  return &(GxB_PLUS_RMINUS_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_RMINUS_UINT8(void)
{
  return &(GxB_PLUS_RMINUS_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_SECONDI1_INT32(void)
{
  return &(GxB_PLUS_SECONDI1_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_SECONDI1_INT64(void)
{
  return &(GxB_PLUS_SECONDI1_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_SECONDI_INT32(void)
{
  return &(GxB_PLUS_SECONDI_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_SECONDI_INT64(void)
{
  return &(GxB_PLUS_SECONDI_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_SECONDJ1_INT32(void)
{
  return &(GxB_PLUS_SECONDJ1_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_SECONDJ1_INT64(void)
{
  return &(GxB_PLUS_SECONDJ1_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_SECONDJ_INT32(void)
{
  return &(GxB_PLUS_SECONDJ_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_SECONDJ_INT64(void)
{
  return &(GxB_PLUS_SECONDJ_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_SECOND_FC32(void)
{
  return &(GxB_PLUS_SECOND_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_SECOND_FC64(void)
{
  return &(GxB_PLUS_SECOND_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_SECOND_FP32(void)
{
  return &(GxB_PLUS_SECOND_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_SECOND_FP64(void)
{
  return &(GxB_PLUS_SECOND_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_SECOND_INT16(void)
{
  return &(GxB_PLUS_SECOND_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_SECOND_INT32(void)
{
  return &(GxB_PLUS_SECOND_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_SECOND_INT64(void)
{
  return &(GxB_PLUS_SECOND_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_SECOND_INT8(void)
{
  return &(GxB_PLUS_SECOND_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_SECOND_UINT16(void)
{
  return &(GxB_PLUS_SECOND_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_SECOND_UINT32(void)
{
  return &(GxB_PLUS_SECOND_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_SECOND_UINT64(void)
{
  return &(GxB_PLUS_SECOND_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_SECOND_UINT8(void)
{
  return &(GxB_PLUS_SECOND_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_TIMES_FC32(void)
{
  return &(GxB_PLUS_TIMES_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_PLUS_TIMES_FC64(void)
{
  return &(GxB_PLUS_TIMES_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_POSITIONI1_INT32(void)
{
  return &(GxB_POSITIONI1_INT32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_POSITIONI1_INT64(void)
{
  return &(GxB_POSITIONI1_INT64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_POSITIONI_INT32(void)
{
  return &(GxB_POSITIONI_INT32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_POSITIONI_INT64(void)
{
  return &(GxB_POSITIONI_INT64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_POSITIONJ1_INT32(void)
{
  return &(GxB_POSITIONJ1_INT32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_POSITIONJ1_INT64(void)
{
  return &(GxB_POSITIONJ1_INT64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_POSITIONJ_INT32(void)
{
  return &(GxB_POSITIONJ_INT32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_POSITIONJ_INT64(void)
{
  return &(GxB_POSITIONJ_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_POW_BOOL(void)
{
  return &(GxB_POW_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_POW_FC32(void)
{
  return &(GxB_POW_FC32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_POW_FC64(void)
{
  return &(GxB_POW_FC64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_POW_FP32(void)
{
  return &(GxB_POW_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_POW_FP64(void)
{
  return &(GxB_POW_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_POW_INT16(void)
{
  return &(GxB_POW_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_POW_INT32(void)
{
  return &(GxB_POW_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_POW_INT64(void)
{
  return &(GxB_POW_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_POW_INT8(void)
{
  return &(GxB_POW_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_POW_UINT16(void)
{
  return &(GxB_POW_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_POW_UINT32(void)
{
  return &(GxB_POW_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_POW_UINT64(void)
{
  return &(GxB_POW_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_POW_UINT8(void)
{
  return &(GxB_POW_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RDIV_BOOL(void)
{
  return &(GxB_RDIV_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RDIV_FC32(void)
{
  return &(GxB_RDIV_FC32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RDIV_FC64(void)
{
  return &(GxB_RDIV_FC64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RDIV_FP32(void)
{
  return &(GxB_RDIV_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RDIV_FP64(void)
{
  return &(GxB_RDIV_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RDIV_INT16(void)
{
  return &(GxB_RDIV_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RDIV_INT32(void)
{
  return &(GxB_RDIV_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RDIV_INT64(void)
{
  return &(GxB_RDIV_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RDIV_INT8(void)
{
  return &(GxB_RDIV_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RDIV_UINT16(void)
{
  return &(GxB_RDIV_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RDIV_UINT32(void)
{
  return &(GxB_RDIV_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RDIV_UINT64(void)
{
  return &(GxB_RDIV_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RDIV_UINT8(void)
{
  return &(GxB_RDIV_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_REMAINDER_FP32(void)
{
  return &(GxB_REMAINDER_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_REMAINDER_FP64(void)
{
  return &(GxB_REMAINDER_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RMINUS_BOOL(void)
{
  return &(GxB_RMINUS_BOOL);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RMINUS_FC32(void)
{
  return &(GxB_RMINUS_FC32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RMINUS_FC64(void)
{
  return &(GxB_RMINUS_FC64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RMINUS_FP32(void)
{
  return &(GxB_RMINUS_FP32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RMINUS_FP64(void)
{
  return &(GxB_RMINUS_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RMINUS_INT16(void)
{
  return &(GxB_RMINUS_INT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RMINUS_INT32(void)
{
  return &(GxB_RMINUS_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RMINUS_INT64(void)
{
  return &(GxB_RMINUS_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RMINUS_INT8(void)
{
  return &(GxB_RMINUS_INT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RMINUS_UINT16(void)
{
  return &(GxB_RMINUS_UINT16);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RMINUS_UINT32(void)
{
  return &(GxB_RMINUS_UINT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RMINUS_UINT64(void)
{
  return &(GxB_RMINUS_UINT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_RMINUS_UINT8(void)
{
  return &(GxB_RMINUS_UINT8);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ROUND_FC32(void)
{
  return &(GxB_ROUND_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ROUND_FC64(void)
{
  return &(GxB_ROUND_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ROUND_FP32(void)
{
  return &(GxB_ROUND_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_ROUND_FP64(void)
{
  return &(GxB_ROUND_FP64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_SECONDI1_INT32(void)
{
  return &(GxB_SECONDI1_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_SECONDI1_INT64(void)
{
  return &(GxB_SECONDI1_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_SECONDI_INT32(void)
{
  return &(GxB_SECONDI_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_SECONDI_INT64(void)
{
  return &(GxB_SECONDI_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_SECONDJ1_INT32(void)
{
  return &(GxB_SECONDJ1_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_SECONDJ1_INT64(void)
{
  return &(GxB_SECONDJ1_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_SECONDJ_INT32(void)
{
  return &(GxB_SECONDJ_INT32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_SECONDJ_INT64(void)
{
  return &(GxB_SECONDJ_INT64);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_SECOND_FC32(void)
{
  return &(GxB_SECOND_FC32);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_SECOND_FC64(void)
{
  return &(GxB_SECOND_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_SIGNUM_FC32(void)
{
  return &(GxB_SIGNUM_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_SIGNUM_FC64(void)
{
  return &(GxB_SIGNUM_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_SIGNUM_FP32(void)
{
  return &(GxB_SIGNUM_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_SIGNUM_FP64(void)
{
  return &(GxB_SIGNUM_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_SINH_FC32(void)
{
  return &(GxB_SINH_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_SINH_FC64(void)
{
  return &(GxB_SINH_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_SINH_FP32(void)
{
  return &(GxB_SINH_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_SINH_FP64(void)
{
  return &(GxB_SINH_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_SIN_FC32(void)
{
  return &(GxB_SIN_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_SIN_FC64(void)
{
  return &(GxB_SIN_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_SIN_FP32(void)
{
  return &(GxB_SIN_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_SIN_FP64(void)
{
  return &(GxB_SIN_FP64);
}

static char * *_cffi_var_GxB_SPEC_ABOUT_STR(void)
{
  return &(GxB_SPEC_ABOUT_STR);
}

static char * *_cffi_var_GxB_SPEC_DATE_STR(void)
{
  return &(GxB_SPEC_DATE_STR);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_SQRT_FC32(void)
{
  return &(GxB_SQRT_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_SQRT_FC64(void)
{
  return &(GxB_SQRT_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_SQRT_FP32(void)
{
  return &(GxB_SQRT_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_SQRT_FP64(void)
{
  return &(GxB_SQRT_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_TANH_FC32(void)
{
  return &(GxB_TANH_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_TANH_FC64(void)
{
  return &(GxB_TANH_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_TANH_FP32(void)
{
  return &(GxB_TANH_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_TANH_FP64(void)
{
  return &(GxB_TANH_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_TAN_FC32(void)
{
  return &(GxB_TAN_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_TAN_FC64(void)
{
  return &(GxB_TAN_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_TAN_FP32(void)
{
  return &(GxB_TAN_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_TAN_FP64(void)
{
  return &(GxB_TAN_FP64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_TGAMMA_FP32(void)
{
  return &(GxB_TGAMMA_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_TGAMMA_FP64(void)
{
  return &(GxB_TGAMMA_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_DIV_FC32(void)
{
  return &(GxB_TIMES_DIV_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_DIV_FC64(void)
{
  return &(GxB_TIMES_DIV_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_DIV_FP32(void)
{
  return &(GxB_TIMES_DIV_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_DIV_FP64(void)
{
  return &(GxB_TIMES_DIV_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_DIV_INT16(void)
{
  return &(GxB_TIMES_DIV_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_DIV_INT32(void)
{
  return &(GxB_TIMES_DIV_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_DIV_INT64(void)
{
  return &(GxB_TIMES_DIV_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_DIV_INT8(void)
{
  return &(GxB_TIMES_DIV_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_DIV_UINT16(void)
{
  return &(GxB_TIMES_DIV_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_DIV_UINT32(void)
{
  return &(GxB_TIMES_DIV_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_DIV_UINT64(void)
{
  return &(GxB_TIMES_DIV_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_DIV_UINT8(void)
{
  return &(GxB_TIMES_DIV_UINT8);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_TIMES_FC32(void)
{
  return &(GxB_TIMES_FC32);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_TIMES_FC32_MONOID(void)
{
  return &(GxB_TIMES_FC32_MONOID);
}

static struct GB_BinaryOp_opaque * *_cffi_var_GxB_TIMES_FC64(void)
{
  return &(GxB_TIMES_FC64);
}

static struct GB_Monoid_opaque * *_cffi_var_GxB_TIMES_FC64_MONOID(void)
{
  return &(GxB_TIMES_FC64_MONOID);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_FIRSTI1_INT32(void)
{
  return &(GxB_TIMES_FIRSTI1_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_FIRSTI1_INT64(void)
{
  return &(GxB_TIMES_FIRSTI1_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_FIRSTI_INT32(void)
{
  return &(GxB_TIMES_FIRSTI_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_FIRSTI_INT64(void)
{
  return &(GxB_TIMES_FIRSTI_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_FIRSTJ1_INT32(void)
{
  return &(GxB_TIMES_FIRSTJ1_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_FIRSTJ1_INT64(void)
{
  return &(GxB_TIMES_FIRSTJ1_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_FIRSTJ_INT32(void)
{
  return &(GxB_TIMES_FIRSTJ_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_FIRSTJ_INT64(void)
{
  return &(GxB_TIMES_FIRSTJ_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_FIRST_FC32(void)
{
  return &(GxB_TIMES_FIRST_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_FIRST_FC64(void)
{
  return &(GxB_TIMES_FIRST_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_FIRST_FP32(void)
{
  return &(GxB_TIMES_FIRST_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_FIRST_FP64(void)
{
  return &(GxB_TIMES_FIRST_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_FIRST_INT16(void)
{
  return &(GxB_TIMES_FIRST_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_FIRST_INT32(void)
{
  return &(GxB_TIMES_FIRST_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_FIRST_INT64(void)
{
  return &(GxB_TIMES_FIRST_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_FIRST_INT8(void)
{
  return &(GxB_TIMES_FIRST_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_FIRST_UINT16(void)
{
  return &(GxB_TIMES_FIRST_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_FIRST_UINT32(void)
{
  return &(GxB_TIMES_FIRST_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_FIRST_UINT64(void)
{
  return &(GxB_TIMES_FIRST_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_FIRST_UINT8(void)
{
  return &(GxB_TIMES_FIRST_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISEQ_FP32(void)
{
  return &(GxB_TIMES_ISEQ_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISEQ_FP64(void)
{
  return &(GxB_TIMES_ISEQ_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISEQ_INT16(void)
{
  return &(GxB_TIMES_ISEQ_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISEQ_INT32(void)
{
  return &(GxB_TIMES_ISEQ_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISEQ_INT64(void)
{
  return &(GxB_TIMES_ISEQ_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISEQ_INT8(void)
{
  return &(GxB_TIMES_ISEQ_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISEQ_UINT16(void)
{
  return &(GxB_TIMES_ISEQ_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISEQ_UINT32(void)
{
  return &(GxB_TIMES_ISEQ_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISEQ_UINT64(void)
{
  return &(GxB_TIMES_ISEQ_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISEQ_UINT8(void)
{
  return &(GxB_TIMES_ISEQ_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISGE_FP32(void)
{
  return &(GxB_TIMES_ISGE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISGE_FP64(void)
{
  return &(GxB_TIMES_ISGE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISGE_INT16(void)
{
  return &(GxB_TIMES_ISGE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISGE_INT32(void)
{
  return &(GxB_TIMES_ISGE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISGE_INT64(void)
{
  return &(GxB_TIMES_ISGE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISGE_INT8(void)
{
  return &(GxB_TIMES_ISGE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISGE_UINT16(void)
{
  return &(GxB_TIMES_ISGE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISGE_UINT32(void)
{
  return &(GxB_TIMES_ISGE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISGE_UINT64(void)
{
  return &(GxB_TIMES_ISGE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISGE_UINT8(void)
{
  return &(GxB_TIMES_ISGE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISGT_FP32(void)
{
  return &(GxB_TIMES_ISGT_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISGT_FP64(void)
{
  return &(GxB_TIMES_ISGT_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISGT_INT16(void)
{
  return &(GxB_TIMES_ISGT_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISGT_INT32(void)
{
  return &(GxB_TIMES_ISGT_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISGT_INT64(void)
{
  return &(GxB_TIMES_ISGT_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISGT_INT8(void)
{
  return &(GxB_TIMES_ISGT_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISGT_UINT16(void)
{
  return &(GxB_TIMES_ISGT_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISGT_UINT32(void)
{
  return &(GxB_TIMES_ISGT_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISGT_UINT64(void)
{
  return &(GxB_TIMES_ISGT_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISGT_UINT8(void)
{
  return &(GxB_TIMES_ISGT_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISLE_FP32(void)
{
  return &(GxB_TIMES_ISLE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISLE_FP64(void)
{
  return &(GxB_TIMES_ISLE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISLE_INT16(void)
{
  return &(GxB_TIMES_ISLE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISLE_INT32(void)
{
  return &(GxB_TIMES_ISLE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISLE_INT64(void)
{
  return &(GxB_TIMES_ISLE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISLE_INT8(void)
{
  return &(GxB_TIMES_ISLE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISLE_UINT16(void)
{
  return &(GxB_TIMES_ISLE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISLE_UINT32(void)
{
  return &(GxB_TIMES_ISLE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISLE_UINT64(void)
{
  return &(GxB_TIMES_ISLE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISLE_UINT8(void)
{
  return &(GxB_TIMES_ISLE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISLT_FP32(void)
{
  return &(GxB_TIMES_ISLT_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISLT_FP64(void)
{
  return &(GxB_TIMES_ISLT_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISLT_INT16(void)
{
  return &(GxB_TIMES_ISLT_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISLT_INT32(void)
{
  return &(GxB_TIMES_ISLT_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISLT_INT64(void)
{
  return &(GxB_TIMES_ISLT_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISLT_INT8(void)
{
  return &(GxB_TIMES_ISLT_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISLT_UINT16(void)
{
  return &(GxB_TIMES_ISLT_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISLT_UINT32(void)
{
  return &(GxB_TIMES_ISLT_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISLT_UINT64(void)
{
  return &(GxB_TIMES_ISLT_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISLT_UINT8(void)
{
  return &(GxB_TIMES_ISLT_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISNE_FP32(void)
{
  return &(GxB_TIMES_ISNE_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISNE_FP64(void)
{
  return &(GxB_TIMES_ISNE_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISNE_INT16(void)
{
  return &(GxB_TIMES_ISNE_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISNE_INT32(void)
{
  return &(GxB_TIMES_ISNE_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISNE_INT64(void)
{
  return &(GxB_TIMES_ISNE_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISNE_INT8(void)
{
  return &(GxB_TIMES_ISNE_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISNE_UINT16(void)
{
  return &(GxB_TIMES_ISNE_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISNE_UINT32(void)
{
  return &(GxB_TIMES_ISNE_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISNE_UINT64(void)
{
  return &(GxB_TIMES_ISNE_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_ISNE_UINT8(void)
{
  return &(GxB_TIMES_ISNE_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LAND_FP32(void)
{
  return &(GxB_TIMES_LAND_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LAND_FP64(void)
{
  return &(GxB_TIMES_LAND_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LAND_INT16(void)
{
  return &(GxB_TIMES_LAND_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LAND_INT32(void)
{
  return &(GxB_TIMES_LAND_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LAND_INT64(void)
{
  return &(GxB_TIMES_LAND_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LAND_INT8(void)
{
  return &(GxB_TIMES_LAND_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LAND_UINT16(void)
{
  return &(GxB_TIMES_LAND_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LAND_UINT32(void)
{
  return &(GxB_TIMES_LAND_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LAND_UINT64(void)
{
  return &(GxB_TIMES_LAND_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LAND_UINT8(void)
{
  return &(GxB_TIMES_LAND_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LOR_FP32(void)
{
  return &(GxB_TIMES_LOR_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LOR_FP64(void)
{
  return &(GxB_TIMES_LOR_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LOR_INT16(void)
{
  return &(GxB_TIMES_LOR_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LOR_INT32(void)
{
  return &(GxB_TIMES_LOR_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LOR_INT64(void)
{
  return &(GxB_TIMES_LOR_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LOR_INT8(void)
{
  return &(GxB_TIMES_LOR_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LOR_UINT16(void)
{
  return &(GxB_TIMES_LOR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LOR_UINT32(void)
{
  return &(GxB_TIMES_LOR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LOR_UINT64(void)
{
  return &(GxB_TIMES_LOR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LOR_UINT8(void)
{
  return &(GxB_TIMES_LOR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LXOR_FP32(void)
{
  return &(GxB_TIMES_LXOR_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LXOR_FP64(void)
{
  return &(GxB_TIMES_LXOR_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LXOR_INT16(void)
{
  return &(GxB_TIMES_LXOR_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LXOR_INT32(void)
{
  return &(GxB_TIMES_LXOR_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LXOR_INT64(void)
{
  return &(GxB_TIMES_LXOR_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LXOR_INT8(void)
{
  return &(GxB_TIMES_LXOR_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LXOR_UINT16(void)
{
  return &(GxB_TIMES_LXOR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LXOR_UINT32(void)
{
  return &(GxB_TIMES_LXOR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LXOR_UINT64(void)
{
  return &(GxB_TIMES_LXOR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_LXOR_UINT8(void)
{
  return &(GxB_TIMES_LXOR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MAX_FP32(void)
{
  return &(GxB_TIMES_MAX_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MAX_FP64(void)
{
  return &(GxB_TIMES_MAX_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MAX_INT16(void)
{
  return &(GxB_TIMES_MAX_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MAX_INT32(void)
{
  return &(GxB_TIMES_MAX_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MAX_INT64(void)
{
  return &(GxB_TIMES_MAX_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MAX_INT8(void)
{
  return &(GxB_TIMES_MAX_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MAX_UINT16(void)
{
  return &(GxB_TIMES_MAX_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MAX_UINT32(void)
{
  return &(GxB_TIMES_MAX_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MAX_UINT64(void)
{
  return &(GxB_TIMES_MAX_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MAX_UINT8(void)
{
  return &(GxB_TIMES_MAX_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MINUS_FC32(void)
{
  return &(GxB_TIMES_MINUS_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MINUS_FC64(void)
{
  return &(GxB_TIMES_MINUS_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MINUS_FP32(void)
{
  return &(GxB_TIMES_MINUS_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MINUS_FP64(void)
{
  return &(GxB_TIMES_MINUS_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MINUS_INT16(void)
{
  return &(GxB_TIMES_MINUS_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MINUS_INT32(void)
{
  return &(GxB_TIMES_MINUS_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MINUS_INT64(void)
{
  return &(GxB_TIMES_MINUS_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MINUS_INT8(void)
{
  return &(GxB_TIMES_MINUS_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MINUS_UINT16(void)
{
  return &(GxB_TIMES_MINUS_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MINUS_UINT32(void)
{
  return &(GxB_TIMES_MINUS_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MINUS_UINT64(void)
{
  return &(GxB_TIMES_MINUS_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MINUS_UINT8(void)
{
  return &(GxB_TIMES_MINUS_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MIN_FP32(void)
{
  return &(GxB_TIMES_MIN_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MIN_FP64(void)
{
  return &(GxB_TIMES_MIN_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MIN_INT16(void)
{
  return &(GxB_TIMES_MIN_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MIN_INT32(void)
{
  return &(GxB_TIMES_MIN_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MIN_INT64(void)
{
  return &(GxB_TIMES_MIN_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MIN_INT8(void)
{
  return &(GxB_TIMES_MIN_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MIN_UINT16(void)
{
  return &(GxB_TIMES_MIN_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MIN_UINT32(void)
{
  return &(GxB_TIMES_MIN_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MIN_UINT64(void)
{
  return &(GxB_TIMES_MIN_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_MIN_UINT8(void)
{
  return &(GxB_TIMES_MIN_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PAIR_FC32(void)
{
  return &(GxB_TIMES_PAIR_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PAIR_FC64(void)
{
  return &(GxB_TIMES_PAIR_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PAIR_FP32(void)
{
  return &(GxB_TIMES_PAIR_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PAIR_FP64(void)
{
  return &(GxB_TIMES_PAIR_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PAIR_INT16(void)
{
  return &(GxB_TIMES_PAIR_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PAIR_INT32(void)
{
  return &(GxB_TIMES_PAIR_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PAIR_INT64(void)
{
  return &(GxB_TIMES_PAIR_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PAIR_INT8(void)
{
  return &(GxB_TIMES_PAIR_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PAIR_UINT16(void)
{
  return &(GxB_TIMES_PAIR_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PAIR_UINT32(void)
{
  return &(GxB_TIMES_PAIR_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PAIR_UINT64(void)
{
  return &(GxB_TIMES_PAIR_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PAIR_UINT8(void)
{
  return &(GxB_TIMES_PAIR_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PLUS_FC32(void)
{
  return &(GxB_TIMES_PLUS_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PLUS_FC64(void)
{
  return &(GxB_TIMES_PLUS_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PLUS_FP32(void)
{
  return &(GxB_TIMES_PLUS_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PLUS_FP64(void)
{
  return &(GxB_TIMES_PLUS_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PLUS_INT16(void)
{
  return &(GxB_TIMES_PLUS_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PLUS_INT32(void)
{
  return &(GxB_TIMES_PLUS_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PLUS_INT64(void)
{
  return &(GxB_TIMES_PLUS_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PLUS_INT8(void)
{
  return &(GxB_TIMES_PLUS_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PLUS_UINT16(void)
{
  return &(GxB_TIMES_PLUS_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PLUS_UINT32(void)
{
  return &(GxB_TIMES_PLUS_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PLUS_UINT64(void)
{
  return &(GxB_TIMES_PLUS_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_PLUS_UINT8(void)
{
  return &(GxB_TIMES_PLUS_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RDIV_FC32(void)
{
  return &(GxB_TIMES_RDIV_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RDIV_FC64(void)
{
  return &(GxB_TIMES_RDIV_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RDIV_FP32(void)
{
  return &(GxB_TIMES_RDIV_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RDIV_FP64(void)
{
  return &(GxB_TIMES_RDIV_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RDIV_INT16(void)
{
  return &(GxB_TIMES_RDIV_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RDIV_INT32(void)
{
  return &(GxB_TIMES_RDIV_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RDIV_INT64(void)
{
  return &(GxB_TIMES_RDIV_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RDIV_INT8(void)
{
  return &(GxB_TIMES_RDIV_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RDIV_UINT16(void)
{
  return &(GxB_TIMES_RDIV_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RDIV_UINT32(void)
{
  return &(GxB_TIMES_RDIV_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RDIV_UINT64(void)
{
  return &(GxB_TIMES_RDIV_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RDIV_UINT8(void)
{
  return &(GxB_TIMES_RDIV_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RMINUS_FC32(void)
{
  return &(GxB_TIMES_RMINUS_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RMINUS_FC64(void)
{
  return &(GxB_TIMES_RMINUS_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RMINUS_FP32(void)
{
  return &(GxB_TIMES_RMINUS_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RMINUS_FP64(void)
{
  return &(GxB_TIMES_RMINUS_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RMINUS_INT16(void)
{
  return &(GxB_TIMES_RMINUS_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RMINUS_INT32(void)
{
  return &(GxB_TIMES_RMINUS_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RMINUS_INT64(void)
{
  return &(GxB_TIMES_RMINUS_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RMINUS_INT8(void)
{
  return &(GxB_TIMES_RMINUS_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RMINUS_UINT16(void)
{
  return &(GxB_TIMES_RMINUS_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RMINUS_UINT32(void)
{
  return &(GxB_TIMES_RMINUS_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RMINUS_UINT64(void)
{
  return &(GxB_TIMES_RMINUS_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_RMINUS_UINT8(void)
{
  return &(GxB_TIMES_RMINUS_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_SECONDI1_INT32(void)
{
  return &(GxB_TIMES_SECONDI1_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_SECONDI1_INT64(void)
{
  return &(GxB_TIMES_SECONDI1_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_SECONDI_INT32(void)
{
  return &(GxB_TIMES_SECONDI_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_SECONDI_INT64(void)
{
  return &(GxB_TIMES_SECONDI_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_SECONDJ1_INT32(void)
{
  return &(GxB_TIMES_SECONDJ1_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_SECONDJ1_INT64(void)
{
  return &(GxB_TIMES_SECONDJ1_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_SECONDJ_INT32(void)
{
  return &(GxB_TIMES_SECONDJ_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_SECONDJ_INT64(void)
{
  return &(GxB_TIMES_SECONDJ_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_SECOND_FC32(void)
{
  return &(GxB_TIMES_SECOND_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_SECOND_FC64(void)
{
  return &(GxB_TIMES_SECOND_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_SECOND_FP32(void)
{
  return &(GxB_TIMES_SECOND_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_SECOND_FP64(void)
{
  return &(GxB_TIMES_SECOND_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_SECOND_INT16(void)
{
  return &(GxB_TIMES_SECOND_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_SECOND_INT32(void)
{
  return &(GxB_TIMES_SECOND_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_SECOND_INT64(void)
{
  return &(GxB_TIMES_SECOND_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_SECOND_INT8(void)
{
  return &(GxB_TIMES_SECOND_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_SECOND_UINT16(void)
{
  return &(GxB_TIMES_SECOND_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_SECOND_UINT32(void)
{
  return &(GxB_TIMES_SECOND_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_SECOND_UINT64(void)
{
  return &(GxB_TIMES_SECOND_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_SECOND_UINT8(void)
{
  return &(GxB_TIMES_SECOND_UINT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_TIMES_FC32(void)
{
  return &(GxB_TIMES_TIMES_FC32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_TIMES_FC64(void)
{
  return &(GxB_TIMES_TIMES_FC64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_TIMES_FP32(void)
{
  return &(GxB_TIMES_TIMES_FP32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_TIMES_FP64(void)
{
  return &(GxB_TIMES_TIMES_FP64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_TIMES_INT16(void)
{
  return &(GxB_TIMES_TIMES_INT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_TIMES_INT32(void)
{
  return &(GxB_TIMES_TIMES_INT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_TIMES_INT64(void)
{
  return &(GxB_TIMES_TIMES_INT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_TIMES_INT8(void)
{
  return &(GxB_TIMES_TIMES_INT8);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_TIMES_UINT16(void)
{
  return &(GxB_TIMES_TIMES_UINT16);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_TIMES_UINT32(void)
{
  return &(GxB_TIMES_TIMES_UINT32);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_TIMES_UINT64(void)
{
  return &(GxB_TIMES_TIMES_UINT64);
}

static struct GB_Semiring_opaque * *_cffi_var_GxB_TIMES_TIMES_UINT8(void)
{
  return &(GxB_TIMES_TIMES_UINT8);
}

static struct GB_SelectOp_opaque * *_cffi_var_GxB_TRIL(void)
{
  return &(GxB_TRIL);
}

static struct GB_SelectOp_opaque * *_cffi_var_GxB_TRIU(void)
{
  return &(GxB_TRIU);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_TRUNC_FC32(void)
{
  return &(GxB_TRUNC_FC32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_TRUNC_FC64(void)
{
  return &(GxB_TRUNC_FC64);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_TRUNC_FP32(void)
{
  return &(GxB_TRUNC_FP32);
}

static struct GB_UnaryOp_opaque * *_cffi_var_GxB_TRUNC_FP64(void)
{
  return &(GxB_TRUNC_FP64);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GxB_VALUEEQ_FC32(void)
{
  return &(GxB_VALUEEQ_FC32);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GxB_VALUEEQ_FC64(void)
{
  return &(GxB_VALUEEQ_FC64);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GxB_VALUENE_FC32(void)
{
  return &(GxB_VALUENE_FC32);
}

static struct GB_IndexUnaryOp_opaque * *_cffi_var_GxB_VALUENE_FC64(void)
{
  return &(GxB_VALUENE_FC64);
}

static const struct _cffi_global_s _cffi_globals[] = {
  { "GB_BinaryOp_new", (void *)_cffi_f_GB_BinaryOp_new, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 439), (void *)_cffi_d_GB_BinaryOp_new },
  { "GB_Iterator_attach", (void *)_cffi_f_GB_Iterator_attach, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 543), (void *)_cffi_d_GB_Iterator_attach },
  { "GB_Iterator_rc_bitmap_next", (void *)_cffi_f_GB_Iterator_rc_bitmap_next, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 540), (void *)_cffi_d_GB_Iterator_rc_bitmap_next },
  { "GB_Iterator_rc_seek", (void *)_cffi_f_GB_Iterator_rc_seek, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 563), (void *)_cffi_d_GB_Iterator_rc_seek },
  { "GB_SelectOp_new", (void *)_cffi_f_GB_SelectOp_new, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2236), (void *)_cffi_d_GB_SelectOp_new },
  { "GB_Type_new", (void *)_cffi_f_GB_Type_new, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2282), (void *)_cffi_d_GB_Type_new },
  { "GB_UnaryOp_new", (void *)_cffi_f_GB_UnaryOp_new, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2336), (void *)_cffi_d_GB_UnaryOp_new },
  { "GB_Vector_Iterator_bitmap_seek", (void *)_cffi_f_GB_Vector_Iterator_bitmap_seek, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 559), (void *)_cffi_d_GB_Vector_Iterator_bitmap_seek },
  { "GRB_SUBVERSION", (void *)_cffi_const_GRB_SUBVERSION, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GRB_VERSION", (void *)_cffi_const_GRB_VERSION, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GrB_ABS_BOOL", (void *)_cffi_var_GrB_ABS_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_ABS_FP32", (void *)_cffi_var_GrB_ABS_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_ABS_FP64", (void *)_cffi_var_GrB_ABS_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_ABS_INT16", (void *)_cffi_var_GrB_ABS_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_ABS_INT32", (void *)_cffi_var_GrB_ABS_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_ABS_INT64", (void *)_cffi_var_GrB_ABS_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_ABS_INT8", (void *)_cffi_var_GrB_ABS_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_ABS_UINT16", (void *)_cffi_var_GrB_ABS_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_ABS_UINT32", (void *)_cffi_var_GrB_ABS_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_ABS_UINT64", (void *)_cffi_var_GrB_ABS_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_ABS_UINT8", (void *)_cffi_var_GrB_ABS_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_AINV_BOOL", (void *)_cffi_var_GrB_AINV_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_AINV_FP32", (void *)_cffi_var_GrB_AINV_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_AINV_FP64", (void *)_cffi_var_GrB_AINV_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_AINV_INT16", (void *)_cffi_var_GrB_AINV_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_AINV_INT32", (void *)_cffi_var_GrB_AINV_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_AINV_INT64", (void *)_cffi_var_GrB_AINV_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_AINV_INT8", (void *)_cffi_var_GrB_AINV_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_AINV_UINT16", (void *)_cffi_var_GrB_AINV_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_AINV_UINT32", (void *)_cffi_var_GrB_AINV_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_AINV_UINT64", (void *)_cffi_var_GrB_AINV_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_AINV_UINT8", (void *)_cffi_var_GrB_AINV_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_ALL", (void *)_cffi_var_GrB_ALL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 696), (void *)0 },
  { "GrB_BAND_INT16", (void *)_cffi_var_GrB_BAND_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BAND_INT32", (void *)_cffi_var_GrB_BAND_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BAND_INT64", (void *)_cffi_var_GrB_BAND_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BAND_INT8", (void *)_cffi_var_GrB_BAND_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BAND_UINT16", (void *)_cffi_var_GrB_BAND_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BAND_UINT32", (void *)_cffi_var_GrB_BAND_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BAND_UINT64", (void *)_cffi_var_GrB_BAND_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BAND_UINT8", (void *)_cffi_var_GrB_BAND_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BLOCKING", (void *)_cffi_const_GrB_BLOCKING, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_BNOT_INT16", (void *)_cffi_var_GrB_BNOT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_BNOT_INT32", (void *)_cffi_var_GrB_BNOT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_BNOT_INT64", (void *)_cffi_var_GrB_BNOT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_BNOT_INT8", (void *)_cffi_var_GrB_BNOT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_BNOT_UINT16", (void *)_cffi_var_GrB_BNOT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_BNOT_UINT32", (void *)_cffi_var_GrB_BNOT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_BNOT_UINT64", (void *)_cffi_var_GrB_BNOT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_BNOT_UINT8", (void *)_cffi_var_GrB_BNOT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_BOOL", (void *)_cffi_var_GrB_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 114), (void *)0 },
  { "GrB_BOR_INT16", (void *)_cffi_var_GrB_BOR_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BOR_INT32", (void *)_cffi_var_GrB_BOR_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BOR_INT64", (void *)_cffi_var_GrB_BOR_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BOR_INT8", (void *)_cffi_var_GrB_BOR_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BOR_UINT16", (void *)_cffi_var_GrB_BOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BOR_UINT32", (void *)_cffi_var_GrB_BOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BOR_UINT64", (void *)_cffi_var_GrB_BOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BOR_UINT8", (void *)_cffi_var_GrB_BOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BXNOR_INT16", (void *)_cffi_var_GrB_BXNOR_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BXNOR_INT32", (void *)_cffi_var_GrB_BXNOR_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BXNOR_INT64", (void *)_cffi_var_GrB_BXNOR_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BXNOR_INT8", (void *)_cffi_var_GrB_BXNOR_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BXNOR_UINT16", (void *)_cffi_var_GrB_BXNOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BXNOR_UINT32", (void *)_cffi_var_GrB_BXNOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BXNOR_UINT64", (void *)_cffi_var_GrB_BXNOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BXNOR_UINT8", (void *)_cffi_var_GrB_BXNOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BXOR_INT16", (void *)_cffi_var_GrB_BXOR_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BXOR_INT32", (void *)_cffi_var_GrB_BXOR_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BXOR_INT64", (void *)_cffi_var_GrB_BXOR_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BXOR_INT8", (void *)_cffi_var_GrB_BXOR_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BXOR_UINT16", (void *)_cffi_var_GrB_BXOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BXOR_UINT32", (void *)_cffi_var_GrB_BXOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BXOR_UINT64", (void *)_cffi_var_GrB_BXOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BXOR_UINT8", (void *)_cffi_var_GrB_BXOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_BinaryOp_error", (void *)_cffi_f_GrB_BinaryOp_error, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 129), (void *)_cffi_d_GrB_BinaryOp_error },
  { "GrB_BinaryOp_free", (void *)_cffi_f_GrB_BinaryOp_free, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 421), (void *)_cffi_d_GrB_BinaryOp_free },
  { "GrB_BinaryOp_new", (void *)_cffi_f_GrB_BinaryOp_new, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 432), (void *)_cffi_d_GrB_BinaryOp_new },
  { "GrB_BinaryOp_wait", (void *)_cffi_f_GrB_BinaryOp_wait, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 456), (void *)_cffi_d_GrB_BinaryOp_wait },
  { "GrB_COLGT", (void *)_cffi_var_GrB_COLGT, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_COLINDEX_INT32", (void *)_cffi_var_GrB_COLINDEX_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_COLINDEX_INT64", (void *)_cffi_var_GrB_COLINDEX_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_COLLE", (void *)_cffi_var_GrB_COLLE, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_COMP", (void *)_cffi_const_GrB_COMP, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_COMPLETE", (void *)_cffi_const_GrB_COMPLETE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_COO_FORMAT", (void *)_cffi_const_GrB_COO_FORMAT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_CSC_FORMAT", (void *)_cffi_const_GrB_CSC_FORMAT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_CSR_FORMAT", (void *)_cffi_const_GrB_CSR_FORMAT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_Col_assign", (void *)_cffi_f_GrB_Col_assign, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1695), (void *)_cffi_d_GrB_Col_assign },
  { "GrB_Col_extract", (void *)_cffi_f_GrB_Col_extract, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3075), (void *)_cffi_d_GrB_Col_extract },
  { "GrB_DESC_C", (void *)_cffi_var_GrB_DESC_C, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_CT0", (void *)_cffi_var_GrB_DESC_CT0, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_CT0T1", (void *)_cffi_var_GrB_DESC_CT0T1, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_CT1", (void *)_cffi_var_GrB_DESC_CT1, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_R", (void *)_cffi_var_GrB_DESC_R, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_RC", (void *)_cffi_var_GrB_DESC_RC, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_RCT0", (void *)_cffi_var_GrB_DESC_RCT0, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_RCT0T1", (void *)_cffi_var_GrB_DESC_RCT0T1, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_RCT1", (void *)_cffi_var_GrB_DESC_RCT1, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_RS", (void *)_cffi_var_GrB_DESC_RS, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_RSC", (void *)_cffi_var_GrB_DESC_RSC, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_RSCT0", (void *)_cffi_var_GrB_DESC_RSCT0, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_RSCT0T1", (void *)_cffi_var_GrB_DESC_RSCT0T1, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_RSCT1", (void *)_cffi_var_GrB_DESC_RSCT1, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_RST0", (void *)_cffi_var_GrB_DESC_RST0, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_RST0T1", (void *)_cffi_var_GrB_DESC_RST0T1, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_RST1", (void *)_cffi_var_GrB_DESC_RST1, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_RT0", (void *)_cffi_var_GrB_DESC_RT0, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_RT0T1", (void *)_cffi_var_GrB_DESC_RT0T1, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_RT1", (void *)_cffi_var_GrB_DESC_RT1, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_S", (void *)_cffi_var_GrB_DESC_S, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_SC", (void *)_cffi_var_GrB_DESC_SC, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_SCT0", (void *)_cffi_var_GrB_DESC_SCT0, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_SCT0T1", (void *)_cffi_var_GrB_DESC_SCT0T1, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_SCT1", (void *)_cffi_var_GrB_DESC_SCT1, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_ST0", (void *)_cffi_var_GrB_DESC_ST0, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_ST0T1", (void *)_cffi_var_GrB_DESC_ST0T1, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_ST1", (void *)_cffi_var_GrB_DESC_ST1, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_T0", (void *)_cffi_var_GrB_DESC_T0, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_T0T1", (void *)_cffi_var_GrB_DESC_T0T1, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DESC_T1", (void *)_cffi_var_GrB_DESC_T1, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 2), (void *)0 },
  { "GrB_DIAG", (void *)_cffi_var_GrB_DIAG, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_DIAGINDEX_INT32", (void *)_cffi_var_GrB_DIAGINDEX_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_DIAGINDEX_INT64", (void *)_cffi_var_GrB_DIAGINDEX_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_DIMENSION_MISMATCH", (void *)_cffi_const_GrB_DIMENSION_MISMATCH, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_DIV_BOOL", (void *)_cffi_var_GrB_DIV_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_DIV_FP32", (void *)_cffi_var_GrB_DIV_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_DIV_FP64", (void *)_cffi_var_GrB_DIV_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_DIV_INT16", (void *)_cffi_var_GrB_DIV_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_DIV_INT32", (void *)_cffi_var_GrB_DIV_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_DIV_INT64", (void *)_cffi_var_GrB_DIV_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_DIV_INT8", (void *)_cffi_var_GrB_DIV_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_DIV_UINT16", (void *)_cffi_var_GrB_DIV_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_DIV_UINT32", (void *)_cffi_var_GrB_DIV_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_DIV_UINT64", (void *)_cffi_var_GrB_DIV_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_DIV_UINT8", (void *)_cffi_var_GrB_DIV_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_DOMAIN_MISMATCH", (void *)_cffi_const_GrB_DOMAIN_MISMATCH, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_Descriptor_error", (void *)_cffi_f_GrB_Descriptor_error, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 133), (void *)_cffi_d_GrB_Descriptor_error },
  { "GrB_Descriptor_free", (void *)_cffi_f_GrB_Descriptor_free, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 466), (void *)_cffi_d_GrB_Descriptor_free },
  { "GrB_Descriptor_new", (void *)_cffi_f_GrB_Descriptor_new, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 466), (void *)_cffi_d_GrB_Descriptor_new },
  { "GrB_Descriptor_set", (void *)_cffi_f_GrB_Descriptor_set, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 473), (void *)_cffi_d_GrB_Descriptor_set },
  { "GrB_Descriptor_wait", (void *)_cffi_f_GrB_Descriptor_wait, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 498), (void *)_cffi_d_GrB_Descriptor_wait },
  { "GrB_EMPTY_OBJECT", (void *)_cffi_const_GrB_EMPTY_OBJECT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_EQ_BOOL", (void *)_cffi_var_GrB_EQ_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_EQ_FP32", (void *)_cffi_var_GrB_EQ_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_EQ_FP64", (void *)_cffi_var_GrB_EQ_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_EQ_INT16", (void *)_cffi_var_GrB_EQ_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_EQ_INT32", (void *)_cffi_var_GrB_EQ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_EQ_INT64", (void *)_cffi_var_GrB_EQ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_EQ_INT8", (void *)_cffi_var_GrB_EQ_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_EQ_UINT16", (void *)_cffi_var_GrB_EQ_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_EQ_UINT32", (void *)_cffi_var_GrB_EQ_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_EQ_UINT64", (void *)_cffi_var_GrB_EQ_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_EQ_UINT8", (void *)_cffi_var_GrB_EQ_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_FIRST_BOOL", (void *)_cffi_var_GrB_FIRST_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_FIRST_FP32", (void *)_cffi_var_GrB_FIRST_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_FIRST_FP64", (void *)_cffi_var_GrB_FIRST_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_FIRST_INT16", (void *)_cffi_var_GrB_FIRST_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_FIRST_INT32", (void *)_cffi_var_GrB_FIRST_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_FIRST_INT64", (void *)_cffi_var_GrB_FIRST_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_FIRST_INT8", (void *)_cffi_var_GrB_FIRST_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_FIRST_UINT16", (void *)_cffi_var_GrB_FIRST_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_FIRST_UINT32", (void *)_cffi_var_GrB_FIRST_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_FIRST_UINT64", (void *)_cffi_var_GrB_FIRST_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_FIRST_UINT8", (void *)_cffi_var_GrB_FIRST_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_FP32", (void *)_cffi_var_GrB_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 114), (void *)0 },
  { "GrB_FP64", (void *)_cffi_var_GrB_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 114), (void *)0 },
  { "GrB_GE_BOOL", (void *)_cffi_var_GrB_GE_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_GE_FP32", (void *)_cffi_var_GrB_GE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_GE_FP64", (void *)_cffi_var_GrB_GE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_GE_INT16", (void *)_cffi_var_GrB_GE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_GE_INT32", (void *)_cffi_var_GrB_GE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_GE_INT64", (void *)_cffi_var_GrB_GE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_GE_INT8", (void *)_cffi_var_GrB_GE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_GE_UINT16", (void *)_cffi_var_GrB_GE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_GE_UINT32", (void *)_cffi_var_GrB_GE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_GE_UINT64", (void *)_cffi_var_GrB_GE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_GE_UINT8", (void *)_cffi_var_GrB_GE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_GT_BOOL", (void *)_cffi_var_GrB_GT_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_GT_FP32", (void *)_cffi_var_GrB_GT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_GT_FP64", (void *)_cffi_var_GrB_GT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_GT_INT16", (void *)_cffi_var_GrB_GT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_GT_INT32", (void *)_cffi_var_GrB_GT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_GT_INT64", (void *)_cffi_var_GrB_GT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_GT_INT8", (void *)_cffi_var_GrB_GT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_GT_UINT16", (void *)_cffi_var_GrB_GT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_GT_UINT32", (void *)_cffi_var_GrB_GT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_GT_UINT64", (void *)_cffi_var_GrB_GT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_GT_UINT8", (void *)_cffi_var_GrB_GT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_IDENTITY_BOOL", (void *)_cffi_var_GrB_IDENTITY_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_IDENTITY_FP32", (void *)_cffi_var_GrB_IDENTITY_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_IDENTITY_FP64", (void *)_cffi_var_GrB_IDENTITY_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_IDENTITY_INT16", (void *)_cffi_var_GrB_IDENTITY_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_IDENTITY_INT32", (void *)_cffi_var_GrB_IDENTITY_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_IDENTITY_INT64", (void *)_cffi_var_GrB_IDENTITY_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_IDENTITY_INT8", (void *)_cffi_var_GrB_IDENTITY_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_IDENTITY_UINT16", (void *)_cffi_var_GrB_IDENTITY_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_IDENTITY_UINT32", (void *)_cffi_var_GrB_IDENTITY_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_IDENTITY_UINT64", (void *)_cffi_var_GrB_IDENTITY_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_IDENTITY_UINT8", (void *)_cffi_var_GrB_IDENTITY_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_INDEX_MAX", (void *)_cffi_const_GrB_INDEX_MAX, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GrB_INDEX_OUT_OF_BOUNDS", (void *)_cffi_const_GrB_INDEX_OUT_OF_BOUNDS, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_INP0", (void *)_cffi_const_GrB_INP0, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_INP1", (void *)_cffi_const_GrB_INP1, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_INSUFFICIENT_SPACE", (void *)_cffi_const_GrB_INSUFFICIENT_SPACE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_INT16", (void *)_cffi_var_GrB_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 114), (void *)0 },
  { "GrB_INT32", (void *)_cffi_var_GrB_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 114), (void *)0 },
  { "GrB_INT64", (void *)_cffi_var_GrB_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 114), (void *)0 },
  { "GrB_INT8", (void *)_cffi_var_GrB_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 114), (void *)0 },
  { "GrB_INVALID_INDEX", (void *)_cffi_const_GrB_INVALID_INDEX, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_INVALID_OBJECT", (void *)_cffi_const_GrB_INVALID_OBJECT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_INVALID_VALUE", (void *)_cffi_const_GrB_INVALID_VALUE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_IndexUnaryOp_error", (void *)_cffi_f_GrB_IndexUnaryOp_error, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 137), (void *)_cffi_d_GrB_IndexUnaryOp_error },
  { "GrB_IndexUnaryOp_free", (void *)_cffi_f_GrB_IndexUnaryOp_free, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 508), (void *)_cffi_d_GrB_IndexUnaryOp_free },
  { "GrB_IndexUnaryOp_new", (void *)_cffi_f_GrB_IndexUnaryOp_new, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 511), (void *)_cffi_d_GrB_IndexUnaryOp_new },
  { "GrB_IndexUnaryOp_wait", (void *)_cffi_f_GrB_IndexUnaryOp_wait, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 527), (void *)_cffi_d_GrB_IndexUnaryOp_wait },
  { "GrB_LAND", (void *)_cffi_var_GrB_LAND, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LAND_LOR_SEMIRING_BOOL", (void *)_cffi_var_GrB_LAND_LOR_SEMIRING_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_LAND_MONOID_BOOL", (void *)_cffi_var_GrB_LAND_MONOID_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_LE_BOOL", (void *)_cffi_var_GrB_LE_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LE_FP32", (void *)_cffi_var_GrB_LE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LE_FP64", (void *)_cffi_var_GrB_LE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LE_INT16", (void *)_cffi_var_GrB_LE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LE_INT32", (void *)_cffi_var_GrB_LE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LE_INT64", (void *)_cffi_var_GrB_LE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LE_INT8", (void *)_cffi_var_GrB_LE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LE_UINT16", (void *)_cffi_var_GrB_LE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LE_UINT32", (void *)_cffi_var_GrB_LE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LE_UINT64", (void *)_cffi_var_GrB_LE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LE_UINT8", (void *)_cffi_var_GrB_LE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LNOT", (void *)_cffi_var_GrB_LNOT, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_LOR", (void *)_cffi_var_GrB_LOR, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LOR_LAND_SEMIRING_BOOL", (void *)_cffi_var_GrB_LOR_LAND_SEMIRING_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_LOR_MONOID_BOOL", (void *)_cffi_var_GrB_LOR_MONOID_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_LT_BOOL", (void *)_cffi_var_GrB_LT_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LT_FP32", (void *)_cffi_var_GrB_LT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LT_FP64", (void *)_cffi_var_GrB_LT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LT_INT16", (void *)_cffi_var_GrB_LT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LT_INT32", (void *)_cffi_var_GrB_LT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LT_INT64", (void *)_cffi_var_GrB_LT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LT_INT8", (void *)_cffi_var_GrB_LT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LT_UINT16", (void *)_cffi_var_GrB_LT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LT_UINT32", (void *)_cffi_var_GrB_LT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LT_UINT64", (void *)_cffi_var_GrB_LT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LT_UINT8", (void *)_cffi_var_GrB_LT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LXNOR", (void *)_cffi_var_GrB_LXNOR, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LXNOR_LOR_SEMIRING_BOOL", (void *)_cffi_var_GrB_LXNOR_LOR_SEMIRING_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_LXNOR_MONOID_BOOL", (void *)_cffi_var_GrB_LXNOR_MONOID_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_LXOR", (void *)_cffi_var_GrB_LXOR, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_LXOR_LAND_SEMIRING_BOOL", (void *)_cffi_var_GrB_LXOR_LAND_SEMIRING_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_LXOR_MONOID_BOOL", (void *)_cffi_var_GrB_LXOR_MONOID_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_MASK", (void *)_cffi_const_GrB_MASK, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_MATERIALIZE", (void *)_cffi_const_GrB_MATERIALIZE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_MAX_BOOL", (void *)_cffi_var_GrB_MAX_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MAX_FIRST_SEMIRING_FP32", (void *)_cffi_var_GrB_MAX_FIRST_SEMIRING_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_FIRST_SEMIRING_FP64", (void *)_cffi_var_GrB_MAX_FIRST_SEMIRING_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_FIRST_SEMIRING_INT16", (void *)_cffi_var_GrB_MAX_FIRST_SEMIRING_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_FIRST_SEMIRING_INT32", (void *)_cffi_var_GrB_MAX_FIRST_SEMIRING_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_FIRST_SEMIRING_INT64", (void *)_cffi_var_GrB_MAX_FIRST_SEMIRING_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_FIRST_SEMIRING_INT8", (void *)_cffi_var_GrB_MAX_FIRST_SEMIRING_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_FIRST_SEMIRING_UINT16", (void *)_cffi_var_GrB_MAX_FIRST_SEMIRING_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_FIRST_SEMIRING_UINT32", (void *)_cffi_var_GrB_MAX_FIRST_SEMIRING_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_FIRST_SEMIRING_UINT64", (void *)_cffi_var_GrB_MAX_FIRST_SEMIRING_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_FIRST_SEMIRING_UINT8", (void *)_cffi_var_GrB_MAX_FIRST_SEMIRING_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_FP32", (void *)_cffi_var_GrB_MAX_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MAX_FP64", (void *)_cffi_var_GrB_MAX_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MAX_INT16", (void *)_cffi_var_GrB_MAX_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MAX_INT32", (void *)_cffi_var_GrB_MAX_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MAX_INT64", (void *)_cffi_var_GrB_MAX_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MAX_INT8", (void *)_cffi_var_GrB_MAX_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MAX_MIN_SEMIRING_FP32", (void *)_cffi_var_GrB_MAX_MIN_SEMIRING_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_MIN_SEMIRING_FP64", (void *)_cffi_var_GrB_MAX_MIN_SEMIRING_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_MIN_SEMIRING_INT16", (void *)_cffi_var_GrB_MAX_MIN_SEMIRING_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_MIN_SEMIRING_INT32", (void *)_cffi_var_GrB_MAX_MIN_SEMIRING_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_MIN_SEMIRING_INT64", (void *)_cffi_var_GrB_MAX_MIN_SEMIRING_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_MIN_SEMIRING_INT8", (void *)_cffi_var_GrB_MAX_MIN_SEMIRING_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_MIN_SEMIRING_UINT16", (void *)_cffi_var_GrB_MAX_MIN_SEMIRING_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_MIN_SEMIRING_UINT32", (void *)_cffi_var_GrB_MAX_MIN_SEMIRING_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_MIN_SEMIRING_UINT64", (void *)_cffi_var_GrB_MAX_MIN_SEMIRING_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_MIN_SEMIRING_UINT8", (void *)_cffi_var_GrB_MAX_MIN_SEMIRING_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_MONOID_FP32", (void *)_cffi_var_GrB_MAX_MONOID_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_MAX_MONOID_FP64", (void *)_cffi_var_GrB_MAX_MONOID_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_MAX_MONOID_INT16", (void *)_cffi_var_GrB_MAX_MONOID_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_MAX_MONOID_INT32", (void *)_cffi_var_GrB_MAX_MONOID_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_MAX_MONOID_INT64", (void *)_cffi_var_GrB_MAX_MONOID_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_MAX_MONOID_INT8", (void *)_cffi_var_GrB_MAX_MONOID_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_MAX_MONOID_UINT16", (void *)_cffi_var_GrB_MAX_MONOID_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_MAX_MONOID_UINT32", (void *)_cffi_var_GrB_MAX_MONOID_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_MAX_MONOID_UINT64", (void *)_cffi_var_GrB_MAX_MONOID_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_MAX_MONOID_UINT8", (void *)_cffi_var_GrB_MAX_MONOID_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_MAX_PLUS_SEMIRING_FP32", (void *)_cffi_var_GrB_MAX_PLUS_SEMIRING_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_PLUS_SEMIRING_FP64", (void *)_cffi_var_GrB_MAX_PLUS_SEMIRING_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_PLUS_SEMIRING_INT16", (void *)_cffi_var_GrB_MAX_PLUS_SEMIRING_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_PLUS_SEMIRING_INT32", (void *)_cffi_var_GrB_MAX_PLUS_SEMIRING_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_PLUS_SEMIRING_INT64", (void *)_cffi_var_GrB_MAX_PLUS_SEMIRING_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_PLUS_SEMIRING_INT8", (void *)_cffi_var_GrB_MAX_PLUS_SEMIRING_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_PLUS_SEMIRING_UINT16", (void *)_cffi_var_GrB_MAX_PLUS_SEMIRING_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_PLUS_SEMIRING_UINT32", (void *)_cffi_var_GrB_MAX_PLUS_SEMIRING_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_PLUS_SEMIRING_UINT64", (void *)_cffi_var_GrB_MAX_PLUS_SEMIRING_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_PLUS_SEMIRING_UINT8", (void *)_cffi_var_GrB_MAX_PLUS_SEMIRING_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_SECOND_SEMIRING_FP32", (void *)_cffi_var_GrB_MAX_SECOND_SEMIRING_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_SECOND_SEMIRING_FP64", (void *)_cffi_var_GrB_MAX_SECOND_SEMIRING_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_SECOND_SEMIRING_INT16", (void *)_cffi_var_GrB_MAX_SECOND_SEMIRING_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_SECOND_SEMIRING_INT32", (void *)_cffi_var_GrB_MAX_SECOND_SEMIRING_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_SECOND_SEMIRING_INT64", (void *)_cffi_var_GrB_MAX_SECOND_SEMIRING_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_SECOND_SEMIRING_INT8", (void *)_cffi_var_GrB_MAX_SECOND_SEMIRING_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_SECOND_SEMIRING_UINT16", (void *)_cffi_var_GrB_MAX_SECOND_SEMIRING_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_SECOND_SEMIRING_UINT32", (void *)_cffi_var_GrB_MAX_SECOND_SEMIRING_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_SECOND_SEMIRING_UINT64", (void *)_cffi_var_GrB_MAX_SECOND_SEMIRING_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_SECOND_SEMIRING_UINT8", (void *)_cffi_var_GrB_MAX_SECOND_SEMIRING_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_TIMES_SEMIRING_FP32", (void *)_cffi_var_GrB_MAX_TIMES_SEMIRING_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_TIMES_SEMIRING_FP64", (void *)_cffi_var_GrB_MAX_TIMES_SEMIRING_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_TIMES_SEMIRING_INT16", (void *)_cffi_var_GrB_MAX_TIMES_SEMIRING_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_TIMES_SEMIRING_INT32", (void *)_cffi_var_GrB_MAX_TIMES_SEMIRING_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_TIMES_SEMIRING_INT64", (void *)_cffi_var_GrB_MAX_TIMES_SEMIRING_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_TIMES_SEMIRING_INT8", (void *)_cffi_var_GrB_MAX_TIMES_SEMIRING_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_TIMES_SEMIRING_UINT16", (void *)_cffi_var_GrB_MAX_TIMES_SEMIRING_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_TIMES_SEMIRING_UINT32", (void *)_cffi_var_GrB_MAX_TIMES_SEMIRING_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_TIMES_SEMIRING_UINT64", (void *)_cffi_var_GrB_MAX_TIMES_SEMIRING_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_TIMES_SEMIRING_UINT8", (void *)_cffi_var_GrB_MAX_TIMES_SEMIRING_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MAX_UINT16", (void *)_cffi_var_GrB_MAX_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MAX_UINT32", (void *)_cffi_var_GrB_MAX_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MAX_UINT64", (void *)_cffi_var_GrB_MAX_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MAX_UINT8", (void *)_cffi_var_GrB_MAX_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MINUS_BOOL", (void *)_cffi_var_GrB_MINUS_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MINUS_FP32", (void *)_cffi_var_GrB_MINUS_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MINUS_FP64", (void *)_cffi_var_GrB_MINUS_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MINUS_INT16", (void *)_cffi_var_GrB_MINUS_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MINUS_INT32", (void *)_cffi_var_GrB_MINUS_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MINUS_INT64", (void *)_cffi_var_GrB_MINUS_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MINUS_INT8", (void *)_cffi_var_GrB_MINUS_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MINUS_UINT16", (void *)_cffi_var_GrB_MINUS_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MINUS_UINT32", (void *)_cffi_var_GrB_MINUS_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MINUS_UINT64", (void *)_cffi_var_GrB_MINUS_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MINUS_UINT8", (void *)_cffi_var_GrB_MINUS_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MINV_BOOL", (void *)_cffi_var_GrB_MINV_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_MINV_FP32", (void *)_cffi_var_GrB_MINV_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_MINV_FP64", (void *)_cffi_var_GrB_MINV_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_MINV_INT16", (void *)_cffi_var_GrB_MINV_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_MINV_INT32", (void *)_cffi_var_GrB_MINV_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_MINV_INT64", (void *)_cffi_var_GrB_MINV_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_MINV_INT8", (void *)_cffi_var_GrB_MINV_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_MINV_UINT16", (void *)_cffi_var_GrB_MINV_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_MINV_UINT32", (void *)_cffi_var_GrB_MINV_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_MINV_UINT64", (void *)_cffi_var_GrB_MINV_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_MINV_UINT8", (void *)_cffi_var_GrB_MINV_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GrB_MIN_BOOL", (void *)_cffi_var_GrB_MIN_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MIN_FIRST_SEMIRING_FP32", (void *)_cffi_var_GrB_MIN_FIRST_SEMIRING_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_FIRST_SEMIRING_FP64", (void *)_cffi_var_GrB_MIN_FIRST_SEMIRING_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_FIRST_SEMIRING_INT16", (void *)_cffi_var_GrB_MIN_FIRST_SEMIRING_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_FIRST_SEMIRING_INT32", (void *)_cffi_var_GrB_MIN_FIRST_SEMIRING_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_FIRST_SEMIRING_INT64", (void *)_cffi_var_GrB_MIN_FIRST_SEMIRING_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_FIRST_SEMIRING_INT8", (void *)_cffi_var_GrB_MIN_FIRST_SEMIRING_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_FIRST_SEMIRING_UINT16", (void *)_cffi_var_GrB_MIN_FIRST_SEMIRING_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_FIRST_SEMIRING_UINT32", (void *)_cffi_var_GrB_MIN_FIRST_SEMIRING_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_FIRST_SEMIRING_UINT64", (void *)_cffi_var_GrB_MIN_FIRST_SEMIRING_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_FIRST_SEMIRING_UINT8", (void *)_cffi_var_GrB_MIN_FIRST_SEMIRING_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_FP32", (void *)_cffi_var_GrB_MIN_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MIN_FP64", (void *)_cffi_var_GrB_MIN_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MIN_INT16", (void *)_cffi_var_GrB_MIN_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MIN_INT32", (void *)_cffi_var_GrB_MIN_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MIN_INT64", (void *)_cffi_var_GrB_MIN_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MIN_INT8", (void *)_cffi_var_GrB_MIN_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MIN_MAX_SEMIRING_FP32", (void *)_cffi_var_GrB_MIN_MAX_SEMIRING_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_MAX_SEMIRING_FP64", (void *)_cffi_var_GrB_MIN_MAX_SEMIRING_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_MAX_SEMIRING_INT16", (void *)_cffi_var_GrB_MIN_MAX_SEMIRING_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_MAX_SEMIRING_INT32", (void *)_cffi_var_GrB_MIN_MAX_SEMIRING_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_MAX_SEMIRING_INT64", (void *)_cffi_var_GrB_MIN_MAX_SEMIRING_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_MAX_SEMIRING_INT8", (void *)_cffi_var_GrB_MIN_MAX_SEMIRING_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_MAX_SEMIRING_UINT16", (void *)_cffi_var_GrB_MIN_MAX_SEMIRING_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_MAX_SEMIRING_UINT32", (void *)_cffi_var_GrB_MIN_MAX_SEMIRING_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_MAX_SEMIRING_UINT64", (void *)_cffi_var_GrB_MIN_MAX_SEMIRING_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_MAX_SEMIRING_UINT8", (void *)_cffi_var_GrB_MIN_MAX_SEMIRING_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_MONOID_FP32", (void *)_cffi_var_GrB_MIN_MONOID_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_MIN_MONOID_FP64", (void *)_cffi_var_GrB_MIN_MONOID_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_MIN_MONOID_INT16", (void *)_cffi_var_GrB_MIN_MONOID_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_MIN_MONOID_INT32", (void *)_cffi_var_GrB_MIN_MONOID_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_MIN_MONOID_INT64", (void *)_cffi_var_GrB_MIN_MONOID_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_MIN_MONOID_INT8", (void *)_cffi_var_GrB_MIN_MONOID_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_MIN_MONOID_UINT16", (void *)_cffi_var_GrB_MIN_MONOID_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_MIN_MONOID_UINT32", (void *)_cffi_var_GrB_MIN_MONOID_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_MIN_MONOID_UINT64", (void *)_cffi_var_GrB_MIN_MONOID_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_MIN_MONOID_UINT8", (void *)_cffi_var_GrB_MIN_MONOID_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_MIN_PLUS_SEMIRING_FP32", (void *)_cffi_var_GrB_MIN_PLUS_SEMIRING_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_PLUS_SEMIRING_FP64", (void *)_cffi_var_GrB_MIN_PLUS_SEMIRING_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_PLUS_SEMIRING_INT16", (void *)_cffi_var_GrB_MIN_PLUS_SEMIRING_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_PLUS_SEMIRING_INT32", (void *)_cffi_var_GrB_MIN_PLUS_SEMIRING_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_PLUS_SEMIRING_INT64", (void *)_cffi_var_GrB_MIN_PLUS_SEMIRING_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_PLUS_SEMIRING_INT8", (void *)_cffi_var_GrB_MIN_PLUS_SEMIRING_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_PLUS_SEMIRING_UINT16", (void *)_cffi_var_GrB_MIN_PLUS_SEMIRING_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_PLUS_SEMIRING_UINT32", (void *)_cffi_var_GrB_MIN_PLUS_SEMIRING_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_PLUS_SEMIRING_UINT64", (void *)_cffi_var_GrB_MIN_PLUS_SEMIRING_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_PLUS_SEMIRING_UINT8", (void *)_cffi_var_GrB_MIN_PLUS_SEMIRING_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_SECOND_SEMIRING_FP32", (void *)_cffi_var_GrB_MIN_SECOND_SEMIRING_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_SECOND_SEMIRING_FP64", (void *)_cffi_var_GrB_MIN_SECOND_SEMIRING_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_SECOND_SEMIRING_INT16", (void *)_cffi_var_GrB_MIN_SECOND_SEMIRING_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_SECOND_SEMIRING_INT32", (void *)_cffi_var_GrB_MIN_SECOND_SEMIRING_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_SECOND_SEMIRING_INT64", (void *)_cffi_var_GrB_MIN_SECOND_SEMIRING_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_SECOND_SEMIRING_INT8", (void *)_cffi_var_GrB_MIN_SECOND_SEMIRING_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_SECOND_SEMIRING_UINT16", (void *)_cffi_var_GrB_MIN_SECOND_SEMIRING_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_SECOND_SEMIRING_UINT32", (void *)_cffi_var_GrB_MIN_SECOND_SEMIRING_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_SECOND_SEMIRING_UINT64", (void *)_cffi_var_GrB_MIN_SECOND_SEMIRING_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_SECOND_SEMIRING_UINT8", (void *)_cffi_var_GrB_MIN_SECOND_SEMIRING_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_TIMES_SEMIRING_FP32", (void *)_cffi_var_GrB_MIN_TIMES_SEMIRING_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_TIMES_SEMIRING_FP64", (void *)_cffi_var_GrB_MIN_TIMES_SEMIRING_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_TIMES_SEMIRING_INT16", (void *)_cffi_var_GrB_MIN_TIMES_SEMIRING_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_TIMES_SEMIRING_INT32", (void *)_cffi_var_GrB_MIN_TIMES_SEMIRING_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_TIMES_SEMIRING_INT64", (void *)_cffi_var_GrB_MIN_TIMES_SEMIRING_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_TIMES_SEMIRING_INT8", (void *)_cffi_var_GrB_MIN_TIMES_SEMIRING_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_TIMES_SEMIRING_UINT16", (void *)_cffi_var_GrB_MIN_TIMES_SEMIRING_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_TIMES_SEMIRING_UINT32", (void *)_cffi_var_GrB_MIN_TIMES_SEMIRING_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_TIMES_SEMIRING_UINT64", (void *)_cffi_var_GrB_MIN_TIMES_SEMIRING_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_TIMES_SEMIRING_UINT8", (void *)_cffi_var_GrB_MIN_TIMES_SEMIRING_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_MIN_UINT16", (void *)_cffi_var_GrB_MIN_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MIN_UINT32", (void *)_cffi_var_GrB_MIN_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MIN_UINT64", (void *)_cffi_var_GrB_MIN_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_MIN_UINT8", (void *)_cffi_var_GrB_MIN_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_Matrix_apply", (void *)_cffi_f_GrB_Matrix_apply, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1620), (void *)_cffi_d_GrB_Matrix_apply },
  { "GrB_Matrix_apply_BinaryOp1st_BOOL", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp1st_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1139), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp1st_BOOL },
  { "GrB_Matrix_apply_BinaryOp1st_FP32", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp1st_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1175), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp1st_FP32 },
  { "GrB_Matrix_apply_BinaryOp1st_FP64", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp1st_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1157), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp1st_FP64 },
  { "GrB_Matrix_apply_BinaryOp1st_INT16", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp1st_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1184), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp1st_INT16 },
  { "GrB_Matrix_apply_BinaryOp1st_INT32", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp1st_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1193), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp1st_INT32 },
  { "GrB_Matrix_apply_BinaryOp1st_INT64", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp1st_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1202), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp1st_INT64 },
  { "GrB_Matrix_apply_BinaryOp1st_INT8", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp1st_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1211), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp1st_INT8 },
  { "GrB_Matrix_apply_BinaryOp1st_Scalar", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp1st_Scalar, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1375), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp1st_Scalar },
  { "GrB_Matrix_apply_BinaryOp1st_UDT", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp1st_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1420), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp1st_UDT },
  { "GrB_Matrix_apply_BinaryOp1st_UINT16", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp1st_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1384), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp1st_UINT16 },
  { "GrB_Matrix_apply_BinaryOp1st_UINT32", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp1st_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1393), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp1st_UINT32 },
  { "GrB_Matrix_apply_BinaryOp1st_UINT64", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp1st_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1402), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp1st_UINT64 },
  { "GrB_Matrix_apply_BinaryOp1st_UINT8", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp1st_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1411), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp1st_UINT8 },
  { "GrB_Matrix_apply_BinaryOp2nd_BOOL", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp2nd_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1220), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp2nd_BOOL },
  { "GrB_Matrix_apply_BinaryOp2nd_FP32", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp2nd_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1256), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp2nd_FP32 },
  { "GrB_Matrix_apply_BinaryOp2nd_FP64", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp2nd_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1238), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp2nd_FP64 },
  { "GrB_Matrix_apply_BinaryOp2nd_INT16", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp2nd_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1265), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp2nd_INT16 },
  { "GrB_Matrix_apply_BinaryOp2nd_INT32", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp2nd_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1274), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp2nd_INT32 },
  { "GrB_Matrix_apply_BinaryOp2nd_INT64", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp2nd_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1283), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp2nd_INT64 },
  { "GrB_Matrix_apply_BinaryOp2nd_INT8", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp2nd_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1292), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp2nd_INT8 },
  { "GrB_Matrix_apply_BinaryOp2nd_Scalar", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp2nd_Scalar, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1310), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp2nd_Scalar },
  { "GrB_Matrix_apply_BinaryOp2nd_UDT", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp2nd_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1366), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp2nd_UDT },
  { "GrB_Matrix_apply_BinaryOp2nd_UINT16", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp2nd_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1330), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp2nd_UINT16 },
  { "GrB_Matrix_apply_BinaryOp2nd_UINT32", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp2nd_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1339), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp2nd_UINT32 },
  { "GrB_Matrix_apply_BinaryOp2nd_UINT64", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp2nd_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1348), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp2nd_UINT64 },
  { "GrB_Matrix_apply_BinaryOp2nd_UINT8", (void *)_cffi_f_GrB_Matrix_apply_BinaryOp2nd_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1357), (void *)_cffi_d_GrB_Matrix_apply_BinaryOp2nd_UINT8 },
  { "GrB_Matrix_apply_IndexOp_BOOL", (void *)_cffi_f_GrB_Matrix_apply_IndexOp_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1429), (void *)_cffi_d_GrB_Matrix_apply_IndexOp_BOOL },
  { "GrB_Matrix_apply_IndexOp_FP32", (void *)_cffi_f_GrB_Matrix_apply_IndexOp_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1465), (void *)_cffi_d_GrB_Matrix_apply_IndexOp_FP32 },
  { "GrB_Matrix_apply_IndexOp_FP64", (void *)_cffi_f_GrB_Matrix_apply_IndexOp_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1447), (void *)_cffi_d_GrB_Matrix_apply_IndexOp_FP64 },
  { "GrB_Matrix_apply_IndexOp_INT16", (void *)_cffi_f_GrB_Matrix_apply_IndexOp_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1474), (void *)_cffi_d_GrB_Matrix_apply_IndexOp_INT16 },
  { "GrB_Matrix_apply_IndexOp_INT32", (void *)_cffi_f_GrB_Matrix_apply_IndexOp_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1483), (void *)_cffi_d_GrB_Matrix_apply_IndexOp_INT32 },
  { "GrB_Matrix_apply_IndexOp_INT64", (void *)_cffi_f_GrB_Matrix_apply_IndexOp_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1492), (void *)_cffi_d_GrB_Matrix_apply_IndexOp_INT64 },
  { "GrB_Matrix_apply_IndexOp_INT8", (void *)_cffi_f_GrB_Matrix_apply_IndexOp_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1501), (void *)_cffi_d_GrB_Matrix_apply_IndexOp_INT8 },
  { "GrB_Matrix_apply_IndexOp_Scalar", (void *)_cffi_f_GrB_Matrix_apply_IndexOp_Scalar, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1510), (void *)_cffi_d_GrB_Matrix_apply_IndexOp_Scalar },
  { "GrB_Matrix_apply_IndexOp_UDT", (void *)_cffi_f_GrB_Matrix_apply_IndexOp_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1555), (void *)_cffi_d_GrB_Matrix_apply_IndexOp_UDT },
  { "GrB_Matrix_apply_IndexOp_UINT16", (void *)_cffi_f_GrB_Matrix_apply_IndexOp_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1519), (void *)_cffi_d_GrB_Matrix_apply_IndexOp_UINT16 },
  { "GrB_Matrix_apply_IndexOp_UINT32", (void *)_cffi_f_GrB_Matrix_apply_IndexOp_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1528), (void *)_cffi_d_GrB_Matrix_apply_IndexOp_UINT32 },
  { "GrB_Matrix_apply_IndexOp_UINT64", (void *)_cffi_f_GrB_Matrix_apply_IndexOp_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1537), (void *)_cffi_d_GrB_Matrix_apply_IndexOp_UINT64 },
  { "GrB_Matrix_apply_IndexOp_UINT8", (void *)_cffi_f_GrB_Matrix_apply_IndexOp_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1546), (void *)_cffi_d_GrB_Matrix_apply_IndexOp_UINT8 },
  { "GrB_Matrix_assign", (void *)_cffi_f_GrB_Matrix_assign, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1571), (void *)_cffi_d_GrB_Matrix_assign },
  { "GrB_Matrix_assign_BOOL", (void *)_cffi_f_GrB_Matrix_assign_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1040), (void *)_cffi_d_GrB_Matrix_assign_BOOL },
  { "GrB_Matrix_assign_FP32", (void *)_cffi_f_GrB_Matrix_assign_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1084), (void *)_cffi_d_GrB_Matrix_assign_FP32 },
  { "GrB_Matrix_assign_FP64", (void *)_cffi_f_GrB_Matrix_assign_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1062), (void *)_cffi_d_GrB_Matrix_assign_FP64 },
  { "GrB_Matrix_assign_INT16", (void *)_cffi_f_GrB_Matrix_assign_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1095), (void *)_cffi_d_GrB_Matrix_assign_INT16 },
  { "GrB_Matrix_assign_INT32", (void *)_cffi_f_GrB_Matrix_assign_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1106), (void *)_cffi_d_GrB_Matrix_assign_INT32 },
  { "GrB_Matrix_assign_INT64", (void *)_cffi_f_GrB_Matrix_assign_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1117), (void *)_cffi_d_GrB_Matrix_assign_INT64 },
  { "GrB_Matrix_assign_INT8", (void *)_cffi_f_GrB_Matrix_assign_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1128), (void *)_cffi_d_GrB_Matrix_assign_INT8 },
  { "GrB_Matrix_assign_Scalar", (void *)_cffi_f_GrB_Matrix_assign_Scalar, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1591), (void *)_cffi_d_GrB_Matrix_assign_Scalar },
  { "GrB_Matrix_assign_UDT", (void *)_cffi_f_GrB_Matrix_assign_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1672), (void *)_cffi_d_GrB_Matrix_assign_UDT },
  { "GrB_Matrix_assign_UINT16", (void *)_cffi_f_GrB_Matrix_assign_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1628), (void *)_cffi_d_GrB_Matrix_assign_UINT16 },
  { "GrB_Matrix_assign_UINT32", (void *)_cffi_f_GrB_Matrix_assign_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1639), (void *)_cffi_d_GrB_Matrix_assign_UINT32 },
  { "GrB_Matrix_assign_UINT64", (void *)_cffi_f_GrB_Matrix_assign_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1650), (void *)_cffi_d_GrB_Matrix_assign_UINT64 },
  { "GrB_Matrix_assign_UINT8", (void *)_cffi_f_GrB_Matrix_assign_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1661), (void *)_cffi_d_GrB_Matrix_assign_UINT8 },
  { "GrB_Matrix_build_BOOL", (void *)_cffi_f_GrB_Matrix_build_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1781), (void *)_cffi_d_GrB_Matrix_build_BOOL },
  { "GrB_Matrix_build_FP32", (void *)_cffi_f_GrB_Matrix_build_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1813), (void *)_cffi_d_GrB_Matrix_build_FP32 },
  { "GrB_Matrix_build_FP64", (void *)_cffi_f_GrB_Matrix_build_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1797), (void *)_cffi_d_GrB_Matrix_build_FP64 },
  { "GrB_Matrix_build_INT16", (void *)_cffi_f_GrB_Matrix_build_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1821), (void *)_cffi_d_GrB_Matrix_build_INT16 },
  { "GrB_Matrix_build_INT32", (void *)_cffi_f_GrB_Matrix_build_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1829), (void *)_cffi_d_GrB_Matrix_build_INT32 },
  { "GrB_Matrix_build_INT64", (void *)_cffi_f_GrB_Matrix_build_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1837), (void *)_cffi_d_GrB_Matrix_build_INT64 },
  { "GrB_Matrix_build_INT8", (void *)_cffi_f_GrB_Matrix_build_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1845), (void *)_cffi_d_GrB_Matrix_build_INT8 },
  { "GrB_Matrix_build_UDT", (void *)_cffi_f_GrB_Matrix_build_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1892), (void *)_cffi_d_GrB_Matrix_build_UDT },
  { "GrB_Matrix_build_UINT16", (void *)_cffi_f_GrB_Matrix_build_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1860), (void *)_cffi_d_GrB_Matrix_build_UINT16 },
  { "GrB_Matrix_build_UINT32", (void *)_cffi_f_GrB_Matrix_build_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1868), (void *)_cffi_d_GrB_Matrix_build_UINT32 },
  { "GrB_Matrix_build_UINT64", (void *)_cffi_f_GrB_Matrix_build_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1876), (void *)_cffi_d_GrB_Matrix_build_UINT64 },
  { "GrB_Matrix_build_UINT8", (void *)_cffi_f_GrB_Matrix_build_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1884), (void *)_cffi_d_GrB_Matrix_build_UINT8 },
  { "GrB_Matrix_clear", (void *)_cffi_f_GrB_Matrix_clear, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 910), (void *)_cffi_d_GrB_Matrix_clear },
  { "GrB_Matrix_deserialize", (void *)_cffi_f_GrB_Matrix_deserialize, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 883), (void *)_cffi_d_GrB_Matrix_deserialize },
  { "GrB_Matrix_diag", (void *)_cffi_f_GrB_Matrix_diag, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 896), (void *)_cffi_d_GrB_Matrix_diag },
  { "GrB_Matrix_dup", (void *)_cffi_f_GrB_Matrix_dup, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 571), (void *)_cffi_d_GrB_Matrix_dup },
  { "GrB_Matrix_eWiseAdd_BinaryOp", (void *)_cffi_f_GrB_Matrix_eWiseAdd_BinaryOp, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1301), (void *)_cffi_d_GrB_Matrix_eWiseAdd_BinaryOp },
  { "GrB_Matrix_eWiseAdd_Monoid", (void *)_cffi_f_GrB_Matrix_eWiseAdd_Monoid, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1582), (void *)_cffi_d_GrB_Matrix_eWiseAdd_Monoid },
  { "GrB_Matrix_eWiseAdd_Semiring", (void *)_cffi_f_GrB_Matrix_eWiseAdd_Semiring, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1611), (void *)_cffi_d_GrB_Matrix_eWiseAdd_Semiring },
  { "GrB_Matrix_eWiseMult_BinaryOp", (void *)_cffi_f_GrB_Matrix_eWiseMult_BinaryOp, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1301), (void *)_cffi_d_GrB_Matrix_eWiseMult_BinaryOp },
  { "GrB_Matrix_eWiseMult_Monoid", (void *)_cffi_f_GrB_Matrix_eWiseMult_Monoid, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1582), (void *)_cffi_d_GrB_Matrix_eWiseMult_Monoid },
  { "GrB_Matrix_eWiseMult_Semiring", (void *)_cffi_f_GrB_Matrix_eWiseMult_Semiring, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1611), (void *)_cffi_d_GrB_Matrix_eWiseMult_Semiring },
  { "GrB_Matrix_error", (void *)_cffi_f_GrB_Matrix_error, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 141), (void *)_cffi_d_GrB_Matrix_error },
  { "GrB_Matrix_exportHint", (void *)_cffi_f_GrB_Matrix_exportHint, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 5), (void *)_cffi_d_GrB_Matrix_exportHint },
  { "GrB_Matrix_exportSize", (void *)_cffi_f_GrB_Matrix_exportSize, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3733), (void *)_cffi_d_GrB_Matrix_exportSize },
  { "GrB_Matrix_export_BOOL", (void *)_cffi_f_GrB_Matrix_export_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3553), (void *)_cffi_d_GrB_Matrix_export_BOOL },
  { "GrB_Matrix_export_FP32", (void *)_cffi_f_GrB_Matrix_export_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3604), (void *)_cffi_d_GrB_Matrix_export_FP32 },
  { "GrB_Matrix_export_FP64", (void *)_cffi_f_GrB_Matrix_export_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3570), (void *)_cffi_d_GrB_Matrix_export_FP64 },
  { "GrB_Matrix_export_INT16", (void *)_cffi_f_GrB_Matrix_export_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3638), (void *)_cffi_d_GrB_Matrix_export_INT16 },
  { "GrB_Matrix_export_INT32", (void *)_cffi_f_GrB_Matrix_export_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3655), (void *)_cffi_d_GrB_Matrix_export_INT32 },
  { "GrB_Matrix_export_INT64", (void *)_cffi_f_GrB_Matrix_export_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3672), (void *)_cffi_d_GrB_Matrix_export_INT64 },
  { "GrB_Matrix_export_INT8", (void *)_cffi_f_GrB_Matrix_export_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3689), (void *)_cffi_d_GrB_Matrix_export_INT8 },
  { "GrB_Matrix_export_UDT", (void *)_cffi_f_GrB_Matrix_export_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3787), (void *)_cffi_d_GrB_Matrix_export_UDT },
  { "GrB_Matrix_export_UINT16", (void *)_cffi_f_GrB_Matrix_export_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3706), (void *)_cffi_d_GrB_Matrix_export_UINT16 },
  { "GrB_Matrix_export_UINT32", (void *)_cffi_f_GrB_Matrix_export_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3723), (void *)_cffi_d_GrB_Matrix_export_UINT32 },
  { "GrB_Matrix_export_UINT64", (void *)_cffi_f_GrB_Matrix_export_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3753), (void *)_cffi_d_GrB_Matrix_export_UINT64 },
  { "GrB_Matrix_export_UINT8", (void *)_cffi_f_GrB_Matrix_export_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3770), (void *)_cffi_d_GrB_Matrix_export_UINT8 },
  { "GrB_Matrix_extract", (void *)_cffi_f_GrB_Matrix_extract, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1571), (void *)_cffi_d_GrB_Matrix_extract },
  { "GrB_Matrix_extractElement_BOOL", (void *)_cffi_f_GrB_Matrix_extractElement_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 72), (void *)_cffi_d_GrB_Matrix_extractElement_BOOL },
  { "GrB_Matrix_extractElement_FP32", (void *)_cffi_f_GrB_Matrix_extractElement_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 245), (void *)_cffi_d_GrB_Matrix_extractElement_FP32 },
  { "GrB_Matrix_extractElement_FP64", (void *)_cffi_f_GrB_Matrix_extractElement_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 187), (void *)_cffi_d_GrB_Matrix_extractElement_FP64 },
  { "GrB_Matrix_extractElement_INT16", (void *)_cffi_f_GrB_Matrix_extractElement_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 303), (void *)_cffi_d_GrB_Matrix_extractElement_INT16 },
  { "GrB_Matrix_extractElement_INT32", (void *)_cffi_f_GrB_Matrix_extractElement_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 332), (void *)_cffi_d_GrB_Matrix_extractElement_INT32 },
  { "GrB_Matrix_extractElement_INT64", (void *)_cffi_f_GrB_Matrix_extractElement_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 361), (void *)_cffi_d_GrB_Matrix_extractElement_INT64 },
  { "GrB_Matrix_extractElement_INT8", (void *)_cffi_f_GrB_Matrix_extractElement_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 390), (void *)_cffi_d_GrB_Matrix_extractElement_INT8 },
  { "GrB_Matrix_extractElement_Scalar", (void *)_cffi_f_GrB_Matrix_extractElement_Scalar, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2196), (void *)_cffi_d_GrB_Matrix_extractElement_Scalar },
  { "GrB_Matrix_extractElement_UDT", (void *)_cffi_f_GrB_Matrix_extractElement_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3868), (void *)_cffi_d_GrB_Matrix_extractElement_UDT },
  { "GrB_Matrix_extractElement_UINT16", (void *)_cffi_f_GrB_Matrix_extractElement_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3399), (void *)_cffi_d_GrB_Matrix_extractElement_UINT16 },
  { "GrB_Matrix_extractElement_UINT32", (void *)_cffi_f_GrB_Matrix_extractElement_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3428), (void *)_cffi_d_GrB_Matrix_extractElement_UINT32 },
  { "GrB_Matrix_extractElement_UINT64", (void *)_cffi_f_GrB_Matrix_extractElement_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3515), (void *)_cffi_d_GrB_Matrix_extractElement_UINT64 },
  { "GrB_Matrix_extractElement_UINT8", (void *)_cffi_f_GrB_Matrix_extractElement_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3823), (void *)_cffi_d_GrB_Matrix_extractElement_UINT8 },
  { "GrB_Matrix_extractTuples_BOOL", (void *)_cffi_f_GrB_Matrix_extractTuples_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3546), (void *)_cffi_d_GrB_Matrix_extractTuples_BOOL },
  { "GrB_Matrix_extractTuples_FP32", (void *)_cffi_f_GrB_Matrix_extractTuples_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3597), (void *)_cffi_d_GrB_Matrix_extractTuples_FP32 },
  { "GrB_Matrix_extractTuples_FP64", (void *)_cffi_f_GrB_Matrix_extractTuples_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3563), (void *)_cffi_d_GrB_Matrix_extractTuples_FP64 },
  { "GrB_Matrix_extractTuples_INT16", (void *)_cffi_f_GrB_Matrix_extractTuples_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3631), (void *)_cffi_d_GrB_Matrix_extractTuples_INT16 },
  { "GrB_Matrix_extractTuples_INT32", (void *)_cffi_f_GrB_Matrix_extractTuples_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3648), (void *)_cffi_d_GrB_Matrix_extractTuples_INT32 },
  { "GrB_Matrix_extractTuples_INT64", (void *)_cffi_f_GrB_Matrix_extractTuples_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3665), (void *)_cffi_d_GrB_Matrix_extractTuples_INT64 },
  { "GrB_Matrix_extractTuples_INT8", (void *)_cffi_f_GrB_Matrix_extractTuples_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3682), (void *)_cffi_d_GrB_Matrix_extractTuples_INT8 },
  { "GrB_Matrix_extractTuples_UDT", (void *)_cffi_f_GrB_Matrix_extractTuples_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3780), (void *)_cffi_d_GrB_Matrix_extractTuples_UDT },
  { "GrB_Matrix_extractTuples_UINT16", (void *)_cffi_f_GrB_Matrix_extractTuples_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3699), (void *)_cffi_d_GrB_Matrix_extractTuples_UINT16 },
  { "GrB_Matrix_extractTuples_UINT32", (void *)_cffi_f_GrB_Matrix_extractTuples_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3716), (void *)_cffi_d_GrB_Matrix_extractTuples_UINT32 },
  { "GrB_Matrix_extractTuples_UINT64", (void *)_cffi_f_GrB_Matrix_extractTuples_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3746), (void *)_cffi_d_GrB_Matrix_extractTuples_UINT64 },
  { "GrB_Matrix_extractTuples_UINT8", (void *)_cffi_f_GrB_Matrix_extractTuples_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3763), (void *)_cffi_d_GrB_Matrix_extractTuples_UINT8 },
  { "GrB_Matrix_free", (void *)_cffi_f_GrB_Matrix_free, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 568), (void *)_cffi_d_GrB_Matrix_free },
  { "GrB_Matrix_import_BOOL", (void *)_cffi_f_GrB_Matrix_import_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 691), (void *)_cffi_d_GrB_Matrix_import_BOOL },
  { "GrB_Matrix_import_FP32", (void *)_cffi_f_GrB_Matrix_import_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 743), (void *)_cffi_d_GrB_Matrix_import_FP32 },
  { "GrB_Matrix_import_FP64", (void *)_cffi_f_GrB_Matrix_import_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 717), (void *)_cffi_d_GrB_Matrix_import_FP64 },
  { "GrB_Matrix_import_INT16", (void *)_cffi_f_GrB_Matrix_import_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 756), (void *)_cffi_d_GrB_Matrix_import_INT16 },
  { "GrB_Matrix_import_INT32", (void *)_cffi_f_GrB_Matrix_import_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 769), (void *)_cffi_d_GrB_Matrix_import_INT32 },
  { "GrB_Matrix_import_INT64", (void *)_cffi_f_GrB_Matrix_import_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 782), (void *)_cffi_d_GrB_Matrix_import_INT64 },
  { "GrB_Matrix_import_INT8", (void *)_cffi_f_GrB_Matrix_import_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 795), (void *)_cffi_d_GrB_Matrix_import_INT8 },
  { "GrB_Matrix_import_UDT", (void *)_cffi_f_GrB_Matrix_import_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 860), (void *)_cffi_d_GrB_Matrix_import_UDT },
  { "GrB_Matrix_import_UINT16", (void *)_cffi_f_GrB_Matrix_import_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 808), (void *)_cffi_d_GrB_Matrix_import_UINT16 },
  { "GrB_Matrix_import_UINT32", (void *)_cffi_f_GrB_Matrix_import_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 821), (void *)_cffi_d_GrB_Matrix_import_UINT32 },
  { "GrB_Matrix_import_UINT64", (void *)_cffi_f_GrB_Matrix_import_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 834), (void *)_cffi_d_GrB_Matrix_import_UINT64 },
  { "GrB_Matrix_import_UINT8", (void *)_cffi_f_GrB_Matrix_import_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 847), (void *)_cffi_d_GrB_Matrix_import_UINT8 },
  { "GrB_Matrix_kronecker_BinaryOp", (void *)_cffi_f_GrB_Matrix_kronecker_BinaryOp, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1301), (void *)_cffi_d_GrB_Matrix_kronecker_BinaryOp },
  { "GrB_Matrix_kronecker_Monoid", (void *)_cffi_f_GrB_Matrix_kronecker_Monoid, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1582), (void *)_cffi_d_GrB_Matrix_kronecker_Monoid },
  { "GrB_Matrix_kronecker_Semiring", (void *)_cffi_f_GrB_Matrix_kronecker_Semiring, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1611), (void *)_cffi_d_GrB_Matrix_kronecker_Semiring },
  { "GrB_Matrix_ncols", (void *)_cffi_f_GrB_Matrix_ncols, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3511), (void *)_cffi_d_GrB_Matrix_ncols },
  { "GrB_Matrix_new", (void *)_cffi_f_GrB_Matrix_new, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 639), (void *)_cffi_d_GrB_Matrix_new },
  { "GrB_Matrix_nrows", (void *)_cffi_f_GrB_Matrix_nrows, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3511), (void *)_cffi_d_GrB_Matrix_nrows },
  { "GrB_Matrix_nvals", (void *)_cffi_f_GrB_Matrix_nvals, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3511), (void *)_cffi_d_GrB_Matrix_nvals },
  { "GrB_Matrix_reduce_BOOL", (void *)_cffi_f_GrB_Matrix_reduce_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 54), (void *)_cffi_d_GrB_Matrix_reduce_BOOL },
  { "GrB_Matrix_reduce_BinaryOp", (void *)_cffi_f_GrB_Matrix_reduce_BinaryOp, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2723), (void *)_cffi_d_GrB_Matrix_reduce_BinaryOp },
  { "GrB_Matrix_reduce_BinaryOp_Scalar", (void *)_cffi_f_GrB_Matrix_reduce_BinaryOp_Scalar, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2168), (void *)_cffi_d_GrB_Matrix_reduce_BinaryOp_Scalar },
  { "GrB_Matrix_reduce_FP32", (void *)_cffi_f_GrB_Matrix_reduce_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 231), (void *)_cffi_d_GrB_Matrix_reduce_FP32 },
  { "GrB_Matrix_reduce_FP64", (void *)_cffi_f_GrB_Matrix_reduce_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 173), (void *)_cffi_d_GrB_Matrix_reduce_FP64 },
  { "GrB_Matrix_reduce_INT16", (void *)_cffi_f_GrB_Matrix_reduce_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 289), (void *)_cffi_d_GrB_Matrix_reduce_INT16 },
  { "GrB_Matrix_reduce_INT32", (void *)_cffi_f_GrB_Matrix_reduce_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 318), (void *)_cffi_d_GrB_Matrix_reduce_INT32 },
  { "GrB_Matrix_reduce_INT64", (void *)_cffi_f_GrB_Matrix_reduce_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 347), (void *)_cffi_d_GrB_Matrix_reduce_INT64 },
  { "GrB_Matrix_reduce_INT8", (void *)_cffi_f_GrB_Matrix_reduce_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 376), (void *)_cffi_d_GrB_Matrix_reduce_INT8 },
  { "GrB_Matrix_reduce_Monoid", (void *)_cffi_f_GrB_Matrix_reduce_Monoid, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3085), (void *)_cffi_d_GrB_Matrix_reduce_Monoid },
  { "GrB_Matrix_reduce_Monoid_Scalar", (void *)_cffi_f_GrB_Matrix_reduce_Monoid_Scalar, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2182), (void *)_cffi_d_GrB_Matrix_reduce_Monoid_Scalar },
  { "GrB_Matrix_reduce_UDT", (void *)_cffi_f_GrB_Matrix_reduce_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3854), (void *)_cffi_d_GrB_Matrix_reduce_UDT },
  { "GrB_Matrix_reduce_UINT16", (void *)_cffi_f_GrB_Matrix_reduce_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3385), (void *)_cffi_d_GrB_Matrix_reduce_UINT16 },
  { "GrB_Matrix_reduce_UINT32", (void *)_cffi_f_GrB_Matrix_reduce_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3414), (void *)_cffi_d_GrB_Matrix_reduce_UINT32 },
  { "GrB_Matrix_reduce_UINT64", (void *)_cffi_f_GrB_Matrix_reduce_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3497), (void *)_cffi_d_GrB_Matrix_reduce_UINT64 },
  { "GrB_Matrix_reduce_UINT8", (void *)_cffi_f_GrB_Matrix_reduce_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3809), (void *)_cffi_d_GrB_Matrix_reduce_UINT8 },
  { "GrB_Matrix_removeElement", (void *)_cffi_f_GrB_Matrix_removeElement, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1900), (void *)_cffi_d_GrB_Matrix_removeElement },
  { "GrB_Matrix_resize", (void *)_cffi_f_GrB_Matrix_resize, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1900), (void *)_cffi_d_GrB_Matrix_resize },
  { "GrB_Matrix_select_BOOL", (void *)_cffi_f_GrB_Matrix_select_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1429), (void *)_cffi_d_GrB_Matrix_select_BOOL },
  { "GrB_Matrix_select_FP32", (void *)_cffi_f_GrB_Matrix_select_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1465), (void *)_cffi_d_GrB_Matrix_select_FP32 },
  { "GrB_Matrix_select_FP64", (void *)_cffi_f_GrB_Matrix_select_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1447), (void *)_cffi_d_GrB_Matrix_select_FP64 },
  { "GrB_Matrix_select_INT16", (void *)_cffi_f_GrB_Matrix_select_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1474), (void *)_cffi_d_GrB_Matrix_select_INT16 },
  { "GrB_Matrix_select_INT32", (void *)_cffi_f_GrB_Matrix_select_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1483), (void *)_cffi_d_GrB_Matrix_select_INT32 },
  { "GrB_Matrix_select_INT64", (void *)_cffi_f_GrB_Matrix_select_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1492), (void *)_cffi_d_GrB_Matrix_select_INT64 },
  { "GrB_Matrix_select_INT8", (void *)_cffi_f_GrB_Matrix_select_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1501), (void *)_cffi_d_GrB_Matrix_select_INT8 },
  { "GrB_Matrix_select_Scalar", (void *)_cffi_f_GrB_Matrix_select_Scalar, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1510), (void *)_cffi_d_GrB_Matrix_select_Scalar },
  { "GrB_Matrix_select_UDT", (void *)_cffi_f_GrB_Matrix_select_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1555), (void *)_cffi_d_GrB_Matrix_select_UDT },
  { "GrB_Matrix_select_UINT16", (void *)_cffi_f_GrB_Matrix_select_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1519), (void *)_cffi_d_GrB_Matrix_select_UINT16 },
  { "GrB_Matrix_select_UINT32", (void *)_cffi_f_GrB_Matrix_select_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1528), (void *)_cffi_d_GrB_Matrix_select_UINT32 },
  { "GrB_Matrix_select_UINT64", (void *)_cffi_f_GrB_Matrix_select_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1537), (void *)_cffi_d_GrB_Matrix_select_UINT64 },
  { "GrB_Matrix_select_UINT8", (void *)_cffi_f_GrB_Matrix_select_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1546), (void *)_cffi_d_GrB_Matrix_select_UINT8 },
  { "GrB_Matrix_serialize", (void *)_cffi_f_GrB_Matrix_serialize, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3887), (void *)_cffi_d_GrB_Matrix_serialize },
  { "GrB_Matrix_serializeSize", (void *)_cffi_f_GrB_Matrix_serializeSize, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3511), (void *)_cffi_d_GrB_Matrix_serializeSize },
  { "GrB_Matrix_setElement_BOOL", (void *)_cffi_f_GrB_Matrix_setElement_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 941), (void *)_cffi_d_GrB_Matrix_setElement_BOOL },
  { "GrB_Matrix_setElement_FP32", (void *)_cffi_f_GrB_Matrix_setElement_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 978), (void *)_cffi_d_GrB_Matrix_setElement_FP32 },
  { "GrB_Matrix_setElement_FP64", (void *)_cffi_f_GrB_Matrix_setElement_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 966), (void *)_cffi_d_GrB_Matrix_setElement_FP64 },
  { "GrB_Matrix_setElement_INT16", (void *)_cffi_f_GrB_Matrix_setElement_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 984), (void *)_cffi_d_GrB_Matrix_setElement_INT16 },
  { "GrB_Matrix_setElement_INT32", (void *)_cffi_f_GrB_Matrix_setElement_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 990), (void *)_cffi_d_GrB_Matrix_setElement_INT32 },
  { "GrB_Matrix_setElement_INT64", (void *)_cffi_f_GrB_Matrix_setElement_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 996), (void *)_cffi_d_GrB_Matrix_setElement_INT64 },
  { "GrB_Matrix_setElement_INT8", (void *)_cffi_f_GrB_Matrix_setElement_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1022), (void *)_cffi_d_GrB_Matrix_setElement_INT8 },
  { "GrB_Matrix_setElement_Scalar", (void *)_cffi_f_GrB_Matrix_setElement_Scalar, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1683), (void *)_cffi_d_GrB_Matrix_setElement_Scalar },
  { "GrB_Matrix_setElement_UDT", (void *)_cffi_f_GrB_Matrix_setElement_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1931), (void *)_cffi_d_GrB_Matrix_setElement_UDT },
  { "GrB_Matrix_setElement_UINT16", (void *)_cffi_f_GrB_Matrix_setElement_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1715), (void *)_cffi_d_GrB_Matrix_setElement_UINT16 },
  { "GrB_Matrix_setElement_UINT32", (void *)_cffi_f_GrB_Matrix_setElement_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1721), (void *)_cffi_d_GrB_Matrix_setElement_UINT32 },
  { "GrB_Matrix_setElement_UINT64", (void *)_cffi_f_GrB_Matrix_setElement_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1905), (void *)_cffi_d_GrB_Matrix_setElement_UINT64 },
  { "GrB_Matrix_setElement_UINT8", (void *)_cffi_f_GrB_Matrix_setElement_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1911), (void *)_cffi_d_GrB_Matrix_setElement_UINT8 },
  { "GrB_Matrix_wait", (void *)_cffi_f_GrB_Matrix_wait, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 913), (void *)_cffi_d_GrB_Matrix_wait },
  { "GrB_Monoid_error", (void *)_cffi_f_GrB_Monoid_error, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 145), (void *)_cffi_d_GrB_Monoid_error },
  { "GrB_Monoid_free", (void *)_cffi_f_GrB_Monoid_free, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 1937), (void *)_cffi_d_GrB_Monoid_free },
  { "GrB_Monoid_new_BOOL", (void *)_cffi_f_GrB_Monoid_new_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1940), (void *)_cffi_d_GrB_Monoid_new_BOOL },
  { "GrB_Monoid_new_FP32", (void *)_cffi_f_GrB_Monoid_new_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1984), (void *)_cffi_d_GrB_Monoid_new_FP32 },
  { "GrB_Monoid_new_FP64", (void *)_cffi_f_GrB_Monoid_new_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1962), (void *)_cffi_d_GrB_Monoid_new_FP64 },
  { "GrB_Monoid_new_INT16", (void *)_cffi_f_GrB_Monoid_new_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1995), (void *)_cffi_d_GrB_Monoid_new_INT16 },
  { "GrB_Monoid_new_INT32", (void *)_cffi_f_GrB_Monoid_new_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2006), (void *)_cffi_d_GrB_Monoid_new_INT32 },
  { "GrB_Monoid_new_INT64", (void *)_cffi_f_GrB_Monoid_new_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2017), (void *)_cffi_d_GrB_Monoid_new_INT64 },
  { "GrB_Monoid_new_INT8", (void *)_cffi_f_GrB_Monoid_new_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2028), (void *)_cffi_d_GrB_Monoid_new_INT8 },
  { "GrB_Monoid_new_UDT", (void *)_cffi_f_GrB_Monoid_new_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2083), (void *)_cffi_d_GrB_Monoid_new_UDT },
  { "GrB_Monoid_new_UINT16", (void *)_cffi_f_GrB_Monoid_new_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2039), (void *)_cffi_d_GrB_Monoid_new_UINT16 },
  { "GrB_Monoid_new_UINT32", (void *)_cffi_f_GrB_Monoid_new_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2050), (void *)_cffi_d_GrB_Monoid_new_UINT32 },
  { "GrB_Monoid_new_UINT64", (void *)_cffi_f_GrB_Monoid_new_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2061), (void *)_cffi_d_GrB_Monoid_new_UINT64 },
  { "GrB_Monoid_new_UINT8", (void *)_cffi_f_GrB_Monoid_new_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2072), (void *)_cffi_d_GrB_Monoid_new_UINT8 },
  { "GrB_Monoid_wait", (void *)_cffi_f_GrB_Monoid_wait, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2098), (void *)_cffi_d_GrB_Monoid_wait },
  { "GrB_NE_BOOL", (void *)_cffi_var_GrB_NE_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_NE_FP32", (void *)_cffi_var_GrB_NE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_NE_FP64", (void *)_cffi_var_GrB_NE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_NE_INT16", (void *)_cffi_var_GrB_NE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_NE_INT32", (void *)_cffi_var_GrB_NE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_NE_INT64", (void *)_cffi_var_GrB_NE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_NE_INT8", (void *)_cffi_var_GrB_NE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_NE_UINT16", (void *)_cffi_var_GrB_NE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_NE_UINT32", (void *)_cffi_var_GrB_NE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_NE_UINT64", (void *)_cffi_var_GrB_NE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_NE_UINT8", (void *)_cffi_var_GrB_NE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_NONBLOCKING", (void *)_cffi_const_GrB_NONBLOCKING, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_NOT_IMPLEMENTED", (void *)_cffi_const_GrB_NOT_IMPLEMENTED, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_NO_VALUE", (void *)_cffi_const_GrB_NO_VALUE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_NULL_POINTER", (void *)_cffi_const_GrB_NULL_POINTER, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_OFFDIAG", (void *)_cffi_var_GrB_OFFDIAG, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_ONEB_BOOL", (void *)_cffi_var_GrB_ONEB_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_ONEB_FP32", (void *)_cffi_var_GrB_ONEB_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_ONEB_FP64", (void *)_cffi_var_GrB_ONEB_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_ONEB_INT16", (void *)_cffi_var_GrB_ONEB_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_ONEB_INT32", (void *)_cffi_var_GrB_ONEB_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_ONEB_INT64", (void *)_cffi_var_GrB_ONEB_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_ONEB_INT8", (void *)_cffi_var_GrB_ONEB_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_ONEB_UINT16", (void *)_cffi_var_GrB_ONEB_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_ONEB_UINT32", (void *)_cffi_var_GrB_ONEB_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_ONEB_UINT64", (void *)_cffi_var_GrB_ONEB_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_ONEB_UINT8", (void *)_cffi_var_GrB_ONEB_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_OUTP", (void *)_cffi_const_GrB_OUTP, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_OUTPUT_NOT_EMPTY", (void *)_cffi_const_GrB_OUTPUT_NOT_EMPTY, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_OUT_OF_MEMORY", (void *)_cffi_const_GrB_OUT_OF_MEMORY, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_PANIC", (void *)_cffi_const_GrB_PANIC, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_PLUS_BOOL", (void *)_cffi_var_GrB_PLUS_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_PLUS_FP32", (void *)_cffi_var_GrB_PLUS_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_PLUS_FP64", (void *)_cffi_var_GrB_PLUS_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_PLUS_INT16", (void *)_cffi_var_GrB_PLUS_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_PLUS_INT32", (void *)_cffi_var_GrB_PLUS_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_PLUS_INT64", (void *)_cffi_var_GrB_PLUS_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_PLUS_INT8", (void *)_cffi_var_GrB_PLUS_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_PLUS_MIN_SEMIRING_FP32", (void *)_cffi_var_GrB_PLUS_MIN_SEMIRING_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_PLUS_MIN_SEMIRING_FP64", (void *)_cffi_var_GrB_PLUS_MIN_SEMIRING_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_PLUS_MIN_SEMIRING_INT16", (void *)_cffi_var_GrB_PLUS_MIN_SEMIRING_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_PLUS_MIN_SEMIRING_INT32", (void *)_cffi_var_GrB_PLUS_MIN_SEMIRING_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_PLUS_MIN_SEMIRING_INT64", (void *)_cffi_var_GrB_PLUS_MIN_SEMIRING_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_PLUS_MIN_SEMIRING_INT8", (void *)_cffi_var_GrB_PLUS_MIN_SEMIRING_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_PLUS_MIN_SEMIRING_UINT16", (void *)_cffi_var_GrB_PLUS_MIN_SEMIRING_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_PLUS_MIN_SEMIRING_UINT32", (void *)_cffi_var_GrB_PLUS_MIN_SEMIRING_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_PLUS_MIN_SEMIRING_UINT64", (void *)_cffi_var_GrB_PLUS_MIN_SEMIRING_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_PLUS_MIN_SEMIRING_UINT8", (void *)_cffi_var_GrB_PLUS_MIN_SEMIRING_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_PLUS_MONOID_FP32", (void *)_cffi_var_GrB_PLUS_MONOID_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_PLUS_MONOID_FP64", (void *)_cffi_var_GrB_PLUS_MONOID_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_PLUS_MONOID_INT16", (void *)_cffi_var_GrB_PLUS_MONOID_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_PLUS_MONOID_INT32", (void *)_cffi_var_GrB_PLUS_MONOID_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_PLUS_MONOID_INT64", (void *)_cffi_var_GrB_PLUS_MONOID_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_PLUS_MONOID_INT8", (void *)_cffi_var_GrB_PLUS_MONOID_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_PLUS_MONOID_UINT16", (void *)_cffi_var_GrB_PLUS_MONOID_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_PLUS_MONOID_UINT32", (void *)_cffi_var_GrB_PLUS_MONOID_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_PLUS_MONOID_UINT64", (void *)_cffi_var_GrB_PLUS_MONOID_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_PLUS_MONOID_UINT8", (void *)_cffi_var_GrB_PLUS_MONOID_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_PLUS_TIMES_SEMIRING_FP32", (void *)_cffi_var_GrB_PLUS_TIMES_SEMIRING_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_PLUS_TIMES_SEMIRING_FP64", (void *)_cffi_var_GrB_PLUS_TIMES_SEMIRING_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_PLUS_TIMES_SEMIRING_INT16", (void *)_cffi_var_GrB_PLUS_TIMES_SEMIRING_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_PLUS_TIMES_SEMIRING_INT32", (void *)_cffi_var_GrB_PLUS_TIMES_SEMIRING_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_PLUS_TIMES_SEMIRING_INT64", (void *)_cffi_var_GrB_PLUS_TIMES_SEMIRING_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_PLUS_TIMES_SEMIRING_INT8", (void *)_cffi_var_GrB_PLUS_TIMES_SEMIRING_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_PLUS_TIMES_SEMIRING_UINT16", (void *)_cffi_var_GrB_PLUS_TIMES_SEMIRING_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_PLUS_TIMES_SEMIRING_UINT32", (void *)_cffi_var_GrB_PLUS_TIMES_SEMIRING_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_PLUS_TIMES_SEMIRING_UINT64", (void *)_cffi_var_GrB_PLUS_TIMES_SEMIRING_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_PLUS_TIMES_SEMIRING_UINT8", (void *)_cffi_var_GrB_PLUS_TIMES_SEMIRING_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GrB_PLUS_UINT16", (void *)_cffi_var_GrB_PLUS_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_PLUS_UINT32", (void *)_cffi_var_GrB_PLUS_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_PLUS_UINT64", (void *)_cffi_var_GrB_PLUS_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_PLUS_UINT8", (void *)_cffi_var_GrB_PLUS_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_REPLACE", (void *)_cffi_const_GrB_REPLACE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_ROWGT", (void *)_cffi_var_GrB_ROWGT, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_ROWINDEX_INT32", (void *)_cffi_var_GrB_ROWINDEX_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_ROWINDEX_INT64", (void *)_cffi_var_GrB_ROWINDEX_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_ROWLE", (void *)_cffi_var_GrB_ROWLE, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_Row_assign", (void *)_cffi_f_GrB_Row_assign, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1705), (void *)_cffi_d_GrB_Row_assign },
  { "GrB_SECOND_BOOL", (void *)_cffi_var_GrB_SECOND_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_SECOND_FP32", (void *)_cffi_var_GrB_SECOND_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_SECOND_FP64", (void *)_cffi_var_GrB_SECOND_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_SECOND_INT16", (void *)_cffi_var_GrB_SECOND_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_SECOND_INT32", (void *)_cffi_var_GrB_SECOND_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_SECOND_INT64", (void *)_cffi_var_GrB_SECOND_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_SECOND_INT8", (void *)_cffi_var_GrB_SECOND_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_SECOND_UINT16", (void *)_cffi_var_GrB_SECOND_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_SECOND_UINT32", (void *)_cffi_var_GrB_SECOND_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_SECOND_UINT64", (void *)_cffi_var_GrB_SECOND_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_SECOND_UINT8", (void *)_cffi_var_GrB_SECOND_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_STRUCTURE", (void *)_cffi_const_GrB_STRUCTURE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_SUCCESS", (void *)_cffi_const_GrB_SUCCESS, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_Scalar_clear", (void *)_cffi_f_GrB_Scalar_clear, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 2119), (void *)_cffi_d_GrB_Scalar_clear },
  { "GrB_Scalar_dup", (void *)_cffi_f_GrB_Scalar_dup, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2111), (void *)_cffi_d_GrB_Scalar_dup },
  { "GrB_Scalar_error", (void *)_cffi_f_GrB_Scalar_error, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 149), (void *)_cffi_d_GrB_Scalar_error },
  { "GrB_Scalar_extractElement_BOOL", (void *)_cffi_f_GrB_Scalar_extractElement_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 78), (void *)_cffi_d_GrB_Scalar_extractElement_BOOL },
  { "GrB_Scalar_extractElement_FP32", (void *)_cffi_f_GrB_Scalar_extractElement_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 251), (void *)_cffi_d_GrB_Scalar_extractElement_FP32 },
  { "GrB_Scalar_extractElement_FP64", (void *)_cffi_f_GrB_Scalar_extractElement_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 193), (void *)_cffi_d_GrB_Scalar_extractElement_FP64 },
  { "GrB_Scalar_extractElement_INT16", (void *)_cffi_f_GrB_Scalar_extractElement_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 309), (void *)_cffi_d_GrB_Scalar_extractElement_INT16 },
  { "GrB_Scalar_extractElement_INT32", (void *)_cffi_f_GrB_Scalar_extractElement_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 338), (void *)_cffi_d_GrB_Scalar_extractElement_INT32 },
  { "GrB_Scalar_extractElement_INT64", (void *)_cffi_f_GrB_Scalar_extractElement_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 367), (void *)_cffi_d_GrB_Scalar_extractElement_INT64 },
  { "GrB_Scalar_extractElement_INT8", (void *)_cffi_f_GrB_Scalar_extractElement_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 396), (void *)_cffi_d_GrB_Scalar_extractElement_INT8 },
  { "GrB_Scalar_extractElement_UDT", (void *)_cffi_f_GrB_Scalar_extractElement_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3878), (void *)_cffi_d_GrB_Scalar_extractElement_UDT },
  { "GrB_Scalar_extractElement_UINT16", (void *)_cffi_f_GrB_Scalar_extractElement_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3405), (void *)_cffi_d_GrB_Scalar_extractElement_UINT16 },
  { "GrB_Scalar_extractElement_UINT32", (void *)_cffi_f_GrB_Scalar_extractElement_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3434), (void *)_cffi_d_GrB_Scalar_extractElement_UINT32 },
  { "GrB_Scalar_extractElement_UINT64", (void *)_cffi_f_GrB_Scalar_extractElement_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3521), (void *)_cffi_d_GrB_Scalar_extractElement_UINT64 },
  { "GrB_Scalar_extractElement_UINT8", (void *)_cffi_f_GrB_Scalar_extractElement_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3829), (void *)_cffi_d_GrB_Scalar_extractElement_UINT8 },
  { "GrB_Scalar_free", (void *)_cffi_f_GrB_Scalar_free, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 2108), (void *)_cffi_d_GrB_Scalar_free },
  { "GrB_Scalar_new", (void *)_cffi_f_GrB_Scalar_new, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2115), (void *)_cffi_d_GrB_Scalar_new },
  { "GrB_Scalar_nvals", (void *)_cffi_f_GrB_Scalar_nvals, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3521), (void *)_cffi_d_GrB_Scalar_nvals },
  { "GrB_Scalar_setElement_BOOL", (void *)_cffi_f_GrB_Scalar_setElement_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2126), (void *)_cffi_d_GrB_Scalar_setElement_BOOL },
  { "GrB_Scalar_setElement_FP32", (void *)_cffi_f_GrB_Scalar_setElement_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2148), (void *)_cffi_d_GrB_Scalar_setElement_FP32 },
  { "GrB_Scalar_setElement_FP64", (void *)_cffi_f_GrB_Scalar_setElement_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2140), (void *)_cffi_d_GrB_Scalar_setElement_FP64 },
  { "GrB_Scalar_setElement_INT16", (void *)_cffi_f_GrB_Scalar_setElement_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2152), (void *)_cffi_d_GrB_Scalar_setElement_INT16 },
  { "GrB_Scalar_setElement_INT32", (void *)_cffi_f_GrB_Scalar_setElement_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2156), (void *)_cffi_d_GrB_Scalar_setElement_INT32 },
  { "GrB_Scalar_setElement_INT64", (void *)_cffi_f_GrB_Scalar_setElement_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2160), (void *)_cffi_d_GrB_Scalar_setElement_INT64 },
  { "GrB_Scalar_setElement_INT8", (void *)_cffi_f_GrB_Scalar_setElement_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2164), (void *)_cffi_d_GrB_Scalar_setElement_INT8 },
  { "GrB_Scalar_setElement_UDT", (void *)_cffi_f_GrB_Scalar_setElement_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2223), (void *)_cffi_d_GrB_Scalar_setElement_UDT },
  { "GrB_Scalar_setElement_UINT16", (void *)_cffi_f_GrB_Scalar_setElement_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2207), (void *)_cffi_d_GrB_Scalar_setElement_UINT16 },
  { "GrB_Scalar_setElement_UINT32", (void *)_cffi_f_GrB_Scalar_setElement_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2211), (void *)_cffi_d_GrB_Scalar_setElement_UINT32 },
  { "GrB_Scalar_setElement_UINT64", (void *)_cffi_f_GrB_Scalar_setElement_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2215), (void *)_cffi_d_GrB_Scalar_setElement_UINT64 },
  { "GrB_Scalar_setElement_UINT8", (void *)_cffi_f_GrB_Scalar_setElement_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2219), (void *)_cffi_d_GrB_Scalar_setElement_UINT8 },
  { "GrB_Scalar_wait", (void *)_cffi_f_GrB_Scalar_wait, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2122), (void *)_cffi_d_GrB_Scalar_wait },
  { "GrB_Semiring_error", (void *)_cffi_f_GrB_Semiring_error, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 157), (void *)_cffi_d_GrB_Semiring_error },
  { "GrB_Semiring_free", (void *)_cffi_f_GrB_Semiring_free, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 2253), (void *)_cffi_d_GrB_Semiring_free },
  { "GrB_Semiring_new", (void *)_cffi_f_GrB_Semiring_new, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2256), (void *)_cffi_d_GrB_Semiring_new },
  { "GrB_Semiring_wait", (void *)_cffi_f_GrB_Semiring_wait, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2261), (void *)_cffi_d_GrB_Semiring_wait },
  { "GrB_TIMES_BOOL", (void *)_cffi_var_GrB_TIMES_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_TIMES_FP32", (void *)_cffi_var_GrB_TIMES_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_TIMES_FP64", (void *)_cffi_var_GrB_TIMES_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_TIMES_INT16", (void *)_cffi_var_GrB_TIMES_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_TIMES_INT32", (void *)_cffi_var_GrB_TIMES_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_TIMES_INT64", (void *)_cffi_var_GrB_TIMES_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_TIMES_INT8", (void *)_cffi_var_GrB_TIMES_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_TIMES_MONOID_FP32", (void *)_cffi_var_GrB_TIMES_MONOID_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_TIMES_MONOID_FP64", (void *)_cffi_var_GrB_TIMES_MONOID_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_TIMES_MONOID_INT16", (void *)_cffi_var_GrB_TIMES_MONOID_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_TIMES_MONOID_INT32", (void *)_cffi_var_GrB_TIMES_MONOID_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_TIMES_MONOID_INT64", (void *)_cffi_var_GrB_TIMES_MONOID_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_TIMES_MONOID_INT8", (void *)_cffi_var_GrB_TIMES_MONOID_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_TIMES_MONOID_UINT16", (void *)_cffi_var_GrB_TIMES_MONOID_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_TIMES_MONOID_UINT32", (void *)_cffi_var_GrB_TIMES_MONOID_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_TIMES_MONOID_UINT64", (void *)_cffi_var_GrB_TIMES_MONOID_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_TIMES_MONOID_UINT8", (void *)_cffi_var_GrB_TIMES_MONOID_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GrB_TIMES_UINT16", (void *)_cffi_var_GrB_TIMES_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_TIMES_UINT32", (void *)_cffi_var_GrB_TIMES_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_TIMES_UINT64", (void *)_cffi_var_GrB_TIMES_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_TIMES_UINT8", (void *)_cffi_var_GrB_TIMES_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GrB_TRAN", (void *)_cffi_const_GrB_TRAN, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_TRIL", (void *)_cffi_var_GrB_TRIL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_TRIU", (void *)_cffi_var_GrB_TRIU, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_Type_error", (void *)_cffi_f_GrB_Type_error, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 161), (void *)_cffi_d_GrB_Type_error },
  { "GrB_Type_free", (void *)_cffi_f_GrB_Type_free, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 2271), (void *)_cffi_d_GrB_Type_free },
  { "GrB_Type_new", (void *)_cffi_f_GrB_Type_new, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2278), (void *)_cffi_d_GrB_Type_new },
  { "GrB_Type_wait", (void *)_cffi_f_GrB_Type_wait, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2317), (void *)_cffi_d_GrB_Type_wait },
  { "GrB_UINT16", (void *)_cffi_var_GrB_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 114), (void *)0 },
  { "GrB_UINT32", (void *)_cffi_var_GrB_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 114), (void *)0 },
  { "GrB_UINT64", (void *)_cffi_var_GrB_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 114), (void *)0 },
  { "GrB_UINT8", (void *)_cffi_var_GrB_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 114), (void *)0 },
  { "GrB_UNINITIALIZED_OBJECT", (void *)_cffi_const_GrB_UNINITIALIZED_OBJECT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GrB_UnaryOp_error", (void *)_cffi_f_GrB_UnaryOp_error, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 165), (void *)_cffi_d_GrB_UnaryOp_error },
  { "GrB_UnaryOp_free", (void *)_cffi_f_GrB_UnaryOp_free, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 2327), (void *)_cffi_d_GrB_UnaryOp_free },
  { "GrB_UnaryOp_new", (void *)_cffi_f_GrB_UnaryOp_new, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2330), (void *)_cffi_d_GrB_UnaryOp_new },
  { "GrB_UnaryOp_wait", (void *)_cffi_f_GrB_UnaryOp_wait, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2351), (void *)_cffi_d_GrB_UnaryOp_wait },
  { "GrB_VALUEEQ_BOOL", (void *)_cffi_var_GrB_VALUEEQ_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEEQ_FP32", (void *)_cffi_var_GrB_VALUEEQ_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEEQ_FP64", (void *)_cffi_var_GrB_VALUEEQ_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEEQ_INT16", (void *)_cffi_var_GrB_VALUEEQ_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEEQ_INT32", (void *)_cffi_var_GrB_VALUEEQ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEEQ_INT64", (void *)_cffi_var_GrB_VALUEEQ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEEQ_INT8", (void *)_cffi_var_GrB_VALUEEQ_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEEQ_UINT16", (void *)_cffi_var_GrB_VALUEEQ_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEEQ_UINT32", (void *)_cffi_var_GrB_VALUEEQ_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEEQ_UINT64", (void *)_cffi_var_GrB_VALUEEQ_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEEQ_UINT8", (void *)_cffi_var_GrB_VALUEEQ_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEGE_BOOL", (void *)_cffi_var_GrB_VALUEGE_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEGE_FP32", (void *)_cffi_var_GrB_VALUEGE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEGE_FP64", (void *)_cffi_var_GrB_VALUEGE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEGE_INT16", (void *)_cffi_var_GrB_VALUEGE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEGE_INT32", (void *)_cffi_var_GrB_VALUEGE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEGE_INT64", (void *)_cffi_var_GrB_VALUEGE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEGE_INT8", (void *)_cffi_var_GrB_VALUEGE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEGE_UINT16", (void *)_cffi_var_GrB_VALUEGE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEGE_UINT32", (void *)_cffi_var_GrB_VALUEGE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEGE_UINT64", (void *)_cffi_var_GrB_VALUEGE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEGE_UINT8", (void *)_cffi_var_GrB_VALUEGE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEGT_BOOL", (void *)_cffi_var_GrB_VALUEGT_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEGT_FP32", (void *)_cffi_var_GrB_VALUEGT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEGT_FP64", (void *)_cffi_var_GrB_VALUEGT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEGT_INT16", (void *)_cffi_var_GrB_VALUEGT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEGT_INT32", (void *)_cffi_var_GrB_VALUEGT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEGT_INT64", (void *)_cffi_var_GrB_VALUEGT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEGT_INT8", (void *)_cffi_var_GrB_VALUEGT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEGT_UINT16", (void *)_cffi_var_GrB_VALUEGT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEGT_UINT32", (void *)_cffi_var_GrB_VALUEGT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEGT_UINT64", (void *)_cffi_var_GrB_VALUEGT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUEGT_UINT8", (void *)_cffi_var_GrB_VALUEGT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUELE_BOOL", (void *)_cffi_var_GrB_VALUELE_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUELE_FP32", (void *)_cffi_var_GrB_VALUELE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUELE_FP64", (void *)_cffi_var_GrB_VALUELE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUELE_INT16", (void *)_cffi_var_GrB_VALUELE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUELE_INT32", (void *)_cffi_var_GrB_VALUELE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUELE_INT64", (void *)_cffi_var_GrB_VALUELE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUELE_INT8", (void *)_cffi_var_GrB_VALUELE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUELE_UINT16", (void *)_cffi_var_GrB_VALUELE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUELE_UINT32", (void *)_cffi_var_GrB_VALUELE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUELE_UINT64", (void *)_cffi_var_GrB_VALUELE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUELE_UINT8", (void *)_cffi_var_GrB_VALUELE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUELT_BOOL", (void *)_cffi_var_GrB_VALUELT_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUELT_FP32", (void *)_cffi_var_GrB_VALUELT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUELT_FP64", (void *)_cffi_var_GrB_VALUELT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUELT_INT16", (void *)_cffi_var_GrB_VALUELT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUELT_INT32", (void *)_cffi_var_GrB_VALUELT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUELT_INT64", (void *)_cffi_var_GrB_VALUELT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUELT_INT8", (void *)_cffi_var_GrB_VALUELT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUELT_UINT16", (void *)_cffi_var_GrB_VALUELT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUELT_UINT32", (void *)_cffi_var_GrB_VALUELT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUELT_UINT64", (void *)_cffi_var_GrB_VALUELT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUELT_UINT8", (void *)_cffi_var_GrB_VALUELT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUENE_BOOL", (void *)_cffi_var_GrB_VALUENE_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUENE_FP32", (void *)_cffi_var_GrB_VALUENE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUENE_FP64", (void *)_cffi_var_GrB_VALUENE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUENE_INT16", (void *)_cffi_var_GrB_VALUENE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUENE_INT32", (void *)_cffi_var_GrB_VALUENE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUENE_INT64", (void *)_cffi_var_GrB_VALUENE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUENE_INT8", (void *)_cffi_var_GrB_VALUENE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUENE_UINT16", (void *)_cffi_var_GrB_VALUENE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUENE_UINT32", (void *)_cffi_var_GrB_VALUENE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUENE_UINT64", (void *)_cffi_var_GrB_VALUENE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_VALUENE_UINT8", (void *)_cffi_var_GrB_VALUENE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GrB_Vector_apply", (void *)_cffi_f_GrB_Vector_apply, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3147), (void *)_cffi_d_GrB_Vector_apply },
  { "GrB_Vector_apply_BinaryOp1st_BOOL", (void *)_cffi_f_GrB_Vector_apply_BinaryOp1st_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2642), (void *)_cffi_d_GrB_Vector_apply_BinaryOp1st_BOOL },
  { "GrB_Vector_apply_BinaryOp1st_FP32", (void *)_cffi_f_GrB_Vector_apply_BinaryOp1st_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2678), (void *)_cffi_d_GrB_Vector_apply_BinaryOp1st_FP32 },
  { "GrB_Vector_apply_BinaryOp1st_FP64", (void *)_cffi_f_GrB_Vector_apply_BinaryOp1st_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2660), (void *)_cffi_d_GrB_Vector_apply_BinaryOp1st_FP64 },
  { "GrB_Vector_apply_BinaryOp1st_INT16", (void *)_cffi_f_GrB_Vector_apply_BinaryOp1st_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2687), (void *)_cffi_d_GrB_Vector_apply_BinaryOp1st_INT16 },
  { "GrB_Vector_apply_BinaryOp1st_INT32", (void *)_cffi_f_GrB_Vector_apply_BinaryOp1st_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2696), (void *)_cffi_d_GrB_Vector_apply_BinaryOp1st_INT32 },
  { "GrB_Vector_apply_BinaryOp1st_INT64", (void *)_cffi_f_GrB_Vector_apply_BinaryOp1st_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2705), (void *)_cffi_d_GrB_Vector_apply_BinaryOp1st_INT64 },
  { "GrB_Vector_apply_BinaryOp1st_INT8", (void *)_cffi_f_GrB_Vector_apply_BinaryOp1st_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2714), (void *)_cffi_d_GrB_Vector_apply_BinaryOp1st_INT8 },
  { "GrB_Vector_apply_BinaryOp1st_Scalar", (void *)_cffi_f_GrB_Vector_apply_BinaryOp1st_Scalar, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2731), (void *)_cffi_d_GrB_Vector_apply_BinaryOp1st_Scalar },
  { "GrB_Vector_apply_BinaryOp1st_UDT", (void *)_cffi_f_GrB_Vector_apply_BinaryOp1st_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2931), (void *)_cffi_d_GrB_Vector_apply_BinaryOp1st_UDT },
  { "GrB_Vector_apply_BinaryOp1st_UINT16", (void *)_cffi_f_GrB_Vector_apply_BinaryOp1st_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2895), (void *)_cffi_d_GrB_Vector_apply_BinaryOp1st_UINT16 },
  { "GrB_Vector_apply_BinaryOp1st_UINT32", (void *)_cffi_f_GrB_Vector_apply_BinaryOp1st_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2904), (void *)_cffi_d_GrB_Vector_apply_BinaryOp1st_UINT32 },
  { "GrB_Vector_apply_BinaryOp1st_UINT64", (void *)_cffi_f_GrB_Vector_apply_BinaryOp1st_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2913), (void *)_cffi_d_GrB_Vector_apply_BinaryOp1st_UINT64 },
  { "GrB_Vector_apply_BinaryOp1st_UINT8", (void *)_cffi_f_GrB_Vector_apply_BinaryOp1st_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2922), (void *)_cffi_d_GrB_Vector_apply_BinaryOp1st_UINT8 },
  { "GrB_Vector_apply_BinaryOp2nd_BOOL", (void *)_cffi_f_GrB_Vector_apply_BinaryOp2nd_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2740), (void *)_cffi_d_GrB_Vector_apply_BinaryOp2nd_BOOL },
  { "GrB_Vector_apply_BinaryOp2nd_FP32", (void *)_cffi_f_GrB_Vector_apply_BinaryOp2nd_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2776), (void *)_cffi_d_GrB_Vector_apply_BinaryOp2nd_FP32 },
  { "GrB_Vector_apply_BinaryOp2nd_FP64", (void *)_cffi_f_GrB_Vector_apply_BinaryOp2nd_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2758), (void *)_cffi_d_GrB_Vector_apply_BinaryOp2nd_FP64 },
  { "GrB_Vector_apply_BinaryOp2nd_INT16", (void *)_cffi_f_GrB_Vector_apply_BinaryOp2nd_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2785), (void *)_cffi_d_GrB_Vector_apply_BinaryOp2nd_INT16 },
  { "GrB_Vector_apply_BinaryOp2nd_INT32", (void *)_cffi_f_GrB_Vector_apply_BinaryOp2nd_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2794), (void *)_cffi_d_GrB_Vector_apply_BinaryOp2nd_INT32 },
  { "GrB_Vector_apply_BinaryOp2nd_INT64", (void *)_cffi_f_GrB_Vector_apply_BinaryOp2nd_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2803), (void *)_cffi_d_GrB_Vector_apply_BinaryOp2nd_INT64 },
  { "GrB_Vector_apply_BinaryOp2nd_INT8", (void *)_cffi_f_GrB_Vector_apply_BinaryOp2nd_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2812), (void *)_cffi_d_GrB_Vector_apply_BinaryOp2nd_INT8 },
  { "GrB_Vector_apply_BinaryOp2nd_Scalar", (void *)_cffi_f_GrB_Vector_apply_BinaryOp2nd_Scalar, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2821), (void *)_cffi_d_GrB_Vector_apply_BinaryOp2nd_Scalar },
  { "GrB_Vector_apply_BinaryOp2nd_UDT", (void *)_cffi_f_GrB_Vector_apply_BinaryOp2nd_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2886), (void *)_cffi_d_GrB_Vector_apply_BinaryOp2nd_UDT },
  { "GrB_Vector_apply_BinaryOp2nd_UINT16", (void *)_cffi_f_GrB_Vector_apply_BinaryOp2nd_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2850), (void *)_cffi_d_GrB_Vector_apply_BinaryOp2nd_UINT16 },
  { "GrB_Vector_apply_BinaryOp2nd_UINT32", (void *)_cffi_f_GrB_Vector_apply_BinaryOp2nd_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2859), (void *)_cffi_d_GrB_Vector_apply_BinaryOp2nd_UINT32 },
  { "GrB_Vector_apply_BinaryOp2nd_UINT64", (void *)_cffi_f_GrB_Vector_apply_BinaryOp2nd_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2868), (void *)_cffi_d_GrB_Vector_apply_BinaryOp2nd_UINT64 },
  { "GrB_Vector_apply_BinaryOp2nd_UINT8", (void *)_cffi_f_GrB_Vector_apply_BinaryOp2nd_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2877), (void *)_cffi_d_GrB_Vector_apply_BinaryOp2nd_UINT8 },
  { "GrB_Vector_apply_IndexOp_BOOL", (void *)_cffi_f_GrB_Vector_apply_IndexOp_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2940), (void *)_cffi_d_GrB_Vector_apply_IndexOp_BOOL },
  { "GrB_Vector_apply_IndexOp_FP32", (void *)_cffi_f_GrB_Vector_apply_IndexOp_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2976), (void *)_cffi_d_GrB_Vector_apply_IndexOp_FP32 },
  { "GrB_Vector_apply_IndexOp_FP64", (void *)_cffi_f_GrB_Vector_apply_IndexOp_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2958), (void *)_cffi_d_GrB_Vector_apply_IndexOp_FP64 },
  { "GrB_Vector_apply_IndexOp_INT16", (void *)_cffi_f_GrB_Vector_apply_IndexOp_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2985), (void *)_cffi_d_GrB_Vector_apply_IndexOp_INT16 },
  { "GrB_Vector_apply_IndexOp_INT32", (void *)_cffi_f_GrB_Vector_apply_IndexOp_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2994), (void *)_cffi_d_GrB_Vector_apply_IndexOp_INT32 },
  { "GrB_Vector_apply_IndexOp_INT64", (void *)_cffi_f_GrB_Vector_apply_IndexOp_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3003), (void *)_cffi_d_GrB_Vector_apply_IndexOp_INT64 },
  { "GrB_Vector_apply_IndexOp_INT8", (void *)_cffi_f_GrB_Vector_apply_IndexOp_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3012), (void *)_cffi_d_GrB_Vector_apply_IndexOp_INT8 },
  { "GrB_Vector_apply_IndexOp_Scalar", (void *)_cffi_f_GrB_Vector_apply_IndexOp_Scalar, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3021), (void *)_cffi_d_GrB_Vector_apply_IndexOp_Scalar },
  { "GrB_Vector_apply_IndexOp_UDT", (void *)_cffi_f_GrB_Vector_apply_IndexOp_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3066), (void *)_cffi_d_GrB_Vector_apply_IndexOp_UDT },
  { "GrB_Vector_apply_IndexOp_UINT16", (void *)_cffi_f_GrB_Vector_apply_IndexOp_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3030), (void *)_cffi_d_GrB_Vector_apply_IndexOp_UINT16 },
  { "GrB_Vector_apply_IndexOp_UINT32", (void *)_cffi_f_GrB_Vector_apply_IndexOp_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3039), (void *)_cffi_d_GrB_Vector_apply_IndexOp_UINT32 },
  { "GrB_Vector_apply_IndexOp_UINT64", (void *)_cffi_f_GrB_Vector_apply_IndexOp_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3048), (void *)_cffi_d_GrB_Vector_apply_IndexOp_UINT64 },
  { "GrB_Vector_apply_IndexOp_UINT8", (void *)_cffi_f_GrB_Vector_apply_IndexOp_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3057), (void *)_cffi_d_GrB_Vector_apply_IndexOp_UINT8 },
  { "GrB_Vector_assign", (void *)_cffi_f_GrB_Vector_assign, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3162), (void *)_cffi_d_GrB_Vector_assign },
  { "GrB_Vector_assign_BOOL", (void *)_cffi_f_GrB_Vector_assign_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2561), (void *)_cffi_d_GrB_Vector_assign_BOOL },
  { "GrB_Vector_assign_FP32", (void *)_cffi_f_GrB_Vector_assign_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2597), (void *)_cffi_d_GrB_Vector_assign_FP32 },
  { "GrB_Vector_assign_FP64", (void *)_cffi_f_GrB_Vector_assign_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2579), (void *)_cffi_d_GrB_Vector_assign_FP64 },
  { "GrB_Vector_assign_INT16", (void *)_cffi_f_GrB_Vector_assign_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2606), (void *)_cffi_d_GrB_Vector_assign_INT16 },
  { "GrB_Vector_assign_INT32", (void *)_cffi_f_GrB_Vector_assign_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2615), (void *)_cffi_d_GrB_Vector_assign_INT32 },
  { "GrB_Vector_assign_INT64", (void *)_cffi_f_GrB_Vector_assign_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2624), (void *)_cffi_d_GrB_Vector_assign_INT64 },
  { "GrB_Vector_assign_INT8", (void *)_cffi_f_GrB_Vector_assign_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2633), (void *)_cffi_d_GrB_Vector_assign_INT8 },
  { "GrB_Vector_assign_Scalar", (void *)_cffi_f_GrB_Vector_assign_Scalar, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3102), (void *)_cffi_d_GrB_Vector_assign_Scalar },
  { "GrB_Vector_assign_UDT", (void *)_cffi_f_GrB_Vector_assign_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3207), (void *)_cffi_d_GrB_Vector_assign_UDT },
  { "GrB_Vector_assign_UINT16", (void *)_cffi_f_GrB_Vector_assign_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3171), (void *)_cffi_d_GrB_Vector_assign_UINT16 },
  { "GrB_Vector_assign_UINT32", (void *)_cffi_f_GrB_Vector_assign_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3180), (void *)_cffi_d_GrB_Vector_assign_UINT32 },
  { "GrB_Vector_assign_UINT64", (void *)_cffi_f_GrB_Vector_assign_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3189), (void *)_cffi_d_GrB_Vector_assign_UINT64 },
  { "GrB_Vector_assign_UINT8", (void *)_cffi_f_GrB_Vector_assign_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3198), (void *)_cffi_d_GrB_Vector_assign_UINT8 },
  { "GrB_Vector_build_BOOL", (void *)_cffi_f_GrB_Vector_build_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3248), (void *)_cffi_d_GrB_Vector_build_BOOL },
  { "GrB_Vector_build_FP32", (void *)_cffi_f_GrB_Vector_build_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3276), (void *)_cffi_d_GrB_Vector_build_FP32 },
  { "GrB_Vector_build_FP64", (void *)_cffi_f_GrB_Vector_build_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3262), (void *)_cffi_d_GrB_Vector_build_FP64 },
  { "GrB_Vector_build_INT16", (void *)_cffi_f_GrB_Vector_build_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3283), (void *)_cffi_d_GrB_Vector_build_INT16 },
  { "GrB_Vector_build_INT32", (void *)_cffi_f_GrB_Vector_build_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3290), (void *)_cffi_d_GrB_Vector_build_INT32 },
  { "GrB_Vector_build_INT64", (void *)_cffi_f_GrB_Vector_build_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3297), (void *)_cffi_d_GrB_Vector_build_INT64 },
  { "GrB_Vector_build_INT8", (void *)_cffi_f_GrB_Vector_build_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3304), (void *)_cffi_d_GrB_Vector_build_INT8 },
  { "GrB_Vector_build_UDT", (void *)_cffi_f_GrB_Vector_build_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3345), (void *)_cffi_d_GrB_Vector_build_UDT },
  { "GrB_Vector_build_UINT16", (void *)_cffi_f_GrB_Vector_build_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3317), (void *)_cffi_d_GrB_Vector_build_UINT16 },
  { "GrB_Vector_build_UINT32", (void *)_cffi_f_GrB_Vector_build_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3324), (void *)_cffi_d_GrB_Vector_build_UINT32 },
  { "GrB_Vector_build_UINT64", (void *)_cffi_f_GrB_Vector_build_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3331), (void *)_cffi_d_GrB_Vector_build_UINT64 },
  { "GrB_Vector_build_UINT8", (void *)_cffi_f_GrB_Vector_build_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3338), (void *)_cffi_d_GrB_Vector_build_UINT8 },
  { "GrB_Vector_clear", (void *)_cffi_f_GrB_Vector_clear, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 2448), (void *)_cffi_d_GrB_Vector_clear },
  { "GrB_Vector_dup", (void *)_cffi_f_GrB_Vector_dup, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2444), (void *)_cffi_d_GrB_Vector_dup },
  { "GrB_Vector_eWiseAdd_BinaryOp", (void *)_cffi_f_GrB_Vector_eWiseAdd_BinaryOp, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2841), (void *)_cffi_d_GrB_Vector_eWiseAdd_BinaryOp },
  { "GrB_Vector_eWiseAdd_Monoid", (void *)_cffi_f_GrB_Vector_eWiseAdd_Monoid, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3093), (void *)_cffi_d_GrB_Vector_eWiseAdd_Monoid },
  { "GrB_Vector_eWiseAdd_Semiring", (void *)_cffi_f_GrB_Vector_eWiseAdd_Semiring, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3138), (void *)_cffi_d_GrB_Vector_eWiseAdd_Semiring },
  { "GrB_Vector_eWiseMult_BinaryOp", (void *)_cffi_f_GrB_Vector_eWiseMult_BinaryOp, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2841), (void *)_cffi_d_GrB_Vector_eWiseMult_BinaryOp },
  { "GrB_Vector_eWiseMult_Monoid", (void *)_cffi_f_GrB_Vector_eWiseMult_Monoid, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3093), (void *)_cffi_d_GrB_Vector_eWiseMult_Monoid },
  { "GrB_Vector_eWiseMult_Semiring", (void *)_cffi_f_GrB_Vector_eWiseMult_Semiring, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3138), (void *)_cffi_d_GrB_Vector_eWiseMult_Semiring },
  { "GrB_Vector_error", (void *)_cffi_f_GrB_Vector_error, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 169), (void *)_cffi_d_GrB_Vector_error },
  { "GrB_Vector_extract", (void *)_cffi_f_GrB_Vector_extract, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3162), (void *)_cffi_d_GrB_Vector_extract },
  { "GrB_Vector_extractElement_BOOL", (void *)_cffi_f_GrB_Vector_extractElement_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 86), (void *)_cffi_d_GrB_Vector_extractElement_BOOL },
  { "GrB_Vector_extractElement_FP32", (void *)_cffi_f_GrB_Vector_extractElement_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 255), (void *)_cffi_d_GrB_Vector_extractElement_FP32 },
  { "GrB_Vector_extractElement_FP64", (void *)_cffi_f_GrB_Vector_extractElement_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 197), (void *)_cffi_d_GrB_Vector_extractElement_FP64 },
  { "GrB_Vector_extractElement_INT16", (void *)_cffi_f_GrB_Vector_extractElement_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 313), (void *)_cffi_d_GrB_Vector_extractElement_INT16 },
  { "GrB_Vector_extractElement_INT32", (void *)_cffi_f_GrB_Vector_extractElement_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 342), (void *)_cffi_d_GrB_Vector_extractElement_INT32 },
  { "GrB_Vector_extractElement_INT64", (void *)_cffi_f_GrB_Vector_extractElement_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 371), (void *)_cffi_d_GrB_Vector_extractElement_INT64 },
  { "GrB_Vector_extractElement_INT8", (void *)_cffi_f_GrB_Vector_extractElement_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 400), (void *)_cffi_d_GrB_Vector_extractElement_INT8 },
  { "GrB_Vector_extractElement_Scalar", (void *)_cffi_f_GrB_Vector_extractElement_Scalar, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2202), (void *)_cffi_d_GrB_Vector_extractElement_Scalar },
  { "GrB_Vector_extractElement_UDT", (void *)_cffi_f_GrB_Vector_extractElement_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3882), (void *)_cffi_d_GrB_Vector_extractElement_UDT },
  { "GrB_Vector_extractElement_UINT16", (void *)_cffi_f_GrB_Vector_extractElement_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3409), (void *)_cffi_d_GrB_Vector_extractElement_UINT16 },
  { "GrB_Vector_extractElement_UINT32", (void *)_cffi_f_GrB_Vector_extractElement_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3438), (void *)_cffi_d_GrB_Vector_extractElement_UINT32 },
  { "GrB_Vector_extractElement_UINT64", (void *)_cffi_f_GrB_Vector_extractElement_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3529), (void *)_cffi_d_GrB_Vector_extractElement_UINT64 },
  { "GrB_Vector_extractElement_UINT8", (void *)_cffi_f_GrB_Vector_extractElement_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3833), (void *)_cffi_d_GrB_Vector_extractElement_UINT8 },
  { "GrB_Vector_extractTuples_BOOL", (void *)_cffi_f_GrB_Vector_extractTuples_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3443), (void *)_cffi_d_GrB_Vector_extractTuples_BOOL },
  { "GrB_Vector_extractTuples_FP32", (void *)_cffi_f_GrB_Vector_extractTuples_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3461), (void *)_cffi_d_GrB_Vector_extractTuples_FP32 },
  { "GrB_Vector_extractTuples_FP64", (void *)_cffi_f_GrB_Vector_extractTuples_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3449), (void *)_cffi_d_GrB_Vector_extractTuples_FP64 },
  { "GrB_Vector_extractTuples_INT16", (void *)_cffi_f_GrB_Vector_extractTuples_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3473), (void *)_cffi_d_GrB_Vector_extractTuples_INT16 },
  { "GrB_Vector_extractTuples_INT32", (void *)_cffi_f_GrB_Vector_extractTuples_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3479), (void *)_cffi_d_GrB_Vector_extractTuples_INT32 },
  { "GrB_Vector_extractTuples_INT64", (void *)_cffi_f_GrB_Vector_extractTuples_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3485), (void *)_cffi_d_GrB_Vector_extractTuples_INT64 },
  { "GrB_Vector_extractTuples_INT8", (void *)_cffi_f_GrB_Vector_extractTuples_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3491), (void *)_cffi_d_GrB_Vector_extractTuples_INT8 },
  { "GrB_Vector_extractTuples_UDT", (void *)_cffi_f_GrB_Vector_extractTuples_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3803), (void *)_cffi_d_GrB_Vector_extractTuples_UDT },
  { "GrB_Vector_extractTuples_UINT16", (void *)_cffi_f_GrB_Vector_extractTuples_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3534), (void *)_cffi_d_GrB_Vector_extractTuples_UINT16 },
  { "GrB_Vector_extractTuples_UINT32", (void *)_cffi_f_GrB_Vector_extractTuples_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3540), (void *)_cffi_d_GrB_Vector_extractTuples_UINT32 },
  { "GrB_Vector_extractTuples_UINT64", (void *)_cffi_f_GrB_Vector_extractTuples_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3740), (void *)_cffi_d_GrB_Vector_extractTuples_UINT64 },
  { "GrB_Vector_extractTuples_UINT8", (void *)_cffi_f_GrB_Vector_extractTuples_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3797), (void *)_cffi_d_GrB_Vector_extractTuples_UINT8 },
  { "GrB_Vector_free", (void *)_cffi_f_GrB_Vector_free, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 2361), (void *)_cffi_d_GrB_Vector_free },
  { "GrB_Vector_new", (void *)_cffi_f_GrB_Vector_new, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2398), (void *)_cffi_d_GrB_Vector_new },
  { "GrB_Vector_nvals", (void *)_cffi_f_GrB_Vector_nvals, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3525), (void *)_cffi_d_GrB_Vector_nvals },
  { "GrB_Vector_reduce_BOOL", (void *)_cffi_f_GrB_Vector_reduce_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 61), (void *)_cffi_d_GrB_Vector_reduce_BOOL },
  { "GrB_Vector_reduce_BinaryOp_Scalar", (void *)_cffi_f_GrB_Vector_reduce_BinaryOp_Scalar, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2175), (void *)_cffi_d_GrB_Vector_reduce_BinaryOp_Scalar },
  { "GrB_Vector_reduce_FP32", (void *)_cffi_f_GrB_Vector_reduce_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 238), (void *)_cffi_d_GrB_Vector_reduce_FP32 },
  { "GrB_Vector_reduce_FP64", (void *)_cffi_f_GrB_Vector_reduce_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 180), (void *)_cffi_d_GrB_Vector_reduce_FP64 },
  { "GrB_Vector_reduce_INT16", (void *)_cffi_f_GrB_Vector_reduce_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 296), (void *)_cffi_d_GrB_Vector_reduce_INT16 },
  { "GrB_Vector_reduce_INT32", (void *)_cffi_f_GrB_Vector_reduce_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 325), (void *)_cffi_d_GrB_Vector_reduce_INT32 },
  { "GrB_Vector_reduce_INT64", (void *)_cffi_f_GrB_Vector_reduce_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 354), (void *)_cffi_d_GrB_Vector_reduce_INT64 },
  { "GrB_Vector_reduce_INT8", (void *)_cffi_f_GrB_Vector_reduce_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 383), (void *)_cffi_d_GrB_Vector_reduce_INT8 },
  { "GrB_Vector_reduce_Monoid_Scalar", (void *)_cffi_f_GrB_Vector_reduce_Monoid_Scalar, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2189), (void *)_cffi_d_GrB_Vector_reduce_Monoid_Scalar },
  { "GrB_Vector_reduce_UDT", (void *)_cffi_f_GrB_Vector_reduce_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3861), (void *)_cffi_d_GrB_Vector_reduce_UDT },
  { "GrB_Vector_reduce_UINT16", (void *)_cffi_f_GrB_Vector_reduce_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3392), (void *)_cffi_d_GrB_Vector_reduce_UINT16 },
  { "GrB_Vector_reduce_UINT32", (void *)_cffi_f_GrB_Vector_reduce_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3421), (void *)_cffi_d_GrB_Vector_reduce_UINT32 },
  { "GrB_Vector_reduce_UINT64", (void *)_cffi_f_GrB_Vector_reduce_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3504), (void *)_cffi_d_GrB_Vector_reduce_UINT64 },
  { "GrB_Vector_reduce_UINT8", (void *)_cffi_f_GrB_Vector_reduce_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3816), (void *)_cffi_d_GrB_Vector_reduce_UINT8 },
  { "GrB_Vector_removeElement", (void *)_cffi_f_GrB_Vector_removeElement, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3352), (void *)_cffi_d_GrB_Vector_removeElement },
  { "GrB_Vector_resize", (void *)_cffi_f_GrB_Vector_resize, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3352), (void *)_cffi_d_GrB_Vector_resize },
  { "GrB_Vector_select_BOOL", (void *)_cffi_f_GrB_Vector_select_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2940), (void *)_cffi_d_GrB_Vector_select_BOOL },
  { "GrB_Vector_select_FP32", (void *)_cffi_f_GrB_Vector_select_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2976), (void *)_cffi_d_GrB_Vector_select_FP32 },
  { "GrB_Vector_select_FP64", (void *)_cffi_f_GrB_Vector_select_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2958), (void *)_cffi_d_GrB_Vector_select_FP64 },
  { "GrB_Vector_select_INT16", (void *)_cffi_f_GrB_Vector_select_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2985), (void *)_cffi_d_GrB_Vector_select_INT16 },
  { "GrB_Vector_select_INT32", (void *)_cffi_f_GrB_Vector_select_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2994), (void *)_cffi_d_GrB_Vector_select_INT32 },
  { "GrB_Vector_select_INT64", (void *)_cffi_f_GrB_Vector_select_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3003), (void *)_cffi_d_GrB_Vector_select_INT64 },
  { "GrB_Vector_select_INT8", (void *)_cffi_f_GrB_Vector_select_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3012), (void *)_cffi_d_GrB_Vector_select_INT8 },
  { "GrB_Vector_select_Scalar", (void *)_cffi_f_GrB_Vector_select_Scalar, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3021), (void *)_cffi_d_GrB_Vector_select_Scalar },
  { "GrB_Vector_select_UDT", (void *)_cffi_f_GrB_Vector_select_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3066), (void *)_cffi_d_GrB_Vector_select_UDT },
  { "GrB_Vector_select_UINT16", (void *)_cffi_f_GrB_Vector_select_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3030), (void *)_cffi_d_GrB_Vector_select_UINT16 },
  { "GrB_Vector_select_UINT32", (void *)_cffi_f_GrB_Vector_select_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3039), (void *)_cffi_d_GrB_Vector_select_UINT32 },
  { "GrB_Vector_select_UINT64", (void *)_cffi_f_GrB_Vector_select_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3048), (void *)_cffi_d_GrB_Vector_select_UINT64 },
  { "GrB_Vector_select_UINT8", (void *)_cffi_f_GrB_Vector_select_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3057), (void *)_cffi_d_GrB_Vector_select_UINT8 },
  { "GrB_Vector_setElement_BOOL", (void *)_cffi_f_GrB_Vector_setElement_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2479), (void *)_cffi_d_GrB_Vector_setElement_BOOL },
  { "GrB_Vector_setElement_FP32", (void *)_cffi_f_GrB_Vector_setElement_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2505), (void *)_cffi_d_GrB_Vector_setElement_FP32 },
  { "GrB_Vector_setElement_FP64", (void *)_cffi_f_GrB_Vector_setElement_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2495), (void *)_cffi_d_GrB_Vector_setElement_FP64 },
  { "GrB_Vector_setElement_INT16", (void *)_cffi_f_GrB_Vector_setElement_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2510), (void *)_cffi_d_GrB_Vector_setElement_INT16 },
  { "GrB_Vector_setElement_INT32", (void *)_cffi_f_GrB_Vector_setElement_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2515), (void *)_cffi_d_GrB_Vector_setElement_INT32 },
  { "GrB_Vector_setElement_INT64", (void *)_cffi_f_GrB_Vector_setElement_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2520), (void *)_cffi_d_GrB_Vector_setElement_INT64 },
  { "GrB_Vector_setElement_INT8", (void *)_cffi_f_GrB_Vector_setElement_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2545), (void *)_cffi_d_GrB_Vector_setElement_INT8 },
  { "GrB_Vector_setElement_Scalar", (void *)_cffi_f_GrB_Vector_setElement_Scalar, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2556), (void *)_cffi_d_GrB_Vector_setElement_Scalar },
  { "GrB_Vector_setElement_UDT", (void *)_cffi_f_GrB_Vector_setElement_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3380), (void *)_cffi_d_GrB_Vector_setElement_UDT },
  { "GrB_Vector_setElement_UINT16", (void *)_cffi_f_GrB_Vector_setElement_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3216), (void *)_cffi_d_GrB_Vector_setElement_UINT16 },
  { "GrB_Vector_setElement_UINT32", (void *)_cffi_f_GrB_Vector_setElement_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3221), (void *)_cffi_d_GrB_Vector_setElement_UINT32 },
  { "GrB_Vector_setElement_UINT64", (void *)_cffi_f_GrB_Vector_setElement_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3356), (void *)_cffi_d_GrB_Vector_setElement_UINT64 },
  { "GrB_Vector_setElement_UINT8", (void *)_cffi_f_GrB_Vector_setElement_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3361), (void *)_cffi_d_GrB_Vector_setElement_UINT8 },
  { "GrB_Vector_size", (void *)_cffi_f_GrB_Vector_size, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3525), (void *)_cffi_d_GrB_Vector_size },
  { "GrB_Vector_wait", (void *)_cffi_f_GrB_Vector_wait, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2451), (void *)_cffi_d_GrB_Vector_wait },
  { "GrB_finalize", (void *)_cffi_f_GrB_finalize, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_N, 3892), (void *)_cffi_d_GrB_finalize },
  { "GrB_getVersion", (void *)_cffi_f_GrB_getVersion, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3838), (void *)_cffi_d_GrB_getVersion },
  { "GrB_init", (void *)_cffi_f_GrB_init, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 9), (void *)_cffi_d_GrB_init },
  { "GrB_mxm", (void *)_cffi_f_GrB_mxm, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1611), (void *)_cffi_d_GrB_mxm },
  { "GrB_mxv", (void *)_cffi_f_GrB_mxv, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3120), (void *)_cffi_d_GrB_mxv },
  { "GrB_transpose", (void *)_cffi_f_GrB_transpose, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1564), (void *)_cffi_d_GrB_transpose },
  { "GrB_vxm", (void *)_cffi_f_GrB_vxm, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3129), (void *)_cffi_d_GrB_vxm },
  { "GxB_ABS_FC32", (void *)_cffi_var_GxB_ABS_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ABS_FC64", (void *)_cffi_var_GxB_ABS_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ACOSH_FC32", (void *)_cffi_var_GxB_ACOSH_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ACOSH_FC64", (void *)_cffi_var_GxB_ACOSH_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ACOSH_FP32", (void *)_cffi_var_GxB_ACOSH_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ACOSH_FP64", (void *)_cffi_var_GxB_ACOSH_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ACOS_FC32", (void *)_cffi_var_GxB_ACOS_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ACOS_FC64", (void *)_cffi_var_GxB_ACOS_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ACOS_FP32", (void *)_cffi_var_GxB_ACOS_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ACOS_FP64", (void *)_cffi_var_GxB_ACOS_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_AINV_FC32", (void *)_cffi_var_GxB_AINV_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_AINV_FC64", (void *)_cffi_var_GxB_AINV_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ALWAYS_HYPER", (void *)_cffi_const_GxB_ALWAYS_HYPER, _CFFI_OP(_CFFI_OP_CONSTANT, 32), (void *)0 },
  { "GxB_ANY_BOOL", (void *)_cffi_var_GxB_ANY_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ANY_BOOL_MONOID", (void *)_cffi_var_GxB_ANY_BOOL_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_ANY_DIV_FC32", (void *)_cffi_var_GxB_ANY_DIV_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_DIV_FC64", (void *)_cffi_var_GxB_ANY_DIV_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_DIV_FP32", (void *)_cffi_var_GxB_ANY_DIV_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_DIV_FP64", (void *)_cffi_var_GxB_ANY_DIV_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_DIV_INT16", (void *)_cffi_var_GxB_ANY_DIV_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_DIV_INT32", (void *)_cffi_var_GxB_ANY_DIV_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_DIV_INT64", (void *)_cffi_var_GxB_ANY_DIV_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_DIV_INT8", (void *)_cffi_var_GxB_ANY_DIV_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_DIV_UINT16", (void *)_cffi_var_GxB_ANY_DIV_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_DIV_UINT32", (void *)_cffi_var_GxB_ANY_DIV_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_DIV_UINT64", (void *)_cffi_var_GxB_ANY_DIV_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_DIV_UINT8", (void *)_cffi_var_GxB_ANY_DIV_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_EQ_BOOL", (void *)_cffi_var_GxB_ANY_EQ_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_EQ_FP32", (void *)_cffi_var_GxB_ANY_EQ_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_EQ_FP64", (void *)_cffi_var_GxB_ANY_EQ_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_EQ_INT16", (void *)_cffi_var_GxB_ANY_EQ_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_EQ_INT32", (void *)_cffi_var_GxB_ANY_EQ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_EQ_INT64", (void *)_cffi_var_GxB_ANY_EQ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_EQ_INT8", (void *)_cffi_var_GxB_ANY_EQ_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_EQ_UINT16", (void *)_cffi_var_GxB_ANY_EQ_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_EQ_UINT32", (void *)_cffi_var_GxB_ANY_EQ_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_EQ_UINT64", (void *)_cffi_var_GxB_ANY_EQ_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_EQ_UINT8", (void *)_cffi_var_GxB_ANY_EQ_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_FC32", (void *)_cffi_var_GxB_ANY_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ANY_FC32_MONOID", (void *)_cffi_var_GxB_ANY_FC32_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_ANY_FC64", (void *)_cffi_var_GxB_ANY_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ANY_FC64_MONOID", (void *)_cffi_var_GxB_ANY_FC64_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_ANY_FIRSTI1_INT32", (void *)_cffi_var_GxB_ANY_FIRSTI1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_FIRSTI1_INT64", (void *)_cffi_var_GxB_ANY_FIRSTI1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_FIRSTI_INT32", (void *)_cffi_var_GxB_ANY_FIRSTI_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_FIRSTI_INT64", (void *)_cffi_var_GxB_ANY_FIRSTI_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_FIRSTJ1_INT32", (void *)_cffi_var_GxB_ANY_FIRSTJ1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_FIRSTJ1_INT64", (void *)_cffi_var_GxB_ANY_FIRSTJ1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_FIRSTJ_INT32", (void *)_cffi_var_GxB_ANY_FIRSTJ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_FIRSTJ_INT64", (void *)_cffi_var_GxB_ANY_FIRSTJ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_FIRST_BOOL", (void *)_cffi_var_GxB_ANY_FIRST_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_FIRST_FC32", (void *)_cffi_var_GxB_ANY_FIRST_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_FIRST_FC64", (void *)_cffi_var_GxB_ANY_FIRST_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_FIRST_FP32", (void *)_cffi_var_GxB_ANY_FIRST_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_FIRST_FP64", (void *)_cffi_var_GxB_ANY_FIRST_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_FIRST_INT16", (void *)_cffi_var_GxB_ANY_FIRST_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_FIRST_INT32", (void *)_cffi_var_GxB_ANY_FIRST_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_FIRST_INT64", (void *)_cffi_var_GxB_ANY_FIRST_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_FIRST_INT8", (void *)_cffi_var_GxB_ANY_FIRST_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_FIRST_UINT16", (void *)_cffi_var_GxB_ANY_FIRST_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_FIRST_UINT32", (void *)_cffi_var_GxB_ANY_FIRST_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_FIRST_UINT64", (void *)_cffi_var_GxB_ANY_FIRST_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_FIRST_UINT8", (void *)_cffi_var_GxB_ANY_FIRST_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_FP32", (void *)_cffi_var_GxB_ANY_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ANY_FP32_MONOID", (void *)_cffi_var_GxB_ANY_FP32_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_ANY_FP64", (void *)_cffi_var_GxB_ANY_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ANY_FP64_MONOID", (void *)_cffi_var_GxB_ANY_FP64_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_ANY_GE_BOOL", (void *)_cffi_var_GxB_ANY_GE_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_GE_FP32", (void *)_cffi_var_GxB_ANY_GE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_GE_FP64", (void *)_cffi_var_GxB_ANY_GE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_GE_INT16", (void *)_cffi_var_GxB_ANY_GE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_GE_INT32", (void *)_cffi_var_GxB_ANY_GE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_GE_INT64", (void *)_cffi_var_GxB_ANY_GE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_GE_INT8", (void *)_cffi_var_GxB_ANY_GE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_GE_UINT16", (void *)_cffi_var_GxB_ANY_GE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_GE_UINT32", (void *)_cffi_var_GxB_ANY_GE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_GE_UINT64", (void *)_cffi_var_GxB_ANY_GE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_GE_UINT8", (void *)_cffi_var_GxB_ANY_GE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_GT_BOOL", (void *)_cffi_var_GxB_ANY_GT_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_GT_FP32", (void *)_cffi_var_GxB_ANY_GT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_GT_FP64", (void *)_cffi_var_GxB_ANY_GT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_GT_INT16", (void *)_cffi_var_GxB_ANY_GT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_GT_INT32", (void *)_cffi_var_GxB_ANY_GT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_GT_INT64", (void *)_cffi_var_GxB_ANY_GT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_GT_INT8", (void *)_cffi_var_GxB_ANY_GT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_GT_UINT16", (void *)_cffi_var_GxB_ANY_GT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_GT_UINT32", (void *)_cffi_var_GxB_ANY_GT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_GT_UINT64", (void *)_cffi_var_GxB_ANY_GT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_GT_UINT8", (void *)_cffi_var_GxB_ANY_GT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_INT16", (void *)_cffi_var_GxB_ANY_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ANY_INT16_MONOID", (void *)_cffi_var_GxB_ANY_INT16_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_ANY_INT32", (void *)_cffi_var_GxB_ANY_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ANY_INT32_MONOID", (void *)_cffi_var_GxB_ANY_INT32_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_ANY_INT64", (void *)_cffi_var_GxB_ANY_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ANY_INT64_MONOID", (void *)_cffi_var_GxB_ANY_INT64_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_ANY_INT8", (void *)_cffi_var_GxB_ANY_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ANY_INT8_MONOID", (void *)_cffi_var_GxB_ANY_INT8_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_ANY_ISEQ_FP32", (void *)_cffi_var_GxB_ANY_ISEQ_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISEQ_FP64", (void *)_cffi_var_GxB_ANY_ISEQ_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISEQ_INT16", (void *)_cffi_var_GxB_ANY_ISEQ_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISEQ_INT32", (void *)_cffi_var_GxB_ANY_ISEQ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISEQ_INT64", (void *)_cffi_var_GxB_ANY_ISEQ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISEQ_INT8", (void *)_cffi_var_GxB_ANY_ISEQ_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISEQ_UINT16", (void *)_cffi_var_GxB_ANY_ISEQ_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISEQ_UINT32", (void *)_cffi_var_GxB_ANY_ISEQ_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISEQ_UINT64", (void *)_cffi_var_GxB_ANY_ISEQ_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISEQ_UINT8", (void *)_cffi_var_GxB_ANY_ISEQ_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISGE_FP32", (void *)_cffi_var_GxB_ANY_ISGE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISGE_FP64", (void *)_cffi_var_GxB_ANY_ISGE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISGE_INT16", (void *)_cffi_var_GxB_ANY_ISGE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISGE_INT32", (void *)_cffi_var_GxB_ANY_ISGE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISGE_INT64", (void *)_cffi_var_GxB_ANY_ISGE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISGE_INT8", (void *)_cffi_var_GxB_ANY_ISGE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISGE_UINT16", (void *)_cffi_var_GxB_ANY_ISGE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISGE_UINT32", (void *)_cffi_var_GxB_ANY_ISGE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISGE_UINT64", (void *)_cffi_var_GxB_ANY_ISGE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISGE_UINT8", (void *)_cffi_var_GxB_ANY_ISGE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISGT_FP32", (void *)_cffi_var_GxB_ANY_ISGT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISGT_FP64", (void *)_cffi_var_GxB_ANY_ISGT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISGT_INT16", (void *)_cffi_var_GxB_ANY_ISGT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISGT_INT32", (void *)_cffi_var_GxB_ANY_ISGT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISGT_INT64", (void *)_cffi_var_GxB_ANY_ISGT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISGT_INT8", (void *)_cffi_var_GxB_ANY_ISGT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISGT_UINT16", (void *)_cffi_var_GxB_ANY_ISGT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISGT_UINT32", (void *)_cffi_var_GxB_ANY_ISGT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISGT_UINT64", (void *)_cffi_var_GxB_ANY_ISGT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISGT_UINT8", (void *)_cffi_var_GxB_ANY_ISGT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISLE_FP32", (void *)_cffi_var_GxB_ANY_ISLE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISLE_FP64", (void *)_cffi_var_GxB_ANY_ISLE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISLE_INT16", (void *)_cffi_var_GxB_ANY_ISLE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISLE_INT32", (void *)_cffi_var_GxB_ANY_ISLE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISLE_INT64", (void *)_cffi_var_GxB_ANY_ISLE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISLE_INT8", (void *)_cffi_var_GxB_ANY_ISLE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISLE_UINT16", (void *)_cffi_var_GxB_ANY_ISLE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISLE_UINT32", (void *)_cffi_var_GxB_ANY_ISLE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISLE_UINT64", (void *)_cffi_var_GxB_ANY_ISLE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISLE_UINT8", (void *)_cffi_var_GxB_ANY_ISLE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISLT_FP32", (void *)_cffi_var_GxB_ANY_ISLT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISLT_FP64", (void *)_cffi_var_GxB_ANY_ISLT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISLT_INT16", (void *)_cffi_var_GxB_ANY_ISLT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISLT_INT32", (void *)_cffi_var_GxB_ANY_ISLT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISLT_INT64", (void *)_cffi_var_GxB_ANY_ISLT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISLT_INT8", (void *)_cffi_var_GxB_ANY_ISLT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISLT_UINT16", (void *)_cffi_var_GxB_ANY_ISLT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISLT_UINT32", (void *)_cffi_var_GxB_ANY_ISLT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISLT_UINT64", (void *)_cffi_var_GxB_ANY_ISLT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISLT_UINT8", (void *)_cffi_var_GxB_ANY_ISLT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISNE_FP32", (void *)_cffi_var_GxB_ANY_ISNE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISNE_FP64", (void *)_cffi_var_GxB_ANY_ISNE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISNE_INT16", (void *)_cffi_var_GxB_ANY_ISNE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISNE_INT32", (void *)_cffi_var_GxB_ANY_ISNE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISNE_INT64", (void *)_cffi_var_GxB_ANY_ISNE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISNE_INT8", (void *)_cffi_var_GxB_ANY_ISNE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISNE_UINT16", (void *)_cffi_var_GxB_ANY_ISNE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISNE_UINT32", (void *)_cffi_var_GxB_ANY_ISNE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISNE_UINT64", (void *)_cffi_var_GxB_ANY_ISNE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_ISNE_UINT8", (void *)_cffi_var_GxB_ANY_ISNE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LAND_BOOL", (void *)_cffi_var_GxB_ANY_LAND_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LAND_FP32", (void *)_cffi_var_GxB_ANY_LAND_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LAND_FP64", (void *)_cffi_var_GxB_ANY_LAND_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LAND_INT16", (void *)_cffi_var_GxB_ANY_LAND_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LAND_INT32", (void *)_cffi_var_GxB_ANY_LAND_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LAND_INT64", (void *)_cffi_var_GxB_ANY_LAND_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LAND_INT8", (void *)_cffi_var_GxB_ANY_LAND_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LAND_UINT16", (void *)_cffi_var_GxB_ANY_LAND_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LAND_UINT32", (void *)_cffi_var_GxB_ANY_LAND_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LAND_UINT64", (void *)_cffi_var_GxB_ANY_LAND_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LAND_UINT8", (void *)_cffi_var_GxB_ANY_LAND_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LE_BOOL", (void *)_cffi_var_GxB_ANY_LE_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LE_FP32", (void *)_cffi_var_GxB_ANY_LE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LE_FP64", (void *)_cffi_var_GxB_ANY_LE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LE_INT16", (void *)_cffi_var_GxB_ANY_LE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LE_INT32", (void *)_cffi_var_GxB_ANY_LE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LE_INT64", (void *)_cffi_var_GxB_ANY_LE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LE_INT8", (void *)_cffi_var_GxB_ANY_LE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LE_UINT16", (void *)_cffi_var_GxB_ANY_LE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LE_UINT32", (void *)_cffi_var_GxB_ANY_LE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LE_UINT64", (void *)_cffi_var_GxB_ANY_LE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LE_UINT8", (void *)_cffi_var_GxB_ANY_LE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LOR_BOOL", (void *)_cffi_var_GxB_ANY_LOR_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LOR_FP32", (void *)_cffi_var_GxB_ANY_LOR_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LOR_FP64", (void *)_cffi_var_GxB_ANY_LOR_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LOR_INT16", (void *)_cffi_var_GxB_ANY_LOR_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LOR_INT32", (void *)_cffi_var_GxB_ANY_LOR_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LOR_INT64", (void *)_cffi_var_GxB_ANY_LOR_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LOR_INT8", (void *)_cffi_var_GxB_ANY_LOR_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LOR_UINT16", (void *)_cffi_var_GxB_ANY_LOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LOR_UINT32", (void *)_cffi_var_GxB_ANY_LOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LOR_UINT64", (void *)_cffi_var_GxB_ANY_LOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LOR_UINT8", (void *)_cffi_var_GxB_ANY_LOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LT_BOOL", (void *)_cffi_var_GxB_ANY_LT_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LT_FP32", (void *)_cffi_var_GxB_ANY_LT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LT_FP64", (void *)_cffi_var_GxB_ANY_LT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LT_INT16", (void *)_cffi_var_GxB_ANY_LT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LT_INT32", (void *)_cffi_var_GxB_ANY_LT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LT_INT64", (void *)_cffi_var_GxB_ANY_LT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LT_INT8", (void *)_cffi_var_GxB_ANY_LT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LT_UINT16", (void *)_cffi_var_GxB_ANY_LT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LT_UINT32", (void *)_cffi_var_GxB_ANY_LT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LT_UINT64", (void *)_cffi_var_GxB_ANY_LT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LT_UINT8", (void *)_cffi_var_GxB_ANY_LT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LXOR_BOOL", (void *)_cffi_var_GxB_ANY_LXOR_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LXOR_FP32", (void *)_cffi_var_GxB_ANY_LXOR_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LXOR_FP64", (void *)_cffi_var_GxB_ANY_LXOR_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LXOR_INT16", (void *)_cffi_var_GxB_ANY_LXOR_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LXOR_INT32", (void *)_cffi_var_GxB_ANY_LXOR_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LXOR_INT64", (void *)_cffi_var_GxB_ANY_LXOR_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LXOR_INT8", (void *)_cffi_var_GxB_ANY_LXOR_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LXOR_UINT16", (void *)_cffi_var_GxB_ANY_LXOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LXOR_UINT32", (void *)_cffi_var_GxB_ANY_LXOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LXOR_UINT64", (void *)_cffi_var_GxB_ANY_LXOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_LXOR_UINT8", (void *)_cffi_var_GxB_ANY_LXOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MAX_FP32", (void *)_cffi_var_GxB_ANY_MAX_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MAX_FP64", (void *)_cffi_var_GxB_ANY_MAX_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MAX_INT16", (void *)_cffi_var_GxB_ANY_MAX_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MAX_INT32", (void *)_cffi_var_GxB_ANY_MAX_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MAX_INT64", (void *)_cffi_var_GxB_ANY_MAX_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MAX_INT8", (void *)_cffi_var_GxB_ANY_MAX_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MAX_UINT16", (void *)_cffi_var_GxB_ANY_MAX_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MAX_UINT32", (void *)_cffi_var_GxB_ANY_MAX_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MAX_UINT64", (void *)_cffi_var_GxB_ANY_MAX_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MAX_UINT8", (void *)_cffi_var_GxB_ANY_MAX_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MINUS_FC32", (void *)_cffi_var_GxB_ANY_MINUS_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MINUS_FC64", (void *)_cffi_var_GxB_ANY_MINUS_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MINUS_FP32", (void *)_cffi_var_GxB_ANY_MINUS_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MINUS_FP64", (void *)_cffi_var_GxB_ANY_MINUS_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MINUS_INT16", (void *)_cffi_var_GxB_ANY_MINUS_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MINUS_INT32", (void *)_cffi_var_GxB_ANY_MINUS_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MINUS_INT64", (void *)_cffi_var_GxB_ANY_MINUS_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MINUS_INT8", (void *)_cffi_var_GxB_ANY_MINUS_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MINUS_UINT16", (void *)_cffi_var_GxB_ANY_MINUS_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MINUS_UINT32", (void *)_cffi_var_GxB_ANY_MINUS_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MINUS_UINT64", (void *)_cffi_var_GxB_ANY_MINUS_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MINUS_UINT8", (void *)_cffi_var_GxB_ANY_MINUS_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MIN_FP32", (void *)_cffi_var_GxB_ANY_MIN_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MIN_FP64", (void *)_cffi_var_GxB_ANY_MIN_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MIN_INT16", (void *)_cffi_var_GxB_ANY_MIN_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MIN_INT32", (void *)_cffi_var_GxB_ANY_MIN_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MIN_INT64", (void *)_cffi_var_GxB_ANY_MIN_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MIN_INT8", (void *)_cffi_var_GxB_ANY_MIN_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MIN_UINT16", (void *)_cffi_var_GxB_ANY_MIN_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MIN_UINT32", (void *)_cffi_var_GxB_ANY_MIN_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MIN_UINT64", (void *)_cffi_var_GxB_ANY_MIN_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_MIN_UINT8", (void *)_cffi_var_GxB_ANY_MIN_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_NE_FP32", (void *)_cffi_var_GxB_ANY_NE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_NE_FP64", (void *)_cffi_var_GxB_ANY_NE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_NE_INT16", (void *)_cffi_var_GxB_ANY_NE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_NE_INT32", (void *)_cffi_var_GxB_ANY_NE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_NE_INT64", (void *)_cffi_var_GxB_ANY_NE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_NE_INT8", (void *)_cffi_var_GxB_ANY_NE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_NE_UINT16", (void *)_cffi_var_GxB_ANY_NE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_NE_UINT32", (void *)_cffi_var_GxB_ANY_NE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_NE_UINT64", (void *)_cffi_var_GxB_ANY_NE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_NE_UINT8", (void *)_cffi_var_GxB_ANY_NE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PAIR_BOOL", (void *)_cffi_var_GxB_ANY_PAIR_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PAIR_FC32", (void *)_cffi_var_GxB_ANY_PAIR_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PAIR_FC64", (void *)_cffi_var_GxB_ANY_PAIR_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PAIR_FP32", (void *)_cffi_var_GxB_ANY_PAIR_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PAIR_FP64", (void *)_cffi_var_GxB_ANY_PAIR_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PAIR_INT16", (void *)_cffi_var_GxB_ANY_PAIR_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PAIR_INT32", (void *)_cffi_var_GxB_ANY_PAIR_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PAIR_INT64", (void *)_cffi_var_GxB_ANY_PAIR_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PAIR_INT8", (void *)_cffi_var_GxB_ANY_PAIR_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PAIR_UINT16", (void *)_cffi_var_GxB_ANY_PAIR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PAIR_UINT32", (void *)_cffi_var_GxB_ANY_PAIR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PAIR_UINT64", (void *)_cffi_var_GxB_ANY_PAIR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PAIR_UINT8", (void *)_cffi_var_GxB_ANY_PAIR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PLUS_FC32", (void *)_cffi_var_GxB_ANY_PLUS_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PLUS_FC64", (void *)_cffi_var_GxB_ANY_PLUS_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PLUS_FP32", (void *)_cffi_var_GxB_ANY_PLUS_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PLUS_FP64", (void *)_cffi_var_GxB_ANY_PLUS_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PLUS_INT16", (void *)_cffi_var_GxB_ANY_PLUS_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PLUS_INT32", (void *)_cffi_var_GxB_ANY_PLUS_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PLUS_INT64", (void *)_cffi_var_GxB_ANY_PLUS_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PLUS_INT8", (void *)_cffi_var_GxB_ANY_PLUS_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PLUS_UINT16", (void *)_cffi_var_GxB_ANY_PLUS_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PLUS_UINT32", (void *)_cffi_var_GxB_ANY_PLUS_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PLUS_UINT64", (void *)_cffi_var_GxB_ANY_PLUS_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_PLUS_UINT8", (void *)_cffi_var_GxB_ANY_PLUS_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RDIV_FC32", (void *)_cffi_var_GxB_ANY_RDIV_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RDIV_FC64", (void *)_cffi_var_GxB_ANY_RDIV_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RDIV_FP32", (void *)_cffi_var_GxB_ANY_RDIV_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RDIV_FP64", (void *)_cffi_var_GxB_ANY_RDIV_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RDIV_INT16", (void *)_cffi_var_GxB_ANY_RDIV_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RDIV_INT32", (void *)_cffi_var_GxB_ANY_RDIV_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RDIV_INT64", (void *)_cffi_var_GxB_ANY_RDIV_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RDIV_INT8", (void *)_cffi_var_GxB_ANY_RDIV_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RDIV_UINT16", (void *)_cffi_var_GxB_ANY_RDIV_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RDIV_UINT32", (void *)_cffi_var_GxB_ANY_RDIV_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RDIV_UINT64", (void *)_cffi_var_GxB_ANY_RDIV_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RDIV_UINT8", (void *)_cffi_var_GxB_ANY_RDIV_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RMINUS_FC32", (void *)_cffi_var_GxB_ANY_RMINUS_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RMINUS_FC64", (void *)_cffi_var_GxB_ANY_RMINUS_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RMINUS_FP32", (void *)_cffi_var_GxB_ANY_RMINUS_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RMINUS_FP64", (void *)_cffi_var_GxB_ANY_RMINUS_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RMINUS_INT16", (void *)_cffi_var_GxB_ANY_RMINUS_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RMINUS_INT32", (void *)_cffi_var_GxB_ANY_RMINUS_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RMINUS_INT64", (void *)_cffi_var_GxB_ANY_RMINUS_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RMINUS_INT8", (void *)_cffi_var_GxB_ANY_RMINUS_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RMINUS_UINT16", (void *)_cffi_var_GxB_ANY_RMINUS_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RMINUS_UINT32", (void *)_cffi_var_GxB_ANY_RMINUS_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RMINUS_UINT64", (void *)_cffi_var_GxB_ANY_RMINUS_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_RMINUS_UINT8", (void *)_cffi_var_GxB_ANY_RMINUS_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_SECONDI1_INT32", (void *)_cffi_var_GxB_ANY_SECONDI1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_SECONDI1_INT64", (void *)_cffi_var_GxB_ANY_SECONDI1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_SECONDI_INT32", (void *)_cffi_var_GxB_ANY_SECONDI_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_SECONDI_INT64", (void *)_cffi_var_GxB_ANY_SECONDI_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_SECONDJ1_INT32", (void *)_cffi_var_GxB_ANY_SECONDJ1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_SECONDJ1_INT64", (void *)_cffi_var_GxB_ANY_SECONDJ1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_SECONDJ_INT32", (void *)_cffi_var_GxB_ANY_SECONDJ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_SECONDJ_INT64", (void *)_cffi_var_GxB_ANY_SECONDJ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_SECOND_BOOL", (void *)_cffi_var_GxB_ANY_SECOND_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_SECOND_FC32", (void *)_cffi_var_GxB_ANY_SECOND_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_SECOND_FC64", (void *)_cffi_var_GxB_ANY_SECOND_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_SECOND_FP32", (void *)_cffi_var_GxB_ANY_SECOND_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_SECOND_FP64", (void *)_cffi_var_GxB_ANY_SECOND_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_SECOND_INT16", (void *)_cffi_var_GxB_ANY_SECOND_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_SECOND_INT32", (void *)_cffi_var_GxB_ANY_SECOND_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_SECOND_INT64", (void *)_cffi_var_GxB_ANY_SECOND_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_SECOND_INT8", (void *)_cffi_var_GxB_ANY_SECOND_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_SECOND_UINT16", (void *)_cffi_var_GxB_ANY_SECOND_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_SECOND_UINT32", (void *)_cffi_var_GxB_ANY_SECOND_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_SECOND_UINT64", (void *)_cffi_var_GxB_ANY_SECOND_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_SECOND_UINT8", (void *)_cffi_var_GxB_ANY_SECOND_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_SPARSITY", (void *)_cffi_const_GxB_ANY_SPARSITY, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_ANY_TIMES_FC32", (void *)_cffi_var_GxB_ANY_TIMES_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_TIMES_FC64", (void *)_cffi_var_GxB_ANY_TIMES_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_TIMES_FP32", (void *)_cffi_var_GxB_ANY_TIMES_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_TIMES_FP64", (void *)_cffi_var_GxB_ANY_TIMES_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_TIMES_INT16", (void *)_cffi_var_GxB_ANY_TIMES_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_TIMES_INT32", (void *)_cffi_var_GxB_ANY_TIMES_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_TIMES_INT64", (void *)_cffi_var_GxB_ANY_TIMES_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_TIMES_INT8", (void *)_cffi_var_GxB_ANY_TIMES_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_TIMES_UINT16", (void *)_cffi_var_GxB_ANY_TIMES_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_TIMES_UINT32", (void *)_cffi_var_GxB_ANY_TIMES_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_TIMES_UINT64", (void *)_cffi_var_GxB_ANY_TIMES_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_TIMES_UINT8", (void *)_cffi_var_GxB_ANY_TIMES_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_ANY_UINT16", (void *)_cffi_var_GxB_ANY_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ANY_UINT16_MONOID", (void *)_cffi_var_GxB_ANY_UINT16_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_ANY_UINT32", (void *)_cffi_var_GxB_ANY_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ANY_UINT32_MONOID", (void *)_cffi_var_GxB_ANY_UINT32_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_ANY_UINT64", (void *)_cffi_var_GxB_ANY_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ANY_UINT64_MONOID", (void *)_cffi_var_GxB_ANY_UINT64_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_ANY_UINT8", (void *)_cffi_var_GxB_ANY_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ANY_UINT8_MONOID", (void *)_cffi_var_GxB_ANY_UINT8_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_API_ABOUT", (void *)_cffi_const_GxB_API_ABOUT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_API_DATE", (void *)_cffi_const_GxB_API_DATE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_API_URL", (void *)_cffi_const_GxB_API_URL, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_API_VERSION", (void *)_cffi_const_GxB_API_VERSION, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_ASINH_FC32", (void *)_cffi_var_GxB_ASINH_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ASINH_FC64", (void *)_cffi_var_GxB_ASINH_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ASINH_FP32", (void *)_cffi_var_GxB_ASINH_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ASINH_FP64", (void *)_cffi_var_GxB_ASINH_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ASIN_FC32", (void *)_cffi_var_GxB_ASIN_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ASIN_FC64", (void *)_cffi_var_GxB_ASIN_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ASIN_FP32", (void *)_cffi_var_GxB_ASIN_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ASIN_FP64", (void *)_cffi_var_GxB_ASIN_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ATAN2_FP32", (void *)_cffi_var_GxB_ATAN2_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ATAN2_FP64", (void *)_cffi_var_GxB_ATAN2_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ATANH_FC32", (void *)_cffi_var_GxB_ATANH_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ATANH_FC64", (void *)_cffi_var_GxB_ATANH_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ATANH_FP32", (void *)_cffi_var_GxB_ATANH_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ATANH_FP64", (void *)_cffi_var_GxB_ATANH_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ATAN_FC32", (void *)_cffi_var_GxB_ATAN_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ATAN_FC64", (void *)_cffi_var_GxB_ATAN_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ATAN_FP32", (void *)_cffi_var_GxB_ATAN_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ATAN_FP64", (void *)_cffi_var_GxB_ATAN_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_AUTO_SPARSITY", (void *)_cffi_const_GxB_AUTO_SPARSITY, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_AxB_DOT", (void *)_cffi_const_GxB_AxB_DOT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_AxB_GUSTAVSON", (void *)_cffi_const_GxB_AxB_GUSTAVSON, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_AxB_HASH", (void *)_cffi_const_GxB_AxB_HASH, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_AxB_METHOD", (void *)_cffi_const_GxB_AxB_METHOD, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_AxB_SAXPY", (void *)_cffi_const_GxB_AxB_SAXPY, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_BACKWARDS", (void *)_cffi_const_GxB_BACKWARDS, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_BAND_BAND_UINT16", (void *)_cffi_var_GxB_BAND_BAND_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BAND_BAND_UINT32", (void *)_cffi_var_GxB_BAND_BAND_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BAND_BAND_UINT64", (void *)_cffi_var_GxB_BAND_BAND_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BAND_BAND_UINT8", (void *)_cffi_var_GxB_BAND_BAND_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BAND_BOR_UINT16", (void *)_cffi_var_GxB_BAND_BOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BAND_BOR_UINT32", (void *)_cffi_var_GxB_BAND_BOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BAND_BOR_UINT64", (void *)_cffi_var_GxB_BAND_BOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BAND_BOR_UINT8", (void *)_cffi_var_GxB_BAND_BOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BAND_BXNOR_UINT16", (void *)_cffi_var_GxB_BAND_BXNOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BAND_BXNOR_UINT32", (void *)_cffi_var_GxB_BAND_BXNOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BAND_BXNOR_UINT64", (void *)_cffi_var_GxB_BAND_BXNOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BAND_BXNOR_UINT8", (void *)_cffi_var_GxB_BAND_BXNOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BAND_BXOR_UINT16", (void *)_cffi_var_GxB_BAND_BXOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BAND_BXOR_UINT32", (void *)_cffi_var_GxB_BAND_BXOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BAND_BXOR_UINT64", (void *)_cffi_var_GxB_BAND_BXOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BAND_BXOR_UINT8", (void *)_cffi_var_GxB_BAND_BXOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BAND_UINT16_MONOID", (void *)_cffi_var_GxB_BAND_UINT16_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_BAND_UINT32_MONOID", (void *)_cffi_var_GxB_BAND_UINT32_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_BAND_UINT64_MONOID", (void *)_cffi_var_GxB_BAND_UINT64_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_BAND_UINT8_MONOID", (void *)_cffi_var_GxB_BAND_UINT8_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_BCLR_INT16", (void *)_cffi_var_GxB_BCLR_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BCLR_INT32", (void *)_cffi_var_GxB_BCLR_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BCLR_INT64", (void *)_cffi_var_GxB_BCLR_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BCLR_INT8", (void *)_cffi_var_GxB_BCLR_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BCLR_UINT16", (void *)_cffi_var_GxB_BCLR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BCLR_UINT32", (void *)_cffi_var_GxB_BCLR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BCLR_UINT64", (void *)_cffi_var_GxB_BCLR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BCLR_UINT8", (void *)_cffi_var_GxB_BCLR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BEGIN", (void *)_cffi_const_GxB_BEGIN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_BGET_INT16", (void *)_cffi_var_GxB_BGET_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BGET_INT32", (void *)_cffi_var_GxB_BGET_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BGET_INT64", (void *)_cffi_var_GxB_BGET_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BGET_INT8", (void *)_cffi_var_GxB_BGET_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BGET_UINT16", (void *)_cffi_var_GxB_BGET_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BGET_UINT32", (void *)_cffi_var_GxB_BGET_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BGET_UINT64", (void *)_cffi_var_GxB_BGET_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BGET_UINT8", (void *)_cffi_var_GxB_BGET_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BITMAP", (void *)_cffi_const_GxB_BITMAP, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_BITMAP_SWITCH", (void *)_cffi_const_GxB_BITMAP_SWITCH, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_BLOCKING_GPU", (void *)_cffi_const_GxB_BLOCKING_GPU, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_BOR_BAND_UINT16", (void *)_cffi_var_GxB_BOR_BAND_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BOR_BAND_UINT32", (void *)_cffi_var_GxB_BOR_BAND_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BOR_BAND_UINT64", (void *)_cffi_var_GxB_BOR_BAND_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BOR_BAND_UINT8", (void *)_cffi_var_GxB_BOR_BAND_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BOR_BOR_UINT16", (void *)_cffi_var_GxB_BOR_BOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BOR_BOR_UINT32", (void *)_cffi_var_GxB_BOR_BOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BOR_BOR_UINT64", (void *)_cffi_var_GxB_BOR_BOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BOR_BOR_UINT8", (void *)_cffi_var_GxB_BOR_BOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BOR_BXNOR_UINT16", (void *)_cffi_var_GxB_BOR_BXNOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BOR_BXNOR_UINT32", (void *)_cffi_var_GxB_BOR_BXNOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BOR_BXNOR_UINT64", (void *)_cffi_var_GxB_BOR_BXNOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BOR_BXNOR_UINT8", (void *)_cffi_var_GxB_BOR_BXNOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BOR_BXOR_UINT16", (void *)_cffi_var_GxB_BOR_BXOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BOR_BXOR_UINT32", (void *)_cffi_var_GxB_BOR_BXOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BOR_BXOR_UINT64", (void *)_cffi_var_GxB_BOR_BXOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BOR_BXOR_UINT8", (void *)_cffi_var_GxB_BOR_BXOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BOR_UINT16_MONOID", (void *)_cffi_var_GxB_BOR_UINT16_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_BOR_UINT32_MONOID", (void *)_cffi_var_GxB_BOR_UINT32_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_BOR_UINT64_MONOID", (void *)_cffi_var_GxB_BOR_UINT64_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_BOR_UINT8_MONOID", (void *)_cffi_var_GxB_BOR_UINT8_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_BSET_INT16", (void *)_cffi_var_GxB_BSET_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BSET_INT32", (void *)_cffi_var_GxB_BSET_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BSET_INT64", (void *)_cffi_var_GxB_BSET_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BSET_INT8", (void *)_cffi_var_GxB_BSET_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BSET_UINT16", (void *)_cffi_var_GxB_BSET_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BSET_UINT32", (void *)_cffi_var_GxB_BSET_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BSET_UINT64", (void *)_cffi_var_GxB_BSET_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BSET_UINT8", (void *)_cffi_var_GxB_BSET_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BSHIFT_INT16", (void *)_cffi_var_GxB_BSHIFT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BSHIFT_INT32", (void *)_cffi_var_GxB_BSHIFT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BSHIFT_INT64", (void *)_cffi_var_GxB_BSHIFT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BSHIFT_INT8", (void *)_cffi_var_GxB_BSHIFT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BSHIFT_UINT16", (void *)_cffi_var_GxB_BSHIFT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BSHIFT_UINT32", (void *)_cffi_var_GxB_BSHIFT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BSHIFT_UINT64", (void *)_cffi_var_GxB_BSHIFT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BSHIFT_UINT8", (void *)_cffi_var_GxB_BSHIFT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_BURBLE", (void *)_cffi_const_GxB_BURBLE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_BXNOR_BAND_UINT16", (void *)_cffi_var_GxB_BXNOR_BAND_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXNOR_BAND_UINT32", (void *)_cffi_var_GxB_BXNOR_BAND_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXNOR_BAND_UINT64", (void *)_cffi_var_GxB_BXNOR_BAND_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXNOR_BAND_UINT8", (void *)_cffi_var_GxB_BXNOR_BAND_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXNOR_BOR_UINT16", (void *)_cffi_var_GxB_BXNOR_BOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXNOR_BOR_UINT32", (void *)_cffi_var_GxB_BXNOR_BOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXNOR_BOR_UINT64", (void *)_cffi_var_GxB_BXNOR_BOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXNOR_BOR_UINT8", (void *)_cffi_var_GxB_BXNOR_BOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXNOR_BXNOR_UINT16", (void *)_cffi_var_GxB_BXNOR_BXNOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXNOR_BXNOR_UINT32", (void *)_cffi_var_GxB_BXNOR_BXNOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXNOR_BXNOR_UINT64", (void *)_cffi_var_GxB_BXNOR_BXNOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXNOR_BXNOR_UINT8", (void *)_cffi_var_GxB_BXNOR_BXNOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXNOR_BXOR_UINT16", (void *)_cffi_var_GxB_BXNOR_BXOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXNOR_BXOR_UINT32", (void *)_cffi_var_GxB_BXNOR_BXOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXNOR_BXOR_UINT64", (void *)_cffi_var_GxB_BXNOR_BXOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXNOR_BXOR_UINT8", (void *)_cffi_var_GxB_BXNOR_BXOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXNOR_UINT16_MONOID", (void *)_cffi_var_GxB_BXNOR_UINT16_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_BXNOR_UINT32_MONOID", (void *)_cffi_var_GxB_BXNOR_UINT32_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_BXNOR_UINT64_MONOID", (void *)_cffi_var_GxB_BXNOR_UINT64_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_BXNOR_UINT8_MONOID", (void *)_cffi_var_GxB_BXNOR_UINT8_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_BXOR_BAND_UINT16", (void *)_cffi_var_GxB_BXOR_BAND_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXOR_BAND_UINT32", (void *)_cffi_var_GxB_BXOR_BAND_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXOR_BAND_UINT64", (void *)_cffi_var_GxB_BXOR_BAND_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXOR_BAND_UINT8", (void *)_cffi_var_GxB_BXOR_BAND_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXOR_BOR_UINT16", (void *)_cffi_var_GxB_BXOR_BOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXOR_BOR_UINT32", (void *)_cffi_var_GxB_BXOR_BOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXOR_BOR_UINT64", (void *)_cffi_var_GxB_BXOR_BOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXOR_BOR_UINT8", (void *)_cffi_var_GxB_BXOR_BOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXOR_BXNOR_UINT16", (void *)_cffi_var_GxB_BXOR_BXNOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXOR_BXNOR_UINT32", (void *)_cffi_var_GxB_BXOR_BXNOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXOR_BXNOR_UINT64", (void *)_cffi_var_GxB_BXOR_BXNOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXOR_BXNOR_UINT8", (void *)_cffi_var_GxB_BXOR_BXNOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXOR_BXOR_UINT16", (void *)_cffi_var_GxB_BXOR_BXOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXOR_BXOR_UINT32", (void *)_cffi_var_GxB_BXOR_BXOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXOR_BXOR_UINT64", (void *)_cffi_var_GxB_BXOR_BXOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXOR_BXOR_UINT8", (void *)_cffi_var_GxB_BXOR_BXOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_BXOR_UINT16_MONOID", (void *)_cffi_var_GxB_BXOR_UINT16_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_BXOR_UINT32_MONOID", (void *)_cffi_var_GxB_BXOR_UINT32_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_BXOR_UINT64_MONOID", (void *)_cffi_var_GxB_BXOR_UINT64_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_BXOR_UINT8_MONOID", (void *)_cffi_var_GxB_BXOR_UINT8_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_BY_COL", (void *)_cffi_const_GxB_BY_COL, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_BY_ROW", (void *)_cffi_const_GxB_BY_ROW, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_BinaryOp_fprint", (void *)_cffi_f_GxB_BinaryOp_fprint, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 460), (void *)_cffi_d_GxB_BinaryOp_fprint },
  { "GxB_BinaryOp_new", (void *)_cffi_f_GxB_BinaryOp_new, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 447), (void *)_cffi_d_GxB_BinaryOp_new },
  { "GxB_BinaryOp_xtype", (void *)_cffi_f_GxB_BinaryOp_xtype, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2293), (void *)_cffi_d_GxB_BinaryOp_xtype },
  { "GxB_BinaryOp_xtype_name", (void *)_cffi_f_GxB_BinaryOp_xtype_name, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 96), (void *)_cffi_d_GxB_BinaryOp_xtype_name },
  { "GxB_BinaryOp_ytype", (void *)_cffi_f_GxB_BinaryOp_ytype, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2293), (void *)_cffi_d_GxB_BinaryOp_ytype },
  { "GxB_BinaryOp_ytype_name", (void *)_cffi_f_GxB_BinaryOp_ytype_name, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 96), (void *)_cffi_d_GxB_BinaryOp_ytype_name },
  { "GxB_BinaryOp_ztype", (void *)_cffi_f_GxB_BinaryOp_ztype, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2293), (void *)_cffi_d_GxB_BinaryOp_ztype },
  { "GxB_BinaryOp_ztype_name", (void *)_cffi_f_GxB_BinaryOp_ztype_name, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 96), (void *)_cffi_d_GxB_BinaryOp_ztype_name },
  { "GxB_CARG_FC32", (void *)_cffi_var_GxB_CARG_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_CARG_FC64", (void *)_cffi_var_GxB_CARG_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_CBRT_FP32", (void *)_cffi_var_GxB_CBRT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_CBRT_FP64", (void *)_cffi_var_GxB_CBRT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_CEIL_FC32", (void *)_cffi_var_GxB_CEIL_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_CEIL_FC64", (void *)_cffi_var_GxB_CEIL_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_CEIL_FP32", (void *)_cffi_var_GxB_CEIL_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_CEIL_FP64", (void *)_cffi_var_GxB_CEIL_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_CHUNK", (void *)_cffi_const_GxB_CHUNK, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_CIMAG_FC32", (void *)_cffi_var_GxB_CIMAG_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_CIMAG_FC64", (void *)_cffi_var_GxB_CIMAG_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_CMPLX_FP32", (void *)_cffi_var_GxB_CMPLX_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_CMPLX_FP64", (void *)_cffi_var_GxB_CMPLX_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_COMPILER_NAME", (void *)_cffi_const_GxB_COMPILER_NAME, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_COMPILER_VERSION", (void *)_cffi_const_GxB_COMPILER_VERSION, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_COMPLETE", (void *)_cffi_const_GxB_COMPLETE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_COMPLETE_VERBOSE", (void *)_cffi_const_GxB_COMPLETE_VERBOSE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_COMPRESSION", (void *)_cffi_const_GxB_COMPRESSION, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_COMPRESSION_DEFAULT", (void *)_cffi_const_GxB_COMPRESSION_DEFAULT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_COMPRESSION_INTEL", (void *)_cffi_const_GxB_COMPRESSION_INTEL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_COMPRESSION_LZ4", (void *)_cffi_const_GxB_COMPRESSION_LZ4, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_COMPRESSION_LZ4HC", (void *)_cffi_const_GxB_COMPRESSION_LZ4HC, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_COMPRESSION_NONE", (void *)_cffi_const_GxB_COMPRESSION_NONE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_COMPRESSION_ZSTD", (void *)_cffi_const_GxB_COMPRESSION_ZSTD, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_CONJ_FC32", (void *)_cffi_var_GxB_CONJ_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_CONJ_FC64", (void *)_cffi_var_GxB_CONJ_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_COPYSIGN_FP32", (void *)_cffi_var_GxB_COPYSIGN_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_COPYSIGN_FP64", (void *)_cffi_var_GxB_COPYSIGN_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_COSH_FC32", (void *)_cffi_var_GxB_COSH_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_COSH_FC64", (void *)_cffi_var_GxB_COSH_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_COSH_FP32", (void *)_cffi_var_GxB_COSH_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_COSH_FP64", (void *)_cffi_var_GxB_COSH_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_COS_FC32", (void *)_cffi_var_GxB_COS_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_COS_FC64", (void *)_cffi_var_GxB_COS_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_COS_FP32", (void *)_cffi_var_GxB_COS_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_COS_FP64", (void *)_cffi_var_GxB_COS_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_CREAL_FC32", (void *)_cffi_var_GxB_CREAL_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_CREAL_FC64", (void *)_cffi_var_GxB_CREAL_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_Col_subassign", (void *)_cffi_f_GxB_Col_subassign, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1695), (void *)_cffi_d_GxB_Col_subassign },
  { "GxB_DEFAULT", (void *)_cffi_const_GxB_DEFAULT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_DESCRIPTOR_CHUNK", (void *)_cffi_const_GxB_DESCRIPTOR_CHUNK, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_DESCRIPTOR_GPU_CHUNK", (void *)_cffi_const_GxB_DESCRIPTOR_GPU_CHUNK, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_DESCRIPTOR_GPU_CONTROL", (void *)_cffi_const_GxB_DESCRIPTOR_GPU_CONTROL, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_DESCRIPTOR_NTHREADS", (void *)_cffi_const_GxB_DESCRIPTOR_NTHREADS, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_DIAG", (void *)_cffi_var_GxB_DIAG, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 155), (void *)0 },
  { "GxB_DIV_FC32", (void *)_cffi_var_GxB_DIV_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_DIV_FC64", (void *)_cffi_var_GxB_DIV_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_Desc_get", (void *)_cffi_const_GxB_Desc_get, _CFFI_OP(_CFFI_OP_CONSTANT, 3980), (void *)0 },
  { "GxB_Desc_get_FP64", (void *)_cffi_f_GxB_Desc_get_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 478), (void *)_cffi_d_GxB_Desc_get_FP64 },
  { "GxB_Desc_get_INT32", (void *)_cffi_f_GxB_Desc_get_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 488), (void *)_cffi_d_GxB_Desc_get_INT32 },
  { "GxB_Desc_set", (void *)_cffi_const_GxB_Desc_set, _CFFI_OP(_CFFI_OP_CONSTANT, 3980), (void *)0 },
  { "GxB_Desc_set_FP64", (void *)_cffi_f_GxB_Desc_set_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 483), (void *)_cffi_d_GxB_Desc_set_FP64 },
  { "GxB_Desc_set_INT32", (void *)_cffi_f_GxB_Desc_set_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 493), (void *)_cffi_d_GxB_Desc_set_INT32 },
  { "GxB_Descriptor_fprint", (void *)_cffi_f_GxB_Descriptor_fprint, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 502), (void *)_cffi_d_GxB_Descriptor_fprint },
  { "GxB_Descriptor_get", (void *)_cffi_f_GxB_Descriptor_get, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 0), (void *)_cffi_d_GxB_Descriptor_get },
  { "GxB_END", (void *)_cffi_const_GxB_END, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_EQ_BOOL_MONOID", (void *)_cffi_var_GxB_EQ_BOOL_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_EQ_EQ_BOOL", (void *)_cffi_var_GxB_EQ_EQ_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_EQ_FP32", (void *)_cffi_var_GxB_EQ_EQ_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_EQ_FP64", (void *)_cffi_var_GxB_EQ_EQ_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_EQ_INT16", (void *)_cffi_var_GxB_EQ_EQ_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_EQ_INT32", (void *)_cffi_var_GxB_EQ_EQ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_EQ_INT64", (void *)_cffi_var_GxB_EQ_EQ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_EQ_INT8", (void *)_cffi_var_GxB_EQ_EQ_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_EQ_UINT16", (void *)_cffi_var_GxB_EQ_EQ_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_EQ_UINT32", (void *)_cffi_var_GxB_EQ_EQ_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_EQ_UINT64", (void *)_cffi_var_GxB_EQ_EQ_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_EQ_UINT8", (void *)_cffi_var_GxB_EQ_EQ_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_FC32", (void *)_cffi_var_GxB_EQ_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_EQ_FC64", (void *)_cffi_var_GxB_EQ_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_EQ_FIRST_BOOL", (void *)_cffi_var_GxB_EQ_FIRST_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_GE_BOOL", (void *)_cffi_var_GxB_EQ_GE_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_GE_FP32", (void *)_cffi_var_GxB_EQ_GE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_GE_FP64", (void *)_cffi_var_GxB_EQ_GE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_GE_INT16", (void *)_cffi_var_GxB_EQ_GE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_GE_INT32", (void *)_cffi_var_GxB_EQ_GE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_GE_INT64", (void *)_cffi_var_GxB_EQ_GE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_GE_INT8", (void *)_cffi_var_GxB_EQ_GE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_GE_UINT16", (void *)_cffi_var_GxB_EQ_GE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_GE_UINT32", (void *)_cffi_var_GxB_EQ_GE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_GE_UINT64", (void *)_cffi_var_GxB_EQ_GE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_GE_UINT8", (void *)_cffi_var_GxB_EQ_GE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_GT_BOOL", (void *)_cffi_var_GxB_EQ_GT_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_GT_FP32", (void *)_cffi_var_GxB_EQ_GT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_GT_FP64", (void *)_cffi_var_GxB_EQ_GT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_GT_INT16", (void *)_cffi_var_GxB_EQ_GT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_GT_INT32", (void *)_cffi_var_GxB_EQ_GT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_GT_INT64", (void *)_cffi_var_GxB_EQ_GT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_GT_INT8", (void *)_cffi_var_GxB_EQ_GT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_GT_UINT16", (void *)_cffi_var_GxB_EQ_GT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_GT_UINT32", (void *)_cffi_var_GxB_EQ_GT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_GT_UINT64", (void *)_cffi_var_GxB_EQ_GT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_GT_UINT8", (void *)_cffi_var_GxB_EQ_GT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LAND_BOOL", (void *)_cffi_var_GxB_EQ_LAND_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LE_BOOL", (void *)_cffi_var_GxB_EQ_LE_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LE_FP32", (void *)_cffi_var_GxB_EQ_LE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LE_FP64", (void *)_cffi_var_GxB_EQ_LE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LE_INT16", (void *)_cffi_var_GxB_EQ_LE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LE_INT32", (void *)_cffi_var_GxB_EQ_LE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LE_INT64", (void *)_cffi_var_GxB_EQ_LE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LE_INT8", (void *)_cffi_var_GxB_EQ_LE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LE_UINT16", (void *)_cffi_var_GxB_EQ_LE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LE_UINT32", (void *)_cffi_var_GxB_EQ_LE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LE_UINT64", (void *)_cffi_var_GxB_EQ_LE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LE_UINT8", (void *)_cffi_var_GxB_EQ_LE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LOR_BOOL", (void *)_cffi_var_GxB_EQ_LOR_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LT_BOOL", (void *)_cffi_var_GxB_EQ_LT_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LT_FP32", (void *)_cffi_var_GxB_EQ_LT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LT_FP64", (void *)_cffi_var_GxB_EQ_LT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LT_INT16", (void *)_cffi_var_GxB_EQ_LT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LT_INT32", (void *)_cffi_var_GxB_EQ_LT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LT_INT64", (void *)_cffi_var_GxB_EQ_LT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LT_INT8", (void *)_cffi_var_GxB_EQ_LT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LT_UINT16", (void *)_cffi_var_GxB_EQ_LT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LT_UINT32", (void *)_cffi_var_GxB_EQ_LT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LT_UINT64", (void *)_cffi_var_GxB_EQ_LT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LT_UINT8", (void *)_cffi_var_GxB_EQ_LT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_LXOR_BOOL", (void *)_cffi_var_GxB_EQ_LXOR_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_NE_FP32", (void *)_cffi_var_GxB_EQ_NE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_NE_FP64", (void *)_cffi_var_GxB_EQ_NE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_NE_INT16", (void *)_cffi_var_GxB_EQ_NE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_NE_INT32", (void *)_cffi_var_GxB_EQ_NE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_NE_INT64", (void *)_cffi_var_GxB_EQ_NE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_NE_INT8", (void *)_cffi_var_GxB_EQ_NE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_NE_UINT16", (void *)_cffi_var_GxB_EQ_NE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_NE_UINT32", (void *)_cffi_var_GxB_EQ_NE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_NE_UINT64", (void *)_cffi_var_GxB_EQ_NE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_NE_UINT8", (void *)_cffi_var_GxB_EQ_NE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_PAIR_BOOL", (void *)_cffi_var_GxB_EQ_PAIR_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_SECOND_BOOL", (void *)_cffi_var_GxB_EQ_SECOND_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_EQ_THUNK", (void *)_cffi_var_GxB_EQ_THUNK, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 155), (void *)0 },
  { "GxB_EQ_ZERO", (void *)_cffi_var_GxB_EQ_ZERO, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 155), (void *)0 },
  { "GxB_ERFC_FP32", (void *)_cffi_var_GxB_ERFC_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ERFC_FP64", (void *)_cffi_var_GxB_ERFC_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ERF_FP32", (void *)_cffi_var_GxB_ERF_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ERF_FP64", (void *)_cffi_var_GxB_ERF_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_EXHAUSTED", (void *)_cffi_const_GxB_EXHAUSTED, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_EXP2_FC32", (void *)_cffi_var_GxB_EXP2_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_EXP2_FC64", (void *)_cffi_var_GxB_EXP2_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_EXP2_FP32", (void *)_cffi_var_GxB_EXP2_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_EXP2_FP64", (void *)_cffi_var_GxB_EXP2_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_EXPM1_FC32", (void *)_cffi_var_GxB_EXPM1_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_EXPM1_FC64", (void *)_cffi_var_GxB_EXPM1_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_EXPM1_FP32", (void *)_cffi_var_GxB_EXPM1_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_EXPM1_FP64", (void *)_cffi_var_GxB_EXPM1_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_EXP_FC32", (void *)_cffi_var_GxB_EXP_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_EXP_FC64", (void *)_cffi_var_GxB_EXP_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_EXP_FP32", (void *)_cffi_var_GxB_EXP_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_EXP_FP64", (void *)_cffi_var_GxB_EXP_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_FAST_IMPORT", (void *)_cffi_const_GxB_FAST_IMPORT, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_FC32", (void *)_cffi_var_GxB_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 114), (void *)0 },
  { "GxB_FC64", (void *)_cffi_var_GxB_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 114), (void *)0 },
  { "GxB_FIRSTI1_INT32", (void *)_cffi_var_GxB_FIRSTI1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_FIRSTI1_INT64", (void *)_cffi_var_GxB_FIRSTI1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_FIRSTI_INT32", (void *)_cffi_var_GxB_FIRSTI_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_FIRSTI_INT64", (void *)_cffi_var_GxB_FIRSTI_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_FIRSTJ1_INT32", (void *)_cffi_var_GxB_FIRSTJ1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_FIRSTJ1_INT64", (void *)_cffi_var_GxB_FIRSTJ1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_FIRSTJ_INT32", (void *)_cffi_var_GxB_FIRSTJ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_FIRSTJ_INT64", (void *)_cffi_var_GxB_FIRSTJ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_FIRST_FC32", (void *)_cffi_var_GxB_FIRST_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_FIRST_FC64", (void *)_cffi_var_GxB_FIRST_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_FLOOR_FC32", (void *)_cffi_var_GxB_FLOOR_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_FLOOR_FC64", (void *)_cffi_var_GxB_FLOOR_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_FLOOR_FP32", (void *)_cffi_var_GxB_FLOOR_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_FLOOR_FP64", (void *)_cffi_var_GxB_FLOOR_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_FLUSH", (void *)_cffi_const_GxB_FLUSH, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_FMOD_FP32", (void *)_cffi_var_GxB_FMOD_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_FMOD_FP64", (void *)_cffi_var_GxB_FMOD_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_FORMAT", (void *)_cffi_const_GxB_FORMAT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_FORMAT_DEFAULT", (void *)_cffi_const_GxB_FORMAT_DEFAULT, _CFFI_OP(_CFFI_OP_CONSTANT, 546), (void *)0 },
  { "GxB_FREXPE_FP32", (void *)_cffi_var_GxB_FREXPE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_FREXPE_FP64", (void *)_cffi_var_GxB_FREXPE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_FREXPX_FP32", (void *)_cffi_var_GxB_FREXPX_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_FREXPX_FP64", (void *)_cffi_var_GxB_FREXPX_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_FULL", (void *)_cffi_const_GxB_FULL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_GE_THUNK", (void *)_cffi_var_GxB_GE_THUNK, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 155), (void *)0 },
  { "GxB_GE_ZERO", (void *)_cffi_var_GxB_GE_ZERO, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 155), (void *)0 },
  { "GxB_GLOBAL_CHUNK", (void *)_cffi_const_GxB_GLOBAL_CHUNK, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_GLOBAL_GPU_CHUNK", (void *)_cffi_const_GxB_GLOBAL_GPU_CHUNK, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_GLOBAL_GPU_CONTROL", (void *)_cffi_const_GxB_GLOBAL_GPU_CONTROL, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_GLOBAL_NTHREADS", (void *)_cffi_const_GxB_GLOBAL_NTHREADS, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_GPU_ALWAYS", (void *)_cffi_const_GxB_GPU_ALWAYS, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_GPU_CHUNK", (void *)_cffi_const_GxB_GPU_CHUNK, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_GPU_CONTROL", (void *)_cffi_const_GxB_GPU_CONTROL, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_GPU_NEVER", (void *)_cffi_const_GxB_GPU_NEVER, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_GT_THUNK", (void *)_cffi_var_GxB_GT_THUNK, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 155), (void *)0 },
  { "GxB_GT_ZERO", (void *)_cffi_var_GxB_GT_ZERO, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 155), (void *)0 },
  { "GxB_Global_Option_get", (void *)_cffi_const_GxB_Global_Option_get, _CFFI_OP(_CFFI_OP_CONSTANT, 3979), (void *)0 },
  { "GxB_Global_Option_get_CHAR", (void *)_cffi_f_GxB_Global_Option_get_CHAR, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 22), (void *)_cffi_d_GxB_Global_Option_get_CHAR },
  { "GxB_Global_Option_get_FP64", (void *)_cffi_f_GxB_Global_Option_get_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 26), (void *)_cffi_d_GxB_Global_Option_get_FP64 },
  { "GxB_Global_Option_get_FUNCTION", (void *)_cffi_f_GxB_Global_Option_get_FUNCTION, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 46), (void *)_cffi_d_GxB_Global_Option_get_FUNCTION },
  { "GxB_Global_Option_get_INT32", (void *)_cffi_f_GxB_Global_Option_get_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 34), (void *)_cffi_d_GxB_Global_Option_get_INT32 },
  { "GxB_Global_Option_get_INT64", (void *)_cffi_f_GxB_Global_Option_get_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 42), (void *)_cffi_d_GxB_Global_Option_get_INT64 },
  { "GxB_Global_Option_set", (void *)_cffi_const_GxB_Global_Option_set, _CFFI_OP(_CFFI_OP_CONSTANT, 3979), (void *)0 },
  { "GxB_Global_Option_set_FP64", (void *)_cffi_f_GxB_Global_Option_set_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 30), (void *)_cffi_d_GxB_Global_Option_set_FP64 },
  { "GxB_Global_Option_set_FP64_ARRAY", (void *)_cffi_f_GxB_Global_Option_set_FP64_ARRAY, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 26), (void *)_cffi_d_GxB_Global_Option_set_FP64_ARRAY },
  { "GxB_Global_Option_set_FUNCTION", (void *)_cffi_f_GxB_Global_Option_set_FUNCTION, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 50), (void *)_cffi_d_GxB_Global_Option_set_FUNCTION },
  { "GxB_Global_Option_set_INT32", (void *)_cffi_f_GxB_Global_Option_set_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 38), (void *)_cffi_d_GxB_Global_Option_set_INT32 },
  { "GxB_Global_Option_set_INT64_ARRAY", (void *)_cffi_f_GxB_Global_Option_set_INT64_ARRAY, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 42), (void *)_cffi_d_GxB_Global_Option_set_INT64_ARRAY },
  { "GxB_HYPERSPARSE", (void *)_cffi_const_GxB_HYPERSPARSE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_HYPER_DEFAULT", (void *)_cffi_const_GxB_HYPER_DEFAULT, _CFFI_OP(_CFFI_OP_CONSTANT, 32), (void *)0 },
  { "GxB_HYPER_SWITCH", (void *)_cffi_const_GxB_HYPER_SWITCH, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_HYPOT_FP32", (void *)_cffi_var_GxB_HYPOT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_HYPOT_FP64", (void *)_cffi_var_GxB_HYPOT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_IDENTITY_FC32", (void *)_cffi_var_GxB_IDENTITY_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_IDENTITY_FC64", (void *)_cffi_var_GxB_IDENTITY_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_IGNORE_DUP", (void *)_cffi_var_GxB_IGNORE_DUP, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_IMPLEMENTATION", (void *)_cffi_const_GxB_IMPLEMENTATION, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_IMPLEMENTATION_ABOUT_STR", (void *)_cffi_var_GxB_IMPLEMENTATION_ABOUT_STR, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 97), (void *)0 },
  { "GxB_IMPLEMENTATION_DATE_STR", (void *)_cffi_var_GxB_IMPLEMENTATION_DATE_STR, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 97), (void *)0 },
  { "GxB_IMPLEMENTATION_LICENSE_STR", (void *)_cffi_var_GxB_IMPLEMENTATION_LICENSE_STR, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 97), (void *)0 },
  { "GxB_IMPLEMENTATION_MAJOR", (void *)_cffi_const_GxB_IMPLEMENTATION_MAJOR, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_IMPLEMENTATION_MINOR", (void *)_cffi_const_GxB_IMPLEMENTATION_MINOR, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_IMPLEMENTATION_NAME_STR", (void *)_cffi_var_GxB_IMPLEMENTATION_NAME_STR, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 97), (void *)0 },
  { "GxB_IMPLEMENTATION_SUB", (void *)_cffi_const_GxB_IMPLEMENTATION_SUB, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_IMPORT", (void *)_cffi_const_GxB_IMPORT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_INC", (void *)_cffi_const_GxB_INC, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_INDEX_MAX", (void *)_cffi_const_GxB_INDEX_MAX, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_ISEQ_BOOL", (void *)_cffi_var_GxB_ISEQ_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISEQ_FC32", (void *)_cffi_var_GxB_ISEQ_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISEQ_FC64", (void *)_cffi_var_GxB_ISEQ_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISEQ_FP32", (void *)_cffi_var_GxB_ISEQ_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISEQ_FP64", (void *)_cffi_var_GxB_ISEQ_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISEQ_INT16", (void *)_cffi_var_GxB_ISEQ_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISEQ_INT32", (void *)_cffi_var_GxB_ISEQ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISEQ_INT64", (void *)_cffi_var_GxB_ISEQ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISEQ_INT8", (void *)_cffi_var_GxB_ISEQ_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISEQ_UINT16", (void *)_cffi_var_GxB_ISEQ_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISEQ_UINT32", (void *)_cffi_var_GxB_ISEQ_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISEQ_UINT64", (void *)_cffi_var_GxB_ISEQ_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISEQ_UINT8", (void *)_cffi_var_GxB_ISEQ_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISFINITE_FC32", (void *)_cffi_var_GxB_ISFINITE_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ISFINITE_FC64", (void *)_cffi_var_GxB_ISFINITE_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ISFINITE_FP32", (void *)_cffi_var_GxB_ISFINITE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ISFINITE_FP64", (void *)_cffi_var_GxB_ISFINITE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ISGE_BOOL", (void *)_cffi_var_GxB_ISGE_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISGE_FP32", (void *)_cffi_var_GxB_ISGE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISGE_FP64", (void *)_cffi_var_GxB_ISGE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISGE_INT16", (void *)_cffi_var_GxB_ISGE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISGE_INT32", (void *)_cffi_var_GxB_ISGE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISGE_INT64", (void *)_cffi_var_GxB_ISGE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISGE_INT8", (void *)_cffi_var_GxB_ISGE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISGE_UINT16", (void *)_cffi_var_GxB_ISGE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISGE_UINT32", (void *)_cffi_var_GxB_ISGE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISGE_UINT64", (void *)_cffi_var_GxB_ISGE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISGE_UINT8", (void *)_cffi_var_GxB_ISGE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISGT_BOOL", (void *)_cffi_var_GxB_ISGT_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISGT_FP32", (void *)_cffi_var_GxB_ISGT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISGT_FP64", (void *)_cffi_var_GxB_ISGT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISGT_INT16", (void *)_cffi_var_GxB_ISGT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISGT_INT32", (void *)_cffi_var_GxB_ISGT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISGT_INT64", (void *)_cffi_var_GxB_ISGT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISGT_INT8", (void *)_cffi_var_GxB_ISGT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISGT_UINT16", (void *)_cffi_var_GxB_ISGT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISGT_UINT32", (void *)_cffi_var_GxB_ISGT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISGT_UINT64", (void *)_cffi_var_GxB_ISGT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISGT_UINT8", (void *)_cffi_var_GxB_ISGT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISINF_FC32", (void *)_cffi_var_GxB_ISINF_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ISINF_FC64", (void *)_cffi_var_GxB_ISINF_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ISINF_FP32", (void *)_cffi_var_GxB_ISINF_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ISINF_FP64", (void *)_cffi_var_GxB_ISINF_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ISLE_BOOL", (void *)_cffi_var_GxB_ISLE_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISLE_FP32", (void *)_cffi_var_GxB_ISLE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISLE_FP64", (void *)_cffi_var_GxB_ISLE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISLE_INT16", (void *)_cffi_var_GxB_ISLE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISLE_INT32", (void *)_cffi_var_GxB_ISLE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISLE_INT64", (void *)_cffi_var_GxB_ISLE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISLE_INT8", (void *)_cffi_var_GxB_ISLE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISLE_UINT16", (void *)_cffi_var_GxB_ISLE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISLE_UINT32", (void *)_cffi_var_GxB_ISLE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISLE_UINT64", (void *)_cffi_var_GxB_ISLE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISLE_UINT8", (void *)_cffi_var_GxB_ISLE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISLT_BOOL", (void *)_cffi_var_GxB_ISLT_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISLT_FP32", (void *)_cffi_var_GxB_ISLT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISLT_FP64", (void *)_cffi_var_GxB_ISLT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISLT_INT16", (void *)_cffi_var_GxB_ISLT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISLT_INT32", (void *)_cffi_var_GxB_ISLT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISLT_INT64", (void *)_cffi_var_GxB_ISLT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISLT_INT8", (void *)_cffi_var_GxB_ISLT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISLT_UINT16", (void *)_cffi_var_GxB_ISLT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISLT_UINT32", (void *)_cffi_var_GxB_ISLT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISLT_UINT64", (void *)_cffi_var_GxB_ISLT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISLT_UINT8", (void *)_cffi_var_GxB_ISLT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISNAN_FC32", (void *)_cffi_var_GxB_ISNAN_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ISNAN_FC64", (void *)_cffi_var_GxB_ISNAN_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ISNAN_FP32", (void *)_cffi_var_GxB_ISNAN_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ISNAN_FP64", (void *)_cffi_var_GxB_ISNAN_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ISNE_BOOL", (void *)_cffi_var_GxB_ISNE_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISNE_FC32", (void *)_cffi_var_GxB_ISNE_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISNE_FC64", (void *)_cffi_var_GxB_ISNE_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISNE_FP32", (void *)_cffi_var_GxB_ISNE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISNE_FP64", (void *)_cffi_var_GxB_ISNE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISNE_INT16", (void *)_cffi_var_GxB_ISNE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISNE_INT32", (void *)_cffi_var_GxB_ISNE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISNE_INT64", (void *)_cffi_var_GxB_ISNE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISNE_INT8", (void *)_cffi_var_GxB_ISNE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISNE_UINT16", (void *)_cffi_var_GxB_ISNE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISNE_UINT32", (void *)_cffi_var_GxB_ISNE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISNE_UINT64", (void *)_cffi_var_GxB_ISNE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ISNE_UINT8", (void *)_cffi_var_GxB_ISNE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_IndexUnaryOp_fprint", (void *)_cffi_f_GxB_IndexUnaryOp_fprint, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 531), (void *)_cffi_d_GxB_IndexUnaryOp_fprint },
  { "GxB_IndexUnaryOp_new", (void *)_cffi_f_GxB_IndexUnaryOp_new, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 518), (void *)_cffi_d_GxB_IndexUnaryOp_new },
  { "GxB_IndexUnaryOp_xtype_name", (void *)_cffi_f_GxB_IndexUnaryOp_xtype_name, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 100), (void *)_cffi_d_GxB_IndexUnaryOp_xtype_name },
  { "GxB_IndexUnaryOp_ytype_name", (void *)_cffi_f_GxB_IndexUnaryOp_ytype_name, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 100), (void *)_cffi_d_GxB_IndexUnaryOp_ytype_name },
  { "GxB_IndexUnaryOp_ztype_name", (void *)_cffi_f_GxB_IndexUnaryOp_ztype_name, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 100), (void *)_cffi_d_GxB_IndexUnaryOp_ztype_name },
  { "GxB_Iterator_free", (void *)_cffi_f_GxB_Iterator_free, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 537), (void *)_cffi_d_GxB_Iterator_free },
  { "GxB_Iterator_get_BOOL", (void *)_cffi_f_GxB_Iterator_get_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3894), (void *)_cffi_d_GxB_Iterator_get_BOOL },
  { "GxB_Iterator_get_FC32", (void *)_cffi_f_GxB_Iterator_get_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3909), (void *)_cffi_d_GxB_Iterator_get_FC32 },
  { "GxB_Iterator_get_FC64", (void *)_cffi_f_GxB_Iterator_get_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3903), (void *)_cffi_d_GxB_Iterator_get_FC64 },
  { "GxB_Iterator_get_FP32", (void *)_cffi_f_GxB_Iterator_get_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3912), (void *)_cffi_d_GxB_Iterator_get_FP32 },
  { "GxB_Iterator_get_FP64", (void *)_cffi_f_GxB_Iterator_get_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3906), (void *)_cffi_d_GxB_Iterator_get_FP64 },
  { "GxB_Iterator_get_INT16", (void *)_cffi_f_GxB_Iterator_get_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3915), (void *)_cffi_d_GxB_Iterator_get_INT16 },
  { "GxB_Iterator_get_INT32", (void *)_cffi_f_GxB_Iterator_get_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3918), (void *)_cffi_d_GxB_Iterator_get_INT32 },
  { "GxB_Iterator_get_INT64", (void *)_cffi_f_GxB_Iterator_get_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3921), (void *)_cffi_d_GxB_Iterator_get_INT64 },
  { "GxB_Iterator_get_INT8", (void *)_cffi_f_GxB_Iterator_get_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3924), (void *)_cffi_d_GxB_Iterator_get_INT8 },
  { "GxB_Iterator_get_UDT", (void *)_cffi_f_GxB_Iterator_get_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3955), (void *)_cffi_d_GxB_Iterator_get_UDT },
  { "GxB_Iterator_get_UINT16", (void *)_cffi_f_GxB_Iterator_get_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3927), (void *)_cffi_d_GxB_Iterator_get_UINT16 },
  { "GxB_Iterator_get_UINT32", (void *)_cffi_f_GxB_Iterator_get_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3930), (void *)_cffi_d_GxB_Iterator_get_UINT32 },
  { "GxB_Iterator_get_UINT64", (void *)_cffi_f_GxB_Iterator_get_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3933), (void *)_cffi_d_GxB_Iterator_get_UINT64 },
  { "GxB_Iterator_get_UINT8", (void *)_cffi_f_GxB_Iterator_get_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3936), (void *)_cffi_d_GxB_Iterator_get_UINT8 },
  { "GxB_Iterator_new", (void *)_cffi_f_GxB_Iterator_new, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 537), (void *)_cffi_d_GxB_Iterator_new },
  { "GxB_LAND_BOOL", (void *)_cffi_var_GxB_LAND_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LAND_EQ_BOOL", (void *)_cffi_var_GxB_LAND_EQ_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_EQ_FP32", (void *)_cffi_var_GxB_LAND_EQ_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_EQ_FP64", (void *)_cffi_var_GxB_LAND_EQ_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_EQ_INT16", (void *)_cffi_var_GxB_LAND_EQ_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_EQ_INT32", (void *)_cffi_var_GxB_LAND_EQ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_EQ_INT64", (void *)_cffi_var_GxB_LAND_EQ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_EQ_INT8", (void *)_cffi_var_GxB_LAND_EQ_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_EQ_UINT16", (void *)_cffi_var_GxB_LAND_EQ_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_EQ_UINT32", (void *)_cffi_var_GxB_LAND_EQ_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_EQ_UINT64", (void *)_cffi_var_GxB_LAND_EQ_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_EQ_UINT8", (void *)_cffi_var_GxB_LAND_EQ_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_FIRST_BOOL", (void *)_cffi_var_GxB_LAND_FIRST_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_FP32", (void *)_cffi_var_GxB_LAND_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LAND_FP64", (void *)_cffi_var_GxB_LAND_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LAND_GE_BOOL", (void *)_cffi_var_GxB_LAND_GE_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_GE_FP32", (void *)_cffi_var_GxB_LAND_GE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_GE_FP64", (void *)_cffi_var_GxB_LAND_GE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_GE_INT16", (void *)_cffi_var_GxB_LAND_GE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_GE_INT32", (void *)_cffi_var_GxB_LAND_GE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_GE_INT64", (void *)_cffi_var_GxB_LAND_GE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_GE_INT8", (void *)_cffi_var_GxB_LAND_GE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_GE_UINT16", (void *)_cffi_var_GxB_LAND_GE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_GE_UINT32", (void *)_cffi_var_GxB_LAND_GE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_GE_UINT64", (void *)_cffi_var_GxB_LAND_GE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_GE_UINT8", (void *)_cffi_var_GxB_LAND_GE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_GT_BOOL", (void *)_cffi_var_GxB_LAND_GT_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_GT_FP32", (void *)_cffi_var_GxB_LAND_GT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_GT_FP64", (void *)_cffi_var_GxB_LAND_GT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_GT_INT16", (void *)_cffi_var_GxB_LAND_GT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_GT_INT32", (void *)_cffi_var_GxB_LAND_GT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_GT_INT64", (void *)_cffi_var_GxB_LAND_GT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_GT_INT8", (void *)_cffi_var_GxB_LAND_GT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_GT_UINT16", (void *)_cffi_var_GxB_LAND_GT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_GT_UINT32", (void *)_cffi_var_GxB_LAND_GT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_GT_UINT64", (void *)_cffi_var_GxB_LAND_GT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_GT_UINT8", (void *)_cffi_var_GxB_LAND_GT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_INT16", (void *)_cffi_var_GxB_LAND_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LAND_INT32", (void *)_cffi_var_GxB_LAND_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LAND_INT64", (void *)_cffi_var_GxB_LAND_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LAND_INT8", (void *)_cffi_var_GxB_LAND_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LAND_LAND_BOOL", (void *)_cffi_var_GxB_LAND_LAND_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_LE_BOOL", (void *)_cffi_var_GxB_LAND_LE_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_LE_FP32", (void *)_cffi_var_GxB_LAND_LE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_LE_FP64", (void *)_cffi_var_GxB_LAND_LE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_LE_INT16", (void *)_cffi_var_GxB_LAND_LE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_LE_INT32", (void *)_cffi_var_GxB_LAND_LE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_LE_INT64", (void *)_cffi_var_GxB_LAND_LE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_LE_INT8", (void *)_cffi_var_GxB_LAND_LE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_LE_UINT16", (void *)_cffi_var_GxB_LAND_LE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_LE_UINT32", (void *)_cffi_var_GxB_LAND_LE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_LE_UINT64", (void *)_cffi_var_GxB_LAND_LE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_LE_UINT8", (void *)_cffi_var_GxB_LAND_LE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_LT_BOOL", (void *)_cffi_var_GxB_LAND_LT_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_LT_FP32", (void *)_cffi_var_GxB_LAND_LT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_LT_FP64", (void *)_cffi_var_GxB_LAND_LT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_LT_INT16", (void *)_cffi_var_GxB_LAND_LT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_LT_INT32", (void *)_cffi_var_GxB_LAND_LT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_LT_INT64", (void *)_cffi_var_GxB_LAND_LT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_LT_INT8", (void *)_cffi_var_GxB_LAND_LT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_LT_UINT16", (void *)_cffi_var_GxB_LAND_LT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_LT_UINT32", (void *)_cffi_var_GxB_LAND_LT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_LT_UINT64", (void *)_cffi_var_GxB_LAND_LT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_LT_UINT8", (void *)_cffi_var_GxB_LAND_LT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_LXOR_BOOL", (void *)_cffi_var_GxB_LAND_LXOR_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_NE_FP32", (void *)_cffi_var_GxB_LAND_NE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_NE_FP64", (void *)_cffi_var_GxB_LAND_NE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_NE_INT16", (void *)_cffi_var_GxB_LAND_NE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_NE_INT32", (void *)_cffi_var_GxB_LAND_NE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_NE_INT64", (void *)_cffi_var_GxB_LAND_NE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_NE_INT8", (void *)_cffi_var_GxB_LAND_NE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_NE_UINT16", (void *)_cffi_var_GxB_LAND_NE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_NE_UINT32", (void *)_cffi_var_GxB_LAND_NE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_NE_UINT64", (void *)_cffi_var_GxB_LAND_NE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_NE_UINT8", (void *)_cffi_var_GxB_LAND_NE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_PAIR_BOOL", (void *)_cffi_var_GxB_LAND_PAIR_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_SECOND_BOOL", (void *)_cffi_var_GxB_LAND_SECOND_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LAND_UINT16", (void *)_cffi_var_GxB_LAND_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LAND_UINT32", (void *)_cffi_var_GxB_LAND_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LAND_UINT64", (void *)_cffi_var_GxB_LAND_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LAND_UINT8", (void *)_cffi_var_GxB_LAND_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LDEXP_FP32", (void *)_cffi_var_GxB_LDEXP_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LDEXP_FP64", (void *)_cffi_var_GxB_LDEXP_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LE_THUNK", (void *)_cffi_var_GxB_LE_THUNK, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 155), (void *)0 },
  { "GxB_LE_ZERO", (void *)_cffi_var_GxB_LE_ZERO, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 155), (void *)0 },
  { "GxB_LGAMMA_FP32", (void *)_cffi_var_GxB_LGAMMA_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LGAMMA_FP64", (void *)_cffi_var_GxB_LGAMMA_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LIBRARY_ABOUT", (void *)_cffi_const_GxB_LIBRARY_ABOUT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_LIBRARY_COMPILE_DATE", (void *)_cffi_const_GxB_LIBRARY_COMPILE_DATE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_LIBRARY_COMPILE_TIME", (void *)_cffi_const_GxB_LIBRARY_COMPILE_TIME, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_LIBRARY_DATE", (void *)_cffi_const_GxB_LIBRARY_DATE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_LIBRARY_LICENSE", (void *)_cffi_const_GxB_LIBRARY_LICENSE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_LIBRARY_NAME", (void *)_cffi_const_GxB_LIBRARY_NAME, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_LIBRARY_OPENMP", (void *)_cffi_const_GxB_LIBRARY_OPENMP, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_LIBRARY_URL", (void *)_cffi_const_GxB_LIBRARY_URL, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_LIBRARY_VERSION", (void *)_cffi_const_GxB_LIBRARY_VERSION, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_LNOT_BOOL", (void *)_cffi_var_GxB_LNOT_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LNOT_FP32", (void *)_cffi_var_GxB_LNOT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LNOT_FP64", (void *)_cffi_var_GxB_LNOT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LNOT_INT16", (void *)_cffi_var_GxB_LNOT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LNOT_INT32", (void *)_cffi_var_GxB_LNOT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LNOT_INT64", (void *)_cffi_var_GxB_LNOT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LNOT_INT8", (void *)_cffi_var_GxB_LNOT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LNOT_UINT16", (void *)_cffi_var_GxB_LNOT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LNOT_UINT32", (void *)_cffi_var_GxB_LNOT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LNOT_UINT64", (void *)_cffi_var_GxB_LNOT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LNOT_UINT8", (void *)_cffi_var_GxB_LNOT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LOG10_FC32", (void *)_cffi_var_GxB_LOG10_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LOG10_FC64", (void *)_cffi_var_GxB_LOG10_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LOG10_FP32", (void *)_cffi_var_GxB_LOG10_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LOG10_FP64", (void *)_cffi_var_GxB_LOG10_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LOG1P_FC32", (void *)_cffi_var_GxB_LOG1P_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LOG1P_FC64", (void *)_cffi_var_GxB_LOG1P_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LOG1P_FP32", (void *)_cffi_var_GxB_LOG1P_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LOG1P_FP64", (void *)_cffi_var_GxB_LOG1P_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LOG2_FC32", (void *)_cffi_var_GxB_LOG2_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LOG2_FC64", (void *)_cffi_var_GxB_LOG2_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LOG2_FP32", (void *)_cffi_var_GxB_LOG2_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LOG2_FP64", (void *)_cffi_var_GxB_LOG2_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LOG_FC32", (void *)_cffi_var_GxB_LOG_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LOG_FC64", (void *)_cffi_var_GxB_LOG_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LOG_FP32", (void *)_cffi_var_GxB_LOG_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LOG_FP64", (void *)_cffi_var_GxB_LOG_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_LOR_BOOL", (void *)_cffi_var_GxB_LOR_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LOR_EQ_BOOL", (void *)_cffi_var_GxB_LOR_EQ_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_EQ_FP32", (void *)_cffi_var_GxB_LOR_EQ_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_EQ_FP64", (void *)_cffi_var_GxB_LOR_EQ_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_EQ_INT16", (void *)_cffi_var_GxB_LOR_EQ_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_EQ_INT32", (void *)_cffi_var_GxB_LOR_EQ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_EQ_INT64", (void *)_cffi_var_GxB_LOR_EQ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_EQ_INT8", (void *)_cffi_var_GxB_LOR_EQ_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_EQ_UINT16", (void *)_cffi_var_GxB_LOR_EQ_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_EQ_UINT32", (void *)_cffi_var_GxB_LOR_EQ_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_EQ_UINT64", (void *)_cffi_var_GxB_LOR_EQ_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_EQ_UINT8", (void *)_cffi_var_GxB_LOR_EQ_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_FIRST_BOOL", (void *)_cffi_var_GxB_LOR_FIRST_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_FP32", (void *)_cffi_var_GxB_LOR_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LOR_FP64", (void *)_cffi_var_GxB_LOR_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LOR_GE_BOOL", (void *)_cffi_var_GxB_LOR_GE_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_GE_FP32", (void *)_cffi_var_GxB_LOR_GE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_GE_FP64", (void *)_cffi_var_GxB_LOR_GE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_GE_INT16", (void *)_cffi_var_GxB_LOR_GE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_GE_INT32", (void *)_cffi_var_GxB_LOR_GE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_GE_INT64", (void *)_cffi_var_GxB_LOR_GE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_GE_INT8", (void *)_cffi_var_GxB_LOR_GE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_GE_UINT16", (void *)_cffi_var_GxB_LOR_GE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_GE_UINT32", (void *)_cffi_var_GxB_LOR_GE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_GE_UINT64", (void *)_cffi_var_GxB_LOR_GE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_GE_UINT8", (void *)_cffi_var_GxB_LOR_GE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_GT_BOOL", (void *)_cffi_var_GxB_LOR_GT_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_GT_FP32", (void *)_cffi_var_GxB_LOR_GT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_GT_FP64", (void *)_cffi_var_GxB_LOR_GT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_GT_INT16", (void *)_cffi_var_GxB_LOR_GT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_GT_INT32", (void *)_cffi_var_GxB_LOR_GT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_GT_INT64", (void *)_cffi_var_GxB_LOR_GT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_GT_INT8", (void *)_cffi_var_GxB_LOR_GT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_GT_UINT16", (void *)_cffi_var_GxB_LOR_GT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_GT_UINT32", (void *)_cffi_var_GxB_LOR_GT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_GT_UINT64", (void *)_cffi_var_GxB_LOR_GT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_GT_UINT8", (void *)_cffi_var_GxB_LOR_GT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_INT16", (void *)_cffi_var_GxB_LOR_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LOR_INT32", (void *)_cffi_var_GxB_LOR_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LOR_INT64", (void *)_cffi_var_GxB_LOR_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LOR_INT8", (void *)_cffi_var_GxB_LOR_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LOR_LE_BOOL", (void *)_cffi_var_GxB_LOR_LE_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_LE_FP32", (void *)_cffi_var_GxB_LOR_LE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_LE_FP64", (void *)_cffi_var_GxB_LOR_LE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_LE_INT16", (void *)_cffi_var_GxB_LOR_LE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_LE_INT32", (void *)_cffi_var_GxB_LOR_LE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_LE_INT64", (void *)_cffi_var_GxB_LOR_LE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_LE_INT8", (void *)_cffi_var_GxB_LOR_LE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_LE_UINT16", (void *)_cffi_var_GxB_LOR_LE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_LE_UINT32", (void *)_cffi_var_GxB_LOR_LE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_LE_UINT64", (void *)_cffi_var_GxB_LOR_LE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_LE_UINT8", (void *)_cffi_var_GxB_LOR_LE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_LOR_BOOL", (void *)_cffi_var_GxB_LOR_LOR_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_LT_BOOL", (void *)_cffi_var_GxB_LOR_LT_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_LT_FP32", (void *)_cffi_var_GxB_LOR_LT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_LT_FP64", (void *)_cffi_var_GxB_LOR_LT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_LT_INT16", (void *)_cffi_var_GxB_LOR_LT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_LT_INT32", (void *)_cffi_var_GxB_LOR_LT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_LT_INT64", (void *)_cffi_var_GxB_LOR_LT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_LT_INT8", (void *)_cffi_var_GxB_LOR_LT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_LT_UINT16", (void *)_cffi_var_GxB_LOR_LT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_LT_UINT32", (void *)_cffi_var_GxB_LOR_LT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_LT_UINT64", (void *)_cffi_var_GxB_LOR_LT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_LT_UINT8", (void *)_cffi_var_GxB_LOR_LT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_LXOR_BOOL", (void *)_cffi_var_GxB_LOR_LXOR_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_NE_FP32", (void *)_cffi_var_GxB_LOR_NE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_NE_FP64", (void *)_cffi_var_GxB_LOR_NE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_NE_INT16", (void *)_cffi_var_GxB_LOR_NE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_NE_INT32", (void *)_cffi_var_GxB_LOR_NE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_NE_INT64", (void *)_cffi_var_GxB_LOR_NE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_NE_INT8", (void *)_cffi_var_GxB_LOR_NE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_NE_UINT16", (void *)_cffi_var_GxB_LOR_NE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_NE_UINT32", (void *)_cffi_var_GxB_LOR_NE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_NE_UINT64", (void *)_cffi_var_GxB_LOR_NE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_NE_UINT8", (void *)_cffi_var_GxB_LOR_NE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_PAIR_BOOL", (void *)_cffi_var_GxB_LOR_PAIR_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_SECOND_BOOL", (void *)_cffi_var_GxB_LOR_SECOND_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LOR_UINT16", (void *)_cffi_var_GxB_LOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LOR_UINT32", (void *)_cffi_var_GxB_LOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LOR_UINT64", (void *)_cffi_var_GxB_LOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LOR_UINT8", (void *)_cffi_var_GxB_LOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LT_THUNK", (void *)_cffi_var_GxB_LT_THUNK, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 155), (void *)0 },
  { "GxB_LT_ZERO", (void *)_cffi_var_GxB_LT_ZERO, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 155), (void *)0 },
  { "GxB_LXOR_BOOL", (void *)_cffi_var_GxB_LXOR_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LXOR_EQ_BOOL", (void *)_cffi_var_GxB_LXOR_EQ_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_EQ_FP32", (void *)_cffi_var_GxB_LXOR_EQ_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_EQ_FP64", (void *)_cffi_var_GxB_LXOR_EQ_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_EQ_INT16", (void *)_cffi_var_GxB_LXOR_EQ_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_EQ_INT32", (void *)_cffi_var_GxB_LXOR_EQ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_EQ_INT64", (void *)_cffi_var_GxB_LXOR_EQ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_EQ_INT8", (void *)_cffi_var_GxB_LXOR_EQ_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_EQ_UINT16", (void *)_cffi_var_GxB_LXOR_EQ_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_EQ_UINT32", (void *)_cffi_var_GxB_LXOR_EQ_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_EQ_UINT64", (void *)_cffi_var_GxB_LXOR_EQ_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_EQ_UINT8", (void *)_cffi_var_GxB_LXOR_EQ_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_FIRST_BOOL", (void *)_cffi_var_GxB_LXOR_FIRST_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_FP32", (void *)_cffi_var_GxB_LXOR_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LXOR_FP64", (void *)_cffi_var_GxB_LXOR_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LXOR_GE_BOOL", (void *)_cffi_var_GxB_LXOR_GE_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_GE_FP32", (void *)_cffi_var_GxB_LXOR_GE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_GE_FP64", (void *)_cffi_var_GxB_LXOR_GE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_GE_INT16", (void *)_cffi_var_GxB_LXOR_GE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_GE_INT32", (void *)_cffi_var_GxB_LXOR_GE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_GE_INT64", (void *)_cffi_var_GxB_LXOR_GE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_GE_INT8", (void *)_cffi_var_GxB_LXOR_GE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_GE_UINT16", (void *)_cffi_var_GxB_LXOR_GE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_GE_UINT32", (void *)_cffi_var_GxB_LXOR_GE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_GE_UINT64", (void *)_cffi_var_GxB_LXOR_GE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_GE_UINT8", (void *)_cffi_var_GxB_LXOR_GE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_GT_BOOL", (void *)_cffi_var_GxB_LXOR_GT_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_GT_FP32", (void *)_cffi_var_GxB_LXOR_GT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_GT_FP64", (void *)_cffi_var_GxB_LXOR_GT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_GT_INT16", (void *)_cffi_var_GxB_LXOR_GT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_GT_INT32", (void *)_cffi_var_GxB_LXOR_GT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_GT_INT64", (void *)_cffi_var_GxB_LXOR_GT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_GT_INT8", (void *)_cffi_var_GxB_LXOR_GT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_GT_UINT16", (void *)_cffi_var_GxB_LXOR_GT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_GT_UINT32", (void *)_cffi_var_GxB_LXOR_GT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_GT_UINT64", (void *)_cffi_var_GxB_LXOR_GT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_GT_UINT8", (void *)_cffi_var_GxB_LXOR_GT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_INT16", (void *)_cffi_var_GxB_LXOR_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LXOR_INT32", (void *)_cffi_var_GxB_LXOR_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LXOR_INT64", (void *)_cffi_var_GxB_LXOR_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LXOR_INT8", (void *)_cffi_var_GxB_LXOR_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LXOR_LE_BOOL", (void *)_cffi_var_GxB_LXOR_LE_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_LE_FP32", (void *)_cffi_var_GxB_LXOR_LE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_LE_FP64", (void *)_cffi_var_GxB_LXOR_LE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_LE_INT16", (void *)_cffi_var_GxB_LXOR_LE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_LE_INT32", (void *)_cffi_var_GxB_LXOR_LE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_LE_INT64", (void *)_cffi_var_GxB_LXOR_LE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_LE_INT8", (void *)_cffi_var_GxB_LXOR_LE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_LE_UINT16", (void *)_cffi_var_GxB_LXOR_LE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_LE_UINT32", (void *)_cffi_var_GxB_LXOR_LE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_LE_UINT64", (void *)_cffi_var_GxB_LXOR_LE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_LE_UINT8", (void *)_cffi_var_GxB_LXOR_LE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_LOR_BOOL", (void *)_cffi_var_GxB_LXOR_LOR_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_LT_BOOL", (void *)_cffi_var_GxB_LXOR_LT_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_LT_FP32", (void *)_cffi_var_GxB_LXOR_LT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_LT_FP64", (void *)_cffi_var_GxB_LXOR_LT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_LT_INT16", (void *)_cffi_var_GxB_LXOR_LT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_LT_INT32", (void *)_cffi_var_GxB_LXOR_LT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_LT_INT64", (void *)_cffi_var_GxB_LXOR_LT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_LT_INT8", (void *)_cffi_var_GxB_LXOR_LT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_LT_UINT16", (void *)_cffi_var_GxB_LXOR_LT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_LT_UINT32", (void *)_cffi_var_GxB_LXOR_LT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_LT_UINT64", (void *)_cffi_var_GxB_LXOR_LT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_LT_UINT8", (void *)_cffi_var_GxB_LXOR_LT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_LXOR_BOOL", (void *)_cffi_var_GxB_LXOR_LXOR_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_NE_FP32", (void *)_cffi_var_GxB_LXOR_NE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_NE_FP64", (void *)_cffi_var_GxB_LXOR_NE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_NE_INT16", (void *)_cffi_var_GxB_LXOR_NE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_NE_INT32", (void *)_cffi_var_GxB_LXOR_NE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_NE_INT64", (void *)_cffi_var_GxB_LXOR_NE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_NE_INT8", (void *)_cffi_var_GxB_LXOR_NE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_NE_UINT16", (void *)_cffi_var_GxB_LXOR_NE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_NE_UINT32", (void *)_cffi_var_GxB_LXOR_NE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_NE_UINT64", (void *)_cffi_var_GxB_LXOR_NE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_NE_UINT8", (void *)_cffi_var_GxB_LXOR_NE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_PAIR_BOOL", (void *)_cffi_var_GxB_LXOR_PAIR_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_SECOND_BOOL", (void *)_cffi_var_GxB_LXOR_SECOND_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_LXOR_UINT16", (void *)_cffi_var_GxB_LXOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LXOR_UINT32", (void *)_cffi_var_GxB_LXOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LXOR_UINT64", (void *)_cffi_var_GxB_LXOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_LXOR_UINT8", (void *)_cffi_var_GxB_LXOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_MAX_DIV_FP32", (void *)_cffi_var_GxB_MAX_DIV_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_DIV_FP64", (void *)_cffi_var_GxB_MAX_DIV_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_DIV_INT16", (void *)_cffi_var_GxB_MAX_DIV_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_DIV_INT32", (void *)_cffi_var_GxB_MAX_DIV_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_DIV_INT64", (void *)_cffi_var_GxB_MAX_DIV_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_DIV_INT8", (void *)_cffi_var_GxB_MAX_DIV_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_DIV_UINT16", (void *)_cffi_var_GxB_MAX_DIV_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_DIV_UINT32", (void *)_cffi_var_GxB_MAX_DIV_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_DIV_UINT64", (void *)_cffi_var_GxB_MAX_DIV_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_DIV_UINT8", (void *)_cffi_var_GxB_MAX_DIV_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_FIRSTI1_INT32", (void *)_cffi_var_GxB_MAX_FIRSTI1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_FIRSTI1_INT64", (void *)_cffi_var_GxB_MAX_FIRSTI1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_FIRSTI_INT32", (void *)_cffi_var_GxB_MAX_FIRSTI_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_FIRSTI_INT64", (void *)_cffi_var_GxB_MAX_FIRSTI_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_FIRSTJ1_INT32", (void *)_cffi_var_GxB_MAX_FIRSTJ1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_FIRSTJ1_INT64", (void *)_cffi_var_GxB_MAX_FIRSTJ1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_FIRSTJ_INT32", (void *)_cffi_var_GxB_MAX_FIRSTJ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_FIRSTJ_INT64", (void *)_cffi_var_GxB_MAX_FIRSTJ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISEQ_FP32", (void *)_cffi_var_GxB_MAX_ISEQ_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISEQ_FP64", (void *)_cffi_var_GxB_MAX_ISEQ_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISEQ_INT16", (void *)_cffi_var_GxB_MAX_ISEQ_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISEQ_INT32", (void *)_cffi_var_GxB_MAX_ISEQ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISEQ_INT64", (void *)_cffi_var_GxB_MAX_ISEQ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISEQ_INT8", (void *)_cffi_var_GxB_MAX_ISEQ_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISEQ_UINT16", (void *)_cffi_var_GxB_MAX_ISEQ_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISEQ_UINT32", (void *)_cffi_var_GxB_MAX_ISEQ_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISEQ_UINT64", (void *)_cffi_var_GxB_MAX_ISEQ_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISEQ_UINT8", (void *)_cffi_var_GxB_MAX_ISEQ_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISGE_FP32", (void *)_cffi_var_GxB_MAX_ISGE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISGE_FP64", (void *)_cffi_var_GxB_MAX_ISGE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISGE_INT16", (void *)_cffi_var_GxB_MAX_ISGE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISGE_INT32", (void *)_cffi_var_GxB_MAX_ISGE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISGE_INT64", (void *)_cffi_var_GxB_MAX_ISGE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISGE_INT8", (void *)_cffi_var_GxB_MAX_ISGE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISGE_UINT16", (void *)_cffi_var_GxB_MAX_ISGE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISGE_UINT32", (void *)_cffi_var_GxB_MAX_ISGE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISGE_UINT64", (void *)_cffi_var_GxB_MAX_ISGE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISGE_UINT8", (void *)_cffi_var_GxB_MAX_ISGE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISGT_FP32", (void *)_cffi_var_GxB_MAX_ISGT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISGT_FP64", (void *)_cffi_var_GxB_MAX_ISGT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISGT_INT16", (void *)_cffi_var_GxB_MAX_ISGT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISGT_INT32", (void *)_cffi_var_GxB_MAX_ISGT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISGT_INT64", (void *)_cffi_var_GxB_MAX_ISGT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISGT_INT8", (void *)_cffi_var_GxB_MAX_ISGT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISGT_UINT16", (void *)_cffi_var_GxB_MAX_ISGT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISGT_UINT32", (void *)_cffi_var_GxB_MAX_ISGT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISGT_UINT64", (void *)_cffi_var_GxB_MAX_ISGT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISGT_UINT8", (void *)_cffi_var_GxB_MAX_ISGT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISLE_FP32", (void *)_cffi_var_GxB_MAX_ISLE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISLE_FP64", (void *)_cffi_var_GxB_MAX_ISLE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISLE_INT16", (void *)_cffi_var_GxB_MAX_ISLE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISLE_INT32", (void *)_cffi_var_GxB_MAX_ISLE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISLE_INT64", (void *)_cffi_var_GxB_MAX_ISLE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISLE_INT8", (void *)_cffi_var_GxB_MAX_ISLE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISLE_UINT16", (void *)_cffi_var_GxB_MAX_ISLE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISLE_UINT32", (void *)_cffi_var_GxB_MAX_ISLE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISLE_UINT64", (void *)_cffi_var_GxB_MAX_ISLE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISLE_UINT8", (void *)_cffi_var_GxB_MAX_ISLE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISLT_FP32", (void *)_cffi_var_GxB_MAX_ISLT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISLT_FP64", (void *)_cffi_var_GxB_MAX_ISLT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISLT_INT16", (void *)_cffi_var_GxB_MAX_ISLT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISLT_INT32", (void *)_cffi_var_GxB_MAX_ISLT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISLT_INT64", (void *)_cffi_var_GxB_MAX_ISLT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISLT_INT8", (void *)_cffi_var_GxB_MAX_ISLT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISLT_UINT16", (void *)_cffi_var_GxB_MAX_ISLT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISLT_UINT32", (void *)_cffi_var_GxB_MAX_ISLT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISLT_UINT64", (void *)_cffi_var_GxB_MAX_ISLT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISLT_UINT8", (void *)_cffi_var_GxB_MAX_ISLT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISNE_FP32", (void *)_cffi_var_GxB_MAX_ISNE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISNE_FP64", (void *)_cffi_var_GxB_MAX_ISNE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISNE_INT16", (void *)_cffi_var_GxB_MAX_ISNE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISNE_INT32", (void *)_cffi_var_GxB_MAX_ISNE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISNE_INT64", (void *)_cffi_var_GxB_MAX_ISNE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISNE_INT8", (void *)_cffi_var_GxB_MAX_ISNE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISNE_UINT16", (void *)_cffi_var_GxB_MAX_ISNE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISNE_UINT32", (void *)_cffi_var_GxB_MAX_ISNE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISNE_UINT64", (void *)_cffi_var_GxB_MAX_ISNE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_ISNE_UINT8", (void *)_cffi_var_GxB_MAX_ISNE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LAND_FP32", (void *)_cffi_var_GxB_MAX_LAND_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LAND_FP64", (void *)_cffi_var_GxB_MAX_LAND_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LAND_INT16", (void *)_cffi_var_GxB_MAX_LAND_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LAND_INT32", (void *)_cffi_var_GxB_MAX_LAND_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LAND_INT64", (void *)_cffi_var_GxB_MAX_LAND_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LAND_INT8", (void *)_cffi_var_GxB_MAX_LAND_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LAND_UINT16", (void *)_cffi_var_GxB_MAX_LAND_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LAND_UINT32", (void *)_cffi_var_GxB_MAX_LAND_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LAND_UINT64", (void *)_cffi_var_GxB_MAX_LAND_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LAND_UINT8", (void *)_cffi_var_GxB_MAX_LAND_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LOR_FP32", (void *)_cffi_var_GxB_MAX_LOR_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LOR_FP64", (void *)_cffi_var_GxB_MAX_LOR_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LOR_INT16", (void *)_cffi_var_GxB_MAX_LOR_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LOR_INT32", (void *)_cffi_var_GxB_MAX_LOR_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LOR_INT64", (void *)_cffi_var_GxB_MAX_LOR_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LOR_INT8", (void *)_cffi_var_GxB_MAX_LOR_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LOR_UINT16", (void *)_cffi_var_GxB_MAX_LOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LOR_UINT32", (void *)_cffi_var_GxB_MAX_LOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LOR_UINT64", (void *)_cffi_var_GxB_MAX_LOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LOR_UINT8", (void *)_cffi_var_GxB_MAX_LOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LXOR_FP32", (void *)_cffi_var_GxB_MAX_LXOR_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LXOR_FP64", (void *)_cffi_var_GxB_MAX_LXOR_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LXOR_INT16", (void *)_cffi_var_GxB_MAX_LXOR_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LXOR_INT32", (void *)_cffi_var_GxB_MAX_LXOR_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LXOR_INT64", (void *)_cffi_var_GxB_MAX_LXOR_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LXOR_INT8", (void *)_cffi_var_GxB_MAX_LXOR_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LXOR_UINT16", (void *)_cffi_var_GxB_MAX_LXOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LXOR_UINT32", (void *)_cffi_var_GxB_MAX_LXOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LXOR_UINT64", (void *)_cffi_var_GxB_MAX_LXOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_LXOR_UINT8", (void *)_cffi_var_GxB_MAX_LXOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_MAX_FP32", (void *)_cffi_var_GxB_MAX_MAX_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_MAX_FP64", (void *)_cffi_var_GxB_MAX_MAX_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_MAX_INT16", (void *)_cffi_var_GxB_MAX_MAX_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_MAX_INT32", (void *)_cffi_var_GxB_MAX_MAX_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_MAX_INT64", (void *)_cffi_var_GxB_MAX_MAX_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_MAX_INT8", (void *)_cffi_var_GxB_MAX_MAX_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_MAX_UINT16", (void *)_cffi_var_GxB_MAX_MAX_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_MAX_UINT32", (void *)_cffi_var_GxB_MAX_MAX_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_MAX_UINT64", (void *)_cffi_var_GxB_MAX_MAX_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_MAX_UINT8", (void *)_cffi_var_GxB_MAX_MAX_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_MINUS_FP32", (void *)_cffi_var_GxB_MAX_MINUS_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_MINUS_FP64", (void *)_cffi_var_GxB_MAX_MINUS_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_MINUS_INT16", (void *)_cffi_var_GxB_MAX_MINUS_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_MINUS_INT32", (void *)_cffi_var_GxB_MAX_MINUS_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_MINUS_INT64", (void *)_cffi_var_GxB_MAX_MINUS_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_MINUS_INT8", (void *)_cffi_var_GxB_MAX_MINUS_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_MINUS_UINT16", (void *)_cffi_var_GxB_MAX_MINUS_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_MINUS_UINT32", (void *)_cffi_var_GxB_MAX_MINUS_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_MINUS_UINT64", (void *)_cffi_var_GxB_MAX_MINUS_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_MINUS_UINT8", (void *)_cffi_var_GxB_MAX_MINUS_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_NAME_LEN", (void *)_cffi_const_GxB_MAX_NAME_LEN, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_MAX_PAIR_FP32", (void *)_cffi_var_GxB_MAX_PAIR_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_PAIR_FP64", (void *)_cffi_var_GxB_MAX_PAIR_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_PAIR_INT16", (void *)_cffi_var_GxB_MAX_PAIR_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_PAIR_INT32", (void *)_cffi_var_GxB_MAX_PAIR_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_PAIR_INT64", (void *)_cffi_var_GxB_MAX_PAIR_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_PAIR_INT8", (void *)_cffi_var_GxB_MAX_PAIR_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_PAIR_UINT16", (void *)_cffi_var_GxB_MAX_PAIR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_PAIR_UINT32", (void *)_cffi_var_GxB_MAX_PAIR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_PAIR_UINT64", (void *)_cffi_var_GxB_MAX_PAIR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_PAIR_UINT8", (void *)_cffi_var_GxB_MAX_PAIR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_RDIV_FP32", (void *)_cffi_var_GxB_MAX_RDIV_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_RDIV_FP64", (void *)_cffi_var_GxB_MAX_RDIV_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_RDIV_INT16", (void *)_cffi_var_GxB_MAX_RDIV_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_RDIV_INT32", (void *)_cffi_var_GxB_MAX_RDIV_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_RDIV_INT64", (void *)_cffi_var_GxB_MAX_RDIV_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_RDIV_INT8", (void *)_cffi_var_GxB_MAX_RDIV_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_RDIV_UINT16", (void *)_cffi_var_GxB_MAX_RDIV_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_RDIV_UINT32", (void *)_cffi_var_GxB_MAX_RDIV_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_RDIV_UINT64", (void *)_cffi_var_GxB_MAX_RDIV_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_RDIV_UINT8", (void *)_cffi_var_GxB_MAX_RDIV_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_RMINUS_FP32", (void *)_cffi_var_GxB_MAX_RMINUS_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_RMINUS_FP64", (void *)_cffi_var_GxB_MAX_RMINUS_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_RMINUS_INT16", (void *)_cffi_var_GxB_MAX_RMINUS_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_RMINUS_INT32", (void *)_cffi_var_GxB_MAX_RMINUS_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_RMINUS_INT64", (void *)_cffi_var_GxB_MAX_RMINUS_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_RMINUS_INT8", (void *)_cffi_var_GxB_MAX_RMINUS_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_RMINUS_UINT16", (void *)_cffi_var_GxB_MAX_RMINUS_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_RMINUS_UINT32", (void *)_cffi_var_GxB_MAX_RMINUS_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_RMINUS_UINT64", (void *)_cffi_var_GxB_MAX_RMINUS_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_RMINUS_UINT8", (void *)_cffi_var_GxB_MAX_RMINUS_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_SECONDI1_INT32", (void *)_cffi_var_GxB_MAX_SECONDI1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_SECONDI1_INT64", (void *)_cffi_var_GxB_MAX_SECONDI1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_SECONDI_INT32", (void *)_cffi_var_GxB_MAX_SECONDI_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_SECONDI_INT64", (void *)_cffi_var_GxB_MAX_SECONDI_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_SECONDJ1_INT32", (void *)_cffi_var_GxB_MAX_SECONDJ1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_SECONDJ1_INT64", (void *)_cffi_var_GxB_MAX_SECONDJ1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_SECONDJ_INT32", (void *)_cffi_var_GxB_MAX_SECONDJ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MAX_SECONDJ_INT64", (void *)_cffi_var_GxB_MAX_SECONDJ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MEMORY_POOL", (void *)_cffi_const_GxB_MEMORY_POOL, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_MINUS_FC32", (void *)_cffi_var_GxB_MINUS_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_MINUS_FC64", (void *)_cffi_var_GxB_MINUS_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_MINV_FC32", (void *)_cffi_var_GxB_MINV_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_MINV_FC64", (void *)_cffi_var_GxB_MINV_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_MIN_DIV_FP32", (void *)_cffi_var_GxB_MIN_DIV_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_DIV_FP64", (void *)_cffi_var_GxB_MIN_DIV_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_DIV_INT16", (void *)_cffi_var_GxB_MIN_DIV_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_DIV_INT32", (void *)_cffi_var_GxB_MIN_DIV_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_DIV_INT64", (void *)_cffi_var_GxB_MIN_DIV_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_DIV_INT8", (void *)_cffi_var_GxB_MIN_DIV_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_DIV_UINT16", (void *)_cffi_var_GxB_MIN_DIV_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_DIV_UINT32", (void *)_cffi_var_GxB_MIN_DIV_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_DIV_UINT64", (void *)_cffi_var_GxB_MIN_DIV_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_DIV_UINT8", (void *)_cffi_var_GxB_MIN_DIV_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_FIRSTI1_INT32", (void *)_cffi_var_GxB_MIN_FIRSTI1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_FIRSTI1_INT64", (void *)_cffi_var_GxB_MIN_FIRSTI1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_FIRSTI_INT32", (void *)_cffi_var_GxB_MIN_FIRSTI_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_FIRSTI_INT64", (void *)_cffi_var_GxB_MIN_FIRSTI_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_FIRSTJ1_INT32", (void *)_cffi_var_GxB_MIN_FIRSTJ1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_FIRSTJ1_INT64", (void *)_cffi_var_GxB_MIN_FIRSTJ1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_FIRSTJ_INT32", (void *)_cffi_var_GxB_MIN_FIRSTJ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_FIRSTJ_INT64", (void *)_cffi_var_GxB_MIN_FIRSTJ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISEQ_FP32", (void *)_cffi_var_GxB_MIN_ISEQ_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISEQ_FP64", (void *)_cffi_var_GxB_MIN_ISEQ_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISEQ_INT16", (void *)_cffi_var_GxB_MIN_ISEQ_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISEQ_INT32", (void *)_cffi_var_GxB_MIN_ISEQ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISEQ_INT64", (void *)_cffi_var_GxB_MIN_ISEQ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISEQ_INT8", (void *)_cffi_var_GxB_MIN_ISEQ_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISEQ_UINT16", (void *)_cffi_var_GxB_MIN_ISEQ_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISEQ_UINT32", (void *)_cffi_var_GxB_MIN_ISEQ_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISEQ_UINT64", (void *)_cffi_var_GxB_MIN_ISEQ_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISEQ_UINT8", (void *)_cffi_var_GxB_MIN_ISEQ_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISGE_FP32", (void *)_cffi_var_GxB_MIN_ISGE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISGE_FP64", (void *)_cffi_var_GxB_MIN_ISGE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISGE_INT16", (void *)_cffi_var_GxB_MIN_ISGE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISGE_INT32", (void *)_cffi_var_GxB_MIN_ISGE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISGE_INT64", (void *)_cffi_var_GxB_MIN_ISGE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISGE_INT8", (void *)_cffi_var_GxB_MIN_ISGE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISGE_UINT16", (void *)_cffi_var_GxB_MIN_ISGE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISGE_UINT32", (void *)_cffi_var_GxB_MIN_ISGE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISGE_UINT64", (void *)_cffi_var_GxB_MIN_ISGE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISGE_UINT8", (void *)_cffi_var_GxB_MIN_ISGE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISGT_FP32", (void *)_cffi_var_GxB_MIN_ISGT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISGT_FP64", (void *)_cffi_var_GxB_MIN_ISGT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISGT_INT16", (void *)_cffi_var_GxB_MIN_ISGT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISGT_INT32", (void *)_cffi_var_GxB_MIN_ISGT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISGT_INT64", (void *)_cffi_var_GxB_MIN_ISGT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISGT_INT8", (void *)_cffi_var_GxB_MIN_ISGT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISGT_UINT16", (void *)_cffi_var_GxB_MIN_ISGT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISGT_UINT32", (void *)_cffi_var_GxB_MIN_ISGT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISGT_UINT64", (void *)_cffi_var_GxB_MIN_ISGT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISGT_UINT8", (void *)_cffi_var_GxB_MIN_ISGT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISLE_FP32", (void *)_cffi_var_GxB_MIN_ISLE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISLE_FP64", (void *)_cffi_var_GxB_MIN_ISLE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISLE_INT16", (void *)_cffi_var_GxB_MIN_ISLE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISLE_INT32", (void *)_cffi_var_GxB_MIN_ISLE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISLE_INT64", (void *)_cffi_var_GxB_MIN_ISLE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISLE_INT8", (void *)_cffi_var_GxB_MIN_ISLE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISLE_UINT16", (void *)_cffi_var_GxB_MIN_ISLE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISLE_UINT32", (void *)_cffi_var_GxB_MIN_ISLE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISLE_UINT64", (void *)_cffi_var_GxB_MIN_ISLE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISLE_UINT8", (void *)_cffi_var_GxB_MIN_ISLE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISLT_FP32", (void *)_cffi_var_GxB_MIN_ISLT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISLT_FP64", (void *)_cffi_var_GxB_MIN_ISLT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISLT_INT16", (void *)_cffi_var_GxB_MIN_ISLT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISLT_INT32", (void *)_cffi_var_GxB_MIN_ISLT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISLT_INT64", (void *)_cffi_var_GxB_MIN_ISLT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISLT_INT8", (void *)_cffi_var_GxB_MIN_ISLT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISLT_UINT16", (void *)_cffi_var_GxB_MIN_ISLT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISLT_UINT32", (void *)_cffi_var_GxB_MIN_ISLT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISLT_UINT64", (void *)_cffi_var_GxB_MIN_ISLT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISLT_UINT8", (void *)_cffi_var_GxB_MIN_ISLT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISNE_FP32", (void *)_cffi_var_GxB_MIN_ISNE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISNE_FP64", (void *)_cffi_var_GxB_MIN_ISNE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISNE_INT16", (void *)_cffi_var_GxB_MIN_ISNE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISNE_INT32", (void *)_cffi_var_GxB_MIN_ISNE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISNE_INT64", (void *)_cffi_var_GxB_MIN_ISNE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISNE_INT8", (void *)_cffi_var_GxB_MIN_ISNE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISNE_UINT16", (void *)_cffi_var_GxB_MIN_ISNE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISNE_UINT32", (void *)_cffi_var_GxB_MIN_ISNE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISNE_UINT64", (void *)_cffi_var_GxB_MIN_ISNE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_ISNE_UINT8", (void *)_cffi_var_GxB_MIN_ISNE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LAND_FP32", (void *)_cffi_var_GxB_MIN_LAND_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LAND_FP64", (void *)_cffi_var_GxB_MIN_LAND_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LAND_INT16", (void *)_cffi_var_GxB_MIN_LAND_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LAND_INT32", (void *)_cffi_var_GxB_MIN_LAND_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LAND_INT64", (void *)_cffi_var_GxB_MIN_LAND_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LAND_INT8", (void *)_cffi_var_GxB_MIN_LAND_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LAND_UINT16", (void *)_cffi_var_GxB_MIN_LAND_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LAND_UINT32", (void *)_cffi_var_GxB_MIN_LAND_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LAND_UINT64", (void *)_cffi_var_GxB_MIN_LAND_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LAND_UINT8", (void *)_cffi_var_GxB_MIN_LAND_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LOR_FP32", (void *)_cffi_var_GxB_MIN_LOR_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LOR_FP64", (void *)_cffi_var_GxB_MIN_LOR_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LOR_INT16", (void *)_cffi_var_GxB_MIN_LOR_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LOR_INT32", (void *)_cffi_var_GxB_MIN_LOR_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LOR_INT64", (void *)_cffi_var_GxB_MIN_LOR_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LOR_INT8", (void *)_cffi_var_GxB_MIN_LOR_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LOR_UINT16", (void *)_cffi_var_GxB_MIN_LOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LOR_UINT32", (void *)_cffi_var_GxB_MIN_LOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LOR_UINT64", (void *)_cffi_var_GxB_MIN_LOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LOR_UINT8", (void *)_cffi_var_GxB_MIN_LOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LXOR_FP32", (void *)_cffi_var_GxB_MIN_LXOR_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LXOR_FP64", (void *)_cffi_var_GxB_MIN_LXOR_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LXOR_INT16", (void *)_cffi_var_GxB_MIN_LXOR_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LXOR_INT32", (void *)_cffi_var_GxB_MIN_LXOR_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LXOR_INT64", (void *)_cffi_var_GxB_MIN_LXOR_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LXOR_INT8", (void *)_cffi_var_GxB_MIN_LXOR_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LXOR_UINT16", (void *)_cffi_var_GxB_MIN_LXOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LXOR_UINT32", (void *)_cffi_var_GxB_MIN_LXOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LXOR_UINT64", (void *)_cffi_var_GxB_MIN_LXOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_LXOR_UINT8", (void *)_cffi_var_GxB_MIN_LXOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_MINUS_FP32", (void *)_cffi_var_GxB_MIN_MINUS_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_MINUS_FP64", (void *)_cffi_var_GxB_MIN_MINUS_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_MINUS_INT16", (void *)_cffi_var_GxB_MIN_MINUS_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_MINUS_INT32", (void *)_cffi_var_GxB_MIN_MINUS_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_MINUS_INT64", (void *)_cffi_var_GxB_MIN_MINUS_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_MINUS_INT8", (void *)_cffi_var_GxB_MIN_MINUS_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_MINUS_UINT16", (void *)_cffi_var_GxB_MIN_MINUS_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_MINUS_UINT32", (void *)_cffi_var_GxB_MIN_MINUS_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_MINUS_UINT64", (void *)_cffi_var_GxB_MIN_MINUS_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_MINUS_UINT8", (void *)_cffi_var_GxB_MIN_MINUS_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_MIN_FP32", (void *)_cffi_var_GxB_MIN_MIN_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_MIN_FP64", (void *)_cffi_var_GxB_MIN_MIN_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_MIN_INT16", (void *)_cffi_var_GxB_MIN_MIN_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_MIN_INT32", (void *)_cffi_var_GxB_MIN_MIN_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_MIN_INT64", (void *)_cffi_var_GxB_MIN_MIN_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_MIN_INT8", (void *)_cffi_var_GxB_MIN_MIN_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_MIN_UINT16", (void *)_cffi_var_GxB_MIN_MIN_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_MIN_UINT32", (void *)_cffi_var_GxB_MIN_MIN_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_MIN_UINT64", (void *)_cffi_var_GxB_MIN_MIN_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_MIN_UINT8", (void *)_cffi_var_GxB_MIN_MIN_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_PAIR_FP32", (void *)_cffi_var_GxB_MIN_PAIR_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_PAIR_FP64", (void *)_cffi_var_GxB_MIN_PAIR_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_PAIR_INT16", (void *)_cffi_var_GxB_MIN_PAIR_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_PAIR_INT32", (void *)_cffi_var_GxB_MIN_PAIR_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_PAIR_INT64", (void *)_cffi_var_GxB_MIN_PAIR_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_PAIR_INT8", (void *)_cffi_var_GxB_MIN_PAIR_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_PAIR_UINT16", (void *)_cffi_var_GxB_MIN_PAIR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_PAIR_UINT32", (void *)_cffi_var_GxB_MIN_PAIR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_PAIR_UINT64", (void *)_cffi_var_GxB_MIN_PAIR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_PAIR_UINT8", (void *)_cffi_var_GxB_MIN_PAIR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_RDIV_FP32", (void *)_cffi_var_GxB_MIN_RDIV_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_RDIV_FP64", (void *)_cffi_var_GxB_MIN_RDIV_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_RDIV_INT16", (void *)_cffi_var_GxB_MIN_RDIV_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_RDIV_INT32", (void *)_cffi_var_GxB_MIN_RDIV_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_RDIV_INT64", (void *)_cffi_var_GxB_MIN_RDIV_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_RDIV_INT8", (void *)_cffi_var_GxB_MIN_RDIV_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_RDIV_UINT16", (void *)_cffi_var_GxB_MIN_RDIV_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_RDIV_UINT32", (void *)_cffi_var_GxB_MIN_RDIV_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_RDIV_UINT64", (void *)_cffi_var_GxB_MIN_RDIV_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_RDIV_UINT8", (void *)_cffi_var_GxB_MIN_RDIV_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_RMINUS_FP32", (void *)_cffi_var_GxB_MIN_RMINUS_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_RMINUS_FP64", (void *)_cffi_var_GxB_MIN_RMINUS_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_RMINUS_INT16", (void *)_cffi_var_GxB_MIN_RMINUS_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_RMINUS_INT32", (void *)_cffi_var_GxB_MIN_RMINUS_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_RMINUS_INT64", (void *)_cffi_var_GxB_MIN_RMINUS_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_RMINUS_INT8", (void *)_cffi_var_GxB_MIN_RMINUS_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_RMINUS_UINT16", (void *)_cffi_var_GxB_MIN_RMINUS_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_RMINUS_UINT32", (void *)_cffi_var_GxB_MIN_RMINUS_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_RMINUS_UINT64", (void *)_cffi_var_GxB_MIN_RMINUS_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_RMINUS_UINT8", (void *)_cffi_var_GxB_MIN_RMINUS_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_SECONDI1_INT32", (void *)_cffi_var_GxB_MIN_SECONDI1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_SECONDI1_INT64", (void *)_cffi_var_GxB_MIN_SECONDI1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_SECONDI_INT32", (void *)_cffi_var_GxB_MIN_SECONDI_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_SECONDI_INT64", (void *)_cffi_var_GxB_MIN_SECONDI_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_SECONDJ1_INT32", (void *)_cffi_var_GxB_MIN_SECONDJ1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_SECONDJ1_INT64", (void *)_cffi_var_GxB_MIN_SECONDJ1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_SECONDJ_INT32", (void *)_cffi_var_GxB_MIN_SECONDJ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MIN_SECONDJ_INT64", (void *)_cffi_var_GxB_MIN_SECONDJ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_MODE", (void *)_cffi_const_GxB_MODE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_Matrix_Iterator_attach", (void *)_cffi_f_GxB_Matrix_Iterator_attach, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 549), (void *)_cffi_d_GxB_Matrix_Iterator_attach },
  { "GxB_Matrix_Iterator_getIndex", (void *)_cffi_f_GxB_Matrix_Iterator_getIndex, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3950), (void *)_cffi_d_GxB_Matrix_Iterator_getIndex },
  { "GxB_Matrix_Iterator_getp", (void *)_cffi_f_GxB_Matrix_Iterator_getp, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3933), (void *)_cffi_d_GxB_Matrix_Iterator_getp },
  { "GxB_Matrix_Iterator_getpmax", (void *)_cffi_f_GxB_Matrix_Iterator_getpmax, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3933), (void *)_cffi_d_GxB_Matrix_Iterator_getpmax },
  { "GxB_Matrix_Iterator_next", (void *)_cffi_f_GxB_Matrix_Iterator_next, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 540), (void *)_cffi_d_GxB_Matrix_Iterator_next },
  { "GxB_Matrix_Iterator_seek", (void *)_cffi_f_GxB_Matrix_Iterator_seek, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 559), (void *)_cffi_d_GxB_Matrix_Iterator_seek },
  { "GxB_Matrix_Option_get", (void *)_cffi_const_GxB_Matrix_Option_get, _CFFI_OP(_CFFI_OP_CONSTANT, 3981), (void *)0 },
  { "GxB_Matrix_Option_get_FP64", (void *)_cffi_f_GxB_Matrix_Option_get_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 921), (void *)_cffi_d_GxB_Matrix_Option_get_FP64 },
  { "GxB_Matrix_Option_get_INT32", (void *)_cffi_f_GxB_Matrix_Option_get_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 931), (void *)_cffi_d_GxB_Matrix_Option_get_INT32 },
  { "GxB_Matrix_Option_set", (void *)_cffi_const_GxB_Matrix_Option_set, _CFFI_OP(_CFFI_OP_CONSTANT, 3981), (void *)0 },
  { "GxB_Matrix_Option_set_FP64", (void *)_cffi_f_GxB_Matrix_Option_set_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 926), (void *)_cffi_d_GxB_Matrix_Option_set_FP64 },
  { "GxB_Matrix_Option_set_INT32", (void *)_cffi_f_GxB_Matrix_Option_set_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 936), (void *)_cffi_d_GxB_Matrix_Option_set_INT32 },
  { "GxB_Matrix_apply_BinaryOp1st", (void *)_cffi_f_GxB_Matrix_apply_BinaryOp1st, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1375), (void *)_cffi_d_GxB_Matrix_apply_BinaryOp1st },
  { "GxB_Matrix_apply_BinaryOp1st_FC32", (void *)_cffi_f_GxB_Matrix_apply_BinaryOp1st_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1166), (void *)_cffi_d_GxB_Matrix_apply_BinaryOp1st_FC32 },
  { "GxB_Matrix_apply_BinaryOp1st_FC64", (void *)_cffi_f_GxB_Matrix_apply_BinaryOp1st_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1148), (void *)_cffi_d_GxB_Matrix_apply_BinaryOp1st_FC64 },
  { "GxB_Matrix_apply_BinaryOp2nd", (void *)_cffi_f_GxB_Matrix_apply_BinaryOp2nd, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1310), (void *)_cffi_d_GxB_Matrix_apply_BinaryOp2nd },
  { "GxB_Matrix_apply_BinaryOp2nd_FC32", (void *)_cffi_f_GxB_Matrix_apply_BinaryOp2nd_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1247), (void *)_cffi_d_GxB_Matrix_apply_BinaryOp2nd_FC32 },
  { "GxB_Matrix_apply_BinaryOp2nd_FC64", (void *)_cffi_f_GxB_Matrix_apply_BinaryOp2nd_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1229), (void *)_cffi_d_GxB_Matrix_apply_BinaryOp2nd_FC64 },
  { "GxB_Matrix_apply_IndexOp_FC32", (void *)_cffi_f_GxB_Matrix_apply_IndexOp_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1456), (void *)_cffi_d_GxB_Matrix_apply_IndexOp_FC32 },
  { "GxB_Matrix_apply_IndexOp_FC64", (void *)_cffi_f_GxB_Matrix_apply_IndexOp_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1438), (void *)_cffi_d_GxB_Matrix_apply_IndexOp_FC64 },
  { "GxB_Matrix_assign_FC32", (void *)_cffi_f_GxB_Matrix_assign_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1073), (void *)_cffi_d_GxB_Matrix_assign_FC32 },
  { "GxB_Matrix_assign_FC64", (void *)_cffi_f_GxB_Matrix_assign_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1051), (void *)_cffi_d_GxB_Matrix_assign_FC64 },
  { "GxB_Matrix_build_FC32", (void *)_cffi_f_GxB_Matrix_build_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1805), (void *)_cffi_d_GxB_Matrix_build_FC32 },
  { "GxB_Matrix_build_FC64", (void *)_cffi_f_GxB_Matrix_build_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1789), (void *)_cffi_d_GxB_Matrix_build_FC64 },
  { "GxB_Matrix_build_Scalar", (void *)_cffi_f_GxB_Matrix_build_Scalar, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1853), (void *)_cffi_d_GxB_Matrix_build_Scalar },
  { "GxB_Matrix_concat", (void *)_cffi_f_GxB_Matrix_concat, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1033), (void *)_cffi_d_GxB_Matrix_concat },
  { "GxB_Matrix_deserialize", (void *)_cffi_f_GxB_Matrix_deserialize, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 889), (void *)_cffi_d_GxB_Matrix_deserialize },
  { "GxB_Matrix_diag", (void *)_cffi_f_GxB_Matrix_diag, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1689), (void *)_cffi_d_GxB_Matrix_diag },
  { "GxB_Matrix_eWiseUnion", (void *)_cffi_f_GxB_Matrix_eWiseUnion, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1319), (void *)_cffi_d_GxB_Matrix_eWiseUnion },
  { "GxB_Matrix_export_BitmapC", (void *)_cffi_f_GxB_Matrix_export_BitmapC, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 583), (void *)_cffi_d_GxB_Matrix_export_BitmapC },
  { "GxB_Matrix_export_BitmapR", (void *)_cffi_f_GxB_Matrix_export_BitmapR, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 583), (void *)_cffi_d_GxB_Matrix_export_BitmapR },
  { "GxB_Matrix_export_CSC", (void *)_cffi_f_GxB_Matrix_export_CSC, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 614), (void *)_cffi_d_GxB_Matrix_export_CSC },
  { "GxB_Matrix_export_CSR", (void *)_cffi_f_GxB_Matrix_export_CSR, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 614), (void *)_cffi_d_GxB_Matrix_export_CSR },
  { "GxB_Matrix_export_FC32", (void *)_cffi_f_GxB_Matrix_export_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3621), (void *)_cffi_d_GxB_Matrix_export_FC32 },
  { "GxB_Matrix_export_FC64", (void *)_cffi_f_GxB_Matrix_export_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3587), (void *)_cffi_d_GxB_Matrix_export_FC64 },
  { "GxB_Matrix_export_FullC", (void *)_cffi_f_GxB_Matrix_export_FullC, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 629), (void *)_cffi_d_GxB_Matrix_export_FullC },
  { "GxB_Matrix_export_FullR", (void *)_cffi_f_GxB_Matrix_export_FullR, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 629), (void *)_cffi_d_GxB_Matrix_export_FullR },
  { "GxB_Matrix_export_HyperCSC", (void *)_cffi_f_GxB_Matrix_export_HyperCSC, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 596), (void *)_cffi_d_GxB_Matrix_export_HyperCSC },
  { "GxB_Matrix_export_HyperCSR", (void *)_cffi_f_GxB_Matrix_export_HyperCSR, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 596), (void *)_cffi_d_GxB_Matrix_export_HyperCSR },
  { "GxB_Matrix_extractElement_FC32", (void *)_cffi_f_GxB_Matrix_extractElement_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 274), (void *)_cffi_d_GxB_Matrix_extractElement_FC32 },
  { "GxB_Matrix_extractElement_FC64", (void *)_cffi_f_GxB_Matrix_extractElement_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 216), (void *)_cffi_d_GxB_Matrix_extractElement_FC64 },
  { "GxB_Matrix_extractTuples_FC32", (void *)_cffi_f_GxB_Matrix_extractTuples_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3614), (void *)_cffi_d_GxB_Matrix_extractTuples_FC32 },
  { "GxB_Matrix_extractTuples_FC64", (void *)_cffi_f_GxB_Matrix_extractTuples_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3580), (void *)_cffi_d_GxB_Matrix_extractTuples_FC64 },
  { "GxB_Matrix_fprint", (void *)_cffi_f_GxB_Matrix_fprint, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 954), (void *)_cffi_d_GxB_Matrix_fprint },
  { "GxB_Matrix_import_BitmapC", (void *)_cffi_f_GxB_Matrix_import_BitmapC, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 645), (void *)_cffi_d_GxB_Matrix_import_BitmapC },
  { "GxB_Matrix_import_BitmapR", (void *)_cffi_f_GxB_Matrix_import_BitmapR, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 645), (void *)_cffi_d_GxB_Matrix_import_BitmapR },
  { "GxB_Matrix_import_CSC", (void *)_cffi_f_GxB_Matrix_import_CSC, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 676), (void *)_cffi_d_GxB_Matrix_import_CSC },
  { "GxB_Matrix_import_CSR", (void *)_cffi_f_GxB_Matrix_import_CSR, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 676), (void *)_cffi_d_GxB_Matrix_import_CSR },
  { "GxB_Matrix_import_FC32", (void *)_cffi_f_GxB_Matrix_import_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 730), (void *)_cffi_d_GxB_Matrix_import_FC32 },
  { "GxB_Matrix_import_FC64", (void *)_cffi_f_GxB_Matrix_import_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 704), (void *)_cffi_d_GxB_Matrix_import_FC64 },
  { "GxB_Matrix_import_FullC", (void *)_cffi_f_GxB_Matrix_import_FullC, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 873), (void *)_cffi_d_GxB_Matrix_import_FullC },
  { "GxB_Matrix_import_FullR", (void *)_cffi_f_GxB_Matrix_import_FullR, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 873), (void *)_cffi_d_GxB_Matrix_import_FullR },
  { "GxB_Matrix_import_HyperCSC", (void *)_cffi_f_GxB_Matrix_import_HyperCSC, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 658), (void *)_cffi_d_GxB_Matrix_import_HyperCSC },
  { "GxB_Matrix_import_HyperCSR", (void *)_cffi_f_GxB_Matrix_import_HyperCSR, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 658), (void *)_cffi_d_GxB_Matrix_import_HyperCSR },
  { "GxB_Matrix_isStoredElement", (void *)_cffi_f_GxB_Matrix_isStoredElement, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1900), (void *)_cffi_d_GxB_Matrix_isStoredElement },
  { "GxB_Matrix_iso", (void *)_cffi_f_GxB_Matrix_iso, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 68), (void *)_cffi_d_GxB_Matrix_iso },
  { "GxB_Matrix_memoryUsage", (void *)_cffi_f_GxB_Matrix_memoryUsage, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 405), (void *)_cffi_d_GxB_Matrix_memoryUsage },
  { "GxB_Matrix_pack_BitmapC", (void *)_cffi_f_GxB_Matrix_pack_BitmapC, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1012), (void *)_cffi_d_GxB_Matrix_pack_BitmapC },
  { "GxB_Matrix_pack_BitmapR", (void *)_cffi_f_GxB_Matrix_pack_BitmapR, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1012), (void *)_cffi_d_GxB_Matrix_pack_BitmapR },
  { "GxB_Matrix_pack_CSC", (void *)_cffi_f_GxB_Matrix_pack_CSC, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1769), (void *)_cffi_d_GxB_Matrix_pack_CSC },
  { "GxB_Matrix_pack_CSR", (void *)_cffi_f_GxB_Matrix_pack_CSR, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1769), (void *)_cffi_d_GxB_Matrix_pack_CSR },
  { "GxB_Matrix_pack_FullC", (void *)_cffi_f_GxB_Matrix_pack_FullC, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1924), (void *)_cffi_d_GxB_Matrix_pack_FullC },
  { "GxB_Matrix_pack_FullR", (void *)_cffi_f_GxB_Matrix_pack_FullR, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1924), (void *)_cffi_d_GxB_Matrix_pack_FullR },
  { "GxB_Matrix_pack_HyperCSC", (void *)_cffi_f_GxB_Matrix_pack_HyperCSC, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1742), (void *)_cffi_d_GxB_Matrix_pack_HyperCSC },
  { "GxB_Matrix_pack_HyperCSR", (void *)_cffi_f_GxB_Matrix_pack_HyperCSR, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1742), (void *)_cffi_d_GxB_Matrix_pack_HyperCSR },
  { "GxB_Matrix_reduce_FC32", (void *)_cffi_f_GxB_Matrix_reduce_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 260), (void *)_cffi_d_GxB_Matrix_reduce_FC32 },
  { "GxB_Matrix_reduce_FC64", (void *)_cffi_f_GxB_Matrix_reduce_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 202), (void *)_cffi_d_GxB_Matrix_reduce_FC64 },
  { "GxB_Matrix_reshape", (void *)_cffi_f_GxB_Matrix_reshape, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 947), (void *)_cffi_d_GxB_Matrix_reshape },
  { "GxB_Matrix_reshapeDup", (void *)_cffi_f_GxB_Matrix_reshapeDup, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 575), (void *)_cffi_d_GxB_Matrix_reshapeDup },
  { "GxB_Matrix_select", (void *)_cffi_f_GxB_Matrix_select, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1602), (void *)_cffi_d_GxB_Matrix_select },
  { "GxB_Matrix_select_FC32", (void *)_cffi_f_GxB_Matrix_select_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1456), (void *)_cffi_d_GxB_Matrix_select_FC32 },
  { "GxB_Matrix_select_FC64", (void *)_cffi_f_GxB_Matrix_select_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1438), (void *)_cffi_d_GxB_Matrix_select_FC64 },
  { "GxB_Matrix_serialize", (void *)_cffi_f_GxB_Matrix_serialize, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3842), (void *)_cffi_d_GxB_Matrix_serialize },
  { "GxB_Matrix_setElement_FC32", (void *)_cffi_f_GxB_Matrix_setElement_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 972), (void *)_cffi_d_GxB_Matrix_setElement_FC32 },
  { "GxB_Matrix_setElement_FC64", (void *)_cffi_f_GxB_Matrix_setElement_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 960), (void *)_cffi_d_GxB_Matrix_setElement_FC64 },
  { "GxB_Matrix_sort", (void *)_cffi_f_GxB_Matrix_sort, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1564), (void *)_cffi_d_GxB_Matrix_sort },
  { "GxB_Matrix_split", (void *)_cffi_f_GxB_Matrix_split, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 901), (void *)_cffi_d_GxB_Matrix_split },
  { "GxB_Matrix_subassign", (void *)_cffi_f_GxB_Matrix_subassign, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1571), (void *)_cffi_d_GxB_Matrix_subassign },
  { "GxB_Matrix_subassign_BOOL", (void *)_cffi_f_GxB_Matrix_subassign_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1040), (void *)_cffi_d_GxB_Matrix_subassign_BOOL },
  { "GxB_Matrix_subassign_FC32", (void *)_cffi_f_GxB_Matrix_subassign_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1073), (void *)_cffi_d_GxB_Matrix_subassign_FC32 },
  { "GxB_Matrix_subassign_FC64", (void *)_cffi_f_GxB_Matrix_subassign_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1051), (void *)_cffi_d_GxB_Matrix_subassign_FC64 },
  { "GxB_Matrix_subassign_FP32", (void *)_cffi_f_GxB_Matrix_subassign_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1084), (void *)_cffi_d_GxB_Matrix_subassign_FP32 },
  { "GxB_Matrix_subassign_FP64", (void *)_cffi_f_GxB_Matrix_subassign_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1062), (void *)_cffi_d_GxB_Matrix_subassign_FP64 },
  { "GxB_Matrix_subassign_INT16", (void *)_cffi_f_GxB_Matrix_subassign_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1095), (void *)_cffi_d_GxB_Matrix_subassign_INT16 },
  { "GxB_Matrix_subassign_INT32", (void *)_cffi_f_GxB_Matrix_subassign_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1106), (void *)_cffi_d_GxB_Matrix_subassign_INT32 },
  { "GxB_Matrix_subassign_INT64", (void *)_cffi_f_GxB_Matrix_subassign_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1117), (void *)_cffi_d_GxB_Matrix_subassign_INT64 },
  { "GxB_Matrix_subassign_INT8", (void *)_cffi_f_GxB_Matrix_subassign_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1128), (void *)_cffi_d_GxB_Matrix_subassign_INT8 },
  { "GxB_Matrix_subassign_Scalar", (void *)_cffi_f_GxB_Matrix_subassign_Scalar, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1591), (void *)_cffi_d_GxB_Matrix_subassign_Scalar },
  { "GxB_Matrix_subassign_UDT", (void *)_cffi_f_GxB_Matrix_subassign_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1672), (void *)_cffi_d_GxB_Matrix_subassign_UDT },
  { "GxB_Matrix_subassign_UINT16", (void *)_cffi_f_GxB_Matrix_subassign_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1628), (void *)_cffi_d_GxB_Matrix_subassign_UINT16 },
  { "GxB_Matrix_subassign_UINT32", (void *)_cffi_f_GxB_Matrix_subassign_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1639), (void *)_cffi_d_GxB_Matrix_subassign_UINT32 },
  { "GxB_Matrix_subassign_UINT64", (void *)_cffi_f_GxB_Matrix_subassign_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1650), (void *)_cffi_d_GxB_Matrix_subassign_UINT64 },
  { "GxB_Matrix_subassign_UINT8", (void *)_cffi_f_GxB_Matrix_subassign_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1661), (void *)_cffi_d_GxB_Matrix_subassign_UINT8 },
  { "GxB_Matrix_type", (void *)_cffi_f_GxB_Matrix_type, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2297), (void *)_cffi_d_GxB_Matrix_type },
  { "GxB_Matrix_type_name", (void *)_cffi_f_GxB_Matrix_type_name, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 104), (void *)_cffi_d_GxB_Matrix_type_name },
  { "GxB_Matrix_unpack_BitmapC", (void *)_cffi_f_GxB_Matrix_unpack_BitmapC, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1002), (void *)_cffi_d_GxB_Matrix_unpack_BitmapC },
  { "GxB_Matrix_unpack_BitmapR", (void *)_cffi_f_GxB_Matrix_unpack_BitmapR, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1002), (void *)_cffi_d_GxB_Matrix_unpack_BitmapR },
  { "GxB_Matrix_unpack_CSC", (void *)_cffi_f_GxB_Matrix_unpack_CSC, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1757), (void *)_cffi_d_GxB_Matrix_unpack_CSC },
  { "GxB_Matrix_unpack_CSR", (void *)_cffi_f_GxB_Matrix_unpack_CSR, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1757), (void *)_cffi_d_GxB_Matrix_unpack_CSR },
  { "GxB_Matrix_unpack_FullC", (void *)_cffi_f_GxB_Matrix_unpack_FullC, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1917), (void *)_cffi_d_GxB_Matrix_unpack_FullC },
  { "GxB_Matrix_unpack_FullR", (void *)_cffi_f_GxB_Matrix_unpack_FullR, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1917), (void *)_cffi_d_GxB_Matrix_unpack_FullR },
  { "GxB_Matrix_unpack_HyperCSC", (void *)_cffi_f_GxB_Matrix_unpack_HyperCSC, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1727), (void *)_cffi_d_GxB_Matrix_unpack_HyperCSC },
  { "GxB_Matrix_unpack_HyperCSR", (void *)_cffi_f_GxB_Matrix_unpack_HyperCSR, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1727), (void *)_cffi_d_GxB_Matrix_unpack_HyperCSR },
  { "GxB_Monoid_fprint", (void *)_cffi_f_GxB_Monoid_fprint, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2102), (void *)_cffi_d_GxB_Monoid_fprint },
  { "GxB_Monoid_identity", (void *)_cffi_f_GxB_Monoid_identity, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3874), (void *)_cffi_d_GxB_Monoid_identity },
  { "GxB_Monoid_new_FC32", (void *)_cffi_f_GxB_Monoid_new_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1973), (void *)_cffi_d_GxB_Monoid_new_FC32 },
  { "GxB_Monoid_new_FC64", (void *)_cffi_f_GxB_Monoid_new_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1951), (void *)_cffi_d_GxB_Monoid_new_FC64 },
  { "GxB_Monoid_operator", (void *)_cffi_f_GxB_Monoid_operator, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 424), (void *)_cffi_d_GxB_Monoid_operator },
  { "GxB_Monoid_terminal", (void *)_cffi_f_GxB_Monoid_terminal, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 91), (void *)_cffi_d_GxB_Monoid_terminal },
  { "GxB_Monoid_terminal_new_BOOL", (void *)_cffi_f_GxB_Monoid_terminal_new_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1945), (void *)_cffi_d_GxB_Monoid_terminal_new_BOOL },
  { "GxB_Monoid_terminal_new_FC32", (void *)_cffi_f_GxB_Monoid_terminal_new_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1978), (void *)_cffi_d_GxB_Monoid_terminal_new_FC32 },
  { "GxB_Monoid_terminal_new_FC64", (void *)_cffi_f_GxB_Monoid_terminal_new_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1956), (void *)_cffi_d_GxB_Monoid_terminal_new_FC64 },
  { "GxB_Monoid_terminal_new_FP32", (void *)_cffi_f_GxB_Monoid_terminal_new_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1989), (void *)_cffi_d_GxB_Monoid_terminal_new_FP32 },
  { "GxB_Monoid_terminal_new_FP64", (void *)_cffi_f_GxB_Monoid_terminal_new_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1967), (void *)_cffi_d_GxB_Monoid_terminal_new_FP64 },
  { "GxB_Monoid_terminal_new_INT16", (void *)_cffi_f_GxB_Monoid_terminal_new_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2000), (void *)_cffi_d_GxB_Monoid_terminal_new_INT16 },
  { "GxB_Monoid_terminal_new_INT32", (void *)_cffi_f_GxB_Monoid_terminal_new_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2011), (void *)_cffi_d_GxB_Monoid_terminal_new_INT32 },
  { "GxB_Monoid_terminal_new_INT64", (void *)_cffi_f_GxB_Monoid_terminal_new_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2022), (void *)_cffi_d_GxB_Monoid_terminal_new_INT64 },
  { "GxB_Monoid_terminal_new_INT8", (void *)_cffi_f_GxB_Monoid_terminal_new_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2033), (void *)_cffi_d_GxB_Monoid_terminal_new_INT8 },
  { "GxB_Monoid_terminal_new_UDT", (void *)_cffi_f_GxB_Monoid_terminal_new_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2088), (void *)_cffi_d_GxB_Monoid_terminal_new_UDT },
  { "GxB_Monoid_terminal_new_UINT16", (void *)_cffi_f_GxB_Monoid_terminal_new_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2044), (void *)_cffi_d_GxB_Monoid_terminal_new_UINT16 },
  { "GxB_Monoid_terminal_new_UINT32", (void *)_cffi_f_GxB_Monoid_terminal_new_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2055), (void *)_cffi_d_GxB_Monoid_terminal_new_UINT32 },
  { "GxB_Monoid_terminal_new_UINT64", (void *)_cffi_f_GxB_Monoid_terminal_new_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2066), (void *)_cffi_d_GxB_Monoid_terminal_new_UINT64 },
  { "GxB_Monoid_terminal_new_UINT8", (void *)_cffi_f_GxB_Monoid_terminal_new_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2077), (void *)_cffi_d_GxB_Monoid_terminal_new_UINT8 },
  { "GxB_NBITMAP_SWITCH", (void *)_cffi_const_GxB_NBITMAP_SWITCH, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_NEVER_HYPER", (void *)_cffi_const_GxB_NEVER_HYPER, _CFFI_OP(_CFFI_OP_CONSTANT, 32), (void *)0 },
  { "GxB_NE_FC32", (void *)_cffi_var_GxB_NE_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_NE_FC64", (void *)_cffi_var_GxB_NE_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_NE_THUNK", (void *)_cffi_var_GxB_NE_THUNK, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 155), (void *)0 },
  { "GxB_NONBLOCKING_GPU", (void *)_cffi_const_GxB_NONBLOCKING_GPU, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_NONZERO", (void *)_cffi_var_GxB_NONZERO, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 155), (void *)0 },
  { "GxB_NO_FORMAT", (void *)_cffi_const_GxB_NO_FORMAT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_NTHREADS", (void *)_cffi_const_GxB_NTHREADS, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_OFFDIAG", (void *)_cffi_var_GxB_OFFDIAG, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 155), (void *)0 },
  { "GxB_ONEB_FC32", (void *)_cffi_var_GxB_ONEB_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ONEB_FC64", (void *)_cffi_var_GxB_ONEB_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ONE_BOOL", (void *)_cffi_var_GxB_ONE_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ONE_FC32", (void *)_cffi_var_GxB_ONE_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ONE_FC64", (void *)_cffi_var_GxB_ONE_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ONE_FP32", (void *)_cffi_var_GxB_ONE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ONE_FP64", (void *)_cffi_var_GxB_ONE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ONE_INT16", (void *)_cffi_var_GxB_ONE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ONE_INT32", (void *)_cffi_var_GxB_ONE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ONE_INT64", (void *)_cffi_var_GxB_ONE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ONE_INT8", (void *)_cffi_var_GxB_ONE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ONE_UINT16", (void *)_cffi_var_GxB_ONE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ONE_UINT32", (void *)_cffi_var_GxB_ONE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ONE_UINT64", (void *)_cffi_var_GxB_ONE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ONE_UINT8", (void *)_cffi_var_GxB_ONE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_PAIR_BOOL", (void *)_cffi_var_GxB_PAIR_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_PAIR_FC32", (void *)_cffi_var_GxB_PAIR_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_PAIR_FC64", (void *)_cffi_var_GxB_PAIR_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_PAIR_FP32", (void *)_cffi_var_GxB_PAIR_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_PAIR_FP64", (void *)_cffi_var_GxB_PAIR_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_PAIR_INT16", (void *)_cffi_var_GxB_PAIR_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_PAIR_INT32", (void *)_cffi_var_GxB_PAIR_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_PAIR_INT64", (void *)_cffi_var_GxB_PAIR_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_PAIR_INT8", (void *)_cffi_var_GxB_PAIR_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_PAIR_UINT16", (void *)_cffi_var_GxB_PAIR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_PAIR_UINT32", (void *)_cffi_var_GxB_PAIR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_PAIR_UINT64", (void *)_cffi_var_GxB_PAIR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_PAIR_UINT8", (void *)_cffi_var_GxB_PAIR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_PLUS_DIV_FC32", (void *)_cffi_var_GxB_PLUS_DIV_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_DIV_FC64", (void *)_cffi_var_GxB_PLUS_DIV_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_DIV_FP32", (void *)_cffi_var_GxB_PLUS_DIV_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_DIV_FP64", (void *)_cffi_var_GxB_PLUS_DIV_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_DIV_INT16", (void *)_cffi_var_GxB_PLUS_DIV_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_DIV_INT32", (void *)_cffi_var_GxB_PLUS_DIV_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_DIV_INT64", (void *)_cffi_var_GxB_PLUS_DIV_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_DIV_INT8", (void *)_cffi_var_GxB_PLUS_DIV_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_DIV_UINT16", (void *)_cffi_var_GxB_PLUS_DIV_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_DIV_UINT32", (void *)_cffi_var_GxB_PLUS_DIV_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_DIV_UINT64", (void *)_cffi_var_GxB_PLUS_DIV_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_DIV_UINT8", (void *)_cffi_var_GxB_PLUS_DIV_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_FC32", (void *)_cffi_var_GxB_PLUS_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_PLUS_FC32_MONOID", (void *)_cffi_var_GxB_PLUS_FC32_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_PLUS_FC64", (void *)_cffi_var_GxB_PLUS_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_PLUS_FC64_MONOID", (void *)_cffi_var_GxB_PLUS_FC64_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_PLUS_FIRSTI1_INT32", (void *)_cffi_var_GxB_PLUS_FIRSTI1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_FIRSTI1_INT64", (void *)_cffi_var_GxB_PLUS_FIRSTI1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_FIRSTI_INT32", (void *)_cffi_var_GxB_PLUS_FIRSTI_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_FIRSTI_INT64", (void *)_cffi_var_GxB_PLUS_FIRSTI_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_FIRSTJ1_INT32", (void *)_cffi_var_GxB_PLUS_FIRSTJ1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_FIRSTJ1_INT64", (void *)_cffi_var_GxB_PLUS_FIRSTJ1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_FIRSTJ_INT32", (void *)_cffi_var_GxB_PLUS_FIRSTJ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_FIRSTJ_INT64", (void *)_cffi_var_GxB_PLUS_FIRSTJ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_FIRST_FC32", (void *)_cffi_var_GxB_PLUS_FIRST_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_FIRST_FC64", (void *)_cffi_var_GxB_PLUS_FIRST_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_FIRST_FP32", (void *)_cffi_var_GxB_PLUS_FIRST_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_FIRST_FP64", (void *)_cffi_var_GxB_PLUS_FIRST_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_FIRST_INT16", (void *)_cffi_var_GxB_PLUS_FIRST_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_FIRST_INT32", (void *)_cffi_var_GxB_PLUS_FIRST_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_FIRST_INT64", (void *)_cffi_var_GxB_PLUS_FIRST_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_FIRST_INT8", (void *)_cffi_var_GxB_PLUS_FIRST_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_FIRST_UINT16", (void *)_cffi_var_GxB_PLUS_FIRST_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_FIRST_UINT32", (void *)_cffi_var_GxB_PLUS_FIRST_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_FIRST_UINT64", (void *)_cffi_var_GxB_PLUS_FIRST_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_FIRST_UINT8", (void *)_cffi_var_GxB_PLUS_FIRST_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISEQ_FP32", (void *)_cffi_var_GxB_PLUS_ISEQ_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISEQ_FP64", (void *)_cffi_var_GxB_PLUS_ISEQ_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISEQ_INT16", (void *)_cffi_var_GxB_PLUS_ISEQ_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISEQ_INT32", (void *)_cffi_var_GxB_PLUS_ISEQ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISEQ_INT64", (void *)_cffi_var_GxB_PLUS_ISEQ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISEQ_INT8", (void *)_cffi_var_GxB_PLUS_ISEQ_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISEQ_UINT16", (void *)_cffi_var_GxB_PLUS_ISEQ_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISEQ_UINT32", (void *)_cffi_var_GxB_PLUS_ISEQ_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISEQ_UINT64", (void *)_cffi_var_GxB_PLUS_ISEQ_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISEQ_UINT8", (void *)_cffi_var_GxB_PLUS_ISEQ_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISGE_FP32", (void *)_cffi_var_GxB_PLUS_ISGE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISGE_FP64", (void *)_cffi_var_GxB_PLUS_ISGE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISGE_INT16", (void *)_cffi_var_GxB_PLUS_ISGE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISGE_INT32", (void *)_cffi_var_GxB_PLUS_ISGE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISGE_INT64", (void *)_cffi_var_GxB_PLUS_ISGE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISGE_INT8", (void *)_cffi_var_GxB_PLUS_ISGE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISGE_UINT16", (void *)_cffi_var_GxB_PLUS_ISGE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISGE_UINT32", (void *)_cffi_var_GxB_PLUS_ISGE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISGE_UINT64", (void *)_cffi_var_GxB_PLUS_ISGE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISGE_UINT8", (void *)_cffi_var_GxB_PLUS_ISGE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISGT_FP32", (void *)_cffi_var_GxB_PLUS_ISGT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISGT_FP64", (void *)_cffi_var_GxB_PLUS_ISGT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISGT_INT16", (void *)_cffi_var_GxB_PLUS_ISGT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISGT_INT32", (void *)_cffi_var_GxB_PLUS_ISGT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISGT_INT64", (void *)_cffi_var_GxB_PLUS_ISGT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISGT_INT8", (void *)_cffi_var_GxB_PLUS_ISGT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISGT_UINT16", (void *)_cffi_var_GxB_PLUS_ISGT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISGT_UINT32", (void *)_cffi_var_GxB_PLUS_ISGT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISGT_UINT64", (void *)_cffi_var_GxB_PLUS_ISGT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISGT_UINT8", (void *)_cffi_var_GxB_PLUS_ISGT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISLE_FP32", (void *)_cffi_var_GxB_PLUS_ISLE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISLE_FP64", (void *)_cffi_var_GxB_PLUS_ISLE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISLE_INT16", (void *)_cffi_var_GxB_PLUS_ISLE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISLE_INT32", (void *)_cffi_var_GxB_PLUS_ISLE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISLE_INT64", (void *)_cffi_var_GxB_PLUS_ISLE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISLE_INT8", (void *)_cffi_var_GxB_PLUS_ISLE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISLE_UINT16", (void *)_cffi_var_GxB_PLUS_ISLE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISLE_UINT32", (void *)_cffi_var_GxB_PLUS_ISLE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISLE_UINT64", (void *)_cffi_var_GxB_PLUS_ISLE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISLE_UINT8", (void *)_cffi_var_GxB_PLUS_ISLE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISLT_FP32", (void *)_cffi_var_GxB_PLUS_ISLT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISLT_FP64", (void *)_cffi_var_GxB_PLUS_ISLT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISLT_INT16", (void *)_cffi_var_GxB_PLUS_ISLT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISLT_INT32", (void *)_cffi_var_GxB_PLUS_ISLT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISLT_INT64", (void *)_cffi_var_GxB_PLUS_ISLT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISLT_INT8", (void *)_cffi_var_GxB_PLUS_ISLT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISLT_UINT16", (void *)_cffi_var_GxB_PLUS_ISLT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISLT_UINT32", (void *)_cffi_var_GxB_PLUS_ISLT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISLT_UINT64", (void *)_cffi_var_GxB_PLUS_ISLT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISLT_UINT8", (void *)_cffi_var_GxB_PLUS_ISLT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISNE_FP32", (void *)_cffi_var_GxB_PLUS_ISNE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISNE_FP64", (void *)_cffi_var_GxB_PLUS_ISNE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISNE_INT16", (void *)_cffi_var_GxB_PLUS_ISNE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISNE_INT32", (void *)_cffi_var_GxB_PLUS_ISNE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISNE_INT64", (void *)_cffi_var_GxB_PLUS_ISNE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISNE_INT8", (void *)_cffi_var_GxB_PLUS_ISNE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISNE_UINT16", (void *)_cffi_var_GxB_PLUS_ISNE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISNE_UINT32", (void *)_cffi_var_GxB_PLUS_ISNE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISNE_UINT64", (void *)_cffi_var_GxB_PLUS_ISNE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_ISNE_UINT8", (void *)_cffi_var_GxB_PLUS_ISNE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LAND_FP32", (void *)_cffi_var_GxB_PLUS_LAND_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LAND_FP64", (void *)_cffi_var_GxB_PLUS_LAND_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LAND_INT16", (void *)_cffi_var_GxB_PLUS_LAND_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LAND_INT32", (void *)_cffi_var_GxB_PLUS_LAND_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LAND_INT64", (void *)_cffi_var_GxB_PLUS_LAND_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LAND_INT8", (void *)_cffi_var_GxB_PLUS_LAND_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LAND_UINT16", (void *)_cffi_var_GxB_PLUS_LAND_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LAND_UINT32", (void *)_cffi_var_GxB_PLUS_LAND_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LAND_UINT64", (void *)_cffi_var_GxB_PLUS_LAND_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LAND_UINT8", (void *)_cffi_var_GxB_PLUS_LAND_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LOR_FP32", (void *)_cffi_var_GxB_PLUS_LOR_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LOR_FP64", (void *)_cffi_var_GxB_PLUS_LOR_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LOR_INT16", (void *)_cffi_var_GxB_PLUS_LOR_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LOR_INT32", (void *)_cffi_var_GxB_PLUS_LOR_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LOR_INT64", (void *)_cffi_var_GxB_PLUS_LOR_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LOR_INT8", (void *)_cffi_var_GxB_PLUS_LOR_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LOR_UINT16", (void *)_cffi_var_GxB_PLUS_LOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LOR_UINT32", (void *)_cffi_var_GxB_PLUS_LOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LOR_UINT64", (void *)_cffi_var_GxB_PLUS_LOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LOR_UINT8", (void *)_cffi_var_GxB_PLUS_LOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LXOR_FP32", (void *)_cffi_var_GxB_PLUS_LXOR_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LXOR_FP64", (void *)_cffi_var_GxB_PLUS_LXOR_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LXOR_INT16", (void *)_cffi_var_GxB_PLUS_LXOR_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LXOR_INT32", (void *)_cffi_var_GxB_PLUS_LXOR_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LXOR_INT64", (void *)_cffi_var_GxB_PLUS_LXOR_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LXOR_INT8", (void *)_cffi_var_GxB_PLUS_LXOR_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LXOR_UINT16", (void *)_cffi_var_GxB_PLUS_LXOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LXOR_UINT32", (void *)_cffi_var_GxB_PLUS_LXOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LXOR_UINT64", (void *)_cffi_var_GxB_PLUS_LXOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_LXOR_UINT8", (void *)_cffi_var_GxB_PLUS_LXOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_MAX_FP32", (void *)_cffi_var_GxB_PLUS_MAX_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_MAX_FP64", (void *)_cffi_var_GxB_PLUS_MAX_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_MAX_INT16", (void *)_cffi_var_GxB_PLUS_MAX_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_MAX_INT32", (void *)_cffi_var_GxB_PLUS_MAX_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_MAX_INT64", (void *)_cffi_var_GxB_PLUS_MAX_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_MAX_INT8", (void *)_cffi_var_GxB_PLUS_MAX_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_MAX_UINT16", (void *)_cffi_var_GxB_PLUS_MAX_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_MAX_UINT32", (void *)_cffi_var_GxB_PLUS_MAX_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_MAX_UINT64", (void *)_cffi_var_GxB_PLUS_MAX_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_MAX_UINT8", (void *)_cffi_var_GxB_PLUS_MAX_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_MINUS_FC32", (void *)_cffi_var_GxB_PLUS_MINUS_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_MINUS_FC64", (void *)_cffi_var_GxB_PLUS_MINUS_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_MINUS_FP32", (void *)_cffi_var_GxB_PLUS_MINUS_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_MINUS_FP64", (void *)_cffi_var_GxB_PLUS_MINUS_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_MINUS_INT16", (void *)_cffi_var_GxB_PLUS_MINUS_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_MINUS_INT32", (void *)_cffi_var_GxB_PLUS_MINUS_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_MINUS_INT64", (void *)_cffi_var_GxB_PLUS_MINUS_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_MINUS_INT8", (void *)_cffi_var_GxB_PLUS_MINUS_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_MINUS_UINT16", (void *)_cffi_var_GxB_PLUS_MINUS_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_MINUS_UINT32", (void *)_cffi_var_GxB_PLUS_MINUS_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_MINUS_UINT64", (void *)_cffi_var_GxB_PLUS_MINUS_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_MINUS_UINT8", (void *)_cffi_var_GxB_PLUS_MINUS_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PAIR_FC32", (void *)_cffi_var_GxB_PLUS_PAIR_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PAIR_FC64", (void *)_cffi_var_GxB_PLUS_PAIR_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PAIR_FP32", (void *)_cffi_var_GxB_PLUS_PAIR_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PAIR_FP64", (void *)_cffi_var_GxB_PLUS_PAIR_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PAIR_INT16", (void *)_cffi_var_GxB_PLUS_PAIR_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PAIR_INT32", (void *)_cffi_var_GxB_PLUS_PAIR_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PAIR_INT64", (void *)_cffi_var_GxB_PLUS_PAIR_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PAIR_INT8", (void *)_cffi_var_GxB_PLUS_PAIR_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PAIR_UINT16", (void *)_cffi_var_GxB_PLUS_PAIR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PAIR_UINT32", (void *)_cffi_var_GxB_PLUS_PAIR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PAIR_UINT64", (void *)_cffi_var_GxB_PLUS_PAIR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PAIR_UINT8", (void *)_cffi_var_GxB_PLUS_PAIR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PLUS_FC32", (void *)_cffi_var_GxB_PLUS_PLUS_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PLUS_FC64", (void *)_cffi_var_GxB_PLUS_PLUS_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PLUS_FP32", (void *)_cffi_var_GxB_PLUS_PLUS_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PLUS_FP64", (void *)_cffi_var_GxB_PLUS_PLUS_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PLUS_INT16", (void *)_cffi_var_GxB_PLUS_PLUS_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PLUS_INT32", (void *)_cffi_var_GxB_PLUS_PLUS_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PLUS_INT64", (void *)_cffi_var_GxB_PLUS_PLUS_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PLUS_INT8", (void *)_cffi_var_GxB_PLUS_PLUS_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PLUS_UINT16", (void *)_cffi_var_GxB_PLUS_PLUS_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PLUS_UINT32", (void *)_cffi_var_GxB_PLUS_PLUS_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PLUS_UINT64", (void *)_cffi_var_GxB_PLUS_PLUS_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_PLUS_UINT8", (void *)_cffi_var_GxB_PLUS_PLUS_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RDIV_FC32", (void *)_cffi_var_GxB_PLUS_RDIV_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RDIV_FC64", (void *)_cffi_var_GxB_PLUS_RDIV_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RDIV_FP32", (void *)_cffi_var_GxB_PLUS_RDIV_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RDIV_FP64", (void *)_cffi_var_GxB_PLUS_RDIV_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RDIV_INT16", (void *)_cffi_var_GxB_PLUS_RDIV_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RDIV_INT32", (void *)_cffi_var_GxB_PLUS_RDIV_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RDIV_INT64", (void *)_cffi_var_GxB_PLUS_RDIV_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RDIV_INT8", (void *)_cffi_var_GxB_PLUS_RDIV_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RDIV_UINT16", (void *)_cffi_var_GxB_PLUS_RDIV_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RDIV_UINT32", (void *)_cffi_var_GxB_PLUS_RDIV_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RDIV_UINT64", (void *)_cffi_var_GxB_PLUS_RDIV_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RDIV_UINT8", (void *)_cffi_var_GxB_PLUS_RDIV_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RMINUS_FC32", (void *)_cffi_var_GxB_PLUS_RMINUS_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RMINUS_FC64", (void *)_cffi_var_GxB_PLUS_RMINUS_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RMINUS_FP32", (void *)_cffi_var_GxB_PLUS_RMINUS_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RMINUS_FP64", (void *)_cffi_var_GxB_PLUS_RMINUS_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RMINUS_INT16", (void *)_cffi_var_GxB_PLUS_RMINUS_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RMINUS_INT32", (void *)_cffi_var_GxB_PLUS_RMINUS_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RMINUS_INT64", (void *)_cffi_var_GxB_PLUS_RMINUS_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RMINUS_INT8", (void *)_cffi_var_GxB_PLUS_RMINUS_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RMINUS_UINT16", (void *)_cffi_var_GxB_PLUS_RMINUS_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RMINUS_UINT32", (void *)_cffi_var_GxB_PLUS_RMINUS_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RMINUS_UINT64", (void *)_cffi_var_GxB_PLUS_RMINUS_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_RMINUS_UINT8", (void *)_cffi_var_GxB_PLUS_RMINUS_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_SECONDI1_INT32", (void *)_cffi_var_GxB_PLUS_SECONDI1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_SECONDI1_INT64", (void *)_cffi_var_GxB_PLUS_SECONDI1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_SECONDI_INT32", (void *)_cffi_var_GxB_PLUS_SECONDI_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_SECONDI_INT64", (void *)_cffi_var_GxB_PLUS_SECONDI_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_SECONDJ1_INT32", (void *)_cffi_var_GxB_PLUS_SECONDJ1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_SECONDJ1_INT64", (void *)_cffi_var_GxB_PLUS_SECONDJ1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_SECONDJ_INT32", (void *)_cffi_var_GxB_PLUS_SECONDJ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_SECONDJ_INT64", (void *)_cffi_var_GxB_PLUS_SECONDJ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_SECOND_FC32", (void *)_cffi_var_GxB_PLUS_SECOND_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_SECOND_FC64", (void *)_cffi_var_GxB_PLUS_SECOND_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_SECOND_FP32", (void *)_cffi_var_GxB_PLUS_SECOND_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_SECOND_FP64", (void *)_cffi_var_GxB_PLUS_SECOND_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_SECOND_INT16", (void *)_cffi_var_GxB_PLUS_SECOND_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_SECOND_INT32", (void *)_cffi_var_GxB_PLUS_SECOND_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_SECOND_INT64", (void *)_cffi_var_GxB_PLUS_SECOND_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_SECOND_INT8", (void *)_cffi_var_GxB_PLUS_SECOND_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_SECOND_UINT16", (void *)_cffi_var_GxB_PLUS_SECOND_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_SECOND_UINT32", (void *)_cffi_var_GxB_PLUS_SECOND_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_SECOND_UINT64", (void *)_cffi_var_GxB_PLUS_SECOND_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_SECOND_UINT8", (void *)_cffi_var_GxB_PLUS_SECOND_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_TIMES_FC32", (void *)_cffi_var_GxB_PLUS_TIMES_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_PLUS_TIMES_FC64", (void *)_cffi_var_GxB_PLUS_TIMES_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_POSITIONI1_INT32", (void *)_cffi_var_GxB_POSITIONI1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_POSITIONI1_INT64", (void *)_cffi_var_GxB_POSITIONI1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_POSITIONI_INT32", (void *)_cffi_var_GxB_POSITIONI_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_POSITIONI_INT64", (void *)_cffi_var_GxB_POSITIONI_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_POSITIONJ1_INT32", (void *)_cffi_var_GxB_POSITIONJ1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_POSITIONJ1_INT64", (void *)_cffi_var_GxB_POSITIONJ1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_POSITIONJ_INT32", (void *)_cffi_var_GxB_POSITIONJ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_POSITIONJ_INT64", (void *)_cffi_var_GxB_POSITIONJ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_POW_BOOL", (void *)_cffi_var_GxB_POW_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_POW_FC32", (void *)_cffi_var_GxB_POW_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_POW_FC64", (void *)_cffi_var_GxB_POW_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_POW_FP32", (void *)_cffi_var_GxB_POW_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_POW_FP64", (void *)_cffi_var_GxB_POW_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_POW_INT16", (void *)_cffi_var_GxB_POW_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_POW_INT32", (void *)_cffi_var_GxB_POW_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_POW_INT64", (void *)_cffi_var_GxB_POW_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_POW_INT8", (void *)_cffi_var_GxB_POW_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_POW_UINT16", (void *)_cffi_var_GxB_POW_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_POW_UINT32", (void *)_cffi_var_GxB_POW_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_POW_UINT64", (void *)_cffi_var_GxB_POW_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_POW_UINT8", (void *)_cffi_var_GxB_POW_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_PRINTF", (void *)_cffi_const_GxB_PRINTF, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_PRINT_1BASED", (void *)_cffi_const_GxB_PRINT_1BASED, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_RANGE", (void *)_cffi_const_GxB_RANGE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_RDIV_BOOL", (void *)_cffi_var_GxB_RDIV_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RDIV_FC32", (void *)_cffi_var_GxB_RDIV_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RDIV_FC64", (void *)_cffi_var_GxB_RDIV_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RDIV_FP32", (void *)_cffi_var_GxB_RDIV_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RDIV_FP64", (void *)_cffi_var_GxB_RDIV_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RDIV_INT16", (void *)_cffi_var_GxB_RDIV_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RDIV_INT32", (void *)_cffi_var_GxB_RDIV_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RDIV_INT64", (void *)_cffi_var_GxB_RDIV_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RDIV_INT8", (void *)_cffi_var_GxB_RDIV_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RDIV_UINT16", (void *)_cffi_var_GxB_RDIV_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RDIV_UINT32", (void *)_cffi_var_GxB_RDIV_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RDIV_UINT64", (void *)_cffi_var_GxB_RDIV_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RDIV_UINT8", (void *)_cffi_var_GxB_RDIV_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_REMAINDER_FP32", (void *)_cffi_var_GxB_REMAINDER_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_REMAINDER_FP64", (void *)_cffi_var_GxB_REMAINDER_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RMINUS_BOOL", (void *)_cffi_var_GxB_RMINUS_BOOL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RMINUS_FC32", (void *)_cffi_var_GxB_RMINUS_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RMINUS_FC64", (void *)_cffi_var_GxB_RMINUS_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RMINUS_FP32", (void *)_cffi_var_GxB_RMINUS_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RMINUS_FP64", (void *)_cffi_var_GxB_RMINUS_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RMINUS_INT16", (void *)_cffi_var_GxB_RMINUS_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RMINUS_INT32", (void *)_cffi_var_GxB_RMINUS_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RMINUS_INT64", (void *)_cffi_var_GxB_RMINUS_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RMINUS_INT8", (void *)_cffi_var_GxB_RMINUS_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RMINUS_UINT16", (void *)_cffi_var_GxB_RMINUS_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RMINUS_UINT32", (void *)_cffi_var_GxB_RMINUS_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RMINUS_UINT64", (void *)_cffi_var_GxB_RMINUS_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_RMINUS_UINT8", (void *)_cffi_var_GxB_RMINUS_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_ROUND_FC32", (void *)_cffi_var_GxB_ROUND_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ROUND_FC64", (void *)_cffi_var_GxB_ROUND_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ROUND_FP32", (void *)_cffi_var_GxB_ROUND_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_ROUND_FP64", (void *)_cffi_var_GxB_ROUND_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_Row_subassign", (void *)_cffi_f_GxB_Row_subassign, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1705), (void *)_cffi_d_GxB_Row_subassign },
  { "GxB_SECONDI1_INT32", (void *)_cffi_var_GxB_SECONDI1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_SECONDI1_INT64", (void *)_cffi_var_GxB_SECONDI1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_SECONDI_INT32", (void *)_cffi_var_GxB_SECONDI_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_SECONDI_INT64", (void *)_cffi_var_GxB_SECONDI_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_SECONDJ1_INT32", (void *)_cffi_var_GxB_SECONDJ1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_SECONDJ1_INT64", (void *)_cffi_var_GxB_SECONDJ1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_SECONDJ_INT32", (void *)_cffi_var_GxB_SECONDJ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_SECONDJ_INT64", (void *)_cffi_var_GxB_SECONDJ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_SECOND_FC32", (void *)_cffi_var_GxB_SECOND_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_SECOND_FC64", (void *)_cffi_var_GxB_SECOND_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_SECURE_IMPORT", (void *)_cffi_const_GxB_SECURE_IMPORT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_SHORT", (void *)_cffi_const_GxB_SHORT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_SHORT_VERBOSE", (void *)_cffi_const_GxB_SHORT_VERBOSE, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_SIGNUM_FC32", (void *)_cffi_var_GxB_SIGNUM_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_SIGNUM_FC64", (void *)_cffi_var_GxB_SIGNUM_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_SIGNUM_FP32", (void *)_cffi_var_GxB_SIGNUM_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_SIGNUM_FP64", (void *)_cffi_var_GxB_SIGNUM_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_SILENT", (void *)_cffi_const_GxB_SILENT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_SINH_FC32", (void *)_cffi_var_GxB_SINH_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_SINH_FC64", (void *)_cffi_var_GxB_SINH_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_SINH_FP32", (void *)_cffi_var_GxB_SINH_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_SINH_FP64", (void *)_cffi_var_GxB_SINH_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_SIN_FC32", (void *)_cffi_var_GxB_SIN_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_SIN_FC64", (void *)_cffi_var_GxB_SIN_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_SIN_FP32", (void *)_cffi_var_GxB_SIN_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_SIN_FP64", (void *)_cffi_var_GxB_SIN_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_SORT", (void *)_cffi_const_GxB_SORT, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_SPARSE", (void *)_cffi_const_GxB_SPARSE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_SPARSITY_CONTROL", (void *)_cffi_const_GxB_SPARSITY_CONTROL, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_SPARSITY_STATUS", (void *)_cffi_const_GxB_SPARSITY_STATUS, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_SPEC_ABOUT_STR", (void *)_cffi_var_GxB_SPEC_ABOUT_STR, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 97), (void *)0 },
  { "GxB_SPEC_DATE_STR", (void *)_cffi_var_GxB_SPEC_DATE_STR, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 97), (void *)0 },
  { "GxB_SPEC_MAJOR", (void *)_cffi_const_GxB_SPEC_MAJOR, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_SPEC_MINOR", (void *)_cffi_const_GxB_SPEC_MINOR, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_SPEC_SUB", (void *)_cffi_const_GxB_SPEC_SUB, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_SPEC_VERSION", (void *)_cffi_const_GxB_SPEC_VERSION, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_SQRT_FC32", (void *)_cffi_var_GxB_SQRT_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_SQRT_FC64", (void *)_cffi_var_GxB_SQRT_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_SQRT_FP32", (void *)_cffi_var_GxB_SQRT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_SQRT_FP64", (void *)_cffi_var_GxB_SQRT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_STDC_VERSION", (void *)_cffi_const_GxB_STDC_VERSION, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_STRIDE", (void *)_cffi_const_GxB_STRIDE, _CFFI_OP(_CFFI_OP_CONSTANT_INT, -1), (void *)0 },
  { "GxB_SUMMARY", (void *)_cffi_const_GxB_SUMMARY, _CFFI_OP(_CFFI_OP_ENUM, -1), (void *)0 },
  { "GxB_Scalar_clear", (void *)_cffi_f_GxB_Scalar_clear, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 2119), (void *)_cffi_d_GxB_Scalar_clear },
  { "GxB_Scalar_dup", (void *)_cffi_f_GxB_Scalar_dup, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2111), (void *)_cffi_d_GxB_Scalar_dup },
  { "GxB_Scalar_error", (void *)_cffi_f_GxB_Scalar_error, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 149), (void *)_cffi_d_GxB_Scalar_error },
  { "GxB_Scalar_extractElement_BOOL", (void *)_cffi_f_GxB_Scalar_extractElement_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 78), (void *)_cffi_d_GxB_Scalar_extractElement_BOOL },
  { "GxB_Scalar_extractElement_FC32", (void *)_cffi_f_GxB_Scalar_extractElement_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 280), (void *)_cffi_d_GxB_Scalar_extractElement_FC32 },
  { "GxB_Scalar_extractElement_FC64", (void *)_cffi_f_GxB_Scalar_extractElement_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 222), (void *)_cffi_d_GxB_Scalar_extractElement_FC64 },
  { "GxB_Scalar_extractElement_FP32", (void *)_cffi_f_GxB_Scalar_extractElement_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 251), (void *)_cffi_d_GxB_Scalar_extractElement_FP32 },
  { "GxB_Scalar_extractElement_FP64", (void *)_cffi_f_GxB_Scalar_extractElement_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 193), (void *)_cffi_d_GxB_Scalar_extractElement_FP64 },
  { "GxB_Scalar_extractElement_INT16", (void *)_cffi_f_GxB_Scalar_extractElement_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 309), (void *)_cffi_d_GxB_Scalar_extractElement_INT16 },
  { "GxB_Scalar_extractElement_INT32", (void *)_cffi_f_GxB_Scalar_extractElement_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 338), (void *)_cffi_d_GxB_Scalar_extractElement_INT32 },
  { "GxB_Scalar_extractElement_INT64", (void *)_cffi_f_GxB_Scalar_extractElement_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 367), (void *)_cffi_d_GxB_Scalar_extractElement_INT64 },
  { "GxB_Scalar_extractElement_INT8", (void *)_cffi_f_GxB_Scalar_extractElement_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 396), (void *)_cffi_d_GxB_Scalar_extractElement_INT8 },
  { "GxB_Scalar_extractElement_UDT", (void *)_cffi_f_GxB_Scalar_extractElement_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3878), (void *)_cffi_d_GxB_Scalar_extractElement_UDT },
  { "GxB_Scalar_extractElement_UINT16", (void *)_cffi_f_GxB_Scalar_extractElement_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3405), (void *)_cffi_d_GxB_Scalar_extractElement_UINT16 },
  { "GxB_Scalar_extractElement_UINT32", (void *)_cffi_f_GxB_Scalar_extractElement_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3434), (void *)_cffi_d_GxB_Scalar_extractElement_UINT32 },
  { "GxB_Scalar_extractElement_UINT64", (void *)_cffi_f_GxB_Scalar_extractElement_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3521), (void *)_cffi_d_GxB_Scalar_extractElement_UINT64 },
  { "GxB_Scalar_extractElement_UINT8", (void *)_cffi_f_GxB_Scalar_extractElement_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3829), (void *)_cffi_d_GxB_Scalar_extractElement_UINT8 },
  { "GxB_Scalar_fprint", (void *)_cffi_f_GxB_Scalar_fprint, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2130), (void *)_cffi_d_GxB_Scalar_fprint },
  { "GxB_Scalar_free", (void *)_cffi_f_GxB_Scalar_free, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 2108), (void *)_cffi_d_GxB_Scalar_free },
  { "GxB_Scalar_memoryUsage", (void *)_cffi_f_GxB_Scalar_memoryUsage, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 409), (void *)_cffi_d_GxB_Scalar_memoryUsage },
  { "GxB_Scalar_new", (void *)_cffi_f_GxB_Scalar_new, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2115), (void *)_cffi_d_GxB_Scalar_new },
  { "GxB_Scalar_nvals", (void *)_cffi_f_GxB_Scalar_nvals, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3521), (void *)_cffi_d_GxB_Scalar_nvals },
  { "GxB_Scalar_setElement_BOOL", (void *)_cffi_f_GxB_Scalar_setElement_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2126), (void *)_cffi_d_GxB_Scalar_setElement_BOOL },
  { "GxB_Scalar_setElement_FC32", (void *)_cffi_f_GxB_Scalar_setElement_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2144), (void *)_cffi_d_GxB_Scalar_setElement_FC32 },
  { "GxB_Scalar_setElement_FC64", (void *)_cffi_f_GxB_Scalar_setElement_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2136), (void *)_cffi_d_GxB_Scalar_setElement_FC64 },
  { "GxB_Scalar_setElement_FP32", (void *)_cffi_f_GxB_Scalar_setElement_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2148), (void *)_cffi_d_GxB_Scalar_setElement_FP32 },
  { "GxB_Scalar_setElement_FP64", (void *)_cffi_f_GxB_Scalar_setElement_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2140), (void *)_cffi_d_GxB_Scalar_setElement_FP64 },
  { "GxB_Scalar_setElement_INT16", (void *)_cffi_f_GxB_Scalar_setElement_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2152), (void *)_cffi_d_GxB_Scalar_setElement_INT16 },
  { "GxB_Scalar_setElement_INT32", (void *)_cffi_f_GxB_Scalar_setElement_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2156), (void *)_cffi_d_GxB_Scalar_setElement_INT32 },
  { "GxB_Scalar_setElement_INT64", (void *)_cffi_f_GxB_Scalar_setElement_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2160), (void *)_cffi_d_GxB_Scalar_setElement_INT64 },
  { "GxB_Scalar_setElement_INT8", (void *)_cffi_f_GxB_Scalar_setElement_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2164), (void *)_cffi_d_GxB_Scalar_setElement_INT8 },
  { "GxB_Scalar_setElement_UDT", (void *)_cffi_f_GxB_Scalar_setElement_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2223), (void *)_cffi_d_GxB_Scalar_setElement_UDT },
  { "GxB_Scalar_setElement_UINT16", (void *)_cffi_f_GxB_Scalar_setElement_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2207), (void *)_cffi_d_GxB_Scalar_setElement_UINT16 },
  { "GxB_Scalar_setElement_UINT32", (void *)_cffi_f_GxB_Scalar_setElement_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2211), (void *)_cffi_d_GxB_Scalar_setElement_UINT32 },
  { "GxB_Scalar_setElement_UINT64", (void *)_cffi_f_GxB_Scalar_setElement_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2215), (void *)_cffi_d_GxB_Scalar_setElement_UINT64 },
  { "GxB_Scalar_setElement_UINT8", (void *)_cffi_f_GxB_Scalar_setElement_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2219), (void *)_cffi_d_GxB_Scalar_setElement_UINT8 },
  { "GxB_Scalar_type", (void *)_cffi_f_GxB_Scalar_type, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2301), (void *)_cffi_d_GxB_Scalar_type },
  { "GxB_Scalar_type_name", (void *)_cffi_f_GxB_Scalar_type_name, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 108), (void *)_cffi_d_GxB_Scalar_type_name },
  { "GxB_Scalar_wait", (void *)_cffi_f_GxB_Scalar_wait, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 2108), (void *)_cffi_d_GxB_Scalar_wait },
  { "GxB_SelectOp_error", (void *)_cffi_f_GxB_SelectOp_error, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 153), (void *)_cffi_d_GxB_SelectOp_error },
  { "GxB_SelectOp_fprint", (void *)_cffi_f_GxB_SelectOp_fprint, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2247), (void *)_cffi_d_GxB_SelectOp_fprint },
  { "GxB_SelectOp_free", (void *)_cffi_f_GxB_SelectOp_free, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 2227), (void *)_cffi_d_GxB_SelectOp_free },
  { "GxB_SelectOp_new", (void *)_cffi_f_GxB_SelectOp_new, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2230), (void *)_cffi_d_GxB_SelectOp_new },
  { "GxB_SelectOp_ttype", (void *)_cffi_f_GxB_SelectOp_ttype, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2305), (void *)_cffi_d_GxB_SelectOp_ttype },
  { "GxB_SelectOp_wait", (void *)_cffi_f_GxB_SelectOp_wait, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2243), (void *)_cffi_d_GxB_SelectOp_wait },
  { "GxB_SelectOp_xtype", (void *)_cffi_f_GxB_SelectOp_xtype, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2305), (void *)_cffi_d_GxB_SelectOp_xtype },
  { "GxB_Semiring_add", (void *)_cffi_f_GxB_Semiring_add, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2094), (void *)_cffi_d_GxB_Semiring_add },
  { "GxB_Semiring_fprint", (void *)_cffi_f_GxB_Semiring_fprint, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2265), (void *)_cffi_d_GxB_Semiring_fprint },
  { "GxB_Semiring_multiply", (void *)_cffi_f_GxB_Semiring_multiply, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 428), (void *)_cffi_d_GxB_Semiring_multiply },
  { "GxB_TANH_FC32", (void *)_cffi_var_GxB_TANH_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_TANH_FC64", (void *)_cffi_var_GxB_TANH_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_TANH_FP32", (void *)_cffi_var_GxB_TANH_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_TANH_FP64", (void *)_cffi_var_GxB_TANH_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_TAN_FC32", (void *)_cffi_var_GxB_TAN_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_TAN_FC64", (void *)_cffi_var_GxB_TAN_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_TAN_FP32", (void *)_cffi_var_GxB_TAN_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_TAN_FP64", (void *)_cffi_var_GxB_TAN_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_TGAMMA_FP32", (void *)_cffi_var_GxB_TGAMMA_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_TGAMMA_FP64", (void *)_cffi_var_GxB_TGAMMA_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_TIMES_DIV_FC32", (void *)_cffi_var_GxB_TIMES_DIV_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_DIV_FC64", (void *)_cffi_var_GxB_TIMES_DIV_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_DIV_FP32", (void *)_cffi_var_GxB_TIMES_DIV_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_DIV_FP64", (void *)_cffi_var_GxB_TIMES_DIV_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_DIV_INT16", (void *)_cffi_var_GxB_TIMES_DIV_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_DIV_INT32", (void *)_cffi_var_GxB_TIMES_DIV_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_DIV_INT64", (void *)_cffi_var_GxB_TIMES_DIV_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_DIV_INT8", (void *)_cffi_var_GxB_TIMES_DIV_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_DIV_UINT16", (void *)_cffi_var_GxB_TIMES_DIV_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_DIV_UINT32", (void *)_cffi_var_GxB_TIMES_DIV_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_DIV_UINT64", (void *)_cffi_var_GxB_TIMES_DIV_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_DIV_UINT8", (void *)_cffi_var_GxB_TIMES_DIV_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_FC32", (void *)_cffi_var_GxB_TIMES_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_TIMES_FC32_MONOID", (void *)_cffi_var_GxB_TIMES_FC32_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_TIMES_FC64", (void *)_cffi_var_GxB_TIMES_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 56), (void *)0 },
  { "GxB_TIMES_FC64_MONOID", (void *)_cffi_var_GxB_TIMES_FC64_MONOID, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 57), (void *)0 },
  { "GxB_TIMES_FIRSTI1_INT32", (void *)_cffi_var_GxB_TIMES_FIRSTI1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_FIRSTI1_INT64", (void *)_cffi_var_GxB_TIMES_FIRSTI1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_FIRSTI_INT32", (void *)_cffi_var_GxB_TIMES_FIRSTI_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_FIRSTI_INT64", (void *)_cffi_var_GxB_TIMES_FIRSTI_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_FIRSTJ1_INT32", (void *)_cffi_var_GxB_TIMES_FIRSTJ1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_FIRSTJ1_INT64", (void *)_cffi_var_GxB_TIMES_FIRSTJ1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_FIRSTJ_INT32", (void *)_cffi_var_GxB_TIMES_FIRSTJ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_FIRSTJ_INT64", (void *)_cffi_var_GxB_TIMES_FIRSTJ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_FIRST_FC32", (void *)_cffi_var_GxB_TIMES_FIRST_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_FIRST_FC64", (void *)_cffi_var_GxB_TIMES_FIRST_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_FIRST_FP32", (void *)_cffi_var_GxB_TIMES_FIRST_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_FIRST_FP64", (void *)_cffi_var_GxB_TIMES_FIRST_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_FIRST_INT16", (void *)_cffi_var_GxB_TIMES_FIRST_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_FIRST_INT32", (void *)_cffi_var_GxB_TIMES_FIRST_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_FIRST_INT64", (void *)_cffi_var_GxB_TIMES_FIRST_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_FIRST_INT8", (void *)_cffi_var_GxB_TIMES_FIRST_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_FIRST_UINT16", (void *)_cffi_var_GxB_TIMES_FIRST_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_FIRST_UINT32", (void *)_cffi_var_GxB_TIMES_FIRST_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_FIRST_UINT64", (void *)_cffi_var_GxB_TIMES_FIRST_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_FIRST_UINT8", (void *)_cffi_var_GxB_TIMES_FIRST_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISEQ_FP32", (void *)_cffi_var_GxB_TIMES_ISEQ_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISEQ_FP64", (void *)_cffi_var_GxB_TIMES_ISEQ_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISEQ_INT16", (void *)_cffi_var_GxB_TIMES_ISEQ_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISEQ_INT32", (void *)_cffi_var_GxB_TIMES_ISEQ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISEQ_INT64", (void *)_cffi_var_GxB_TIMES_ISEQ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISEQ_INT8", (void *)_cffi_var_GxB_TIMES_ISEQ_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISEQ_UINT16", (void *)_cffi_var_GxB_TIMES_ISEQ_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISEQ_UINT32", (void *)_cffi_var_GxB_TIMES_ISEQ_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISEQ_UINT64", (void *)_cffi_var_GxB_TIMES_ISEQ_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISEQ_UINT8", (void *)_cffi_var_GxB_TIMES_ISEQ_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISGE_FP32", (void *)_cffi_var_GxB_TIMES_ISGE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISGE_FP64", (void *)_cffi_var_GxB_TIMES_ISGE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISGE_INT16", (void *)_cffi_var_GxB_TIMES_ISGE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISGE_INT32", (void *)_cffi_var_GxB_TIMES_ISGE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISGE_INT64", (void *)_cffi_var_GxB_TIMES_ISGE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISGE_INT8", (void *)_cffi_var_GxB_TIMES_ISGE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISGE_UINT16", (void *)_cffi_var_GxB_TIMES_ISGE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISGE_UINT32", (void *)_cffi_var_GxB_TIMES_ISGE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISGE_UINT64", (void *)_cffi_var_GxB_TIMES_ISGE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISGE_UINT8", (void *)_cffi_var_GxB_TIMES_ISGE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISGT_FP32", (void *)_cffi_var_GxB_TIMES_ISGT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISGT_FP64", (void *)_cffi_var_GxB_TIMES_ISGT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISGT_INT16", (void *)_cffi_var_GxB_TIMES_ISGT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISGT_INT32", (void *)_cffi_var_GxB_TIMES_ISGT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISGT_INT64", (void *)_cffi_var_GxB_TIMES_ISGT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISGT_INT8", (void *)_cffi_var_GxB_TIMES_ISGT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISGT_UINT16", (void *)_cffi_var_GxB_TIMES_ISGT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISGT_UINT32", (void *)_cffi_var_GxB_TIMES_ISGT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISGT_UINT64", (void *)_cffi_var_GxB_TIMES_ISGT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISGT_UINT8", (void *)_cffi_var_GxB_TIMES_ISGT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISLE_FP32", (void *)_cffi_var_GxB_TIMES_ISLE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISLE_FP64", (void *)_cffi_var_GxB_TIMES_ISLE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISLE_INT16", (void *)_cffi_var_GxB_TIMES_ISLE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISLE_INT32", (void *)_cffi_var_GxB_TIMES_ISLE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISLE_INT64", (void *)_cffi_var_GxB_TIMES_ISLE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISLE_INT8", (void *)_cffi_var_GxB_TIMES_ISLE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISLE_UINT16", (void *)_cffi_var_GxB_TIMES_ISLE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISLE_UINT32", (void *)_cffi_var_GxB_TIMES_ISLE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISLE_UINT64", (void *)_cffi_var_GxB_TIMES_ISLE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISLE_UINT8", (void *)_cffi_var_GxB_TIMES_ISLE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISLT_FP32", (void *)_cffi_var_GxB_TIMES_ISLT_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISLT_FP64", (void *)_cffi_var_GxB_TIMES_ISLT_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISLT_INT16", (void *)_cffi_var_GxB_TIMES_ISLT_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISLT_INT32", (void *)_cffi_var_GxB_TIMES_ISLT_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISLT_INT64", (void *)_cffi_var_GxB_TIMES_ISLT_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISLT_INT8", (void *)_cffi_var_GxB_TIMES_ISLT_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISLT_UINT16", (void *)_cffi_var_GxB_TIMES_ISLT_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISLT_UINT32", (void *)_cffi_var_GxB_TIMES_ISLT_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISLT_UINT64", (void *)_cffi_var_GxB_TIMES_ISLT_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISLT_UINT8", (void *)_cffi_var_GxB_TIMES_ISLT_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISNE_FP32", (void *)_cffi_var_GxB_TIMES_ISNE_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISNE_FP64", (void *)_cffi_var_GxB_TIMES_ISNE_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISNE_INT16", (void *)_cffi_var_GxB_TIMES_ISNE_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISNE_INT32", (void *)_cffi_var_GxB_TIMES_ISNE_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISNE_INT64", (void *)_cffi_var_GxB_TIMES_ISNE_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISNE_INT8", (void *)_cffi_var_GxB_TIMES_ISNE_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISNE_UINT16", (void *)_cffi_var_GxB_TIMES_ISNE_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISNE_UINT32", (void *)_cffi_var_GxB_TIMES_ISNE_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISNE_UINT64", (void *)_cffi_var_GxB_TIMES_ISNE_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_ISNE_UINT8", (void *)_cffi_var_GxB_TIMES_ISNE_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LAND_FP32", (void *)_cffi_var_GxB_TIMES_LAND_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LAND_FP64", (void *)_cffi_var_GxB_TIMES_LAND_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LAND_INT16", (void *)_cffi_var_GxB_TIMES_LAND_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LAND_INT32", (void *)_cffi_var_GxB_TIMES_LAND_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LAND_INT64", (void *)_cffi_var_GxB_TIMES_LAND_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LAND_INT8", (void *)_cffi_var_GxB_TIMES_LAND_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LAND_UINT16", (void *)_cffi_var_GxB_TIMES_LAND_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LAND_UINT32", (void *)_cffi_var_GxB_TIMES_LAND_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LAND_UINT64", (void *)_cffi_var_GxB_TIMES_LAND_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LAND_UINT8", (void *)_cffi_var_GxB_TIMES_LAND_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LOR_FP32", (void *)_cffi_var_GxB_TIMES_LOR_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LOR_FP64", (void *)_cffi_var_GxB_TIMES_LOR_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LOR_INT16", (void *)_cffi_var_GxB_TIMES_LOR_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LOR_INT32", (void *)_cffi_var_GxB_TIMES_LOR_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LOR_INT64", (void *)_cffi_var_GxB_TIMES_LOR_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LOR_INT8", (void *)_cffi_var_GxB_TIMES_LOR_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LOR_UINT16", (void *)_cffi_var_GxB_TIMES_LOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LOR_UINT32", (void *)_cffi_var_GxB_TIMES_LOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LOR_UINT64", (void *)_cffi_var_GxB_TIMES_LOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LOR_UINT8", (void *)_cffi_var_GxB_TIMES_LOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LXOR_FP32", (void *)_cffi_var_GxB_TIMES_LXOR_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LXOR_FP64", (void *)_cffi_var_GxB_TIMES_LXOR_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LXOR_INT16", (void *)_cffi_var_GxB_TIMES_LXOR_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LXOR_INT32", (void *)_cffi_var_GxB_TIMES_LXOR_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LXOR_INT64", (void *)_cffi_var_GxB_TIMES_LXOR_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LXOR_INT8", (void *)_cffi_var_GxB_TIMES_LXOR_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LXOR_UINT16", (void *)_cffi_var_GxB_TIMES_LXOR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LXOR_UINT32", (void *)_cffi_var_GxB_TIMES_LXOR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LXOR_UINT64", (void *)_cffi_var_GxB_TIMES_LXOR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_LXOR_UINT8", (void *)_cffi_var_GxB_TIMES_LXOR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MAX_FP32", (void *)_cffi_var_GxB_TIMES_MAX_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MAX_FP64", (void *)_cffi_var_GxB_TIMES_MAX_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MAX_INT16", (void *)_cffi_var_GxB_TIMES_MAX_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MAX_INT32", (void *)_cffi_var_GxB_TIMES_MAX_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MAX_INT64", (void *)_cffi_var_GxB_TIMES_MAX_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MAX_INT8", (void *)_cffi_var_GxB_TIMES_MAX_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MAX_UINT16", (void *)_cffi_var_GxB_TIMES_MAX_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MAX_UINT32", (void *)_cffi_var_GxB_TIMES_MAX_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MAX_UINT64", (void *)_cffi_var_GxB_TIMES_MAX_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MAX_UINT8", (void *)_cffi_var_GxB_TIMES_MAX_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MINUS_FC32", (void *)_cffi_var_GxB_TIMES_MINUS_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MINUS_FC64", (void *)_cffi_var_GxB_TIMES_MINUS_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MINUS_FP32", (void *)_cffi_var_GxB_TIMES_MINUS_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MINUS_FP64", (void *)_cffi_var_GxB_TIMES_MINUS_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MINUS_INT16", (void *)_cffi_var_GxB_TIMES_MINUS_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MINUS_INT32", (void *)_cffi_var_GxB_TIMES_MINUS_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MINUS_INT64", (void *)_cffi_var_GxB_TIMES_MINUS_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MINUS_INT8", (void *)_cffi_var_GxB_TIMES_MINUS_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MINUS_UINT16", (void *)_cffi_var_GxB_TIMES_MINUS_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MINUS_UINT32", (void *)_cffi_var_GxB_TIMES_MINUS_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MINUS_UINT64", (void *)_cffi_var_GxB_TIMES_MINUS_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MINUS_UINT8", (void *)_cffi_var_GxB_TIMES_MINUS_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MIN_FP32", (void *)_cffi_var_GxB_TIMES_MIN_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MIN_FP64", (void *)_cffi_var_GxB_TIMES_MIN_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MIN_INT16", (void *)_cffi_var_GxB_TIMES_MIN_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MIN_INT32", (void *)_cffi_var_GxB_TIMES_MIN_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MIN_INT64", (void *)_cffi_var_GxB_TIMES_MIN_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MIN_INT8", (void *)_cffi_var_GxB_TIMES_MIN_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MIN_UINT16", (void *)_cffi_var_GxB_TIMES_MIN_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MIN_UINT32", (void *)_cffi_var_GxB_TIMES_MIN_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MIN_UINT64", (void *)_cffi_var_GxB_TIMES_MIN_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_MIN_UINT8", (void *)_cffi_var_GxB_TIMES_MIN_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PAIR_FC32", (void *)_cffi_var_GxB_TIMES_PAIR_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PAIR_FC64", (void *)_cffi_var_GxB_TIMES_PAIR_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PAIR_FP32", (void *)_cffi_var_GxB_TIMES_PAIR_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PAIR_FP64", (void *)_cffi_var_GxB_TIMES_PAIR_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PAIR_INT16", (void *)_cffi_var_GxB_TIMES_PAIR_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PAIR_INT32", (void *)_cffi_var_GxB_TIMES_PAIR_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PAIR_INT64", (void *)_cffi_var_GxB_TIMES_PAIR_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PAIR_INT8", (void *)_cffi_var_GxB_TIMES_PAIR_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PAIR_UINT16", (void *)_cffi_var_GxB_TIMES_PAIR_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PAIR_UINT32", (void *)_cffi_var_GxB_TIMES_PAIR_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PAIR_UINT64", (void *)_cffi_var_GxB_TIMES_PAIR_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PAIR_UINT8", (void *)_cffi_var_GxB_TIMES_PAIR_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PLUS_FC32", (void *)_cffi_var_GxB_TIMES_PLUS_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PLUS_FC64", (void *)_cffi_var_GxB_TIMES_PLUS_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PLUS_FP32", (void *)_cffi_var_GxB_TIMES_PLUS_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PLUS_FP64", (void *)_cffi_var_GxB_TIMES_PLUS_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PLUS_INT16", (void *)_cffi_var_GxB_TIMES_PLUS_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PLUS_INT32", (void *)_cffi_var_GxB_TIMES_PLUS_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PLUS_INT64", (void *)_cffi_var_GxB_TIMES_PLUS_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PLUS_INT8", (void *)_cffi_var_GxB_TIMES_PLUS_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PLUS_UINT16", (void *)_cffi_var_GxB_TIMES_PLUS_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PLUS_UINT32", (void *)_cffi_var_GxB_TIMES_PLUS_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PLUS_UINT64", (void *)_cffi_var_GxB_TIMES_PLUS_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_PLUS_UINT8", (void *)_cffi_var_GxB_TIMES_PLUS_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RDIV_FC32", (void *)_cffi_var_GxB_TIMES_RDIV_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RDIV_FC64", (void *)_cffi_var_GxB_TIMES_RDIV_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RDIV_FP32", (void *)_cffi_var_GxB_TIMES_RDIV_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RDIV_FP64", (void *)_cffi_var_GxB_TIMES_RDIV_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RDIV_INT16", (void *)_cffi_var_GxB_TIMES_RDIV_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RDIV_INT32", (void *)_cffi_var_GxB_TIMES_RDIV_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RDIV_INT64", (void *)_cffi_var_GxB_TIMES_RDIV_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RDIV_INT8", (void *)_cffi_var_GxB_TIMES_RDIV_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RDIV_UINT16", (void *)_cffi_var_GxB_TIMES_RDIV_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RDIV_UINT32", (void *)_cffi_var_GxB_TIMES_RDIV_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RDIV_UINT64", (void *)_cffi_var_GxB_TIMES_RDIV_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RDIV_UINT8", (void *)_cffi_var_GxB_TIMES_RDIV_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RMINUS_FC32", (void *)_cffi_var_GxB_TIMES_RMINUS_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RMINUS_FC64", (void *)_cffi_var_GxB_TIMES_RMINUS_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RMINUS_FP32", (void *)_cffi_var_GxB_TIMES_RMINUS_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RMINUS_FP64", (void *)_cffi_var_GxB_TIMES_RMINUS_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RMINUS_INT16", (void *)_cffi_var_GxB_TIMES_RMINUS_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RMINUS_INT32", (void *)_cffi_var_GxB_TIMES_RMINUS_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RMINUS_INT64", (void *)_cffi_var_GxB_TIMES_RMINUS_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RMINUS_INT8", (void *)_cffi_var_GxB_TIMES_RMINUS_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RMINUS_UINT16", (void *)_cffi_var_GxB_TIMES_RMINUS_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RMINUS_UINT32", (void *)_cffi_var_GxB_TIMES_RMINUS_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RMINUS_UINT64", (void *)_cffi_var_GxB_TIMES_RMINUS_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_RMINUS_UINT8", (void *)_cffi_var_GxB_TIMES_RMINUS_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_SECONDI1_INT32", (void *)_cffi_var_GxB_TIMES_SECONDI1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_SECONDI1_INT64", (void *)_cffi_var_GxB_TIMES_SECONDI1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_SECONDI_INT32", (void *)_cffi_var_GxB_TIMES_SECONDI_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_SECONDI_INT64", (void *)_cffi_var_GxB_TIMES_SECONDI_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_SECONDJ1_INT32", (void *)_cffi_var_GxB_TIMES_SECONDJ1_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_SECONDJ1_INT64", (void *)_cffi_var_GxB_TIMES_SECONDJ1_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_SECONDJ_INT32", (void *)_cffi_var_GxB_TIMES_SECONDJ_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_SECONDJ_INT64", (void *)_cffi_var_GxB_TIMES_SECONDJ_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_SECOND_FC32", (void *)_cffi_var_GxB_TIMES_SECOND_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_SECOND_FC64", (void *)_cffi_var_GxB_TIMES_SECOND_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_SECOND_FP32", (void *)_cffi_var_GxB_TIMES_SECOND_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_SECOND_FP64", (void *)_cffi_var_GxB_TIMES_SECOND_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_SECOND_INT16", (void *)_cffi_var_GxB_TIMES_SECOND_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_SECOND_INT32", (void *)_cffi_var_GxB_TIMES_SECOND_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_SECOND_INT64", (void *)_cffi_var_GxB_TIMES_SECOND_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_SECOND_INT8", (void *)_cffi_var_GxB_TIMES_SECOND_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_SECOND_UINT16", (void *)_cffi_var_GxB_TIMES_SECOND_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_SECOND_UINT32", (void *)_cffi_var_GxB_TIMES_SECOND_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_SECOND_UINT64", (void *)_cffi_var_GxB_TIMES_SECOND_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_SECOND_UINT8", (void *)_cffi_var_GxB_TIMES_SECOND_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_TIMES_FC32", (void *)_cffi_var_GxB_TIMES_TIMES_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_TIMES_FC64", (void *)_cffi_var_GxB_TIMES_TIMES_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_TIMES_FP32", (void *)_cffi_var_GxB_TIMES_TIMES_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_TIMES_FP64", (void *)_cffi_var_GxB_TIMES_TIMES_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_TIMES_INT16", (void *)_cffi_var_GxB_TIMES_TIMES_INT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_TIMES_INT32", (void *)_cffi_var_GxB_TIMES_TIMES_INT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_TIMES_INT64", (void *)_cffi_var_GxB_TIMES_TIMES_INT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_TIMES_INT8", (void *)_cffi_var_GxB_TIMES_TIMES_INT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_TIMES_UINT16", (void *)_cffi_var_GxB_TIMES_TIMES_UINT16, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_TIMES_UINT32", (void *)_cffi_var_GxB_TIMES_TIMES_UINT32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_TIMES_UINT64", (void *)_cffi_var_GxB_TIMES_TIMES_UINT64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TIMES_TIMES_UINT8", (void *)_cffi_var_GxB_TIMES_TIMES_UINT8, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 159), (void *)0 },
  { "GxB_TRIL", (void *)_cffi_var_GxB_TRIL, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 155), (void *)0 },
  { "GxB_TRIU", (void *)_cffi_var_GxB_TRIU, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 155), (void *)0 },
  { "GxB_TRUNC_FC32", (void *)_cffi_var_GxB_TRUNC_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_TRUNC_FC64", (void *)_cffi_var_GxB_TRUNC_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_TRUNC_FP32", (void *)_cffi_var_GxB_TRUNC_FP32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_TRUNC_FP64", (void *)_cffi_var_GxB_TRUNC_FP64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 118), (void *)0 },
  { "GxB_Type_fprint", (void *)_cffi_f_GxB_Type_fprint, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2321), (void *)_cffi_d_GxB_Type_fprint },
  { "GxB_Type_from_name", (void *)_cffi_f_GxB_Type_from_name, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2274), (void *)_cffi_d_GxB_Type_from_name },
  { "GxB_Type_name", (void *)_cffi_f_GxB_Type_name, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 112), (void *)_cffi_d_GxB_Type_name },
  { "GxB_Type_new", (void *)_cffi_f_GxB_Type_new, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2287), (void *)_cffi_d_GxB_Type_new },
  { "GxB_Type_size", (void *)_cffi_f_GxB_Type_size, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 413), (void *)_cffi_d_GxB_Type_size },
  { "GxB_UnaryOp_fprint", (void *)_cffi_f_GxB_UnaryOp_fprint, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2355), (void *)_cffi_d_GxB_UnaryOp_fprint },
  { "GxB_UnaryOp_new", (void *)_cffi_f_GxB_UnaryOp_new, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2343), (void *)_cffi_d_GxB_UnaryOp_new },
  { "GxB_UnaryOp_xtype", (void *)_cffi_f_GxB_UnaryOp_xtype, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2309), (void *)_cffi_d_GxB_UnaryOp_xtype },
  { "GxB_UnaryOp_xtype_name", (void *)_cffi_f_GxB_UnaryOp_xtype_name, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 116), (void *)_cffi_d_GxB_UnaryOp_xtype_name },
  { "GxB_UnaryOp_ztype", (void *)_cffi_f_GxB_UnaryOp_ztype, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2309), (void *)_cffi_d_GxB_UnaryOp_ztype },
  { "GxB_UnaryOp_ztype_name", (void *)_cffi_f_GxB_UnaryOp_ztype_name, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 116), (void *)_cffi_d_GxB_UnaryOp_ztype_name },
  { "GxB_VALUEEQ_FC32", (void *)_cffi_var_GxB_VALUEEQ_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GxB_VALUEEQ_FC64", (void *)_cffi_var_GxB_VALUEEQ_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GxB_VALUENE_FC32", (void *)_cffi_var_GxB_VALUENE_FC32, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GxB_VALUENE_FC64", (void *)_cffi_var_GxB_VALUENE_FC64, _CFFI_OP(_CFFI_OP_GLOBAL_VAR_F, 102), (void *)0 },
  { "GxB_Vector_Iterator_attach", (void *)_cffi_f_GxB_Vector_Iterator_attach, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 554), (void *)_cffi_d_GxB_Vector_Iterator_attach },
  { "GxB_Vector_Iterator_getIndex", (void *)_cffi_f_GxB_Vector_Iterator_getIndex, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3933), (void *)_cffi_d_GxB_Vector_Iterator_getIndex },
  { "GxB_Vector_Iterator_getp", (void *)_cffi_f_GxB_Vector_Iterator_getp, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3933), (void *)_cffi_d_GxB_Vector_Iterator_getp },
  { "GxB_Vector_Iterator_getpmax", (void *)_cffi_f_GxB_Vector_Iterator_getpmax, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3933), (void *)_cffi_d_GxB_Vector_Iterator_getpmax },
  { "GxB_Vector_Iterator_next", (void *)_cffi_f_GxB_Vector_Iterator_next, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 540), (void *)_cffi_d_GxB_Vector_Iterator_next },
  { "GxB_Vector_Iterator_seek", (void *)_cffi_f_GxB_Vector_Iterator_seek, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 559), (void *)_cffi_d_GxB_Vector_Iterator_seek },
  { "GxB_Vector_Option_get", (void *)_cffi_const_GxB_Vector_Option_get, _CFFI_OP(_CFFI_OP_CONSTANT, 3982), (void *)0 },
  { "GxB_Vector_Option_get_FP64", (void *)_cffi_f_GxB_Vector_Option_get_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2459), (void *)_cffi_d_GxB_Vector_Option_get_FP64 },
  { "GxB_Vector_Option_get_INT32", (void *)_cffi_f_GxB_Vector_Option_get_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2469), (void *)_cffi_d_GxB_Vector_Option_get_INT32 },
  { "GxB_Vector_Option_set", (void *)_cffi_const_GxB_Vector_Option_set, _CFFI_OP(_CFFI_OP_CONSTANT, 3982), (void *)0 },
  { "GxB_Vector_Option_set_FP64", (void *)_cffi_f_GxB_Vector_Option_set_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2464), (void *)_cffi_d_GxB_Vector_Option_set_FP64 },
  { "GxB_Vector_Option_set_INT32", (void *)_cffi_f_GxB_Vector_Option_set_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2474), (void *)_cffi_d_GxB_Vector_Option_set_INT32 },
  { "GxB_Vector_apply_BinaryOp1st", (void *)_cffi_f_GxB_Vector_apply_BinaryOp1st, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2731), (void *)_cffi_d_GxB_Vector_apply_BinaryOp1st },
  { "GxB_Vector_apply_BinaryOp1st_FC32", (void *)_cffi_f_GxB_Vector_apply_BinaryOp1st_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2669), (void *)_cffi_d_GxB_Vector_apply_BinaryOp1st_FC32 },
  { "GxB_Vector_apply_BinaryOp1st_FC64", (void *)_cffi_f_GxB_Vector_apply_BinaryOp1st_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2651), (void *)_cffi_d_GxB_Vector_apply_BinaryOp1st_FC64 },
  { "GxB_Vector_apply_BinaryOp2nd", (void *)_cffi_f_GxB_Vector_apply_BinaryOp2nd, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2821), (void *)_cffi_d_GxB_Vector_apply_BinaryOp2nd },
  { "GxB_Vector_apply_BinaryOp2nd_FC32", (void *)_cffi_f_GxB_Vector_apply_BinaryOp2nd_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2767), (void *)_cffi_d_GxB_Vector_apply_BinaryOp2nd_FC32 },
  { "GxB_Vector_apply_BinaryOp2nd_FC64", (void *)_cffi_f_GxB_Vector_apply_BinaryOp2nd_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2749), (void *)_cffi_d_GxB_Vector_apply_BinaryOp2nd_FC64 },
  { "GxB_Vector_apply_IndexOp_FC32", (void *)_cffi_f_GxB_Vector_apply_IndexOp_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2967), (void *)_cffi_d_GxB_Vector_apply_IndexOp_FC32 },
  { "GxB_Vector_apply_IndexOp_FC64", (void *)_cffi_f_GxB_Vector_apply_IndexOp_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2949), (void *)_cffi_d_GxB_Vector_apply_IndexOp_FC64 },
  { "GxB_Vector_assign_FC32", (void *)_cffi_f_GxB_Vector_assign_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2588), (void *)_cffi_d_GxB_Vector_assign_FC32 },
  { "GxB_Vector_assign_FC64", (void *)_cffi_f_GxB_Vector_assign_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2570), (void *)_cffi_d_GxB_Vector_assign_FC64 },
  { "GxB_Vector_build_FC32", (void *)_cffi_f_GxB_Vector_build_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3269), (void *)_cffi_d_GxB_Vector_build_FC32 },
  { "GxB_Vector_build_FC64", (void *)_cffi_f_GxB_Vector_build_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3255), (void *)_cffi_d_GxB_Vector_build_FC64 },
  { "GxB_Vector_build_Scalar", (void *)_cffi_f_GxB_Vector_build_Scalar, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3311), (void *)_cffi_d_GxB_Vector_build_Scalar },
  { "GxB_Vector_deserialize", (void *)_cffi_f_GxB_Vector_deserialize, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2437), (void *)_cffi_d_GxB_Vector_deserialize },
  { "GxB_Vector_diag", (void *)_cffi_f_GxB_Vector_diag, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2550), (void *)_cffi_d_GxB_Vector_diag },
  { "GxB_Vector_eWiseUnion", (void *)_cffi_f_GxB_Vector_eWiseUnion, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2830), (void *)_cffi_d_GxB_Vector_eWiseUnion },
  { "GxB_Vector_export_Bitmap", (void *)_cffi_f_GxB_Vector_export_Bitmap, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2364), (void *)_cffi_d_GxB_Vector_export_Bitmap },
  { "GxB_Vector_export_CSC", (void *)_cffi_f_GxB_Vector_export_CSC, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2376), (void *)_cffi_d_GxB_Vector_export_CSC },
  { "GxB_Vector_export_Full", (void *)_cffi_f_GxB_Vector_export_Full, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2389), (void *)_cffi_d_GxB_Vector_export_Full },
  { "GxB_Vector_extractElement_FC32", (void *)_cffi_f_GxB_Vector_extractElement_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 284), (void *)_cffi_d_GxB_Vector_extractElement_FC32 },
  { "GxB_Vector_extractElement_FC64", (void *)_cffi_f_GxB_Vector_extractElement_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 226), (void *)_cffi_d_GxB_Vector_extractElement_FC64 },
  { "GxB_Vector_extractTuples_FC32", (void *)_cffi_f_GxB_Vector_extractTuples_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3467), (void *)_cffi_d_GxB_Vector_extractTuples_FC32 },
  { "GxB_Vector_extractTuples_FC64", (void *)_cffi_f_GxB_Vector_extractTuples_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3455), (void *)_cffi_d_GxB_Vector_extractTuples_FC64 },
  { "GxB_Vector_fprint", (void *)_cffi_f_GxB_Vector_fprint, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2484), (void *)_cffi_d_GxB_Vector_fprint },
  { "GxB_Vector_import_Bitmap", (void *)_cffi_f_GxB_Vector_import_Bitmap, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2403), (void *)_cffi_d_GxB_Vector_import_Bitmap },
  { "GxB_Vector_import_CSC", (void *)_cffi_f_GxB_Vector_import_CSC, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2415), (void *)_cffi_d_GxB_Vector_import_CSC },
  { "GxB_Vector_import_Full", (void *)_cffi_f_GxB_Vector_import_Full, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2428), (void *)_cffi_d_GxB_Vector_import_Full },
  { "GxB_Vector_isStoredElement", (void *)_cffi_f_GxB_Vector_isStoredElement, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3352), (void *)_cffi_d_GxB_Vector_isStoredElement },
  { "GxB_Vector_iso", (void *)_cffi_f_GxB_Vector_iso, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 82), (void *)_cffi_d_GxB_Vector_iso },
  { "GxB_Vector_memoryUsage", (void *)_cffi_f_GxB_Vector_memoryUsage, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 417), (void *)_cffi_d_GxB_Vector_memoryUsage },
  { "GxB_Vector_pack_Bitmap", (void *)_cffi_f_GxB_Vector_pack_Bitmap, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2535), (void *)_cffi_d_GxB_Vector_pack_Bitmap },
  { "GxB_Vector_pack_CSC", (void *)_cffi_f_GxB_Vector_pack_CSC, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3237), (void *)_cffi_d_GxB_Vector_pack_CSC },
  { "GxB_Vector_pack_Full", (void *)_cffi_f_GxB_Vector_pack_Full, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3373), (void *)_cffi_d_GxB_Vector_pack_Full },
  { "GxB_Vector_reduce_FC32", (void *)_cffi_f_GxB_Vector_reduce_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 267), (void *)_cffi_d_GxB_Vector_reduce_FC32 },
  { "GxB_Vector_reduce_FC64", (void *)_cffi_f_GxB_Vector_reduce_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 209), (void *)_cffi_d_GxB_Vector_reduce_FC64 },
  { "GxB_Vector_select", (void *)_cffi_f_GxB_Vector_select, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3111), (void *)_cffi_d_GxB_Vector_select },
  { "GxB_Vector_select_FC32", (void *)_cffi_f_GxB_Vector_select_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2967), (void *)_cffi_d_GxB_Vector_select_FC32 },
  { "GxB_Vector_select_FC64", (void *)_cffi_f_GxB_Vector_select_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2949), (void *)_cffi_d_GxB_Vector_select_FC64 },
  { "GxB_Vector_serialize", (void *)_cffi_f_GxB_Vector_serialize, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3848), (void *)_cffi_d_GxB_Vector_serialize },
  { "GxB_Vector_setElement_FC32", (void *)_cffi_f_GxB_Vector_setElement_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2500), (void *)_cffi_d_GxB_Vector_setElement_FC32 },
  { "GxB_Vector_setElement_FC64", (void *)_cffi_f_GxB_Vector_setElement_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2490), (void *)_cffi_d_GxB_Vector_setElement_FC64 },
  { "GxB_Vector_sort", (void *)_cffi_f_GxB_Vector_sort, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3155), (void *)_cffi_d_GxB_Vector_sort },
  { "GxB_Vector_subassign", (void *)_cffi_f_GxB_Vector_subassign, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3162), (void *)_cffi_d_GxB_Vector_subassign },
  { "GxB_Vector_subassign_BOOL", (void *)_cffi_f_GxB_Vector_subassign_BOOL, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2561), (void *)_cffi_d_GxB_Vector_subassign_BOOL },
  { "GxB_Vector_subassign_FC32", (void *)_cffi_f_GxB_Vector_subassign_FC32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2588), (void *)_cffi_d_GxB_Vector_subassign_FC32 },
  { "GxB_Vector_subassign_FC64", (void *)_cffi_f_GxB_Vector_subassign_FC64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2570), (void *)_cffi_d_GxB_Vector_subassign_FC64 },
  { "GxB_Vector_subassign_FP32", (void *)_cffi_f_GxB_Vector_subassign_FP32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2597), (void *)_cffi_d_GxB_Vector_subassign_FP32 },
  { "GxB_Vector_subassign_FP64", (void *)_cffi_f_GxB_Vector_subassign_FP64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2579), (void *)_cffi_d_GxB_Vector_subassign_FP64 },
  { "GxB_Vector_subassign_INT16", (void *)_cffi_f_GxB_Vector_subassign_INT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2606), (void *)_cffi_d_GxB_Vector_subassign_INT16 },
  { "GxB_Vector_subassign_INT32", (void *)_cffi_f_GxB_Vector_subassign_INT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2615), (void *)_cffi_d_GxB_Vector_subassign_INT32 },
  { "GxB_Vector_subassign_INT64", (void *)_cffi_f_GxB_Vector_subassign_INT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2624), (void *)_cffi_d_GxB_Vector_subassign_INT64 },
  { "GxB_Vector_subassign_INT8", (void *)_cffi_f_GxB_Vector_subassign_INT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2633), (void *)_cffi_d_GxB_Vector_subassign_INT8 },
  { "GxB_Vector_subassign_Scalar", (void *)_cffi_f_GxB_Vector_subassign_Scalar, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3102), (void *)_cffi_d_GxB_Vector_subassign_Scalar },
  { "GxB_Vector_subassign_UDT", (void *)_cffi_f_GxB_Vector_subassign_UDT, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3207), (void *)_cffi_d_GxB_Vector_subassign_UDT },
  { "GxB_Vector_subassign_UINT16", (void *)_cffi_f_GxB_Vector_subassign_UINT16, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3171), (void *)_cffi_d_GxB_Vector_subassign_UINT16 },
  { "GxB_Vector_subassign_UINT32", (void *)_cffi_f_GxB_Vector_subassign_UINT32, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3180), (void *)_cffi_d_GxB_Vector_subassign_UINT32 },
  { "GxB_Vector_subassign_UINT64", (void *)_cffi_f_GxB_Vector_subassign_UINT64, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3189), (void *)_cffi_d_GxB_Vector_subassign_UINT64 },
  { "GxB_Vector_subassign_UINT8", (void *)_cffi_f_GxB_Vector_subassign_UINT8, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3198), (void *)_cffi_d_GxB_Vector_subassign_UINT8 },
  { "GxB_Vector_type", (void *)_cffi_f_GxB_Vector_type, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2313), (void *)_cffi_d_GxB_Vector_type },
  { "GxB_Vector_type_name", (void *)_cffi_f_GxB_Vector_type_name, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 120), (void *)_cffi_d_GxB_Vector_type_name },
  { "GxB_Vector_unpack_Bitmap", (void *)_cffi_f_GxB_Vector_unpack_Bitmap, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 2525), (void *)_cffi_d_GxB_Vector_unpack_Bitmap },
  { "GxB_Vector_unpack_CSC", (void *)_cffi_f_GxB_Vector_unpack_CSC, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3226), (void *)_cffi_d_GxB_Vector_unpack_CSC },
  { "GxB_Vector_unpack_Full", (void *)_cffi_f_GxB_Vector_unpack_Full, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 3366), (void *)_cffi_d_GxB_Vector_unpack_Full },
  { "GxB_colIterator_attach", (void *)_cffi_f_GxB_colIterator_attach, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 549), (void *)_cffi_d_GxB_colIterator_attach },
  { "GxB_colIterator_getColIndex", (void *)_cffi_f_GxB_colIterator_getColIndex, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3933), (void *)_cffi_d_GxB_colIterator_getColIndex },
  { "GxB_colIterator_getRowIndex", (void *)_cffi_f_GxB_colIterator_getRowIndex, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3933), (void *)_cffi_d_GxB_colIterator_getRowIndex },
  { "GxB_colIterator_kount", (void *)_cffi_f_GxB_colIterator_kount, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3933), (void *)_cffi_d_GxB_colIterator_kount },
  { "GxB_colIterator_kseek", (void *)_cffi_f_GxB_colIterator_kseek, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 559), (void *)_cffi_d_GxB_colIterator_kseek },
  { "GxB_colIterator_nextCol", (void *)_cffi_f_GxB_colIterator_nextCol, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 540), (void *)_cffi_d_GxB_colIterator_nextCol },
  { "GxB_colIterator_nextRow", (void *)_cffi_f_GxB_colIterator_nextRow, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 540), (void *)_cffi_d_GxB_colIterator_nextRow },
  { "GxB_colIterator_seekCol", (void *)_cffi_f_GxB_colIterator_seekCol, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 559), (void *)_cffi_d_GxB_colIterator_seekCol },
  { "GxB_deserialize_type_name", (void *)_cffi_f_GxB_deserialize_type_name, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 124), (void *)_cffi_d_GxB_deserialize_type_name },
  { "GxB_init", (void *)_cffi_f_GxB_init, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 12), (void *)_cffi_d_GxB_init },
  { "GxB_pack_HyperHash", (void *)_cffi_f_GxB_pack_HyperHash, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1028), (void *)_cffi_d_GxB_pack_HyperHash },
  { "GxB_rowIterator_attach", (void *)_cffi_f_GxB_rowIterator_attach, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 549), (void *)_cffi_d_GxB_rowIterator_attach },
  { "GxB_rowIterator_getColIndex", (void *)_cffi_f_GxB_rowIterator_getColIndex, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3933), (void *)_cffi_d_GxB_rowIterator_getColIndex },
  { "GxB_rowIterator_getRowIndex", (void *)_cffi_f_GxB_rowIterator_getRowIndex, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3933), (void *)_cffi_d_GxB_rowIterator_getRowIndex },
  { "GxB_rowIterator_kount", (void *)_cffi_f_GxB_rowIterator_kount, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 3933), (void *)_cffi_d_GxB_rowIterator_kount },
  { "GxB_rowIterator_kseek", (void *)_cffi_f_GxB_rowIterator_kseek, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 559), (void *)_cffi_d_GxB_rowIterator_kseek },
  { "GxB_rowIterator_nextCol", (void *)_cffi_f_GxB_rowIterator_nextCol, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 540), (void *)_cffi_d_GxB_rowIterator_nextCol },
  { "GxB_rowIterator_nextRow", (void *)_cffi_f_GxB_rowIterator_nextRow, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_O, 540), (void *)_cffi_d_GxB_rowIterator_nextRow },
  { "GxB_rowIterator_seekRow", (void *)_cffi_f_GxB_rowIterator_seekRow, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 559), (void *)_cffi_d_GxB_rowIterator_seekRow },
  { "GxB_unpack_HyperHash", (void *)_cffi_f_GxB_unpack_HyperHash, _CFFI_OP(_CFFI_OP_CPYTHON_BLTN_V, 1028), (void *)_cffi_d_GxB_unpack_HyperHash },
};

static const struct _cffi_struct_union_s _cffi_struct_unions[] = {
  { "GB_BinaryOp_opaque", 3985, _CFFI_F_OPAQUE,
    (size_t)-1, -1, -1, 0 /* opaque */ },
  { "GB_Descriptor_opaque", 3986, _CFFI_F_OPAQUE,
    (size_t)-1, -1, -1, 0 /* opaque */ },
  { "GB_IndexUnaryOp_opaque", 3987, _CFFI_F_OPAQUE,
    (size_t)-1, -1, -1, 0 /* opaque */ },
  { "GB_Iterator_opaque", 3988, _CFFI_F_OPAQUE,
    (size_t)-1, -1, -1, 0 /* opaque */ },
  { "GB_Matrix_opaque", 3989, _CFFI_F_OPAQUE,
    (size_t)-1, -1, -1, 0 /* opaque */ },
  { "GB_Monoid_opaque", 3990, _CFFI_F_OPAQUE,
    (size_t)-1, -1, -1, 0 /* opaque */ },
  { "GB_Scalar_opaque", 3991, _CFFI_F_OPAQUE,
    (size_t)-1, -1, -1, 0 /* opaque */ },
  { "GB_SelectOp_opaque", 3992, _CFFI_F_OPAQUE,
    (size_t)-1, -1, -1, 0 /* opaque */ },
  { "GB_Semiring_opaque", 3993, _CFFI_F_OPAQUE,
    (size_t)-1, -1, -1, 0 /* opaque */ },
  { "GB_Type_opaque", 3994, _CFFI_F_OPAQUE,
    (size_t)-1, -1, -1, 0 /* opaque */ },
  { "GB_UnaryOp_opaque", 3995, _CFFI_F_OPAQUE,
    (size_t)-1, -1, -1, 0 /* opaque */ },
  { "GB_Vector_opaque", 3996, _CFFI_F_OPAQUE,
    (size_t)-1, -1, -1, 0 /* opaque */ },
  { "_IO_FILE", 3978, _CFFI_F_OPAQUE,
    (size_t)-1, -1, -1, 0 /* opaque */ },
};

static const struct _cffi_enum_s _cffi_enums[] = {
  { "$GrB_Desc_Field", 3, _cffi_prim_int(sizeof(GrB_Desc_Field), ((GrB_Desc_Field)-1) <= 0),
    "GrB_OUTP,GrB_MASK,GrB_INP0,GrB_INP1,GxB_DESCRIPTOR_NTHREADS,GxB_DESCRIPTOR_CHUNK,GxB_DESCRIPTOR_GPU_CONTROL,GxB_DESCRIPTOR_GPU_CHUNK,GxB_AxB_METHOD,GxB_SORT,GxB_COMPRESSION,GxB_IMPORT" },
  { "$GrB_Desc_Value", 476, _cffi_prim_int(sizeof(GrB_Desc_Value), ((GrB_Desc_Value)-1) <= 0),
    "GxB_DEFAULT,GrB_REPLACE,GrB_COMP,GrB_STRUCTURE,GrB_TRAN,GxB_GPU_ALWAYS,GxB_GPU_NEVER,GxB_AxB_GUSTAVSON,GxB_AxB_DOT,GxB_AxB_HASH,GxB_AxB_SAXPY,GxB_SECURE_IMPORT" },
  { "$GrB_Format", 702, _cffi_prim_int(sizeof(GrB_Format), ((GrB_Format)-1) <= 0),
    "GrB_CSR_FORMAT,GrB_CSC_FORMAT,GrB_COO_FORMAT" },
  { "$GrB_Info", 3983, _cffi_prim_int(sizeof(GrB_Info), ((GrB_Info)-1) <= 0),
    "GrB_SUCCESS,GrB_NO_VALUE,GxB_EXHAUSTED,GrB_UNINITIALIZED_OBJECT,GrB_NULL_POINTER,GrB_INVALID_VALUE,GrB_INVALID_INDEX,GrB_DOMAIN_MISMATCH,GrB_DIMENSION_MISMATCH,GrB_OUTPUT_NOT_EMPTY,GrB_NOT_IMPLEMENTED,GrB_PANIC,GrB_OUT_OF_MEMORY,GrB_INSUFFICIENT_SPACE,GrB_INVALID_OBJECT,GrB_INDEX_OUT_OF_BOUNDS,GrB_EMPTY_OBJECT" },
  { "$GrB_Mode", 10, _cffi_prim_int(sizeof(GrB_Mode), ((GrB_Mode)-1) <= 0),
    "GrB_NONBLOCKING,GrB_BLOCKING,GxB_NONBLOCKING_GPU,GxB_BLOCKING_GPU" },
  { "$GrB_WaitMode", 458, _cffi_prim_int(sizeof(GrB_WaitMode), ((GrB_WaitMode)-1) <= 0),
    "GrB_COMPLETE,GrB_MATERIALIZE" },
  { "$GxB_Format_Value", 546, _cffi_prim_int(sizeof(GxB_Format_Value), ((GxB_Format_Value)-1) <= 0),
    "GxB_BY_ROW,GxB_BY_COL,GxB_NO_FORMAT" },
  { "$GxB_Option_Field", 20, _cffi_prim_int(sizeof(GxB_Option_Field), ((GxB_Option_Field)-1) <= 0),
    "GxB_HYPER_SWITCH,GxB_BITMAP_SWITCH,GxB_FORMAT,GxB_MODE,GxB_LIBRARY_NAME,GxB_LIBRARY_VERSION,GxB_LIBRARY_DATE,GxB_LIBRARY_ABOUT,GxB_LIBRARY_URL,GxB_LIBRARY_LICENSE,GxB_LIBRARY_COMPILE_DATE,GxB_LIBRARY_COMPILE_TIME,GxB_API_VERSION,GxB_API_DATE,GxB_API_ABOUT,GxB_API_URL,GxB_COMPILER_VERSION,GxB_COMPILER_NAME,GxB_LIBRARY_OPENMP,GxB_GLOBAL_NTHREADS,GxB_GLOBAL_CHUNK,GxB_BURBLE,GxB_PRINTF,GxB_FLUSH,GxB_MEMORY_POOL,GxB_PRINT_1BASED,GxB_SPARSITY_STATUS,GxB_SPARSITY_CONTROL,GxB_GLOBAL_GPU_CONTROL,GxB_GLOBAL_GPU_CHUNK" },
  { "$GxB_Print_Level", 463, _cffi_prim_int(sizeof(GxB_Print_Level), ((GxB_Print_Level)-1) <= 0),
    "GxB_SILENT,GxB_SUMMARY,GxB_SHORT,GxB_COMPLETE,GxB_SHORT_VERBOSE,GxB_COMPLETE_VERBOSE" },
};

static const struct _cffi_typename_s _cffi_typenames[] = {
  { "FILE", 3978 },
  { "GrB_BinaryOp", 56 },
  { "GrB_Desc_Field", 3 },
  { "GrB_Desc_Value", 476 },
  { "GrB_Descriptor", 2 },
  { "GrB_Format", 702 },
  { "GrB_Index", 75 },
  { "GrB_IndexUnaryOp", 102 },
  { "GrB_Info", 3983 },
  { "GrB_Matrix", 7 },
  { "GrB_Mode", 10 },
  { "GrB_Monoid", 57 },
  { "GrB_Scalar", 80 },
  { "GrB_Semiring", 159 },
  { "GrB_Type", 114 },
  { "GrB_UnaryOp", 118 },
  { "GrB_Vector", 65 },
  { "GrB_WaitMode", 458 },
  { "GxB_FC32_t", 974 },
  { "GxB_FC64_t", 962 },
  { "GxB_Format_Value", 546 },
  { "GxB_Iterator", 541 },
  { "GxB_Option_Field", 20 },
  { "GxB_Print_Level", 463 },
  { "GxB_Scalar", 80 },
  { "GxB_SelectOp", 155 },
  { "GxB_binary_function", 434 },
  { "GxB_index_unary_function", 513 },
  { "GxB_select_function", 2232 },
  { "GxB_unary_function", 2332 },
};

static const struct _cffi_type_context_s _cffi_type_context = {
  _cffi_types,
  _cffi_globals,
  NULL,  /* no fields */
  _cffi_struct_unions,
  _cffi_enums,
  _cffi_typenames,
  3341,  /* num_globals */
  13,  /* num_struct_unions */
  9,  /* num_enums */
  30,  /* num_typenames */
  NULL,  /* no includes */
  3999,  /* num_types */
  0,  /* flags */
};

#ifdef __GNUC__
#  pragma GCC visibility push(default)  /* for -fvisibility= */
#endif

#ifdef PYPY_VERSION
PyMODINIT_FUNC
_cffi_pypyinit__graphblas(const void *p[])
{
    p[0] = (const void *)0x2601;
    p[1] = &_cffi_type_context;
#if PY_MAJOR_VERSION >= 3
    return NULL;
#endif
}
#  ifdef _MSC_VER
     PyMODINIT_FUNC
#  if PY_MAJOR_VERSION >= 3
     PyInit__graphblas(void) { return NULL; }
#  else
     init_graphblas(void) { }
#  endif
#  endif
#elif PY_MAJOR_VERSION >= 3
PyMODINIT_FUNC
PyInit__graphblas(void)
{
  return _cffi_init("suitesparse_graphblas._graphblas", 0x2601, &_cffi_type_context);
}
#else
PyMODINIT_FUNC
init_graphblas(void)
{
  _cffi_init("suitesparse_graphblas._graphblas", 0x2601, &_cffi_type_context);
}
#endif

#ifdef __GNUC__
#  pragma GCC visibility pop
#endif
